package com.boylesoftware.cb2;

import java.sql.Statement;
import java.sql.SQLWarning;
import java.sql.SQLException;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.util.sql.ConnectionWrapper;
import com.boylesoftware.cb2.util.pool.AdaptiveObjectPool;


/**
 * Application context wraps database connections that it returns into this
 * class, which provides internal service for statistics and additional
 * control.
 *
 * <p>All JDBC operations are logged on debug level using
 * <code>com.boylesoftware.cb2.ACTSQL</code> channel.
 *
 * @author Lev Himmelfarb
 * @version $Id: ActiveConnection.java,v 1.4 2004/04/02 21:30:46 levahim Exp $
 */
final class ActiveConnection
    extends ConnectionWrapper {

    /**
     * Subsystem name for logging JDBC operations.
     */
    private static final String LOG_SUBSYS = "com.boylesoftware.cb2.ACTSQL";


    /**
     * Id for the next active statement.
     */
    private static long nextActiveStatementId = 1L;


    /**
     * Reference to the pool this object was borrowed from.
     */
    private final AdaptiveObjectPool pool;

    /**
     * Reference to out application context.
     */
    private final ApplicationContext appCtx;

    /**
     * Reference to the pool of <code>ActiveStatement</code> objects
     * located in the <code>ActiveConnectionFactory</code> instance.
     */
    private final AdaptiveObjectPool stmtPool;

    /**
     * Debug log.
     */
    final Log log;


    /**
     * Creates an instance.
     *
     * @param pool reference to the pool, from which this object was borrowed.
     * @param appCtx reference to our application context.
     * @param stmtPool reference to the <code>ActiveStatement</code>
     * objects pool located in the <code>ActiveConnectionFactory</code>
     * instance.
     */
    ActiveConnection(AdaptiveObjectPool pool,
                     ApplicationContext appCtx,
                     AdaptiveObjectPool stmtPool) {

        this.pool = pool;
        this.appCtx = appCtx;
        this.stmtPool = stmtPool;
        this.log = appCtx.getLog(LOG_SUBSYS);
    }


    /**
     * Logs the call and calls the underlying connection method.
     */
    public void setAutoCommit(boolean autoCommit)
        throws SQLException {

        if(this.log.isDebugEnabled())
            this.log.debug("con #" + this.getRealConnectionId() +
                           ": setting auto-commit mode to " + autoCommit);

        this.con.setAutoCommit(autoCommit);
    }

    /**
     * Logs the call and calls the underlying connection method.
     */
    public void commit()
        throws SQLException {

        if(this.log.isDebugEnabled())
            this.log.debug("con #" + this.getRealConnectionId() +
                           ": committing transaction");

        this.con.commit();
    }

    /**
     * Logs the call and calls the underlying connection method.
     */
    public void rollback()
        throws SQLException {

        if(this.log.isDebugEnabled())
            this.log.debug("con #" + this.getRealConnectionId() +
                           ": rolling transaction back");

        this.con.rollback();
    }

    /**
     * Logs the call and calls the underlying connection method.
     */
    public void setReadOnly(boolean readOnly)
        throws SQLException {

        if(this.log.isDebugEnabled())
            this.log.debug("con #" + this.getRealConnectionId() +
                           ": setting read-only mode to " + readOnly);

        this.con.setReadOnly(readOnly);
    }

    /**
     * Logs the call and calls the underlying connection method.
     */
    public void setCatalog(String catalog)
        throws SQLException {

        if(this.log.isDebugEnabled())
            this.log.debug("con #" + this.getRealConnectionId() +
                           ": setting current catalog to [" + catalog + "]");

        this.con.setCatalog(catalog);
    }

    /**
     * Just calls the underlying connection. We override it so the
     * <code>postCallHook</code> is not called, which is otherwise
     * the default behaviour of <code>ConnectionWrapper</code>.
     */
    public SQLWarning getWarnings()
        throws SQLException {

        return this.con.getWarnings();
    }

    /**
     * Just calls the underlying connection. We override it so the
     * <code>postCallHook</code> is not called, which is otherwise
     * the default behaviour of <code>ConnectionWrapper</code>.
     */
    public void clearWarnings()
        throws SQLException {

        this.con.clearWarnings();
    }

    /**
     * Closes the wrapped connection and returns this instance of the
     * wrapper back to the pool. Don't use the instance after this
     * method call because it can be reassigned to another borrower.
     *
     * @throws SQLException if the wrapped connection could not be closed.
     * The instance is returned back to the pool and becomes unusable anyway.
     */
    public void close()
        throws SQLException {

        try {
            if(this.con != null)
                this.con.close();
        } finally {
            this.con = null;
            this.pool.returnObject(this);
        }
    }

    /**
     * Wraps the statement into an <code>ActiveStatement</code> instance
     * and returns it.
     *
     * @param stmt statement returned by the underlying connection.
     * @param sql the SQL text or <code>null</code>.
     *
     * @return statement wrapped into an <code>ActiveStatement</code>
     * instance.
     */
    protected Statement createStatementHook(Statement stmt, String sql) {

        ActiveStatement astmt = (ActiveStatement)this.stmtPool.get();
        Long id;
        synchronized(this.getClass()) {
            id = new Long(nextActiveStatementId++);
        }
        astmt.setup(this, stmt, id, sql);
        this.appCtx.registerStatement(id, astmt);

        return astmt;
    }

    /**
     * Checks for SQL warnings and logs them if there are any.
     *
     * @throws SQLException if underlying <code>getWarnings</code>
     * throws an exception.
     */
    protected void postCallHook()
        throws SQLException {

        boolean clear = false;
        for(SQLWarning warn = this.con.getWarnings(); warn != null; warn = warn.getNextWarning()) {
            clear = true;
            this.log.warn("con #" + this.getRealConnectionId() +
                           ": error code " + warn.getErrorCode() +
                           ", sql state " + warn.getSQLState() +
                           ", " + warn.getMessage());
        }
        if(clear)
            this.con.clearWarnings();
    }

    /**
     * Finds the underlying connection object descending through all
     * <code>ConnectionWrapper</code> wrappers and gets its hash code.
     * This method is used to identify connections returned from the
     * underlying connection pool, which is useful in logging to determine
     * if some events happen on the same connection.
     *
     * @return hash code of the connection stripped from all known to CB2
     * wrappers.
     */
    int getRealConnectionId() {

        java.sql.Connection c = this.con;
        while(c instanceof ConnectionWrapper)
            c = ((ConnectionWrapper)c).getWrappedConnection();

        return c.hashCode();
    }

    /**
     * Returns name of the specified transaction isolation level. Used
     * in the <code>ApplicationContext</code>'s <code>getStats</code> method.
     *
     * @param level one of <code>Connection.TRANSACTION_XXX</code>
     * constants.
     *
     * @return human readable name of the specified level.
     */
    static final String getTxIsoLevelName(int level) {

        switch(level) {
        case TRANSACTION_NONE:
            return "NONE";
        case TRANSACTION_READ_COMMITTED:
            return "READ COMMITTED";
        case TRANSACTION_READ_UNCOMMITTED:
            return "READ UNCOMMITTED";
        case TRANSACTION_REPEATABLE_READ:
            return "REPEATABLE READ";
        case TRANSACTION_SERIALIZABLE:
            return "SERIALIZABLE";
        default:
            return "UNKNOWN";
        }
    }
}
