package com.boylesoftware.cb2;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.CallableStatement;
import java.sql.ResultSet;
import java.sql.SQLWarning;
import java.sql.SQLException;

import com.boylesoftware.cb2.util.TextUtils;
import com.boylesoftware.cb2.util.sql.CallableStatementWrapper;
import com.boylesoftware.cb2.util.pool.AdaptiveObjectPool;


/**
 * Wrapper around database statement providing statistics
 * and additional control. <code>ActiveConnection</code> wraps
 * all statements that it creates into this class.
 *
 * @author Lev Himmelfarb
 * @version $Id: ActiveStatement.java,v 1.7 2004/10/11 18:17:36 levahim Exp $
 */
final class ActiveStatement
    extends CallableStatementWrapper
    implements StatusProvider {

    // statement states
    private static final int STATE_CREATING          = -1;
    private static final int STATE_CREATED           = 0;
    private static final int STATE_EXECUTING         = 1;
    private static final int STATE_PROCESSING_RESULT = 2;
    private static final int STATE_CANCELLED         = 3;
    private static final int STATE_CLOSED            = 4;


    /**
     * Lock object for the statement's state.
     */
    private final Object stateLock = new Object();

    /**
     * Reference to the pool, from which this object was borrowed.
     */
    private final AdaptiveObjectPool pool;

    /**
     * Reference to our application context.
     */
    private final ApplicationContext appCtx;

    /**
     * Id of the registered statement.
     */
    private Long id = null;

    /**
     * Name of the thread, by which this statement was created.
     */
    private String createdByThreadName;

    /**
     * Current state, one of <code>STATE_xxx</code> constants.
     */
    private int state = STATE_CREATING; // this state is actually never shown

    /**
     * Timestamp of the last state change.
     */
    private long lastStateChangeTs = System.currentTimeMillis();

    /**
     * SQL text of the query.
     */
    private String sql = null;

    /**
     * Query parameters.
     */
    private Object [] params = null;

    /**
     * Maximum index of a set param in the <code>params</code>.
     */
    private int maxParamIndex = -1;


    /**
     * Creates an instance.
     *
     * @param pool reference to the pool, to which instances should be returned
     * when closed. It is the pool in <code>ActiveConnectionFactory</code>.
     * @param appCtx reference to our application context.
     */
    ActiveStatement(AdaptiveObjectPool pool,
                    ApplicationContext appCtx) {

        this.pool = pool;
        this.appCtx = appCtx;
    }


    /**
     * Setup the object after it is borrowed from the pool but not yet
     * returned to the client, nor registered.
     *
     * @param con the parent active connection used to create this
     * active statement.
     * @param stmt the statement to wrap, can be any
     * <code>java.sql.Statement</code>, <code>java.sql.PreparedStatement</code>
     * or <code>java.sql.CallableStatement</code> instance.
     * @param id registered active statement id.
     * @param sql SQL text of the query or <code>null</code>.
     */
    void setup(ActiveConnection con,
               Statement stmt,
               Long id,
               String sql) {

        this.con = con;
        if(stmt instanceof CallableStatement)
            this.stmt = this.pstmt = this.cstmt = (CallableStatement)stmt;
        else if(stmt instanceof PreparedStatement)
            this.stmt = this.pstmt = (PreparedStatement)stmt;
        else
            this.stmt = stmt;
        this.id = id;
        this.sql = sql;
        if(sql != null) {
            int nQs = 0;
            int i = -1;
            while((i = sql.indexOf('?', i + 1)) >= 0)
                nQs++;

            this.params = new Object[nQs];
            this.maxParamIndex = -1;
        }
        this.state = STATE_CREATED;
        this.lastStateChangeTs = System.currentTimeMillis();
        this.createdByThreadName = Thread.currentThread().getName();
    }


    ///////////////////////////////////////////////////////////////////////////
    // Called while the statement is used by the client.
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Remembers query parameter in the <code>params</code> array and calls
     * the underlying statement.
     *
     * @param index parameter index.
     * @param param parameter value.
     *
     * @return the <code>param</code>.
     */
    protected Object setParameterHook(int index, Object param) {

        if(this.params != null) {
            synchronized(this.stateLock) {
                this.params[index - 1] = param;
                if(this.maxParamIndex < index)
                    this.maxParamIndex = index;
            }
        }

        return param;
    }

    /**
     * Clears the <code>params</code> array and calls the underlying statement.
     *
     * @throws SQLException if the underlying statement throws it.
     */
    public void clearParameters()
        throws SQLException {

        if(this.params != null) {
            synchronized(this.stateLock) {
                for(int i = 0; i < this.maxParamIndex; i++)
                    this.params[i] = null;
                this.maxParamIndex = -1;
            }
        }

        this.pstmt.clearParameters();
    }

    /**
     * Logs the call and calls the underlying connection method.
     */
    public void addBatch(String sql)
        throws SQLException {

        if(((ActiveConnection)this.con).log.isDebugEnabled()) {
            ((ActiveConnection)this.con).log.
                debug("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                      ": adding sql to batch [" + sql + "]");
        }

        this.stmt.addBatch(sql);
    }

    /**
     * Logs the call and calls the underlying connection method.
     */
    public void addBatch()
        throws SQLException {

        if(((ActiveConnection)this.con).log.isDebugEnabled()) {
            ((ActiveConnection)this.con).log.
                debug("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                      ": adding sql to batch [" + this.sql + "]" +
                      (this.params != null ?
                          "\nparams are " + TextUtils.arrayToString(this.params, 0, this.maxParamIndex) :
                          ""));
        }

        this.pstmt.addBatch();
    }

    /**
     * Logs the call and calls the underlying connection method.
     */
    public void clearBatch()
        throws SQLException {

        if(((ActiveConnection)this.con).log.isDebugEnabled()) {
            ((ActiveConnection)this.con).log.
                debug("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                      ": clearing the batch");
        }

        this.stmt.clearBatch();
    }

    /**
     * Calls <code>cancel</code> on the underlying statement and updates
     * the active statement's state if successful.
     */
    public void cancel()
        throws SQLException {

        this.stmt.cancel();

        synchronized(this.stateLock) {
            this.state = STATE_CANCELLED;
            this.lastStateChangeTs = System.currentTimeMillis();
        }
    }

    /**
     * Just calls the underlying statement. We override it so the
     * <code>postCallHook</code> is not called, which is otherwise the
     * default behaviour of <code>StatementWrapper</code>.
     */
    public SQLWarning getWarnings()
        throws SQLException {

        return this.con.getWarnings();
    }

    /**
     * Just calls the underlying statement. We override it so the
     * <code>postCallHook</code> is not called, which is otherwise the
     * default behaviour of <code>StatementWrapper</code>.
     */
    public void clearWarnings()
        throws SQLException {

        this.con.clearWarnings();
    }

    /**
     * Closes the wrapped statement and returns this active statement instance
     * back to the pool. Don't use the instance after this method call,
     * because it can be reassigned to another borrower.
     *
     * @throws SQLException if the wrapped statement could not be closed.
     * The instance is returned back to the pool and becomes unusable anyway.
     */
    public void close()
        throws SQLException {

        // we don't synchronize anything here, because this method
        // can proceed only after removing the statement from the
        // list of registered statements in the app context, and access
        // to that list is always synchronized and getStats is called
        // from the synchronized block too

        this.appCtx.removeStatement(this.id);

        try {
            this.stmt.close();
            this.state = STATE_CLOSED;
            this.lastStateChangeTs = System.currentTimeMillis();
        } finally {
            this.stmt = this.pstmt = this.cstmt = null;
            this.id = null;
            this.sql = null;
            this.params = null;
            this.maxParamIndex = -1;
            this.pool.returnObject(this);
        }
    }

    /**
     * Sets the active statement's state to <code>STATE_EXECUTING</code>
     * and logs the call.
     *
     * @param executeType type of the execute call.
     * @param sql SQL text passed for execution from the caller.
     *
     * @return the <code>sql</code> argument.
     */
    protected String preexecuteHook(int executeType, String sql) {

        synchronized(this.stateLock) {
            if(this.sql == null)
                this.sql = sql;
            this.state = STATE_EXECUTING;
            this.lastStateChangeTs = System.currentTimeMillis();

            if(((ActiveConnection)this.con).log.isDebugEnabled()) {
                switch(executeType) {
                case EXECUTE_TYPE_QUERY:
                    ((ActiveConnection)this.con).log.
                        debug("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                              ": executing query [" + this.sql + "]" +
                              (sql == null ?
                                  ((this.params != null) && (this.params.length > 0) ?
                                      "\nparams are " + TextUtils.arrayToString(this.params, 0, this.maxParamIndex) :
                                      "") :
                                  ""));
                    break;
                case EXECUTE_TYPE_UPDATE:
                    ((ActiveConnection)this.con).log.
                        debug("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                              ": executing update [" + this.sql + "]" +
                              (sql == null ?
                                  ((this.params != null) && (this.params.length > 0) ?
                                      "\nparams are " + TextUtils.arrayToString(this.params, 0, this.maxParamIndex) :
                                      "") :
                                  ""));
                    break;
                case EXECUTE_TYPE_GENERIC:
                    ((ActiveConnection)this.con).log.
                        debug("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                              ": executing sql [" + this.sql + "]" +
                              (sql == null ?
                                  ((this.params != null) && (this.params.length > 0) ?
                                      "\nparams are " + TextUtils.arrayToString(this.params, 0, this.maxParamIndex) :
                                      "") :
                                  ""));
                    break;
                default: // EXECUTE_TYPE_BATCH:
                    ((ActiveConnection)this.con).log.
                        debug("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                              ": executing batch");
                }
            }
        }

        return sql;
    }

    /**
     * Sets the active statement's state to
     * <code>STATE_PROCESSING_RESULT</code>.
     *
     * @param rs the result set.
     * @param cnt update count.
     * @param isResultSet is result set flag.
     */
    protected void postexecuteHook(ResultSet rs, int cnt, boolean isResultSet) {

        synchronized(this.stateLock) {
            this.state = STATE_PROCESSING_RESULT;
            this.lastStateChangeTs = System.currentTimeMillis();
        }
    }

    /**
     * Checks for SQL warnings and logs them if there are any.
     *
     * @throws SQLException if underlying <code>getWarnings</code>
     * throws an exception.
     */
    protected void postCallHook()
        throws SQLException {

        boolean clear = false;
        for(SQLWarning warn = this.stmt.getWarnings(); warn != null; warn = warn.getNextWarning()) {
            clear = true;
            ((ActiveConnection)this.con).log.
                warn("con #" + ((ActiveConnection)this.con).getRealConnectionId() +
                     ": error code " + warn.getErrorCode() +
                     ", sql state " + warn.getSQLState() +
                     ", " + warn.getMessage());
        }
        if(clear)
            this.stmt.clearWarnings();
    }


    ///////////////////////////////////////////////////////////////////////////
    // Status provider interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Build a partial XML describing this statement. This method is called
     * from <code>ApplicationContext</code> and its output is included into the
     * application context's stats.
     *
     * @return XML fragment describing this active statement.
     *
     * @throws UnsupportedEncodingException if UTF-8 is not supported on this
     * platform, which is practically impossible.
     */
    public byte [] getStats()
        throws UnsupportedEncodingException {

        ByteArrayOutputStream out = new ByteArrayOutputStream(1024);
        PrintStream res = new PrintStream(out, false, "UTF-8");

        res.print("    <stmt id=\""); res.print(this.id); res.println("\">");
        res.print("      <thread><![CDATA["); res.print(this.createdByThreadName); res.println("]]></thread>");
        synchronized(this.stateLock) {
            res.print("      <stateChangeTs>"); res.print(StatusProvider.TIMESTAMP_FORMAT.format(new java.util.Date(this.lastStateChangeTs))); res.println("</stateChangeTs>");
            if(this.sql != null) {
                res.print("      <query><![CDATA["); res.print(this.sql); res.println("]]></query>");
            }
            if((this.params != null) && (this.params.length > 0)) {
                res.print("      <queryParams><![CDATA[");
                res.print(TextUtils.arrayToString(this.params, 0, this.maxParamIndex));
                res.println("]]></queryParams>");
            }
            res.print("      <status>"); res.print(getActiveStatementStatusName(this.state)); res.println("</status>");
        }
        res.println("    </stmt>\n");

        res.flush();
        return out.toByteArray();
    }

    /**
     * Returns name of the specified active statement state, which is one
     * of <code>STATE_xxx</code> constants.
     *
     * @param state the state's code.
     *
     * @return human readable state name.
     */
    private static String getActiveStatementStatusName(int state) {

        switch(state) {
        case ActiveStatement.STATE_CREATING:
            return "Creating";
        case ActiveStatement.STATE_CREATED:
            return "Created";
        case ActiveStatement.STATE_EXECUTING:
            return "Executing";
        case ActiveStatement.STATE_PROCESSING_RESULT:
            return "Processing result";
        case ActiveStatement.STATE_CANCELLED:
            return "Cancelled";
        case ActiveStatement.STATE_CLOSED:
            return "Closed";
        default:
            return "Unknown";
        }
    }

    /**
     * Not used in this implementation.
     *
     * @param eventName ignored.
     *
     * @return <code>null</code>.
     */
    public com.boylesoftware.cb2.util.EventStatistics getEventStats(String eventName) {

        return null;
    }
}
