package com.boylesoftware.cb2;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.Serializable;
import java.io.PrintStream;
import java.io.IOException;
import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.security.AccessControlException;
import java.lang.reflect.InvocationTargetException;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.beanutils.BeanUtils;

import com.boylesoftware.cb2.util.AccessGuard;
import com.boylesoftware.cb2.util.TextUtils;
import com.boylesoftware.cb2.util.EventStatistics;
import com.boylesoftware.cb2.util.pool.AdaptiveObjectPool;


/**
 * <code>ApplicationContext</code> class provides low-level
 * application-wide services available to all application parts at all times.
 * The services include access to database connections, transaction
 * management, access to the application configuration and logging facilities,
 * broadcast messaging. Application context object is an application-wide
 * "singleton", that is there is only one instance per application.
 *
 * <p>During initialization the application context reads a configuration file
 * called <code>cb2app.properties</code>, which contains application properties,
 * available to the user application code, as well as a set of special
 * properties configuring the application context itself. The file
 * should be in a format very similar to the standard Java properties file
 * format, except that it does not support multiline properies (the backslash
 * syntax) and adds some extensions such as conditional inclusion tags and
 * macros (see below). Also, the set of application properties can be
 * supplemented by values loaded from a database, however there are some
 * basic properties, including those configuring the database connections,
 * that can be specified only in the <code>cb2app.properties</code> file.
 *
 * <h4>Standard Application Properties</h4>
 *
 * <p>The following properties in <code>cb2app.properties</code> configure
 * application context's operation:
 *
 * <ul>
 *
 * <li><code><b>com.boylesoftware.cb2.dataSource.<i>xxx</i>.jndiPath</b></code> -
 * Name in JNDI context where to find a <code>javax.sql.DataSource</code>
 * object to be used by the application for getting database connections.
 * The <code><i>xxx</i></code> is the data source logical name. An application
 * can have multiple database connection sources configured to connect to
 * different databases and data source logical names identify the data sources.
 * The name can be "default" for the default or the only data source in the
 * application. If none of such nor
 * <code>com.boylesoftware.cb2.dataSource.<i>xxx</i>.class</code> properties
 * is specified, the database related services of the application context
 * will be disabled.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.dataSource.<i>xxx</i>.class</b></code> -
 * name of the <code>javax.sql.DataSource</code> interface implementation
 * class. Alternatively to getting the data source object from the JNDI by
 * specifying <code>com.boylesoftware.cb2.dataSource.<i>xxx</i>.jndiPath</code>
 * property, you can create the data source directly by specifying this
 * property, which may be useful for application that do not have access to
 * the JNDI. If this property is specified the application context will create
 * an instance of named class and then will use
 * <code>com.boylesoftware.cb2.dataSource.<i>xxx</i>.property.<i>nnn</i></code>
 * properties to set properties on it as on a Java Bean. Note, that only one
 * <code>jndiPath</code> or <code>class</code> can be specified for one data
 * source.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.dataSource.<i>xxx</i>.property.<i>nnn</i></b></code> -
 * used with <code>com.boylesoftware.cb2.dataSource.<i>xxx</i>.class</code>
 * property to set parameters on the data source. The <code><i>nnn</i></code>
 * is used as Java Bean property name.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.closeDataSourcesAtShutdown</b></code> -
 * if set to <code>true</code> the application context will try to close
 * data sources when shutting down (the <code>destroy</code> method
 * is called). Some database connection pool implementations provide a
 * <code>close</code> method that explicitely closes all database connections
 * in the pool. The application context can try to find such a method a call
 * it when shutting down. By default this property is <code>false</code>,
 * meaning that nothing special is done with the data sources on shutdown.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.closeJNDIContexts</b></code> -
 * if this property is set to <code>false</code> the appliction context will
 * not try to close JNDI <code>InitialContext</code> each time after use.
 * Since <code>InitialContext</code> represents a connection to the JNDI
 * provider it is a good practice to close it explicitly after it is no longer
 * needed, and then, instantiation of a another <code>InitialContext</code>
 * reestablishes the connection. Some broken Servlet Containers (for example
 * ServletExec by New Atlanta Communications) shutdown the whole JNDI subsystem
 * when <code>close</code> is called on <code>InitialContext</code> so the
 * application can never do lookups on any new instance of
 * <code>InitialContext</code> ever since. To fix this problem you can set
 * this property to <code>false</code> and CB2 will not be closing
 * <code>InitialContext</code> instances after use, but will be just discarding
 * them. By default this property is <code>true</code>. This property can be
 * specified only in <code>cb2app.properties</code> file and not the
 * database.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.txContext</b></code> -
 * Name of the class implementing <code>TransactionContext</code> interface.
 * If this parameter is not specified
 * <code>com.boylesoftware.cb2.SimpleTransactionContext</code> is used by
 * default. CB2 has a few transaction context implementations packaged with
 * it, which provide different features. See standard trasaction context
 * implementations description below.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.appProperties.dataSource</b></code> -
 * logical name of the data source to be used with the query defined by
 * <code>com.boylesoftware.cb2.appProperties.query</code>
 * property to read additional application properties. This property is
 * optional, if it is not defined the "default" data source will be used.
 * See the next property description for more details.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.appProperties.query</b></code> -
 * SQL query that reads additional application properties from the
 * database. The properties, fetched by the query, will be mixed together
 * with the properties from the <code>cb2app.properties</code> configuration
 * file, overriding any existing values, into the final set of application
 * properties. The first column in the result set returned by this query
 * should be a property name and the second column should be the property
 * value. If this property is not set no additional application properties
 * will be loaded from the database and the contents of the
 * <code>cb2app.properties</code> file will fully determine the set of
 * application properties.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.txControlMode</b></code> -
 * can be "relaxed", "updates" or "strict" and determines how the
 * <code>getConnection</code> methods behaive when called without an
 * active transaction context associated with the calling thread, that is when
 * no transaction has been started before the call. When the value is "relaxed"
 * and one of <code>getConnection</code> methods is called while not in a
 * transaction, the method returns a raw database connection directly from the
 * undelrying data source, which is similar to calling
 * <code>getNoTXConnection</code>. When the value is "updates", which is the
 * default when the property is not specified, the methods throw an
 * exception when called for database updates (the <code>forUpdate</code>
 * method argument is not specified or is <code>true</code>). When the value
 * is "strict" the methods throw an exception when called without an active tx
 * context. This parameter may come from the database-stored part of
 * the application properties set.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.txIsolationLevel</b></code> -
 * defines the transaction isolation level, which is set by the application
 * context on all database connections before giving them to the application
 * code. Possible values are: 1 - READ UNCOMMITTED, 2 - READ COMMITTED,
 * 3 - REPEATABLE READ and 4 - SERIALIZABLE. If not specified then application
 * context does not set transaction isolation level on connections and it is
 * left in its default state provided by the undelrying data source or
 * JDBC driver implementation.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.broadcastMessaging.provider.class</b></code> -
 * name of the CB2 Broadcast Messaging provider implementation class. The
 * class must implement the <code>BroadcastMessagingProvider</code>
 * interface. This property is optional and if it is not specified the
 * <code>broadcastBCMMessage</code> method does nothing. You must
 * provide CB2 with a provider if you build a cluster of CB2-based
 * applications, because BCM is the mechanism used for communication between
 * cluster members. CB2 provides some standard provider implementations,
 * see packages under <code>com.boylesoftware.cb2.bcmproviders</code>.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.initializeLogging</b></code> -
 * CB2 uses Apache Commons Logging as the logging subsystem interface and it
 * assumes that the actual log kit is initialized independently in
 * an implementation specific way. On the other hand, application context
 * knows how to initialize some logging kits and can do it if this optional
 * property is specified so the user application code does not have to do
 * anything about initializing logging. If this property is specified and
 * names a supported logging kit the application context will attempt to
 * perform initialization of it during (re)configuration. Currently
 * two values are supported (case-insensitive): "log4j" and "jdk14". If the
 * property is not specified no underlying logging imeplementation
 * initialization is performed by the application context. Note, that if the
 * property is specified, some additional application properties may be
 * required and processed depending on the log kit chosen.</li>
 *
 * <li><code><b>org.apache.commons.logging.Log</b></code> -
 * this property defines Apache Commons Logging log implementation to be used
 * with the application. See Apache Commons Logging documentation for more
 * details. If this property is not defined the logging service will be
 * disabled and <code>getLog</code> method will be returning <code>null</code>
 * upon call.</li>
 *
 * <li><code><b>org.apache.log4j.config.file</b></code> -
 * this optional property is used when
 * <code>com.boylesoftware.cb2.initializeLogging</code> is set to "log4j",
 * it specifies the Log4J configuration file. If the value starts with a slash
 * (/) it is assumed to be an absolute path, otherwise it is relative to the
 * application context config directory. If the filename ends with ".xml"
 * (case-insensitive) then the file format is assumed to be XML
 * and <code>org.apache.log4j.xml.DOMConfigurator</code> is used.
 * Otherwise it is assumed to be a standard properties file and
 * <code>org.apache.log4j.PropertyConfigurator</code> is used. If
 * <code>com.boylesoftware.cb2.initializeLogging</code> is set to "log4j" and
 * this property is not specified, then <code>PropertyConfigurator</code>
 * is used and preprocessed (that is all macros expanded, all conditions
 * processed) <code>cb2app.properties</code> file is fed to it.</li>
 *
 * <li><code><b>java.util.logging.config.file</b></code> -
 * this optional property is used when
 * <code>com.boylesoftware.cb2.initializeLogging</code> is set to "jdk14",
 * it specifies the logging configuration file in standard properties format
 * used to configure the <code>java.util.logging.LogManager</code>. If the
 * value starts with a slash (/) it is assumed to be an absolute path,
 * otherwise it is relative to the application context config directory. If
 * <code>com.boylesoftware.cb2.initializeLogging</code> is set to "jdk14" and
 * this property is not specified, then preprocessed
 * <code>cb2app.properties</code> file itself is used to configure the
 * log manager.</li>
 *
 * </ul>
 *
 * <h4>Transaction Contexts</h4>
 *
 * <p>There are several standard transaction context implementations included
 * in CB2. Each of them has a different set of features and is more suitable
 * for this or that particular application. Here is their description:
 *
 * <ul>
 *
 * <li><code><b>com.boylesoftware.cb2.SimpleTransactionContext</b></code> -
 * this implementation is used by default when no
 * <code>com.boylesoftware.cb2.txContext</code> application property is
 * specified. As its name suggests, it is a very simple implementation that
 * uses JDBC's <code>setAutoCommit</code>, <code>commit</code> and
 * <code>rollback</code> methods on <code>java.sql.Connection</code> to manage
 * transactions. In many cases using this implementation is the best choice
 * in terms of efficiency and reliability. However, it has certain limitations:
 * it does not support distributed transactions - an error is raised when
 * trying to execute commands on different database connections within single
 * transaction. Also, it does not support nested transactions - it is an error
 * to try to start new transaction in a thread that already has started one
 * and has not finished it yet by committing or rolling it back. Use this
 * implementation if your application works with just one database, does not
 * require distributed transactions, does not want to include other than
 * database operations (like JMS operations for example) to transactions.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.JTATransactionContext</b></code> -
 * uses JTA's <code>javax.transaction.UserTransaction</code> interface, which
 * it looks up in JNDI, to manage transactions. This transaction context,
 * depending on the underlying JTA implementation, usually supports distributed
 * transactions, nested transactions, and including JMS and other non-database
 * operations to transactions. By default this implementation looks up
 * <code>UserTransaction</code> object in JNDI under name
 * "java:comp/UserTransaction". It can be overridden by specifying a different
 * name with <code>com.boylesoftware.cb2.userTransaction.jndiPath</code>
 * application property.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.LanguageTransactionContext</b></code> -
 * this implementation uses SQL language commands on database connections to
 * manage transactions. Sometimes it is better to use this implementation
 * rather than <code>com.boylesoftware.cb2.SimpleTransactionContext</code> to
 * workaround bugs in JDBC drivers, to allow, if supported by the database,
 * nested transactions, or for better performance. By default this
 * implementation uses <code>BEGIN TRANSACTION</code> command to begin a new
 * transaction, <code>COMMIT TRANSACTION</code> to commit it, and
 * <code>ROLLBACK TRANSACTION</code> to roll it back. These commands can be
 * overridden with <code>com.boylesoftware.cb2.txContext.command.begin</code>,
 * <code>com.boylesoftware.cb2.txContext.command.commit</code> and
 * <code>com.boylesoftware.cb2.txContext.command.rollback</code> application
 * properties respectively.</li>
 *
 * </ul>
 *
 * <h4>Appliaction Properties File Extended Syntax</h4>
 *
 * <p>The <code>cb2app.properties</code> file allows some extensions
 * to the basic properties syntax. One extension allows you to conditionally
 * include or exclude different parts of the file basing on values of other
 * properties in the same file or system properties. A single line in the file
 * can contain a condition tag in one of the following three forms:
 *
 * <ul>
 *
 * <li><code>[<i>property name</i>]</code> - properties that follow this line
 * until next condition tag or until the end of the file will be included
 * only if the named property is defined in system properties or in this
 * file above this line.</li>
 *
 * <li><code>[^<i>property name</i>]</code> - properties that follow this line
 * until next condition tag or until the end of the file will be included
 * only if the named property is not defined either in system properties or
 * in this file above this line.</li>
 *
 * <li><code>[<i>property name</i>:<i>property value</i>]</code> - properties
 * that follow this line until next condition tag or until the end of the file
 * will be included only if the named property is defined and equals to
 * the specified value.</li>
 *
 * <li><code>[^<i>property name</i>:<i>property value</i>]</code> - properties
 * that follow this line until next condition tag or until the end of the file
 * will be included only if the named property is not defined or its value is
 * not equal to the specified value.</li>
 *
 * <li><code>[/]</code> - properties that follow this line until next
 * condition tag or until the end of the file will be included regarless
 * of any conditions (end of current condition tag).</li>
 *
 * </ul>
 *
 * <p>Another extention allows macros. If you put <code>${<i>name</i>}</code>
 * construct in a property value it will be replaced with the value of another
 * property named <code><i>name</i></code>. If there is no such property the
 * tag will be left intact. Note, that you can use only properties defined
 * above the place the macro appears, or system properties.
 *
 * <h4>Internal Log Messages</h4>
 *
 * <p>Application context logs debug messages under
 * <code>com.boylesoftware.cb2.APPCTX</code> channel.
 *
 * @author Lev Himmelfarb
 * @version $Id: ApplicationContext.java,v 1.22 2004/10/11 18:17:36 levahim Exp $
 */
public final class ApplicationContext
    implements StatusProvider, BroadcastMessagingTerminal {

    /**
     * Subsystem name for the logger.
     */
    private final static String LOG_SUBSYS = "com.boylesoftware.cb2.APPCTX";

    /**
     * Path to the application initialization parameters file. This
     * path is relative to <code>configDirPath</code> passed to the init method.
     */
    private final static String CB2_CONFIG_FILE = "/cb2app.properties";

    /**
     * Name of BCM channel on which <code>setProperty</code> method broadcasts
     * application property change events.
     */
    private final static String CHANGE_PROPERTY_CHANNEL =
        "com.boylesoftware.cb2.ApplicationContext.CHANGE_PROPERTY";

    /**
     * Constant for the "relaxed" tx control mode. The
     * <code>getConnection</code> methods return a raw connection when
     * called without an active tx context associated with the calling
     * thread.
     */
    private final static int TXCTLMODE_RELAXED = 1;

    /**
     * Constant for the "updates" tx control mode. The
     * <code>getConnection</code> methods throw an exception when
     * called for database updates without an active tx context
     * associated with the calling thread. This is the default mode.
     */
    private final static int TXCTLMODE_UPDATES = 2;

    /**
     * Constant for the "strict" tx control mode. The
     * <code>getConnection</code> methods throw an exception when
     * called without an active tx context associated with the calling
     * thread.
     */
    private final static int TXCTLMODE_STRICT = 3;

    /**
     * Pattern for conditional inclusion tags in the
     * <code>cb2app.properties</code> file. 
     */
    private final static Pattern CONDTAG_PATTERN =
        Pattern.compile("^\\[([^\\[\\]:]+)(:([^\\[\\]]*))?\\]$");


    ///////////////////////////////////////////////////////////////////////////
    // Context data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Application name.
     */
    private final String appName;

    /**
     * Application context access guard.
     */
    private final AccessGuard guard;

    /**
     * Filesystem path to a directory that contains various
     * application configuration files including
     * <code>cb2app.properties</code>.
     */
    private String configDirPath;

    /**
     * The application properties from the
     * <code>cb2app.properties</code> file.
     */
    private Properties basicProperties;

    /**
     * The application properties from the the
     * <code>cb2app.properties</code> file plus properties optionally
     * read from the database.
     */
    private Properties properties;

    /**
     * Collection of <code>DataSource</code> objects. The keys are
     * data source logical names (objects of <code>String</code>) as
     * specified in the configuration file, the values are data source
     * objects (objects of <code>UniversalDataSource</code>). This map
     * is populated once during application context re/initialization.
     * If this field is <code>null</code> it means that the
     * database-related services are not available.
     */
    private Map dataSources;

    /**
     * Set to <code>true</code> if we shuold close InitialContext after use.
     */
    private boolean closeJNDIContexts;

    /**
     * Apache Commons Logging log factory.
     */
    private LogFactory logFactory;

    /**
     * Implementation class for transaction contexts.
     */
    private Class txContextClass;

    /**
     * Storage of transaction contexts associated with threads.
     */
    private static ThreadLocal txContexts = new ThreadLocal();

    /**
     * The transaction control mode. One of TXCTLMODE_xxx constants.
     */
    private int txControlMode;

    /**
     * Transaction isolation level. <code>Connection.TRANSACTION_NONE</code>
     * if not used.
     */
    private int txIsolationLevel;

    /**
     * Contains all registered Broadcast Messaging terminals. The kays
     * are channel names (objects of <code>String</code>) and the
     * values are terminal instances (objects of
     * <code>BroadcastMessagingTerminal</code>).
     */
    private final Map bcmTerminals;

    /**
     * The Broadcast Messaging provider instance from the
     * <code>com.boylesoftware.cb2.broadcastMessaging.provider.class</code>
     * application property.
     */
    private BroadcastMessagingProvider bcmProvider;

    /**
     * Indicates if the BCM provider is present and initialized. This flag
     * is used in the complex scheme of BCM provider reinitialization.
     * See <code>reconfigure</code> and <code>broadcastBCMMessage</code>
     * methods.
     */
    private boolean isBCMProviderInitialized;

    /**
     * Pool of <code>ActiveConnection</code> objects.
     */
    private final AdaptiveObjectPool actionConnectionPool;

    /**
     * Collection of active statements. The keys are statement ids
     * (objects of <code>Long</code>) and the values
     * are ative statement instances (objects of
     * <code>ActiveStatement</code>).
     */
    private final Map activeStatements;

    /**
     * Tells if the <code>destroy</code> method has been called and
     * the application context is in the process of shutting itself down.
     */
    private boolean isShuttingDown;


    ///////////////////////////////////////////////////////////////////////////
    // Context lifecycle
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Create an application context instance and initializes
     * it. There should be only one instance of application context in
     * the application, which has to be created during the application
     * startup before any other subsystem starts using it. Later, in the
     * runtime, you can call <code>reconfigure</code> to have the context
     * reload application properties and reinitialize itself.
     *
     * @param appName the application name, should be unique within the JVM.
     * When you have multiple logical applications in the same JVM, such as for
     * example multiple CB2 web-applications running under the same servlet
     * container, it is necessary to assign different names to them.
     * @param configDirPath points to the directory which contains application
     * configuration files including <code>cb2app.properties</code>.
     *
     * @throws ApplicationException there was a fatal error during
     * context initialization. The context, and thus the whole
     * application, is unavailable.
     */
    public ApplicationContext(String appName, String configDirPath)
        throws ApplicationException {

        this.appName = appName;
        this.guard = new AccessGuard();
        this.bcmTerminals = new HashMap();
        this.isBCMProviderInitialized = false;
        this.actionConnectionPool =
            new AdaptiveObjectPool(new ActiveConnectionFactory(this));
        this.activeStatements = new HashMap();
        this.isShuttingDown = false;
        this.logFactory = null;

        startupLog("creating new application context");

        this.registerBCMTerminal(CHANGE_PROPERTY_CHANNEL, this);

        this.reconfigure(configDirPath);
    }

    /**
     * Re/configure the context. This method reloads all application
     * properties and reinitializes all internal services.
     *
     * @param configDirPath points to the directory which contains application
     * configuration files including <code>cb2app.properties</code>.
     *
     * @throws ApplicationException there was a fatal error during
     * context configuration. The context, and thus the whole
     * application, is unavailable and has to be shut down
     * immediately.
     */
    public void reconfigure(String configDirPath)
        throws ApplicationException {

        try {
            this.guard.enterExclusive();

            startupLog("starting application context configuration");

            this.configDirPath = configDirPath;

            // read the application properties
            this.reloadBasicProperties();

            // set essential flags
            String closeJNDIContextsS = this.basicProperties.
                getProperty("com.boylesoftware.cb2.closeJNDIContexts");
            if(closeJNDIContextsS == null)
                closeJNDIContextsS = "true";
            this.closeJNDIContexts = Boolean.valueOf(closeJNDIContextsS).booleanValue();

            // create data sources
            this.createDataSources();

            // set the basic properties
            try {
                this.properties = new Properties(System.getProperties());
            } catch(AccessControlException e) {
                startupLog("WARNING! Listing system properties is not permitted by AccessController," +
                           " the application properties will not be backed by the system properties.");
                this.properties = new Properties();
            }
            this.properties.putAll(this.basicProperties);

            // add properties stored in the database to the application properties set
            this.loadDatabaseProperties();

            // find and load transaction context implementation class
            String txContextClassName = this.properties.
                getProperty("com.boylesoftware.cb2.txContext");
            if(txContextClassName == null) {
                this.txContextClass = SimpleTransactionContext.class;
            } else {
                try {
                    this.txContextClass = Class.forName(txContextClassName);
                    if(!(TransactionContext.class).isAssignableFrom(this.txContextClass)) {
                        startupLog("ERROR! specified tx context class does not implement TransactionContext interface.");
                        throw new ApplicationException("Specified tx context class does not implement TransactionContext interface.");
                    }
                } catch(ClassNotFoundException e) {
                    startupLog("ERROR! cannot load tx context implementation class " + txContextClassName +
                               ": " + e);
                    throw new ApplicationException("Cannot load tx context implementation class " + txContextClassName + ".", e);
                }
            }

            // get the tx control mode
            String txControlModeS = this.properties.getProperty("com.boylesoftware.cb2.txControlMode");
            if((txControlModeS == null) || "updates".equalsIgnoreCase(txControlModeS))
                this.txControlMode = TXCTLMODE_UPDATES;
            else if("relaxed".equalsIgnoreCase(txControlModeS))
                this.txControlMode = TXCTLMODE_RELAXED;
            else if("strict".equalsIgnoreCase(txControlModeS))
                this.txControlMode = TXCTLMODE_STRICT;
            else {
                startupLog("ERROR! invalid value of the com.boylesoftware.cb2.txControlMode parameter [" + txControlModeS + "]");
                throw new ApplicationException("Invalid value of the com.boylesoftware.cb2.txControlMode parameter [" +
                                               txControlModeS + "]");
            }

            // get the tx isolation level
            this.txIsolationLevel = Connection.TRANSACTION_NONE;
            String txIsolationLevelS = this.properties.getProperty("com.boylesoftware.cb2.txIsolationLevel");
            if(txIsolationLevelS != null) {
                try {
                    switch(Integer.parseInt(txIsolationLevelS)) {
                    case 1:
                        this.txIsolationLevel = Connection.TRANSACTION_READ_UNCOMMITTED;
                        break;
                    case 2:
                        this.txIsolationLevel = Connection.TRANSACTION_READ_COMMITTED;
                        break;
                    case 3:
                        this.txIsolationLevel = Connection.TRANSACTION_REPEATABLE_READ;
                        break;
                    case 4:
                        this.txIsolationLevel = Connection.TRANSACTION_SERIALIZABLE;
                        break;
                    default:
                        startupLog("ERROR! invalid value of the com.boylesoftware.cb2.txIsolationLevel parameter [" +
                                   txIsolationLevelS + "]");
                        throw new ApplicationException("Invalid value of the com.boylesoftware.cb2.txIsolationLevel parameter [" +
                                                       txIsolationLevelS + "]");
                    }
                } catch(NumberFormatException e) {
                    startupLog("ERROR! invalid value of the com.boylesoftware.cb2.txIsolationLevel parameter [" +
                               txIsolationLevelS + "]");
                    throw new ApplicationException("Invalid value of the com.boylesoftware.cb2.txIsolationLevel parameter [" +
                                                   txIsolationLevelS + "]", e);
                }
            }

            // get the Broadcast Messaging provider
            if(this.bcmProvider != null) {
                this.isBCMProviderInitialized = false;
                this.bcmProvider.destroy();
                this.bcmProvider = null;
            }
            String bcmProviderClassName = this.properties.getProperty("com.boylesoftware.cb2.broadcastMessaging.provider.class");
            if(bcmProviderClassName != null) {
                try {
                    startupLog("instantiating BCM provider " + bcmProviderClassName);
                    this.bcmProvider = (BroadcastMessagingProvider)Class.
                        forName(bcmProviderClassName).newInstance();
                } catch(ClassNotFoundException e) {
                    startupLog("ERROR! specified BCM provider class was not found: " + e);
                    throw new ApplicationException("Could not load the BCM provider class.", e);
                } catch(IllegalAccessException e) {
                    startupLog("ERROR! could not instantiate the specified BCM provider class: " + e);
                    throw new ApplicationException("Could not instantiate the BCM provider class.", e);
                } catch(InstantiationException e) {
                    startupLog("ERROR! could not instantiate the specified BCM provider class: " + e);
                    throw new ApplicationException("Could not instantiate the BCM provider class.", e);
                }
            } else
                startupLog("no broadcast messaging provider is specified, the BCM and clustering will be disabled");

            // configure the logging subsystem
            this.reconfigureLogging();

        } finally {
            this.guard.leaveExclusive();
        }

        // initialize the BCM provider (we do it separately bacause
        // it requires access to the application context resources and
        // thus the application context has to be configured)
        try {
            this.guard.enterShared();

            if(this.bcmProvider != null) {
                synchronized(this.bcmProvider) {
                    if(!this.isBCMProviderInitialized) {
                        startupLog("initializing the BCM provider");
                        this.bcmProvider.init(this);
                        this.isBCMProviderInitialized = true;
                    }
                }
            }

        } finally {
            this.guard.leaveShared();
        }

        // all done
        startupLog("the application context has been successfully configured.");
    }

    /**
     * Loads basic application properties from the
     * <code>cb2app.properties</code> file. If there are already basic
     * properties cached in the application context, the method
     * <strong>clears</strong> it and replaces with a new object.
     *
     * @throws ApplicationException if an error happens while loading the
     * properties.
     */
    private void reloadBasicProperties()
        throws ApplicationException {

        if(this.basicProperties != null)
            this.basicProperties.clear();
        this.basicProperties = new Properties();

        String configFilePath = this.configDirPath + CB2_CONFIG_FILE;
        startupLog("reading initial configuration from " + configFilePath);
        try {

            BufferedReader in = null;
            try {
                in = new BufferedReader(new FileReader(configFilePath));

                String str;
                boolean inSkipMode = false;
                int curLine = 0;
                while((str = in.readLine()) != null) {
                    curLine++;
                    if((str = str.trim()).length() > 0) {
                        Matcher matcher = CONDTAG_PATTERN.matcher(str);
                        boolean isCondTag = matcher.matches();
                        if(inSkipMode) {
                            if(isCondTag)
                                inSkipMode = false;
                            else
                                continue;                                
                        }
                        if(isCondTag) {
                            String name = matcher.group(1);
                            if(name.equals("/"))
                                continue;
                            boolean inverse = false;
                            if(name.charAt(0) == '^') {
                                name = name.substring(1).trim();
                                inverse = true;
                            }
                            String value = this.basicProperties.getProperty(name);
                            if(value == null)
                                value = System.getProperty(name);
                            if(value == null) {
                                inSkipMode = !inverse;
                                continue; 
                            }
                            if(matcher.group(3) == null) {
                                inSkipMode = inverse;
                                continue;
                            }
                            inSkipMode = !(matcher.group(3).trim().equals(value) ^ inverse);

                        } else if((str.charAt(0) == '#') || (str.charAt(0) == '!')) {
                            continue;

                        } else {
                            int ind = str.indexOf('=');
                            int ind0 = str.indexOf(':');
                            if((ind0 >= 0) && (ind0 < ind))
                                ind = ind0;
                            if(ind <= 0)
                                throw new ApplicationException("Invalid syntax in cb2app.properties file, line " +
                                                               curLine);
                            String propName = str.substring(0, ind).trim();
                            String propValue = TextUtils.replaceMacros(str.substring(ind + 1).trim(),
                                                                       new TextUtils.MapValueGetter(this.basicProperties) {

                                                                           public String get(String name) {

                                                                               String value = super.get(name);

                                                                               return (value != null ? value : System.getProperty(name));
                                                                           }
                                                                       },
                                                                       true);
                            this.basicProperties.setProperty(propName, propValue);
                        }
                    }
                }

            } finally {
                if(in != null)
                    in.close();
            }

        } catch(IOException e) {
            startupLog("ERROR! got input/output exception reading configuration file: " + e);
            throw new ApplicationException("Could not load configuration file.", e);
        }
    }

    /**
     * Creates data sources defined in the basic application properties. This
     * method does not destroy, nor close any existing data sources.
     *
     * @throws ApplicationException if an error happens creating data sources.
     */
    private void createDataSources()
        throws ApplicationException {

        this.dataSources = new HashMap();

        startupLog("creating data sources");
        try {

            InitialContext ic = null;
            Connection con = null;
            try {

                Properties dsProps = new Properties();
                Enumeration en = this.basicProperties.propertyNames();
                while(en.hasMoreElements()) {
                    String propName = (String)en.nextElement();
                    if(propName.startsWith("com.boylesoftware.cb2.dataSource.")) {
                        int p = propName.indexOf('.', "com.boylesoftware.cb2.dataSource.".length());
                        if(p < 0) {
                            startupLog("ERROR! Invalid property " + propName);
                            throw new ApplicationException("Invalid property " + propName);
                        }
                        String dsName = propName.substring("com.boylesoftware.cb2.dataSource.".length(), p);
                        if(!this.dataSources.containsKey(dsName)) {
                            String prefix = "com.boylesoftware.cb2.dataSource." + dsName + ".";

                            String dsJndiPath = null;
                            String dsClassName = null;
                            DataSource ds;

                            Enumeration en1 = this.basicProperties.propertyNames();
                            while(en1.hasMoreElements()) {
                                propName = (String)en1.nextElement();
                                if(propName.startsWith(prefix)) {
                                    String suffix = propName.substring(prefix.length());
                                    if(suffix.equals("jndiPath")) {
                                        dsJndiPath = this.basicProperties.getProperty(propName);
                                    } else if(suffix.equals("class")) {
                                        dsClassName = this.basicProperties.getProperty(propName);
                                    } else if(suffix.startsWith("property.")) {
                                        String dsPropName = propName.substring(prefix.length() + "property.".length());
                                        dsProps.setProperty(dsPropName, this.basicProperties.getProperty(propName));
                                    } else {
                                        startupLog("ERROR! Invalid property " + propName);
                                        throw new ApplicationException("Invalid property " + propName);
                                    }
                                }
                            }

                            if((dsJndiPath != null) && (dsClassName != null)) {
                                startupLog("ERROR! You cannot specify both jndiName and class properties for data source " + dsName);
                                throw new ApplicationException("You cannot specify both jndiName and class properties for data source " +
                                                               dsName);
                            }

                            if(dsJndiPath != null) {
                                startupLog("getting " + dsName + " DataSource object from JNDI path " + dsJndiPath);
                                if(ic == null)
                                    ic = new InitialContext();
                                ds = (DataSource)ic.lookup(dsJndiPath);
                                if(ds == null) {
                                    startupLog("ERROR! Could not find DataSource in JNDI under " + dsJndiPath);
                                    throw new ApplicationException("Could not find DataSource in JNDI under " + dsJndiPath);
                                }
                                startupLog("ok, got it.");

                            } else if(dsClassName != null) {
                                startupLog("creating " + dsName + " DataSource object from class " + dsClassName);
                                try {
                                    ds = (DataSource)Class.forName(dsClassName).newInstance();
                                    en1 = dsProps.propertyNames();
                                    while(en1.hasMoreElements()) {
                                        String dsPropName = (String)en1.nextElement();
                                        BeanUtils.setProperty(ds, dsPropName, dsProps.getProperty(dsPropName));
                                    }
                                } catch(ClassNotFoundException e) {
                                    startupLog("ERROR! Could not load implementation class for data source " + dsName);
                                    throw new ApplicationException("Could not load implementation class for data source " + dsName, e);
                                } catch(InstantiationException e) {
                                    startupLog("ERROR! Could not instantiate implementation class for data source " + dsName);
                                    throw new ApplicationException("Could not instantiate implementation class for data source " +
                                                                   dsName, e);
                                } catch(IllegalAccessException e) {
                                    startupLog("ERROR! Error accessing implementation class for data source " + dsName);
                                    throw new ApplicationException("Error accessing implementation class for data source " + dsName, e);
                                } catch(InvocationTargetException e) {
                                    startupLog("ERROR! Error setting property on implementation class for data source " + dsName);
                                    throw new ApplicationException("Error setting property on implementation class for data source " +
                                                                   dsName, e);
                                }
                                startupLog("ok, got it.");
                            } else {
                                startupLog("ERROR! Either jndiName and class property has to be specified for data source " + dsName);
                                throw new ApplicationException("Either jndiName and class property has to be specified for data source " +
                                                               dsName);
                            }

                            dsProps.clear();

                            // store the reference to the DataSource found
                            this.dataSources.put(dsName, new UniversalDataSource(ds));

                            // try to obtain a connection from the data source
                            startupLog("trying to get a test database connection...");
                            con = ds.getConnection();
                            con.close();
                            con = null;
                            startupLog("ok, got it.");
                        }
                    }
                }

            } finally {
                if(con != null)
                    con.close();
                if(this.closeJNDIContexts && (ic != null))
                    ic.close();
            }

        } catch(NamingException e) {
            startupLog("ERROR! got JNDI exception during application context configuration: " + e);
            throw new ApplicationException("Could not configure application context.", e);

        } catch(SQLException e) {
            startupLog("ERROR! got JDBC exception during application context configuration: " + e);
            throw new ApplicationException("Could not configure application context.", e);
        }
    }

    /**
     * Loads application properties from the database and adds them to the main
     * set of application properties.
     *
     * @throws ApplicationException if an error happens loading the properties.
     */
    private void loadDatabaseProperties()
        throws ApplicationException {

        String getPropertiesQuery = this.basicProperties.
            getProperty("com.boylesoftware.cb2.appProperties.query");
        if((this.dataSources != null) && (getPropertiesQuery != null)) {
            String getPropertiesDSName = this.basicProperties.
                getProperty("com.boylesoftware.cb2.appProperties.dataSource");
            if(getPropertiesDSName == null)
                getPropertiesDSName = "default";
            DataSource ds = (DataSource)this.dataSources.get(getPropertiesDSName);
            if(ds == null) {
                startupLog("ERROR! Data source named \"" + getPropertiesDSName +
                           "\" and supposed to be used for additional app properties, does not exist.");
                throw new ApplicationException("Data source for additional application properties does not exist.");
            }
            startupLog("reading additional properties from the " + getPropertiesDSName +
                       " database with [" + getPropertiesQuery + "]");
            Properties addProps = new Properties();
            try {
                Connection con = null;
                Statement stmt = null;
                ResultSet rs = null;
                try {
                    con = ds.getConnection();
                    stmt = con.createStatement();
                    rs = stmt.executeQuery(getPropertiesQuery);
                    while(rs.next()) {
                        addProps.setProperty(rs.getString(1), rs.getString(2));
                    }
                } finally {
                    if(rs != null)
                        rs.close();
                    if(stmt != null)
                        stmt.close();
                    if(con != null)
                        con.close();
                }
            } catch(SQLException e) {
                startupLog("ERROR! got JDBC exception reading properties from the database: " + e);
                throw new ApplicationException("Could not configure application context.", e);
            }
            this.properties.putAll(addProps);
            startupLog("added/replaced " + addProps.size() + " properties");
        }
    }

    /**
     * Reconfigure logging subsystem. This method releases all logs in the
     * Apache Commons log factory, initializes the underlying log kit
     * implementation, and finally gets new Commons log factory and sets it up.
     *
     * @throws ApplicationException if an error happens.
     */
    private void reconfigureLogging()
        throws ApplicationException {

        startupLog("configuring logging subsystem");

        // release existing logs
        LogFactory.releaseAll();
        this.logFactory = null;

        // initialize underlying logging kit implementation
        String logImpl = this.properties.getProperty("com.boylesoftware.cb2.initializeLogging");
        if(logImpl != null) {
            if(logImpl.equalsIgnoreCase("log4j")) {
                startupLog("configuring Log4J");
                org.apache.log4j.LogManager.resetConfiguration();
                String cfn = this.properties.getProperty("org.apache.log4j.config.file");
                if(cfn != null) {
                    if(!cfn.startsWith("/"))
                        cfn = this.configDirPath + "/" + cfn;
                    if(cfn.trim().toLowerCase().endsWith(".xml"))
                        org.apache.log4j.xml.DOMConfigurator.configure(cfn);
                    else
                        org.apache.log4j.PropertyConfigurator.configure(cfn);
                } else {
                    org.apache.log4j.PropertyConfigurator.configure(this.properties);
                }
            } else if(logImpl.equalsIgnoreCase("jdk14")) {
                startupLog("configuring JDK14 logging");
                java.util.logging.LogManager lm = java.util.logging.LogManager.getLogManager();
                lm.reset();
                try {
                    InputStream is = null;
                    try {
                        String cfn = this.properties.getProperty("java.util.logging.config.file");
                        if(cfn != null) {
                            if(!cfn.startsWith("/"))
                                cfn = this.configDirPath + "/" + cfn;
                            is = new FileInputStream(cfn);
                        } else {
                            ByteArrayOutputStream baos = new ByteArrayOutputStream();
                            this.properties.store(baos, null);
                            baos.flush();
                            is = new ByteArrayInputStream(baos.toByteArray());
                            baos.close();
                        }
                        lm.readConfiguration(is);
                    } finally {
                        if(is != null)
                            is.close();
                    }
                } catch(IOException e) {
                    startupLog("ERROR! Could not process log kit configuration: " + e);
                    throw new ApplicationException("Could not process log kit configuration.", e);
                }
            } else {
                startupLog("ERROR! Unsupported log kit implementation [" + logImpl + "]");
                throw new ApplicationException("Unsupported log kit implementation [" + logImpl + "]");
            }
        }

        // initialize commons logging
        String logImplClassName = this.properties.
            getProperty("org.apache.commons.logging.Log");
        if(logImplClassName != null) {
            LogFactory.release(this.getClass().getClassLoader());
            this.logFactory = LogFactory.getFactory();
            this.logFactory.setAttribute("org.apache.commons.logging.Log", logImplClassName);
        }
    }

    /**
     * Destroy the application context. This method is usually called
     * as the last thing when the application is shutting down. The
     * application context becomes unusable after this call.
     */
    public void destroy() {

        try {
            this.guard.enterExclusive();

            startupLog("the application context is being destroyed");

            this.isShuttingDown = true;

            this.shutdownDatabase();

            if(this.bcmProvider != null) {
                this.bcmProvider.destroy();
                this.bcmProvider = null;
            }
            this.basicProperties.clear();
            this.basicProperties = null;
            this.properties.clear();
            this.properties = null;
            if(this.dataSources != null) {
                this.dataSources.clear();
                this.dataSources = null;
            }
            this.txContextClass = null;
            this.logFactory = null;

            startupLog("the application context has been destroyed.");

        } finally {
            this.guard.leaveExclusive();
        }
    }

    /**
     * Gracefully shuts down the database access subsystem. Called from
     * the <code>destroy</code> method.
     */
    private void shutdownDatabase() {

        // check and wait for active statements
        for(int i = 0; i < 10; i++) {
            int n;
            synchronized(this.activeStatements) {
                n = this.activeStatements.size();
            }
            if(n == 0)
                break;
            if(i == 0)
                this.startupLog("waiting for active database statements to finish...");
            try {
                Thread.sleep(1000);
            } catch(InterruptedException e) { /* do nothing */ }
        }

        // cancel active statements if there are still any
        synchronized(this.activeStatements) {
            if(this.activeStatements.size() > 0) {
                this.startupLog("cancelling active database statements...");

                Iterator i = this.activeStatements.values().iterator();
                while(i.hasNext()) {
                    try {
                        ((ActiveStatement)i.next()).cancel();
                    } catch(SQLException e) {
                        this.startupLog("got error cancelling an active statement: " + e.getMessage());
                    }
                }
                this.activeStatements.clear();

                this.startupLog("giving the cancelled statements 10 seconds to get cancelled...");
                try {
                    Thread.sleep(10000);
                } catch(InterruptedException e) { /* do nothing */ }
            }
        }

        // close database connection pools
        if((new Boolean(this.properties.getProperty("com.boylesoftware.cb2.closeDataSourcesAtShutdown"))).booleanValue()) {
            synchronized(this.dataSources) {
                Iterator i = this.dataSources.values().iterator();
                while(i.hasNext()) {
                    UniversalDataSource ds = (UniversalDataSource)i.next();
                    this.startupLog("closing database connection pool...");
                    try {
                        ds.close();
                    } catch(SQLException e) {
                        this.startupLog("got error closing a database connection pool: " + e.getMessage());
                    }
                }
            }
        }
    }

    /**
     * This method is used internally by the
     * <code>ApplicationContext</code> class for logging. It posts a
     * message to standard system output stream since the loggin
     * subsystem may not be available.
     *
     * @param msg the message to post.
     */
    private void startupLog(String msg) {

        System.out.println("[APPCTX] [" + this.appName + "] " + msg);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Application context service interface
    ///////////////////////////////////////////////////////////////////////////

	/**
	 * Gets the application name originally set through <code>appName</code>
     * parameter of the application context constructor.
	 *
	 * @return the application name.
	 */
	public String getAppName() {

		return this.appName;
	}

    /**
     * Returns filesystem path to the directory that contains application
     * configuration files including <code>cb2app.properties</code>. Your
     * application may use the directory to store its own specific
     * configuration files and then find them in the runtime using this method.
     *
     * @return path to the configuration directory.
     */
    public String getConfigDirPath() {

        try {
            this.guard.enterShared();

            return this.configDirPath;

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Gets an application property value. Application properties are read from
     * the <code>cb2app.properties</code> configuration file and optionally
     * from the database during application context configuration. The
     * properties then are cached in the application context so this method is
     * fast. If this method cannot find named property among the application
     * properties, it tries to find it among system properties.
     *
     * Note, that sometimes looking up some system property can be disallowed
     * by the security manager
     * (see <code>java.security.AccessController</code>), in which case this
     * method throws a <code>java.security.AccessControlException</code>.
     *
     * @param name the property name.
     *
     * @return the property value or <code>null</code> if there is no
     * such property.
     *
     * @see #getProperties()
     */
    public String getProperty(String name) {

        String res = null;

        try {
            this.guard.enterShared();

            res = this.properties.getProperty(name);

        } finally {
            this.guard.leaveShared();
        }

        return (res != null ? res : System.getProperty(name));
    }

    /**
     * Gets all application properties. The returned <code>Properties</code>
     * object is backed up by the system properties, so if some application
     * property is not specified, the default value for it can be provided by
     * a system property with the same name.
     *
     * <p>Note, that sometimes getting all system properties at once, required
     * to back up application properties with system properties, is prohibited
     * by the security manager
     * (see <code>java.security.AccessController</code>), in which case the
     * <code>Properties</code> object returned by this method is
     * <strong>not</strong> becked up by system properties.
     *
     * @return application properties.
     *
     * @see #getProperty(String)
     */
    public Properties getProperties() {

        try {
            this.guard.enterShared();

            return this.properties;

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Gets value of an application property defined in the
     * <code>cb2app.properties</code> file. This method does not return values
     * of properties loaded from the database (see
     * <code>com.boylesoftware.cb2.appProperties.query</code> application
     * property), only those defined in the file. Also, it always returns the
     * original value of the property even if it was redefined later via a
     * <code>setProperty</code> call. If a property is defined in both the file
     * and the database, the value from the file is returned by this method.
     *
     * <p>This method is not supposed to be used by the application code,
     * however it can be useful for a subsystem, which manages application
     * properties, such as <code>PropertySharedBLO</code> from
     * <code>com.boylesoftware.cb2.commons.property</code> package.
     *
     * @param name the basic application property's name.
     *
     * @return the original value as it is specified in the
     * <code>cb2app.properties</code> file, or <code>null</code> if the
     * property is not defined there.
     */
    public String getBasicProperty(String name) {

        try {
            this.guard.enterShared();

            return this.basicProperties.getProperty(name);

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Changes value of an existing application property or sets up a new
     * application property. This method does not update any underlying
     * property storages - neither the <code>cb2app.property</code> file, nor
     * the database. It merely sets value for a cached in the application
     * context's memory property. It also means that a property, set by this
     * method, does not survive a <code>reconfigure</code> call, unless it was
     * changed in the underlying storage as well.
     *
     * <p>If the specified value is <code>null</code>, the method removes the
     * property from the set of application properties.
     *
     * <p>This method, after the property has been successfully set, sends a
     * message via BCM to all cluster members, so they set their cached
     * properties too.
     *
     * @param name the property name.
     * @param value new property value.
     */
    public void setProperty(String name, String value) {

        try {
            this.guard.enterShared();

            if(value != null)
                this.properties.setProperty(name, value);
            else
                this.properties.remove(name);

        } finally {
            this.guard.leaveShared();
        }

        this.broadcastBCMMessage(CHANGE_PROPERTY_CHANNEL,
                                 new String [] { name, value });
    }

    /**
     * Processes an internal application context BCM message. Do not call this
     * method.
     */
    public void processMessage(String channelName,
                               String source,
                               Serializable messageParam) {

        if(channelName.equals(CHANGE_PROPERTY_CHANNEL)) {
            String [] nameValue = (String [])messageParam;

            try {
                this.guard.enterShared();

                if(nameValue[1] != null)
                    this.properties.setProperty(nameValue[0], nameValue[1]);
                else
                    this.properties.remove(nameValue[0]);

            } finally {
                this.guard.leaveShared();
            }
        }
    }

    /**
     * Returns log for the specified channel. Application context uses Apache
     * Commons Logging as a generic interface to different supported log kits.
     * This method, in fact, results in a call to
     * <code>org.apache.commons.logging.LogFactory</code> configured and
     * maintained inside the application context.
     *
     * @param subsystem name of the log channel (or log category name in terms
     * of some log kit implementations).
     *
     * @return reference to the log object or <code>null</code> if
     * logging is not supported by the context.
     */
    public Log getLog(String subsystem) {

        try {
            this.guard.enterShared();

            if(this.logFactory == null)
                return null;

            return this.logFactory.getInstance(subsystem);

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Convenience method returning value of
     * <code>com.boylesoftware.cb2.closeJNDIContexts</code> application
     * property as a boolean. If the property is not specified, this
     * method returns the default value, which is <code>true</code>.
     *
     * <p>Well-written applications that work with JNDI should consult
     * this method when they don't need <code>InitialContext</code> object
     * anymore and have to decide whether to call <code>close</code> method
     * on it.
     *
     * @return <code>true</code> if <code>InitialContext</code> objects
     * should be closed explicitely after use.
     */
    public boolean getCloseJNDIContexts() {

        try {
            this.guard.enterShared();

            return this.closeJNDIContexts;

        } finally {
            this.guard.leaveShared();
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Transactional database access
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Starts new transaction. In CB2 transaction is represented by a
     * transaction context instance, which is the interface to manage the
     * transaction. One transaction context can represent only one active
     * top-level transaction, however, depending on the implementation, nested
     * transactions within one transaction context may be supported.
     *
     * <p>Optionally, by setting <code>bindToThread</code> parameter to
     * <code>true</code>, you can associate new transaction context with the
     * current thread, which enables calling other transaction management
     * methods from within this thread without specifying transaction context
     * instance. This, in fact, is the most common use case, so there is a
     * version of this method which takes no <code>bindToThread</code>
     * parameter assuming it's <code>true</code> provided for conevnience.
     *
     * @param bindToThread tells whether the new transaction
     * context should be associated with the calling thread. The context is
     * bound to the thread if this parameter is <code>true</code>.
     *
     * @return transaction context instance, which represents the transaction
     * and can be passed to other transaction management methods to identify on
     * which transaction they should perform their actions.
     *
     * @throws ApplicationException if transaction cannot be started.
     *
     * @see #beginTransaction()
     */
    public TransactionContext beginTransaction(boolean bindToThread)
        throws ApplicationException {

        TransactionContext txCtx = null;

        // already have tx context bound to the thread
        if(bindToThread)
            txCtx = (TransactionContext)txContexts.get();

        try {
            this.guard.enterShared();

            // create new txContext if not available
            if(txCtx == null) {
                txCtx = (TransactionContext)this.txContextClass.newInstance();
                txCtx.setup(this);

                // bind it to the thread
                if(bindToThread) {
                    txContexts.set(txCtx);
                    this.getLog(LOG_SUBSYS).debug("bound new transaction context to the thread.");
                }
            }

            // check if unbound txs are supported
            if(!bindToThread && !txCtx.supportsUnboundTransactions())
                throw new UnsupportedOperationException("Unbound from thread transactions are not supported" +
                                                        " by this transaction context implementation.");

            // begin transaction
            txCtx.begin();

        } catch(IllegalAccessException e) {
            throw new ApplicationException("Could not instantiate new transaction context.", e);

        } catch(InstantiationException e) {
            throw new ApplicationException("Could not instantiate new transaction context.", e);

        } finally {
            this.guard.leaveShared();
        }

        return txCtx;
    }

    /**
     * Starts new transaction and binds its transaction context to the calling
     * thread. This is a convenience method, which calls
     * <code>beginTransaction(true)</code> internally.
     *
     * @throws ApplicationException if transaction cannot be started.
     *
     * @see #beginTransaction(boolean)
     */
    public void beginTransaction()
        throws ApplicationException {

        this.beginTransaction(true);
    }

    /**
     * Commits transaction associated with the specified transaction
     * context.
     *
     * @param txCtx the trnasaction context.
     *
     * @throws ApplicationException if underlying subsystem error
     * happens.
     */
    public static void commitTransaction(TransactionContext txCtx)
        throws ApplicationException {

        txCtx.commit();
    }

    /**
     * Commits transaction associated with the calling thread.
     *
     * @throws ApplicationException if underlying subsystem error
     * happens.
     * @throws IllegalStateException if there is no transaction
     * context associated with the thread.
     */
    public static void commitTransaction()
        throws ApplicationException {

        TransactionContext txCtx = (TransactionContext)txContexts.get();
        if(txCtx == null)
            throw new IllegalStateException("There is no transaction context associated with the thread.");

        txCtx.commit();
    }

    /**
     * Rolls back transaction associated with the specified transaction
     * context.
     *
     * @param txCtx the transaction context.
     *
     * @throws ApplicationException if underlying subsystem error
     * happens.
     */
    public static void rollbackTransaction(TransactionContext txCtx)
        throws ApplicationException {

        txCtx.rollback();
    }

    /**
     * Rolls back transaction associated with the calling thread.
     *
     * @throws ApplicationException if underlying subsystem error
     * happens.
     * @throws IllegalStateException if there is no transaction
     * context associated with the thread.
     */
    public static void rollbackTransaction()
        throws ApplicationException {

        TransactionContext txCtx = (TransactionContext)txContexts.get();
        if(txCtx == null)
            throw new IllegalStateException("There is no transaction context associated with the thread.");

        txCtx.rollback();
    }

    /**
     * Marks transaction associated with the specified transaction
     * context as for rollback only. If a transaction is marked as for rollback
     * only, when commit is requested it actually gets rolled back instead. It
     * is useful when at some point an error happens, but you want to continue
     * operation, so you don't throw an exception, while the transaction has to
     * be rolled back anyway when the processing finishes.
     *
     * @param txCtx the trnasaction context.
     *
     * @throws ApplicationException if underlying transaction manager
     * error happens.
     */
    public static void setRollbackOnly(TransactionContext txCtx)
        throws ApplicationException {

        txCtx.setRollbackOnly();
    }

    /**
     * Marks transaction associated with the calling thread for
     * rollback only. Does nothing if there is no transaction context
     * for the thread.
     *
     * @throws ApplicationException if underlying subsystem error
     * happens.
     */
    public static void setRollbackOnly()
        throws ApplicationException {

        TransactionContext txCtx = (TransactionContext)txContexts.get();
        if(txCtx != null)
            txCtx.setRollbackOnly();
    }

    /**
     * Tells if transaction associated with the specified transaction
     * context is marked for rollback only.
     *
     * @param txCtx the trnasaction context.
     *
     * @return <code>true</code> if the transaction is marked for
     * rollback only.
     *
     * @throws ApplicationException if underlying subsystem
     * error happens.
     *
     * @see #setRollbackOnly(TransactionContext)
     */
    public static boolean isRollbackOnly(TransactionContext txCtx)
        throws ApplicationException {

        return txCtx.isRollbackOnly();
    }

    /**
     * Tells if transaction associated with the calling thread is
     * marked for rollback only. Always returns <code>false</code> if
     * there is no transaction context for the thread.
     *
     * @return <code>true</code> if the transaction exists and is marked
     * for rollback only.
     *
     * @throws ApplicationException if underlying subsystem
     * error happens.
     *
     * @see #setRollbackOnly()
     */
    public static boolean isRollbackOnly()
        throws ApplicationException {

        TransactionContext txCtx = (TransactionContext)txContexts.get();
        if(txCtx == null)
            return false;

        return txCtx.isRollbackOnly();
    }

    /**
     * Gets database connection enrolled in the specified transaction context.
     *
     * <p>Never call any transaction management methods on the connection
     * returned by this method - the transaction is managed automatically
     * by the transaction context. Use <code>getNoTXConnection</code> low-level
     * method whenever you need a connection, which is not managed by the
     * application context's transaction management subsystem.
     *
     * <p>Note, that depending on the transaction context implementation, you
     * may not be allowed to call this method to get connections from different
     * data sources within the same transaction.
     *
     * @param txCtx the transaction context.
     * @param dataSourceName name of the data source as defined in the
     * <code>cb2app.properties</code> configuration file. If <code>null</code>,
     * "default" is assumed.
     *
     * @return database connection enrolled in the specified transaction
     * context.
     *
     * @throws SQLException if connection cannnot be retrieved due to the
     * underlying data source error.
     *
     * @see #getNoTXConnection(String)
     */
    public static Connection getConnection(TransactionContext txCtx,
                                           String dataSourceName)
        throws SQLException {

        return txCtx.getConnection(dataSourceName != null ? dataSourceName : "default");
    }

    /**
     * Gets database connection.
     *
     * <p>When this method is called and there is an active transaction
     * associated with the calling thread, that is
     * <code>beginTransaction</code> was called to start transaction and bind
     * the transaction context to the thread, the returned connection is
     * enrolled in the transaction context. If there is no active transaction
     * context for the thread, then the behaviour of this method is determined
     * by the configured transaction control mode (see description of
     * <code>com.boylesoftware.cb2.txControlMode</code> application property)
     * and <code>forUpdate</code> parameter: it may return a raw connection
     * directly from the underlying data source, like
     * <code>getNoTXConnection</code> does, or it may throw an exception.
     *
     * @param dataSourceName name of the data source as defined in the
     * <code>cb2app.properties</code> configuration file. If <code>null</code>,
     * "default" is assumed.
     * @param forUpdate should be set to <code>true</code> if the
     * requested connection is going to be used for database updates,
     * that is any operations that modify data in the database and thus
     * better be performed in a transaction. If the connection is going
     * to be used only to read data from the database, this parameter
     * should be <code>false</code>.
     *
     * @return connection to the database, possibly enrolled in the current
     * transaction.
     *
     * @throws SQLException if the connection cannot be retrieved due
     * to the underlying data source error.
     * @throws IllegalStateException if there is no transaction
     * context associated with the thread and the current transaction
     * control mode requires the method to throw an error.
     *
     * @see #getConnection(TransactionContext, String)
     */
    public Connection getConnection(String dataSourceName, boolean forUpdate)
        throws SQLException {

        TransactionContext txCtx = (TransactionContext)txContexts.get();

        if((txCtx != null) && txCtx.isActive())
            return txCtx.getConnection(dataSourceName != null ? dataSourceName : "default");

        try {
            this.guard.enterShared();

            if((this.txControlMode == TXCTLMODE_STRICT) ||
               ((this.txControlMode == TXCTLMODE_UPDATES) && forUpdate))
                throw new IllegalStateException("There is no transaction context associated with the thread" +
                                                 " or the transaction has not been begun.");

            return this.getNoTXConnection(dataSourceName);

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Convenience method, which calls
     * <code>getConnection(txCtx, "default")</code> internally.
     *
     * @param txCtx the transaction context.
     *
     * @return connection to the default database enrolled in the specified
     * transaction context.
     *
     * @throws SQLException if the connection cannot be retrieved due
     * to the underlying data source error.
     *
     * @see #getConnection(TransactionContext, String)
     */
    public static Connection getConnection(TransactionContext txCtx)
        throws SQLException {

        return txCtx.getConnection("default");
    }

    /**
     * Convenience method, which calls
     * <code>getConnection("default", forUpdate)</code> internally.
     *
     * @param forUpdate tells if the connection is going to be used for
     * modifying data in the database.
     *
     * @return connection to the default database enrolled in the current
     * transaction if there is any.
     *
     * @throws SQLException if the connection cannot be retrieved due
     * to the underlying data source error.
     * @throws IllegalStateException if there is no transaction
     * context associated with the thread and the current transaction
     * control mode requires the method to throw an error.
     *
     * @see #getConnection(String, boolean)
     */
    public Connection getConnection(boolean forUpdate)
        throws SQLException {

        return this.getConnection("default", forUpdate);
    }

    /**
     * Convenience method, which calls
     * <code>getConnection("default", true)</code> internally.
     *
     * @return connection to the default database enrolled in the current
     * transaction if there is any.
     *
     * @throws SQLException if the connection cannot be retrieved due
     * to the underlying data source error.
     * @throws IllegalStateException if there is no transaction
     * context associated with the thread and the current transaction
     * control mode requires the method to throw an error.
     *
     * @see #getConnection(String, boolean)
     */
    public Connection getConnection()
        throws SQLException {

        return this.getConnection("default", true);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Low-level database connection access
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Tells if the database subsystem is enabled for this application context.
     * This method returns <code>false</code> if no data sources are configured
     * in the <code>cb2app.properties</code> file.
     *
     * @return <code>true</code> if the database services are supported.
     */
    public boolean isDatabaseAvailable() {

        try {
            this.guard.enterShared();

            return (this.dataSources != null);

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Gets database connection directly from the data source. Connection
     * returned by this method is not managed by the application context's
     * transaction management subsystem, as opposed to connection returned by
     * <code>getConnection</code> methods. Usually there is no reason to call
     * this method from the application code unless it is really necessary to
     * manage transactions in some specific way, which is not supported by
     * the framework.
     *
     * @param dataSourceName name of the data source as defined in the
     * <code>cb2app.properties</code> configuration file. If <code>null</code>,
     * "default" is assumed.
     *
     * @return database connection or <code>null</code> if the
     * database access subsystem is turned off for this application
     * context (see <code>isDatabaseAvailable</code> method).
     *
     * @throws SQLException if error happens in the undelying data source
     * object.
     * @throws IllegalArgumentException if there is no data source with the
     * specified name.
     */
    public Connection getNoTXConnection(String dataSourceName)
        throws SQLException {

        if(dataSourceName == null)
            dataSourceName = "default";

        try {
            this.guard.enterShared();

            if(this.dataSources == null)
                return null;

            DataSource ds = (DataSource)this.dataSources.get(dataSourceName);
            if(ds == null)
                throw new IllegalArgumentException("There is no data source named " + dataSourceName);

            ActiveConnection ac = (ActiveConnection)this.actionConnectionPool.get();
            Connection c = ds.getConnection();
            if((this.txIsolationLevel != Connection.TRANSACTION_NONE) &&
               (c.getTransactionIsolation() != this.txIsolationLevel))
                c.setTransactionIsolation(this.txIsolationLevel);
            ac.setWrappedConnection(c);
            return ac;

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Convenience method, which calls
     * <code>getNoTXConnection("default")</code> internally.
     *
     * @return connection to the default database or <code>null</code> if the
     * database access subsystem is turned off for this application context.
     *
     * @throws SQLException if error happens in the undelying data source
     * object.
     * @throws IllegalArgumentException if there is no default data source.
     *
     * @see #getNoTXConnection(String)
     */
    public Connection getNoTXConnection()
        throws SQLException {

        return this.getNoTXConnection("default");
    }


    ///////////////////////////////////////////////////////////////////////////
    // Interface of the Broadcast Messaging subsystem
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Tells if the CB2 Broadcast Messaging subsystem (the BCM) is enabled.
     * The BCM is disabled if no
     * <code>com.boylesoftware.cb2.broadcastMessaging.provider.class</code>
     * application property is provided. If the BCM is disabled, the
     * <code>broadcastBCMMessage</code> method does nothing and returns
     * immediately.
     *
     * @return <code>true</code> if the BCM is enabled and the
     * <code>broadcastBCMMessage</code> method can be used to deliver
     * broadcast messages to the cluster members.
     */
    public boolean isBCMEnabled() {

        try {
            this.guard.enterShared();

            return (this.bcmProvider != null);

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Registers a BCM terminal on the specified channel. Application context
     * will call the terminal's <code>processMessage</code> method each time a
     * message is received on the channel.
     * 
     * <p>If there was already a terminal registered on the channel, it is
     * overwritten by the new one. ?nly one terminal can be subscribed to a
     * channel at a time.
     *
     * @param channelName name of the BCM channel, to which the terminal will
     * be subscribed.
     * @param bcmTerminal the terminal object.
     *
     * @see #broadcastBCMMessage
     */
    public void registerBCMTerminal(String channelName,
                                    BroadcastMessagingTerminal bcmTerminal) {

        synchronized(this.bcmTerminals) {
            this.bcmTerminals.put(channelName, bcmTerminal);
        }
    }

    /**
     * Unregisters the BCM terminal set on the specified channel. Does nothing
     * if no terminal is registered on the channel.
     *
     * @param channelName the BCM channel name.
     */
    public void unregisterBCMTerminal(String channelName) {

        synchronized(this.bcmTerminals) {
            this.bcmTerminals.remove(channelName);
        }
    }

    /**
     * Sends a message to all cluster members through the CB2 Broadcast
     * Messaging subsystem (the BCM). BCM terminals registered on the specified
     * channel in all CB2 application instances that are members of the cluster
     * will receive the message. This includes the local instance, which sends
     * the message. Note, however, that depending on the BCM provider
     * imlementation, it is possible that this method returns
     * <strong>before</strong> the message is received and processed by the BCM
     * terminal on the local instance (and any remote instance as well).
     *
     * @param channelName BCM channel, to which the method sends the message.
     * @param messageParam optional general purpose message parameter, may be
     * <code>null</code>.
     */
    public void broadcastBCMMessage(String channelName,
                                    Serializable messageParam) {

        try {
            this.guard.enterShared();

            if(this.bcmProvider != null) {

                // wait until the provider is initialized
                boolean inited = false;
                do {
                    synchronized(this.bcmProvider) {
                        inited = this.isBCMProviderInitialized;
                    }
                    if(!inited) {
                        try {
                            Thread.sleep(100);
                        } catch(InterruptedException e) { /* ignore it */ }
                    }
                } while(!inited);

                // send the message
                this.bcmProvider.broadcastMessage(channelName, messageParam);
            }

        } finally {
            this.guard.leaveShared();
        }
    }

    /**
     * Processes a received through the CB2 Broadcast Messaging
     * subsystem message by calling <code>processMessage</code> method
     * on the appropriate terminal. This method is a part of SPI, there is
     * no need for the user application code to call this method. Instead,
     * Broadcast Messaging provider implementations call this method when
     * they receive a message to deliver it to the appropriate terminal.
     *
     * @param channelName channel on which the message is received.
     * @param source identifier, specific to the Broadcast Messaging provider
     * implementation, which identifies the JVM that sent the message.
     * @param messageParam optional parameter received with the message, may be
     * <code>null</code>.
     *
     * @throws ApplicationException if an unexpected error happens while
     * processing the message in the terminal. The provider implementation
     * should log the error and continue operation.
     */
    public void processBCMMessage(String channelName,
                                  String source,
                                  Serializable messageParam)
        throws ApplicationException {

        BroadcastMessagingTerminal bcmTerminal;

        synchronized(this.bcmTerminals) {
            bcmTerminal = (BroadcastMessagingTerminal)this.bcmTerminals.get(channelName);
        }

        if(bcmTerminal != null) {
            bcmTerminal.processMessage(channelName, source, messageParam);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Interface for managing active statements
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Registers an active database statement.
     *
     * @param id unique id for the statement.
     * @param astmt reference to the active statement object.
     */
    void registerStatement(Long id, ActiveStatement astmt) {

        if(this.isShuttingDown)
            throw new IllegalStateException("The application is going down.");

        synchronized(this.activeStatements) {
            this.activeStatements.put(id, astmt);
        }
    }

    /**
     * Unregisters a statement registered with <code>registerStatement<code>
     * call. Does nothing if there is no statement with the specified
     * id registered.
     *
     * @param id the statement id returned by the
     * <code>registerStatement</code> call.
     */
    void removeStatement(Long id) {

        synchronized(this.activeStatements) {
            this.activeStatements.remove(id);
        }
    }

    /**
     * Tries to cancel an active database statement. Application context keeps
     * track of all statements executed on any database connection aquired
     * through the application context. List of currently active statements
     * with their ids can be retrieved from the output of <code>getStats</code>
     * method.
     *
     * <p>This method does nothing if there is no statement with the specified
     * id.
     *
     * <p>If an error happens cancelling the statement the method logs the
     * error message to the application log but does not throw any exception.
     * If the statement is cancelled successfully, the thread, which was
     * executing it, will most likely get an exception. The exact behaviour
     * depends on the JDBC driver implementation.
     *
     * @param id the statement's id.
     */
    public void cancelStatement(Long id) {

        ActiveStatement astmt;
        synchronized(this.activeStatements) {
            astmt = (ActiveStatement)this.activeStatements.get(id);
        }

        if(astmt == null)
            return;

        Log log = this.getLog(LOG_SUBSYS);
        log.info("processing request to cancel an active database statement with id " + id);
        try {
            astmt.cancel();
            log.info("active database statement with id " + id + " has been successfully cancelled.");
        } catch(SQLException e) {
            log.error("Could not cancel statement with id " + id + ", ignoring the error", e);
        }
    }

    /**
     * Tries to cancel all currently active database statements.
     *
     * @see #cancelStatement
     */
    public void cancelAllStatements() {

        Log log = this.getLog(LOG_SUBSYS);
        log.info("processing request to cancel all active database statements");
        Map.Entry [] entries;
        synchronized(this.activeStatements) {
            entries = (Map.Entry [])this.activeStatements.entrySet().
                toArray(new Map.Entry[this.activeStatements.size()]);
        }
        for(int i = 0; i < entries.length; i++) {
            try {
                ((ActiveStatement)entries[i].getValue()).cancel();
                log.info("active database statement with id " + entries[i].getKey() + " has been successfully cancelled.");
            } catch(SQLException e) {
                log.error("Could not cancel statement with id " + entries[i].getKey() + ", ignoring the error", e);
            }
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Status XMLs
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Retrieves application context information and returns it as XML.
     * The data inlcudes basic JVM information, application and system
     * properties, class loader chain, contents of JNDI, information on
     * the data sources, active database statements.
     *
     * @return complete application context information XML.
     */
    public byte [] getStats()
        throws Exception {

        ByteArrayOutputStream out = new ByteArrayOutputStream(1024);
        PrintStream res = new PrintStream(out, false, "UTF-8");

        try {
            this.guard.enterShared();

            res.println("<stats>");
            Runtime runtime = Runtime.getRuntime();
            res.println("  <jvmInfo>");
            res.print("    <availProcs>"); res.print(runtime.availableProcessors()); res.println("</availProcs>");
            res.print("    <totalMemory>"); res.print(runtime.totalMemory()); res.println("</totalMemory>");
            res.print("    <freeMemory>"); res.print(runtime.freeMemory()); res.println("</freeMemory>");
            try {
                InetAddress localHost = InetAddress.getLocalHost();
                res.print("    <host>"); res.print(localHost.getHostName()); res.print(" ("); res.print(localHost.getHostAddress()); res.println(")</host>");
            } catch(UnknownHostException e) { /* ignore it */ }
            res.println("  </jvmInfo>");
            res.println("  <impls>");
            javax.xml.parsers.DocumentBuilderFactory dbf = javax.xml.parsers.DocumentBuilderFactory.newInstance();
            res.print("    <dbfClass>"); res.print(dbf.getClass().getName()); res.println("</dbfClass>");
            javax.xml.parsers.DocumentBuilder db = dbf.newDocumentBuilder();
            res.print("    <dbClass>"); res.print(db.getClass().getName()); res.println("</dbClass>");
            org.w3c.dom.DOMImplementation domImpl = db.getDOMImplementation();
            res.print("    <domImplClass>"); res.print(domImpl.getClass().getName()); res.println("</domImplClass>");
            res.println("  </impls>");
            res.println("  <threads>");
            ThreadGroup myThreadGroup = Thread.currentThread().getThreadGroup();
            Thread [] myThreads = new Thread[myThreadGroup.activeCount()];
            try {
                myThreadGroup.enumerate(myThreads, true);
                for(int i = 0; i < myThreads.length; i++) {
                    if(myThreads[i] != null) {
                        res.println("    <thread>");
                        res.print("      <name><![CDATA["); res.print(myThreads[i].getName()); res.println("]]></name>");
                        res.print("      <group><![CDATA["); res.print(myThreads[i].getThreadGroup().getName()); res.println("]]></group>");
                        res.print("      <priority>"); res.print(myThreads[i].getPriority()); res.println("</priority>");
                        res.print("      <daemon>"); res.print(myThreads[i].isDaemon()); res.println("</daemon>");
                        res.println("    </thread>");
                    }
                }
            } catch(SecurityException e) { /* ignore it */ }
            res.println("  </threads>");
            res.println("  <sysProps>");
            java.util.Iterator i;
            try {
                i = System.getProperties().entrySet().iterator();
                while(i.hasNext()) {
                    Map.Entry entry = (Map.Entry)i.next();
                    res.print("    <prop name=\""); res.print((String)entry.getKey()); res.print("\"><![CDATA[");
                    res.print((String)entry.getValue());
                    res.println("]]></prop>");
                }
            } catch(AccessControlException e) {
                // ignore it
            }
            res.println("  </sysProps>");
            res.println("  <appProps>");
            i = this.getProperties().entrySet().iterator();
            while(i.hasNext()) {
                Map.Entry entry = (Map.Entry)i.next();
                res.print("    <prop name=\""); res.print((String)entry.getKey()); res.print("\"><![CDATA[");
                res.print((String)entry.getValue());
                res.println("]]></prop>");
            }
            res.println("  </appProps>");
            res.println("  <clChain>");
            int n = 0;
            for(ClassLoader cl = this.getClass().getClassLoader(); cl != null; cl = cl.getParent()) {
                res.print("    <cl level=\""); res.print(n++); res.println("\">");
                res.print("      <class>"); res.print(cl.getClass().getName()); res.println("</class>");
                res.print("      <stringValue><![CDATA["); res.print(cl.toString()); res.println("]]></stringValue>");
                res.println("    </cl>");
            }
            res.println("  </clChain>");
            res.println("  <jndi>");
            ByteArrayOutputStream out1 = new ByteArrayOutputStream(1024);
            PrintStream res1 = new PrintStream(out1, false, "UTF-8");
            try {
                this.statsDumpJndiContext(res1, "java:comp");
            } catch(javax.naming.NoInitialContextException e) {
                // ignore it and abandon JNDI processing
            } catch(javax.naming.NamingException e) {
                res1.flush();
                if(out1.toByteArray().length == 0)
                    this.statsDumpJndiContext(res1, ""); // try a different approach
                else
                    throw e;
            }
            res1.flush();
            res.write(out1.toByteArray());
            res1.close();
            res.println("  </jndi>");
            if(this.dataSources != null) {
                res.println("  <db>");
                i = this.dataSources.entrySet().iterator();
                while(i.hasNext()) {
                    Map.Entry entry = (Map.Entry)i.next();
                    UniversalDataSource ds = (UniversalDataSource)entry.getValue();
                    res.print("    <dataSource name=\""); res.print((String)entry.getKey()); res.println("\">");
                    Connection con = null;
                    try {
                        boolean conAvail = false;
                        try {
                            con = ds.getConnection();
                            conAvail = true;
                        } catch(SQLException e) { /* ignore */ }

                        if(conAvail) {
                            res.println("      <info>");
                            res.print("        <dsClass>");
                            res.print(ds.getDataSource().getClass().getName());
                            res.println("</dsClass>");
                            res.print("        <conClass>");
                            res.print(con.getClass().getName());
                            res.println("</conClass>");
                            java.sql.DatabaseMetaData dbm = con.getMetaData();
                            res.print("        <dbProduct><![CDATA[");
                            res.print(dbm.getDatabaseProductName());
                            res.println("]]></dbProduct>");
                            res.print("        <dbProductVersion><![CDATA[");
                            res.print(dbm.getDatabaseProductVersion());
                            res.println("]]></dbProductVersion>");
                            res.print("        <driverName><![CDATA[");
                            res.print(dbm.getDriverName());
                            res.println("]]></driverName>");
                            res.print("        <driverVersion><![CDATA[");
                            res.print(dbm.getDriverVersion());
                            res.println("]]></driverVersion>");
                            res.print("        <url><![CDATA[");
                            res.print(dbm.getURL());
                            res.println("]]></url>");
                            res.print("        <user><![CDATA[");
                            res.print(dbm.getUserName());
                            res.println("]]></user>");
                            res.print("        <txIsolation>");
                            res.print(ActiveConnection.getTxIsoLevelName(dbm.getDefaultTransactionIsolation()));
                            res.println("</txIsolation>");
                            res.print("        <extraNameChars><![CDATA[");
                            res.print(dbm.getExtraNameCharacters());
                            res.println("]]></extraNameChars>");
                            res.print("        <idQuoteString><![CDATA[");
                            res.print(dbm.getIdentifierQuoteString());
                            res.println("]]></idQuoteString>");
                            res.println("      </info>");
                        } else {
                            res.println("      <info>");
                            res.println("        <dsClass>n/a</dsClass>");
                            res.println("        <conClass>n/a</conClass>");
                            res.println("        <dbProduct>n/a</dbProduct>");
                            res.println("        <dbProductVersion>n/a</dbProductVersion>");
                            res.println("        <driverName>n/a</driverName>");
                            res.println("        <driverVersion>n/a</driverVersion>");
                            res.println("        <url>n/a</url>");
                            res.println("        <user>n/a</user>");
                            res.println("        <txIsolation>n/a</txIsolation>");
                            res.println("        <extraNameChars>n/a</extraNameChars>");
                            res.println("        <idQuoteString>n/a</idQuoteString>");
                            res.println("      </info>");
                        }

                    } finally {
                        if(con != null) {
                            con.close();
                            con = null;
                        }
                    }
                    res.write(ds.getStats());
                    res.println("    </dataSource>");
                }
                res.println("  </db>");
            }
            res.println("  <activeStmts>");
            synchronized(this.activeStatements) {
                i = this.activeStatements.values().iterator();
                while(i.hasNext()) {
                    ActiveStatement astmt = (ActiveStatement)i.next();
                    res.write(astmt.getStats());
                }
            }
            res.println("  </activeStmts>");
            res.println("</stats>");

        } finally {
            this.guard.leaveShared();
        }

        res.flush();
        return out.toByteArray();
    }

    /**
     * Recursively dumps contents of the specified JNDI context to XML
     * format. Used from the <code>getStats</code> method.
     *
     * @param res output for XML.
     * @param ctxPath path to the JNDI context to dump.
     *
     * @throws NamingException if JNDI error has happened.
     */
    private void statsDumpJndiContext(PrintStream res, String ctxPath)
        throws NamingException {

        java.util.List subContexts = new java.util.ArrayList();
        InitialContext ic = null;
        try {
            ic = new InitialContext();
            javax.naming.Context ctx = (javax.naming.Context)ic.lookup(ctxPath);

            javax.naming.NamingEnumeration en = ctx.listBindings("");
            while(en.hasMore()) {
                javax.naming.Binding b = (javax.naming.Binding)en.next();
                String path = ctxPath + "/" + b.getName();

                res.print("    <binding name=\"");
                res.print(path);
                res.println("\">");
                Object v = b.getObject();
                if(v != null) {
                    res.print("      <class>");
                    res.print(v.getClass().getName());
                    res.println("</class>");
                    res.print("      <stringValue><![CDATA[");
                    res.print(v.toString());
                    res.println("]]></stringValue>");

                    if(v instanceof javax.naming.Context)
                        subContexts.add(path);
                }
                res.println("    </binding>");
            }

        } finally {
            if(this.closeJNDIContexts && (ic != null))
                ic.close();
        }

        java.util.Iterator i = subContexts.iterator();
        while(i.hasNext()) {
            this.statsDumpJndiContext(res, (String)i.next());
        }
    }

    /**
     * Gets event statistics for various events. This method understands the
     * following <code>eventName</code> values:
     *
     * <ul>
     *
     * <li><code>DS_<i>&lt;data source name&gt;</i>_getConnection</code> -
     * statistics for get connection calls on the specified data source.</li>
     *
     * </ul>
     */
    public EventStatistics getEventStats(String eventName) {

        if(eventName.startsWith("DS_")) {
            int lastUnderscoreInd = eventName.lastIndexOf('_');
            if(lastUnderscoreInd == 2)
                throw new IllegalArgumentException("Invalid data source event name [" +
                                                   eventName + "]");
            String dsName = eventName.substring(3, lastUnderscoreInd);
            String dsEventName = eventName.substring(lastUnderscoreInd + 1);
            try {
                this.guard.enterShared();

                UniversalDataSource ds = (UniversalDataSource)this.dataSources.
                    get(dsName);
                if(ds == null)
                    throw new IllegalArgumentException("Invalid data source name [" +
                                                       dsName + "] in the data source event name [" +
                                                       eventName + "]");

                return ds.getEventStats(dsEventName);

            } finally {
                this.guard.leaveShared();
            }
        }

        return null;
    }
}
