package com.boylesoftware.cb2;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;


/**
 * Container for Business Level errors reported by a BLO. If a non-critical
 * error happens in a BLO, it creates a <code>BLErrors</code> object in its
 * business method, populates it with the error codes and possibly error
 * descriptions, and then saves it in the BLO using <code>setErrors</code>
 * protected method. In the business method result the BLO indicates to the
 * caller, in a call-specific way, that there were errors. Given such a result,
 * the caller then calls <code>getLastErrors</code> method on the BLO, which
 * returns last saved <code>BLErrors</code> object with information about the
 * errors and clears the internal BLO variable that stores
 * <code>BLErrors</code> objects, so any consequent call to
 * <code>getLastErrors</code> returns nothing until another business method
 * sets the errors via a <code>setErrors</code> call.
 *
 * <p>Reporting errors using <code>BLErrors</code> objects, as opposed to
 * throwing exceptions, is reserved to errors that are, in fact, normal
 * business situations, rather then unexpected failures of the system. For
 * example, let's assume that there is a BLO that represents some entity stored
 * in the database and the entity has a name, which is unique. The user tries
 * to create a new record in the database and enters a name that is already
 * used. This is a normal business situation, quite expected, so the BLO, when
 * requested to save the new record, does not throw any exception, but sets
 * <code>BLErrors</code> instead and tells the presentation level (the caller
 * of the business method) that it cannot save the record. The presentation
 * level then asks the user to enter a different name and try to save the
 * record again. However, if the record is accepted by the BLO, but during its
 * attempt to actually insert the record into the database an unexpected
 * underlying subsystem error happens, such as database connection is lost and
 * the JDBC driver has thrown an <code>SQLException</code>, then the reaction
 * of the BLO is to throw a <code>BLException</code> to indicate a critical
 * unexpected failure.
 *
 * <p>Errors are identified by error codes and error descriptions are very
 * optional. This is because error messages belong to the presentation level,
 * so it is left on the presentation level implementation to interpret the
 * error code and display correct error message to the user (or do something
 * else to recover after the error), so the business level does not have to
 * worry about how to formulate the error message right and in what language it
 * should be if it is a multilingual application.
 *
 * @author Lev Himmelfarb
 * @version $Id: BLErrors.java,v 1.3 2004/04/16 21:43:27 levahim Exp $
 */
public class BLErrors
    implements java.io.Serializable {

    /**
     * The errors container. Keys are error codes (objects of
     * <code>Integer</code>) and values are error descriptions (objects of
     * <code>String</code>). An error description can be empty string, but
     * never <code>null</code>.
     */
    protected final Map errors;


    /**
     * Creates an empty <code>BLErrors</code> object. Use
     * <code>BLErrors(int initialCapacity)</code> constructor when you know in
     * advance the number of errors you are going to put into the object - it
     * helps to save memory and makes application more efficient.
     */
    public BLErrors() {

        this.errors = new HashMap();
    }

    /**
     * Creates an empty <code>BLErrors</code> object. Use this constructor when
     * you know in advance the number of errors you are going to put into the
     * object.
     *
     * @param initialCapacity number of errors you are going to put into the
     * object.
     */
    public BLErrors(int initialCapacity) {

        this.errors = new HashMap(initialCapacity);
    }


    /**
     * Adds an error to the container. The error description is set to an empty
     * string by this method.
     *
     * @param error the error code.
     */
    public void addError(int error) {

        this.addError(error, "");
    }

    /**
     * Adds an error and its description to the container.
     *
     * @param error the error code.
     * @param description the error description, cannot be <code>null</code>,
     * but can be an empty string.
     */
    public void addError(int error, String description) {

        this.errors.put(new Integer(error), description);
    }

    /**
     * Checks if there is the specified error in this <code>BLErrors</code>
     * instance.
     *
     * @param error the error code.
     *
     * @return <code>true</code> if the specified error is among the errors
     * contained within this instance.
     */
    public boolean containsError(int error) {

        return this.errors.containsKey(new Integer(error));
    }

    /**
     * Gets all errors in this instance as an array or error codes.
     *
     * @return array of error codes.
     */
    public int [] getErrors() {

        int [] res = new int[this.errors.size()];
        Iterator i = this.errors.keySet().iterator();
        int n = 0;
        while(i.hasNext())
            res[n++] = ((Integer)i.next()).intValue();

        return res;
    }

    /**
     * Gets the specified error's description.
     *
     * @param error the error code.
     *
     * @return the specified error's description or <code>null</code>
     * if there is not such error in this instance.
     */
    public String getErrorDescription(int error) {

        return (String)this.errors.get(new Integer(error));
    }

    /**
     * Gets total number of errors in this instance.
     *
     * @return number of errors.
     */
    public int size() {

        return this.errors.size();
    }
}
