package com.boylesoftware.cb2;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Properties;
import java.util.Date;
import java.util.Iterator;
import java.io.Serializable;
import java.io.File;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.util.XMLUtils;
import com.boylesoftware.cb2.util.EventStatistics;
import com.boylesoftware.cb2.util.CruelErrorHandler;
import com.boylesoftware.cb2.util.LocalResourceEntityResolver;
import com.boylesoftware.cb2.util.UniqueIdGenerator;
import com.boylesoftware.cb2.util.ExceptionUtils;


/**
 * Business Level Manager singleton, which represents the Business Level of a
 * CB2-based application. One instance of this class exists in the application,
 * usually created at the application startup, and it is the second level up
 * from the application context in terms of provided services. The BL Manager
 * supports notion of user sessions, represented by BLO containers, and
 * Business Level Objects (BLOs) that populate BLO containers and implement the
 * application's business logic. A dedicated single BLO container is maintained
 * by the BL Manager for shared BLOs - the Shared BLO Container. Also, the
 * BL Manager maintains DAO instances for each configured in the application
 * context data source, which it provides to the BLOs for database operations.
 *
 * <p>Clients of the BL Manager, usually the Presentation Level implementation,
 * have responsibility to manage user sessions within the BL Manager creating
 * and destroying them via <code>createUserSession</code> and
 * <code>destroyUserSession</code> methods. It is also a responsibility of the
 * client, if required by the application, to notify the BL Manager about user
 * authentication events, such as user login and logout, by calling appropriate
 * methods.
 *
 * <p>When the client needs to perform a business level operation, that is to
 * call a business method on a BLO, it first requests the BLO container for
 * the user session by calling <code>getBLOContainer</code> method and passing
 * to it the session id, previously aquired from the
 * <code>createUserSession</code> call. Then, the required BLO is looked up by
 * its deployment name in the BLO container and appropriate business methods
 * are called on it. When the BL Manager returns the BLO container to the
 * <code>getBLOContainer</code> caller, it locks the container, so any
 * consequent call to <code>getBLOContainer</code> with the same session id
 * will block until the BLO container is released by the previous requester by
 * calling <code>releaseBLOContainer</code> method (or <code>release</code>
 * convenience method on the container itself). This mechanism allows to forget
 * about providing access synchronization for the BLOs if used correctly, that
 * is if references to BLOs are not stored somewhere outside BLO container and
 * they are not called without aquiring and locking the BLO container via
 * <code>getBLOContainer</code> first. This does not concern the Shared BLO
 * Container, which never gets locked providing concurring access to the shared
 * BLOs. That is why appropriate access synchronization should be implemented
 * in the BLO level for shared BLOs.
 *
 * <p>The BL Manager is configured by file called <code>blo-config.xml</code>
 * and located in the configuration directory (see application context's
 * <code>getConfigDirPath</code> method). This configuration file contains
 * BLO deployment descriptors and configuration for the DAOs. The DTD can be
 * found at
 * <a href="http://www.cb2project.com/dtd/blo-config_1_0.dtd" target="_blank">http://www.cb2project.com/dtd/blo-config_1_0.dtd</a>.
 *
 * <p>A few application properties control the BL Manager's behaviour:
 *
 * <ul>
 *
 * <li><code><b>com.boylesoftware.cb2.maxWaitingThreads</b></code> -
 * allows to limit maximum number of threads waiting in the
 * <code>getBLOContainer</code> method for the BLO container to get unlocked.
 * If this property is not specified, there is no limit. Otherwise, if a thread
 * tries to get BLO container, which is locked by another thread, and there
 * are already <code>com.boylesoftware.cb2.maxWaitingThreads</code> threads
 * waiting, a <code>TooManyWaitingThreadsException</code> will be thrown by the
 * <code>getBLOConatiner</code> method.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.keepBLSessions</b></code> -
 * allows to disable destroying of BL Manager's state when the BL Manager gets
 * destroyed. Normally, when BL Manager is being destroyed via
 * <code>destroy</code> method call, it explicitely destroys all BLO containers
 * in it and all BLOs in the containers. If this application property is set
 * to "true" it does not happen, which may be useful if you (or your
 * presentation level framework) serializes BLO containers and stores them in a
 * persistent storage to restore user sessions after restart (see
 * <code>restoreUserSession</code> method).</li>
 *
 * </ul>
 *
 * <p>This class logs messages related to its operation under
 * "com.boylesoftware.cb2.BLOMGR" log channel.
 *
 * @author Lev Himmelfarb
 * @version $Id: BLManager.java,v 1.12 2004/10/11 18:17:36 levahim Exp $
 */
public final class BLManager
    implements StatusProvider {

    /**
     * Subsystem name for the logger.
     */
    private final static String LOG_SUBSYS = "com.boylesoftware.cb2.BLOMGR";

    /**
     * Config dir relative path to the file containing BLO descriptors
     * and the DAO configuration. During initialization this file is
     * parsed and used for configuration.
     */
    private final static String BLO_CONFIG_FILE = "/blo-config.xml";

    /**
     * List of public ids and local DTDs for the configuration XML
     * files.
     */
    private final static String [] LOCAL_DTDS = new String[] {
        "-//Boyle Software, Inc.//DTD CB2 Common Entities 1.0//EN",
        "/com/boylesoftware/cb2/common-entities_1_0.dtd",
        "-//Boyle Software, Inc.//DTD CB2 Common DAO Configuration Elements 1.0//EN",
        "/com/boylesoftware/cb2/common-dao-config_1_0.dtd",
        "-//Boyle Software, Inc.//DTD CB2 Business Level Configuration 1.0//EN",
        "/com/boylesoftware/cb2/blo-config_1_0.dtd",
        "-//Boyle Software, Inc.//DTD CB2 DAO External Configuration 1.0//EN",
        "/com/boylesoftware/cb2/dao-config_1_0.dtd" };


    ///////////////////////////////////////////////////////////////////////////
    // Internal helper classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Represents a single descriptor of a BLO. The
     * <code>bloDescriptors</code> member variable hold instances of
     * this class for all the registered BLOs. It gets populated
     * during the <code>blo-config.xml</code> configuration file
     * initial parsing.
     */
    static class BLODescriptor
        implements Serializable {

        /**
         * The BLO's name.
         */
        public final String name;

        /**
         * BLO's implementation class.
         */
        public final Class bloClass;

        /**
         * True if the BLO is a shared BLO.
         */
        public final boolean isShared;

        /**
         * The BLO's initialization parameters.
         */
        public final Properties initParams;

        /**
         * Statistics on lookups of this BLO (<code>getBLO</code>
         * method calls).
         */
        public final EventStatistics requestStats;


        /**
         * Constructs a descriptor from a configuration XML file
         * element.
         */
        BLODescriptor(Element bloElement)
            throws ApplicationException {

            this.isShared = (new Boolean(bloElement.getAttribute("shared"))).booleanValue();
            this.name = bloElement.getAttribute("name");
            String bloClassName = XMLUtils.getXMLSubelementText(bloElement, "class");
            if((this.name == null) || (bloClassName == null))
                throw new ApplicationException("Invalid syntax in the BLO configuration file.");
            try {
                this.bloClass = Class.forName(bloClassName);
            } catch(ClassNotFoundException e) {
                throw new ApplicationException("Could not load BLO class for [" + this.name + "].", e);
            }
            this.initParams = new Properties();
            NodeList bloParams = bloElement.getElementsByTagName("init-param");
            for(int i = 0; i < bloParams.getLength(); i++) {
                Element paramElement = (Element)bloParams.item(i);
                this.initParams.setProperty(XMLUtils.getXMLSubelementText(paramElement, "param-name"),
                                            XMLUtils.getXMLSubelementText(paramElement, "param-value"));
            }
            this.requestStats = new EventStatistics();
        }
    }


    /**
     * Wrapper around a BLO container stored in a session. Basically
     * it is the representation of a user session.
     *
     * <p>This class adds a BLO container lock flag. The
     * <code>BLManager</code> stores BLO containers in the sessions
     * storage wrapped in this class which allows
     * <code>BLOContainer</code> locking.
     */
    private class BLOContainerWrapper {

        /**
         * The BLO container itself.
         */
        private final BLOContainer bloContainer;

        /**
         * Maximum number of threads allowed to wait for a BLO container
         * instance.
         */
        private final int maxWaitingThreads;

        /**
         * True if the BLO container is locked, i.e. assigned to some
         * client through the getBLOContainer method of BLManager.
         */
        private boolean isBLOContainerLocked;

        /**
         * Number of threads currently locked on the container waiting
         * for its release.
         */
        private int nWaitingThreads;

        /**
         * Timestamp when the session was started.
         */
        private final Date startedOn;

        /**
         * Set to <code>true</code> if the container has been
         * destroyed.
         */
        private boolean isDestroyed;


        /**
         * Constructs a wrapper for the specified BLO
         * container. Initially the BLO container is not locked.
         */
        public BLOContainerWrapper(BLOContainer bloContainer,
                                   ApplicationContext appCtx)
            throws BLException {

            this.bloContainer = bloContainer;
            this.isBLOContainerLocked = false;
            this.nWaitingThreads = 0;
            this.startedOn = new Date();
            this.isDestroyed = false;

            String maxWaitingThreadsS = appCtx.getProperty("com.boylesoftware.cb2.maxWaitingThreads");
            if(maxWaitingThreadsS != null) {
                try {
                    this.maxWaitingThreads = Integer.parseInt(maxWaitingThreadsS);
                } catch(NumberFormatException e) {
                    throw new BLException("Property com.boylesoftware.cb2.maxWaitingThreads " +
                                          "must be an integer number.", e);
                }
            } else {
                this.maxWaitingThreads = -1;
            }
        }

        /**
         * Returns the BLO container wrapped by this wrapper.
         */
        public BLOContainer getBLOContainer() {

            return this.bloContainer;
        }

        /**
         * Returns <code>true</code> if the BLO container wrapped by
         * this wrapper is locked.
         */
        public boolean isBLOContainerLocked() {

            return this.isBLOContainerLocked;
        }

        /**
         * Sets the <code>isBLOContainerLocked</code> flag.
         */
        public void setIsBLOContainerLocked(boolean isBLOContainerLocked) {

            this.isBLOContainerLocked = isBLOContainerLocked;
        }

        /**
         * Registers a thread waiting for BLO container.
         *
         * @throws TooManyWaitingThreadsException if the limit is exceeded.
         */
        public void incrementWaitingThreads()
            throws TooManyWaitingThreadsException {

            if((this.maxWaitingThreads > 0) &&
               (this.nWaitingThreads >= this.maxWaitingThreads))
                throw new TooManyWaitingThreadsException();

            this.nWaitingThreads++;
        }

        /**
         * Unregisters a waiting thread.
         */
        public void decrementWaitingThreads() {

            this.nWaitingThreads--;
        }

        /**
         * Returns the number of threads that are waiting for this BLO
         * container to get released.
         *
         * @return the number of waiting threads.
         */
        public int getNWaitingThreads() {

            return this.nWaitingThreads;
        }

        /**
         * Returns timestamp when this session was started.
         */
        public Date getStartedOn() {

            return this.startedOn;
        }

        /**
         * Returns <code>true</code> if the container has been
         * destroyed.
         */
        public boolean isDestroyed() {

            return this.isDestroyed;
        }

        /**
         * Sets the <code>isDestroyed</code> flag value.
         */
        public void setIsDestroyed(boolean isDestroyed) {

            this.isDestroyed = isDestroyed;
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // BLManager's internal data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Our application context.
     */
    private final ApplicationContext appCtx;

    /**
     * All registered BLOs' descriptors read from the configuration
     * file in the order they are defined in it.
     */
    private final BLODescriptor [] bloDescriptors;

    /**
     * Collection of DAO configuration instances. The keys are data
     * source names (objects of <code>String</code>), the values are
     * config instances (objects of <code>DAOConfig</code>).
     */
    private final Map daoConfigs;

    /**
     * Collection of database access object instances. The keys are
     * data source names (objects of <code>String</code>), the values
     * are DAO instances (objects of <code>DAO</code>).
     */
    private final Map daos;

    /**
     * This map contains all active user sessions and associated with
     * them BLO containers. The keys are user session ids (objects of
     * <code>Object</code>) and the values are BLO container wrappers
     * (objects of <code>BLManager.BLOContainerWrapper</code>)
     * containing <code>BLOContainer</code> instances.
     */
    private final Map userSessions;

    /**
     * This is the single BLO container for shared BLOs.
     */
    private final BLOContainer sharedBLOContainer;

    /**
     * XML document builder to be used for parsing configuration
     * files.
     */
    private final DocumentBuilder documentBuilder;

    /**
     * The BLManager's log.
     */
    private final Log log;

    /**
     * Session id generator.
     */
    private final UniqueIdGenerator sessionIdGenerator;


    ///////////////////////////////////////////////////////////////////////////
    // BLO subsystem initialization and internal service
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Creates <code>BLManager</code> and initializes the BLO
     * subsystem. There should be only one <code>BLManager</code>
     * instance in the application. Usually an instance of
     * <code>BLManager</code> created once during the application
     * startup just after the application context is created and
     * intialized.
     *
     * @param appCtx the application context.
     *
     * @throws ApplicationException the BLO subsystem cannot be initialized and
     * is unavailable.
     */
    public BLManager(ApplicationContext appCtx)
        throws ApplicationException {

        this.appCtx = appCtx;

        // get the log
        this.log = appCtx.getLog(LOG_SUBSYS);

        // tell everybody we are getting up
        this.log.info("initializing BLO subsystem");

        // create the document builder
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setIgnoringComments(true);
            dbf.setValidating(true);
            this.documentBuilder = dbf.newDocumentBuilder();
            this.documentBuilder.setEntityResolver(new LocalResourceEntityResolver(this.getClass(), LOCAL_DTDS));
            this.documentBuilder.setErrorHandler(new CruelErrorHandler());
        } catch(ParserConfigurationException e) {
            this.log.fatal("Could not create XML parser.", e);
            throw new ApplicationException("Could not create XML parser.", e);
        }

        // find and parse the Business Level configuration file
        Document bloConfig = this.parseBLOConfigFile();

        // load BLO descriptors
        this.log.debug("loading BLO descriptors");
        List bloDescriptorsList = new ArrayList();
        Set addedNames = new HashSet();
        NodeList bloElements = bloConfig.getDocumentElement().getElementsByTagName("blo");
        for(int i = 0; i < bloElements.getLength(); i++) {
            BLODescriptor bloDescriptor = new BLODescriptor((Element)bloElements.item(i));
            if(!addedNames.add(bloDescriptor.name)) {
                this.log.fatal("There are two or more BLOs named \"" + bloDescriptor.name +
                                "\" defined in the blo-config.xml.");
                throw new ApplicationException("There are two or more BLOs named \"" + bloDescriptor.name +
                                               "\" defined in the blo-config.xml.");
            }
            bloDescriptorsList.add(bloDescriptor);
        }
        this.bloDescriptors = (BLODescriptor [])bloDescriptorsList.
            toArray(new BLODescriptor[bloDescriptorsList.size()]);

        // initialize the DAL
        this.daoConfigs = new HashMap();
        this.daos = new HashMap();
        this.reconfigureDAL(bloConfig);

        // create and initialize the shared BLO container
        this.log.debug("initializing the shared BLO container");
        this.sharedBLOContainer = new BLOContainer(this, null);
        this.sharedBLOContainer.init();

        // initialize user sessions container
        this.sessionIdGenerator = new UniqueIdGenerator();
        this.userSessions = new HashMap();

        // all done
        this.log.debug("the BLO subsystem has been successfully initialized.");
    }

    /**
     * Parses the <code>blo-config.xml</code> file.
     *
     * @return parsed document built from the configuration file.
     *
     * @throws BLException the configuration file cannot be parsed.
     */
    private Document parseBLOConfigFile()
        throws BLException {

        try {
            String configFilePath = this.appCtx.getConfigDirPath() + BLO_CONFIG_FILE;

            this.log.debug("parsing BLO configuration file at " + configFilePath);
            return this.documentBuilder.parse(new File(configFilePath));

        } catch(SAXException e) {
            this.log.fatal("Could not parse the BLO configuration file.", e);
            throw new BLException("Could not parse the BLO configuration file.", e);
        } catch(IOException e) {
            this.log.fatal("Could not read the BLO configuration file.", e);
            throw new BLException("Could not read the BLO configuration file.", e);
        }
    }

    /**
     * Re/initializes the database access level. The BLO configuration
     * file is reread and new DAOs along with DAO config objects are
     * created and configured for each data source.
     *
     * @param bloConfig a preparsed <code>blo-config.xml</code>
     * document or <code>null</code> if you want the method to find
     * and parse the file itself.
     *
     * @throws BLException the DAL cannot be reconfigured, the
     * application is probably left in an unstable state.
     */
    public void reconfigureDAL(Document bloConfig)
        throws BLException {

        this.log.debug("loading DAL configuration");

        // parse BLO config file if not provided
        if(bloConfig == null)
            bloConfig = this.parseBLOConfigFile();

        // load DAO configurations
         synchronized(this.daoConfigs) {
            this.daoConfigs.clear();

            NodeList daoElements = bloConfig.getDocumentElement().getElementsByTagName("dao-config");
            for(int i = 0; i < daoElements.getLength(); i++) {
                Element daoConfigElement = (Element)daoElements.item(i);
                String dataSourceName = daoConfigElement.getAttribute("dataSource");
                if(dataSourceName.length() == 0)
                    dataSourceName = "default";

                // check for duplicates
                if(this.daoConfigs.containsKey(dataSourceName)) {
                    this.log.fatal("There are two or more dao-config elements for data source \"" + dataSourceName +
                                   "\" in the blo-config.xml file.");
                    throw new BLException("There are two or more dao-config elements for data source \"" + dataSourceName +
                                          "\" in the blo-config.xml file.");
                }

                List daoConfigElements = new ArrayList();

                // add main DAO config entry
                daoConfigElements.add(daoConfigElement);

                // external DAO config files
                try {

                    NodeList elements = daoConfigElement.getElementsByTagName("external-dao-config");
                    for(int j = 0; j < elements.getLength(); j++) {
                        Element element = (Element)elements.item(j);
                        String fileName = element.getAttribute("file");

                        if(!fileName.startsWith("/"))
                            fileName = this.appCtx.getConfigDirPath() + "/" + fileName;

                        this.log.info("parsing external DAO configuration file at " + fileName);
                        daoConfigElement = this.documentBuilder.parse(new File(fileName)).getDocumentElement();
                        daoConfigElements.add(daoConfigElement);
                    }

                } catch(SAXException e) {
                    this.log.fatal("Could not parse external DAO configuration file.", e);
                    throw new BLException("Could not parse external DAO configuration file.", e);
                } catch(IOException e) {
                    this.log.fatal("Could not read external DAO configuration file.", e);
                    throw new BLException("Could not read external DAO configuration file.", e);
                }

                // create the DAO config instance
                this.log.debug("instantiating the DAO config for " + dataSourceName + " data source");
                this.daoConfigs.put(dataSourceName,
                                    new DAOConfig(this.appCtx,
                                                  dataSourceName,
                                                  (Element [])daoConfigElements.
                                                  toArray(new Element[daoConfigElements.size()])));
            }
        }

        // create DAO instances
         synchronized(this.daos) {
            this.daos.clear();

            synchronized(this.daoConfigs) {

                Iterator i = this.daoConfigs.values().iterator();
                while(i.hasNext()) {
                    DAOConfig config = (DAOConfig)i.next();

                    // create the DAO instance
                    this.log.debug("instantiating the DAO for " + config.getDataSourceName() + " data source");
                    this.daos.put(config.getDataSourceName(),
                                  new DAO(config, true));
                }
            }
        }
    }

    /**
     * Destroys the BLO subsystem. This method is usually called when
     * the application is shutting down.
     *
     * <p>Unless <code>com.boylesoftware.cb2.keepBLSessions</code> application
     * property is set to "true", calling this method causes the BL Manager to
     * explicitely destroy all BLO containers within it, which, in turn, causes
     * the BLO containers to explicitely destroy all BLOs contained within them
     * by calling their <code>destroy</code> methods.
     *
     * <p>If at the moment of this call threads exist that are waiting for BLO
     * containers in <code>getBLOContainer</code> method, they will get a
     * <code>NoBLOContainerException</code>. Basically, it is a good practice
     * to write the presentation level in a way that there is a guarantee that
     * no BLO containers are being used at the moment of the
     * <code>destroy</code> call.
     */
    public void destroy() {

        this.log.info("the BLO subsystem is shutting down");

        boolean destroyBLSessions = !Boolean.
            valueOf(this.appCtx.getProperty("com.boylesoftware.cb2.keepBLSessions")).booleanValue();

        if(!destroyBLSessions)
            this.log.info("keeping BL sessions and not destrying them, as specified by com.boylesoftware.cb2.keepBLSessions property");

        synchronized(this.userSessions) {
            if(destroyBLSessions) {
                Iterator i = this.userSessions.values().iterator();
                while(i.hasNext()) {
                    ((BLOContainerWrapper)i.next()).getBLOContainer().destroy();
                }
            }

            this.userSessions.clear();
        }

        if(destroyBLSessions)
            this.sharedBLOContainer.destroy();

        for(int i = 0; i < this.bloDescriptors.length; i++)
            this.bloDescriptors[i] = null;

        this.daoConfigs.clear();
        this.daos.clear();

        this.log.debug("the BLO subsystem has been shut down.");
    }

    /**
     * Gets a shared BLO from the Shared BLO Container.
     *
     * <p>All BLO containers perform lookup in the Shared BLO Container if
     * there is no requested BLO in them, so it is a relatively rare case when
     * this method needs to be called directly. The need may arise when you
     * don't have any context of current user session, or for some reason don't
     * want to lock the session's BLO container, while all you need is a shared
     * BLO. Then it can be aquired via this method.
     *
     * @param bloName case-insensitive deployment name of a shared BLO.
     *
     * @return reference to the shared BLO or <code>null</code> if there is no
     * BLO under the specified name in the Shared BLO Container.
     */
    public BLObject getSharedBLO(String bloName) {

        return this.sharedBLOContainer.getBLO(bloName);
    }

    /**
     * Checks if the specified by name BLO exists in the shared BLO
     * container. The BLO names are case-insensitive.
     *
     * @param bloName the requested BLO's deployment name.
     *
     * @return <code>true</code> if the specified BLO is present in
     * the shared BLO container.
     */
    boolean hasSharedBLO(String bloName) {

        return this.sharedBLOContainer.hasBLO(bloName);
    }

    /**
     * Gets application context to which this BL Manager belongs.
     *
     * @return reference to the application context.
     */
    public ApplicationContext getApplicationContext() {

        return this.appCtx;
    }

    /**
     * Returns all BLO descriptors.
     *
     * @return array of all registered BLOs' descriptors in the order
     * they are defined in the configuration file.
     */
    BLODescriptor [] getBLODescriptors() {

        return this.bloDescriptors;
    }


    ///////////////////////////////////////////////////////////////////////////
    // BLO subsystem service interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Binds the specified BLO container instance with the specified session
     * id effectively creating a new session.
     *
     * <p>This method can be used to restore a session from a serialized BLO
     * container. It also internally "activates" the specified BLO container
     * instance, which causes <code>activate</code> methods to be called on the
     * BLOs in the container.
     *
     * @param sessionId session id, to which to bind the container. Must not
     * already exist, or a <code>BLException</code> is thrown.
     * @param bloContainer BLO container instance restored from the serialized
     * form (transient fields are not set).
     *
     * @throws BLException if the session cannot be restored.
     */
    public void restoreUserSession(String sessionId, BLOContainer bloContainer)
        throws BLException {

        int nSessions;
        synchronized(this.userSessions) {

            if(this.userSessions.containsKey(sessionId))
                throw new BLException("There is already a session with id [" + sessionId +
                                      "] registered in the BL Manager.");

            BLOContainerWrapper wrapper = new BLOContainerWrapper(bloContainer, this.appCtx);
            bloContainer.activate(this);
            this.userSessions.put(sessionId, wrapper);

            nSessions = this.userSessions.size();
        }

        this.log.debug("BLO container for user session [" + sessionId +
                       "] has been restored from serialized form, " +
                       "now there are " + nSessions + " active sessions");
    }

    /**
     * Creates and initializes a new user session with BLO container and all
     * the BLOs. When a session gets created the BLOs are instantiated, or
     * borrowed from the pool, and initialized through calling their
     * <code>init</code> methods.
     *
     * @return the new session's id.
     *
     * @throws BLException could not initialize the BLO container or one of the
     * BLOs (a BLO's <code>init</code> method has thrown an exception).
     */
    public String createUserSession()
        throws BLException {

        int nSessions;
        String sessionId = this.sessionIdGenerator.next();
        synchronized(this.userSessions) {

            BLOContainer bloContainer = new BLOContainer(this, sessionId);
            BLOContainerWrapper wrapper = new BLOContainerWrapper(bloContainer, this.appCtx);
            bloContainer.init();
            this.userSessions.put(sessionId, wrapper);

            nSessions = this.userSessions.size();
        }

        this.log.debug("BLO container for user session [" + sessionId +
                       "] has been initialized, " +
                       "now there are " + nSessions + " active sessions");

        return sessionId;
    }

    /**
     * Destroyes a user session including its BLO container and all the BLOs in
     * it. When a session gets destroyed, the BLOs are notified through calling
     * their <code>destroy</code> methods.
     *
     * <p>Note, that if the BLO container is locked at the moment of the call,
     * the method blocks and waits indefinitely until the BLO container gets
     * unlocked by another thread.
     *
     * @param sessionId id of the session to destroy.
     * 
     * @return <code>true</code> if the session has been successfully
     * destroyed by this call, <code>false</code> if no session with
     * the specified id exists or the session is already being destroyed by
     * another call from a different thread.
     */
    public boolean destroyUserSession(String sessionId) {

        BLOContainerWrapper bloContainerWrapper;
        synchronized(this.userSessions) {
            bloContainerWrapper = (BLOContainerWrapper)this.userSessions.get(sessionId);
        }

        if(bloContainerWrapper == null) {
            this.log.warn("Called destroyUserSession for session " +
                          sessionId + ", which does not exist.");
            return false;
        }

        synchronized(bloContainerWrapper) {

            if(bloContainerWrapper.isDestroyed()) {
                this.log.warn("Called destroyUserSession for session " +
                              sessionId + ", which is already being destroyed.");
                return false;
            }

            if(bloContainerWrapper.isBLOContainerLocked())
                try {
                    bloContainerWrapper.wait();
                } catch(InterruptedException e) {
                    // never can happen
                }

            bloContainerWrapper.getBLOContainer().destroy();

            bloContainerWrapper.setIsDestroyed(true);

            bloContainerWrapper.setIsBLOContainerLocked(false);

            bloContainerWrapper.notifyAll();
        }

        int nSessions;
        synchronized(this.userSessions) {
            this.userSessions.remove(sessionId);
            nSessions = this.userSessions.size();
        }

        this.log.debug("User session [" + sessionId + "] has been destroyed, " +
                       nSessions + " active sessions left");

        return true;
    }

    /**
     * Gets BLO container for the specified user session.
     *
     * <p>When this method returns a BLO container it locks it, meaning that
     * any subsequent <code>getBLOContainer</code> call for the same session
     * will block and wait indefinitely until the BLO container is released by
     * the previous requester through calling <code>releaseBLOContainer</code>
     * method (or <code>release</code> convenience method on the BLO container
     * itself). We suggest having an invocation of <code>release</code> method
     * in a <code>finally</code> block of your client code so there is a
     * guarantee that the BLO container always gets unlocked.
     *
     * <p>Never keep references to BLOs that are got from a BLO container after
     * the BLO container is released, nor any references to the BLO container
     * itself - always reaquire the BLO container when you need any service
     * provided by the BLOs.
     *
     * @param sessionId id of the user session, for which we are requesting the
     * BLO container.
     *
     * @return reference to the BLO container corresponding the session.
     *
     * @throws NoBLOContainerException if there is no BLO container for the
     * requested session. Note, that the session can be destroyed by another
     * thread during this method call. If that happens, this exception is
     * thrown with an appropriate message.
     * @throws TooManyWaitingThreadsException if the BLO container is locked
     * and there are already too many threads waiting for its release. The
     * maximum number of waiting threads is specified by
     * <code>com.boylesoftware.cb2.maxWaitingThreads</code> application
     * property. If it is not specified, the number of waiting threads is not
     * limited.
     * @throws BLException if an error happens in a BLO implementing
     * <code>BLOContainerStateListener</code> interface and it throws an
     * exception.
     *
     * @see #releaseBLOContainer(String)
     * @see BLOContainer#release()
     */
    public BLOContainer getBLOContainer(String sessionId)
        throws NoBLOContainerException, TooManyWaitingThreadsException, BLException {

        BLOContainerWrapper bloContainerWrapper;
        synchronized(this.userSessions) {
            bloContainerWrapper = (BLOContainerWrapper)this.userSessions.get(sessionId);
        }

        if(bloContainerWrapper == null)
            throw new NoBLOContainerException("There is no BLO container instance for the session.");

        synchronized(bloContainerWrapper) {

            if(bloContainerWrapper.isBLOContainerLocked()) {
                bloContainerWrapper.incrementWaitingThreads();
                try {
                    bloContainerWrapper.wait();
                } catch(InterruptedException e) {
                    // never can happen
                }
                bloContainerWrapper.decrementWaitingThreads();
            }

            if(bloContainerWrapper.isDestroyed())
                throw new NoBLOContainerException("The BLO container has been destroyed while waiting for the lock.");

            bloContainerWrapper.setIsBLOContainerLocked(true);
        }

        this.log.debug("locked BLO container for session " + sessionId.toString());

        BLOContainer bloContainer = bloContainerWrapper.getBLOContainer();
        try {
            bloContainer.processBLOContainerLock();
        } catch(Throwable e) {
            try {
                bloContainer.processBLOContainerUnlock();
            } catch(Throwable e1) { // skip it
                this.log.error("Got error while notifying BLOs about BLO conatiner unlock" +
                               " after unsuccessful attempt to lock it, this exception will" +
                               " not be thrown, the original exception caused lock failure" +
                               " will be thrown instead.", e1);
            } finally {
                synchronized(bloContainerWrapper) {
                    bloContainerWrapper.setIsBLOContainerLocked(false);
                    bloContainerWrapper.notify();
                }
                ExceptionUtils.
                    rethrowBLException(e, "Got error notifying BLOs about BLO container lock.");
            }
        }

        return bloContainer;
    }

    /**
     * Releases BLO container that was previously locked by a
     * <code>getBLOContainer</code> call. After the BLO conatiner has been
     * unlocked by this call, it can be reaquired again by another
     * <code>getBLOContainer</code> mathod invocation.
     *
     * <p>Note, that it is usually more convenient to call
     * <code>release</code> method on the BLO container itself, which calls
     * this method internally, but does not require session id.
     *
     * @param sessionId id of the user session, whose BLO container we are
     * releasing.
     *
     * @throws BLException if an error happens in a BLO implementing
     * <code>BLOContainerStateListener</code> interface and it throws an
     * exception. Note, that the BLO container gets unlocked anyway.
     * @throws IllegalArgumentException if there is no session with the
     * specified session id.
     *
     * @see #getBLOContainer(String)
     * @see BLOContainer#release()
     */
    public void releaseBLOContainer(String sessionId)
        throws BLException {

        BLOContainerWrapper bloContainerWrapper;
        synchronized(this.userSessions) {
            bloContainerWrapper = (BLOContainerWrapper)this.userSessions.get(sessionId);
        }

        if(bloContainerWrapper == null)
            throw new IllegalArgumentException("There is no BLO container instance for the session.");

        boolean unlocked = false;
        try {
            synchronized(bloContainerWrapper) {
                if(bloContainerWrapper.isBLOContainerLocked()) {
                    try {
                        bloContainerWrapper.getBLOContainer().
                            processBLOContainerUnlock();
                    } catch(Throwable e) {
                        ExceptionUtils.
                            rethrowBLException(e, "Got error notifying BLOs about BLO container unlock, the BLO container will be unlocked anyway.");
                    } finally {
                        bloContainerWrapper.setIsBLOContainerLocked(false);
                        bloContainerWrapper.notify();
                        unlocked = true;
                    }
                }
            }
        } finally {
            // note that this message may appear in the log AFTER a message
            // about successful lock waited for this unlock
            if(unlocked)
                this.log.debug("unlocked BLO container for session " + sessionId.toString());
        }
    }

    /**
     * Returns the DAO config instance for the specified data source.
     *
     * @param dataSourceName name of the data source.
     *
     * @return reference to the DAO config.
     */
    DAOConfig getDAOConfig(String dataSourceName) {

        synchronized(this.daoConfigs) {
            return (DAOConfig)this.daoConfigs.get(dataSourceName);
        }
    }

    /**
     * Returns the DAO instance for the specified data source.
     *
     * @param dataSourceName name of the data source.
     *
     * @return reference to the DAO.
     */
    DAO getDAO(String dataSourceName) {

        synchronized(this.daos) {
            return (DAO)this.daos.get(dataSourceName);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Status XMLs
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Retrieves information about the Business Level as an XML text.
     *
     * @return XML for the general BL info.
     */
    public byte [] getStats()
        throws Exception {

        ByteArrayOutputStream out = new ByteArrayOutputStream(8192);
        PrintStream res = new PrintStream(out, false, "UTF-8");

        res.println("<stats>");
        res.println("  <blos>");
        for(int i = 0; i < this.bloDescriptors.length; i++) {
            BLODescriptor desc = this.bloDescriptors[i];
            res.print("    <blo name=\""); res.print(desc.name); res.println("\">");
            res.print("      <class>"); res.print(desc.bloClass.getName()); res.println("</class>");
            res.print("      <shared>"); res.print(desc.isShared); res.println("</shared>");
            res.print("      <initParams>"); res.print(desc.initParams.toString()); res.println("</initParams>");
            res.print("      <nRequests>"); res.print(desc.requestStats.getNTimes()); res.println("</nRequests>");
            res.print("      <reqsPerMinute>"); res.print(desc.requestStats.getEventsPerMinuteAvg()); res.println("</reqsPerMinute>");
            res.println("    </blo>");
        }
        res.println("  </blos>");
        res.println("  <sessions>");
        synchronized(this.userSessions) {
            Iterator i = this.userSessions.entrySet().iterator();
            while(i.hasNext()) {
                Map.Entry entry = (Map.Entry)i.next();
                res.print("    <session id=\""); res.print((String)entry.getKey()); res.println("\">");
                res.print("      <startedOn>");
                res.print(StatusProvider.TIMESTAMP_FORMAT.format(((BLOContainerWrapper)entry.getValue()).getStartedOn()));
                res.println("</startedOn>");
                res.print("      <nWaitingThreads>");
                res.print(((BLOContainerWrapper)entry.getValue()).getNWaitingThreads());
                res.println("</nWaitingThreads>");
                res.println("    </session>");
            }
        }
        res.println("  </sessions>");
        synchronized(this.daos) {
            if((this.daos != null) && (this.daos.size() > 0)) {
                res.println("  <daos>");
                Iterator i = this.daos.entrySet().iterator();
                while(i.hasNext()) {
                    Map.Entry entry = (Map.Entry)i.next();
                    DAO dao = (DAO)entry.getValue();
                    res.print("    <dao name=\""); res.print((String)entry.getKey()); res.println("\">");
                    res.write(dao.getStats());
                    res.println("    </dao>");
                }
                res.println("  </daos>");
            }
        }
        res.println("</stats>");

        res.flush();
        return out.toByteArray();
    }

    /**
     * Not used in this class, always returns <code>null</code>.
     */
    public EventStatistics getEventStats(String eventName) {

        return null;
    }
}
