package com.boylesoftware.cb2;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.security.Principal;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.util.ExceptionUtils;


/**
 * As its name suggests, BLO container contains BLOs. BLOs never exist out of
 * BLO container context and the only way to access a BLO is to find first the
 * BLO container and then lookup the BLO in it by the BLO's deployment name.
 * The BL Manager creates, maintains and destroys BLO containers as user
 * sessions are created and destroyed. Since there is always one dedicated BLO
 * container corresponding to every user session, in a sense, a BLO container
 * is a representation of a user session in the Business Level. There is one
 * exception to that though - the Shared BLO Container, which is an instance of
 * <code>BLOContainer</code> class which contains shared BLOs and does not
 * correspond to any particular user session. The instance of the Shared BLO
 * Container is maintained by the BL Manager as well.
 *
 * <p>The only way to get a BLO container for the application code is through
 * the BL Manager's <code>getBLOContainer</code> method. The
 * <code>getBLOContainer</code> method also locks the BLO container so no other
 * thread can access it until it is release by the current owner via
 * <code>release</code> method call (or BL Manager's
 * <code>releaseBLOContainer</code>). Because of that, all methods of this
 * class are absolutely <strong>not</strong> thread-safe, counting on
 * proper usage, which is never store references to BLO containers in your
 * application. Instead, you should always call BLO container's methods between
 * <code>getBLOContainer</code> and <code>release</code> calls, that is while
 * it is locked.
 *
 * <p>This class logs messages related to its operation under
 * "com.boylesoftware.cb2.BLOCNR" log channel.
 *
 * @author Lev Himmelfarb
 * @version $Id: BLOContainer.java,v 1.7 2004/04/23 17:49:55 levahim Exp $
 *
 * @see BLManager
 * @see BLObject
 */
public final class BLOContainer
    implements java.io.Serializable {

    /**
     * Subsystem name for the logger.
     */
    private final static String LOG_SUBSYS = "com.boylesoftware.cb2.BLOCNR";


    /**
     * Reference to the <code>BLManager</code> that created and
     * manages this container.
     */
    private transient BLManager blManager;

    /**
     * Id of the user session this BLO container instance belongs
     * to. Is <code>null</code> for the shared BLO container.
     */
    private final String sessionId;

    /**
     * This map containes instances of all BLOs in this
     * container. Keys are BLO deployment names (objects of
     * <code>String</code>) and values are instances of BLOs (objects
     * of <code>BLObject</code>). The stored names are all in upper
     * case.
     */
    private final Map blos;

    /**
     * Currently authenticated user for the BLO container or
     * <code>null</code> if the container is in the
     * "non-authenticated" state.
     */
    private Principal userPrincipal;


    ///////////////////////////////////////////////////////////////////////////
    // Internal interface for BLO subsystem
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructs and initializes a new instance of BLO container. It
     * also creates and initializes all BLOs and stores them
     * internally. This constructor is called from the
     * <code>BLManager</code>'s <code>createUserSession()</code>
     * method. If the <code>sessionId</code> argument is
     * <code>null</code> it means that it is the BLO container for
     * shared BLOs and is not associated with any particular user
     * session.
     *
     * <p>Note that new BLO container is unusable until <code>init</code>
     * is called.
     *
     * @param blManager the parent <code>BLManager</code>.
     * @param sessionId the user session id this BLO container is
     * associated with or <code>null</code> for shared BLO container.
     */
    BLOContainer(BLManager blManager, String sessionId) {

        this.blManager = blManager;
        this.sessionId = sessionId;
        this.userPrincipal = null;
        this.blos = new HashMap();
    }

    /**
     * Initializes a new instance of BLO container, populates it with
     * BLOs and initializes the BLOs.
     *
     * @throws BLException if BLOs could not be initialized.
     */
    void init()
        throws BLException {

        ApplicationContext appCtx = this.blManager.getApplicationContext();

        // process BLO deployment descriptors and load BLOs
        BLManager.BLODescriptor [] bloDescriptors = this.blManager.getBLODescriptors();
        for(int i = 0; i < bloDescriptors.length; i++) {
            BLManager.BLODescriptor bloDescriptor = bloDescriptors[i];

            if((bloDescriptor.isShared && this.isShared()) ||
               (!bloDescriptor.isShared && !this.isShared())) {
                BLObject blo = this.getBLOInstance(bloDescriptor);
                blo.setInternals(bloDescriptor,
                                 appCtx,
                                 this);
                this.blos.put(bloDescriptor.name.toUpperCase(), blo);
            }
        }

        // initialize BLOs (we do it separately because init
        // methods of the BLOs may need service of other BLOs
        // to register event listeners).

        // get the log
        Log log = this.blManager.getApplicationContext().getLog(LOG_SUBSYS);

        // check if want transactions for BLO inits
        boolean doTX = appCtx.isDatabaseAvailable();
        boolean inTX = false;

        try {

            // call inits
            for(int i = 0; i < bloDescriptors.length; i++) {
                BLManager.BLODescriptor bloDescriptor = bloDescriptors[i];
                BLObject blo = (BLObject)this.blos.get(bloDescriptor.name.toUpperCase());
                if(blo != null) {

                    // begin tx if available
                    if(doTX) {
                        appCtx.beginTransaction();
                        inTX = true;
                    }

                    // call BLO's init
                    log.debug("initializing BLO \"" + bloDescriptor.name +
                              "\" for session [" + this.sessionId +
                              "], init params are " + bloDescriptor.initParams);
                    blo.init();

                    // commint tx if available
                    if(inTX) {
                        ApplicationContext.commitTransaction();
                        inTX = false;
                    }
                }
            }

        } catch(Throwable e) {

            // log the error
            log.error("Got error initializing BLOs for the session.", e);

            // rollback tx if available
            if(inTX) {
                try {
                    ApplicationContext.rollbackTransaction();
                } catch(ApplicationException e2) {
                    throw new BLException("Encountered error rolling back a transaction due to another error.", e2);
                }
            }

            // pass it up
            ExceptionUtils.
                rethrowBLException(e, "Got an error performing BLO initialization.");
        }
    }

    /**
     * This notification method should be called when there is no more
     * need in this BLO container giving it a chance to release any
     * used resources. This implementation also invokes destroy
     * methods on all BLOs contained in the container. This method is
     * called from the <code>BLManager</code>'s
     * <code>destroyUserSession()</code> method.
     */
    void destroy() {

        // do logout
        if(this.userPrincipal != null) {
            try {
                this.performAuthenticationNotification(null, false);
            } catch(BLException e) {
                this.blManager.getApplicationContext().getLog(LOG_SUBSYS).
                    error("Could not perform logout, destroying the BLO container anyway.", e);
            }
        }

        // call destroy methods on all BLOs
        BLManager.BLODescriptor [] bloDescriptors = this.blManager.getBLODescriptors();
        for(int i = bloDescriptors.length - 1; i >= 0; i--) {
            BLObject blo = (BLObject)this.blos.get(bloDescriptors[i].name.toUpperCase());
            if(blo != null)
                blo.internalDestroy();
        }

        // let the garbage collector to collect the garbage
        this.blos.clear();
    }

    /**
     * Returns an instance of the specified BLO to be used in this
     * session.
     *
     * @param desc descriptor of the BLO we are requesting.
     *
     * @return the requested BLO instance, not initialized.
     *
     * @throws BLException if the BLO instance could not be retrieved.
     */
    private BLObject getBLOInstance(BLManager.BLODescriptor desc)
        throws BLException {

        try {
            return (BLObject)desc.bloClass.newInstance();

        } catch(InstantiationException e) {
            throw new BLException("Could not instantiate a BLO.", e);

        } catch(IllegalAccessException e) {
            throw new BLException("Could not instantiate a BLO.", e);
        }
    }

    /**
     * Sets the owning BL manager. The internal BL manager reference is
     * transient, so this method should be called after an instance of
     * BLO container is restored from the serialized form. Also, this
     * method "activates" all its BLOs.
     *
     * @param blManager reference to the BL manager instance that owns
     * this BLO container.
     *
     * @throws BLException if one of BLOs could not be "activated".
     */
    void activate(BLManager blManager)
        throws BLException {

        this.blManager = blManager;

        ApplicationContext appCtx = blManager.getApplicationContext();
        BLManager.BLODescriptor [] bloDescriptors = blManager.getBLODescriptors();
        int nActivated = 0;
        for(int i = 0; i < bloDescriptors.length; i++) {
            BLObject blo = (BLObject)this.blos.get(bloDescriptors[i].name.toUpperCase());
            if(blo != null) {
                blo.internalActivate(appCtx, this);
                nActivated++;
            }
        }

        if(nActivated != this.blos.size())
            throw new BLException("The BLO container contains a different from the current configurarion set of BLOs.");
    }

    /**
     * Checks if any BLOs implement <code>BLOContainerStateListener</code>
     * interface and calls <code>processBLOContainerLock</code> on them.
     * The BL Manager calls this method right after the BLO container is
     * locked by the <code>getBLOContainer</code> call, but not yet returned
     * to the caller.
     *
     * @throws BLException if a BLO implementing
     * <code>BLOContainerStateListener</code> interface throws it. If it
     * happens, the BL Manager unlocks the BLO container immediately and
     * rethrows the exception to the caller. Also, it does not prevent from
     * calling <code>processBLOContainerUnlock</code> methods on ALL
     * listeners regardless whether <code>processBLOContainerLock</code>
     * was called on them or not.
     */
    void processBLOContainerLock()
        throws BLException {

        Iterator i = this.blos.values().iterator();
        while(i.hasNext()) {
            BLObject blo = (BLObject)i.next();
            if(blo instanceof BLOContainerStateListener)
                ((BLOContainerStateListener)blo).processBLOContainerLock();
        }
    }

    /**
     * Checks if any BLOs implement <code>BLOContainerStateListener</code>
     * interface and calls <code>processBLOContainerUnlock</code> on them.
     * The BL Manager calls this method just before the BLO container gets
     * unlocked by the <code>releaseBLOContainer</code> call.
     *
     * @throws BLException if a BLO implementing
     * <code>BLOContainerStateListener</code> interface throws it. If it
     * happens, the BL Manager unlocks the BLO container anyway and
     * rethrows the exception to the caller.
     */
    void processBLOContainerUnlock()
        throws BLException {

        Throwable exceptionToThrow = null;
        Iterator i = this.blos.values().iterator();
        while(i.hasNext()) {
            BLObject blo = (BLObject)i.next();
            if(blo instanceof BLOContainerStateListener) {
                try {
                    ((BLOContainerStateListener)blo).processBLOContainerUnlock();
                } catch(Throwable e) {
                    if(exceptionToThrow == null)
                        exceptionToThrow = e;
                }
            }
        }

        if(exceptionToThrow != null)
            ExceptionUtils.
                rethrowBLException(exceptionToThrow, "Got error notifying BLOs about BLO container unlock.");
    }


    ///////////////////////////////////////////////////////////////////////////
    // BLO container interface for clients
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets a BLO from this container or the Shared BLO Container. First, it
     * checks if there is a BLO under the specified deployment name in this
     * BLO container. If it finds it, it returns it immediately. If not, and
     * this is not the Shared BLO Container, it tries to find and return the
     * BLO from the Shared BLO Container. If it's not there either, the method
     * throws an exception. Note however, that if it is the Shared BLO
     * Container and it does not contain the requested BLO, this method does
     * not throw an exception, but returns <code>null</code> instead. 
     *
     * @param bloName case-insensitive deployment name of a BLO.
     *
     * @return reference to the named BLO or <code>null</code> if there is no
     * such BLO and this is the Shared BLO Container.
     *
     * @throws IllegalArgumentException if there is no BLO with the specified
     * name and this container is not the Shared BLO Container.
     */
    public BLObject getBLO(String bloName) {

        BLObject blo = (BLObject)this.blos.get(bloName.toUpperCase());
        if((blo == null) && !this.isShared()) {
            blo = this.blManager.getSharedBLO(bloName);
            if(blo == null)
                throw new IllegalArgumentException("There is no BLO under name \"" + bloName + "\".");
        } else if(blo != null) {
            // update stats
            BLManager.BLODescriptor bloDescriptor = blo.getBLODescriptor();
            if(bloDescriptor == null) // this should actually never happen
                throw new IllegalStateException("getBLO is called for BLO \"" + bloName +
                                                "\" and its essential internal fields are not set.");
            bloDescriptor.requestStats.eventStarted();
        }

        return blo;
    }

    /**
     * Checks if there is a BLO under the specified name in this container or
     * in the Shared BLO Container.
     *
     * @param bloName BLO's deployment name as specified in the
     * <code>blo-config.xml</code> file.
     *
     * @return <code>true</code> if the specified BLO exists and can be
     * retrieved via <code>getBLO</code> method.
     */
    public boolean hasBLO(String bloName) {

        boolean res = this.blos.containsKey(bloName.toUpperCase());
        if(!res && !this.isShared())
            res = this.blManager.hasSharedBLO(bloName);

        return res;
    }

    /**
     * Releases this BLO container after it has been locked by the BL Manager's
     * <code>getBLOContainer</code> method. This is a convenience method, which
     * calls <code>releaseBLOContainer</code> on the BL Manager internally. It
     * is often more convenient to call this method rather than the
     * BL Manager's, because it does not require the session id.
     *
     * @throws BLException if an error happens in a BLO implementing
     * <code>BLOContainerStateListener</code> interface and it throws an
     * exception. Note, that the BLO container gets unlocked anyway.
     * @throws UnsupportedOperationException if called on the Shared BLO
     * Container.
     *
     * @see BLManager#releaseBLOContainer
     */
    public void release()
        throws BLException {

        if(this.isShared())
            throw new UnsupportedOperationException("You can not release the shared BLO container.");

        this.blManager.releaseBLOContainer(this.sessionId);
    }

    /**
     * Notifies the BLO container about a user login event for the session. The
     * BLO container calls <code>processUserLogin</code> method on all its
     * BLOs that implement <code>AuthenticationListener</code> interface. A
     * transaction is started for this operation and all BLOs are called within
     * that one single transaction. This method checks if it has been already
     * successfully called earlier and no user logout happened since then. If
     * so, an <code>IllegalStateException</code> is thrown.
     *
     * <p>Standard CB2 Presentation Level implementations, such as the one
     * located in <code>com.boylesoftware.cb2.presentation.servlet</code>
     * package, call this method automatically when appropriate. But if you are
     * implementing the presentation level yourself, you should take care of
     * calling this method upon each successful user login (if you care about
     * the notification at all).
     *
     * <p>Note also, that this method is never called on the Shared BLO
     * Container, because the Shared BLO Container does not correspond to any
     * user session. It makes impractical for shared BLOs to implement
     * <code>AuthenticationListener</code> interface.
     *
     * @param principal the pricipal, who logged in. Should not be
     * <code>null</code>, or a <code>IllegalArgumentException</code> is thrown.
     *
     * @throws BLException if one of the BLOs implementing
     * <code>AuthenticationListener</code> interface threw an exception. It may
     * also mean that not all listeners have been notified, and it certainly
     * means that the trasnaction got rolled back.
     * @throws IllegalStateException if there is already a logged in user.
     *
     * @see AuthenticationListener
     */
    public void processUserLogin(Principal principal)
        throws BLException {

        if(principal == null)
            throw new IllegalArgumentException("Logged in user principal cannot be null.");

        if(this.userPrincipal != null)
            throw new IllegalStateException("The user has aready logged in; " +
                                            "current user is [" + this.userPrincipal.getName() +
                                            "], trying to login user [" + principal.getName() + "].");

        this.performAuthenticationNotification(principal, true);
    }

    /**
     * Notifies the BLO container about a user logout event for the session.
     * This method works very similarly to the <code>processUserLogin</code>
     * method, with the difference that <code>processUserLogout</code> method
     * is called on the BLOs implementing <code>AuthenticationListener</code>
     * interface, and that it does nothing if <code>processUserLogin</code> has
     * not been successfully called earlier.
     *
     * @throws BLException if one of the BLOs implementing
     * <code>AuthenticationListener</code> interface threw an exception. It may
     * also mean that not all listeners have been notified, and it certainly
     * means that the trasnaction got rolled back.
     *
     * @see AuthenticationListener
     * @see #processUserLogin
     */
    public void processUserLogout()
        throws BLException {

        if(this.userPrincipal != null) {
            this.performAuthenticationNotification(null, false);
        }
    }

    /**
     * Performes the user login/logout processing.
     *
     * @param principal the user principal if it is login, or
     * <code>null</code> is it is logout. It will be assigned to the
     * <code>userPrincipal</code> member variable.
     * @param isLogin <code>true</code> if it is login,
     * <code>false</code> if logout.
     *
     * @throws BLException if one of the BLOs implementing
     * <code>AuthenticationListener</code> interface threw an
     * exception.
     */
    private void performAuthenticationNotification(Principal principal, boolean isLogin)
        throws BLException {

        // notify the BLOs in a transaction
        ApplicationContext appCtx = this.blManager.getApplicationContext();
        boolean doTX = appCtx.isDatabaseAvailable();
        boolean inTX = false;
        try {

            // begin tx if available
            if(doTX) {
                appCtx.beginTransaction();
                inTX = true;
            }

            // call the BLOs
            Iterator i = this.blos.values().iterator();
            while(i.hasNext()) {
                BLObject blo = (BLObject)i.next();
                if(blo instanceof AuthenticationListener) {
                    if(isLogin)
                        ((AuthenticationListener)blo).processUserLogin(principal);
                    else
                        ((AuthenticationListener)blo).processUserLogout();
                }
            }

            // commint tx if available
            if(inTX) {
                ApplicationContext.commitTransaction();
                inTX = false;
            }

        } catch(Throwable e) {

            // rollback tx if available
            if(inTX) {
                try {
                    ApplicationContext.rollbackTransaction();
                } catch(ApplicationException e2) {
                    throw new BLException("Encountered error rolling back a transaction due to another error.", e2);
                }
            }

            // pass it up
            ExceptionUtils.
                rethrowBLException(e, "Got error notifying BLOs about user login/logout.");
        }

        // done successfully
        this.userPrincipal = principal;
    }


    ///////////////////////////////////////////////////////////////////////////
    // BLO container interface for BLOs
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets the BL Manager.
     *
     * @return reference to the BL Manager.
     */
    public BLManager getBLManager() {

        return this.blManager;
    }

    /**
     * Gets id of the user session, to which this container corresponds.
     *
     * @return id of the session, or <code>null</code> if it is the Shared BLO
     * Container.
     */
    public Object getSessionId() {

        return this.sessionId;
    }

    /**
     * Gets currently authenticated for the session user principal set earlier
     * by the <code>processUserLogin</code> call.
     *
     * @return the user principal or <code>null</code> if the session is not
     * authenticated.
     */
    public Principal getUserPrincipal() {

        return this.userPrincipal;
    }

    /**
     * Tells if this is the Shared BLO Container.
     *
     * @return <code>true</code> if this is the Shared BLO Container.
     */
    public boolean isShared() {

        return (this.sessionId == null);
    }
}
