package com.boylesoftware.cb2;

import org.apache.commons.logging.Log;


/**
 * Parent abstract class for Business Level Objects (the BLOs), that are
 * software components in the CB2 architecture that represent various business
 * entities and implement application's business logic.
 *
 * <p>This class provides API for BLO implementations as a set of protected
 * methods. The public interface of a BLO implementation represents its
 * business interface.
 *
 * @author Lev Himmelfarb
 * @version $Id: BLObject.java,v 1.6 2004/10/11 18:17:36 levahim Exp $
 *
 * @see BLManager
 * @see BLOContainer
 */
public abstract class BLObject
    implements java.io.Serializable {

    /**
     * Subsystem name prefix for the logger. The
     * <code>BLOContainer</code> adds the BLO's name to this prefix to
     * get a logger for the BLO.
     */
    private final static String LOG_SUBSYS_PREFIX = "com.boylesoftware.cb2.BLOBJT.";


    /**
     * The BLO descriptor used for this BLO.
     */
    private BLManager.BLODescriptor bloDescriptor;

    /**
     * Reference to the <code>ApplicationContext</code>.
     */
    private transient ApplicationContext appCtx;

    /**
     * Reference to the BLO container.
     *
     * @deprecated use <code>getBLOContainer</code> or <code>getBLO</code>
     * protected methods instead. Exposing this variable makes possible for a
     * BLO to change it, which is illegal.
     */
    protected transient BLOContainer bloc;

    /**
     * Business errors for the last invoked business method.
     */
    private BLErrors errors;

    /**
     * Log to be used in the BLO's methods. The log is preconfigured to log
     * messages to channel called "com.boylesoftware.cb2.BLOBJT.<i>name</i>",
     * where <i>name</i> is the BLO's deployment name.
     */
    protected transient Log log;


    ///////////////////////////////////////////////////////////////////////////
    // Lifecycle
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Sets the BLO's internal fields. This method is always called before
     * any <code>init</code> method on any BLO from the same container. Thus
     * it guarantees that the if a BLO accesses another BLO in its
     * <code>init</code> method the essential fields will be set in the
     * other BLO even if it was not initialized yet.
     *
     * @param bloDescriptor reference to the BLO descriptor used for this BLO.
     * @param appCtx out application context.
     * @param bloContainer our BLO container.
     */
    final void setInternals(BLManager.BLODescriptor bloDescriptor,
                            ApplicationContext appCtx,
                            BLOContainer bloContainer) {

        this.bloDescriptor = bloDescriptor;
        this.appCtx = appCtx;
        this.bloc = bloContainer;
        this.log = appCtx.getLog(LOG_SUBSYS_PREFIX + this.bloDescriptor.name);
    }

    /**
     * Gets the reference to the BLO descriptor used for this BLO, which is
     * stored internally in the <code>BLObject</code>.
     *
     * @return reference to the BLO descriptor.
     */
    final BLManager.BLODescriptor getBLODescriptor() {

        return this.bloDescriptor;
    }

    /**
     * Initializes the BLO. This method is called automatically on the BLO to
     * notify it that a new BLO container has been created and this BLO is
     * assigned to it. Subclasses redefine this method to implement BLO
     * initialization logic. The default implementation does nothing.
     *
     * <p>Shared BLOs are created and initialized when the BL Manager is
     * created and initialized, usually during application startup, making
     * <code>init</code> methods of shared BLOs an ideal place for application
     * initialization code.
     *
     * <p>If <code>init</code> method of a BLO throws an exception, it cancels
     * the whole BLO container initialization - BLO container cannot be created
     * and used unless all BLOs went through their <code>init</code> methods
     * successfully.
     *
     * <p>BLOs are initialized in a transaction - one transaction to call
     * <code>init</code> methods of all BLOs in the BLO container. If any of
     * the BLOs throws an exception, the whole transaction, which includes
     * anything done in <code>init</code> methods of other BLOs that has
     * already been successfully called, gets rolled back.
     *
     * <p>Note, that BLOs are always initialized in the order their deployment
     * descriptors appear in the <code>blo-config.xml</code> file. If one BLO
     * needs service of another BLO in its <code>init</code> method, place
     * the other BLO's deployment descriptor above the deployment descriptor of
     * the first one in the <code>blo-config.xml</code> - it will guarantee
     * that the other BLO will be initialized first.
     *
     * @throws BLException if the BLO cannot be initialized.
     */
    protected void init()
        throws BLException { /* do nothing */ }

    /**
     * The BLO container invokes this method when it does not need
     * this BLO anymore thus giving it a chance to release any used
     * resources. This method also calls the user defined
     * <code>destroy()</code> method internally.
     */
    final void internalDestroy() {

        this.log.debug("removing from session [" + this.bloc.getSessionId() + "]");

        this.destroy();

        this.appCtx = null;
        this.bloc = null;
        this.errors = null;
        this.log = null;
    }

    /**
     * Destroyes the BLO. This method is called automatically when BLO
     * container, to which this BLO belongs, is destroyed. BLO containers get
     * destroyed when their respective user sessions finish or when the whole
     * application shuts down. Subclasses redefine this method to free any used
     * resources or to perform any other operations related to the fact that
     * the BLO is no longer used. The default implementation does nothing.
     *
     * <p>Note, that the Shared BLO Container gets destroyed when the BL
     * Manager gets destroyed, usually at the application shutdown, making
     * <code>destroy</code> methods of shared BLOs an ideal place for code,
     * which needs to be executed before the application goes down.
     *
     * <p>No database transaction is created for <code>destroy</code> methods,
     * so it is not the best place to perform database operations.
     *
     * <p>BLOs are always destroyed in the <strong>reverse</strong> order their
     * deployment descriptors appear in the <code>blo-config.xml</code> file
     * (reverse order from how they are initialized).
     */
    protected void destroy() { /* do nothing */ }

    /**
     * This method sets internal transient fields and calls user-defined
     * methos <code>activate</code>. This method shuold be called after an
     * instance is restored from the serialized form.
     *
     * @param appCtx our application context.
     * @param bloContainer the parent BLO container instance.
     *
     * @throws BLException if user-defined <code>activate</code> method
     * throws an exception thus indicating that the instance cannot be
     * restored.
     */
    final void internalActivate(ApplicationContext appCtx,
                                BLOContainer bloContainer)
        throws BLException {

        this.appCtx = appCtx;
        this.bloc = bloContainer;
        this.log = appCtx.getLog(LOG_SUBSYS_PREFIX + this.bloDescriptor.name);

        this.log.debug("activating for session [" + this.bloc.getSessionId() + "]");

        this.activate();
    }

    /**
     * Restore BLO after it has been recreated from the serialized form. This
     * method is called automatically when a user session is being restored in
     * the BL Manager via <code>restoreUserSession</code> call. Subclasses
     * redefine this method to implement any logic specific to such an event.
     * The default implementation does nothing.
     *
     * <p>Note, that activation is performed in non-transactional context,
     * so avoid doing database related operations in <code>activate</code>
     * method implementations. Also, keep the method implementation short,
     * because it is called from a synchronized block, which prevents getting
     * and locking BLO containers for <strong>all</strong> user sessions in the
     * application.
     *
     * <p>BLOs are always activated in the order their deployment descriptors
     * appear in the <code>blo-config.xml</code> file (the same order they are
     * initialized).
     *
     * @throws BLException if the instance cannot be restored. Throwing
     * exception cancels the whole user session restoration attempt - the
     * <code>restoreUserSession</code> method on the BL Manager will throw an
     * exception to the caller.
     *
     * @see BLManager#restoreUserSession
     */
    protected void activate()
        throws BLException { /* do nothing */ }


    ///////////////////////////////////////////////////////////////////////////
    // Service
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets errors for the last BLO business method call. A business method
     * creates <code>BLErrors</code> object and sets it in the BLO via
     * <code>setErrors</code> protected method. Then it returns to the caller
     * a special, method specific result indicating that there were errors and
     * the call was not successful. The caller then call
     * <code>getLastErrors</code> method on the BLO to find out what were the
     * errors.
     *
     * <p>Calling this method resets the internal BLO variable that stores
     * <code>BLErrors</code> object after last <code>setErrors</code> call.
     * After it has been reset, all subsequent calls to
     * <code>getLastErrors</code> will be returning an empty
     * <code>BLErrors</code> object until another <code>setErrors</code> call
     * sets new <code>BLErrors</code>.
     *
     * @return errors for the last business method call. Never returns
     * <code>null</code>, but an empty errors object if there are no errors.
     */
    public BLErrors getLastErrors() {

        BLErrors res = this.errors;
        if(res == null)
            res = new BLErrors(0);

        this.errors = null;

        return res;
    }

    /**
     * Sets errors for the last BLO business method call. If a business method
     * wants to indicate to the caller that the call was not successful, it
     * creates a <code>BLErrors</code> object, sets it in the BLO using this
     * method, and then returns to the caller a special kind of result, which
     * tells the caller that there were errors and it should call
     * <code>getLastErrors</code> on the BLO to find out what exactly it was.
     *
     * <p>If there was already a <code>BLErrors</code> object set by an earlier
     * call, it is overwritten by this call.
     *
     * @param errors the errors bundle.
     */
    protected final void setErrors(BLErrors errors) {

        this.errors = errors;
    }

    /**
     * Gets a BLO initialization parameter. A BLO can have custom
     * initialization parameters specified in its deployment descriptor in the
     * <code>blo-config.xml</code> file via <code>&lt;init-param&gt;</code>
     * elements. They can be used to configure the BLO for the particular
     * application in which it is used, so the BLO implementation can be reused
     * in other applications with some functional changes.
     *
     * @param parameterName the init parameter's name as specified by the
     * <code>&lt;param-name&gt;</code> element.
     *
     * @return the specified parameter's value or <code>null</code> if no such
     * init parameter is defined in the BLO's deployment descriptor.
     */
    protected final String getInitParameter(String parameterName) {

        return this.bloDescriptor.initParams.getProperty(parameterName);
    }

    /**
     * Marks current transaction for rollback only, or does nothing if
     * there is no current transaction. This is a convenience method, which
     * calls application context's appropriate method internally.
     *
     * @throws BLException could not mark current transaction for rollback
     * only.
     *
     * @see ApplicationContext#setRollbackOnly()
     */
    protected void setRollbackOnly()
        throws BLException {

        try {
            ApplicationContext.setRollbackOnly();
        } catch(ApplicationException e) {
            throw new BLException("Could not set transaction for rollack only.", e);
        }
    }

    /**
     * Checks if the current transaction is marked for rollback only. Always
     * returns <code>false</code> if there is no current transaction. This is a
     * convenience method, which calls application context's appropriate method
     * internally.
     *
     * @return <code>true</code> if the current transaction is marked for
     * rollback only.
     *
     * @throws BLException could not get the current transaction's status.
     *
     * @see ApplicationContext#isRollbackOnly()
     */
    protected boolean isRollbackOnly()
        throws BLException {

        try {
            return ApplicationContext.isRollbackOnly();
        } catch(ApplicationException e) {
            throw new BLException("Could not get transaction status.", e);
        }
    }

    /**
     * Gets DAO config object for the specified data source. DAO config objects
     * are created from <code>&lt;dao-config&gt;</code> elements in the
     * <code>blo-config.xml</code> file and are maintained by the BL Manager -
     * one for each configured data source.
     *
     * <p>A BLO may need this object if it wants to create its own custom DAO
     * for some reason, which is supposed to be a rather rare case.
     *
     * @param dataSourceName name of the data source as specified in the
     * application context configuration and <code>dataSource</code> attribute
     * of the <code>&lt;dao-config&gt;</code> element.
     *
     * @return reference to the appropriate DAO config object.
     */
    protected DAOConfig getDAOConfig(String dataSourceName) {

        return this.bloc.getBLManager().
            getDAOConfig(dataSourceName);
    }

    /**
     * Gets DAO config object for the default data source.
     *
     * @return reference to the DAO config object or <code>null</code> if the
     * default source is not configured.
     */
    protected DAOConfig getDAOConfig() {

        return this.getDAOConfig("default");
    }

    /**
     * Gets DAO for the specified data source. The DAO can be used in BLO's
     * business methods to perform database operations. Although it is possible
     * for a BLO to perform database operations directly via the JDBC interface
     * (the application context provides access to configured database
     * connections), it is more convenient to use the DAO, since it provides a
     * higher level interface helping in the most use cases.
     *
     * <p>The BL Manager maintains a single DAO instance for each configured
     * data source. The DAO's are built using DAO config objects created out
     * of <code>&lt;dao-config&gt;</code> elements defined in the
     * <code>blo-config.xml</code> file. This method allows to get reference to
     * the appropriate DAO, maintained by the BL Manager, to be used in a BLO's
     * business method.
     *
     * @param dataSourceName name of the data source as specified in the
     * application context configuration and <code>dataSource</code> attribute
     * of the <code>&lt;dao-config&gt;</code> element.
     *
     * @return reference to the appropriate DAO.
     */
    protected DAO getDAO(String dataSourceName) {

        return this.bloc.getBLManager().
            getDAO(dataSourceName);
    }

    /**
     * Gets DAO for the default data source.
     *
     * @return reference to the DAO or <code>null</code> if the default data
     * source is not configured.
     */
    protected DAO getDAO() {

        return this.getDAO("default");
    }

    /**
     * Gets BLO container to which this BLO belongs. Often BLOs need service of
     * other BLOs in their methods. The <code>getBLOContainer</code> method
     * provides access to the BLO container, which can be used to lookup other
     * BLOs that belong to the same user session as well as shared BLOs.
     *
     * <p>The BLO container's state is properly maintained by the BL Manager -
     * there is no need to release the BLO container aquired via this protected
     * method, the reference can be abandoned after it is no longer needed.
     *
     * @return reference to the BLO container.
     */
    protected final BLOContainer getBLOContainer() {

        return this.bloc;
    }

    /**
     * Gets a BLO from the same or the shared BLO container. This is a
     * convenience method, which is equivalent to calling
     * <code>this.getBLOContainer().getBLO(bloName)</code>.
     *
     * @param bloName the BLO's deployment name.
     *
     * @return the BLO.
     */
    protected final BLObject getBLO(String bloName) {

        return this.bloc.getBLO(bloName);
    }

    /**
     * Gets application context. This method is the way for the BLO's methods
     * to access application context's services.
     *
     * @return reference to the application context.
     */
    protected final ApplicationContext getApplicationContext() {

        return this.appCtx;
    }

    /**
     * Gets the BLO's deployment name as specified by <code>name</code>
     * attribute of the corresponding <code>&lt;blo&gt;</code> element in the
     * <code>blo-config.xml</code> file. This is the name, under which the BLO
     * can be looked up in a BLO container.
     *
     * @return the BLO's deployment name.
     */
    public final String getDeploymentName() {

        return this.bloDescriptor.name;
    }
}
