package com.boylesoftware.cb2;

import java.io.Serializable;


/**
 * Interface for CB2 Broadcast Messaging provider implementations.
 * CB2 Broadcast Messaging subsystem (the BCM), supported by the application
 * context, helps building server clusters providing means for sending
 * broadcast messages to application instances in the cluster. A BCM provider
 * is an object behind application context's <code>broadcastBCMMessage</code>
 * method, and that is the provider, which actually does all the work of the
 * delivery of messages to the cluster members.
 *
 * <p>When provider receives a message for the local instance, it should call
 * <code>processBCMMessage</code> on the application context to deliver it to
 * the appropriate BCM terminal.
 *
 * <p>The provider implementation used by the application context is specified
 * by <code>com.boylesoftware.cb2.broadcastMessaging.provider.class</code>
 * application property.
 *
 * <p>CB2 has several BCM provider implementations packaged with it - see
 * <code>com.boylesoftware.cb2.bcmproviders.*</code> packages.
 *
 * @author Lev Himmelfarb
 * @version $Id: BroadcastMessagingProvider.java,v 1.3 2004/04/16 21:43:27 levahim Exp $
 *
 * @see ApplicationContext#broadcastBCMMessage
 * @see ApplicationContext#processBCMMessage
 */
public interface BroadcastMessagingProvider {

    /**
     * Channel for the logger. Provider implementations should use this channel
     * name for logging.
     */
    static final String LOG_SUBSYS = "com.boylesoftware.cb2.BCMPRV";


    /**
     * Initializes the provider. The application context calls this method once
     * just after the provider has been instantiated.
     *
     * @param appCtx reference to the application context.
     *
     * @throws ApplicationException if a serious error happens, the provider
     * cannot be initialized and the application should shut down itself.
     */
    void init(ApplicationContext appCtx)
        throws ApplicationException;

    /**
     * Destroys the provider and frees any used resources. The application
     * context calls this method once when it does not need the provider
     * anymore, usually when the application goes down or is being
     * reinitialized.
     */
    void destroy();

    /**
     * Makes the provider to broadcast a message to all registered instances in
     * the cluster including the local instance. Note, that depending on the
     * BCM provider implementation this method may return before the message is
     * actually delivered to the local, as well as to any remote cluster
     * member.
     *
     * @param channelName name of the channel, to which the message is
     * targeted.
     * @param messageParam optional general purpose message parameter, may be
     * <code>null</code>.
     */
    void broadcastMessage(String channelName, Serializable messageParam);
}
