package com.boylesoftware.cb2;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.lang.reflect.Field;
import java.lang.reflect.Array;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.util.EventStatistics;
import com.boylesoftware.cb2.util.TextUtils;
import com.boylesoftware.cb2.util.pool.AdaptiveObjectPool;
import com.boylesoftware.cb2.util.pool.PooledObjectFactoryAdapter;
import com.boylesoftware.cb2.util.sql.ConnectionWrapper;


/**
 * Database Access Object (the DAO) provides high-level API for database
 * operations. Internally, it is based on JDBC, but working through the DAO
 * allows application code not to worry about opening and closing database
 * connections, creating and executing statements, processing result sets. The
 * DAO works with Data Models (DMs), which are very simple objects extending
 * <code>DataModel</code> abstract class and containing public data fields that
 * match to result set columns. When a fetch call (corresponding to SQL
 * <code>SELECT</code>) is performed, the DAO processes the result set, for
 * each row it create a DM, stores data from the result set columns to the
 * appropriate fields of the DM, and then returns the while fetch result to the
 * caller as an array of populated DMs. When an update or insert is performed,
 * the DAO constructs an appropriate SQL query automatically basing on the DM's
 * set of fields, and then executes it. DMs fields a mapped to result set
 * columns by names - in the simpliest case, the DAO finds the result set
 * column's label via standard JDBC method <code>getColumnLabel</code> on
 * <code>ResultSetMetaData</code> and then matches it to the DM's field with
 * the same name. In reality, before the DAO matches column label to the DM
 * field, it chops off prefix from the label. Column label prefix is a string,
 * which prepends the corresponding DM field name in a column label and is
 * separated from it by a special character sequence called <i>field name
 * separator</i>, which is specified in the DAO configuration. The prefix plays
 * no role in result set column to DM field matching and its actual usage is
 * dicussed later.
 *
 * <h4>DAO Configuration</h4>
 *
 * Although the DAO provides low-level methods called <code>executeFetch</code>
 * and <code>executeUpdate</code>, which allow the application to pass SQL
 * query text to be executed directly to the method, that is not the usual DAO
 * usage. Instead, the majority of the DAO's methods work with queries
 * specified externally in the DAO configuration and identified by names. This
 * approach allows to isolate SQL code from the application's Java code. The
 * DAO configuration is represented by an instance of <code>DAOConfig</code>
 * class, which is built from a <code>&lt;dao-config&gt;</code> element in the
 * <code>blo-config.xml</code> file. In order to create an instance of DAO it
 * is always required to specify an appropriate DAO config object. Beside query
 * definitions, the DAO configuration also contains descriptors for all data
 * models. The descriptors provide some meta data for the DMs, such as to what
 * database table a DM corresponds, and also allow to specify in query
 * definitions what DM class this or that fetch query returns. See
 * <a href="http://www.cb2project.com/dtd/blo-config_1_0.dtd">http://www.cb2project.com/dtd/blo-config_1_0.dtd</a>
 * for the <code>blo-config.xml</code> file structure.
 *
 * <h4>DMs and Nested DMs</h4>
 *
 * It is the best approach first to create DMs that directly correspond to
 * database tables used in the application - the logical meaning of a Data
 * Model is to represent data in the way it is stored in the database. In some
 * cases small hierarchies can be created to provide short versions of DMs to
 * be used with fetches that return lists of DMs with only most basic fields
 * populated - a search result is a good example. The full version of the DM,
 * which extends the short one, then can be used for fetching full details of
 * a record and updating/creating new records.
 *
 * <p>It is often required to fetch data not from one table, but from a number
 * of joined tables. To make it easy, DMs can have nested DMs, to which data
 * from the joined tables go. In order to allow the DAO to match result set
 * columns to the appropriate DMs, the result set must be structured in a
 * specific way. The columns in the result set are always processed from left
 * to right, so the order of the select list is important. First, columns are
 * matched against the top DM's fields. Then, if a column labeled so it matches
 * to the DM's field which is in turn another DM, the context switches to the
 * other DM. Depending on whether it is one to zero-or-one or one to
 * zero-or-more relationship, the nested field may be represented by a single
 * DM class field or by an array of DMs. The result set column labeled by the
 * nested DM field's name, which is called <i>the header column</i>, must have
 * different values for these two cases. In the case of a single nested DM the
 * header column's value must be the number of result set columns that follow
 * this one and must be stored to the nested DM. In the case of an array of
 * nested DMs, the header column's value must be something, that identifies
 * the parent record, so consequent result set rows that have the same value in
 * the header column will go to one array nested in one parent DM until the
 * header column's value changes and a new parent DM is created with a new
 * nested array in it. Note, that rows belonging to the same nested array must
 * immediately follow each other in the result set, which usually requires an
 * appropriate <code>ORDER BY</code> or <code>GROUP BY</code> clause in the SQL
 * query. Also, the first column in the sequence of columns that correspond to
 * the nested array must not be nullable - it it is <code>NULL</code> in the
 * result set, which is possible in the case of an outer join, the nested array
 * in the parent DM is left empty. In the case of nested DM arrays, the DAO
 * considers all columns following the header column until the last column in
 * the result set row belonging to the nested DM. It means that columns that
 * go to the nested array must be all groupped together at the very end of the
 * select list.
 *
 * <p>Nested DM fields, either single nested DMs or nested DM arrays, are never
 * set to <code>null</code> by the DAO - in case of an outer join when no
 * record exists corresponding to the nested field, if the field is a single
 * DM, an instance of the DM class will be created and all fields in it will be
 * set to <code>null</code>. If it is a nested array, an epmty array will be
 * set into the parent DM's field.
 *
 * <p>The DAO also correctly processes nested DMs in other nested DMs allowing
 * unlimited nesting levels. However, please note that due to the nature of
 * relational databases, it is impossible to have more than one nested array
 * on the same nesting level.
 *
 * <p>Sometimes the database does not allow having columns within one select
 * list labeled with identical names. On the other hand, in the case of nested
 * DMs it is possible that two DMs have fields with the same names. To help
 * solve the problem, column labels can be prepended by prefixes, that will
 * make them unique. A prefix is any string separated from the rest of the
 * column label, which is the corresponding DM field name, by a <i>field name
 * separator</i> sequence, which can be configured by a
 * <code>&lt;field-name-separator&gt;</code> element in the DAO configuration.
 * Suppose the field name separator is an underscore character, then column
 * labeled "person_firstName" will be matched to a DM field named "firstName",
 * while the "person_" prefix is discarded.
 *
 * <p>Using nested DMs is a relatively complex issue. You should read the CB2
 * User Manual, available on the web-site, for better understanding.
 *
 * <p>Note, that it is safe to add nested DM fields to DMs that are used for
 * DAO updates and inserts - those calls ignore fields that are of type derived
 * from <code>DataModel</code> and arrays of such types when constructing SQL
 * statements.
 *
 * <p>Note also, that because of the DAO Extended Syntax, described below, it
 * is rarely (or never?) needed to manually construct proper select lists in
 * queries to structure result sets.
 *
 * <h4>DAO Extended Syntax</h4>
 *
 * In the DAO configuration XML, where SQL query bodies are defined, it is
 * allowed to have special <code>{...}</code> contructs, similar to JDBC escape
 * syntax, that are interpreted by the DAO. Below is the list of such
 * constructs:
 *
 * <ul>
 *
 * <li><code><b>{macro ...}</b></code> - allows to define a piece of SQL code
 * in a <code>&lt;sqlmacro&gt;</code> element and then insert it into
 * different SQL queries. The syntax for the construct is as follows:
 *
 * <pre>
 * {macro <i>&lt;macro name&gt;</i> [(<i>&lt;param 1 value&gt;</i> [, <i>&lt;param 2 value&gt;</i> ...])]}</pre>
 *
 * Every occurence of this construct will be replaced by the body of the
 * defined by an <code>&lt;sqlmacro&gt;</code> element macro with approprite
 * <code>{mparam ...}</code> constucts replaced by the provided macro parameter
 * values.</li>
 *
 * <p><li><code><b>{mparam ...}</b></code> - used in <code>&lt;sql&gt;</code>
 * subelements of <code>&lt;sqlmacro&gt;</code> elements to identify macro
 * parameter placeholders. The syntax is:
 *
 * <pre>
 * {mparam <i>&lt;param name&gt;</i>}</pre>
 *
 * Where <code><i>&lt;param name&gt;</i></code> is the name of a parameter
 * defined by a <code>&lt;param&gt;</code> element.</li>
 *
 * <p><li><code><b>{dm ...}</b></code> - renders select list for the DM,
 * associated with the query by <code>usedm</code> attribute of the
 * <code>&lt;query&gt;</code> element, which may include select lists for
 * nested DMs as well. The syntax is as follows:
 *
 * <pre>
 * {dm [from <i>&lt;table name&gt;</i>]
 *     [prefix <i>&lt;column label prefix&gt;</i>]
 *     [excluding (<i>&lt;DM field 1&gt;</i> [, <i>&lt;DM field 2&gt;</i> ...])|only (<i>&lt;DM field 1&gt;</i> [, <i>&lt;DM field 2&gt;</i> ...])]
 *     [{<i>&lt;nested DM field&gt;</i>
 *         [from <i>&lt;table name&gt;</i>]
 *         [prefix <i>&lt;column label prefix&gt;</i>]
 *         [excluding (<i>&lt;DM field 1&gt;</i> [, <i>&lt;DM field 2&gt;</i> ...])|only (<i>&lt;DM field 1&gt;</i> [, <i>&lt;DM field 2&gt;</i> ...])]
 *         [by <i>&lt;parent DM field&gt;</i>]
 *         [{<i>&lt;nested DM field&gt;</i> ...}
 *          ...
 *         ]
 *      }
 *      ...
 *     ]
 * }</pre>
 *
 * Here is the meaning of the optional parameters:
 *
 * <p><ul>
 *
 * <li><code><b>from</b></code> - specifies table from which DM fields should
 * be selected. By default, the tables associated with the DM by
 * <code>table</code> attribute of its <code>&lt;dm&gt;</code> element is
 * used.</li>
 *
 * <li><code><b>prefix</b></code> - prefix to be added to the column labels. By
 * default no prefix is added to the top DM fields and a prefix consisting of
 * parent DM field names up to the toppest one separated by the field name
 * separator sequence is added to nested DMs.</li>
 *
 * <li><code><b>excluding</b></code> - allows to exclude named DM fields from
 * the select list. By default all DM fields that are not nested DMs and not
 * arrays of nested DMs are included.</li>
 *
 * <li><code><b>only</b></code> - opposite to <code>excluding</code>, which
 * tells to include <strong>only</strong> the named DM fields to the select
 * list.</li>
 *
 * <li><code><b>by</b></code> - in the context of a nested DM array specifies
 * the name of the parent DM's field which should be used as the value of the
 * header column. By default the first, or the only id field (specified by an
 * <code>&lt;idfield&gt;</code> element in the DM descriptor) of the parent DM
 * is used.</li>
 *
 * <li><code><b>{<i>&lt;nested DM field&gt;</i> ...}</b></code> - allows to
 * render select list for a nested DM including the appropriate header
 * column.</li>
 *
 * </ul>
 *
 * <p>See CB2 User Manual for examples.</li>
 *
 * <p><li><code><b>{cond ...}</b></code> - allows to have dynamic SQL. Some
 * variations of the DAO's methods take <code>conds</code> argument, which
 * is a set of condition names to include in the query. Bodies of
 * <code>{cond ...}</code> constructs are included to the final SQL query
 * only if the construct's name is present in the <code>conds</code> set.
 * Here is the full syntax:
 *
 * <pre>
 * {cond (<i>&lt;cond name&gt;</i>) <i>&lt;SQL text&gt;</i>}</pre>
 *
 * Where <code><i>&lt;cond name&gt;</i></code> is the condition name to be put
 * to the <code>conds</code> set if the condition's body should be
 * included.</li>
 *
 * <p><li><code><b>{? ...}</b></code> and <code><b>{?? ...}</b></code> - named
 * query parameter placeholders. DAO methods <code>fetchWithNameParams</code>
 * and <code>updateWithNamedParams</code> allow to specify query parameters
 * not by positions, but by names. These constructs identify where the named
 * parameters go in the query. Here is the syntax:
 *
 * <pre>
 * {? <i>&lt;param name&gt;</i> [[+]<i>&lt;SQL type&gt;</i>]}
 * {?? <i>&lt;param name&gt;</i> [[+]<i>&lt;SQL type&gt;</i>]}</pre>
 *
 * Specifying parameter's SQL type plays role when the DAO tries to set the
 * parameter to <code>null</code>. The JDBC's <code>setNull</code> method on
 * the <code>PreparedStatement</code> object requires the the SQL type of the
 * target parameter is specified. By default, it the SQL type is not in the
 * parameter's placeholder, the DAO uses <code>java.sql.Types.VARCHAR</code>,
 * which some JDBC drivers may dislike when used with, say, a numeric column.
 * If the SQL type is specified with the placeholder, the DAO uses it when
 * calling <code>setNull</code> method. Names of SQL types are case-insensitive
 * names of constants in the <code>java.sql.Types</code> class. A plus sign
 * in from of a type name makes the DAO to use it not only with
 * <code>setNull</code> method, but also with <code>setObject</code> method
 * when the parameter's value is not <code>null</code>.
 *
 * <p>See below for the explanation of what double question mark means.</li>
 *
 * </ul>
 *
 * <p>Another extension of standard JDBC is double question marks as parameter
 * placeholders. It allows to have arrays as query parameters - the
 * corresponding double question mark will be expanded by the DAO to a sequence
 * of comma-separated single question marks, which is the standard JDBC query
 * parameter placeholder mark, with the exactly the same length as the number
 * of elements in the query parameter array, so the values from the array will
 * be set to the appropriate places. This is especially useful with SQL
 * <code>IN</code> conditions.
 *
 * <h4>DAO Plugins</h4>
 *
 * The DAO is a very flexible tool and it provides even more flexibility with
 * plugins mechanism. You can plug in special modules, sort of hooks, to the
 * DAO and it will be calling them at different times to perform this or that
 * elementary operation. Plugins are installed in the DAO configuration via
 * <code>&lt;plugin&gt</code> elements. To see what kinds of plugins DAO
 * supports see subinterfaces of the <code>DAOPlugin</code> abstract interface.
 *
 * <p>There is also another kind of hook, which is not actually a DAO plugin,
 * but is used quite similarly. Objects implementing
 * <code>FetchResultProcessor</code> can be used with the DAO fetch calls to
 * perform certain application specific operations on DMs as the DAO creates
 * them in its internal result set processing loop. If can even completely
 * change the fetch call's effect. For example, you have a query, which returns
 * hundreds of thousands of rows and what you need is to process the data and
 * store the information in a file. If you let the fetch call return the result
 * as an array of DMs, first, the memory consumption will be irrational, second
 * you will have two loops going over the whole array - first time when the DAO
 * processes the result set rows, second time when you loop through the
 * resulting array to process the returned DMs. The correct way to solve such a
 * task is to have a <code>FetchResultProcessor</code> implementation, which
 * processes individual DMs as they come out of the DAO, append data to the
 * file, and then discard the DM instance so it does not take memory.
 *
 * <h4>Result Set Pagination</h4>
 *
 * The DAO's fetch methods are able to select and return only a certain page
 * from the whole result set returned by the query. The pagination is
 * performed by a plugin implementing <code>FetchExecutor</code> interface. The
 * default CB2 implementation, <code>GenericFetchExecutor</code>, uses
 * scrollable result sets to scroll to and then fetch the requested page. Note,
 * that unfortunately not all JDBC drivers implement scrollable result sets
 * correctly, in which case it may require a database specific approach,
 * usually rather complex and tricky, if your application needs the
 * functionality of result set pagination. CB2 has some database specific
 * solutions packaged with it, see <code>com.boylesoftware.cb2.db.*</code>
 * packages.
 *
 * <p><em>The result set pagination mechanism assumes that one row in the
 * result set corresponds to one record, which is not true if nested DM arrays
 * are used, where a sequence of rows following each other may be packed into
 * one DM. The result set pagination does not work correctly with nested DM
 * arrays. It is not a bug, but certainly a limitation.</em>
 *
 * <h4>Accessing the DAO</h4>
 *
 * It is rarely needed for the application code to create instances of the DAO.
 * The BL Manager maintains an instance of DAO for each configured data source.
 * A reference to a BL Manager's DAO can be retrieved in any BLO via its
 * <code>getDAO</code> protected method. Also, the DAO class is completly
 * thread safe, so it is perfectly fine if one instance is used by multiple
 * threads at the same time.
 *
 * <p>However, there are advanced cases when the application would need to
 * create its own instance of the DAO. For example, if it needs to make the DAO
 * to work with a particular database connection not available through the
 * application context. Note, that sometimes there is a need to execute a
 * sequence of database calls with a guarantee that they all are executed on
 * the same connection. That situation is not a reason for creating an instance
 * of DAO - the BL Manager's DAO can be still used provided that the connection
 * is "pinned down" to the thread by a <code>pinConnection</code> call. Another
 * way to execute several SQL statements on the same connection is to define
 * multiple <code>&lt;sql&gt;</code> subelements (so-called "sql chunks")
 * within one <code>&lt;query&gt;</code> element in the DAO configuration.
 *
 * <p>When an instance of DAO needs to be created, it requires specifying the
 * appropriate DAO config object. The DAO config objects are also maintained
 * by the BL Manager, one for each configured data source, and can be retieved
 * from within any BLO via its <code>getDAOConfig</code> protected method.
 *
 * <h4>Transactions</h4>
 *
 * The DAO does not perform <strong>any</strong> transaction management.
 *
 * <h4>Logging</h4>
 *
 * <p>The DAO logs messages related to its operation under
 * "com.boylesoftware.cb2.DAOBJT" log channel.
 *
 * @author Lev Himmelfarb
 * @version $Id: DAO.java,v 1.18 2004/10/11 18:17:36 levahim Exp $
 */
public final class DAO
    implements StatusProvider {

    /**
     * Subsystem name prefix for the logger.
     */
    private final static String LOG_SUBSYS = "com.boylesoftware.cb2.DAOBJT";


    ///////////////////////////////////////////////////////////////////////////
    // Public constants
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Used with fetch methods to indicate that automatic <code>ORDER BY</code>
     * clause generation is not used, or it is used, but no <code>ASC</code>
     * nor <code>DESC</code> keywords should be added to the column names in
     * the clause.
     */
    public final static int ORDER_UNUSED = 0;

    /**
     * Same as <code>ORDER_UNUSED</code>.
     *
     * @deprecated use <code>ORDER_UNUSED</code> instead.
     */
    public final static int ORDER_NOORDER = ORDER_UNUSED;

    /**
     * Used with fetch methods to indicate that the caller wants ascending
     * order in the result set. SQL keyword <code>ASC</code> will then be added
     * to all column names in the automatically generated <code>ORDER BY</code>
     * clause.
     */
    public final static int ORDER_ASC = 1;

    /**
     * Used with fetch methods to indicate that the caller wants descending
     * order in the result set. SQL keyword <code>DESC</code> will then be
     * added to all column names in the automatically generated
     * <code>ORDER BY</code> clause.
     */
    public final static int ORDER_DESC = 2;


    ///////////////////////////////////////////////////////////////////////////
    // Utility classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Connection wrapper that disables the <code>close()</code>
     * method.
     */
    private static final class UnclosableConnection
        extends ConnectionWrapper {

        /**
         * Constructs an "unclosable" connection.
         */
        public UnclosableConnection(Connection con) {
            super(con);
        }

        /**
         * Does nothing.
         */
        public void close() { /* do nothing */ }
    }


    /**
     * A pinned DB connection wrapper to be stored in the
     * <code>pinnedConnections</code> storage for threads.
     */
    private static final class PinnedConnection {

        /**
         * The connection.
         */
        private final UnclosableConnection con;

        /**
         * Timestamp when the connection was pinned.
         */
        public final long pinnedOn;

        /**
         * Indicates if the connection has been pinned for running
         * updates on it.
         */
        private final boolean forUpdate;

        /**
         * How many times this pinned connection has been requested
         * since it was pinned.
         */
        private int nGets;


        /**
         * Creates a wrapper for a connection.
         */
        public PinnedConnection(Connection con, boolean forUpdate) {

            this.con = new UnclosableConnection(con);
            this.pinnedOn = System.currentTimeMillis();
            this.forUpdate = forUpdate;
            this.nGets = 0;
        }


        /**
         * Returns the wrapped connection and increments the
         * <code>nGets</code> counter.
         */
        public UnclosableConnection getConnection(boolean forUpdate) {

            if(forUpdate && !this.forUpdate)
                throw new IllegalArgumentException("Tried to get database connection for updating data" +
                                                   " while the connection was pinned to the thread" +
                                                   " only for reading data.");

            this.nGets++;
            return this.con;
        }

        /**
         * Closes the wrapped connection.
         *
         * @throws SQLException if the connection could not be closed.
         */
        public void close()
            throws SQLException {

            this.con.getWrappedConnection().close();
        }

        /**
         * Returns the counter indicating how many times this pinned
         * connection has been requested since it was pinned.
         */
        public int getNGets() {

            return this.nGets;
        }
    }


    /**
     * Used by the <code>singleElementArrayListPool</code> object
     * pool.
     */
    private static final class SingleElementArrayListFactory
        extends PooledObjectFactoryAdapter {

        /**
         * Instantiates and returns a single element empty
         * <code>ArrayList</code>.
         *
         * @return the new array list.
         */
        public Object createObject() {

            return new ArrayList(1);
        }

        /**
         * Clears the specified <code>ArrayList</code>.
         *
         * @param o the object to clear, must be an instance of
         * <code>ArrayList</code>.
         */
        public void resetObject(Object o) {

            ((ArrayList)o).clear();
        }
    }


    /**
     * Base class for standard type convertors used in the
     * <code>setField</code> method.
     */
    private static abstract class TypeConvertor {

        /**
         * Convert the specified object to a certain type. The input object
         * should be of a specific type accepted by the concrete method
         * implementation.
         *
         * @param value the value to convert.
         *
         * @return the value converted to the specific type.
         */
        public abstract Object convert(Object value);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Inner classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Used by the <code>executeFetch</code> method to store the
     * current state.
     */
    private final class FetchContext {

        /**
         * Parameters of a field in the result set.
         */
        public final class FieldDescriptor {

            /**
             * Field type, regular field.
             */
            public final static int FT_REGULAR = 0;

            /**
             * Field type, nested array of DMs.
             */
            public final static int FT_ARRAY = 1;

            /**
             * Field type, single nested DM.
             */
            public final static int FT_NESTEDDM = 2;

            /**
             * Field type, to be ignored.
             */
            public final static int FT_IGNORE = 3;


            /**
             * The corresponding field object. May be
             * <code>null</code>.
             */
            public final Field field;

            /**
             * The field's kind, one of the FT_xxx constants.
             */
            public final int type;

            /**
             * Name of the DM field.
             */
            public final String fieldName;

            /**
             * Type of the DM field.
             */
            public final Class fieldClass;


            /**
             * Constructs a descriptor.
             *
             * @param dmClass class of the DM.
             * @param columnLabel the column lable in the result set.
             */
            public FieldDescriptor(Class dmClass, String columnLabel)
                throws BLException {

                if(dmClass == DynamicDataModel.class) {
                    this.field = null;
                    this.type = FT_REGULAR;
                    this.fieldName = columnLabel;
                    this.fieldClass = Object.class;
                } else {
                    this.field = config.getDMField(dmClass, columnLabel);
                    if(this.field == null) {
                        this.type = FT_IGNORE;
                        this.fieldName = null;
                        this.fieldClass = null;
                    } else {
                        this.fieldName = this.field.getName();
                        this.fieldClass = this.field.getType();
                        if(this.field.getType().isArray())
                            this.type = FT_ARRAY;
                        else if((DataModel.class).isAssignableFrom(this.field.getType()))
                            this.type = FT_NESTEDDM;
                        else
                            this.type = FT_REGULAR;
                    }
                }
            }
        }


        /**
         * The result set.
         */
        public final ResultSet rs;

        /**
         * The result set's meta data.
         */
        public final ResultSetMetaData rsmeta;

        /**
         * Total number of columns in the result set.
         */
        public final int columnCount;

        /**
         * Descriptors for all fields in the result set.
         */
        private final FieldDescriptor [] fieldDescriptors;

        /**
         * Callbacks.
         */
        public final FetchResultProcessorHandler [] hooks;

        /**
         * Current DM nesting level starting from zero.
         */
        public int level;

        /**
         * Current list of parent DMs keys' inedexes.
         */
        public final List upperKeysInds;

        /**
         * Current list of parent DMs keys' values.
         */
        public final List upperKeysVals;

        /**
         * Indicates if a <code>rs.next()</code> has not hit the end
         * of the result set yet.
         */
        public boolean onValidRow;

        /**
         * Number fo rows processed. Each call to the
         * <code>next</code> method incremements this counter.
         */
        public int rowsProcessed;


        /**
         * Constructs a new context. When this constructor is called
         * the result set is already set to the first row to fetch.
         *
         * @param rs the result set to process.
         * @param hooks callbacks.
         */
        public FetchContext(ResultSet rs,
                            FetchResultProcessorHandler [] hooks)
            throws SQLException {

            this.rs = rs;
            this.rsmeta = rs.getMetaData();
            this.columnCount = this.rsmeta.getColumnCount();
            this.fieldDescriptors = new FieldDescriptor[this.columnCount];
            this.hooks = hooks;
            this.level = 0;
            this.upperKeysInds = new ArrayList();
            this.upperKeysVals = new ArrayList();
            this.onValidRow = true;
            this.rowsProcessed = 0;
        }


        /**
         * Gets a descriptor for a result set column.
         *
         * @param dmClass class of the DM being processed.
         * @param index index of the column in the result set.
         *
         * @return the field descriptor for the specified column.
         */
        public FieldDescriptor getFieldDescriptor(Class dmClass, int index)
            throws BLException, SQLException {

            FieldDescriptor res = this.fieldDescriptors[index - 1];

            if(res == null) {
                res = new FieldDescriptor(dmClass, this.rsmeta.getColumnLabel(index));
                this.fieldDescriptors[index - 1] = res;
            }

            return res;
        }

        /**
         * Moves the cursor one row forward and stores if the is a
         * result in the result set in the <code>onValidRow</code>
         * member variable.
         *
         * @return <code>true</code> if the cursor is pointed to a
         * result now, <code>false</code> if the end of the result set
         * has been reached.
         *
         * @throws SQLException if the was a database error.
         */
        public boolean next()
            throws SQLException {

            this.rowsProcessed++;

            return (this.onValidRow = this.rs.next());
        }
    }


    /**
     * Used by the <code>Context</code> object pool.
     */
    private static final class ContextFactory
        extends PooledObjectFactoryAdapter {

        /**
         * Instantiates and returns a
         * <code>FetchPrimaryChunkExecutor</code>.
         *
         * @return the new <code>FetchPrimaryChunkExecutor</code>.
         */
        public Object createObject() {

            return new Context();
        }

        /**
         * Clears the specified object.
         *
         * @param o the object to clear.
         */
        public void resetObject(Object o) {

            ((Context)o).reset();
        }
    }


    /**
     * Context of a single DAO call, which is passed all through the stack
     * of internal method calls.
     */
    private static final class Context {

        /**
         * The query descriptor or <code>null</code>.
         */
        public DAOConfig.QueryDescriptor queryDesc;

        /**
         * The conditions passed with the call, may be <code>null</code>.
         */
        public Set conditions;

        /**
         * The query name.
         */
        public String queryName;

        /**
         * SQL types for query parameters when using <code>setObject</code>.
         */
        public Integer [] paramSqlTypes;

        /**
         * Sql types for query parameters when using <code>setNull</code>.
         */
        public Integer [] paramSqlTypesForNulls;


        /**
         * Sets up the context for a named query execution.
         *
         * @param config our DAO configuration.
         * @param queryName the query name.
         * @param conditions the conditions passed with the call.
         *
         * @throws BLException if no query exists with the specified name.
         */
        public void setup(DAOConfig config,
                          String queryName,
                          Set conditions)
            throws BLException {

            this.queryDesc = config.getQueryDescriptor(queryName);
            this.conditions = conditions;
            this.queryName = this.queryDesc.name;

            List placeholders = new ArrayList(this.queryDesc.paramPlaceholders.length);
            for(int i = 0; i < this.queryDesc.paramPlaceholders.length; i++) {
                if((this.queryDesc.paramPlaceholders[i].conditionName == null) ||
                   ((conditions != null) && conditions.contains(this.queryDesc.paramPlaceholders[i].conditionName)))
                    placeholders.add(this.queryDesc.paramPlaceholders[i]);
            }

            this.paramSqlTypes = new Integer[placeholders.size()];
            this.paramSqlTypesForNulls = new Integer[placeholders.size()];
            Iterator it = placeholders.iterator();
            int i = 0;
            while(it.hasNext()) {
                DAOConfig.QueryDescriptor.ParamPlaceholder pp =
                    (DAOConfig.QueryDescriptor.ParamPlaceholder)it.next();
                this.paramSqlTypes[i] = pp.sqlType;
                this.paramSqlTypesForNulls[i] = pp.sqlTypeForNulls;
                i++;
            }
        }

        /**
         * Clears all context fields before sending it back to the pool.
         */
        void reset() {

            this.queryDesc = null;
            this.conditions = null;
            this.queryName = null;
            this.paramSqlTypes = null;
            this.paramSqlTypesForNulls = null;
        }
    }


    /**
     * Used by the <code>fetchPrimaryChunkExecutorPool</code> object
     * pool.
     */
    private static final class FetchPrimaryChunkExecutorFactory
        extends PooledObjectFactoryAdapter {

        /**
         * Our DAO.
         */
        private final DAO dao;


        /**
         * Constructs a factory.
         */
        public FetchPrimaryChunkExecutorFactory(DAO dao) {

            this.dao = dao;
        }


        /**
         * Instantiates and returns a
         * <code>FetchPrimaryChunkExecutor</code>.
         *
         * @return the new <code>FetchPrimaryChunkExecutor</code>.
         */
        public Object createObject() {

            return new FetchPrimaryChunkExecutor(this.dao);
        }

        /**
         * Clears the specified object.
         *
         * @param o the object to clear.
         */
        public void resetObject(Object o) {

            ((PrimaryChunkExecutor)o).reset();
        }
    }


    /**
     * Used by the <code>updatePrimaryChunkExecutorPool</code> object
     * pool.
     */
    private static final class UpdatePrimaryChunkExecutorFactory
        extends PooledObjectFactoryAdapter {

        /**
         * Our DAO.
         */
        private final DAO dao;


        /**
         * Constructs a factory.
         */
        public UpdatePrimaryChunkExecutorFactory(DAO dao) {

            this.dao = dao;
        }


        /**
         * Instantiates and returns a
         * <code>UpdatePrimaryChunkExecutor</code>.
         *
         * @return the new <code>UpdatePrimaryChunkExecutor</code>.
         */
        public Object createObject() {

            return new UpdatePrimaryChunkExecutor(this.dao);
        }

        /**
         * Clears the specified object.
         *
         * @param o the object to clear.
         */
        public void resetObject(Object o) {

            ((PrimaryChunkExecutor)o).reset();
        }
    }


    /**
     * Common abstract parent used by <code>executeNamedQuery</code>
     * private method to generalize execution of named fetches and
     * updates.
     */
    private static abstract class PrimaryChunkExecutor {

        /**
         * Reference to our parent DAO.
         */
        protected final DAO dao;

        /**
         * The call context.
         */
        protected Context ctx;


        /**
         * Constructs a stub.
         */
        protected PrimaryChunkExecutor(DAO dao) {

            this.dao = dao;
        }


        /**
         * Basic reset.
         */
        public void reset() {

            this.ctx = null;
        }


        /**
         * Sets basic query info.
         */
        public final void setQueryInfo(Context ctx) {

            this.ctx = ctx;
        }

        /**
         * Executes the primary sql chunk.
         *
         * @param sqlChunk the descriptor of the primary sql chunk.
         * @param chunkParams query parameters for the chunk.
         *
         * @throws BLException if an error happened.
         */
        public abstract void execute(DAOConfig.QueryDescriptor.SqlChunk sqlChunk,
                                     Object [] chunkParams)
            throws BLException;
    }


    /**
     * Primary sql chunk executor for fetches.
     */
    private static final class FetchPrimaryChunkExecutor
        extends PrimaryChunkExecutor {

        // additional request info
        private String [] orderBy;
        private int order;
        private int page;
        private int pageSize;
        private FetchResultProcessorHandler [] hooks;
        private FetchResultDescriptor fetchResultDescriptor;

        /**
         * The query execution result.
         */
        public DataModel [] res;


        /**
         * Constructs an executor.
         */
        public FetchPrimaryChunkExecutor(DAO dao) {
            super(dao);
        }


        /**
         * Reset.
         */
        public void reset() {
            super.reset();

            this.orderBy = null;
            this.hooks = null;
            this.fetchResultDescriptor = null;
        }


        /**
         * Sets basic query info.
         */
        public void setQueryInfo(Context ctx,
                                 String [] orderBy,
                                 int order,
                                 int page,
                                 int pageSize,
                                 FetchResultProcessorHandler [] hooks,
                                 FetchResultDescriptor fetchResultDescriptor) {
            this.setQueryInfo(ctx);

            this.orderBy = orderBy;
            this.order = order;
            this.page = page;
            this.pageSize = pageSize;
            this.hooks = hooks;
            this.fetchResultDescriptor = fetchResultDescriptor;
        }

        /**
         * Executes the primary sql chunk.
         */
        public void execute(DAOConfig.QueryDescriptor.SqlChunk sqlChunk,
                            Object [] chunkParams)
            throws BLException {

            // get the DM class
            Class dmClass = this.dao.config.getQueryDMClass(this.ctx.queryName);

            // get the query text into a string buffer
            StringBuffer query = new StringBuffer(sqlChunk.getSql(this.ctx.conditions));

            // add "order by" clause if needed
            if((this.orderBy != null) && (this.orderBy.length > 0)) {
                query.append(" ORDER BY ");
                for(int i = 0; i < this.orderBy.length; i++) {
                    if(i > 0)
                        query.append(", ");
                    query.append(this.orderBy[i]);
                    switch(this.order) {
                    case ORDER_ASC:
                        query.append(" ASC");
                        break;
                    case ORDER_DESC:
                        query.append(" DESC");
                    }
                }
            }

            // execute the query
            this.res = this.dao.executeFetch_Internal(this.ctx,
                                                      dmClass,
                                                      query.toString(),
                                                      chunkParams,
                                                      this.page,
                                                      this.pageSize,
                                                      this.hooks,
                                                      this.fetchResultDescriptor);
        }
    }


    /**
     * Primary sql chunk executor for updates.
     */
    private static final class UpdatePrimaryChunkExecutor
        extends PrimaryChunkExecutor {

        /**
         * The query execution result.
         */
        int res;


        /**
         * Constructs an executor.
         */
        public UpdatePrimaryChunkExecutor(DAO dao) {
            super(dao);
        }


        /**
         * Executes the primary sql chunk.
         */
        public void execute(DAOConfig.QueryDescriptor.SqlChunk sqlChunk,
                            Object [] chunkParams)
            throws BLException {

            this.res = this.dao.executeUpdate_Internal(this.ctx,
                                                       sqlChunk.getSql(this.ctx.conditions),
                                                       chunkParams,
                                                       true,
                                                       true);
        }
    }


    /**
     * Represents statistics for one SQL query.
     */
    private static final class QueryStatsRecord {

        /**
         * The query name, can be <code>null</code>.
         */
        public final String name;

        /**
         * The query text.
         */
        public final String queryText;

        /**
         * Statistics for query executions.
         */
        public final EventStatistics execStats;

        /**
         * Statistics for query result set processings. Stays just
         * initialized for queries that do not generate any result
         * sets.
         */
        public final EventStatistics procStats;


        /**
         * Constructs new stats record for the specified query.
         *
         * @param name name of the query, may be <code>null</code> for
         * direct queries.
         * @param queryText the query's SQL text.
         */
        public QueryStatsRecord(String name, String queryText) {

            this.name = name;
            this.queryText = queryText;
            this.execStats = new EventStatistics();
            this.procStats = new EventStatistics();
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // The DAO internal static data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Type convertors for objects of <code>java.lang.Numeric</code>.
     */
    private static final Map NUMERIC_TYPE_CONVERTORS;
    static {
        NUMERIC_TYPE_CONVERTORS = new HashMap(16);
        NUMERIC_TYPE_CONVERTORS.put(Boolean.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return new Boolean(((Number)value).intValue() != 0);
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Boolean.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new Boolean(((Number)value).intValue() != 0);
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Byte.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Byte ? value : new Byte(((Number)value).byteValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Byte.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Byte ? value : new Byte(((Number)value).byteValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Short.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Short ? value : new Short(((Number)value).shortValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Short.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Short ? value : new Short(((Number)value).shortValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Integer.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Integer ? value : new Integer(((Number)value).intValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Integer.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Integer ? value : new Integer(((Number)value).intValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Long.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Long ? value : new Long(((Number)value).longValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Long.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Long ? value : new Long(((Number)value).longValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Float.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Float ? value : new Float(((Number)value).floatValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Float.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Float ? value : new Float(((Number)value).floatValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Double.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Double ? value : new Double(((Number)value).doubleValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(Double.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof Double ? value : new Double(((Number)value).doubleValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(BigDecimal.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof BigDecimal ? value : new BigDecimal(((Number)value).doubleValue()));
            }
        });
        NUMERIC_TYPE_CONVERTORS.put(BigInteger.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (value instanceof BigInteger ? value : new BigInteger(String.valueOf(((Number)value).longValue())));
            }
        });
    }

    /**
     * Type convertors for objects of <code>java.lang.Boolean</code>.
     */
    private static final Map BOOLEAN_TYPE_CONVERTORS;
    static {
        BOOLEAN_TYPE_CONVERTORS = new HashMap(14);
        BOOLEAN_TYPE_CONVERTORS.put(Byte.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return new Byte(((Boolean)value).booleanValue() ? (byte)1 : (byte)0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Byte.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new Byte(((Boolean)value).booleanValue() ? (byte)1 : (byte)0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Short.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return new Short(((Boolean)value).booleanValue() ? (short)1 : (short)0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Short.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new Short(((Boolean)value).booleanValue() ? (short)1 : (short)0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Integer.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return new Integer(((Boolean)value).booleanValue() ? 1 : 0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Integer.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new Integer(((Boolean)value).booleanValue() ? 1 : 0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Long.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return new Long(((Boolean)value).booleanValue() ? 1L : 0L);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Long.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new Long(((Boolean)value).booleanValue() ? 1L : 0L);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Float.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return new Float(((Boolean)value).booleanValue() ? 1.0F : 0.0F);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Float.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new Float(((Boolean)value).booleanValue() ? 1.0F : 0.0F);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Double.TYPE, new TypeConvertor() {
            public Object convert(Object value) {
                return new Double(((Boolean)value).booleanValue() ? 1.0 : 0.0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(Double.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new Double(((Boolean)value).booleanValue() ? 1.0 : 0.0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(BigDecimal.class, new TypeConvertor() {
            public Object convert(Object value) {
                return new BigDecimal(((Boolean)value).booleanValue() ? 1.0 : 0.0);
            }
        });
        BOOLEAN_TYPE_CONVERTORS.put(BigInteger.class, new TypeConvertor() {
            public Object convert(Object value) {
                return (((Boolean)value).booleanValue() ? BigInteger.ONE : BigInteger.ZERO);
            }
        });
    }


    ///////////////////////////////////////////////////////////////////////////
    // The DAO configuration and data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Our config object.
     */
    final DAOConfig config;

    /**
     * Log to be used for DAO diagnostic messages.
     */
    private final Log log;

    /**
     * Set to <code>true</code> if this <code>DAO</code> instance
     * should use transactional database connections from the
     * application context. Otherwise, the <code>DAO</code> uses
     * <code>getNoTXConnection</code> method on the application
     * context to get database connections. This field is ignored when
     * the <code>DAO</code> is configured to use a particular database
     * connection.
     */
    private final boolean isTransactional;

    /**
     * Database connection to be used or <code>null</code> if a
     * database connection should be retrived from the data source
     * each time it is needed. If this connection is set it is
     * "unclosable" so calls to <code>close()</code> method in the
     * DAO's internal methods have no effect.
     */
    private final Connection con;

    /**
     * Pool of DAO call contexts (objects of <code>Context</code>).
     */
    private final AdaptiveObjectPool contextsPool;

    /**
     * Pool of <code>ArrayList</code> objects configured for a single
     * element. They are used in the <code>processResultSetRow</code>
     * method.
     */
    private final AdaptiveObjectPool singleElementArrayListPool;

    /**
     * Pool of <code>FetchPrimaryChunkExecutor</code> objects.
     */
    private final AdaptiveObjectPool fetchPrimaryChunkExecutorPool;

    /**
     * Pool of <code>UpdatePrimaryChunkExecutor</code> objects.
     */
    private final AdaptiveObjectPool updatePrimaryChunkExecutorPool;

    /**
     * Contains stats records for each query. The keys are query SQL
     * texts (values of <code>String</code>) and the values are the
     * stats records (objects of <code>QueryStatsRecord</code>).
     */
    private final Map queryStats;

    /**
     * Storage of pinned connections associated with threads. Objects
     * of <code>PinnedConnection</code> are stored in this storage.
     */
    private static final ThreadLocal pinnedConnections = new ThreadLocal();


    ///////////////////////////////////////////////////////////////////////////
    // The DAO initialization
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Common constructor called by all other constructors.
     *
     * @param config value for the <code>config</code> field.
     * @param isTransactional value for the
     * <code>isTransactional</code> field.
     * @param con value for the <code>con</code> field.
     */
    private DAO(DAOConfig config, boolean isTransactional, Connection con) {

        this.config = config;
        this.log = config.getApplicationContext().getLog(LOG_SUBSYS);
        this.isTransactional = isTransactional;
        this.con = con;
        this.contextsPool = new AdaptiveObjectPool(new ContextFactory());
        this.singleElementArrayListPool = new AdaptiveObjectPool(new SingleElementArrayListFactory());
        this.fetchPrimaryChunkExecutorPool = new AdaptiveObjectPool(new FetchPrimaryChunkExecutorFactory(this));
        this.updatePrimaryChunkExecutorPool = new AdaptiveObjectPool(new UpdatePrimaryChunkExecutorFactory(this));
        this.queryStats = new HashMap();
    }

    /**
     * Constructs a DAO instance, which automatically retrieves database
     * connections from the application context every time it needs it. The
     * data source used is the one associated with the specified DAO config
     * object.
     *
     * <p>Depending on the value of the <code>isTransactional</code> parameter,
     * the DAO will be using application context's standard
     * <code>getConnection</code> method, which returns transactional database
     * connections, or it will be using low-level
     * <code>getNoTXConnection</code> method.
     * 
     * @param config the <code>DAOConfig</code> object.
     * @param isTransactional set to <code>true</code> if this DAO should use
     * transactional database connections, which is the normal mode. Otherwise,
     * the DAO will be using <code>getNoTXConnection</code> method on the
     * application context to get database connections.
     */
    public DAO(DAOConfig config, boolean isTransactional) {
        this(config, isTransactional, null);
    }

    /**
     * Constructs a DAO instance configured to use a certain database
     * connection. The application code outside the DAO must completely take
     * care of opening, managing and closing the database connection.
     *
     * @param config the <code>DAOConfig</code> object.
     * @param con the database connection to be used by the DAO. Note, that it
     * should be a connection borrowed from the same data source, with which
     * the specified <code>DAOConfig</code> object is associated. It is
     * required, because the <code>DAOConfig</code> gathers information about
     * the database connection's features when it is created. That information
     * is later used by the DAO to adjust the way it operates on the
     * connection.
     */
    public DAO(DAOConfig config, Connection con) {
        this(config, false, new UnclosableConnection(con));
    }


    ///////////////////////////////////////////////////////////////////////////
    // Internal service
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Returns a database connection.
     *
     * @param forUpdate should be set to <code>true</code> if the
     * requested connection will be used for database updates.
     *
     * @return database connection.
     *
     * @throws SQLException if the connection cannot be retrieved
     * because of the database level error.
     */
    private Connection getConnection(boolean forUpdate)
        throws SQLException {

        if(this.con != null)
            return this.con;

        PinnedConnection pcon = (PinnedConnection)pinnedConnections.get();
        if(pcon != null)
            return pcon.getConnection(forUpdate);

        return (this.isTransactional ?
                this.config.getApplicationContext().
                getConnection(this.config.getDataSourceName(), forUpdate) :
                this.config.getApplicationContext().
                getNoTXConnection(this.config.getDataSourceName()));
    }

    /**
     * Closes the specified database connection resources.
     *
     * @param con DB connection to close or <code>null</code>.
     * @param stmt statement to close or <code>null</code>.
     * @param rs result set to close or <code>null</code>.
     *
     * @return exception to throw if could not close any one of the
     * specified resources, otherwise returns <code>null</code>.
     */
    private BLException closeDBResources(Connection con,
                                         Statement stmt,
                                         ResultSet rs) {

        BLException eToThrow = null;

        if(rs != null)
            try {
                rs.close();
            } catch(SQLException e) {
                eToThrow = new BLException("Could not close the result set in the DAO.", e);
            }
        if(stmt != null)
            try {
                stmt.close();
            } catch(SQLException e) {
                if(eToThrow == null)
                    eToThrow = new BLException("Could not close the statement in the DAO.", e);
            }
        if(con != null)
            try {
                con.close();
            } catch(SQLException e) {
                if(eToThrow == null)
                    eToThrow = new BLException("Could not close the database connection in the DAO.", e);
            }

        return eToThrow;
    }

    /**
     * Sets field on a dynamic DM.
     *
     * @param dm the target dynamic DM instance.
     * @param fieldName field name.
     * @param value new value for the field.
     *
     * @throws BLException if an error happens, for example if the field is
     * already set in the DM (duplicate labels in the result set).
     */
    private Object setDynamicField(DynamicDataModel dm,
                                   String fieldName,
                                   Object value)
        throws BLException {

        // call the plugins
        ResultSetFieldProcessor [] plugins = this.config.getResultSetFieldProcessors();
        if(plugins != null) {
            for(int i = 0; i < plugins.length; i++)
                value = plugins[i].processResultSetField(fieldName,
                                                         Object.class,
                                                         value);
        }

        // set the field
        if(dm.put(fieldName, value) != null)
            throw new BLException("Result set has two or more columns named \"" +
                                  fieldName + "\" so it cannot be used with a dynamic DM.");

        // return what we've set
        return value;
    }

    /**
     * Sets field on a DM via reflection.
     *
     * <p>This method is capable of performing some basic type conversions
     * between all <code>java.lang.Number</code> types with addition of
     * <code>boolean/Boolean</code>.
     *
     * @param dm the target DM instance.
     * @param field the field.
     * @param value the value.
     *
     * @return the value actually set into the field, which may be something
     * different from what we've passed.
     *
     * @throws BLException if any result set field processor plugin has thrown
     * it.
     * @throws IllegalAccessException when there are problems accessing the
     * specified field in the specified object through the Java reflection
     * mechanism.
     * @throws IllegalArgumentException when we are trying to set a value of an
     * incompatible with the DM field type.
     * @throws NullPointerException if we are setting <code>null</code> to a
     * primitive field.
     */
    private Object setReflectField(DataModel dm,
                                   Field field,
                                   Object value)
        throws BLException, IllegalAccessException {

        // call the plugins
        ResultSetFieldProcessor [] plugins = this.config.getResultSetFieldProcessors();
        if(plugins != null) {
            String fieldName = field.getName();
            Class fieldClass = field.getType();
            for(int i = 0; i < plugins.length; i++)
                value = plugins[i].processResultSetField(fieldName,
                                                         fieldClass,
                                                         value);
        }

        // do standard type conversions
        if((value != null) &&
           !field.getType().isAssignableFrom(value.getClass())) {

            TypeConvertor tc = null;
            if(value instanceof Number) {
                tc = (TypeConvertor)NUMERIC_TYPE_CONVERTORS.get(field.getType());
            } else if(value instanceof Boolean) {
                tc = (TypeConvertor)BOOLEAN_TYPE_CONVERTORS.get(field.getType());
            }
            if(tc != null)
                value = tc.convert(value);
        }

        // try to set the field
        try {
            field.set(dm, value);
        } catch(IllegalArgumentException e) {
            this.log.error("Incompatibale type for DM [" + dm.getClass().getName() +
                           "] field " + field.getName() +
                           " [" + field.getType().getName() +
                           "] and result set column [" +
                           (value != null ?
                            value.getClass().getName() + "/" + value.toString() :
                            "null") + "]");
            throw e;
        } catch(NullPointerException e) {
            if((value == null) && field.getType().isPrimitive())
                this.log.error("Cannot set null to primitive field " + field.getName() +
                               " of DM [" + dm.getClass().getName() +
                               "], use a wrapper class to define the field to make it nullable.");
            throw e;
        }

        // return what we've set
        return value;
    }

    /**
     * For each array element (an element which has an array type) in
     * the <code>queryParams</code> array finds a double question mark
     * sequence ("??") in the specified SQL query and replaces it with
     * a comma separated sequence of single question marks according
     * to the number of elements in the array.
     *
     * @param query SQL query text possibly containing double question
     * marks.
     * @param queryParams array of query parameters possibly
     * containing array elements. May be <code>null</code>.
     *
     * @return the SQL query text with double question marks expanded.
     *
     * @throws BLException if number of double question mark sequences
     * in the query does not match the number of array elements in the
     * <code>queryParams</code> array.
     */
    private String expandDoubleQuestionMarks(String query,
                                             Object [] queryParams)
        throws BLException {

        if(queryParams == null)
            return query;

        StringBuffer expandedQuery = null;

        int lastDQIndex = 0;
        for(int i = 0; i < queryParams.length; i++) {
            if((queryParams[i] != null) && queryParams[i].getClass().isArray()) {
                int dqIndex = query.indexOf("??", lastDQIndex);
                if(dqIndex < 0)
                    throw new BLException("Could not find a ?? in the query for an array query parameter.");
                if(expandedQuery == null)
                    expandedQuery = new StringBuffer();
                expandedQuery.append(query.substring(lastDQIndex, dqIndex));
                lastDQIndex = dqIndex + 2;
                Object [] subParams = (Object [])queryParams[i];
                for(int j = 0; j < subParams.length; j++) {
                    if(j > 0)
                        expandedQuery.append(", ?");
                    else
                        expandedQuery.append('?');
                }
            }
        }

        if(expandedQuery == null)
            return query;

        expandedQuery.append(query.substring(lastDQIndex));
        return expandedQuery.toString();
    }

    /**
     * Sets parameters in the specified JDBC prepared statement to the
     * values from the <code>queryParams</code> array.
     *
     * @param ctx the call context.
     * @param pstmt the JDBC prepared statement where to set the
     * params.
     * @param queryParams array of query parameters. The elements can
     * be regular objects or arrays of objects. May be
     * <code>null</code>.
     *
     * @throws SQLException if could not set a prepared statement
     * parameter.
     */
    private void setQueryParams(Context ctx,
                                PreparedStatement pstmt,
                                Object [] queryParams)
        throws SQLException {

        int phIndex = 1;
        for(int i = 0; i < queryParams.length; i++) {
            Object value = queryParams[i];

            // call plugins
            QueryParamProcessor [] plugins = this.config.getQueryParamProcessors();
            if(plugins != null) {
                for(int j = 0; j < plugins.length; j++)
                    value = plugins[j].processQueryParam(pstmt, phIndex, value);
            }

            // set the param
            phIndex = this.setQueryParam(ctx, pstmt, phIndex, i, value);
        }
    }

    /**
     * Sets a parameter in a prepared statement. Correctly processes
     * array values.
     *
     * @param ctx the call context.
     * @param pstmt the prepared statement.
     * @param phIndex index of a parameter placeholder in the prepared
     * statement.
     * @param paramIndex parameter index in the original query parameters
     * array.
     * @param value the parameter value, which can be an array.
     *
     * @return index of the next unset parameter placeholder in the
     * prepared statement.
     *
     * @throws SQLException if could not set a parameter in the prepared
     * statement.
     */
    private int setQueryParam(Context ctx,
                              PreparedStatement pstmt,
                              int phIndex,
                              int paramIndex,
                              Object value)
        throws SQLException {

        if(value != null) {

            if(value.getClass().isArray()) { // array element
                Object [] subParams = (Object [])value;
                for(int i = 0; i < subParams.length; i++)
                    phIndex = this.setQueryParam(ctx, pstmt, phIndex, paramIndex, subParams[i]);

            } else {                         // regular object
                QueryParamSetter queryParamSetter = this.config.getQueryParamSetter();
                if(queryParamSetter != null) {
                    queryParamSetter.setQueryParam(pstmt,
                                                   phIndex++,
                                                   value,
                                                   (ctx.paramSqlTypes != null ? ctx.paramSqlTypes[paramIndex] : null));
                } else {
                    if((ctx.paramSqlTypes != null) && (ctx.paramSqlTypes[paramIndex] != null))
                        pstmt.setObject(phIndex++, value, ctx.paramSqlTypes[paramIndex].intValue());
                    else
                        pstmt.setObject(phIndex++, value);
                }
            }

        } else {                             // a null
            if((ctx.paramSqlTypesForNulls != null) && (ctx.paramSqlTypesForNulls[paramIndex] != null))
                pstmt.setNull(phIndex++, ctx.paramSqlTypesForNulls[paramIndex].intValue());
            else
                pstmt.setNull(phIndex++, java.sql.Types.VARCHAR);
        }

        return phIndex;
    }

    /**
     * Returns stats record for the specified query.
     *
     * @param queryName name of the query, may be <code>null</code>
     * for direct queries.
     * @param queryText the query's SQL text.
     *
     * @return stats record to be used.
     */
    private QueryStatsRecord getQueryStatsRecord(String queryName,
                                                 String queryText) {

        QueryStatsRecord stats;
        synchronized(this.queryStats) {
            stats = (QueryStatsRecord)this.queryStats.get(queryText);
            if(stats == null) {
                stats = new QueryStatsRecord(queryName, queryText);
                this.queryStats.put(queryText, stats);
            }
        }

        return stats;
    }

    /**
     * Converts a query parameters map to array suitable for passing
     * to <code>executeXXX</code> methods.
     *
     * If a parameter is not present in the specified map a
     * <code>null</code> will be set in the corresponding element of
     * the resulting array.
     *
     * @param ctx the call context.
     * @param queryParamsMap query parameters as a map. The keys are
     * parameters names (objects <code>String</code>), the values are
     * parameter values.
     *
     * @return query parameters as an array for passing to
     * <code>executeXXX</code> methods.
     */
    private Object [] getQueryParamsArray(Context ctx,
                                          Map queryParamsMap) {

        if(queryParamsMap == null)
            return null;

        List resList = new ArrayList(ctx.queryDesc.paramPlaceholders.length);
        for(int i = 0; i < ctx.queryDesc.paramPlaceholders.length; i++) {
            if((ctx.queryDesc.paramPlaceholders[i].conditionName == null) ||
               ((ctx.conditions != null) && ctx.conditions.contains(ctx.queryDesc.paramPlaceholders[i].conditionName)))
                resList.add(queryParamsMap.get(ctx.queryDesc.paramPlaceholders[i].name));
        }

        return resList.toArray();
    }

    /**
     * Gets the body of the primary sql chunk of the named query. No
     * condition tags are processed.
     *
     * @param queryName the name of the query.
     *
     * @return the primary (result) sql chunk of the named query.
     *
     * @throws BLException if there is no query with the specified
     * name.
     */
    private String getPrimaryQuery(String queryName)
        throws BLException {

        DAOConfig.QueryDescriptor qd =
            this.config.getQueryDescriptor(queryName);

        for(int i = 0; i < qd.sqls.length; i++)
            if(qd.sqls[i].isResult)
                return qd.sqls[i].getSql(null);

        // actually never can happen
        return qd.sqls[qd.sqls.length - 1].getSql(null);
    }

    /**
     * Finds corresponding SQL type for a Java type according to the
     * JDBC specification.
     *
     * @param javaType the Java type.
     *
     * @return wrapped SQL type (see <code>java.sql.Types</code>).
     */
    private Integer getSQLTypeForJavaType(Class javaType) {

        if((String.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.VARCHAR);
        else if((javaType == Integer.TYPE) || (Integer.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.INTEGER);
        else if((javaType == Long.TYPE) || (Long.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.BIGINT);
        else if((java.math.BigDecimal.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.NUMERIC);
        else if((java.sql.Timestamp.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.TIMESTAMP);
        else if((java.sql.Date.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.DATE);
        else if((java.sql.Time.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.TIME);
        else if((javaType == Boolean.TYPE) || (Boolean.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.BIT);
        else if((javaType == Character.TYPE) || (Character.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.CHAR);
        else if((javaType == Double.TYPE) || (Double.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.DOUBLE);
        else if((javaType == Float.TYPE) || (Float.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.REAL);
        else if((javaType == Byte.TYPE) || (Byte.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.TINYINT);
        else if((javaType == Short.TYPE) || (Short.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.SMALLINT);
        else if((java.sql.Blob.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.BLOB);
        else if((java.sql.Clob.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.CLOB);
        else if((java.sql.Array.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.ARRAY);
        else if((java.sql.Struct.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.STRUCT);
        else if((java.sql.Ref.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.REF);
        else if((java.net.URL.class).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.DATALINK);
        else if(javaType.isArray() && (Byte.TYPE).isAssignableFrom(javaType))
            return new Integer(java.sql.Types.VARBINARY);
        else
            return new Integer(java.sql.Types.JAVA_OBJECT);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Special public service interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * "Pins down" a database connection to the current thread. All subsequent
     * calls to any DAO instance (except ones created via
     * <code>DAO(DAOConfig config, Connection con)</code> constructor) from the
     * same thread will be using the same connection until
     * <code>unpinConnection</code> is called from within the thread.
     *
     * <p>It is highly recommended to have an <code>unpinConnection</code> call
     * in a <code>finally</code> block!
     *
     * <p>This method does nothing if called on a DAO created via
     * <code>DAO(DAOConfig config, Connection con)</code> constructor.
     *
     * @param forUpdate indicates if your are going to run queries that modify
     * data in the database. Set to <code>false</code> if you are going to
     * execute only queries that read data (fetch calls) while the connection
     * is "pinned".
     *
     * @throws BLException if the connection could not be retrieved from the
     * application context.
     *
     * @see #unpinConnection
     */
    public void pinConnection(boolean forUpdate)
        throws BLException {

        if(this.con != null)
            return;

        if(pinnedConnections.get() != null)
            throw new IllegalStateException("The thread already has a pinned DB connection associated with it.");

        try {
            pinnedConnections.set(new PinnedConnection(this.getConnection(forUpdate), forUpdate));
        } catch(SQLException e) {
            throw new BLException("The connection could not be retrieved.", e);
        }

        this.log.debug("pinned database connection down to the thread.");
    }

    /**
     * Releases and closes the database connection "pinned" by an earlier
     * <code>pinConnection</code> call.
     *
     * <p>This method does nothing if called on a DAO created via
     * <code>DAO(DAOConfig config, Connection con)</code> constructor.
     *
     * @throws BLException if the connection could not be closed. Even
     * if an exception is thrown the connection still becomes disassociated
     * from the thread, that is "unpinned".
     * @throws IllegalStateException if no connection has been "pinned"
     * earlier.
     *
     * @see #pinConnection
     */
    public void unpinConnection()
        throws BLException {

        if(this.con != null)
            return;

        PinnedConnection pcon = (PinnedConnection)pinnedConnections.get();
        if(pcon == null)
            throw new IllegalStateException("The thread has no pinned DB connection associated with it.");

        pinnedConnections.set(null);

        this.log.debug("unpinned database connection from the thread.");

        try {
            pcon.close();
        } catch(SQLException e) {
            throw new BLException("The connection could not be closed.", e);
        }
    }

    /**
     * Gets the query's primary (result) sql chunk text with all macros
     * expanded.
     *
     * @param queryName the query name.
     *
     * @return the query's primary sql chunk body.
     *
     * @throws BLException if there is no query with the specified name.
     *
     * @deprecated there is no direct substitution for this method. If you
     * needed this method for building dynamic SQL queries you should consider
     * using <code>{cond ...}</code> constructs of the DAO Extended Syntax.
     */
    public String getQuery(String queryName)
        throws BLException {

        return this.getPrimaryQuery(queryName);
    }


    ///////////////////////////////////////////////////////////////////////////
    // The DAO main low-level "workhorses"
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Internal implementation of <code>executeFetch</code> method.
     * This method is not a wrapper and does the actual job.
     */
    DataModel [] executeFetch_Internal(Context ctx,
                                       Class dmClass,
                                       String query,
                                       Object [] queryParams,
                                       int page,
                                       int pageSize,
                                       FetchResultProcessorHandler [] hooks,
                                       FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        // expand queryParams arrays into ? sequences
        query = this.expandDoubleQuestionMarks(query, queryParams);

        BLException eToThrow = null;

        List res = (pageSize > 0 ? new ArrayList(pageSize) : new ArrayList(100));

        int rowsProcessed = 0;
        long execMillis = 0;
        long rpMillis = 0;
        FetchExecutor fe = null;
        Connection con = null;
        Statement stmt = null;
        ResultSet rs = null;
        try {

            // get the database connection
            con = this.getConnection(false);

            // get and init the fetch executor
            fe = this.config.getFetchExecutor();
            fe.init(this.config, con, dmClass, query, queryParams, page, pageSize, fetchResultDescriptor);

            // prepare the statement
            stmt = fe.prepareStatement();

            // set parameters
            if((queryParams != null) && (queryParams.length > 0))
                this.setQueryParams(ctx, (PreparedStatement)stmt, queryParams);

            // stats
            QueryStatsRecord stats = this.getQueryStatsRecord(ctx.queryName, query);
            long startTS = stats.execStats.eventStarted();

            // execute the query
            stmt = fe.execute();
            rs = fe.getResultSet();

            // stats
            execMillis = stats.execStats.eventFinished(startTS);
            startTS = stats.procStats.eventStarted();

            // process the result set's rows
            if(fe.setToTheFirstRow()) {

                // create a fetch context
                FetchContext fc = new FetchContext(rs, hooks);

                // process the result set, we are at its first row
                int e = pageSize;
                do {

                    if(processResultSetRow(res,
                                           fc,
                                           true,
                                           dmClass,
                                           1,
                                           fc.columnCount))
                        fc.next();

                } while(fc.onValidRow && ((pageSize <= 0) || (e-- > 0)));

                // free the fetch executor
                fe.finish(fc.rowsProcessed, res);

                // for log message at the end of the method
                rowsProcessed = fc.rowsProcessed;
            }

            // stats
            rpMillis = stats.procStats.eventFinished(startTS);

            // all done, leave the try block

        } catch(SQLException e) {
            eToThrow = new BLException("Database error in the DAO.", e);

        } catch(InstantiationException e) {
            eToThrow = new BLException("Could not instantiate a DM in the DAO.", e);

        } catch(IllegalAccessException e) {
            eToThrow = new BLException("Something is wrong with the DM class in the DAO.", e);

        } finally {

            if(fe != null) {
                fe.destroy();
                this.config.releaseFetchExecutor(fe);
            }

            BLException e = this.closeDBResources(con, stmt, rs);
            if((e != null) && (eToThrow == null))
                eToThrow = e;
        }

        // throw exception if there was an error
        if(eToThrow != null)
            throw eToThrow;

        // return the result as an array
        if(this.log.isDebugEnabled()) {
            this.log.debug("the query fetched " + res.size() + " record(s) (" + rowsProcessed +
                           " row(s) processed, exec times: " + execMillis + "+" +
                           rpMillis + ":" + (execMillis + rpMillis) + "ms)");
        }
        return (DataModel [])res.toArray((DataModel [])Array.newInstance(dmClass, 0));
    }

    /**
     * Process current row of the result set according to the DAO
     * result set specification, build the DM instance and adds it to
     * the <code>result</code> list.
     *
     * @return <code>true</code> if the current row in the result set
     * is the last row processed for the current DM,
     * <code>false</code> if the current row in the result set is the
     * first row for the next DM.
     */
    private boolean processResultSetRow(List result,
                                        FetchContext fc,
                                        boolean callHooks,
                                        Class dmClass,
                                        int firstColumn,
                                        int lastColumn)
        throws BLException, SQLException, InstantiationException, IllegalAccessException {

        // get the result set field getter if defined
        ResultSetFieldGetter resultSetFieldGetter = this.config.getResultSetFieldGetter();

        // create a DM instance
        DataModel dm;
        if(dmClass == DynamicDataModel.class)
            dm = new DynamicDataModel(fc.columnCount);
        else
            dm = (DataModel)dmClass.newInstance();

        // initially we are on the row being processed
        boolean doNext = true;

        // process each column
        columnByColumn:
        for(int i = firstColumn; i <= lastColumn; i++) {
            FetchContext.FieldDescriptor fd = fc.getFieldDescriptor(dmClass, i);

            switch(fd.type) {
            case FetchContext.FieldDescriptor.FT_ARRAY: // is it a nested array?

                // process the rest of the columns as a nested DM array and exit the loop
                fd.field.set(dm, this.processNestedArray(fc,
                                                         fd.field.getType().getComponentType(),
                                                         i + 1,
                                                         lastColumn));
                doNext = false;
                break columnByColumn;

            case FetchContext.FieldDescriptor.FT_NESTEDDM: // is it a nested DM?

                // process next specified in this field number of columns as another DM
                int nFields = fc.rs.getInt(i);
                List inRes = null;
                try {
                    inRes = (List)this.singleElementArrayListPool.get();
                    doNext = this.processResultSetRow(inRes,
                                                      fc,
                                                      false,
                                                      fd.field.getType(),
                                                      i + 1,
                                                      i + nFields);
                    fd.field.set(dm, inRes.get(0));
                } finally {
                    if(inRes != null)
                        this.singleElementArrayListPool.returnObject(inRes);
                }
                i += nFields;

                // finish processing this row if there was a nested array in the internal DM
                if(!doNext)
                    break columnByColumn;

                break;

            case FetchContext.FieldDescriptor.FT_REGULAR: // it is a regular field

                // get the value
                Object value =
                    (resultSetFieldGetter != null ?
                     resultSetFieldGetter.getResultSetField(fc.rs,
                                                            i,
                                                            fd.fieldName,
                                                            fd.fieldClass) :
                     fc.rs.getObject(i));

                // set the filed value in the DM
                if(dmClass == DynamicDataModel.class)
                    this.setDynamicField((DynamicDataModel)dm,
                                         fd.fieldName,
                                         value);
                else
                    this.setReflectField(dm, fd.field, value);

                break;
            }
        }

        // call the hooks
        if((callHooks) && (fc.hooks != null)) {
            for(int i = 0; i < fc.hooks.length; i++) {
                if(fc.hooks[i].getLevel() == fc.level) {
                    if((dm = fc.hooks[i].getProcessor().
                        processDM(dm, fc.rs, firstColumn, lastColumn)) == null)
                        break;
                }
            }
        }

        // add the DM to the result
        if(dm != null)
            result.add(dm);

        // done, tell if we are on the same row
        return doNext;
    }

    /**
     * Processes the specified columns range in the result set
     * according to the DM nested array specification. Returns an
     * array, which may be empty, of nested DMs. The result set is
     * expected to be set on the first row of the nested array data.
     */
    private DataModel [] processNestedArray(FetchContext fc,
                                            Class nestedDMClass,
                                            int firstColumn,
                                            int lastColumn)
        throws BLException, SQLException, InstantiationException, IllegalAccessException {

        // prepare the result storage
        List res = new ArrayList();

        // outer join with no records?
        if(fc.rs.getObject(firstColumn) != null) {

            // go down to the next nesting level
            fc.level++;

            // add the upper keys to the list
            fc.upperKeysInds.add(new Integer(firstColumn - 1));
            fc.upperKeysVals.add(fc.rs.getObject(firstColumn - 1));

            // process the rows
            boolean cont;
            do {

                if(processResultSetRow(res,
                                       fc,
                                       true,
                                       nestedDMClass,
                                       firstColumn,
                                       lastColumn))
                    fc.next();

                cont = fc.onValidRow;
                if(cont) {
                    for(int i = 0; i < fc.level; i++) {
                        if(!fc.rs.getObject(((Integer)fc.upperKeysInds.get(i)).intValue()).
                           equals(fc.upperKeysVals.get(i))) {
                            cont = false;
                            break;
                        }
                    }
                }

            } while(cont);

            // remove the upper keys from the list
            fc.upperKeysInds.remove(fc.level - 1);
            fc.upperKeysVals.remove(fc.level - 1);

            // go up one level
            fc.level--;

        } else
            fc.next();

        return (DataModel [])res.toArray((DataModel [])Array.newInstance(nestedDMClass, res.size()));
    }

    /**
     * Internal <code>executeUpdate</code> implementation.
     * This method is not a wrapper and does the actual job.
     *
     * @param ctx the call context.
     * @param query the SQL query text.
     * @param queryParams query parameters.
     * @param forUpdate tells what type f the database connection
     * should be requested -- for update or read-only.
     * @param useExecuteUpdate tells if <code>executeUpdate</code>
     * or <code>execute</code> call should be used.
     *
     * @return number of affected rows. The values is guaranteed
     * to be right only if <code>executeUpdate</code> was requested
     * by specifying the <code>useExecuteUpdate</code> argument
     * as <code>true</code>.
     *
     * @throws BLException if an error happened executing the query.
     */
    int executeUpdate_Internal(Context ctx,
                               String query,
                               Object [] queryParams,
                               boolean forUpdate,
                               boolean useExecuteUpdate)
        throws BLException {

        // expand double question marks
        query = this.expandDoubleQuestionMarks(query, queryParams);

        int res = -1; // make the javac to shut up
        long execMillis = 0;

        BLException eToThrow = null;

        Connection con = null;
        Statement stmt = null;

        try {

            // get the database connection
            con = this.getConnection(forUpdate);

            // what type of statement to use?
            boolean usePreparedStatement = ((queryParams != null) && (queryParams.length > 0));

            // create the statement and set query parameters
            if(usePreparedStatement) {
                stmt = con.prepareStatement(query);
                this.setQueryParams(ctx, (PreparedStatement)stmt, queryParams);
            } else {
                stmt = con.createStatement();
            }

            // stats
            QueryStatsRecord stats = this.getQueryStatsRecord(ctx.queryName, query);
            long startTS = stats.execStats.eventStarted();

            // execute the query
            if(useExecuteUpdate) {
                res = (usePreparedStatement ?
                       ((PreparedStatement)stmt).executeUpdate() :
                       stmt.executeUpdate(query));
            } else {
                if(usePreparedStatement)
                    ((PreparedStatement)stmt).execute();
                else
                    stmt.execute(query);
                res = stmt.getUpdateCount();
            }

            // stats
            execMillis = stats.execStats.eventFinished(startTS);

        } catch(SQLException e) {
            eToThrow = new BLException("Database error in the DAO.", e);

        } finally {
            BLException e = this.closeDBResources(con, stmt, null);
            if((e != null) && (eToThrow == null))
                eToThrow = e;
        }

        // throw exception if there was an error
        if(eToThrow != null)
            throw eToThrow;

        // return the result
        this.log.debug("the query affected " + (res < 0 ? "no" : String.valueOf(res)) +
                       " rows (exec time: " + execMillis + "ms)");
        return res;
    }


    ///////////////////////////////////////////////////////////////////////////
    // The DAO named queries "workhorses"
    ///////////////////////////////////////////////////////////////////////////

    /**
     * The internal version of <code>fetch</code>, performs a fetch call
     * using <code>executeNamedQuery</code>.
     */
    private DataModel [] fetch_Internal(Context ctx,
                                        Object [] queryParams,
                                        String [] orderBy,
                                        int order,
                                        int page,
                                        int pageSize,
                                        FetchResultProcessorHandler [] hooks,
                                        FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        this.log.debug("fetching data with query \"" + ctx.queryName + "\"");

        FetchPrimaryChunkExecutor pce = null;
        try {
            pce = (FetchPrimaryChunkExecutor)this.fetchPrimaryChunkExecutorPool.get();
            pce.setQueryInfo(ctx,
                             orderBy,
                             order,
                             page,
                             pageSize,
                             hooks,
                             fetchResultDescriptor);

            this.executeNamedQuery(ctx,
                                   queryParams,
                                   pce,
                                   false);

            return pce.res;

        } finally {
            if(pce != null)
                this.fetchPrimaryChunkExecutorPool.returnObject(pce);
        }
    }

    /**
     * Internal version of <code>update</code>, performs an update call
     * using <code>executeNamedQuery</code>.
     */
    private int update_Internal(Context ctx,
                                Object [] queryParams)
        throws BLException {

        this.log.debug("updating data with query \"" + ctx.queryName + "\"");

        UpdatePrimaryChunkExecutor pce = null;
        try {
            pce = (UpdatePrimaryChunkExecutor)this.updatePrimaryChunkExecutorPool.get();
            pce.setQueryInfo(ctx);

            this.executeNamedQuery(ctx,
                                   queryParams,
                                   pce,
                                   true);

            return pce.res;

        } finally {
            if(pce != null)
                this.updatePrimaryChunkExecutorPool.returnObject(pce);
        }
    }

    /**
     * Executes chunks of a named query using primary chunk executors and
     * <code>executeUpdate_Internal</code> for non-primary chunks.
     *
     * @param ctx the call context.
     * @param queryParams values for the query parameters.
     * @param primaryChunkExecutor primary chunk executor
     * implementation.
     * @param isUpdate indicates if the query can modify underlying
     * database data.
     *
     * @throws BLException if an error happens.
     */
    private void executeNamedQuery(Context ctx,
                                   Object [] queryParams,
                                   PrimaryChunkExecutor primaryChunkExecutor,
                                   boolean isUpdate)
        throws BLException {

        if(ctx.queryDesc.sqls.length == 1) { // light-weight version for single sql chunk

            primaryChunkExecutor.execute(ctx.queryDesc.sqls[0], queryParams);

        } else {                  // full version for multiple sql chunks

            // pin down the connection
            this.pinConnection(isUpdate);

            // process the sql chunks
            try {

                int firstParamIndex = 0;
                Set succeededChunks = new HashSet(ctx.queryDesc.sqls.length);
                Throwable exceptionToThrow = null;
                for(int i = 0; i < ctx.queryDesc.sqls.length; i++) {

                    // prepare params sublist
                    int nParams = ctx.queryDesc.sqls[i].getNParams(ctx.conditions);
                    Object [] chunkParams = null;
                    if(nParams > 0) {
                        chunkParams = new Object[nParams];
                        System.arraycopy(queryParams, firstParamIndex, chunkParams, 0, nParams);
                        firstParamIndex += nParams;
                    }

                    // check dependencies
                    if(!succeededChunks.containsAll(ctx.queryDesc.sqls[i].depends)) {
                        this.log.info("skipping execution of sql chunk \"" + ctx.queryDesc.sqls[i].name +
                                      "\" because of errors in previous sql chunks, which this one depends on.");
                        continue;
                    }

                    // try to execute the chunk
                    this.log.debug("executing chunk \"" + ctx.queryDesc.sqls[i].name + "\"");
                    try {

                        if(ctx.queryDesc.sqls[i].isResult) {

                            primaryChunkExecutor.execute(ctx.queryDesc.sqls[i], chunkParams);

                        } else {

                            this.executeUpdate_Internal(ctx,
                                                        ctx.queryDesc.sqls[i].getSql(ctx.conditions),
                                                        chunkParams,
                                                        isUpdate,
                                                        false);
                        }

                        // add it to the list of successfully executed chunks
                        succeededChunks.add(ctx.queryDesc.sqls[i].name);

                    } catch(Throwable e) {
                        if(exceptionToThrow == null)
                            exceptionToThrow = e;

                        this.log.error("Got an error executing query \"" + ctx.queryName +
                                       "\" chunk \"" + ctx.queryDesc.sqls[i].name + "\".", e);
                    }
                }

                // check if there were errors
                if(exceptionToThrow != null) {
                    if(exceptionToThrow instanceof BLException)
                        throw (BLException)exceptionToThrow;
                    else if(exceptionToThrow instanceof RuntimeException)
                        throw (RuntimeException)exceptionToThrow;
                    else
                        throw (Error)exceptionToThrow;
                }

            } finally {

                // unpin the connection anyway
                this.unpinConnection();
            }
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // The DAO service interface
    ///////////////////////////////////////////////////////////////////////////

    // named queries

    /**
     * Version of {@link #fetch(String, Set, Object [], String [], int, int,
     * int, FetchResultProcessorHandler [], FetchResultDescriptor)}, which
     * works with named query parameters. Instead of object array, this method
     * takes query parameters as a map, which allows to associate names with
     * values. In the query's body <code>{? <i>&lt;param name&gt;</i>}</code>
     * and <code>{?? <i>&lt;param name&gt;</i>}</code> DAO Extended Syntax
     * constructs are used to mark parameter placeholders. As opposed to the
     * standard JDBC way of associating parameters with placeholders basing on
     * the placeholders' positions in the query text, using named query
     * parameters allows to disassociate the SQL query's structure from the way
     * it is called from Java. The disadvantage of it though is a slight lost
     * in performace, which can be ignored in the most cases.
     *
     * <p>Unnamed parameter placeholders, that is <code>?</code> and
     * <code>??</code>, still can be used with this call. For such unnamed
     * placeholders the parameter name in the map should be a string
     * representation of its position starting from 1.
     *
     * @param queryParams collection of values for the query parameters. The
     * keys are parameters names (objects of <code>String</code>), the values
     * are parameter values. A value can be an object, a <code>null</code>, or
     * an array of objects for a double-question mark placeholder. If value is
     * not present in the map for a parameter, it is assumed to be
     * <code>null</code>. This argument is optional and can be
     * <code>null</code> if the query does not have parameters.
     *
     * @return array of DMs built from the result set. The class of the DMs is
     * determined by <code>usedm</code> attribute of the
     * <code>&lt;query&gt;</code> element in the DAO configuration. Never
     * returns <code>null</code>, but can return an empty array.
     *
     * @throws BLException if an unexpected error happens and the query cannot
     * be executed.
     */
    public DataModel [] fetchWithNamedParams(String queryName,
                                             Set conditions,
                                             Map queryParams,
                                             String [] orderBy,
                                             int order,
                                             int page,
                                             int pageSize,
                                             FetchResultProcessorHandler [] hooks,
                                             FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        Context ctx = (Context)this.contextsPool.get();
        try {
            ctx.setup(this.config, queryName, conditions);
            return this.fetch_Internal(ctx,
                                       this.getQueryParamsArray(ctx, queryParams),
                                       orderBy,
                                       order,
                                       page,
                                       pageSize,
                                       hooks,
                                       fetchResultDescriptor);
        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }

    /**
     * Fetches data from the database using a query from the DAO configuration
     * and returns the result as an array of DMs.
     *
     * <p>This method, if requested, can append an <code>ORDER BY</code> clause
     * the query. If does so if the <Code>orderBy</code> method argument is not
     * <code>null</code> and is not an empty array. The method simply appends
     * each element in the array to the generated <code>ORDER BY</code> clause
     * separating the elements by commas. If the <code>order</code> method
     * argument is not <code>ORDER_UNUSED</code>, it also appends
     * <code>ASC</code> or <code>DESC</code> after each element. Since no
     * checks are performed, the <code>orderBy</code> array can contain whole
     * sequencies of column names separated by commas with their individual
     * <code>ASC</code> and <code>DESC</code> modifiers, or it can be used so
     * each element contains a single column name, and then using the
     * <code>order</code> argument may make sense.
     *
     * @param queryName name of the SQL query to execute defined in the DAO
     * configuration.
     * @param conditions set of names (objects of <code>String</code>) of
     * <code>{cond ...}</code> DAO Extended Syntax elements that should be
     * included in the query. Can be <code>null</code> if the query contains
     * no <code>{cond ...}</code> elements, should not be <code>null</code>
     * otherwise.
     * @param queryParams array of values for the query parameters matched to
     * the parameter placeholders in the query's body basing on their
     * positions. An element in this array may be an object, a
     * <code>null</code>, or an array of objects for double-question mark
     * parameter placeholders. Can be <code>null</code> if the query has no
     * parameters.
     * @param orderBy array of elements that will be added to the automatically
     * generated <code>ORDER BY</code> clause separated by commas. If automatic
     * <code>ORDER BY</code> clause should not be generated and appended to the
     * query, this parameter should be <code>null</code> or an empty array.
     * @param order one of <code>ORDER_<i>xxx</i></code> constants telling if
     * <code>ASC</code>, <code>DESC</code> or nothing should be added to each
     * element in the <code>orderBy</code> array when generating an automatic
     * <code>ORDER BY</code> clause.
     * @param page number of the result set page to return starting from zero.
     * Ignored if <code>pageSize</code> parameter is equal or less than zero.
     * @param pageSize number of records in one result set page. If this
     * parameter is equal or less than zero, no result set pagination is
     * performed and the whole result set is processed and returned.
     * @param hooks array of hooks called by the DAO from within its internal
     * result set processing loop for each DM on the specified by the hook
     * DM nesting level. Can be <code>null</code> or an empty array.
     * @param fetchResultDescriptor output parameter, which is updated after
     * the method call and contains additional information about the result
     * set. The method ignores this parameter if non-paginated result set is
     * requested (<code>pageSize</code> parameter is equal or less than zero).
     * Can be <code>null</code>.
     *
     * @return array of DMs built from the result set. The class of the DMs is
     * determined by <code>usedm</code> attribute of the
     * <code>&lt;query&gt;</code> element in the DAO configuration. Never
     * returns <code>null</code>, but can return an empty array.
     *
     * @throws BLException if an unexpected error happens and the query cannot
     * be executed.
     */
    public DataModel [] fetch(String queryName,
                              Set conditions,
                              Object [] queryParams,
                              String [] orderBy,
                              int order,
                              int page,
                              int pageSize,
                              FetchResultProcessorHandler [] hooks,
                              FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        Context ctx = (Context)this.contextsPool.get();
        try {
            ctx.setup(this.config, queryName, conditions);
            return this.fetch_Internal(ctx,
                                       queryParams,
                                       orderBy,
                                       order,
                                       page,
                                       pageSize,
                                       hooks,
                                       fetchResultDescriptor);
        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }

    /**
     * Version of {@link #update(String, Set, Object [])}, which works with
     * named query parameters.
     *
     * <p>See {@link #fetchWithNamedParams(String, Set, Map, String[], int,
     * int, int, FetchResultProcessorHandler[], FetchResultDescriptor)} for
     * more information on named query parameters.
     *
     * @return number of affected rows. If the database does not return any
     * update counts, returns <code>-1</code>.
     *
     * @throws BLException if an unexpected error happens and the query cannot
     * be executed.
     */
    public int updateWithNamedParams(String queryName,
                                     Set conditions,
                                     Map queryParams)
        throws BLException {

        Context ctx = (Context)this.contextsPool.get();
        try {
            ctx.setup(this.config, queryName, conditions);
            return this.update_Internal(ctx, this.getQueryParamsArray(ctx, queryParams));
        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }

    /**
     * Executes a query from the DAO configuration as an update query and
     * returns the number of affected rows.
     *
     * @param queryName name of the SQL query to execute defined in the DAO
     * configuration.
     * @param conditions set of names (objects of <code>String</code>) of
     * <code>{cond ...}</code> DAO Extended Syntax elements that should be
     * included in the query. Can be <code>null</code> if the query contains
     * no <code>{cond ...}</code> elements, should not be <code>null</code>
     * otherwise.
     * @param queryParams array of values for the query parameters matched to
     * the parameter placeholders in the query's body basing on their
     * positions. An element in this array may be an object, a
     * <code>null</code>, or an array of objects for double-question mark
     * parameter placeholders. Can be <code>null</code> if the query has no
     * parameters.
     *
     * @return number of affected rows. If the database does not return any
     * update counts, returns <code>-1</code>.
     *
     * @throws BLException if an unexpected error happens and the query cannot
     * be executed.
     */
    public int update(String queryName, Set conditions, Object [] queryParams)
        throws BLException {

        Context ctx = (Context)this.contextsPool.get();
        try {
            ctx.setup(this.config, queryName, conditions);
            return this.update_Internal(ctx, queryParams);
        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }


    // autostatements

    /**
     * Constructs and executes an SQL <code>INSERT</code> statement for
     * the specified DM. The DM must have <code>table</code> attribute in its
     * descriptor in the DAO configuration, as well as appropriate definition
     * of its id fields (<code>&lt;idfield&gt;</code> subelements). This method
     * updates id fields in the specified DM instance with the values inserted
     * into the table. The way the new values are aquired for the id fields is
     * determined by <code>srcquery</code> and <code>srcorder</code> attributes
     * of the appropriate <code>&lt;idfield&gt;</code> elements.
     *
     * <p>This method ignores any nested DM and nested DM array fields so the
     * same DM class can be used for complex fetch calls with joined tables as
     * well as for inserts.
     *
     * @param dm DM with data to be inserted into the associated with the DM
     * database table. The id fields are ignored and then updated with the
     * values inserted into the table.
     *
     * @return number of inserted rows (normally one).
     *
     * @throws BLException if an error happens and the query cannot be
     * constructed or executed.
     */
    public int insert(DataModel dm)
        throws BLException {

        this.log.debug("performing auto insert for " + dm.getClass().getName());

        int res = -1; // make the javac to shut up
        long execMillis = 0;

        BLException eToThrow = null;

        Connection con = null;
        Statement stmt = null;
        ResultSet rs = null;

        Context ctx = (Context)this.contextsPool.get();

        try {

            DAOConfig.DMDescriptor dmDescriptor = this.config.getDMDescriptor(dm.getClass());

            if((dmDescriptor.tableName == null) || (dmDescriptor.tableName.length() == 0))
                throw new BLException("Attribute \"table\" is not defined for DM \"" +
                                      dmDescriptor.name + "\" in the configuration file.");

            List params = new ArrayList();
            List paramTypes = new ArrayList();
            StringBuffer querySB = new StringBuffer();

            // build the query
            querySB.append("INSERT INTO ");
            querySB.append(dmDescriptor.tableName);
            querySB.append(" (");
            Iterator i = dmDescriptor.fields.values().iterator();
            boolean fieldAdded = false;
            int pos = 0;
            Map constFieldValues = new HashMap(dmDescriptor.constFields.size());
            while(i.hasNext()) {
                Field field = (Field)i.next();
                String fieldName = field.getName();
                DAOConfig.DMDescriptor.ConstFieldDescriptor constFieldDesc =
                    (DAOConfig.DMDescriptor.ConstFieldDescriptor)dmDescriptor.constFields.get(fieldName);
                if((constFieldDesc == null) ||
                   (constFieldDesc.applyTo == DAOConfig.DMDescriptor.ConstFieldDescriptor.APPLYTO_UPDATES)) { // not a const field
                    if(!dmDescriptor.idFieldNamesSet.contains(fieldName) &&
                       !field.getType().isArray() &&
                       !(DataModel.class).isAssignableFrom(field.getType())) { // a regular field
                        if(fieldAdded)
                            querySB.append(", ");
                        else
                            fieldAdded = true;
                        querySB.append(fieldName);
                        params.add(field.get(dm));
                        paramTypes.add(this.getSQLTypeForJavaType(field.getType()));
                        pos++;
                    }
                } else { // a const field
                    if(fieldAdded)
                        querySB.append(", ");
                    else
                        fieldAdded = true;
                    querySB.append(fieldName);
                    constFieldValues.put(new Integer(pos), constFieldDesc.sql);
                    pos++;
                }
            }
            int nFieldsAdded = pos;
            StringBuffer preIds = new StringBuffer();
            for(int j = 0; j < dmDescriptor.idFields.length; j++) {
                DAOConfig.DMDescriptor.IDFieldDescriptor idFieldDesc = dmDescriptor.idFields[j];
                if((idFieldDesc.srcQuery == null) ||
                   (idFieldDesc.srcQuery.length() == 0)) { // no srcquery
                    if(fieldAdded)
                        querySB.append(", ");
                    else
                        fieldAdded = true;
                    querySB.append(idFieldDesc.name);
                    Field field = (Field)dmDescriptor.fields.get(idFieldDesc.name.toUpperCase());
                    params.add(field.get(dm));
                    paramTypes.add(this.getSQLTypeForJavaType(field.getType()));
                    nFieldsAdded++;

                } else if(!idFieldDesc.isPost) { // pre srcquery
                    if(preIds.length() > 0)
                        preIds.append(", ");
                    preIds.append(idFieldDesc.name);
                    nFieldsAdded++;
                }
            }
            if(preIds.length() > 0) {
                if(fieldAdded)
                    querySB.append(", ");
                querySB.append(preIds);
                fieldAdded = true;
            }
            querySB.append(") VALUES (");
            for(int j = 0; j < nFieldsAdded; j++) {
                if(j > 0)
                    querySB.append(", ");
                String constFieldText = (String)constFieldValues.get(new Integer(j));
                if(constFieldText == null)
                    querySB.append('?');
                else
                    querySB.append(constFieldText);
            }
            querySB.append(')');

            // get the database connection
            con = this.getConnection(true);

            // get the result set field getter if defined
            ResultSetFieldGetter resultSetFieldGetter = this.config.getResultSetFieldGetter();

            // get values for all "pre" ids
            for(int j = 0; j < dmDescriptor.idFields.length; j++) {
                DAOConfig.DMDescriptor.IDFieldDescriptor idFieldDesc = dmDescriptor.idFields[j];
                if(!idFieldDesc.isPost &&
                   (idFieldDesc.srcQuery != null) &&
                   (idFieldDesc.srcQuery.length() > 0)) {

                    // get the field
                    Field field = (Field)dmDescriptor.fields.get(idFieldDesc.name.toUpperCase());

                    // get value
                    String preQuery = this.getPrimaryQuery(idFieldDesc.srcQuery);
                    stmt = con.createStatement();
                    if(this.log.isDebugEnabled())
                        this.log.debug("executing " + idFieldDesc.name +
                                       " id field getter pre query [" + preQuery + "]");
                    rs = stmt.executeQuery(preQuery);
                    if(!rs.next())
                        throw new BLException("Id getter pre query for field " + idFieldDesc.name +
                                              " of DM " + dmDescriptor.name + " returned no result.");
                    Object idValue = (resultSetFieldGetter != null ?
                                      resultSetFieldGetter.getResultSetField(rs,
                                                                             1,
                                                                             field.getName(),
                                                                             field.getType()) :
                                      rs.getObject(1));
                    if(this.log.isDebugEnabled())
                        this.log.debug("id field getter pre query returned [" + idValue + "]");
                    rs.close();
                    rs = null;
                    stmt.close();
                    stmt = null;

                    // set the field
                    idValue = this.setReflectField(dm, field, idValue);
                    params.add(idValue);
                    paramTypes.add(this.getSQLTypeForJavaType(field.getType()));
                }
            }

            String query = querySB.toString();
            Object [] queryParams = params.toArray();

            // set param types in the context
            ctx.paramSqlTypesForNulls = (Integer [])paramTypes.toArray(new Integer[paramTypes.size()]);

            // compile the query text
            stmt = con.prepareStatement(query);

            // set query parameters
            this.setQueryParams(ctx, (PreparedStatement)stmt, queryParams);

            // stats
            QueryStatsRecord stats = this.getQueryStatsRecord("autoinsert for " + dm.getClass().getName(), query);
            long startTS = stats.execStats.eventStarted();

            // execute the query
            res = ((PreparedStatement)stmt).executeUpdate();

            // stats
            execMillis = stats.execStats.eventFinished(startTS);

            // close it, we gonna reuse it
            stmt.close();
            stmt = null;

            // get values for all "post" ids
            for(int j = 0; j < dmDescriptor.idFields.length; j++) {
                if(dmDescriptor.idFields[j].isPost) {

                    // get the field
                    Field field = (Field)dmDescriptor.fields.get(dmDescriptor.idFields[j].name.toUpperCase());

                    // get value
                    String postQuery = this.getPrimaryQuery(dmDescriptor.idFields[j].srcQuery);
                    stmt = con.createStatement();
                    if(this.log.isDebugEnabled())
                        this.log.debug("executing " + dmDescriptor.idFields[j].name +
                                       " id field getter post query [" + postQuery + "]");
                    rs = stmt.executeQuery(postQuery);
                    if(!rs.next())
                        throw new BLException("Id getter post query for field " + dmDescriptor.idFields[j].name +
                                              " of DM " + dmDescriptor.name + " returned no result.");
                    Object idValue = (resultSetFieldGetter != null ?
                                      resultSetFieldGetter.getResultSetField(rs,
                                                                             1,
                                                                             field.getName(),
                                                                             field.getType()) :
                                      rs.getObject(1));
                    if(this.log.isDebugEnabled())
                        this.log.debug("id field getter post query returned [" + idValue + "]");
                    rs.close();
                    rs = null;
                    stmt.close();
                    stmt = null;

                    // set the field
                    idValue = this.setReflectField(dm, field, idValue);
                }
            }

        } catch(BLException e) {
            eToThrow = e;

        } catch(IllegalAccessException e) {
            eToThrow = new BLException("Something is wrong with the DM " + dm.getClass().getName() + ".", e);

        } catch(SQLException e) {
            eToThrow = new BLException("Database error in the DAO.", e);

        } finally {
            this.contextsPool.returnObject(ctx);

            BLException e = this.closeDBResources(con, stmt, rs);
            if((e != null) && (eToThrow == null))
                eToThrow = e;
        }

        // throw exception if there was an error
        if(eToThrow != null)
            throw eToThrow;

        // return the result
        this.log.debug("auto insert for " + dm.getClass().getName() + " inserted " +
                       res + " record(s) (exec time: " + execMillis + "ms)");
        return res;
    }

    /**
     * Constructs and executes an SQL <code>UPDATE</code> statement for
     * the specified DM. The DM must have <code>table</code> attribute in its
     * descriptor in the DAO configuration, as well as appropriate definition
     * of its id fields (<code>&lt;idfield&gt;</code> subelements). The rows
     * to be updated are selected by values in the specified instance's id
     * fields - the id fields and their values go to the statement's
     * <code>WHERE</code> clause to select appropriate record, while other
     * fields go to the <code>SET</code> clause. If the DM has multiple id
     * fields, the condition in the <code>WHERE</code> clause uses
     * <code>AND</code> logical operation.
     *
     * <p>This method ignores any nested DM and nested DM array fields so the
     * same DM class can be used for complex fetch calls with joined tables as
     * well as for updates.
     *
     * @param dm DM with data to be set into the associated with the DM
     * database table. The rows to be updated are selected by values in the
     * DM's id fields.
     *
     * @return number of updated rows (normally one or zero).
     *
     * @throws BLException if an error happens and the query cannot be
     * constructed or executed.
     */
    public int update(DataModel dm)
        throws BLException {

        return this.update(dm, null, null);
    }

    /**
     * Similar to {@link #update(DataModel)}, except that only fields listed in
     * the <code>fields</code> set are included into the generated
     * <code>UPDATE</code> statement.
     *
     * <p>Note, that id fields, even if don't appear in the set, are always
     * included into the statement's <code>WHERE</code> clause.
     *
     * @param dm DM with data to be set into the associated with the DM
     * database table. The rows to be updated are selected by values in the
     * DM's id fields. Values in fields not listed in the <code>fields</code>
     * set are ignored.
     * @param fields set of DM field names (objects of <code>String</code>)
     * that should be updated by the generated statement. Fields that don't
     * appear in the set are left untouched.
     *
     * @return number of updated rows (normally one or zero).
     *
     * @throws BLException if an error happens and the query cannot be
     * constructed or executed.
     *
     * @see #updateExcludingFields(DataModel, Set)
     */
    public int updateFields(DataModel dm, Set fields)
        throws BLException {

        return this.update(dm, fields, null);
    }

    /**
     * Similar to {@link #update(DataModel)}, except that only fields not
     * listed in the <code>fields</code> set are included into the generated
     * <code>UPDATE</code> statement.
     *
     * <p>Note, that id fields, even if appear in the set, are always included
     * into the statement's <code>WHERE</code> clause.
     *
     * @param dm DM with data to be set into the associated with the DM
     * database table. The rows to be updated are selected by values in the
     * DM's id fields. Values in fields listed in the <code>fields</code>
     * set are ignored.
     * @param fields set of DM field names (objects of <code>String</code>)
     * that should be excluded from the update. Fields that appear in the set
     * are left untouched.
     *
     * @return number of updated rows (normally one or zero).
     *
     * @throws BLException if an error happens and the query cannot be
     * constructed or executed.
     *
     * @see #updateFields(DataModel, Set)
     */
    public int updateExcludingFields(DataModel dm, Set fields)
        throws BLException {

        return this.update(dm, null, fields);
    }

    /**
     * Similar to {@link #update(DataModel)}, but update only those fields,
     * that have different values in the two specified DMs. This method uses
     * <code>diff</code> on <code>DataModel</code> to compare the DMs so all
     * rules that apply to the <code>diff</code> method do apply to this method
     * as well. If the DMs are identical, the method does nothing.
     *
     * <p>Note, that this method does not check if the id fields are identical,
     * although logically they should always be. In any case, the id field
     * values from the <code>newDm</code> are taken when creating the
     * statement.
     * 
     * @param oldDm DM with "old" data.
     * @param newDm DM with "new" data, that will be stored into the database.
     *
     * @return number of updated rows (normally one or zero).
     *
     * @throws BLException if an error happens and the query cannot be
     * constructed or executed.
     * @throws IllegalArgumentException if types of the specified DMs are
     * different.
     *
     * @see DataModel#diff
     */
    public int updateDiff(DataModel oldDm, DataModel newDm)
        throws BLException {

        String [] diff = newDm.diff(oldDm);
        if(diff.length > 0) {
            Set fields = new HashSet(diff.length);
            for(int i = 0; i < diff.length; i++)
                fields.add(diff[i]);
            return this.update(newDm, fields, null);
        }

        this.log.debug("DMs are identical, nothing to update");
        return 0;
    }

    /**
     * Internal function behind <code>update(DataModel)</code>,
     * <code>updateFields</code> and
     * <code>updateExcludingFields</code> public methods.
     *
     * <p>Note, that at least one of the parameters
     * <code>includedFields</code> and <code>excludedFields</code> has
     * to be <code>null</code>.
     *
     * <p>Note, that id fields, if appear in any of the sets, are
     * ignored and they are always included into the
     * <code>WHERE</code> clause of the generated statement.
     *
     * @param dm the data model instance containing new data to be
     * stored in the associated with it in the blo-config.xml file
     * database table. The rows to be updated are selected by the DM's
     * "id" fields.
     * @param includedFields set of DM field names (objects of
     * <code>String</code>) that will be updated by the generated
     * statement. Fields not appearing in the set will not be touched.
     * @param excludedFields set of DM field names (objects of
     * <code>String</code>) that will NOT be updated by the generated
     * statement.
     *
     * @return the number of updated rows.
     *
     * @throws BLException indicates a serious error, the query can
     * not be executed for some reason.
     */
    private int update(DataModel dm,
                       Set includedFields,
                       Set excludedFields)
        throws BLException {

        this.log.debug("performing auto update for " + dm.getClass().getName());

        Context ctx = (Context)this.contextsPool.get();
        try {
            ctx.queryName = "autoupdate for " + dm.getClass().getName();

            // build the query

            DAOConfig.DMDescriptor dmDescriptor = this.config.getDMDescriptor(dm.getClass());

            if((dmDescriptor.tableName == null) || (dmDescriptor.tableName.length() == 0))
                throw new BLException("Attribute \"table\" is not defined for DM \"" +
                                      dmDescriptor.name + "\" in the configuration file.");

            List params = new ArrayList();
            List paramTypes = new ArrayList();
            StringBuffer query = new StringBuffer();
            try {

                // generate the UPDATE clause
                query.append("UPDATE ");
                query.append(dmDescriptor.tableName);

                // generate the SETs clauses
                query.append(" SET ");
                Iterator i = dmDescriptor.fields.values().iterator();
                boolean fieldAdded = false;
                while(i.hasNext()) {
                    Field field = (Field)i.next();
                    String fieldName = field.getName();

                    if(((includedFields != null) && !includedFields.contains(fieldName)) ||
                       ((excludedFields != null) && excludedFields.contains(fieldName)))
                        continue;

                    DAOConfig.DMDescriptor.ConstFieldDescriptor constFieldDesc =
                        (DAOConfig.DMDescriptor.ConstFieldDescriptor)dmDescriptor.constFields.get(fieldName);
                    if((constFieldDesc == null) ||
                       (constFieldDesc.applyTo == DAOConfig.DMDescriptor.ConstFieldDescriptor.APPLYTO_INSERTS)) {
                        if(!dmDescriptor.idFieldNamesSet.contains(fieldName) &&
                           !field.getType().isArray() &&
                           !(DataModel.class).isAssignableFrom(field.getType())) {
                            if(fieldAdded)
                                query.append(", ");
                            else
                                fieldAdded = true;
                            query.append(fieldName);
                            query.append(" = ?");
                            params.add(field.get(dm));
                            paramTypes.add(this.getSQLTypeForJavaType(field.getType()));
                        }
                    } else {
                        if(fieldAdded)
                            query.append(", ");
                        else
                            fieldAdded = true;
                        query.append(fieldName);
                        query.append(" = ");
                        query.append(constFieldDesc.sql);
                    }
                }

                // check if we got fields to update
                if(!fieldAdded)
                    throw new BLException("There were no fields in the DM to be set by the UPDATE statement.");

                // generate the WHERE clause
                if(dmDescriptor.idFields.length > 0) {
                    query.append(" WHERE ");
                    for(int j = 0; j < dmDescriptor.idFields.length; j++) {
                        if(j > 0)
                            query.append(" AND ");
                        query.append(dmDescriptor.idFields[j].name);
                        query.append(" = ?");
                        Field field = (Field)dmDescriptor.fields.get(dmDescriptor.idFields[j].name.toUpperCase());
                        params.add(field.get(dm));
                        paramTypes.add(this.getSQLTypeForJavaType(field.getType()));
                    }
                }

            } catch(IllegalAccessException e) {
                throw new BLException("Something is wrong with the DM " + dm.getClass().getName() + ".", e);
            }

            // put parameter types to the context
            ctx.paramSqlTypesForNulls = (Integer [])paramTypes.toArray(new Integer[paramTypes.size()]);

            // execute the query
            int res = this.executeUpdate_Internal(ctx,
                                                  query.toString(),
                                                  params.toArray(),
                                                  true,
                                                  true);

            // return the result
            return res;

        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }

    /**
     * Constructs and executes an SQL <code>DELETE</code> statement for
     * the specified DM. The DM must have <code>table</code> attribute in its
     * descriptor in the DAO configuration, as well as appropriate definition
     * of its id fields (<code>&lt;idfield&gt;</code> subelements). The rows
     * to be deleted are selected by values in the specified instance's id
     * fields - the id fields and their values go to the statement's
     * <code>WHERE</code> clause to select appropriate record, while other
     * fields are ignored. If the DM has multiple id fields, the condition in
     * the <code>WHERE</code> clause uses <code>AND</code> logical operation.
     *
     * @param dm DM with id fields set to select rows to delete.
     *
     * @return number of deleted rows (normally one or zero).
     *
     * @throws BLException if an error happens and the query cannot be
     * constructed or executed.
     */
    public int delete(DataModel dm)
        throws BLException {

        this.log.debug("performing auto delete for " + dm.getClass().getName());

        Context ctx = (Context)this.contextsPool.get();
        try {
            ctx.queryName = "autodelete for " + dm.getClass().getName();

            // build the query

            DAOConfig.DMDescriptor dmDescriptor = this.config.getDMDescriptor(dm.getClass());

            if((dmDescriptor.tableName == null) || (dmDescriptor.tableName.length() == 0))
                throw new BLException("Attribute \"table\" is not defined for DM \"" +
                                      dmDescriptor.name + "\" in the configuration file.");

            List params = new ArrayList();
            StringBuffer query = new StringBuffer();
            try {

                query.append("DELETE FROM ");
                query.append(dmDescriptor.tableName);
                if(dmDescriptor.idFields.length > 0) {
                    query.append(" WHERE ");
                    for(int j = 0; j < dmDescriptor.idFields.length; j++) {
                        if(j > 0)
                            query.append(" AND ");
                        query.append(dmDescriptor.idFields[j].name);
                        query.append(" = ?");
                        Field field = (Field)dmDescriptor.fields.get(dmDescriptor.idFields[j].name.toUpperCase());
                        params.add(field.get(dm));
                    }
                }

            } catch(IllegalAccessException e) {
                throw new BLException("Something is wrong with the DM " + dm.getClass().getName() + ".", e);
            }

            // execute the query
            int res = this.executeUpdate_Internal(ctx,
                                                  query.toString(),
                                                  params.toArray(),
                                                  true,
                                                  true);

            // return the result
            return res;

        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }


    // low-level executors that take query text as a parameter

    /**
     * Fetches data from the database using the specified SQL query and returns
     * the result as an array of DMs. This is a very low-level method that
     * does not require query definition in the DAO configuration - instead,
     * it takes the query's SQL as an argument. The DM, corresponding to the
     * specified DM class, still has to be defined in the DAO configuration and
     * have a <code>&lt;dm&gt;</code> element.
     *
     * <p>Note, that this low-level method does not support DAO Extended Syntax
     * constructs in the query's SQL except double question marks.
     *
     * <p>See {@link #fetch(String, Set, Object [], String [], int, int, int,
     * FetchResultProcessorHandler [], FetchResultDescriptor)} for description
     * of common method parameters.
     *
     * @param dmClass class of the DM this query returns. Should be a subclass
     * of <code>DataModel</code>.
     * @param query the query's SQL text.
     *
     * @return array of DMs built from the result set. Never returns
     * <code>null</code>, but can return an empty array.
     *
     * @throws BLException if an unexpected error happens and the query cannot
     * be executed.
     */
    public DataModel [] executeFetch(Class dmClass,
                                     String query,
                                     Object [] queryParams,
                                     int page,
                                     int pageSize,
                                     FetchResultProcessorHandler [] hooks,
                                     FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        Context ctx = (Context)this.contextsPool.get();
        try {
            return this.executeFetch_Internal(ctx,
                                              dmClass,
                                              query,
                                              queryParams,
                                              page,
                                              pageSize,
                                              hooks,
                                              fetchResultDescriptor);
        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }

    /**
     * Executes the specified SQL query and returns the number of affected
     * rows. This is a very low-level method that does not require query
     * definition in the DAO configuration - instead, it takes the query's SQL
     * as an argument.
     *
     * <p>Note, that this low-level method does not support DAO Extended Syntax
     * constructs in the query's SQL except double question marks.
     *
     * <p>See {@link #update(String, Set, Object[])} for description of common
     * method parameters.
     *
     * @param query the query's SQL text.
     *
     * @return number of affected rows. If the database does not return any
     * update counts, returns <code>-1</code>.
     *
     * @throws BLException if an unexpected error happens and the query cannot
     * be executed.
     */
    public int executeUpdate(String query, Object [] queryParams)
        throws BLException {

        Context ctx = (Context)this.contextsPool.get();
        try {
            return this.executeUpdate_Internal(ctx,
                                               query,
                                               queryParams,
                                               true,
                                               true);
        } finally {
            this.contextsPool.returnObject(ctx);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Convenience wrapper methods
    ///////////////////////////////////////////////////////////////////////////

    // fetch series

    /**
     * Calls <code>fetch(queryName, null, queryParams, null, DAO.ORDER_UNUSED,
     * 0, 0, null, null)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName, Object [] queryParams)
        throws BLException {

        return this.fetch(queryName, null,
                          queryParams,
                          null, ORDER_UNUSED,
                          0, 0,
                          null,
                          null);
    }

    /**
     * Calls <code>fetch(queryName, null, queryParams, orderBy, order, 0, 0,
     * null, null)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName, Object [] queryParams, String [] orderBy, int order)
        throws BLException {

        return this.fetch(queryName, null,
                          queryParams,
                          orderBy, order,
                          0, 0,
                          null,
                          null);
    }

    /**
     * Calls <code>fetch(queryName, null, queryParams, null, DAO.ORDER_UNUSED,
     * page, pageSize, null, fetchResultDescriptor)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName,
                              Object [] queryParams,
                              int page,
                              int pageSize,
                              FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetch(queryName, null,
                          queryParams,
                          null, DAO.ORDER_UNUSED,
                          page, pageSize,
                          null,
                          fetchResultDescriptor);
    }

    /**
     * Calls <code>fetch(queryName, null, queryParams, orderBy, order, page,
     * pageSize, null, fetchResultDescriptor)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName,
                              Object [] queryParams,
                              String [] orderBy,
                              int order,
                              int page,
                              int pageSize,
                              FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetch(queryName, null,
                          queryParams,
                          orderBy, order,
                          page, pageSize,
                          null,
                          fetchResultDescriptor);
    }

    /**
     * Calls <code>fetch(queryName, conditions, queryParams, null,
     * DAO.ORDER_UNUSED, 0, 0, null, null)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName, Set conditions, Object [] queryParams)
        throws BLException {

        return this.fetch(queryName, conditions,
                          queryParams,
                          null, ORDER_UNUSED,
                          0, 0,
                          null,
                          null);
    }

    /**
     * Calls <code>fetch(queryName, conditions, queryParams, orderBy, order, 0,
     * 0, null, null)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName,
                              Set conditions,
                              Object [] queryParams,
                              String [] orderBy,
                              int order)
        throws BLException {

        return this.fetch(queryName, conditions,
                          queryParams,
                          orderBy, order,
                          0, 0,
                          null,
                          null);
    }

    /**
     * Calls <code>fetch(queryName, conditions, queryParams, null,
     * DAO.ORDER_UNUSED, page, pageSize, null, fetchResultDescriptor)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName,
                              Set conditions,
                              Object [] queryParams,
                              int page,
                              int pageSize,
                              FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetch(queryName, conditions,
                          queryParams,
                          null, DAO.ORDER_UNUSED,
                          page, pageSize,
                          null,
                          fetchResultDescriptor);
    }

    /**
     * Calls <code>fetch(queryName, conditions, queryParams, orderBy, order,
     * page, pageSize, null, fetchResultDescriptor)</code>.
     *
     * @see #fetch(String, Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetch(String queryName,
                              Set conditions,
                              Object [] queryParams,
                              String [] orderBy,
                              int order,
                              int page,
                              int pageSize,
                              FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetch(queryName, conditions,
                          queryParams,
                          orderBy, order,
                          page, pageSize,
                          null,
                          fetchResultDescriptor);
    }


    // fetchWithNamedParams series

    /**
     * Calls <code>fetchWithNamedParams(queryName, null, queryParams, null,
     * DAO.ORDER_UNUSED, 0, 0, null, null)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName, Map queryParams)
        throws BLException {

        return this.fetchWithNamedParams(queryName, null,
                                         queryParams,
                                         null, ORDER_UNUSED,
                                         0, 0,
                                         null,
                                         null);
    }

    /**
     * Calls <code>fetchWithNamedParams(queryName, null, queryParams, orderBy,
     * order, 0, 0, null, null)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName, Map queryParams, String [] orderBy, int order)
        throws BLException {

        return this.fetchWithNamedParams(queryName, null,
                                         queryParams,
                                         orderBy, order,
                                         0, 0,
                                         null,
                                         null);
    }

    /**
     * Calls <code>fetchWithNamedParams(queryName, null, queryParams, null,
     * DAO.ORDER_UNUSED, page, pageSize, null, fetchResultDescriptor)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName,
                                             Map queryParams,
                                             int page,
                                             int pageSize,
                                             FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetchWithNamedParams(queryName, null,
                                         queryParams,
                                         null, DAO.ORDER_UNUSED,
                                         page, pageSize,
                                         null,
                                         fetchResultDescriptor);
    }

    /**
     * Calls <code>fetchWithNamedParams(queryName, null, queryParams, orderBy,
     * order, page, pageSize, null, fetchResultDescriptor)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName,
                                             Map queryParams,
                                             String [] orderBy,
                                             int order,
                                             int page,
                                             int pageSize,
                                             FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetchWithNamedParams(queryName, null,
                                         queryParams,
                                         orderBy, order,
                                         page, pageSize,
                                         null,
                                         fetchResultDescriptor);
    }

    /**
     * Calls <code>fetchWithNamedParams(queryName, conditions, queryParams,
     * null, DAO.ORDER_UNUSED, 0, 0, null, null)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName, Set conditions, Map queryParams)
        throws BLException {

        return this.fetchWithNamedParams(queryName, conditions,
                                         queryParams,
                                         null, ORDER_UNUSED,
                                         0, 0,
                                         null,
                                         null);
    }

    /**
     * Calls <code>fetchWithNamedParams(queryName, conditions, queryParams,
     * orderBy, order, 0, 0, null, null)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName,
                                             Set conditions,
                                             Map queryParams,
                                             String [] orderBy,
                                             int order)
        throws BLException {

        return this.fetchWithNamedParams(queryName, conditions,
                                         queryParams,
                                         orderBy, order,
                                         0, 0,
                                         null,
                                         null);
    }

    /**
     * Calls <code>fetchWithNamedParams(queryName, conditions, queryParams,
     * null, DAO.ORDER_UNUSED, page, pageSize, null,
     * fetchResultDescriptor)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName,
                                             Set conditions,
                                             Map queryParams,
                                             int page,
                                             int pageSize,
                                             FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetchWithNamedParams(queryName, conditions,
                                         queryParams,
                                         null, DAO.ORDER_UNUSED,
                                         page, pageSize,
                                         null,
                                         fetchResultDescriptor);
    }

    /**
     * Calls <code>fetchWithNamedParams(queryName, conditions, queryParams,
     * orderBy, order, page, pageSize, null, fetchResultDescriptor)</code>.
     *
     * @see #fetchWithNamedParams(String, Set, Map, String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] fetchWithNamedParams(String queryName,
                                             Set conditions,
                                             Map queryParams,
                                             String [] orderBy,
                                             int order,
                                             int page,
                                             int pageSize,
                                             FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.fetchWithNamedParams(queryName, conditions,
                                         queryParams,
                                         orderBy, order,
                                         page, pageSize,
                                         null,
                                         fetchResultDescriptor);
    }


    // executeFetch series

    /**
     * Calls <code>executeFetch(dmClass, query, queryParams, 0, 0, null,
     * null)</code>.
     *
     * @see #executeFetch(Class, String, Object [], int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] executeFetch(Class dmClass,
                                     String query,
                                     Object [] queryParams)
        throws BLException {

        return this.executeFetch(dmClass,
                                 query,
                                 queryParams,
                                 0, 0,
                                 null,
                                 null);
    }

    /**
     * Calls <code>executeFetch(dmClass, query, queryParams, page, pageSize,
     * null, fetchResultDescriptor)</code>.
     *
     * @see #executeFetch(Class, String, Object [], int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
     */
    public DataModel [] executeFetch(Class dmClass,
                                     String query,
                                     Object [] queryParams,
                                     int page,
                                     int pageSize,
                                     FetchResultDescriptor fetchResultDescriptor)
        throws BLException {

        return this.executeFetch(dmClass,
                                 query,
                                 queryParams,
                                 page, pageSize,
                                 null,
                                 fetchResultDescriptor);
    }


    // updates

    /**
     * Calls <code>update(queryName, null, queryParams)</code>.
     *
     * @see #update(String, Set, Object [])
     */
    public int update(String queryName, Object [] queryParams)
        throws BLException {

        return this.update(queryName, null, queryParams);
    }

    /**
     * Calls <code>updateWithNamedParams(queryName, null, queryParams)</code>.
     *
     * @see #updateWithNamedParams(String, Set, Map)
     */
    public int updateWithNamedParams(String queryName, Map queryParams)
        throws BLException {

        return this.updateWithNamedParams(queryName, null, queryParams);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Status XMLs
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Retrieves information about the DAO and returns it as an XML text.
     *
     * @return XML for the DAO info and stats.
     *
     * @throws UnsupportedEncodingException if UTF-8 is not supported on this
     * platform, which is practically impossible.
     */
    public byte [] getStats()
        throws UnsupportedEncodingException {

        ByteArrayOutputStream out = new ByteArrayOutputStream(4096);
        PrintStream res = new PrintStream(out, false, "UTF-8");

        res.println("<daostats>");
        synchronized(this.queryStats) {
            Iterator i = this.queryStats.values().iterator();
            int n = 0;
            while(i.hasNext()) {
                QueryStatsRecord stats = (QueryStatsRecord)i.next();
                res.print("  <query n=\""); res.print(n++); res.println("\">");
                res.print("    <sql><![CDATA["); res.print(stats.queryText); res.println("]]></sql>");
                if(stats.name != null) {
                    res.print("    <name><![CDATA["); res.print(stats.name); res.println("]]></name>");
                }
                res.print("    <nExecs>"); res.print(stats.execStats.getNTimes()); res.println("</nExecs>");
                res.print("    <execsPerMin>"); res.print(stats.execStats.getEventsPerMinuteAvg()); res.println("</execsPerMin>");
                res.print("    <execTimeMin>"); res.print(stats.execStats.getProcTimeMin()); res.println("</execTimeMin>");
                res.print("    <execTimeAvg>"); res.print(stats.execStats.getProcTimeAvg()); res.println("</execTimeAvg>");
                res.print("    <execTimeMax>"); res.print(stats.execStats.getProcTimeMax()); res.println("</execTimeMax>");
                res.print("    <procTimeMin>"); res.print(stats.procStats.getProcTimeMin()); res.println("</procTimeMin>");
                res.print("    <procTimeAvg>"); res.print(stats.procStats.getProcTimeAvg()); res.println("</procTimeAvg>");
                res.print("    <procTimeMax>"); res.print(stats.procStats.getProcTimeMax()); res.println("</procTimeMax>");
                long timeMillis = (stats.execStats.getProcTimeAvg() + stats.procStats.getProcTimeAvg())*stats.execStats.getNTimes();
                res.print("    <timeMillis>"); res.print(timeMillis); res.println("</timeMillis>");
                res.print("    <time>"); res.print(TextUtils.timeIntervalToString(timeMillis)); res.println("</time>");
                res.println("  </query>");
            }
        }
        res.println("</daostats>");

        res.flush();
        return out.toByteArray();
    }

    /**
     * Not used in this class, always returns <code>null</code>.
     */
    public EventStatistics getEventStats(String eventName) {

        return null;
    }
}
