package com.boylesoftware.cb2;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.SQLException;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.util.XMLUtils;
import com.boylesoftware.cb2.util.TextUtils;
import com.boylesoftware.cb2.util.pool.AdaptiveObjectPool;


/**
 * DAO configuration built from a <code>&lt;dao-config&gt;</code> element in
 * the <code>blo-config.xml</code> file. A <code>DAOConfig</code> object is
 * required to create any instance of DAO. Multiple DAO instances can share
 * the same DAO configuration object. The reason behind separating DAO
 * configuration from the DAO itself, is that parsing a
 * <code>&lt;dao-config&gt;</code> element and bulding a <code>DAOConfig</code>
 * object out of it is a very expensive operation, so having a separate object
 * allows to perform it once and then quickly create DAO instances on its
 * basis.
 *
 * <p>The BL Manager, being initialized, creates and then keeps instances of
 * <code>DAOConfig</code> for each configured data source. References to those
 * objects can be retrieved from within any BLO via its
 * <code>getDAOConfig</code> protected method.
 *
 * <p>The <code>DAOConfig</code> logs messages related to its operation under
 * "com.boylesoftware.cb2.DAOCFG" log channel.
 *
 * @author Lev Himmelfarb
 * @version $Id: DAOConfig.java,v 1.10 2004/10/11 18:17:36 levahim Exp $
 */
public final class DAOConfig {

    /**
     * Subsystem name prefix for the logger.
     */
    private final static String LOG_SUBSYS = "com.boylesoftware.cb2.DAOCFG";

    /**
     * The default field name separator if no
     * <code>field-name-separator</code> element is specified in the
     * <code>dao-config</code> element of the
     * <code>blo-config.xml</code> file.
     */
    private final static String DEFAULT_FIELD_NAME_SEPARATOR = "$";


    ///////////////////////////////////////////////////////////////////////////
    // Configuration elements descriptors
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Represents a descriptor of a single DM as defined in the DAO
     * configuration file.
     */
    static class DMDescriptor {

        /**
         * Descriptor of a single "id" field.
         */
        static class IDFieldDescriptor {

            /**
             * The field name.
             */
            public final String name;

            /**
             * Name of the SQL query in the configuration file that is
             * used as the source of new values for the id field.
             */
            public final String srcQuery;

            /**
             * True if the srcQuery should be executed after the
             * INSERT statement.
             */
            public final boolean isPost;


            /**
             * The constructor.
             */
            public IDFieldDescriptor(String name, String srcQuery, String srcOrder) {

                this.name = name;
                this.srcQuery = srcQuery;
                this.isPost = srcOrder.equals("post");
            }
        }


        /**
         * Descriptor of a single "const" field.
         */
        static class ConstFieldDescriptor {

            public static final int APPLYTO_ALL     = 0;
            public static final int APPLYTO_INSERTS = 1;
            public static final int APPLYTO_UPDATES = 2;

            /**
             * The field name.
             */
            public final String name;

            /**
             * The applyTo property. One of <code>APPLYTO_XXX</code>
             * constants.
             */
            public final int applyTo;

            /**
             * The field's value.
             */
            public final String sql;


            /**
             * The constructor.
             */
            public ConstFieldDescriptor(String name, String applyTo, String sql) {

                this.name = name;
                if("inserts".equals(applyTo))
                    this.applyTo = APPLYTO_INSERTS;
                else if("updates".equals(applyTo))
                    this.applyTo = APPLYTO_UPDATES;
                else
                    this.applyTo = APPLYTO_ALL;
                this.sql = sql;
            }
        }


        /**
         * The DM's name.
         */
        public final String name;

        /**
         * The name of the database table associated with this DM or
         * <code>null</code>.
         */
        public final String tableName;

        /**
         * List of fields that uniquely define this DM instance in the
         * database table. The elements are in the same order as they
         * appear in the configuration file.
         */
        public final IDFieldDescriptor [] idFields;

        /**
         * List of fields that substituted with constant SQL code in
         * INSERT and UPDATE queries generated by the DAO. The keys
         * are field names (objects of <code>String</code>) and the
         * values are field descriptors (objects of
         * <code>ConstFieldDescriptor</code>).
         */
        public final Map constFields;

        /**
         * The id fields' names in a set, so we can easily check if
         * some name is an id field name.
         */
        public final Set idFieldNamesSet;

        /**
         * The keys are all upper-case DM class field names (objects
         * of <code>String</code>) and the values are the respective
         * field reflection objects (objects of <code>Field</code>).
         */
        public final Map fields;


        /**
         * Constructs a desctriptor. Used for debug (in test <code>main</code>
         * methods).
         */
        public DMDescriptor(String name,
                            String tableName,
                            IDFieldDescriptor [] idFields,
                            Map constFields,
                            Class dmClass) {

            // basic attributes
            this.name = name;
            this.tableName = tableName;

            // dm class fields
            this.fields =  new HashMap(this.getDMFields(dmClass));

            // id and const fields
            this.idFields = idFields;
            this.constFields = constFields;

            // id field names set
            this.idFieldNamesSet = new HashSet(idFields.length);
            for(int i = 0; i < idFields.length; i++)
                this.idFieldNamesSet.add(idFields[i].name);
        }

        /**
         * Constructs a descriptor.
         *
         * @throws BLException if the element is invalid.
         */
        public DMDescriptor(Element dmElement, Class dmClass)
            throws BLException {

            // basic attributes
            this.name = dmElement.getAttribute("name");
            this.tableName = TextUtils.getNullIfEmpty(dmElement.getAttribute("table"));

            // dm class fields
            this.fields =  new HashMap(this.getDMFields(dmClass));

            // id fields list
            List idFieldsList = new ArrayList();
            NodeList elements = dmElement.getElementsByTagName("idfield");
            for(int i = 0; i < elements.getLength(); i++) {
                Element element = (Element)elements.item(i);
                String fieldName = element.getAttribute("name");
                if(!this.fields.containsKey(fieldName.toUpperCase()))
                    throw new BLException("Id field \"" + fieldName +
                                          "\" is not present in the DM class " +
                                          dmClass.getName() + ".");
                idFieldsList.add(new IDFieldDescriptor(fieldName,
                                                       element.getAttribute("srcquery"),
                                                       element.getAttribute("srcorder")));
            }
            this.idFields = (IDFieldDescriptor [])idFieldsList.
                toArray(new IDFieldDescriptor[idFieldsList.size()]);

            // id field names set
            this.idFieldNamesSet = new HashSet(this.idFields.length);
            for(int i = 0; i < this.idFields.length; i++) {
                if(!this.idFieldNamesSet.add(this.idFields[i].name))
                    throw new BLException("Two or more idfields named \"" +
                                          this.idFields[i].name + "\" are defined for DM " +
                                          this.name + ".");
            }

            // const fields list
            this.constFields = new HashMap();
            elements = dmElement.getElementsByTagName("constfield");
            for(int i = 0; i < elements.getLength(); i++) {
                Element element = (Element)elements.item(i);
                String fieldName = element.getAttribute("name");
                if(!this.fields.containsKey(fieldName.toUpperCase()))
                    throw new BLException("Const field \"" + fieldName +
                                          "\" is not present in the DM class " +
                                          dmClass.getName() + ".");
                ConstFieldDescriptor cfd = new ConstFieldDescriptor(fieldName,
                                                                    element.getAttribute("applyto"),
                                                                    XMLUtils.getXMLSubelementText(element, "sql"));
                if(this.constFields.put(cfd.name, cfd) != null)
                    throw new BLException("Two or more constfields named \"" +
                                          cfd.name + "\" are defined for DM " +
                                          this.name + ".");
            }
        }

        /**
         * Enumerates valid DM fields and returns them as a map.
         *
         * @param dmClass the DM class.
         *
         * @return map where keys are all uppercase field names and values are
         * respective field reflection objects.
         */
        private Map getDMFields(Class dmClass) {

            Field [] fieldsArray = dmClass.getFields();
            Map fields =  new HashMap(fieldsArray.length);
            for(int i = 0; i < fieldsArray.length; i++) {
                int fieldMods = fieldsArray[i].getModifiers();
                if(Modifier.isPublic(fieldMods) &&
                   !(Modifier.isStatic(fieldMods) || Modifier.isFinal(fieldMods)))
                    fields.put(fieldsArray[i].getName().toUpperCase(),
                               fieldsArray[i]);
            }

            return fields;
        }
    }


    /**
     * Represents a SQL macro definition.
     */
    static class SQLMacroDescriptor {

        /**
         * The macro's name.
         */
        public final String name;

        /**
         * List of parameter names in the order as they appear in the
         * macro definition.
         */
        public final String [] params;

        /**
         * The SQL body of the macro.
         */
        public final String sql;


        /**
         * Constructs a descriptor. Used for debug (in test <code>main</code>
         * methods).
         *
         * @param name the macro name.
         * @param params the param names.
         * @param sql the macro body.
         */
        public SQLMacroDescriptor(String name,
                                  String [] params,
                                  String sql) {

            this.name = name;
            this.params = params;
            this.sql = sql;
        }

        /**
         * Constructs a descriptor for a specified
         * <code>sqlmacro</code> XML element.
         *
         * @param sqlmacroElement the <code>sqlmacro</code> XML
         * element.
         */
        public SQLMacroDescriptor(Element sqlmacroElement)
            throws BLException {

            this.name = sqlmacroElement.getAttribute("name");
            NodeList elements = sqlmacroElement.getElementsByTagName("param");
            this.params = new String[elements.getLength()];
            Set addedNames = new HashSet(elements.getLength());
            for(int i = 0; i < this.params.length; i++) {
                String paramName = ((Element)elements.item(i)).getAttribute("name");
                if(addedNames.contains(paramName))
                    throw new BLException("There are two or more parameters named \"" + paramName +
                                          "\" defined for macro \"" + this.name + "\".");
                this.params[i] = paramName;
                addedNames.add(paramName);
            }
            this.sql = XMLUtils.getXMLSubelementText(sqlmacroElement, "sql").trim();
        }
    }


    /**
     * Represents a query descriptor in the configuration file.
     */
    static class QueryDescriptor {

        /**
         * Descriptor of a condition included in the SQL query text.
         */
        static class Condition {

            /**
             * Name of the condition.
             */
            public final String name;

            /**
             * Insertion point for the condition body in the query
             * text.
             */
            public final int index;

            /**
             * The condition's body text.
             */
            public final String text;


            /**
             * Constructs a condition descriptor.
             */
            public Condition(String name, int index, String text) {

                this.name = name;
                this.index = index;
                this.text = text;
            }


            /**
             * Gets string representation of this condition. For debug
             * purposes.
             */
            public String toString() {

                return "COND{NAME=[" + this.name +
                       "],INDEX=[" + this.index +
                       "],TEXT=[" + this.text + "]}";
            }
        }


        /**
         * Descriptor of a single parameter placeholder in the query
         * text.
         */
        static class ParamPlaceholder {

            /**
             * Name of the parameter. For unnamed parameters this
             * field is set to the placeholder's position index
             * starting from 1.
             */
            public final String name;

            /**
             * SQL type of the parameter (see <code>java.sql.Types</code>)
             * or <code>null</code> if not specified.
             */
            public final Integer sqlType;

            /**
             * SQL type of the parameter to be used with the
             * <code>setNull</code> method, or <code>null</code> if not
             * specified. If not specified,
             * <code>java.sql.Types.VARCHAR</code> will be used.
             */
            public final Integer sqlTypeForNulls;

            /**
             * If the placeholder appears in a condition context (that
             * is inside <code>$[condition:...]</code> structure) the
             * name of the condition goes into this field, otherwise
             * this field is <code>null</code>.
             */
            public final String conditionName;

            /**
             * Set to <code>true</code> for array parameters
             * placeholders (<code>??</code> and
             * <code>${??:name}</code>).
             */
            public final boolean isArray;


            /**
             * Constructs a placeholder descriptor.
             */
            public ParamPlaceholder(String name,
                                    Integer sqlType,
                                    Integer sqlTypeForNulls,
                                    String conditionName,
                                    boolean isArray) {

                this.name = name;
                this.sqlType = sqlType;
                this.sqlTypeForNulls = sqlTypeForNulls;
                this.conditionName = conditionName;
                this.isArray = isArray;
            }


            /**
             * Gets string representation of this placeholder. For debug
             * purposes.
             */
            public String toString() {

                return "QPARAM{NAME=[" + this.name +
                       "],TYPE=[" + this.sqlType +
                       "],COND=[" + this.conditionName +
                       "],ARRAY=[" + this.isArray + "]}";
            }
        }


        /**
         * Descriptor for a single sql chunk within the query.
         */
        static class SqlChunk {

            /**
             * Reference to our parent query descriptor.
             */
            private final QueryDescriptor queryDescriptor;

            /**
             * Optional name, is chunk's position starting from 1 for
             * unnamed chunks.
             */
            public final String name;

            /**
             * Collection of names (objects of <code>String</code>) of
             * sql chunks this one "depends" on.
             */
            public final Set depends;

            /**
             * Set to <code>true</code> for the primary chunk.
             */
            public final boolean isResult;

            /**
             * Index of the first param placeholder in the parent's
             * <code>paramPlaceholders</code> array which belongs to
             * this chunk.
             */
            public final int firstParamPlaceholderIndex;

            /**
             * Index of the last param placeholder in the parent's
             * <code>paramPlaceholders</code> array which belongs to
             * this chunk.
             */
            public final int lastParamPlaceholderIndex;

            /**
             * The chunk's body.
             */
            private final String sql;

            /**
             * List of conditions defined in the chunk.
             */
            private final Condition [] conditions;


            /**
             * Constructs an sql chunk descriptor.
             */
            public SqlChunk(QueryDescriptor queryDescriptor,
                            String name,
                            String depends,
                            boolean isResult,
                            String sql,
                            Condition [] conditions,
                            int firstParamPlaceholderIndex,
                            int lastParamPlaceholderIndex) {

                this.queryDescriptor = queryDescriptor;
                this.name = name;
                if(depends != null) {
                    StringTokenizer st = new StringTokenizer(depends);
                    this.depends = new HashSet(st.countTokens());
                    while(st.hasMoreTokens())
                        this.depends.add(st.nextToken());
                } else {
                    this.depends = new HashSet(0);
                }
                this.isResult = isResult;
                this.sql = sql;
                this.conditions = conditions;
                this.firstParamPlaceholderIndex = firstParamPlaceholderIndex;
                this.lastParamPlaceholderIndex = lastParamPlaceholderIndex;
            }

            /**
             * Gets sql text of the chunk with all macroses expanded
             * and conditions processed.
             *
             * @param conditions set of condition
             * (<code>$[condition:&lt;condition name&gt;:&lt;SQL
             * text&gt;]</code> tags in the query text) names (objects
             * of <code>String</code>) that should be included into
             * the query. If this parameter is <code>null</code> no
             * condition tags processing is performed on the query
             * text. Note the difference between <code>null</code> and
             * an empty set!
             *
             * @return the sql text.
             */
            public String getSql(Set conditions) {

                if(conditions == null)
                    return this.sql;

                StringBuffer resQuery = new StringBuffer();
                int pos = 0;
                for(int i = 0; i < this.conditions.length; i++) {
                    resQuery.append(this.sql.substring(pos, this.conditions[i].index));
                    if(conditions.contains(this.conditions[i].name))
                        resQuery.append(this.conditions[i].text);
                    pos = this.conditions[i].index;
                }
                if(pos < this.sql.length())
                    resQuery.append(this.sql.substring(pos));

                return resQuery.toString();
            }

            /**
             * Counts number of param placeholders for this chunk with
             * the given conditions.
             *
             * @param conditions collection of condition names to
             * include.
             *
             * @return number of param placeholders.
             */
            public int getNParams(Set conditions) {

                int res = 0;
                for(int i = this.firstParamPlaceholderIndex; i <= this.lastParamPlaceholderIndex; i++) {
                    if((this.queryDescriptor.paramPlaceholders[i].conditionName == null) ||
                       ((conditions != null) && conditions.contains(this.queryDescriptor.paramPlaceholders[i].conditionName)))
                        res++;
                }

                return res;
            }
        }


        /**
         * Name of the query.
         */
        public final String name;

        /**
         * List of sql chunks defined for the query.
         */
        public final SqlChunk [] sqls;

        /**
         * List of query parameter placeholders in the query text.
         */
        public final ParamPlaceholder [] paramPlaceholders;


        /**
         * Constructs a descriptor.
         */
        public QueryDescriptor(Element queryElement,
                               Map dms,
                               Map dmDescriptors,
                               Map queryDMClasses,
                               Map sqlmacros,
                               String fieldNameSeparator)
            throws BLException {

            // get the query name
            this.name = queryElement.getAttribute("name");

            // process the usedm attribute and add the DM class to queryDMClasses
            if(queryElement.hasAttribute("usedm")) {
                Class dmClass = (Class)dms.get(queryElement.getAttribute("usedm"));
                if(dmClass == null)
                    throw new BLException("Cannot find dm element \"" + queryElement.getAttribute("usedm") +
                                          "\" for usedm attribute of query \"" + this.name +
                                          "\".");
                queryDMClasses.put(this.name, dmClass);
            }

            // get the sql chunks
            NodeList nl = queryElement.getElementsByTagName("sql");
            this.sqls = new SqlChunk[nl.getLength()];
            boolean assignedPrimary = false;
            Set definedNames = new HashSet(this.sqls.length);
            List paramPlaceholders = new ArrayList();
            for(int i = 0; i < this.sqls.length; i++) {
                Element element = (Element)nl.item(i);

                // get the chunk body
                String chunkText = XMLUtils.getXMLNodeText(element);
                if(chunkText == null)
                    throw new BLException("No text is defined in an sql subelement of query \"" +
                                          this.name + "\".");

                // process the sql XML element and create a descriptor
                if("true".equals(element.getAttribute("oldsyntax")))
                    this.sqls[i] = DAOSpecialSyntaxParserOld.
                        parseSqlChunk(element,
                                      i + 1,
                                      (i + 1) >= this.sqls.length,
                                      assignedPrimary,
                                      this,
                                      sqlmacros,
                                      dms,
                                      dmDescriptors,
                                      queryDMClasses,
                                      fieldNameSeparator,
                                      paramPlaceholders);
                else
                    this.sqls[i] = DAOSpecialSyntaxParser.
                        parseSqlChunk(this,
                                      (element.hasAttribute("name") ? element.getAttribute("name") : String.valueOf(i + 1)),
                                      "true".equals(element.getAttribute("result")) || (!assignedPrimary && ((i + 1) >= this.sqls.length)),
                                      chunkText,
                                      (element.hasAttribute("depends") ? element.getAttribute("depends") : null),
                                      sqlmacros,
                                      dms,
                                      dmDescriptors,
                                      queryDMClasses,
                                      fieldNameSeparator,
                                      paramPlaceholders);

                // check if everything is ok
                if(definedNames.contains(this.sqls[i].name))
                    throw new BLException("More than one sql chunk has name \"" + this.sqls[i].name +
                                          "\" in query \"" + this.name + "\".");
                definedNames.add(this.sqls[i].name);
                if(assignedPrimary && this.sqls[i].isResult)
                    throw new BLException("More than one sql chunk is marked as the resulting in query \"" + this.name + "\".");
                assignedPrimary = (assignedPrimary || this.sqls[i].isResult);
                if(!definedNames.containsAll(this.sqls[i].depends))
                    throw new BLException("Invalid reference in the \"depends\" attribute of sql chunk " +
                                          this.sqls[i].name + " in query \"" + this.name + "\".");
            }

            // the param placeholders list is ready
            this.paramPlaceholders = (ParamPlaceholder [])paramPlaceholders.
                toArray(new ParamPlaceholder[paramPlaceholders.size()]);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // The DAO configuration
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Our application context.
     */
    private final ApplicationContext appCtx;

    /**
     * Logical name of the data source this config object is built
     * for.
     */
    private final String dataSourceName;

    /**
     * Map between SQL query names and query descriptors. Keys are
     * query names (objects of <code>String</code>) as they appear in
     * the configuration file and values are the query descriptors
     * (objects of <code>QueryDescriptor</code>).
     */
    private final Map queries;

    /**
     * Map between SQL query names and used by the query DM class
     * objects. Keys are query names (objects of <code>String</code>)
     * as they appear in the configuration file and values are the DM
     * classes (objects of <code>Class</code>).
     */
    private final Map queryDMClasses;

    /**
     * Maps DM classes and DM descriptors from the configuration
     * file. The keys are DM classes (objects of <code>Class</code>),
     * the values are DM descriptors (objects of
     * <code>DMDescriptor</code> inner class).
     */
    private final Map dmDescriptors;

    /**
     * Set to <code>true</code> if the database driver supports
     * <code>setFetchSize</code> method on <code>Statement</code>
     * interface.
     */
    private final boolean driverSupportsSetFetchSize;

    /**
     * String that can be used to prefix DM field names in result
     * sets. It can be useful when you have fields with the same name
     * in one result set.
     */
    private final String fieldNameSeparator;

    /**
     * Pool of fetch executor instances.
     */
    private final AdaptiveObjectPool fetchExecutorsPool;

    /**
     * List of query param processor plugins.
     */
    private final QueryParamProcessor [] queryParamProcessors;

    /**
     * The query param setter plugin, or <code>null</code> to use
     * the default implementation.
     */
    private final QueryParamSetter queryParamSetter;

    /**
     * List of result set field processor plugins.
     */
    private final ResultSetFieldProcessor [] resultSetFieldProcessors;

    /**
     * The result set field getter plugin, or <code>null</code> to use
     * the default implementation.
     */
    private final ResultSetFieldGetter resultSetFieldGetter;


    ///////////////////////////////////////////////////////////////////////////
    // The initialization
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructs a DAO config instance from a number of
     * <code>&lt;dao-config&gt;</code> XML elements. Contents of the specified
     * <code>&lt;dao-config&gt;</code> elements are combined together and
     * threated as one.
     *
     * @param appCtx the application context.
     * @param dataSourceName logical name of the data source, for which this
     * config is built. Any <code>dataSource</code> attributes of the specified
     * <code>&lt;dao-config&gt;</code> elements are ignored.
     * @param daoConfigs array of parsed <code>&lt;dao-config&gt;</code> XML
     * elements.
     *
     * @throws BLException if an error happens preventing the DAO config from
     * being built.
     */
    public DAOConfig(ApplicationContext appCtx,
                     String dataSourceName,
                     Element [] daoConfigs)
        throws BLException {

        // basic config
        this.appCtx = appCtx;
        this.dataSourceName = dataSourceName;

        // get the log
        Log log = appCtx.getLog(LOG_SUBSYS);

        // tell that we are starting working
        log.info("creating an instance of a DAO config object");

        // init basic configuration storages
        this.queries = new HashMap();
        this.queryDMClasses = new HashMap();
        this.dmDescriptors = new HashMap();

        // determine the field name separator
        String fieldNameSeparatorTmp = null;
        for(int i = 0; i < daoConfigs.length; i++) {
            fieldNameSeparatorTmp = XMLUtils.getXMLSubelementText(daoConfigs[i], "field-name-separator");
            if(fieldNameSeparatorTmp != null)
                break;
        }
        this.fieldNameSeparator = (fieldNameSeparatorTmp != null ?
                                   fieldNameSeparatorTmp :
                                   DEFAULT_FIELD_NAME_SEPARATOR);

        // process the plugins
        Class fetchExecutorClass = null;
        List queryParamProcessorsList = new ArrayList();
        QueryParamSetter queryParamSetter = null;
        List resultSetFieldProcessorsList = new ArrayList();
        ResultSetFieldGetter resultSetFieldGetter = null;
        for(int i = 0; i < daoConfigs.length; i++) {
            NodeList elements = daoConfigs[i].getElementsByTagName("plugin");
            if(elements.getLength() > 0) {
                try {
                    for(int j = 0; j < elements.getLength(); j++) {
                        String pluginClassName = XMLUtils.
                            getXMLSubelementText((Element)elements.item(j), "class");
                        Class pluginClass = Class.forName(pluginClassName);
                        Object plugin = pluginClass.newInstance();

                        boolean isValid = false;

                        if(plugin instanceof FetchExecutor) {
                            if(fetchExecutorClass != null)
                                throw new BLException("You cannot have more than one FetchExecutor plugin defined for a DAO.");

                            fetchExecutorClass = pluginClass;
                            isValid = true;
                        }

                        if(plugin instanceof QueryParamProcessor) {
                            queryParamProcessorsList.add(plugin);
                            isValid = true;
                        }
                        if(plugin instanceof QueryParamSetter) {
                            if(queryParamSetter != null)
                                throw new BLException("You cannot have more than one QueryParamSetter plugin defined for a DAO.");
                            queryParamSetter = (QueryParamSetter)plugin;
                            isValid = true;
                        }
                        if(plugin instanceof ResultSetFieldProcessor) {
                            resultSetFieldProcessorsList.add(plugin);
                            isValid = true;
                        }
                        if(plugin instanceof ResultSetFieldGetter) {
                            if(resultSetFieldGetter != null)
                                throw new BLException("You cannot have more than one ResultSetFieldGetter plugin defined for a DAO.");
                            resultSetFieldGetter = (ResultSetFieldGetter)plugin;
                            isValid = true;
                        }

                        if(!isValid)
                            throw new BLException("A DAO plugin implementation class " + pluginClass +
                                                  " does not implement any of the DAOPlugin descendant interfaces.");

                        ((DAOPlugin)plugin).setApplicationContext(this.appCtx);
                    }
                } catch(ClassNotFoundException e) {
                    throw new BLException("Could not load a DAO plugin implementation class.", e);
                } catch(IllegalAccessException e) {
                    throw new BLException("Could not instantiate a DAO plugin implementation class.", e);
                } catch(InstantiationException e) {
                    throw new BLException("Could not instantiate a DAO plugin implementation class.", e);
                }

                break;
            }
        }

        this.fetchExecutorsPool = new AdaptiveObjectPool(fetchExecutorClass != null ?
                                                         fetchExecutorClass :
                                                         GenericFetchExecutor.class);
        this.queryParamProcessors = (queryParamProcessorsList.size() > 0 ?
                                     (QueryParamProcessor [])queryParamProcessorsList.
                                     toArray(new QueryParamProcessor [queryParamProcessorsList.size()]) :
                                     null);
        this.queryParamSetter = queryParamSetter;
        this.resultSetFieldProcessors = (resultSetFieldProcessorsList.size() > 0 ?
                                         (ResultSetFieldProcessor [])resultSetFieldProcessorsList.
                                         toArray(new ResultSetFieldProcessor [resultSetFieldProcessorsList.size()]) :
                                         null);
        this.resultSetFieldGetter = resultSetFieldGetter;

        // process configuration elements
        Map dms = new HashMap();
        Map sqlmacros = new HashMap();
        for(int i = 0; i < daoConfigs.length; i++) {
            this.processDAOConfigElement(daoConfigs[i], dms, sqlmacros);
        }
        dms.clear();
        sqlmacros.clear();

        // done
        log.info("the DAO config has been successfully created");

        // test the JDBC driver capabilities
        log.info("testing the database driver capabilities");
        try {
            Connection con = null;
            Statement stmt = null;
            try {
                con = this.appCtx.getNoTXConnection(dataSourceName);
                stmt = con.createStatement();

                boolean driverSupportsSetFetchSize = true;
                try {
                    stmt.setFetchSize(5);
                    log.info("setFetchSize is supported");
                } catch(Exception e) {
                    log.info("setFetchSize is not supported");
                    driverSupportsSetFetchSize = false;
                }
                this.driverSupportsSetFetchSize = driverSupportsSetFetchSize;

            } finally {
                 if(stmt != null)
                     stmt.close();
                if(con != null)
                    con.close();
            }
        } catch(SQLException e) {
            log.error("Got SQLException while checking the database driver capabilities.", e);
            throw new BLException("Could not build the DAO config.", e);
        }

        // all done
        log.info("an instance of a DAO config was successfully created");
    }

    /**
     * Processes contents of the specified <code>dao-config</code> XML
     * element and populates the corresponding internal data
     * structures.
     */
    private void processDAOConfigElement(Element daoConfig, Map dms, Map sqlmacros)
        throws BLException {

        try {

            // read DM descriptors
            NodeList elements = daoConfig.getElementsByTagName("dm");
            for(int i = 0; i < elements.getLength(); i++) {
                Element element = (Element)elements.item(i);
                Class dmClass = Class.forName(XMLUtils.getXMLSubelementText(element, "class"));
                if(this.dmDescriptors.put(dmClass, new DMDescriptor(element, dmClass)) != null)
                    throw new BLException("DM class " + dmClass.getName() +
                                          " has more than one DM descriptor in the configuration file.");
                if(dms.put(element.getAttribute("name"), dmClass) != null)
                    throw new BLException("There is more than one DM descriptor named " +
                                          element.getAttribute("name"));
            }

            // read sqlmacors
            elements = daoConfig.getElementsByTagName("sqlmacro");
            for(int i = 0; i < elements.getLength(); i++) {
                Element element = (Element)elements.item(i);
                SQLMacroDescriptor desc = new SQLMacroDescriptor(element);
                if(sqlmacros.put(desc.name, desc) != null)
                    throw new BLException("There is more than one sqlmacro element named \"" +
                                          desc.name + "\".");
            }

            // read queries
            elements = daoConfig.getElementsByTagName("query");
            for(int i = 0; i < elements.getLength(); i++) {
                Element element = (Element)elements.item(i);
                QueryDescriptor desc = new QueryDescriptor(element,
                                                           dms,
                                                           this.dmDescriptors,
                                                           this.queryDMClasses,
                                                           sqlmacros,
                                                           this.fieldNameSeparator);
                 if(this.queries.put(desc.name, desc) != null)
                    throw new BLException("There are two or more queries named \"" + desc.name + "\".");
            }

        } catch(ClassNotFoundException e) {
            throw new BLException("Could not load class for a DM.", e);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Service interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets the application context associated with this config object.
     *
     * @return reference to the application context.
     */
    public ApplicationContext getApplicationContext() {

        return this.appCtx;
    }

    /**
     * Gets the data source name this object is configured for.
     *
     * @return the data source name.
     */
    String getDataSourceName() {

        return this.dataSourceName;
    }

    /**
     * Checks if the database driver of the configured data source
     * supports the <code>setFetchSize</code> method on
     * <code>Statement</code>. The driver is tested and the flag is
     * set internally in the config object during its initialization.
     *
     * @return <code>true</code> if the database driver supports
     * <code>setFetchSize</code> method.
     */
    boolean doesDriverSupportSetFetchSize() {

        return this.driverSupportsSetFetchSize;
    }

    /**
     * Returns an instance of the fetch executor as configured.
     *
     * @return a fetch executor instance.
     */
    FetchExecutor getFetchExecutor() {

        return (FetchExecutor)this.fetchExecutorsPool.get();
    }

    /**
     * Releases a fetch executor instance got through the
     * <code>getFetchExecutor</code> method.
     */
    void releaseFetchExecutor(FetchExecutor fe) {

        this.fetchExecutorsPool.returnObject(fe);
    }

    /**
     * Returns the list of installed query param processor plugins.
     *
     * @return the list of installed query param processor plugins or
     * <code>null</code>.
     */
    QueryParamProcessor [] getQueryParamProcessors() {

        return this.queryParamProcessors;
    }

    /**
     * Returns the instance of the query param setter plugin or
     * <code>null</code> if none defined.
     *
     * @return reference to the query param setter plugin instance
     * or <code>null</code>.
     */
    QueryParamSetter getQueryParamSetter() {

        return this.queryParamSetter;
    }

    /**
     * Returns the list of installed result set field processor
     * plugins.
     *
     * @return the list of installed result set field processor
     * plugins or <code>null</code>.
     */
    ResultSetFieldProcessor [] getResultSetFieldProcessors() {

        return this.resultSetFieldProcessors;
    }

    /**
     * Returns the instance of the result set field getter plugin or
     * <code>null</code> if none defined.
     *
     * @return reference to the result set field getter plugin instance
     * or <code>null</code>.
     */
    ResultSetFieldGetter getResultSetFieldGetter() {

        return this.resultSetFieldGetter;
    }

    /**
     * Returns SQL query descriptor from the DAO configuration.
     *
     * @param queryName name of the query as it is specified in the
     * DAO configuration.
     *
     * @return the SQL query descriptor.
     *
     * @throws BLException if there is no query with the specified
     * name.
     */
    QueryDescriptor getQueryDescriptor(String queryName)
        throws BLException {

        QueryDescriptor res = (QueryDescriptor)this.queries.get(queryName);

        if(res == null)
            throw new BLException("There is no SQL query defined under name \"" + queryName + "\".");

        return res;
    }

    /**
     * Finds class object for the DM associated with the specified SQL
     * query in the DAO configuration.
     *
     * @param queryName name of the query as it is specified in the
     * DAO configuration.
     *
     * @return class of the DM associated with the query through
     * <code>usedm</code> attribute in the configuration.
     *
     * @throws BLException if there is no query with the specified
     * name or it does not have a <code>usedm</code> attribute.
     */
    Class getQueryDMClass(String queryName)
        throws BLException {

        Class res = (Class)this.queryDMClasses.get(queryName);

        if(res == null)
            throw new BLException("There is no SQL query defined under name \"" + queryName +
                                  "\" or it does not have a usedm attribute.");

        return res;
    }

    /**
     * Finds descriptor for the specified by its class DM.
     *
     * @param dmClass class object of a DM.
     *
     * @return DM desriptor instance.
     *
     * @throws BLException if there is no DM descriptor for the
     * specified class in the DAO configuration.
     */
    DMDescriptor getDMDescriptor(Class dmClass)
        throws BLException {

        DMDescriptor res = (DMDescriptor)this.dmDescriptors.get(dmClass);

        if(res == null)
            throw new BLException("There is no DM descriptor for class " + dmClass.getName() + ".");

        return res;
    }

    /**
     * Finds field object for the specified by name field of the
     * specified DM class. The field name is case-insensitive. If the
     * <code>ciFieldName</code> ends with the field separator sequence
     * it means that the column should be ignored by the DAO and this
     * method returns <code>null</code> to indicate that.
     *
     * @param dmClass class object of the DM.
     * @param ciFieldName case-insensitive name of a field in the
     * specified by its class object DM.
     *
     * @return field object for the specified DM field or
     * <code>null</code> if the column should be ignored by the DAO.
     *
     * @throws BLException if there is no DM descriptor for the
     * specified class in the DAO configuration or the DM does not
     * contain the specified field.
     */
    Field getDMField(Class dmClass, String ciFieldName)
        throws BLException {

        // ignore fields that end with the separator sequence
        if(ciFieldName.endsWith(this.fieldNameSeparator))
            return null;

        // extract the actual field name from the column label
        int nameStartIndex = ciFieldName.lastIndexOf(this.fieldNameSeparator);
        if(nameStartIndex >= 0)
            ciFieldName = ciFieldName.substring(nameStartIndex + this.fieldNameSeparator.length());

        // find the field in the DM class
        Field res = (Field)this.getDMDescriptor(dmClass).fields.get(ciFieldName.toUpperCase());
        if(res == null)
            throw new BLException("There is no field \"" + ciFieldName + "\" in " + dmClass.getName() + ".");

        // return the field
        return res;
    }
}
