package com.boylesoftware.cb2;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.lang.reflect.Field;


/**
 * This helper class encapsulates the functionality for parsing DAO
 * special syntax, which may appear in the bodies of SQL queries defined
 * in the DAO configuration. This class is used by the
 * <code>DAOConfig</code> class.
 *
 * @author Lev Himmelfarb
 * @version $Id: DAOSpecialSyntaxParser.java,v 1.6 2004/10/11 18:17:36 levahim Exp $
 */
final class DAOSpecialSyntaxParser {


    ///////////////////////////////////////////////////////////////////////////
    // The Lexical Analyzer
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Represents the input SQL text and provides functionality of a
     * basic lexical analyzer.
     */
    static final class Input {

        // token types

        /**
         * End of data.
         */
        public static final int TOKEN_EOD = 0;

        /**
         * Any single character which is not a part of any other token.
         */
        public static final int TOKEN_SINGLE_CHAR = 1;

        /**
         * Whitespace.
         */
        public static final int TOKEN_SPACE = 2;

        /**
         * A word (letters, digits and the underscore).
         */
        public static final int TOKEN_WORD = 3;

        /**
         * An SQL string literal (text in single quotes).
         */
        public static final int TOKEN_LITERAL = 4;

        /**
         * SQL comment.
         */
        public static final int TOKEN_COMMENT = 5;

        /**
         * A "\," sequence.
         */
        public static final int TOKEN_ESCAPED_COMMA = 6;

        /**
         * A double-question mark ("??" sequence).
         */
        public static final int TOKEN_DOUBLE_QMARK = 7;

        /**
         * Opening of a "macro" tag.
         */
        public static final int TOKEN_TAG_MACRO = 8;

        /**
         * Opening of a "dm" tag.
         */
        public static final int TOKEN_TAG_DM = 9;

        /**
         * Opening of a "cond" tag.
         */
        public static final int TOKEN_TAG_COND = 10;

        /**
         * Opening of a named query param tag.
         */
        public static final int TOKEN_TAG_QPARAM = 11;

        /**
         * A "(".
         */
        public static final int TOKEN_LRB = 12;

        /**
         * A ")".
         */
        public static final int TOKEN_RRB = 13;

        /**
         * A "{".
         */
        public static final int TOKEN_LCB = 14;

        /**
         * A "}".
         */
        public static final int TOKEN_RCB = 15;

        /**
         * A "?".
         */
        public static final int TOKEN_QMARK = 16;

        /**
         * A ",".
         */
        public static final int TOKEN_COMMA = 17;


        // tag opening regexp

        /**
         * Matches opening of a special syntax tag.
         */
        private static final Pattern PATTERN_TAG_OPENING =
            Pattern.compile("\\{\\s*(?:cb2\\s*:\\s*)?(macro|dm|cond|\\?\\??)");


        // input context

        /**
         * Represents input context - a set of attributes forming
         * a single piece of input text.
         */
        private static final class InputContext {

            /**
             * The input text as a string.
             */
            public final String inputTextS;

            /**
             * The input text as an array.
             */
            public final char [] inputText;

            /**
             * Tag opening matcher.
             */
            public final Matcher tagOpeningMatcher;

            /**
             * Index of a character in the input text from which to start scanning
             * for the next token.
             */
            public int tailIndex;


            /**
             * Constructs a context for the specified string.
             *
             * @param inputText the input text.
             * @param tagOpeningMatcher matcher for tag openings pointed to
             * the input text.
             */
            public InputContext(String inputText, Matcher tagOpeningMatcher) {

                this.inputTextS = inputText;
                this.inputText = inputText.toCharArray();
                this.tagOpeningMatcher = tagOpeningMatcher;
                this.tailIndex = 0;
            }
        }


        // data

        /**
         * Stack of contexts to extract tokens from.
         */
        private final LinkedList contextsStack;

        /**
         * Type of the current token as returned by the last call
         * to <code>nextToken</code>.
         */
        private int tokenType;

        /**
         * Text of the token found by the last successful
         * <code>nextToken</code> call.
         */
        private StringBuffer token;


        // constructor

        /**
         * Constructs the input text tokenizer.
         *
         * @param inputText the input text.
         */
        public Input(String inputText) {

            this.contextsStack = new LinkedList();
            this.tokenType = -1;
            this.token = new StringBuffer();

            // push the top-level context
            this.pushInput(inputText);
        }


        // methods

        /**
         * Scans the input from the last point and returns the next token.
         *
         * @return type of the token found (one of <code>TOKEN_*</code>
         * constants) or <code>TOKEN_EOD</code> if the end of the input
         * text has been reached.
         *
         * @throws DAOSpecialSyntaxException if a syntax error is detected.
         */
        public int nextToken()
            throws DAOSpecialSyntaxException {

            // erase the old token
            this.tokenType = -1;
            this.token.setLength(0);

            // get current context
            InputContext ctx = null;
            do {
                // no input anymore?
                if(this.contextsStack.size() <= 0) {
                    this.tokenType = TOKEN_EOD;
                    break;
                }

                // get the context on the top of the stack
                ctx = (InputContext)this.contextsStack.getLast();

                // switch context if the top one is depleted
                if(ctx.tailIndex >= ctx.inputText.length) {
                    this.contextsStack.removeLast();
                    ctx = null;
                }

            } while(ctx == null);

            // got input
            if(this.tokenType < 0) {

                // get the char
                char c = ctx.inputText[ctx.tailIndex++];
                this.token.append(c);

                // whitespace?
                if(Character.isWhitespace(c)) {
                    while((ctx.tailIndex < ctx.inputText.length) &&
                          (Character.isWhitespace(ctx.inputText[ctx.tailIndex])))
                        this.token.append(ctx.inputText[ctx.tailIndex++]);
                    this.tokenType = TOKEN_SPACE;

                // string literal?
                } else if(c == '\'') {
                    while(ctx.tailIndex < ctx.inputText.length) {
                        c = ctx.inputText[ctx.tailIndex++];
                        this.token.append(c);
                        if(c == '\'') {
                            if((ctx.tailIndex >= ctx.inputText.length) ||
                               (ctx.inputText[ctx.tailIndex] != '\'')) {
                                this.tokenType = TOKEN_LITERAL;
                                break;
                            }
                            this.token.append(ctx.inputText[ctx.tailIndex++]);
                        }
                    }
                    if(this.tokenType < 0)
                        throw new DAOSpecialSyntaxException("Unexpected end of text in a string literal.");

                // single-line comment?
                } else if((c == '-') &&
                          (ctx.tailIndex < ctx.inputText.length) &&
                          (ctx.inputText[ctx.tailIndex] == '-')) {
                    this.token.append(ctx.inputText[ctx.tailIndex++]);
                    while(ctx.tailIndex < ctx.inputText.length) {
                        c = ctx.inputText[ctx.tailIndex++];
                        this.token.append(c);
                        if(c == '\n')
                            break;
                    }
                    this.tokenType = TOKEN_COMMENT;

                // multi-line comment?
                } else if((c == '/') &&
                          (ctx.tailIndex < ctx.inputText.length) &&
                          (ctx.inputText[ctx.tailIndex] == '*')) {

                    int e = ctx.inputTextS.indexOf("*/", ctx.tailIndex + 1);
                    if(e < 0)
                        throw new DAOSpecialSyntaxException("Unexpected end of text in a comment.");
                    e += 2;
                    this.token.append(ctx.inputText, ctx.tailIndex, e - ctx.tailIndex);
                    ctx.tailIndex = e;
                    this.tokenType = TOKEN_COMMENT;

                // double-question mark?
                } else if((c == '?') &&
                          (ctx.tailIndex < ctx.inputText.length) &&
                          (ctx.inputText[ctx.tailIndex] == '?')) {
                    this.token.append(ctx.inputText[ctx.tailIndex++]);
                    this.tokenType = TOKEN_DOUBLE_QMARK;

                // escaped comma?
                } else if((c == '\\') &&
                          (ctx.tailIndex < ctx.inputText.length) &&
                          (ctx.inputText[ctx.tailIndex] == ',')) {
                    this.token.append(ctx.inputText[ctx.tailIndex++]);
                    this.tokenType = TOKEN_ESCAPED_COMMA;

                // CB2 tag opening?
                } else if(ctx.tagOpeningMatcher.find(ctx.tailIndex - 1) &&
                          (ctx.tagOpeningMatcher.start() == ctx.tailIndex - 1)) {
                    String macroName = ctx.tagOpeningMatcher.group(1);
                    if(macroName.equals("macro"))
                        this.tokenType = TOKEN_TAG_MACRO;
                    else if(macroName.equals("dm"))
                        this.tokenType = TOKEN_TAG_DM;
                    else if(macroName.equals("cond"))
                        this.tokenType = TOKEN_TAG_COND;
                    else if(macroName.startsWith("?"))
                        this.tokenType = TOKEN_TAG_QPARAM;
                    else // should never happen
                        throw new Error("Regexp has matched, but the tag name is wrong.");

                    this.token.append(ctx.inputText,
                                      ctx.tailIndex,
                                      ctx.tagOpeningMatcher.group().length() - 1);
                    ctx.tailIndex = ctx.tagOpeningMatcher.end();

                // a word?
                } else if(Character.isUnicodeIdentifierPart(c)) {
                    while((ctx.tailIndex < ctx.inputText.length) &&
                          (Character.isUnicodeIdentifierPart(ctx.inputText[ctx.tailIndex])))
                        this.token.append(ctx.inputText[ctx.tailIndex++]);
                    this.tokenType = TOKEN_WORD;

                // return as a single-char token
                } else {
                    switch(c) {
                    case '(':
                        this.tokenType = TOKEN_LRB;
                        break;
                    case ')':
                        this.tokenType = TOKEN_RRB;
                        break;
                    case '{':
                        this.tokenType = TOKEN_LCB;
                        break;
                    case '}':
                        this.tokenType = TOKEN_RCB;
                        break;
                    case '?':
                        this.tokenType = TOKEN_QMARK;
                        break;
                    case ',':
                        this.tokenType = TOKEN_COMMA;
                        break;
                    default:
                        this.tokenType = TOKEN_SINGLE_CHAR;
                    }
                }
            }

            // done, return
            return this.tokenType;
        }

        /**
         * Get the last found token's type.
         *
         * @return the token type.
         */
        public int getTokenType() {

            return this.tokenType;
        }

        /**
         * Get the last found token's text.
         *
         * @return the token text.
         */
        public String getToken() {

            return this.token.toString();
        }

        /**
         * Pushes text on top of the input context stack so the next call to
         * <code>nextToken</code> will start scanning the just pushed text.
         * After the pushed text is finished the scanning resumes from the
         * point it stopped in the previous input context.
         *
         * @param inputString text to push on to the top of the stack.
         */
        public void pushInput(String inputString) {

            this.contextsStack.add(new InputContext(inputString,
                                               PATTERN_TAG_OPENING.matcher(inputString)));
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // The State Machine Classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Parser context.
     */
    static final class Context {

        /**
         * DM context.
         */
        static final class DMContext {

            /**
             * Parent DM context or <code>null</code> for the top-level DM.
             */
            public final DMContext parent;

            /**
             * The DM descriptor.
             */
            public final DAOConfig.DMDescriptor desc;

            /**
             * Field in the parent DM to which we are storing this nested
             * DM, or <code>null</code> for the top-level DM.
             */
            public final Field parentField;

            /**
             * Table name.
             */
            public String tableName;

            /**
             * Field alias prefix.
             */
            public String prefix;

            /**
             * Set of names of fields to exclude.
             */
            public final Set excludeFields;

            /**
             * Set of names of fields to include.
             */
            public final Set onlyFields;

            /**
             * For array nested DMs name of the field in the parent DM
             * to use in the linking header select column. If not specified
             * the first id field is used.
             */
            public Field byField;

            /**
             * Number of columns in all nested DMs.
             */
            public int nNestedFields;

            /**
             * Number of nested arrays in this DM.
             */
            public int nNestedArrays;

            /**
             * Tells if the last nested DM in this DM was an array.
             */
            public boolean wasNestedArray;


            /**
             * Constructs a context.
             *
             * @param parent parent DM context or <code>null</code> for the
             * top-level DM.
             * @param desc the DM descriptor.
             * @param parentField field in the parent DM to which we are
             * storing this nested DM, or <code>null</code> for the
             * top-level DM.
             * @param dmFieldSeparator DM field alias separator string.
             */
            public DMContext(DMContext parent,
                             DAOConfig.DMDescriptor desc,
                             Field parentField,
                             String dmFieldSeparator) {

                this.parent = parent;
                this.desc = desc;
                this.parentField = parentField;
                this.tableName = desc.tableName;
                if(parent != null) {
                    if(parent.prefix != null)
                        this.prefix = parent.prefix + dmFieldSeparator + parentField.getName();
                    else
                        this.prefix = parentField.getName();
                } else
                    this.prefix = null;
                this.excludeFields = new HashSet();
                this.onlyFields = new HashSet();
                if((parent != null) && parentField.getType().isArray()) {
                    if(parent.desc.idFields.length > 0)
                        this.byField = (Field)parent.desc.fields.
                            get(parent.desc.idFields[0].name.toUpperCase());
                } else
                    this.byField = null;
                this.nNestedFields = 0;
                this.nNestedArrays = 0;
                this.wasNestedArray = false;
            }
        }


        // keyword modes, used as value in currentDMKWMode
        public static final int DMKW_MODE_NOKW = 0;
        public static final int DMKW_MODE_FROM = 1;
        public static final int DMKW_MODE_PREFIX = 2;
        public static final int DMKW_MODE_BY = 3;
        public static final int DMKW_MODE_EXCLUDING = 4;
        public static final int DMKW_MODE_ONLY = 5;


        // input

        /**
         * The parser input.
         */
        public final Input in;


        // output

        /**
         * The parser output.
         */
        public StringBuffer out;

        /**
         * First param placeholder index.
         */
        public final int firstParamPlaceholderIndex;


        // reference info

        public final String queryName;
        public final Map sqlMacros;
        public final Map dmDescriptors;
        public final Map queryDMClasses;
        public final String fieldNameSeparator;

        // output

        /**
         * Collected condition descriptors.
         */
        public final List conds;

        /**
         * Collected parameter placeholder descriptors.
         */
        public final List paramPlaceholders;


        // context

        /**
         * Stack of paired tokens.
         */
        public final char [] tokenStack;

        /**
         * Index of the top element on the <code>tokenStack</code> or
         * <code>-1</code> if the stack is empty.
         */
        public int tokenStackTop;

        /**
         * Current list, elements are of type <code>StringBuffer</code>.
         */
        public final LinkedList currentList;

        /**
         * The "real" output saved temporarily.
         */
        public StringBuffer savedOut;

        /**
         * Name of the current condition, that is the condition whose
         * body we are parsing at this moment, or <code>null</code> if
         * we are not in a condition body context.
         */
        public String currentConditionName;

        /**
         * Index in the output SQL, from which all conditions are cut out,
         * of where the currently being processed condition should be
         * inserted.
         */
        public int currentConditionIndex;

        /**
         * SQL type of the current query param, or <code>null</code>.
         */
        public Integer currentQParamType;

        /**
         * Tells if the current query parameter type is meant for both
         * <code>setNull</code> and <code>setObject</code>, otherwise
         * it is used only for <code>setNull</code>.
         */
        public boolean currentQParamTypeForBoth;

        /**
         * Name of current named query param, or <code>null</code>.
         */
        public String currentQParamName;

        /**
         * Descriptor of the macro, which reference we are currently
         * processing, or <code>null</code>.
         */
        public DAOConfig.SQLMacroDescriptor currentMacroDesc;

        /**
         * Stack of nested DM descriptors, elements are of type
         * <code>DMContext</code>.
         */
        public final LinkedList dmDescStack;

        /**
         * List of all DM descriptors in the order they should be
         * renedered to the out; elements are of type <code>DMContext</code>.
         */
        public final List dmDescList;

        /**
         * Current dm keyword mode, is set after a dm tag keyword is received
         * indicating that the next token should be the keyword argument.
         */
        public int currentDMKWMode;


        /**
         * Constructs a context.
         *
         * @param in the parser input.
         * @param paramPlaceholders parameter placeholder descriptors.
         * @param queryName name of the query we are parsing.
         * @param sqlMacros sql macro descriptors by names.
         * @param dmDescriptors DM classes to DM decriptors.
         * @param queryDMClasses query names to used DM classes.
         * @param fieldNameSeparator the field name separator string.
         */
        public Context(Input in,
                       List paramPlaceholders,
                       String queryName,
                       Map sqlMacros,
                       Map dmDescriptors,
                       Map queryDMClasses,
                       String fieldNameSeparator) {

            this.in = in;
            this.out = new StringBuffer();
            this.firstParamPlaceholderIndex = paramPlaceholders.size();

            this.queryName = queryName;
            this.sqlMacros = sqlMacros;
            this.dmDescriptors = dmDescriptors;
            this.queryDMClasses = queryDMClasses;
            this.fieldNameSeparator = fieldNameSeparator;

            this.conds = new ArrayList();
            this.paramPlaceholders = paramPlaceholders;

            this.tokenStack = new char[128];
            this.tokenStackTop = -1;
            this.currentList = new LinkedList();
            this.savedOut = null;
            this.currentConditionName = null;
            this.currentConditionIndex = -1;
            this.currentQParamType = null;
            this.currentQParamTypeForBoth = false;
            this.currentQParamName = null;
            this.currentMacroDesc = null;
            this.dmDescStack = new LinkedList();
            this.dmDescList = new ArrayList();
            this.currentDMKWMode = DMKW_MODE_NOKW;
        }
    }


    /**
     * Action to take upon arrival to a state. All concrete actions are
     * derived from this base class.
     */
    static abstract class Action {

        /**
         * Invoke the action.
         *
         * @param ctx the current parser context.
         *
         * @return <code>true</code> if the alternative next state should
         * be used by the machine.
         *
         * @throws DAOSpecialSyntaxException if an error is detected.
         */
        public abstract boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException;
    }


    /**
     * State transition in the state matrix. Basically, it is a pair of
     * the action to take upon arrival to this state, and the next state.
     */
    private static final class StateTrans {

        /**
         * Action to invoke upon arrival to this state.
         */
        public final Action action;

        /**
         * Next state if the action is successful. If it is less than zero
         * it is a final state.
         */
        public final int nextState;

        /**
         * Alternate next state.
         */
        public final int altNextState;


        /**
         * Constructs a state transition object to put into the state
         * matrix.
         *
         * @param action action to take upon arrival to this state.
         * @param nextState the next state.
         * @param altNextState alternate next state.
         */
        public StateTrans(Action action, int nextState, int altNextState) {

            this.action = action;
            this.nextState = nextState;
            this.altNextState = altNextState;
        }

        /**
         * Constructs a state transition object with the same next state and
         * the alternate next state.
         *
         * @param action the state action.
         * @param nextState the next state.
         */
        public StateTrans(Action action, int nextState) {
            this(action, nextState, nextState);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // State Actions
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Does nothing.
     */
    static final Action ACTION_SKIP = new Action() {

        public boolean invoke(Context ctx) {

            return false;
        }
    };

    /**
     * Copies the current token to the output.
     */
    static final Action ACTION_COPY = new Action() {

        public boolean invoke(Context ctx) {

            ctx.out.append(ctx.in.getToken());

            return false;
        }
    };

    /**
     * Push a "(" on the stack and copy the current token to the output.
     */
    static final Action ACTION_PUSH_LRB = new Action() {

        public boolean invoke(Context ctx) {

            ctx.tokenStack[++ctx.tokenStackTop] = '(';

            ctx.out.append(ctx.in.getToken());

            return false;
        }
    };

    /**
     * Push a "{" on the stack and copy the current token to the output.
     */
    static final Action ACTION_PUSH_LCB = new Action() {

        public boolean invoke(Context ctx) {

            ctx.tokenStack[++ctx.tokenStackTop] = '{';

            ctx.out.append(ctx.in.getToken());

            return false;
        }
    };

    /**
     * Processes a ")".
     */
    static final Action ACTION_POP_RRB = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            if(ctx.tokenStackTop < 0)
                throw new DAOSpecialSyntaxException("Unexpected \")\".");

            switch(ctx.tokenStack[ctx.tokenStackTop--]) {

            case '(': // close ()
                ACTION_COPY.invoke(ctx);
                return false;

            case 'l': // end of a list
                ctx.out = ctx.savedOut;
                ctx.savedOut = null;
                return true;

            default: // nothing to close
                throw new DAOSpecialSyntaxException("Unexpected \")\".");
            }
        }
    };

    /**
     * Processes a "}".
     */
    static final Action ACTION_POP_RCB = new Action() {

        /**
         * Pattern for macro parameter placeholder tags in macro bodies.
         * Group 1 gives the param name.
         */
        private final Pattern PARAM_PATTERN =
            Pattern.compile("\\{\\s*(?:cb2\\s*:\\s*)?mparam\\s+(\\w+)\\s*\\}");

        /**
         * Do the action.
         */
        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            if(ctx.tokenStackTop < 0)
                throw new DAOSpecialSyntaxException("Unexpected \"}\".");

            switch(ctx.tokenStack[ctx.tokenStackTop--]) {

            case '{': // close {}
                ACTION_COPY.invoke(ctx);
                break;

            case 'c': // end of a condition tag body
                ctx.conds.add(new DAOConfig.QueryDescriptor.Condition(ctx.currentConditionName,
                                                                      ctx.currentConditionIndex,
                                                                      ctx.out.substring(ctx.currentConditionIndex).trim()));
                ctx.out.setLength(ctx.currentConditionIndex);
                ctx.currentConditionName = null;
                ctx.currentConditionIndex = -1;
                break;

            case 'm': // end of a macro reference tag
                ctx.in.pushInput(this.getMacroText(ctx));
                ctx.currentMacroDesc = null;
                ctx.currentList.clear();
                break;

            case 'd': // end of a dm macro tag
                Context.DMContext dc = (Context.DMContext)ctx.dmDescStack.removeLast();
                if((dc.nNestedArrays > 0) && !dc.wasNestedArray)
                    throw new DAOSpecialSyntaxException("Nested array must be the last nested DM in the list if nested DMs.");
                if(dc.parentField != null) { // nested dm
                    Context.DMContext parentDc = (Context.DMContext)ctx.dmDescStack.getLast();
                    parentDc.nNestedFields += dc.nNestedFields + this.countFields(dc) + 1;
                    return true;
                }
                // top-level dm
                this.renderDMSelectList(ctx);
                break;

            case '?': // end of a single named query param tag
                ctx.paramPlaceholders.
                    add(new DAOConfig.QueryDescriptor.ParamPlaceholder((ctx.currentQParamName != null ? ctx.currentQParamName : String.valueOf(ctx.paramPlaceholders.size() + 1)),
                                                                       (ctx.currentQParamTypeForBoth ? ctx.currentQParamType : null),
                                                                       ctx.currentQParamType,
                                                                       ctx.currentConditionName,
                                                                       false));
                ctx.out.append('?');
                ctx.currentQParamName = null;
                ctx.currentQParamType = null;
                ctx.currentQParamTypeForBoth = false;
                break;

            case 'a': // end of an array named query param tag
                ctx.paramPlaceholders.
                    add(new DAOConfig.QueryDescriptor.ParamPlaceholder((ctx.currentQParamName != null ? ctx.currentQParamName : String.valueOf(ctx.paramPlaceholders.size() + 1)),
                                                                       (ctx.currentQParamTypeForBoth ? ctx.currentQParamType : null),
                                                                       ctx.currentQParamType,
                                                                       ctx.currentConditionName,
                                                                       true));
                ctx.out.append("??");
                ctx.currentQParamName = null;
                ctx.currentQParamType = null;
                ctx.currentQParamTypeForBoth = false;
                break;

            default: // nothing to close
                throw new DAOSpecialSyntaxException("Unexpected \"}\".");
            }

            return false;
        }

        /**
         * Prepares the macro text checking and replacing the parameters.
         *
         * @param ctx the parser context.
         *
         * @return the prepared macro text ready to be put to the output.
         *
         * @throws DAOSpecialSyntaxException if something is wrong with the
         * parameters (invalid number of params provided in the reference
         * tag).
         */
        private String getMacroText(Context ctx)
            throws DAOSpecialSyntaxException {

            if(ctx.currentMacroDesc.params.length != ctx.currentList.size())
                throw new DAOSpecialSyntaxException("Invalid number of parameters is provided in \"" +
                                                    ctx.currentMacroDesc.name + "\" macro reference tag, expected " +
                                                    ctx.currentMacroDesc.params.length + " params, but received " +
                                                    ctx.currentList.size() + " params.");

            Map paramValues = new HashMap(ctx.currentMacroDesc.params.length);
            for(int i = 0; i < ctx.currentMacroDesc.params.length; i++)
                paramValues.put(ctx.currentMacroDesc.params[i],
                                ((StringBuffer)ctx.currentList.get(i)).toString());

            Matcher m = this.PARAM_PATTERN.matcher(ctx.currentMacroDesc.sql);
            StringBuffer res = new StringBuffer(ctx.currentMacroDesc.sql.length());
            while(m.find()) {
                String paramValue = (String)paramValues.get(m.group(1));
                if(paramValue == null)
                    throw new DAOSpecialSyntaxException("Reference to nonexistent param \"" +
                                                        m.group(1) + "\" in macro \"" +
                                                        ctx.currentMacroDesc.name + "\".");
                m.appendReplacement(res, paramValue.trim());
            }
            m.appendTail(res);

            return res.toString();
        }

        /**
         * Renders a select list for the DMs in the DM list and clears
         * the list.
         *
         * @param ctx the parser context.
         *
         * @throws DAOSpecialSyntaxException if an error is detected.
         */
        private void renderDMSelectList(Context ctx)
            throws DAOSpecialSyntaxException {

            // get the whitespace
            StringBuffer ws = new StringBuffer();
            int p = ctx.out.length() - 1;
            char c;
            while((p >= 0) &&
                  ((c = ctx.out.charAt(p--)) != '\n')) {
                ws.append(Character.isWhitespace(c) ? c : ' ');
            }

            // main loop through the DMs in the list
            Iterator dci = ctx.dmDescList.iterator();
            boolean topLevel = true;
            while(dci.hasNext()) {
                Context.DMContext dc = (Context.DMContext)dci.next();

                int nAdded = 0;

                // render header column for nested DMs
                if(!topLevel) {

                    ctx.out.append(",\n");
                    ctx.out.append(ws);

                    if(dc.parentField.getType().isArray()) { // nested DM array
                        if(dc.byField == null)
                            throw new DAOSpecialSyntaxException("Cannot join nested DM array - no id fields are specified for the parent DM and no \"by\" clause is present.");
                        if(dc.parent.tableName != null) {
                            ctx.out.append(dc.parent.tableName);
                            ctx.out.append(".");
                        }
                        ctx.out.append(dc.byField.getName());
                        ctx.out.append(" AS ");
                        if(dc.parent.prefix != null) {
                            ctx.out.append(dc.parent.prefix);
                            ctx.out.append(ctx.fieldNameSeparator);
                        }
                        ctx.out.append(dc.parentField.getName());

                    } else { // nested single DM
                        ctx.out.append(this.countFields(dc) + dc.nNestedFields); //???
                        ctx.out.append(" AS ");
                        if(dc.parent.prefix != null) {
                            ctx.out.append(dc.parent.prefix);
                            ctx.out.append(ctx.fieldNameSeparator);
                        }
                        ctx.out.append(dc.parentField.getName());
                    }

                    nAdded++;
                }

                // add id fields first
                Iterator i = dc.desc.idFieldNamesSet.iterator();
                while(i.hasNext()) {
                    String fieldName = (String)i.next();

                    if(((dc.onlyFields.size() == 0) && !dc.excludeFields.contains(fieldName)) ||
                       dc.onlyFields.contains(fieldName)) {

                        if(nAdded > 0) {
                            ctx.out.append(",\n");
                            ctx.out.append(ws);
                        }

                        if(dc.tableName != null) {
                            ctx.out.append(dc.tableName);
                            ctx.out.append('.');
                        }
                        ctx.out.append(fieldName);
                        ctx.out.append(" AS ");
                        if(dc.prefix != null) {
                            ctx.out.append(dc.prefix);
                            ctx.out.append(ctx.fieldNameSeparator);
                        }
                        ctx.out.append(fieldName);

                        nAdded++;
                    }
                }

                // add non-id fields
                i = dc.desc.fields.values().iterator();
                while(i.hasNext()) {
                    Field field = (Field)i.next();
                    if(!field.getType().isArray() &&
                       !(DataModel.class).isAssignableFrom(field.getType())) {

                        String fieldName = field.getName();
                        if(!dc.desc.idFieldNamesSet.contains(fieldName) &&
                           (((dc.onlyFields.size() == 0) && !dc.excludeFields.contains(fieldName)) ||
                            dc.onlyFields.contains(fieldName))) {

                            if(nAdded > 0) {
                                ctx.out.append(",\n");
                                ctx.out.append(ws);
                            }

                            if(dc.tableName != null) {
                                ctx.out.append(dc.tableName);
                                ctx.out.append('.');
                            }
                            ctx.out.append(fieldName);
                            ctx.out.append(" AS ");
                            if(dc.prefix != null) {
                                ctx.out.append(dc.prefix);
                                ctx.out.append(ctx.fieldNameSeparator);
                            }
                            ctx.out.append(fieldName);

                            nAdded++;
                        }
                    }
                }

                // if it was top-level, it's not anymore
                topLevel = false;
            }

            // done, purge the list
            ctx.dmDescList.clear();
        }

        /**
         * Counts the number of fields that should be rendered to the
         * output for the specified DM.
         *
         * @param dc the DM context.
         */
        private int countFields(Context.DMContext dc) {

            int res = 0;

            Iterator i = dc.desc.fields.values().iterator();
            while(i.hasNext()) {
                Field field = (Field)i.next();
                String fieldName = field.getName();
                if(!field.getType().isArray() &&
                   !(DataModel.class).isAssignableFrom(field.getType()) &&
                   (((dc.onlyFields.size() == 0) && !dc.excludeFields.contains(fieldName)) ||
                    dc.onlyFields.contains(fieldName))) {
                        res++;
                }
            }

            return res;
        }
    };

    /**
     * Finish parsing (performs some final checks).
     */
    static final Action ACTION_FINISH = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            if(ctx.tokenStackTop >= 0)
                throw new DAOSpecialSyntaxException("Unclosed (), {} or tag.");

            return false;
        }
    };

    /**
     * Got an unexpected token - throws an exception.
     */
    static final Action ACTION_UNEXPECTED_TOKEN = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            throw new DAOSpecialSyntaxException("An unexpected \"" +
                                                ctx.in.getToken() + "\".");
        }
    };

    /**
     * Got an unexpected end of data - throws an exception.
     */
    static final Action ACTION_UNEXPECTED_EOD = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            throw new DAOSpecialSyntaxException("An unexpected end of data.");
        }
    };

    /**
     * Accepts a "(" and starts a list.
     */
    static final Action ACTION_START_LIST = new Action() {

        public boolean invoke(Context ctx) {

            ctx.savedOut = ctx.out;
            ctx.out = new StringBuffer();
            ctx.currentList.add(ctx.out);

            ctx.tokenStack[++ctx.tokenStackTop] = 'l';

            return false;
        }
    };

    /**
     * Processes a comma and start a new element in the current list.
     */
    static final Action ACTION_LIST_NEXT = new Action() {

        public boolean invoke(Context ctx) {

            ctx.out = new StringBuffer();
            ctx.currentList.add(ctx.out);

            return false;
        }
    };

    /**
     * Copies a "," to the output.
     */
    static final Action ACTION_COPY_COMMA = new Action() {

        public boolean invoke(Context ctx) {

            ctx.out.append(',');

            return false;
        }
    };

    /**
     * Processes a standard, not named query parameter placeholder
     * (standalone ? and ??).
     */
    static final Action ACTION_STANDARD_QPARAM = new Action() {

        public boolean invoke(Context ctx) {

            ctx.paramPlaceholders.
                add(new DAOConfig.QueryDescriptor.ParamPlaceholder(String.valueOf(ctx.paramPlaceholders.size() + 1),
                                                                                  null,
                                                                                  null,
                                                                                  ctx.currentConditionName,
                                                                                  ctx.in.getToken().length() > 1));

            ctx.out.append(ctx.in.getToken());

            return false;
        }
    };

    /**
     * Start a condition body context.
     */
    static final Action ACTION_START_CONDITION = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            if(ctx.currentConditionName != null)
                throw new DAOSpecialSyntaxException("Cannot have nested conditions.");

            ctx.currentConditionIndex = ctx.out.length();

            ctx.tokenStack[++ctx.tokenStackTop] = 'c';

            return false;
        }
    };

    /**
     * Set condition name.
     */
    static final Action ACTION_SET_CONDITION_NAME = new Action() {

        public boolean invoke(Context ctx) {

            ctx.currentConditionName = ctx.in.getToken();

            return false;
        }
    };

    /**
     * Start named query param tag.
     */
    static final Action ACTION_START_NAMED_QPARAM = new Action() {

        public boolean invoke(Context ctx) {

            ctx.tokenStack[++ctx.tokenStackTop] =
                (ctx.in.getToken().endsWith("??") ? 'a' : '?');

            return false;
        }
    };

    /**
     * Accepts a '+' and marks the query param type as for both nulls and
     * objects.
     */
    static final Action ACTION_SET_QPARAM_NTYPE = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            if(!ctx.in.getToken().equals("+"))
                throw new DAOSpecialSyntaxException("Unexcpected token \"" +
                                                    ctx.in.getToken() + "\".");

            ctx.currentQParamTypeForBoth = true;

            return false;
        }
    };

    /**
     * Set named query parameter type.
     */
    static final Action ACTION_SET_QPARAM_TYPE = new Action() {

        private final Map TYPES;
        {
            this.TYPES = new HashMap(30);
            this.TYPES.put("ARRAY", new Integer(java.sql.Types.ARRAY));
            this.TYPES.put("BIGINT", new Integer(java.sql.Types.BIGINT));
            this.TYPES.put("BINARY", new Integer(java.sql.Types.BINARY));
            this.TYPES.put("BIT", new Integer(java.sql.Types.BIT));
            this.TYPES.put("BLOB", new Integer(java.sql.Types.BLOB));
            this.TYPES.put("BOOLEAN", new Integer(java.sql.Types.BOOLEAN));
            this.TYPES.put("CHAR", new Integer(java.sql.Types.CHAR));
            this.TYPES.put("CLOB", new Integer(java.sql.Types.CLOB));
            this.TYPES.put("DATALINK", new Integer(java.sql.Types.DATALINK));
            this.TYPES.put("DATE", new Integer(java.sql.Types.DATE));
            this.TYPES.put("DECIMAL", new Integer(java.sql.Types.DECIMAL));
            this.TYPES.put("DISTINCT", new Integer(java.sql.Types.DISTINCT));
            this.TYPES.put("DOUBLE", new Integer(java.sql.Types.DOUBLE));
            this.TYPES.put("FLOAT", new Integer(java.sql.Types.FLOAT));
            this.TYPES.put("INTEGER", new Integer(java.sql.Types.INTEGER));
            this.TYPES.put("JAVA_OBJECT", new Integer(java.sql.Types.JAVA_OBJECT));
            this.TYPES.put("LONGVARBINARY", new Integer(java.sql.Types.LONGVARBINARY));
            this.TYPES.put("LONGVARCHAR", new Integer(java.sql.Types.LONGVARCHAR));
            this.TYPES.put("NULL", new Integer(java.sql.Types.NULL));
            this.TYPES.put("NUMERIC", new Integer(java.sql.Types.NUMERIC));
            this.TYPES.put("OTHER", new Integer(java.sql.Types.OTHER));
            this.TYPES.put("REAL", new Integer(java.sql.Types.REAL));
            this.TYPES.put("REF", new Integer(java.sql.Types.REF));
            this.TYPES.put("SMALLINT", new Integer(java.sql.Types.SMALLINT));
            this.TYPES.put("STRUCT", new Integer(java.sql.Types.STRUCT));
            this.TYPES.put("TIME", new Integer(java.sql.Types.TIME));
            this.TYPES.put("TIMESTAMP", new Integer(java.sql.Types.TIMESTAMP));
            this.TYPES.put("TINYINT", new Integer(java.sql.Types.TINYINT));
            this.TYPES.put("VARBINARY", new Integer(java.sql.Types.VARBINARY));
            this.TYPES.put("VARCHAR", new Integer(java.sql.Types.VARCHAR));
        }

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            if((ctx.currentQParamType = (Integer)this.TYPES.get(ctx.in.getToken().toUpperCase())) == null)
                throw new DAOSpecialSyntaxException("Invalid SQL type name \"" +
                                                    ctx.in.getToken() + "\", see java.sql.Types.");

            return false;
        }
    };

    /**
     * Set named query parameter name.
     */
    static final Action ACTION_SET_QPARAM_NAME = new Action() {

        public boolean invoke(Context ctx) {

            ctx.currentQParamName = ctx.in.getToken();

            return false;
        }
    };

    /**
     * Start macro reference tag.
     */
    static final Action ACTION_START_MACRO = new Action() {

        public boolean invoke(Context ctx) {

            ctx.tokenStack[++ctx.tokenStackTop] = 'm';

            return false;
        }
    };

    /**
     * Set macro name for the current macro reference.
     */
    static final Action ACTION_SET_MACRO_NAME = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            if((ctx.currentMacroDesc = (DAOConfig.SQLMacroDescriptor)ctx.sqlMacros.
                   get(ctx.in.getToken())) == null)
                throw new DAOSpecialSyntaxException("Reference to a nonexistant macro \"" +
                                                    ctx.in.getToken() + "\".");

            return false;
        }
    };

    /**
     * Start dm macro tag.
     */
    static final Action ACTION_START_DM = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            Class dmClass = (Class)ctx.queryDMClasses.get(ctx.queryName);
            if(dmClass == null)
                throw new DAOSpecialSyntaxException("Cannot use DM macro in a query with no usedm attribute.");

            Context.DMContext dc =
                new Context.DMContext(null,
                                      (DAOConfig.DMDescriptor)ctx.dmDescriptors.get(dmClass),
                                      null,
                                      ctx.fieldNameSeparator);
            ctx.dmDescStack.add(dc);
            ctx.dmDescList.add(dc);

            ctx.tokenStack[++ctx.tokenStackTop] = 'd';

            return false;
        }
    };

    /**
     * Process a word or a keyword inside a dm macro tag.
     */
    static final Action ACTION_PROCESS_DM_KW = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            String word = ctx.in.getToken();

            Context.DMContext dc;
            switch(ctx.currentDMKWMode) {

            case Context.DMKW_MODE_FROM:
                dc = (Context.DMContext)ctx.dmDescStack.getLast();
                dc.tableName = word;
                ctx.currentDMKWMode = Context.DMKW_MODE_NOKW;
                break;

            case Context.DMKW_MODE_PREFIX:
                dc = (Context.DMContext)ctx.dmDescStack.getLast();
                dc.prefix = word;
                ctx.currentDMKWMode = Context.DMKW_MODE_NOKW;
                break;

            case Context.DMKW_MODE_BY:
                dc = (Context.DMContext)ctx.dmDescStack.getLast();
                if(dc.parentField == null)
                    throw new DAOSpecialSyntaxException("Clause \"by\" is not allowed in top level DM.");
                if(!dc.parentField.getType().isArray())
                    throw new DAOSpecialSyntaxException("Clause \"by\" is applicable only to array nested DMs.");
                Context.DMContext parDc = (Context.DMContext)ctx.dmDescStack.
                    get(ctx.dmDescStack.size() - 2);
                if((dc.byField = (Field)parDc.desc.fields.get(word.toUpperCase())) == null)
                    throw new DAOSpecialSyntaxException("Field \"" + word +
                                                        "\" specified in the \"by\" clause is not present in the parent DM.");
                ctx.currentDMKWMode = Context.DMKW_MODE_NOKW;
                break;

            default: // Context.DMKW_MODE_NOKW
                if(word.equalsIgnoreCase("from"))
                    ctx.currentDMKWMode = Context.DMKW_MODE_FROM;
                else if(word.equalsIgnoreCase("prefix"))
                    ctx.currentDMKWMode = Context.DMKW_MODE_PREFIX;
                else if(word.equalsIgnoreCase("excluding")) {
                    ctx.currentDMKWMode = Context.DMKW_MODE_EXCLUDING;
                    return true;
                } else if(word.equalsIgnoreCase("only")) {
                    ctx.currentDMKWMode = Context.DMKW_MODE_ONLY;
                    return true;
                } else if(word.equalsIgnoreCase("by"))
                    ctx.currentDMKWMode = Context.DMKW_MODE_BY;
                else
                    throw new DAOSpecialSyntaxException("Unexpected token \"" + word + "\".");
            }

            return false;
        }
    };

    /**
     * Processes an element in the excluded fields list, adds the word
     * to the set of excluded fields in the current DM.
     */
    static final Action ACTION_ADD_FIELD = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            Context.DMContext dc = (Context.DMContext)ctx.dmDescStack.getLast();

            Field f = (Field)dc.desc.fields.get(ctx.in.getToken().toUpperCase());
            if(f == null)
                throw new DAOSpecialSyntaxException("List of fields in the \"" +
                                                    (ctx.currentDMKWMode == Context.DMKW_MODE_EXCLUDING ?
                                                     "excluding" : "only") +
                                                    "\" clause contains \"" +
                                                    ctx.in.getToken() + "\", which is not present in the DM.");

            if(ctx.currentDMKWMode == Context.DMKW_MODE_EXCLUDING)
                dc.excludeFields.add(f.getName());
            else // Context.DMKW_MODE_ONLY
                dc.onlyFields.add(f.getName());

            return false;
        }
    };

    /**
     * Processes closing round bracket after list of fields in excluding
     * or only clause.
     */
    static final Action ACTION_CLOSE_FIELD_LIST = new Action() {

        public boolean invoke(Context ctx) {

            ctx.currentDMKWMode = Context.DMKW_MODE_NOKW;

            return false;
        }
    };

    /**
     * Start a nested dm in a dm macro tag.
     */
    static final Action ACTION_SET_NESTED_FIELD = new Action() {

        public boolean invoke(Context ctx)
            throws DAOSpecialSyntaxException {

            Context.DMContext parentDc = (Context.DMContext)ctx.dmDescStack.getLast();

            Field parentField = (Field)parentDc.desc.fields.
                get(ctx.in.getToken().toUpperCase());
            if(parentField == null)
                throw new DAOSpecialSyntaxException("DM field \"" + ctx.in.getToken() +
                                                    "\", specified in a nested DM tag is not present in the parent DM.");
            if(!(DataModel.class).isAssignableFrom(parentField.getType()) &&
               !(parentField.getType().isArray() && (DataModel.class).isAssignableFrom(parentField.getType().getComponentType())))
                throw new DAOSpecialSyntaxException("DM field \"" + ctx.in.getToken() +
                                                    "\" in the parent DM is neither a DataModel nor an array of DataModels.");

            if(parentField.getType().isArray()) {
                if(++parentDc.nNestedArrays > 1)
                    throw new DAOSpecialSyntaxException("Cannot have more than one nested DM array in a single parent DM.");
                parentDc.wasNestedArray = true;
            } else {
                parentDc.wasNestedArray = false;
            }

            DAOConfig.DMDescriptor dmDesc = (DAOConfig.DMDescriptor)ctx.dmDescriptors.
                get(parentField.getType().isArray() ?
                    parentField.getType().getComponentType() :
                    parentField.getType());
            if(dmDesc == null)
                throw new DAOSpecialSyntaxException("No DM descriptor is defined for \"" +
                                                    parentField.getName() + "\" field in the parent DM.");

            Context.DMContext dc = new Context.DMContext(parentDc,
                                                         dmDesc,
                                                         parentField,
                                                         ctx.fieldNameSeparator);
            ctx.dmDescStack.add(dc);
            ctx.dmDescList.add(dc);

            ctx.tokenStack[++ctx.tokenStackTop] = 'd';

            return false;
        }
    };


    ///////////////////////////////////////////////////////////////////////////
    // The State Matrix
    ///////////////////////////////////////////////////////////////////////////

    /**
     * The state transition matrix for the parser.
     */
    private static final StateTrans[][] STATES = {
        {   // 0 - normal input outside any special tags
            new StateTrans(ACTION_FINISH, -1),             // EOD
            new StateTrans(ACTION_COPY, 0),                // SINGLE_CHAR
            new StateTrans(ACTION_COPY, 0),                // SPACE
            new StateTrans(ACTION_COPY, 0),                // WORD
            new StateTrans(ACTION_COPY, 0),                // LITERAL
            new StateTrans(ACTION_COPY, 0),                // COMMENT
            new StateTrans(ACTION_COPY, 0),                // ESCAPED_COMMA
            new StateTrans(ACTION_STANDARD_QPARAM, 0),     // DOUBLE_QMARK
            new StateTrans(ACTION_START_MACRO, 7),         // TAG_MACRO
            new StateTrans(ACTION_START_DM, 11),           // TAG_DM
            new StateTrans(ACTION_START_CONDITION, 1),     // TAG_COND
            new StateTrans(ACTION_START_NAMED_QPARAM, 4),  // TAG_QPARAM
            new StateTrans(ACTION_PUSH_LRB, 0),            // LRB
            new StateTrans(ACTION_POP_RRB, 0),             // RRB
            new StateTrans(ACTION_PUSH_LCB, 0),            // LCB
            new StateTrans(ACTION_POP_RCB, 0),             // RCB
            new StateTrans(ACTION_STANDARD_QPARAM, 0),     // QMARK
            new StateTrans(ACTION_COPY, 0)                 // COMMA
        },
        {   // 1 - got a condition tag opening, now need a "("
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 1),                // SPACE
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 1),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_SKIP, 2),                // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 2 - got the "(", now need the condition name
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 2),                // SPACE
            new StateTrans(ACTION_SET_CONDITION_NAME, 3),  // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 2),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 3 - got condition name, now need ')'
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 3),                // SPACE
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 3),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_SKIP, 0),                // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 4 - got a named query param tag opening
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 4),                // SPACE
            new StateTrans(ACTION_SET_QPARAM_NAME, 5),     // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 4),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_POP_RCB, 0),             // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 5 - got a named query param name, can take type or '}'
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_SET_QPARAM_NTYPE, 5),    // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 5),                // SPACE
            new StateTrans(ACTION_SET_QPARAM_TYPE, 6),     // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 5),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_POP_RCB, 0),             // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 6 - got a named query param name and type, now need a '}'
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 6),                // SPACE
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 6),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_POP_RCB, 0),             // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 7 - got a macro reference tag opening, need macro name now
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 7),                // SPACE
            new StateTrans(ACTION_SET_MACRO_NAME, 8),      // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 7),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 8 - got a macro name, can start a list or close the tag now
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 8),                // SPACE
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 8),                // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_START_LIST, 9),          // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_POP_RCB, 0),             // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 9 - collect list
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_COPY, 9),                // SINGLE_CHAR
            new StateTrans(ACTION_COPY, 9),                // SPACE
            new StateTrans(ACTION_COPY, 9),                // WORD
            new StateTrans(ACTION_COPY, 9),                // LITERAL
            new StateTrans(ACTION_COPY, 9),                // COMMENT
            new StateTrans(ACTION_COPY_COMMA, 9),          // ESCAPED_COMMA
            new StateTrans(ACTION_COPY, 9),                // DOUBLE_QMARK
            new StateTrans(ACTION_PUSH_LCB, 9),            // TAG_MACRO
            new StateTrans(ACTION_PUSH_LCB, 9),            // TAG_DM
            new StateTrans(ACTION_PUSH_LCB, 9),            // TAG_COND
            new StateTrans(ACTION_PUSH_LCB, 9),            // TAG_QPARAM
            new StateTrans(ACTION_PUSH_LRB, 9),            // LRB
            new StateTrans(ACTION_POP_RRB, 9, 10),         // RRB
            new StateTrans(ACTION_PUSH_LCB, 9),            // LCB
            new StateTrans(ACTION_POP_RCB, 9),             // RCB
            new StateTrans(ACTION_COPY, 9),                // QMARK
            new StateTrans(ACTION_LIST_NEXT, 9)            // COMMA
        },
        {   // 10 - got the params list, now need '}'
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 10),               // SPACE
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // WORD
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // LITERAL
            new StateTrans(ACTION_SKIP, 10),               // COMMENT
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // LRB
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // RRB
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // LCB
            new StateTrans(ACTION_POP_RCB, 0),             // RCB
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // QMARK
            new StateTrans(ACTION_UNEXPECTED_EOD, -1)      // COMMA
        },
        {   // 11 - got a dm macro tag opening or nested dm opening with field
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 11),               // SPACE
            new StateTrans(ACTION_PROCESS_DM_KW, 11, 12),  // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 11),               // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_SKIP, 14),               // LCB
            new StateTrans(ACTION_POP_RCB, 0, 11),         // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 12 - got "excluding" keyword, need '(' now
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 12),               // SPACE
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 12),               // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_SKIP, 13),               // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        },
        {   // 13 - got '(', collecting the fields list now
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 13),               // SPACE
            new StateTrans(ACTION_ADD_FIELD, 13),          // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 13),               // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_CLOSE_FIELD_LIST, 11),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_SKIP, 13)                // COMMA
        },
        {   // 14 - got nested DM opening '{', now need the field name
            new StateTrans(ACTION_UNEXPECTED_EOD, -1),     // EOD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // SINGLE_CHAR
            new StateTrans(ACTION_SKIP, 14),               // SPACE
            new StateTrans(ACTION_SET_NESTED_FIELD, 11),   // WORD
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LITERAL
            new StateTrans(ACTION_SKIP, 14),               // COMMENT
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // ESCAPED_COMMA
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // DOUBLE_QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_MACRO
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_DM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_COND
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // TAG_QPARAM
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RRB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // LCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // RCB
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1),   // QMARK
            new StateTrans(ACTION_UNEXPECTED_TOKEN, -1)    // COMMA
        }
    };


    ///////////////////////////////////////////////////////////////////////////
    // The Main Interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Parse sql chunk and create a descriptor.
     *
     * @param queryDescriptor descriptor of the query this chunk
     * belongs to.
     * @param chunkName name of the chunk.
     * @param isResult tells if this is the resulting chunk.
     * @param chunkText source text of the chunk from the configuration
     * XML element.
     * @param depends space separated list of chunks on which this one
     * depends, may be <code>null</code>.
     * @param sqlMacros defined macros.
     * @param dms DM names to DM classes.
     * @param dmDescriptors DM classes to DM decriptors.
     * @param queryDMClasses query names to used DM classes.
     * @param fieldNameSeparator the field name separator string.
     * @param paramPlaceholders output parameter for query parameter
     * placeholder descriptors.
     *
     * @return descriptor for the sql chunk.
     *
     * @throws BLException if an error happens and the sql chunk cannot
     * be parsed/used.
     */
    public static DAOConfig.QueryDescriptor.SqlChunk parseSqlChunk(DAOConfig.QueryDescriptor queryDescriptor,
                                                                   String chunkName,
                                                                   boolean isResult,
                                                                   String chunkText,
                                                                   String depends,
                                                                   Map sqlMacros,
                                                                   Map dms,
                                                                   Map dmDescriptors,
                                                                   Map queryDMClasses,
                                                                   String fieldNameSeparator,
                                                                   List paramPlaceholders)
        throws BLException {

        // create parser context
        Context ctx = new Context(new Input(chunkText),
                                  paramPlaceholders,
                                  queryDescriptor.name,
                                  sqlMacros,
                                  dmDescriptors,
                                  queryDMClasses,
                                  fieldNameSeparator);

        // state machine main loop
        try {
            int state = 0;

            while(state >= 0) {
                StateTrans t = STATES[state][ctx.in.nextToken()];
                state = (t.action.invoke(ctx) ? t.altNextState : t.nextState);
            }

        } catch(DAOSpecialSyntaxException e) {
            throw new BLException("Could not parse special DAO syntax in SQL text of \"" +
                                  queryDescriptor.name + "\" query, chunk \"" +
                                  chunkName + "\".", e);
        }

        // done, create descriptor and return it
        return new DAOConfig.QueryDescriptor.SqlChunk(queryDescriptor,
                                                      chunkName,
                                                      depends,
                                                      isResult,
                                                      ctx.out.toString(),
                                                      (DAOConfig.QueryDescriptor.Condition [])ctx.conds.
                                                          toArray(new DAOConfig.QueryDescriptor.Condition[ctx.conds.size()]),
                                                      ctx.firstParamPlaceholderIndex,
                                                      ctx.paramPlaceholders.size() - 1);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Test
    ///////////////////////////////////////////////////////////////////////////

//    private static class TestDM
//        extends DataModel {
//
//        public int testId;
//        public String name;
//        public Integer mainAddressId;
//
//        public TestAddressDM homeAddress;
//        public TestAddressDM workAddress;
//
//        public TestAddressDM [] allAddresses;
//
//        public TestPhoneDM [] allPhones;
//    }
//
//    private static class TestAddressDM
//        extends DataModel {
//
//        public int addressId;
//        public String address;
//        public Integer phoneId;
//
//        public TestPhoneDM phone;
//    }
//
//    private static class TestPhoneDM
//        extends DataModel {
//
//        public int phoneId;
//        public String phone;
//    }
//
//    /**
//     * Test entry point.
//     */
//    public static void main(String [] args)
//        throws Exception {
//
//        java.io.StringWriter chunkTextSW = new java.io.StringWriter();
//        int c;
//        while((c = System.in.read()) > 0)
//            chunkTextSW.write(c);
//
//        System.out.println("Got input, parsing it...");
//
//        List paramPlaceholders = new ArrayList();
//        Map sqlMacros = new HashMap();
//        sqlMacros.put("mamber", new DAOConfig.SQLMacroDescriptor("mamber",
//                                                                 new String[] {"p1", "p2", "p3", "p4", "p5"},
//                                                                 "[ZZZ p1=[{mparam p1}] p2=[{mparam p2}] p3=[{mparam p3}] p4=[{mparam p4}] p5=[{mparam p5}] - {macro num}]"));
//        sqlMacros.put("num", new DAOConfig.SQLMacroDescriptor("num",
//                                                                 new String[] {},
//                                                                 "GONZO"));
//        Map dmDescriptors = new HashMap();
//        dmDescriptors.put(TestDM.class,
//                          new DAOConfig.DMDescriptor("test",
//                                                     "tb_Test",
//                                                     new DAOConfig.DMDescriptor.IDFieldDescriptor [] {
//                                                        new DAOConfig.DMDescriptor.IDFieldDescriptor("testId", null, "post")
//                                                     },
//                                                     new HashMap(),
//                                                     TestDM.class));
//        dmDescriptors.put(TestAddressDM.class,
//                          new DAOConfig.DMDescriptor("testAddress",
//                                                     "tb_TestAddress",
//                                                     new DAOConfig.DMDescriptor.IDFieldDescriptor [] {
//                                                        new DAOConfig.DMDescriptor.IDFieldDescriptor("addressId", null, "post")
//                                                     },
//                                                     new HashMap(),
//                                                     TestAddressDM.class));
//        dmDescriptors.put(TestPhoneDM.class,
//                          new DAOConfig.DMDescriptor("testPhone",
//                                                     "tb_TestPhone",
//                                                     new DAOConfig.DMDescriptor.IDFieldDescriptor [] {
//                                                        new DAOConfig.DMDescriptor.IDFieldDescriptor("phoneId", null, "post")
//                                                     },
//                                                     new HashMap(),
//                                                     TestPhoneDM.class));
//        Map queryDMClasses = new HashMap();
//        queryDMClasses.put("testQuery", TestDM.class);
//        Context ctx = new Context(new Input(chunkTextSW.toString()),
//                                  paramPlaceholders,
//                                  "testQuery",
//                                  sqlMacros,
//                                  dmDescriptors,
//                                  queryDMClasses,
//                                  "$");
//
//        int state = 0;
//        while(state >= 0) {
//            int token = ctx.in.nextToken();
//            StateTrans t = STATES[state][token];
//            //System.out.print("#" + state + " [" + ctx.in.getToken() + "]=" + token);
//            state = (t.action.invoke(ctx) ? t.altNextState : t.nextState);
//            //System.out.println(" -> #" + state);
//        }
//
//        System.out.println("ORIGINAL SQL:");
//        System.out.println("===========================================");
//        System.out.println(chunkTextSW.toString());
//        System.out.println("===========================================");
//        System.out.println("\nTRANSFORMED SQL:");
//        System.out.println("===========================================");
//        System.out.println(ctx.out.toString());
//        System.out.println("===========================================");
//        System.out.println("\nCONDITIONS:");
//        java.util.Iterator i = ctx.conds.iterator();
//        while(i.hasNext()) {
//            System.out.println("-> " + i.next());
//        }
//        System.out.println("\nPARAMETERS:");
//        i = ctx.paramPlaceholders.iterator();
//        while(i.hasNext()) {
//            System.out.println("-> " + i.next());
//        }
//    }
}
