package com.boylesoftware.cb2;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.lang.reflect.Field;
import org.w3c.dom.Element;

import com.boylesoftware.cb2.util.XMLUtils;


/**
 * This helper class encapsulates the functionality for parsing DAO
 * special syntax, which may appear in the bodies of SQL queries defined
 * in the DAO configuration. This class is used by the
 * <code>DAOConfig</code> class.
 *
 * <p>This is the implementation of the old, obsolete DAO syntax.
 *
 * @author Lev Himmelfarb
 * @version $Id: DAOSpecialSyntaxParserOld.java,v 1.2 2004/10/11 18:17:36 levahim Exp $
 */
final class DAOSpecialSyntaxParserOld {

    /**
     * Pattern for refrences to custom macros in query texts.
     */
    private static final Pattern CUSTOM_MACRO_REFERENCE_PATTERN =
        Pattern.compile("\\$\\{(?!dm\\}|dm:|nesteddm:|\\?\\??\\}|\\?\\??:)((\\\\\\{|\\\\\\}|[^{}])+)\\}");

    /**
     * Index of the group within
     * <code>CUSTOM_MACRO_REFERENCE_PATTERN</code> containing the
     * macro reference content.
     */
    private static final int CUSTOM_MACRO_REFERENCE_GROUP = 1;

    /**
     * Pattern used for splitting comma separated lists in ${...}
     * constructs.
     */
    private final static Pattern COMMASEP_PATTERN = Pattern.compile("(?<!\\\\),");

    /**
     * Pattern for replacing escaped with a backslash characters.
     */
    private final static Pattern ESCAPEREP_PATTERN = Pattern.compile("\\\\(.)");

    /**
     * Pattern used for finding parameter placeholders in sqlmacro
     * bodies.
     */
    private final static Pattern PARAM_PATTERN = Pattern.compile("\\$\\{param:([^{}]+)\\}");

    /**
     * Used in sqlmacro replacement mechanism.
     */
    private final static Pattern DOLLARSIGN_PATTERN = Pattern.compile("\\$");

    /**
     * Pattern for standard special tags in query texts.
     */
    private static final Pattern SPECIAL_TAG_PATTERN =
        Pattern.compile("\\$\\[condition:|(?<!\\\\)\\]|(\\$\\{)?((?<=\\$\\{)(dm|\\?\\??|(dm|nesteddm|\\?\\??):[^{}]+)(?=\\})|(?<!\\$\\{)\\?\\??(?!\\}))\\}?");

    /**
     * Index of the group within <code>SPECIAL_TAG_PATTERN</code>
     * conatining the tag's content.
     */
    private static final int SPECIAL_TAG_GROUP = 2;

    /**
     * Pattern for condition structures in query texts.
     */
    private static final Pattern CONDITION_PATTERN =
        Pattern.compile("\\$\\[condition:(.+?)(?<!\\\\)\\]", Pattern.DOTALL);

    /**
     * Index of the group within <code>CONDITION_PATTERN</code>
     * conatining the condition's content.
     */
    private static final int CONDITION_GROUP = 1;


    /**
     * Parse sql XML element and create a descriptor from it. This
     * method parses old DAO syntax (<code>${...}</code>).
     *
     * @param sqlElement XML element describing the chunk.
     * @param chunkNumber number of the chunk among all chunks in
     * defined within the query, starting from 1.
     * @param isLastChunk <code>true</code> if it is the last chunk
     * within the query.
     * @param assignedPrimary <code>true</code> if the primary result
     * chunk has already been defined for the query.
     * @param queryDescriptor descriptor of the query this chunk
     * belongs to.
     * @param sqlmacros defined macros.
     * @param dms DM names to DM classes.
     * @param dmDescriptors DM classes to DM decriptors.
     * @param queryDMClasses query names to used DM classes.
     * @param fieldNameSeparator the field name separator string.
     * @param paramPlaceholders output parameter for query parameter
     * placeholder descriptors.
     *
     * @return descriptor for the sql chunk.
     *
     * @throws BLException if an error happens and the sql chunk cannot
     * be parsed/used.
     */
    public static DAOConfig.QueryDescriptor.SqlChunk parseSqlChunk(Element sqlElement,
                                                                   int chunkNumber,
                                                                   boolean isLastChunk,
                                                                   boolean assignedPrimary,
                                                                   DAOConfig.QueryDescriptor queryDescriptor,
                                                                   Map sqlmacros,
                                                                   Map dms,
                                                                   Map dmDescriptors,
                                                                   Map queryDMClasses,
                                                                   String fieldNameSeparator,
                                                                   List paramPlaceholders)
        throws BLException {

        // get the chunk body (it is never null, 'cause we've checked it before the call
        String sqlSrc = XMLUtils.getXMLNodeText(sqlElement);

        // frist pass: expand references to custom macroses
        sqlSrc = processCustomMacros(sqlSrc,
                                     sqlmacros,
                                     queryDescriptor.name);

        // second pass: expand standard tags
        int firstParamPlaceholderIndex = paramPlaceholders.size();
        sqlSrc = processSpecialTags(sqlSrc,
                                    paramPlaceholders,
                                    dms,
                                    dmDescriptors,
                                    queryDMClasses,
                                    fieldNameSeparator,
                                    queryDescriptor.name);
        int lastParamPlaceholderIndex = paramPlaceholders.size() - 1;

        // final pass: process conditions
        List conditions = new ArrayList();
        sqlSrc = processConditions(sqlSrc,
                                   conditions,
                                   queryDescriptor.name);

        // create the chunk descriptor and return it
        return new DAOConfig.QueryDescriptor.SqlChunk(queryDescriptor,
                                                      (sqlElement.hasAttribute("name") ? sqlElement.getAttribute("name") : String.valueOf(chunkNumber)),
                                                      (sqlElement.hasAttribute("depends") ? sqlElement.getAttribute("depends") : null),
                                                      "true".equals(sqlElement.getAttribute("result")) || (!assignedPrimary && isLastChunk),
                                                      sqlSrc,
                                                      (DAOConfig.QueryDescriptor.Condition [])conditions.toArray(new DAOConfig.QueryDescriptor.Condition[conditions.size()]),
                                                      firstParamPlaceholderIndex,
                                                      lastParamPlaceholderIndex);
    }


    /**
     * Processes the source SQL text and expands all references to
     * custom macros (defined with <code>sqlmacro</code>
     * elements).
     *
     * @param sqlSrc the SQL text with references to custom
     * macros.
     * @param sqlmacros collection of custom SQL macros
     * descriptors. The keys are macro names (objects of
     * <code>String</code>), the values are macro descriptors
     * (objects of <code>SQLMacroDescriptor</code>).
     * @param queryName name of the query <code>sqlSrc</code>
     * belongs to.
     *
     * @return processed SQL text with expanded references to
     * custom macros.
     *
     * @throws BLException if error happened expanding macro
     * references.
     */
    private static String processCustomMacros(String sqlSrc,
                                              Map sqlmacros,
                                              String queryName)
        throws BLException {

        StringBuffer sqlRes = new StringBuffer(sqlSrc.length());

        Matcher matcher = CUSTOM_MACRO_REFERENCE_PATTERN.matcher(sqlSrc);
        while(matcher.find()) {
            String macro = matcher.group(CUSTOM_MACRO_REFERENCE_GROUP);

            int colInd = macro.indexOf(':');
            String macroName = (colInd < 0 ? macro : macro.substring(0, colInd));
            DAOConfig.SQLMacroDescriptor desc = (DAOConfig.SQLMacroDescriptor)sqlmacros.get(macroName);
            if(desc == null)
                throw new BLException("Reference to nonexistent macro \"" + macroName +
                                      "\" in query \"" + queryName + "\".");
            String macroValue;
            if((colInd < 0) || (colInd == macro.length() - 1)) {
                if(desc.params.length > 0)
                    throw new BLException("Invalid number of parameters in macro \"" + macroName +
                                          "\" call in query \"" + queryName + "\".");
                macroValue = desc.sql;

            } else {
                String [] paramValues = COMMASEP_PATTERN.split(macro.substring(colInd + 1));
                if(paramValues.length != desc.params.length)
                    throw new BLException("Invalid number of parameters in macro \"" + macroName +
                                          "\" call in query \"" + queryName + "\".");
                Map paramValuesMap = new HashMap(desc.params.length);
                for(int j = 0; j < desc.params.length; j++) {
                    Matcher m = ESCAPEREP_PATTERN.matcher(paramValues[j].trim());
                    paramValuesMap.put(desc.params[j], m.replaceAll("$1"));
                }
                StringBuffer res = new StringBuffer();
                Matcher m = PARAM_PATTERN.matcher(desc.sql);
                while(m.find()) {
                    String paramValue = (String)paramValuesMap.get(m.group(1));
                    if(paramValue == null)
                        throw new BLException("Reference to nonexistent param \"" + m.group(1) +
                                              "\" in macro \"" + macroName + "\".");
                    m.appendReplacement(res, paramValue);
                }
                m.appendTail(res);
                macroValue = res.toString();
            }

            matcher.appendReplacement(sqlRes, DOLLARSIGN_PATTERN.matcher(macroValue).replaceAll("\\\\\\$"));
        }
        matcher.appendTail(sqlRes);

        return sqlRes.toString();
    }

    /**
     * Processes the source SQL text and expands all standard
     * special tags.
     *
     * @param sqlSrc the SQL text with special tags.
     * @param queryName name of the query <code>sqlSrc</code>
     * belongs to.
     *
     * @return processed SQL text with expanded special tags.
     *
     * @throws BLException if error happened expanding the tags.
     */
    private static String processSpecialTags(String sqlSrc,
                                             List paramPlaceholdersList,
                                             Map dms,
                                             Map dmDescriptors,
                                             Map queryDMClasses,
                                             String fieldNameSeparator,
                                             String queryName)
        throws BLException {

        StringBuffer sqlRes = new StringBuffer(sqlSrc.length());

        DAOConfig.DMDescriptor currentDMDesc = null;
        if(queryDMClasses.containsKey(queryName))
            currentDMDesc = (DAOConfig.DMDescriptor)dmDescriptors.
                get(queryDMClasses.get(queryName));
        String currentConditionName = null;
        Matcher matcher = SPECIAL_TAG_PATTERN.matcher(sqlSrc);
        while(matcher.find()) {

            if(matcher.group().equals("$[condition:")) {
                int colInd = sqlSrc.indexOf(':', matcher.start() + 12);
                if((colInd < 0) || (colInd >= sqlSrc.length() - 2))
                    throw new BLException("Invalid syntax of a condition structure in query \"" + queryName + "\".");
                currentConditionName = sqlSrc.substring(matcher.start() + 12, colInd).trim();
                matcher.appendReplacement(sqlRes, "\\$[condition:");
                continue;
            } else if(matcher.group().equals("]")) {
                currentConditionName = null;
                matcher.appendReplacement(sqlRes, matcher.group());
                continue;
            }

            String macro = matcher.group(SPECIAL_TAG_GROUP);

            if(macro.equals("dm")) { // default dm macro
                if(currentDMDesc == null)
                    throw new BLException("Cannot reference default DM when \"usedm\" attribute" +
                                          " is not defined in query \"" + queryName + "\".");
                StringBuffer expanded = new StringBuffer();
                expandDMFieldsMacro(expanded, currentDMDesc.name, null, dms, dmDescriptors, fieldNameSeparator);
                matcher.appendReplacement(sqlRes, DOLLARSIGN_PATTERN.matcher(expanded.toString()).replaceAll("\\\\\\$"));

            } else if(macro.startsWith("dm:")) { // dm macro
                StringBuffer expanded = new StringBuffer();
                expandDMFieldsMacro(expanded, macro.substring(3), null, dms, dmDescriptors, fieldNameSeparator);
                matcher.appendReplacement(sqlRes, DOLLARSIGN_PATTERN.matcher(expanded.toString()).replaceAll("\\\\\\$"));

            } else if(macro.startsWith("nesteddm:")) { // nesteddm macro
                if(currentDMDesc == null)
                    throw new BLException("Cannot use nesteddm macro when \"usedm\" attribute" +
                                          " is not defined in query \"" + queryName + "\".");
                int splitIndex = macro.indexOf(',', 9);
                String joinToFieldName = (splitIndex >= 0 ?
                                          macro.substring(9, splitIndex).trim() :
                                          macro.substring(9).trim());
                String joinToFieldNameFull = joinToFieldName;
                String joinToFieldPrefix = null;
                int sepIndex = joinToFieldName.lastIndexOf(fieldNameSeparator);
                if(sepIndex >= 0) {
                    joinToFieldPrefix = joinToFieldName.substring(0, sepIndex + 1);
                    joinToFieldName = joinToFieldName.substring(sepIndex + 1);
                }
                Field joinToField = (Field)currentDMDesc.fields.get(joinToFieldName.toUpperCase());
                if(joinToField == null)
                    throw new BLException("Reference to nonexistent DM field \"" + joinToFieldName +
                                          "\" in query \"" + queryName + "\".");

                if(joinToField.getType().isArray()) {
                    String nestedDMName;
                    DAOConfig.DMDescriptor nestedDMDesc;
                    if(splitIndex >= 0) {
                        nestedDMName = macro.substring(splitIndex + 1).trim();
                        sepIndex = nestedDMName.lastIndexOf(fieldNameSeparator);
                        Class nestedDMClass = (Class)dms.get(sepIndex < 0 ?
                                                             nestedDMName :
                                                             nestedDMName.substring(sepIndex + 1));
                        if(nestedDMClass == null)
                            throw new BLException("Reference to nonexistent DM \"" + nestedDMName +
                                                  "\" in query \"" + queryName + "\".");
                        nestedDMDesc = (DAOConfig.DMDescriptor)dmDescriptors.get(nestedDMClass);
                    } else {
                        Class nestedDMClass = joinToField.getType().getComponentType();
                        nestedDMDesc = (DAOConfig.DMDescriptor)dmDescriptors.get(nestedDMClass);
                        if(nestedDMDesc == null)
                            throw new BLException("There is no DM descriptor for DM class \"" + nestedDMClass.getName() +
                                                  "\" referenced from nesteddm macro in query \"" + queryName + "\".");
                        nestedDMName = nestedDMDesc.name;
                    }
                    StringBuffer expanded = new StringBuffer();
                    int nFields = expandDMFieldsMacro(expanded,
                                                      nestedDMName,
                                                      joinToFieldNameFull + fieldNameSeparator,
                                                      dms,
                                                      dmDescriptors,
                                                      fieldNameSeparator);
                    StringBuffer headerAndExpanded = new StringBuffer();
                    if(currentDMDesc.idFields.length == 0)
                        throw new BLException("DM \"" + currentDMDesc.name +
                                              "\" does not have id fields required for joining nested arrays.");
                    if((currentDMDesc.tableName == null) || (currentDMDesc.tableName.length() == 0))
                        throw new BLException("DM \"" + currentDMDesc.name +
                                              "\" does not have \"table\" attribute.");
                    headerAndExpanded.append(currentDMDesc.tableName);
                    headerAndExpanded.append('.');
                    headerAndExpanded.append(currentDMDesc.idFields[0].name);
                    headerAndExpanded.append(" AS ");
                    if(joinToFieldPrefix != null)
                        headerAndExpanded.append(joinToFieldPrefix);
                    headerAndExpanded.append(joinToFieldName);
                    if(nFields > 0) {
                        headerAndExpanded.append(",\n");
                        headerAndExpanded.append(expanded);
                    }

                    matcher.appendReplacement(sqlRes, DOLLARSIGN_PATTERN.matcher(headerAndExpanded.toString()).replaceAll("\\\\\\$"));

                    currentDMDesc = nestedDMDesc;

                } else if((DataModel.class).isAssignableFrom(joinToField.getType())) {
                    String nestedDMName;
                    if(splitIndex >= 0) {
                        nestedDMName = macro.substring(splitIndex + 1).trim();
                    } else {
                        Class nestedDMClass = joinToField.getType();
                        DAOConfig.DMDescriptor nestedDMDesc = (DAOConfig.DMDescriptor)dmDescriptors.get(nestedDMClass);
                        if(nestedDMDesc == null)
                            throw new BLException("There is no DM descriptor for DM class \"" + nestedDMClass.getName() +
                                                  "\" referenced from nesteddm macro in query \"" + queryName + "\".");
                        nestedDMName = nestedDMDesc.name;
                    }
                    StringBuffer expanded = new StringBuffer();
                    int nFields = expandDMFieldsMacro(expanded,
                                                      nestedDMName,
                                                      joinToFieldNameFull + fieldNameSeparator,
                                                      dms,
                                                      dmDescriptors,
                                                      fieldNameSeparator);
                    StringBuffer headerAndExpanded = new StringBuffer();
                    headerAndExpanded.append(nFields);
                    headerAndExpanded.append(" AS ");
                    if(joinToFieldPrefix != null)
                        headerAndExpanded.append(joinToFieldPrefix);
                    headerAndExpanded.append(joinToFieldName);
                    if(nFields > 0) {
                        headerAndExpanded.append(",\n");
                        headerAndExpanded.append(expanded);
                    }

                    matcher.appendReplacement(sqlRes, DOLLARSIGN_PATTERN.matcher(headerAndExpanded.toString()).replaceAll("\\\\\\$"));

                } else
                    throw new BLException("Cannot join nested DM to regular field \"" +
                                          joinToFieldName + "\".");

            } else if(macro.equals("?")) { // unnamed single query parameter placeholder

                matcher.appendReplacement(sqlRes, "?");

                paramPlaceholdersList.add(new DAOConfig.QueryDescriptor.ParamPlaceholder(String.valueOf(paramPlaceholdersList.size() + 1),
                                                                                         null,
                                                                                         null,
                                                                                         currentConditionName,
                                                                                         false));

            } else if(macro.equals("??")) { // unnamed array query parameter placeholder

                matcher.appendReplacement(sqlRes, "??");

                paramPlaceholdersList.add(new DAOConfig.QueryDescriptor.ParamPlaceholder(String.valueOf(paramPlaceholdersList.size() + 1),
                                                                                         null,
                                                                                         null,
                                                                                         currentConditionName,
                                                                                         true));

            } else if(macro.startsWith("?:")) { // named single query parameter placeholder

                if(macro.length() == 2)
                    throw new BLException("Invalid syntax of a parameter placeholder tag in query \"" + queryName + "\".");

                matcher.appendReplacement(sqlRes, "?");

                paramPlaceholdersList.add(new DAOConfig.QueryDescriptor.ParamPlaceholder(macro.substring(2),
                                                                                         null,
                                                                                         null,
                                                                                         currentConditionName,
                                                                                         false));

            } else if(macro.startsWith("??:")) { // named array query parameter placeholder

                if(macro.length() == 3)
                    throw new BLException("Invalid syntax of a parameter placeholder tag in query \"" + queryName + "\".");

                matcher.appendReplacement(sqlRes, "??");

                paramPlaceholdersList.add(new DAOConfig.QueryDescriptor.ParamPlaceholder(macro.substring(3),
                                                                                         null,
                                                                                         null,
                                                                                         currentConditionName,
                                                                                         true));

            } else { // should actually never happen
                throw new RuntimeException("Could not recognize standard special tag in query \"" + queryName +
                                           "\", the matched pattern was [" + matcher.group() + "].");
            }
        }
        matcher.appendTail(sqlRes);

        return sqlRes.toString();
    }

    /**
     * Processes condition structures
     * (<code>$[condition:name:...]</code>) in the source SQL
     * text.
     *
     * @param sqlSrc the SQL text with conditions.
     * @param queryName name of the query <code>sqlSrc</code>
     * belongs to.
     *
     * @return processed SQL text with conditions stripped.
     *
     * @throws BLException if error happened expanding the tags.
     */
    private static String processConditions(String sqlSrc,
                                            List conditionsList,
                                            String queryName)
        throws BLException {

        StringBuffer sqlRes = new StringBuffer(sqlSrc.length());

        Matcher matcher = CONDITION_PATTERN.matcher(sqlSrc);
        while(matcher.find()) {
            String macro = matcher.group(CONDITION_GROUP);

            int colInd = macro.indexOf(':');
            if((colInd < 0) || (colInd >= (macro.length() - 1)))
                throw new BLException("Invalid syntax of a condition tag in query \"" + queryName + "\".");

            matcher.appendReplacement(sqlRes, "");

            conditionsList.add(new DAOConfig.QueryDescriptor.Condition(macro.substring(0, colInd).trim(),
                                                                       sqlRes.length(),
                                                                       ESCAPEREP_PATTERN.matcher(macro.substring(colInd + 1).trim()).replaceAll("$1")));
        }
        matcher.appendTail(sqlRes);

        return sqlRes.toString();
    }

    /**
     * Expands the named DM to a list if its fields suitable for a
     * SELECT query.
     *
     * @param expanded output string buffer that receives the expanded
     * list of fields.
     * @param dmName logical name of the DM.
     * @param prefix optional prefix for returned column names. May be
     * <code>null</code>.
     * @param dms map of DM names and their implementation classes.
     *
     * @return number of fields added the result.
     *
     * @throws BLException if there is an error and the list cannot be
     * generated.
     */
    private static int expandDMFieldsMacro(StringBuffer expanded,
                                           String dmName,
                                           String prefix,
                                           Map dms,
                                           Map dmDescriptors,
                                           String fieldNameSeparator)
        throws BLException {

        // check for prefix in the DM name
        int sepIndex = dmName.lastIndexOf(fieldNameSeparator);
        if(sepIndex >= 0) {
            prefix = dmName.substring(0, sepIndex + 1);
            dmName = dmName.substring(sepIndex + 1);
        }

        // find the DM descriptor
        Class dmClass = (Class)dms.get(dmName);
        if(dmClass == null)
            throw new BLException("Reference to nonexistent DM \"" + dmName + "\".");
        DAOConfig.DMDescriptor dmDesc = (DAOConfig.DMDescriptor)dmDescriptors.get(dmClass);

        if((dmDesc.tableName == null) || (dmDesc.tableName.length() == 0))
            throw new BLException("DM \"" + dmName +
                                  "\" cannot be referenced from a query since its" +
                                  " \"table\" attribute is not defined.");

        // add id fields first
        int nAdded = 0;
        Iterator it = dmDesc.idFieldNamesSet.iterator();
        while(it.hasNext()) {
            String fieldName = (String)it.next();
            if(nAdded > 0)
                expanded.append(",\n");
            expanded.append(dmDesc.tableName);
            expanded.append('.');
            expanded.append(fieldName);
            expanded.append(" AS ");
            if(prefix != null)
                expanded.append(prefix);
            expanded.append(fieldName);
            nAdded++;
        }

        // add all other fields
        it = dmDesc.fields.values().iterator();
        while(it.hasNext()) {
            Field fieldField = (Field)it.next();
            if(!fieldField.getType().isArray() &&
               !(DataModel.class).isAssignableFrom(fieldField.getType())) {
                String fieldName = fieldField.getName();
                if(!dmDesc.idFieldNamesSet.contains(fieldName)) {
                    if(nAdded > 0)
                        expanded.append(",\n");
                    expanded.append(dmDesc.tableName);
                    expanded.append('.');
                    expanded.append(fieldName);
                    expanded.append(" AS ");
                    if(prefix != null)
                        expanded.append(prefix);
                    expanded.append(fieldName);
                    nAdded++;
                }
            }
        }

        // all done
        return nAdded;
    }
}
