package com.boylesoftware.cb2;

import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.List;
import java.util.ArrayList;


/**
 * Parent class for Data Models used by the DAO.
 *
 * @author Lev Himmelfarb
 * @version $Id: DataModel.java,v 1.6 2004/10/11 18:17:36 levahim Exp $
 *
 * @see DAO
 */
public abstract class DataModel
    implements Serializable {

    /**
     * Returns hash code of the data model. This implementation
     * actually returns the hash code of the model's string
     * representation returned by <code>toString</code> method making
     * two data models equal if they contain the same data.
     *
     * <p>Note, that this method is not very effective. Its performace
     * greatly depends on the amount of data in the DM, so try to
     * avoid using large DMs in hash-like data structures.
     *
     * @return the model's hash code.
     */
    public int hashCode() {

        return this.toString().hashCode();
    }

    /**
     * Indicates whether some other object is "equal to" this data
     * model. This implementation checks if <code>obj</code> is a
     * reference to the same instance as this one, then, if it is not so, it
     * checks if <code>obj</code> belongs to the exactly same class and if
     * its string representation is the same.
     *
     * @param obj data model, to which to compare this one.
     *
     * @return <code>true</code> if <code>obj</code> equals to this
     * data model, that is it is not <code>null</code>, has the same
     * type and the same string representation.
     */
    public boolean equals(Object obj) {

        return ((obj != null) &&
                ((obj == this) ||
                 (obj.getClass().getName().equals(this.getClass().getName()) &&
                  obj.toString().equals(this.toString()))));
    }

    /**
     * Gives string representation of the data model including values of all
     * its fields. The output includes nested DMs as well.
     *
     * <p>Note, that the resulting string generated by this method can be HUGE.
     *
     * @return string representation of the data model.
     */
    public String toString() {

        StringBuffer res = new StringBuffer();

        res.append('{');

        Field [] fields = this.getClass().getFields();
        for(int i = 0; i < fields.length; i++) {
            if(i > 0)
                res.append(", ");
            res.append(fields[i].getName());
            res.append("=[");
            res.append(fields[i].getType().getName());
            res.append('/');
            try {
                Object fieldVal = fields[i].get(this);
                if(fieldVal != null) {
                    if(fields[i].getType().isArray()) {
                        Object [] vals = (Object [])fieldVal;
                        for(int j = 0; j < vals.length; j++) {
                            res.append('[');
                            if(vals[j] != null) {
                                res.append(vals[j].getClass().getName());
                                res.append('/');
                                res.append(vals[j].toString());
                            } else {
                                res.append("<null>");
                            }
                            res.append(']');
                        }
                    } else {
                        res.append(fieldVal.toString());
                    }
                } else {
                    res.append("<null>");
                }
            } catch(IllegalAccessException e) {
                res.append("<ERROR: ILLEGAL ACCESS!>");
            }
            res.append("]");
        }

        res.append('}');

        return res.toString();
    }

    /**
     * Finds fields that have different values in this DM and the specified
     * one, which must belong to the exactly same class.
     *
     * <p>This method ignores all array and nested DM fields.
     *
     * @param dm another DM.
     *
     * @return array of names of fields that have different values in this and
     * the specified DM. Returns an empty array if the DMs are equal (equal,
     * but excluding nested DMs, as opposed to the <code>equals</code> method,
     * which compares string representations of the DMs that include values
     * from all nested DMs and arrays of nested DMs).
     *
     * @throws IllegalArgumentException if type of the specified DM is
     * different from the type of this one.
     */
    public String [] diff(DataModel dm) {

        if(dm.getClass() != this.getClass())
            throw new IllegalArgumentException("The type of the specified DM is " +
                                               dm.getClass().getName() +
                                               " and is different from the type of this DM, which is " +
                                               this.getClass().getName() + ".");

        if(dm == this)
            return new String[0];

        Field [] fields = this.getClass().getFields();
        List res = new ArrayList(fields.length);
        for(int i = 0; i < fields.length; i++) {
            Class fieldType = fields[i].getType();
            if(!(DataModel.class).isAssignableFrom(fieldType) &&
               !fieldType.isArray()) {
                try {
                    Object val1 = fields[i].get(this);
                    Object val2 = fields[i].get(dm);
                    if(val1 != null) {
                        if(!val1.equals(val2))
                            res.add(fields[i].getName());
                    } else {
                        if(val2 != null)
                            res.add(fields[i].getName());
                    }
                } catch(IllegalAccessException e) { /* should never actually happen */ }
            }
        }

        return (String [])res.toArray(new String[res.size()]);
    }

    /**
     * Gets a field's value by its name.
     *
     * @param name name of a DM field.
     *
     * @return the field's value.
     *
     * @throws NoSuchFieldException if no public field with the specified name
     * is defined in the DM.
     */
    public Object getField(String name)
        throws NoSuchFieldException {

        Field field = this.getClass().getField(name);
        try {
            return field.get(this);
        } catch(IllegalAccessException e) { // can never happen
            throw new Error("Cannot access DM field through reflection.", e);
        }
    }

    /**
     * Sets new value to a field via reflection. This method uses
     * <code>set</code> method of <code>java.lang.reflect.Field</code>, and
     * therefore all features of the the <code>set</code> method apply to
     * this method as well.
     *
     * @param name DM field's name.
     * @param value new value.
     *
     * @throws NoSuchFieldException if no public field with the specified name
     * is defined in the DM.
     */
    public void setField(String name, Object value)
        throws NoSuchFieldException {

        Field field = this.getClass().getField(name);
        try {
            field.set(this, value);
        } catch(IllegalAccessException e) { // can never happen
            throw new Error("Cannot access DM field through reflection.", e);
        }
    }
}
