package com.boylesoftware.cb2;

import java.util.Collection;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;


/**
 * Special implementation of Data Model, which is not associated with any
 * specific result set, but is able to contain any data in a form of name-value
 * mappings. A query can be associated with <code>DynamicDataModel</code> and
 * then used with a DAO fetch call. The DAO then will return instances of
 * <code>DynamicDataModel</code> containing data for all columns in the result
 * set. The keys in the data model's map will be column labels aquired via
 * <code>getColumnLabel</code> call on the JDBC's
 * <code>ResultSetMetaData</code> object, and the values - objects returned by
 * <code>getObject</code> method on the <code>ResultSet</code>.
 *
 * <p>Note, that although it is very tempting to use this class in many cases,
 * it is significantly less efficient than using normal DMs, so it should be
 * done only in very special cases.
 *
 * <p>This class is not synchronized.
 *
 * @author Lev Himmelfarb
 * @version $Id: DynamicDataModel.java,v 1.3 2004/05/19 14:24:05 levahim Exp $
 */
public final class DynamicDataModel
    extends DataModel
    implements Map {

    /**
     * The row data.
     */
    private final Map fields;


    /**
     * Constructs a DM.
     *
     * @param capacity the internal map's capacity.
     */
    DynamicDataModel(int capacity) {

        this.fields = new HashMap(capacity);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Override standard DM methods
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gives string representation of the data model including values of all
     * its fields
     *
     * @return string representation of the data model.
     */
    public String toString() {

        StringBuffer res = new StringBuffer();

        res.append('{');

        Iterator i = this.fields.entrySet().iterator();
        int n = 0;
        while(i.hasNext()) {
            Map.Entry entry = (Map.Entry)i.next();
            String fieldName = entry.getKey().toString();
            Object fieldValue = entry.getValue();
            if(n++ > 0)
                res.append(", ");
            res.append(fieldName);
            res.append("=[");
            if(fieldValue != null) {
                res.append(fieldValue.getClass().getName());
                res.append('/');
                res.append(fieldValue.toString());
            } else {
                res.append("<null>");
            }
            res.append("]");
        }

        res.append('}');

        return res.toString();
    }

    /**
     * Finds fields that have different values in this DM and the specified
     * one, which must be a <code>DynamicDataModel</code> too.
     *
     * @param dm another dynamic DM.
     *
     * @return array of names of fields that have different values in this and
     * the specified DM. Returns an empty array if the DMs are equal.
     *
     * @throws IllegalArgumentException if type of the specified DM is not
     * <code>DynamicDataModel</code>.
     */
    public String [] diff(DataModel dm) {

        if(!(dm instanceof DynamicDataModel))
            throw new IllegalArgumentException("The type of the specified DM is " +
                                               dm.getClass().getName() +
                                               " and is different from the type of this DM, which is " +
                                               this.getClass().getName() + ".");

        if(dm == this)
            return new String[0];

        Iterator i = this.fields.entrySet().iterator();
        List res = new ArrayList(this.fields.size());
        while(i.hasNext()) {
            Map.Entry entry = (Map.Entry)i.next();
            String fieldName = entry.getKey().toString();
            Object val1 = entry.getValue();
            Object val2 = ((DynamicDataModel)dm).get(fieldName);
            if(val1 != null) {
                if(!val1.equals(val2))
                    res.add(fieldName);
            } else {
                if(val2 != null)
                    res.add(fieldName);
            }
        }

        return (String [])res.toArray(new String[res.size()]);
    }

    /**
     * Gets a field's value by its name. In the case of a
     * <code>DynamicDataModel</code> this does almost the same as the
     * <code>get</code> method from the <code>java.util.Map</code> interface
     * implemented by this class. The difference is that this method throws a
     * <code>NoSuchFieldException</code> if the requested field does not exist,
     * while the <code>get</code> method simply returns <code>null</code>.
     *
     * @param name name of a DM field.
     *
     * @return the field's value.
     *
     * @throws NoSuchFieldException if the requested field is not present in
     * the DM.
     */
    public Object getField(String name)
        throws NoSuchFieldException {

        if(!this.fields.containsKey(name))
            throw new NoSuchFieldException("Field \"" + name +
                                           "\" is not present in the DM.");

        return this.fields.get(name);
    }

    /**
     * Sets a field's value by the field's name. In the case of a
     * <code>DynamicDataModel</code> this is a complete analogue of
     * <code>put</code> method from the <code>java.util.Map</code> interface
     * implemented by this class.
     *
     * @param name name of a DM field.
     * @param value new value.
     */
    public void setField(String name, Object value) {

        this.fields.put(name, value);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Map interface
    ///////////////////////////////////////////////////////////////////////////

    public int size() {

        return this.fields.size();
    }

    public void clear() {

        this.fields.clear();
    }

    public boolean isEmpty() {

        return this.fields.isEmpty();
    }

    public boolean containsKey(Object key) {

        return this.fields.containsKey(key);
    }

    public boolean containsValue(Object value) {

        return this.fields.containsValue(value);
    }

    public Collection values() {

        return this.fields.values();
    }

    public void putAll(Map t) {

        this.fields.putAll(t);
    }

    public Set entrySet() {

        return this.fields.entrySet();
    }

    public Set keySet() {

        return this.fields.keySet();
    }

    public Object get(Object key) {

        return this.fields.get(key);
    }

    public Object remove(Object key) {

        return this.fields.remove(key);
    }

    public Object put(Object key, Object value) {

        return this.fields.put(key, value);
    }
}
