package com.boylesoftware.cb2;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;


/**
 * Interface for modules that are used by the DAO's fetch calls to prepare and
 * execute SQL statements.
 *
 * <p>A custom imlpementation of <code>FetchExecutor</code> can be installed to
 * the DAO as a DAO plugin. Only one such plugin is allowed for a DAO. By
 * default, if no custom <code>FetchExecutor</code> plugin is installed, the
 * DAO uses <code>GenericFetchExecutor</code>.
 *
 * <p>Unlike many other DAO plugin types, the DAO does not allow concurrent
 * threads into the same instance of the fetch executor. Also, the fetch
 * executor can be stateful - the DAO guarantees that single instance will be
 * used during one fetch call. 
 *
 * <p>Here is the protocol of how the DAO calls the fetch executor:
 *
 * <ol>
 *
 * <li>When a fetch method is called on the DAO, an instance of the currently
 * configured for this DAO <code>FetchExecutor</code> implementation (
 * <code>GenericFetchExecutor</code> by default) is taken from the pool. Then
 * the <code>init</code> method is called on it and the fetch call parameters
 * are passed to it so it can initialize its internals.</li>
 *
 * <li>The <code>prepareStatement</code> is called and it returns the
 * <code>java.sql.Statement</code> object to be executed by the DAO. If there
 * are query parameters, then the method must return an instance of
 * <code>java.sql.PreparedStatement</code>, so the DAO can set the parameters
 * on it.</li>
 *
 * <li>If there are query parameters, the DAO sets them on the statement
 * returned by the previous call. The fetch executor is not involved in this
 * process (but <code>QueryParamProcessor<code> and
 * <code>QueryParamSetter</code> plugins are, if defined for the DAO).</li>
 *
 * <li>The <code>execute</code> is called and the fetch executor actually
 * executes the query, which returnes a result set. The result set is stored
 * then internally in the fetch executor.</li>
 *
 * <li>Immediately, the <code>getResultSet</code> is called and it
 * returns a reference to the just fetched result set.</li>
 *
 * <li><code>setToTheFirstRow</code> is called.</li>
 *
 * <li>If the previous call returned <code>true</code>, then the DAO processes
 * the result set rows, and, when it is finished, it calls <code>finish</code>
 * method on the fetch executor. Otherwise, if the
 * <code>setToTheFirstRow</code> call returned <code>false</code>, this step is
 * skipped and the <code>finish</code> method is never called.</li>
 *
 * <li>Finally, <code>destroy</code> method is called and the fetch executor
 * instance is returned back to the pool. Note, that the <code>destroy</code>
 * method gets called always, even if there were exceptions thrown by other
 * methods.</li>
 *
 * </ol>
 *
 * @author Lev Himmelfarb
 * @version $Id: FetchExecutor.java,v 1.3 2004/10/11 18:17:36 levahim Exp $
 *
 * @see DAO#fetch(String, java.util.Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
 */
public interface FetchExecutor
    extends DAOPlugin {

    /**
     * Initializes the executor before calling any other methods.
     *
     * @param daoConfig the DAO configuration.
     * @param con the database connection to be used.
     * @param dmClass class of the top-level DMs to be returned by the call.
     * @param query the SQL query text.
     * @param queryParams values for the query parameters, <code>null</code> or
     * empty array if the query needs no parameters.
     * @param page number of the result set page requested, starting from zero.
     * @param pageSize number of records in one result set page, equal or less
     * than zero if result set pagination is not requested.
     * @param fetchResultDescriptor the <code>FetchResultDescriptor</code>
     * instance or <code>null</code> if no result info requested.
     *
     * @throws SQLException if the executor cannot be initialized.
     */
    void init(DAOConfig daoConfig,
              Connection con,
              Class dmClass,
              String query,
              Object [] queryParams,
              int page,
              int pageSize,
              FetchResultDescriptor fetchResultDescriptor)
        throws SQLException;

    /**
     * Prepares a statement object before the <code>execute</code> method call.
     * Note, that if the query requires parameters, that is the
     * <code>queryParams</code> passed to the <code>init</code> method was not
     * <code>null</code> and not empty, then this method is required to return
     * an instance of <code>java.sql.PreparedStatement</code>, so the DAO can
     * set query parameters on it before calling <code>execute</code>.
     *
     * @return the statement object, which may be an instance of
     * <code>java.sql.PreparedStatement</code>.
     *
     * @throws SQLException if a database error happens.
     */
    Statement prepareStatement()
        throws SQLException;

    /**
     * Executes the query and returns the statement, which was actually used
     * for getting the result set (possibly a different statement from what
     * <code>prepareStatement</code> returned). The <code>getResultSet</code>
     * method is called later to retrieve the result set.
     *
     * @return the statement owning the result set.
     *
     * @throws SQLException if a database error happens.
     */
    Statement execute()
        throws SQLException;

    /**
     * Gets the result set after <code>execute</code> method call.
     *
     * @return result set with the requested data.
     */
    ResultSet getResultSet();

    /**
     * Positions the result set cursor to the first row of the requested data.
     *
     * @return <code>true</code> if the cursor has been set or
     * <code>false</code> if there is not data (for example the result set is
     * empty or the requested page does not exist).
     *
     * @throws SQLException if a database error happens.
     */
    boolean setToTheFirstRow()
        throws SQLException;

    /**
     * Called after everything has been processed and the resulting array of
     * DMs formed. This method does not get called if the
     * <code>setToTheFirstRow</code> returned <code>false</code>.
     *
     * @param nRowsProcessed number of result set rows processed by the DAO.
     * @param result the result, which will be returned by the DAO to the
     * caller as a <code>java.util.List</code> of DMs.
     *
     * @throws SQLException if a database error happens.
     */
    void finish(int nRowsProcessed, List result)
        throws SQLException;

    /**
     * Releases any used resources, but does not close any
     * <strong>exposed</strong> to the DAO database resources like the result
     * set, the statement returned by the <code>execute</code> method and the
     * connection.
     */
    void destroy();
}
