package com.boylesoftware.cb2;

import java.sql.ResultSet;
import java.sql.SQLException;


/**
 * Classes that implement this interface can be passed along with a DAO fetch
 * call to be installed into the DAO's result set processing loop to perform
 * certain operations on the DMs as they are coming out of the DAO and before
 * they get packed into an array and returned back to the caller. These fetch
 * result processors need to be wrapped into
 * <code>FetchResultProcessorHandler</code> objects before used with a fetch
 * call. The <code>FetchResultProcessorHandler</code> allows to install the
 * processor on a certain DM nesting level, so it is called only for the DMs
 * formed on that level.
 *
 * @author Lev Himmelfarb
 * @version $Id: FetchResultProcessor.java,v 1.3 2004/10/11 18:17:36 levahim Exp $
 *
 * @see FetchResultProcessorHandler
 * @see DAO#fetch(String, java.util.Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
 */
public interface FetchResultProcessor {

    /**
     * Called by the DAO for each DM on a certain DM nesting level just after
     * it has processed a result set row and populated the DM's fields.
     *
     * @param dm the DM created an populated by the DAO.
     * @param rs reference to the result set.
     * @param firstColumnIndex index of a column in the result set, which is
     * the first column that belongs to the DM.
     * @param lastColumnIndex index of a column in the result set, which is
     * the last column that belongs to the DM.
     *
     * @return DM, which has to be taken by the DAO as the result of the row
     * processing. If the implementation returns <code>null</code>, the DAO
     * unconditionally aborts current row processing and nothing is added to the
     * resulting array of DMs for this row. If multiple processors are
     * installed on the same nesting level, others do not get called if one of
     * them returns <code>null</code>.
     *
     * @throws SQLException if a database error happens.
     * @throws BLException if a serious error happens during the DM processing.
     */
    DataModel processDM(DataModel dm, ResultSet rs, int firstColumnIndex, int lastColumnIndex)
        throws BLException, SQLException;
}
