package com.boylesoftware.cb2;


/**
 * Wrapper for <code>FetchResultPocessor</code> interface implementations, so
 * they can be passed to the DAO along with a fetch call. The wrapper adds
 * DM nesting level, on which the processor should be installed. If the level
 * is zero, the processor will be called for each top-level DM, which goes to
 * the array returned by the fetch call back to the caller. If the level is
 * one, then the processor will be called for DMs in the nested DM array in the
 * top-level DM. And so on.
 *
 * @author Lev Himmelfarb
 * @version $Id: FetchResultProcessorHandler.java,v 1.3 2004/10/11 18:17:36 levahim Exp $
 *
 * @see DAO#fetch(String, java.util.Set, Object [], String [], int, int, int, FetchResultProcessorHandler [], FetchResultDescriptor)
 */
public final class FetchResultProcessorHandler {

    /**
     * Processor hook implementation, which provides DM processing
     * callback.
     */
    private final FetchResultProcessor processor;

    /**
     * Nested array level, starting from <code>0</code>.
     */
    private final int level;


    /**
     * Creates a handler, which can be passed to the DAO with a fetch call.
     *
     * @param processor processor implementation.
     * @param level nested DM array level starting from zero.
     */
    public FetchResultProcessorHandler(FetchResultProcessor processor,
                                       int level) {

        this.processor = processor;
        this.level = level;
    }


    /**
     * Returns the processor hook implementation, which provides DM
     * processing callback.
     */
    FetchResultProcessor getProcessor() {

        return this.processor;
    }

    /**
     * Returns the nested array level, starting from <code>0</code>.
     */
    int getLevel() {

        return this.level;
    }
}
