package com.boylesoftware.cb2;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;


/**
 * Generic <code>FetchExecutor</code> implementation used by the DAO by default
 * when no other DAO plugin implementing <code>FetchExecutor</code> interface
 * is installed. This implementation uses JDBC scrollable result sets for
 * paginated results. In general, it tries to use the most standard JDBC
 * features to implement functionality.
 *
 * @author Lev Himmelfarb
 * @version $Id: GenericFetchExecutor.java,v 1.6 2004/10/11 18:17:36 levahim Exp $
 */
public class GenericFetchExecutor
    implements FetchExecutor {

    /**
     * The application context.
     */
    protected ApplicationContext appCtx;

    /**
     * The DAO configuration. Set by the <code>init</code> method.
     */
    protected DAOConfig daoConfig;

    /**
     * The database connection. Set by the <code>init</code> method.
     */
    protected Connection con;

    /**
     * Text of the SQL query to be executed. Set by the <code>init</code>
     * method.
     */
    protected String query;

    /**
     * The query parameters as passed to the DAO. Set by the <code>init</code>
     * method.
     */
    protected Object [] queryParams;

    /**
     * Requested page starting from zero. Set by the <code>init</code> method.
     */
    protected int page;

    /**
     * Requested page size. Is less or equal to zero if no result set
     * pagination is requested. Set by the <code>init</code> method.
     */
    protected int pageSize;

    /**
     * Reference to the fetch result descriptor instance passed to the DAO for
     * the call. May be <code>null</code>. Set by the <code>init</code> method.
     */
    protected FetchResultDescriptor fetchResultDescriptor;

    /**
     * The statement object.
     */
    protected Statement stmt;

    /**
     * Set to <code>true</code> if the <code>stmt</code> is an instance of
     * <code>java.sql.PreparedStatement</code>.
     */
    protected boolean usePreparedStatement;

    /**
     * The result set object.
     */
    protected ResultSet rs;


    /**
     * Sets the application context.
     *
     * @param appCtx the application context.
     */
    public void setApplicationContext(ApplicationContext appCtx) {

        this.appCtx = appCtx;
    }


    /**
     * Initializes the executor setting up internal member variables.
     */
    public void init(DAOConfig daoConfig,
                     Connection con,
                     Class dmClass,
                     String query,
                     Object [] queryParams,
                     int page,
                     int pageSize,
                     FetchResultDescriptor fetchResultDescriptor) {

        this.daoConfig = daoConfig;
        this.con = con;
        this.query = query;
        this.queryParams = queryParams;
        this.page = page;
        this.pageSize = pageSize;
        this.fetchResultDescriptor = fetchResultDescriptor;

        this.stmt = null;
        this.rs = null;

        this.usePreparedStatement = ((queryParams != null) && (queryParams.length > 0));
    }

    /**
     * Creates a statement object. This method sets the <code>stmt</code>
     * member variable.
     *
     * @throws SQLException if a JDBC error happens.
     */
    public Statement prepareStatement()
        throws SQLException {

        try {

            this.stmt = ((this.pageSize > 0) && ((this.page > 0) || (this.fetchResultDescriptor != null)) ?
                         (this.usePreparedStatement ?
                          this.con.prepareStatement(this.query, ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY) :
                          this.con.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY)) :
                         (this.usePreparedStatement ?
                          this.con.prepareStatement(this.query, ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY) :
                          this.con.createStatement(ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY)));

            if(this.daoConfig.doesDriverSupportSetFetchSize() && this.pageSize > 0) {
                this.stmt.setFetchSize(this.pageSize + 1);
            }

        } catch(SQLException e) {
            if(this.stmt != null) {
                try {
                    this.stmt.close();
                } catch(SQLException e1) { /* ignore it */ }
            }
            throw e;
        }

        return this.stmt;
    }

    /**
     * Executes the <code>stmt</code> statement and sets the <code>rs</code>
     * member variable.
     *
     * @throws SQLException if a JDBC error happens.
     */
    public Statement execute()
        throws SQLException {

        this.rs = (this.usePreparedStatement ?
                   ((PreparedStatement)this.stmt).executeQuery() :
                   this.stmt.executeQuery(this.query));

        return this.stmt;
    }

    /**
     * Gets the result set.
     */
    public ResultSet getResultSet() {

        return this.rs;
    }

    /**
     * Positions the result set cursor to the first row of the requested data.
     *
     * @throws SQLException if a JDBC error happens.
     */
    public boolean setToTheFirstRow()
        throws SQLException {

        if(!this.rs.next()) {

            if(this.fetchResultDescriptor != null)
                this.fetchResultDescriptor.setRowsTotal(0);

            return false;
        }

        // determine the total number of records and scroll to the page
        if(this.pageSize > 0) {

            if(this.fetchResultDescriptor != null) {

                // get the number of rows
                this.rs.last();
                this.fetchResultDescriptor.setRowsTotal(this.rs.getRow());

                // scroll to the page
                return this.rs.absolute((this.page * this.pageSize) + 1);
            }

            // scroll to the page
            if(this.page > 0)
                return this.rs.relative(this.page * this.pageSize);
        }

        return true;
    }

    /**
     * Finishes query execution and result set processing and sets up fetch
     * result descriptor in <code>fetchResultDescriptor</code> member
     * variable if it is not <code>null</code>.
     */
    public void finish(int nRowsProcessed, List result) {

        if((this.pageSize <= 0) && (this.fetchResultDescriptor != null))
            this.fetchResultDescriptor.setRowsTotal(nRowsProcessed);
    }

    /**
     * Sets <code>null</code> to all internal member variables.
     */
    public void destroy() {

        this.daoConfig = null;
        this.con = null;
        this.query = null;
        this.fetchResultDescriptor = null;
        this.stmt = null;
        this.rs = null;
    }
}
