package com.boylesoftware.cb2;

import java.sql.Connection;
import java.sql.SQLException;
import javax.transaction.UserTransaction;
import javax.transaction.Status;
import javax.transaction.NotSupportedException;
import javax.transaction.SystemException;
import javax.transaction.RollbackException;
import javax.transaction.HeuristicMixedException;
import javax.transaction.HeuristicRollbackException;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.commons.logging.Log;


/**
 * Transaction context implementation that works through JTA's
 * <code>UserTransaction</code> interface, which it takes from the JNDI.
 *
 * <p>By default this implementation looks for <code>UserTransaction</code>
 * object in JNDI under name "java:comp/UserTransaction", but it can be
 * overriden by specifying
 * <code>com.boylesoftware.cb2.userTransaction.jndiPath</code> application
 * property.
 *
 * @author Lev Himmelfarb
 * @version $Id: JTATransactionContext.java,v 1.3 2004/10/11 18:17:36 levahim Exp $
 */
final class JTATransactionContext
    implements TransactionContext {

    /**
     * Reference to the application context.
     */
    private ApplicationContext appCtx;

    /**
     * The <code>UserTransaction</code> object.
     */
    private UserTransaction userTX;

    /**
     * Tells if the transaction is active.
     */
    private boolean inTX;

    /**
     * Current trasaction nesting level.
     */
    private int nestedLevel;


    /**
     * Setup the instance.
     *
     * @param appCtx the application context.
     *
     * @throws ApplicationException if cannot find <code>UserTransaction</code>
     * object in the JNDI.
     */
    public void setup(ApplicationContext appCtx)
        throws ApplicationException {

        this.appCtx = appCtx;
        this.inTX = false;
        this.nestedLevel = 0;

        String utJndiPath = appCtx.
            getProperty("com.boylesoftware.cb2.userTransaction.jndiPath");
        if(utJndiPath == null)
            utJndiPath = "java:comp/UserTransaction";

        try {
            InitialContext ic = null;
            try {
                ic = new InitialContext();

                this.userTX = (UserTransaction)ic.lookup(utJndiPath);
                if(this.userTX == null)
                    throw new ApplicationException("Cannot find UserTransaction object in the JNDI under name \"" +
                                                   utJndiPath + "\".");

            } finally {
                if(appCtx.getCloseJNDIContexts() && (ic != null))
                    ic.close();
            }
        } catch(NamingException e) {
            throw new ApplicationException("Got error retrieving UserTransaction object from JNDI.", e);
        }
    }


    /**
     * Begins a transaction.
     *
     * @throws ApplicationException if the transaction cannot be
     * started due to an underlying transaction manager error.
     */
    public synchronized void begin()
        throws ApplicationException {

        Log log = this.appCtx.getLog(LOG_SUBSYS);
        try {
            log.debug("starting transaction, tx status is [" + getTXStatusName(this.userTX.getStatus()) + "]");
            this.userTX.begin();
            log.debug("transaction started, tx status is [" + getTXStatusName(this.userTX.getStatus()) + "]");
        } catch(NotSupportedException e) {
            throw new ApplicationException("Nested transactions are not supported.", e);
        } catch(SystemException e) {
            throw new ApplicationException("Could not begin new transaction.", e);
        }

        this.nestedLevel++;
        this.inTX = true;

        log.debug("tx is active");
    }

    /**
     * Gets the database connection enrolled in the transaction.
     *
     * @param dataSourceName logical name of the datasource.
     *
     * @return the database connection.
     *
     * @throws SQLException if the connection cannnot be retrieved due to the
     * underlying data source error.
     * @throws IllegalStateException if the transaction is not ative.
     */
    public synchronized Connection getConnection(String dataSourceName)
        throws SQLException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        return this.appCtx.getNoTXConnection(dataSourceName);
    }

    /**
     * Commits the transaction.
     *
     * @return <code>true</code>.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not ative.
     */
    public synchronized boolean commit()
        throws ApplicationException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        Log log = this.appCtx.getLog(LOG_SUBSYS);

        try { // for SystemException
            try {
                log.debug("committing transaction, tx status is [" + getTXStatusName(this.userTX.getStatus()) + "]");
                this.userTX.commit();
                log.debug("transaction committed, tx status is [" + getTXStatusName(this.userTX.getStatus()) + "]");
            } catch(RollbackException e) {
                log.debug("transaction was marked for rollback only and has been rolled back, tx status is [" +
                          getTXStatusName(this.userTX.getStatus()) + "]");
            } catch(HeuristicMixedException e) {
                log.warn("Heuristic decision was made committing distributed transaction, tx status is [" +
                         getTXStatusName(this.userTX.getStatus()) + "]", e);
            } catch(HeuristicRollbackException e) {
                log.warn("Heuristic decision was made committing distributed transaction, tx status is [" +
                         getTXStatusName(this.userTX.getStatus()) + "]", e);
            }
        } catch(SystemException e) {
            throw new ApplicationException("Could not commit transaction due to a serious error" +
                                           " in the underlying subsystem.", e);
        } finally {
            if(--this.nestedLevel <= 0) {
                this.inTX = false;
                this.appCtx.getLog(LOG_SUBSYS).debug("tx is inactive");
            }
        }

        return !this.inTX;
    }

    /**
     * Rollbacks the transaction.
     *
     * @return <code>true</code>.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not ative.
     */
    public synchronized boolean rollback()
        throws ApplicationException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        Log log = this.appCtx.getLog(LOG_SUBSYS);

        try {
            log.debug("rolling distributed transaction back, tx status is [" + getTXStatusName(this.userTX.getStatus()) + "]");
            this.userTX.rollback();
            log.debug("distributed transaction rolled back, tx status is [" + getTXStatusName(this.userTX.getStatus()) + "]");
        } catch(SystemException e) {
            throw new ApplicationException("Could not rollback distributed transaction due to a serious error" +
                                           " in the underlying subsystems.", e);
        } finally {
            if(--this.nestedLevel <= 0) {
                this.inTX = false;
                this.appCtx.getLog(LOG_SUBSYS).debug("tx is inactive");
            }
        }

        return !this.inTX;
    }

    /**
     * Marks the transaction for rollback only.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     */
    public synchronized void setRollbackOnly()
        throws ApplicationException {

        if(this.inTX) {

            Log log = this.appCtx.getLog(LOG_SUBSYS);

            try {
                log.debug("marking distributed transaction for rollback only, tx status is [" +
                          getTXStatusName(this.userTX.getStatus()) + "]");
                this.userTX.setRollbackOnly();
                log.debug("distributed transaction marked for rollback only, tx status is [" +
                          getTXStatusName(this.userTX.getStatus()) + "]");
            } catch(SystemException e) {
                throw new ApplicationException("Could not mark distributed transaction for rollback only due to a serious error" +
                                               " in the underlying subsystems.", e);
            }
        }
    }

    /**
     * Tells if the transaction is marked for rollback only.
     *
     * @return <code>true</code> if the transaction is marked for
     * rollback only. Always returns <code>false</code> if there no active
     * transaction.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     */
    public synchronized boolean isRollbackOnly()
        throws ApplicationException {

        try {
            return (this.inTX ? (this.userTX.getStatus() == Status.STATUS_MARKED_ROLLBACK) : false);
        } catch(SystemException e) {
            throw new ApplicationException("Could not get distributed transaction status due to a serious error" +
                                           " in the underlying subsystems.", e);
        }
    }

    /**
     * Tells if the transaction is active.
     *
     * @return <code>true</code> if the transaction is active.
     */
    public synchronized boolean isActive() {

        return this.inTX;
    }

    /**
     * Tells if unbound transactions are supported.
     *
     * @return <code>false</code>.
     */
    public synchronized boolean supportsUnboundTransactions() {

        return false;
    }


    /**
     * Gets human-readable string representaion of the specified
     * transaction status from <code>javax.transaction.Status</code>
     * interface definition. Used for logging.
     *
     * @param status transaction status code as returns by
     * <code>UserTransaction.getStatus</code> method.
     *
     * @return name of the specified transaction status.
     */
    private final static String getTXStatusName(int status) {

        switch(status) {
        case Status.STATUS_ACTIVE:
            return "ACTIVE";
        case Status.STATUS_COMMITTED:
            return "COMMITTED";
        case Status.STATUS_COMMITTING:
            return "COMMITTING";
        case Status.STATUS_MARKED_ROLLBACK:
            return "MARKED_ROLLBACK";
        case Status.STATUS_NO_TRANSACTION:
            return "NO_TRANSACTION";
        case Status.STATUS_PREPARED:
            return "PREPARED";
        case Status.STATUS_PREPARING:
            return "PREPARING";
        case Status.STATUS_ROLLEDBACK:
            return "ROLLEDBACK";
        case Status.STATUS_ROLLING_BACK:
            return "ROLLING_BACK";
        case Status.STATUS_UNKNOWN:
            return "UNKNOWN";
        default:
            return "invalid tx status";
        }
    }
}
