package com.boylesoftware.cb2;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.SQLException;

import org.apache.commons.logging.Log;


/**
 * Transaction context implementation that uses SQL language commands on
 * database connections to manage transactions. By default this
 * implementation uses <code>BEGIN TRANSACTION</code> command to begin a new
 * transaction, <code>COMMIT TRANSACTION</code> to commit it, and
 * <code>ROLLBACK TRANSACTION</code> to roll it back. These commands can be
 * overridden with <code>com.boylesoftware.cb2.txContext.command.begin</code>,
 * <code>com.boylesoftware.cb2.txContext.command.commit</code> and
 * <code>com.boylesoftware.cb2.txContext.command.rollback</code> application
 * properties respectively.
 *
 * <p>This implementation does not support distributed transactions, but,
 * depending on the underlying database, may support nested transactions.
 *
 * @author Lev Himmelfarb
 * @version $Id: LanguageTransactionContext.java,v 1.4 2004/10/11 18:17:36 levahim Exp $
 */
final class LanguageTransactionContext
    implements TransactionContext {

    /**
     * Reference to the application context.
     */
    private ApplicationContext appCtx;

    /**
     * SQL command to begin a transaction.
     */
    private String beginCommand;

    /**
     * SQL command to commit a transaction.
     */
    private String commitCommand;

    /**
     * SQL command to rollback a transaction.
     */
    private String rollbackCommand;

    /**
     * Tells if the transaction is active.
     */
    private boolean inTX;

    /**
     * Current trasaction nesting level.
     */
    private int nestedLevel;

    /**
     * Number fo transactions actually begun.
     */
    private int nTxsBegun;

    /**
     * The database connection bound to the transaction context when the
     * transaction is active.
     */
    private TxConnection con;

    /**
     * Name of the data source, from which the <code>con</code> connection was
     * retrieved.
     */
    private String dataSourceName;

    /**
     * Tells if the transaction has been marked for rollback only.
     */
    private boolean isRollbackOnly;


    /**
     * Setup the instance.
     *
     * @param appCtx the application context.
     */
    public void setup(ApplicationContext appCtx) {

        this.appCtx = appCtx;
        if((this.beginCommand = appCtx.getProperty("com.boylesoftware.cb2.txContext.command.begin")) == null)
            this.beginCommand = "BEGIN TRANSACTION";
        if((this.commitCommand = appCtx.getProperty("com.boylesoftware.cb2.txContext.command.commit")) == null)
            this.commitCommand = "COMMIT TRANSACTION";
        if((this.rollbackCommand = appCtx.getProperty("com.boylesoftware.cb2.txContext.command.rollback")) == null)
            this.rollbackCommand = "ROLLBACK TRANSACTION";
        this.inTX = false;
        this.nestedLevel = 0;
        this.nTxsBegun = 0;
        this.con = null;
        this.isRollbackOnly = false;
    }

    /**
     * Begins a transaction.
     *
     * @throws ApplicationException if the transaction cannot be
     * started.
     */
    public synchronized void begin()
        throws ApplicationException {

        if(this.nestedLevel++ == 0)
            this.isRollbackOnly = false;
        this.inTX = true;

        this.appCtx.getLog(LOG_SUBSYS).debug("tx is active");
    }

    /**
     * Gets the database connection associated with the transaction.
     *
     * @param dataSourceName logical name of the datasource.
     *
     * @return the database connection.
     *
     * @throws SQLException if the connection cannnot be retrieved due to the
     * underlying data source error.
     * @throws IllegalStateException if the transaction is not ative.
     * @throws UnsupportedOperationException if trying to get connection
     * from a different data source within the same transaction.
     */
    public synchronized Connection getConnection(String dataSourceName)
        throws SQLException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        Log log = this.appCtx.getLog(LOG_SUBSYS);

        if(this.con == null) {
            log.debug("opening transactional database connection");
            this.con = new TxConnection(this.appCtx.getNoTXConnection(dataSourceName));
            this.dataSourceName = dataSourceName;

        } else {
            if(!dataSourceName.equals(this.dataSourceName)) {
                throw new UnsupportedOperationException("Cannot request connection to a different data source" +
                                                        " within same transaction, this transaction context implementation" +
                                                        " does not support distributed transactions.");
            }
        }

        if(this.nTxsBegun < this.nestedLevel) {
            Statement stmt = this.con.getWrappedConnection().createStatement();
            while(this.nTxsBegun < this.nestedLevel) {
                stmt.execute(this.beginCommand);
                this.nTxsBegun++;
            }
            stmt.close();
        }

        return this.con;
    }

    /**
     * Commits the transaction.
     *
     * @return <code>true</code>.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not ative.
     */
    public synchronized boolean commit()
        throws ApplicationException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        Log log = this.appCtx.getLog(LOG_SUBSYS);

        try {
            if(this.nTxsBegun > 0) {
                Statement stmt = null;
                try {
                    stmt = this.con.getWrappedConnection().createStatement();

                    if(this.isRollbackOnly) {
                        log.debug("transaction is marked for rollback only, rolling transaction back");
                        stmt.execute(this.rollbackCommand);
                        log.debug("transaction rolled back");
                    } else {
                        log.debug("committing transaction");
                        stmt.execute(this.commitCommand);
                        log.debug("transaction committed");
                    }

                    this.nTxsBegun--;

                } finally {
                    if(stmt != null)
                        stmt.close(); 
                }
            }
        } catch(SQLException e) {
            throw new ApplicationException("Could not commit/rollback transaction.", e);

        } finally {
            if(--this.nestedLevel < this.nTxsBegun)
                this.nTxsBegun = this.nestedLevel;
            if(this.nestedLevel == 0) {
                this.inTX = false;
                if(this.con != null) {
                    log.debug("closing transactional database connection");
                    try {
                        this.con.getWrappedConnection().close();
                    } catch(SQLException e) {
                        log.error("Could not close connection, ignoring the error and just throwing the connection away.", e);
                    }
                    this.con = null;
                }
                this.appCtx.getLog(LOG_SUBSYS).debug("tx is inactive");
            }
        }

        return !this.inTX;
    }

    /**
     * Rollbacks the transaction.
     *
     * @return <code>true</code>.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not ative.
     */
    public synchronized boolean rollback()
        throws ApplicationException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        Log log = this.appCtx.getLog(LOG_SUBSYS);

        try {
            if(this.nTxsBegun > 0) {
                Statement stmt = null;
                try {
                    stmt = this.con.getWrappedConnection().createStatement();

                    log.debug("rolling transaction back");
                    stmt.execute(this.rollbackCommand);
                    log.debug("transaction rolled back");

                    this.nTxsBegun--;

                } finally {
                    if(stmt != null)
                        stmt.close(); 
                }
            }
        } catch(SQLException e) {
            throw new ApplicationException("Could not rollback transaction.", e);

        } finally {
            if(--this.nestedLevel < this.nTxsBegun)
                this.nTxsBegun = this.nestedLevel;
            if(this.nestedLevel == 0) {
                this.inTX = false;
                if(this.con != null) {
                    log.debug("closing transactional database connection");
                    try {
                        this.con.getWrappedConnection().close();
                    } catch(SQLException e) {
                        log.error("Could not close connection, ignoring the error and just throwing the connection away.", e);
                    }
                    this.con = null;
                }
                this.appCtx.getLog(LOG_SUBSYS).debug("tx is inactive");
            }
        }

        return !this.inTX;
    }

    /**
     * Marks the transaction for rollback only.
     */
    public synchronized void setRollbackOnly() {

        if(this.inTX) {
            this.appCtx.getLog(LOG_SUBSYS).debug("marking the current transaction for rollback only");
            this.isRollbackOnly = true;
        }
    }

    /**
     * Tells if the transaction is marked for rollback only.
     *
     * @return <code>true</code> if the transaction is marked for
     * rollback only. Always returns <code>false</code> if there no active
     * transaction.
     */
    public synchronized boolean isRollbackOnly() {

        return (this.inTX ? this.isRollbackOnly : false);
    }

    /**
     * Tells if the transaction is active.
     *
     * @return <code>true</code> if the transaction is active.
     */
    public synchronized boolean isActive() {

        return this.inTX;
    }

    /**
     * Tells if unbound transactions are supported.
     *
     * @return <code>true</code>.
     */
    public synchronized boolean supportsUnboundTransactions() {

        return true;
    }
}
