package com.boylesoftware.cb2;

import java.sql.PreparedStatement;
import java.sql.SQLException;


/**
 * Interface for DAO plugins that perform additional processing of query
 * parameters just before they get set into a prepared statement during a DAO
 * call.
 *
 * <p>Multiple <code>QueryParamProcessor</code> plugins can be installed for
 * one DAO. In this case they will be called in the order, in which their
 * <code>&lt;plugin&gt;</code> elements appear in the DAO configuration. Each
 * processor gets at its input parameter value returned by the previous one.
 *
 * <p>One instance of a <code>QueryParamProcessor</code> plugin can be shared
 * by multiple threads at the same time.
 *
 * @author Lev Himmelfarb
 * @version $Id: QueryParamProcessor.java,v 1.3 2004/04/20 16:00:58 levahim Exp $
 */
public interface QueryParamProcessor
    extends DAOPlugin {

    /**
     * This method is called by the DAO just before a query parameter gets set
     * into a <code>java.sql.PreparedStatement</code>. The method returns an
     * object, which will be actually used by the DAO and set into the
     * statement.
     *
     * <p>Note, that for array query parameters the method is called once for
     * the whole array and then one time for each of its element.
     *
     * <p>This method should not set the parameter by itself. The
     * <code>pstmt</code> argument is given to it only for reference.
     *
     * @param pstmt reference to the prepared statement object, into which the
     * parameter is going to be set.
     * @param placeholderIndex index of the placeholder in the prepared
     * statement, into which the parameter is going to be set.
     * @param queryParam the original value of the parameter. May be
     * <code>null</code>, an object, or an array of objects.
     *
     * @return value of the parameter to be actually set into the statement.
     *
     * @throws SQLException if an error happens.
     */
    public Object processQueryParam(PreparedStatement pstmt, int placeholderIndex, Object queryParam)
        throws SQLException;
}
