package com.boylesoftware.cb2;

import java.sql.PreparedStatement;
import java.sql.SQLException;


/**
 * Interface for the DAO plugin, which is responsible for setting query
 * parameters into JDBC prepared statements during a DAO call. The DAO does not
 * call this plugin for nulls, nor it calls it for arrays. In the case of nulls
 * the DAO always calls <code>setNull</code> on the
 * <code>java.sql.PreparedStatement</code> internally. For arrays, it
 * automatically breaks them down onto individual elements and then calls the
 * plugin for each element.
 *
 * <p>Only one plugin of this type can be installed in a DAO configuration.
 * If none installed, the DAO will use a simple internal implementation, which
 * calls <code>setObject</code> on the <code>java.sql.PreparedStatement</code>.
 *
 * <p>Note, that one instance of the plugin can be shared by multiple threads
 * at the same time.
 *
 * @author Lev Himmelfarb
 * @version $Id: QueryParamSetter.java,v 1.3 2004/04/20 16:00:58 levahim Exp $
 */
public interface QueryParamSetter
    extends DAOPlugin {

    /**
     * Sets the specified query parameter into the specified prepared
     * statement.
     *
     * @param pstmt the statement.
     * @param placeholderIndex the parameter index.
     * @param queryParam the value to set, never can be an array nor
     * <code>null</code>.
     * @param sqlType target SQL type (see <code>java.sql.Types</code>), or
     * <code>null</code> if not specified.
     *
     * @throws SQLException if an error happens.
     */
    public void setQueryParam(PreparedStatement pstmt,
                              int placeholderIndex,
                              Object queryParam,
                              Integer sqlType)
        throws SQLException;
}
