package com.boylesoftware.cb2;


/**
 * Interface for DAO plugins that perform additional processing of data from
 * result set columns just before they get set into DM fields during a DAO
 * fetch call.
 *
 * <p>Multiple <code>ResultSetFieldProcessor</code> plugins can be installed
 * for one DAO. In this case they will be called in the order, in which their
 * <code>&lt;plugin&gt;</code> elements appear in the DAO configuration. Each
 * processor gets at its input value returned by the previous one.
 *
 * <p>Note, that one instance of the processor can be shared by multiple
 * threads at the same time.
 *
 * @author Lev Himmelfarb
 * @version $Id: ResultSetFieldProcessor.java,v 1.4 2004/05/12 15:41:17 levahim Exp $
 */
public interface ResultSetFieldProcessor
    extends DAOPlugin {

    /**
     * This method is called by the DAO just before a value from a result set
     * column gets set into a DM field. The method returns an object, which
     * will be actually used by the DAO and set into the DM.
     *
     * <p>Note, that if this method returns an object, which has type
     * incomaptible with the target DM field, the DAO will throw an
     * <code>IllegalArgumentException</code> runtime exception when it tries to
     * actually set it.
     *
     * @param fieldName name of the target DM field.
     * @param fieldClass type of the target DM field. When the DM is a
     * <code>DynamicDataModel</code>, generic <code>Object</code> class is
     * passed from the DAO.
     * @param value the original value intended to be set into the DM field.
     *
     * @return value to be actually set into the DM field.
     *
     * @throws BLException if an unexpected error happens.
     */
    public Object processResultSetField(String fieldName,
                                        Class fieldClass,
                                        Object value)
        throws BLException;
}
