package com.boylesoftware.cb2;

import java.sql.Connection;
import java.sql.SQLException;

import org.apache.commons.logging.Log;


/**
 * Simple transaction context implementation that uses JDBC
 * <code>setAutoCommit</code>, <code>commit</code> and <code>rollback</code>
 * methods on <code>java.sql.Connection</code> object for transaction
 * management thus relying on JDBC driver implementation. This approach
 * requires binding database connection to the transaction context for
 * the whole duration of the transaction, guaranteeing that all commands
 * in the transaction are executed on the same connection. It is implemented
 * in this class transparently for the application.
 *
 * <p>This implementation does not support distributed transactions, nor
 * it does support nested transactions.
 *
 * <p>This is the default transaction context implementation used by the
 * application context unless it is overridden by specifying
 * <code>com.boylesoftware.cb2.txContext</code> application property.
 *
 * @author Lev Himmelfarb
 * @version $Id: SimpleTransactionContext.java,v 1.2 2004/10/11 18:17:36 levahim Exp $
 */
final class SimpleTransactionContext
    implements TransactionContext {

    /**
     * Reference to the application context.
     */
    private ApplicationContext appCtx;

    /**
     * Tells if the transaction is active.
     */
    private boolean inTX;

    /**
     * The database connection bound to the transaction context when the
     * transaction is active.
     */
    private TxConnection con;

    /**
     * Name of the data source, from which the <code>con</code> connection was
     * retrieved.
     */
    private String dataSourceName;

    /**
     * Original auto commit mode for the <code>con</code> connection before
     * if was set to <code>false</code> in order to start the transaction.
     */
    private boolean originalAutoCommit;

    /**
     * Tells if the transaction has been marked for rollback only.
     */
    private boolean isRollbackOnly;


    /**
     * Setup the instance.
     *
     * @param appCtx the application context.
     */
    public void setup(ApplicationContext appCtx) {

        this.appCtx = appCtx;
        this.inTX = false;
        this.con = null;
        this.isRollbackOnly = false;
    }

    /**
     * Begins a transaction.
     *
     * @throws ApplicationException if the transaction cannot be
     * started.
     */
    public synchronized void begin()
        throws ApplicationException {

        if(this.inTX)
            throw new ApplicationException("Already in a transaction, this transaction context" +
                                           " implementation does not allow nested transactions.");

        this.inTX = true;
        this.isRollbackOnly = false;

        this.appCtx.getLog(LOG_SUBSYS).debug("tx is active");
    }

    /**
     * Gets the database connection associated with the transaction.
     *
     * @param dataSourceName logical name of the datasource.
     *
     * @return the database connection.
     *
     * @throws SQLException if the connection cannnot be retrieved due to the
     * underlying data source error.
     * @throws IllegalStateException if the transaction is not ative.
     * @throws UnsupportedOperationException if trying to get connection
     * from a different data source within the same transaction.
     */
    public synchronized Connection getConnection(String dataSourceName)
        throws SQLException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        Log log = this.appCtx.getLog(LOG_SUBSYS);

        if(this.con == null) {
            log.debug("opening transactional database connection");
            Connection c = this.appCtx.getNoTXConnection(dataSourceName);
            this.originalAutoCommit = c.getAutoCommit();
            if(this.originalAutoCommit) {
                log.debug("current connection autocommit mode is true, setting it to false");
                c.setAutoCommit(false);
            } else {
                log.debug("current connection autocommit mode is already false, continuing within this existing trasaction");
            }
            this.con = new TxConnection(c);
            this.dataSourceName = dataSourceName;
        } else {
            if(!dataSourceName.equals(this.dataSourceName)) {
                throw new UnsupportedOperationException("Cannot request connection to a different data source" +
                                                        " within same transaction, this transaction context implementation" +
                                                        " does not support distributed transactions.");
            }
        }

        return this.con;
    }

    /**
     * Commits the transaction.
     *
     * @return <code>true</code>.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not ative.
     */
    public synchronized boolean commit()
        throws ApplicationException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        this.inTX = false;

        if(this.con != null) {
            Log log = this.appCtx.getLog(LOG_SUBSYS);
            Connection c = this.con.getWrappedConnection();
            try {

                if(this.isRollbackOnly) {
                    log.debug("transaction is marked for rollback only, rolling transaction back");
                    c.rollback();
                    log.debug("transaction rolled back");
                } else {
                    log.debug("committing transaction");
                    c.commit();
                    log.debug("transaction committed");
                }

                if(this.originalAutoCommit) {
                    log.debug("restoring connection autocommit mode back to true");
                    c.setAutoCommit(this.originalAutoCommit);
                } else {
                    log.debug("the original connection autocommit mode was false, leaving it as it is");
                }

            } catch(SQLException e) {
                throw new ApplicationException("Could not commit/rollback transaction, closing connection anyway and throwing exception.", e);

            } finally {
                this.con = null;
                log.debug("closing transactional database connection");
                try {
                    c.close();
                } catch(SQLException e) {
                    log.error("Could not close connection, ignoring the error and just throwing the connection away.", e);
                }
                this.appCtx.getLog(LOG_SUBSYS).debug("tx is inactive");
            }
        }

        return true;
    }

    /**
     * Rollbacks the transaction.
     *
     * @return <code>true</code>.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not ative.
     */
    public synchronized boolean rollback()
        throws ApplicationException {

        if(!this.inTX)
            throw new IllegalStateException("The transaction has not been begun.");

        this.inTX = false;

        if(this.con != null) {
            Log log = this.appCtx.getLog(LOG_SUBSYS);
            Connection c = this.con.getWrappedConnection();
            try {

                log.debug("rolling transaction back");
                c.rollback();
                log.debug("transaction rolled back");

                if(this.originalAutoCommit) {
                    log.debug("restoring connection autocommit mode back to true");
                    c.setAutoCommit(this.originalAutoCommit);
                } else {
                    log.debug("the original connection autocommit mode was false, leaving it as it is");
                }

            } catch(SQLException e) {
                throw new ApplicationException("Could not rollback transaction, closing connection anyway and throwing exception.", e);

            } finally {
                this.con = null;
                log.debug("closing transactional database connection");
                try {
                    c.close();
                } catch(SQLException e) {
                    log.error("Could not close connection, ignoring the error and just throwing the connection away.", e);
                }
                this.appCtx.getLog(LOG_SUBSYS).debug("tx is inactive");
            }
        }

        return true;
    }

    /**
     * Marks the transaction for rollback only.
     */
    public synchronized void setRollbackOnly() {

        if(this.inTX) {
            this.appCtx.getLog(LOG_SUBSYS).debug("marking the current transaction for rollback only");
            this.isRollbackOnly = true;
        }
    }

    /**
     * Tells if the transaction is marked for rollback only.
     *
     * @return <code>true</code> if the transaction is marked for
     * rollback only. Always returns <code>false</code> if there no active
     * transaction.
     */
    public synchronized boolean isRollbackOnly() {

        return (this.inTX ? this.isRollbackOnly : false);
    }

    /**
     * Tells if the transaction is active.
     *
     * @return <code>true</code> if the transaction is active.
     */
    public synchronized boolean isActive() {

        return this.inTX;
    }

    /**
     * Tells if unbound transactions are supported.
     *
     * @return <code>true</code>.
     */
    public synchronized boolean supportsUnboundTransactions() {

        return true;
    }
}
