package com.boylesoftware.cb2;

import java.text.DateFormat;
import java.text.SimpleDateFormat;

import com.boylesoftware.cb2.util.EventStatistics;


/**
 * Interface for objects that provide information about their internal status,
 * environment, and operation statistics. This interface is implemented by some
 * key internal CB2 objects that represent various subsystems, such as the BL
 * Manager, the Application Context and others, so important information about
 * their operation can be retrieved at runtime by some internal monitoring
 * subsystem.
 *
 * @author Lev Himmelfarb
 * @version $Id: StatusProvider.java,v 1.6 2004/04/30 18:54:08 levahim Exp $
 */
public interface StatusProvider {

    /**
     * Date format, which should be used by status provider implementators for
     * formating timestamps in the generated by the <code>getStats</code>
     * method XMLs.
     */
    static final DateFormat TIMESTAMP_FORMAT =
        new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");


    /**
     * Gets status/statistics report as an XML text. The structure of the XML
     * is specific to the implementator, but it is recommended to have one
     * top-level XML element and all other elements nested in it. The generated
     * XML should not contain the <code>&lt;?xml ...?&gt;</code> descriptor.
     * The idea is that it can be used by the caller as a fragment of a bigger
     * XML document, and the caller is responsible for such "finishing touches"
     * as adding an appropriate <code>&lt;?xml ...?&gt;</code> descriptor
     * including document encoding.
     *
     * <p>This method returns XML as a byte array. The characters in the byte
     * array <strong>must</strong> be in UTF-8 encoding.
     *
     * @return XML with status and statistics information.
     *
     * @throws Exception if an unexpected error happens.
     */
    byte [] getStats()
        throws Exception;

    /**
     * Gets statistics for the specified event. The object, implementing this
     * interface, may gather statistics for various internal events. This
     * method allows to access gathered data for a specified event, identified
     * by event name. Event names are free-form and are specific to the
     * particular status provider implementation. If the implementation does
     * not recognize the specified event name, it may return <code>null</code>.
     *
     * @param eventName implementation specific event name, for which current
     * statistics are requested.
     *
     * @return statistics information for the specified event or
     * <code>null</code> if the event is unknown.
     */
    EventStatistics getEventStats(String eventName);
}
