package com.boylesoftware.cb2;

import java.sql.Connection;
import java.sql.SQLException;


/**
 * Interface for transaction context implementations used by the application
 * context to manage database transactions. Application context reads
 * <code>com.boylesoftware.cb2.txContext</code> application property, which
 * names a class that implements this interface, and uses it as transaction
 * context with its transaction management methods.
 *
 * <p>When a new thread tries to perform a transactional database operation,
 * the application context checks if it already has a transaction context
 * object instance associated with it and if not, creates an instance of
 * transaction context and binds it to the thread. Right after the new
 * instance is created the application context calls <code>setup</code> on it
 * and then other methods are called according to the protocol. The instance
 * stays bound to the thread until the thread dies.
 *
 * <p>There are also transaction management methods in the application context
 * that operate on transaction contexts that are not bound to any thread.
 *
 * <p>Note, that implementations should be thread-safe providing their own
 * means of access synchronization. Multiple threads can be trying to call
 * different methods at the same time.
 *
 * <p>CB2 has several standard transaction context implementations. See
 * description of <code>com.boylesoftware.cb2.txContext</code> application
 * property in the <code>ApplicationContext</code> class description.
 *
 * @author Lev Himmelfarb
 * @version $Id: TransactionContext.java,v 1.7 2004/04/20 16:00:58 levahim Exp $
 */
public interface TransactionContext {

    /**
     * Channel name for the logger. Interface implementators should use it when
     * logging messages.
     */
    final static String LOG_SUBSYS = "com.boylesoftware.cb2.TXCTXT";


    /**
     * Setup the instance before using it. This method is called by the
     * application context once for every new instance.
     *
     * @param appCtx the application context.
     *
     * @throws ApplicationException if an error happens.
     */
    void setup(ApplicationContext appCtx)
        throws ApplicationException;

    /**
     * Begin a new transaction.
     *
     * @throws ApplicationException if an error happens, if there is already
     * an active transaction and the transaction context does not support
     * nested transactions.
     */
    void begin()
        throws ApplicationException;

    /**
     * Get a database connection managed by this transaction context.
     * This method can be called only between <code>begin</code> and
     * <code>commit</code> or <code>rollback</code> method calls, that is
     * when transaction is active.
     *
     * @param dataSourceName logical name of the datasource, from which
     * to get the connection.
     *
     * @return the database connection.
     *
     * @throws SQLException if the connection cannnot be retrieved due
     * to the underlying data source error.
     * @throws IllegalStateException if the transaction is not
     * ative. In other words if the transaction has not been begun or
     * is already committed or rolled back.
     */
    Connection getConnection(String dataSourceName)
        throws SQLException;

    /**
     * Commit the transaction.
     *
     * @return <code>true</code> if there are no active transactions anymore
     * after successful commit. Can return <code>false</code> if nested
     * transactions are supported and there are still active transactions
     * in the stack after committing the topmost one.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not
     * ative. In other words if the transaction has not been begun or
     * is already committed or rolledback.
     */
    boolean commit()
        throws ApplicationException;

    /**
     * Rollback the transaction.
     *
     * @return <code>true</code> if there are no active transactions anymore
     * after successful rollback. Can return <code>false</code> if nested
     * transactions are supported and there are still active transactions
     * in the stack after rolling back the topmost one.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     * @throws IllegalStateException if the transaction is not
     * ative. In other words if the transaction has not been begun or
     * is already committed or rolledback.
     */
    boolean rollback()
        throws ApplicationException;

    /**
     * Marks the transaction for rollback only. When the application tries
     * to commit a transaction which is marked for rollback only, the
     * transaction context will roll it back instead of committing.
     * Calling this method when there are no active transactions does nothing
     * and is perfectly valid.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     */
    void setRollbackOnly()
        throws ApplicationException;

    /**
     * Tells if the transaction is marked for rollback only.
     *
     * @return <code>true</code> if the transaction is marked for
     * rollback only. Always returns <code>false</code> if there are no
     * active transactions.
     *
     * @throws ApplicationException when the underlying subsystem
     * error happens.
     *
     * @see #setRollbackOnly
     */
    boolean isRollbackOnly()
        throws ApplicationException;

    /**
     * Tells if there is an active transaction at the moment of the call.
     *
     * @return <code>true</code> if the transaction is active.
     */
    boolean isActive();

    /**
     * Tells if this implementation supports transaction contexts that are
     * not bound to a thread.
     *
     * @return <code>true</code> if unbound transaction contexts can be
     * created.
     */
    boolean supportsUnboundTransactions();
}
