package com.boylesoftware.cb2;

import java.io.PrintWriter;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.sql.Connection;
import java.sql.SQLException;
import javax.sql.DataSource;

import com.boylesoftware.cb2.util.EventStatistics;
import com.boylesoftware.cb2.util.EventHistory;


/**
 * Universal database connection pool wrapper, which unifies interface
 * of different known database connection pool implementations. Currently
 * this class supports the following connection pool implementations:
 *
 * <ul>
 *
 * <li><a href="http://jakarta.apache.org/commons/dbcp">Jakarta Commons DBCP</a></li>
 *
 * <li><a href="http://www.caucho.com/resin">Caucho Resin DBPool</a></li>
 *
 * </ul>
 *
 * @author Lev Himmelfarb
 * @version $Id: UniversalDataSource.java,v 1.5 2004/10/11 18:17:36 levahim Exp $
 */
class UniversalDataSource
    implements DataSource, StatusProvider {

    /**
     * Type for uknown pool implementations.
     */
    public static final int TYPE_UNKNOWN = 0;

    /**
     * Type for Jakarta Commons DBCP pool implementation.
     */
    public static final int TYPE_DBCP = 1;

    /**
     * Type for Caucho Resin DBPool pool implementation.
     */
    public static final int TYPE_RESIN = 2;


    /**
     * The wrapped database connection pool.
     */
    private final DataSource ds;

    /**
     * The pool implementation type. One of <code>TYPE_*</code> constants.
     */
    private final int type;

    /**
     * Statistics for <code>getConnection</code> calls.
     */
    private final EventStatistics getConnectionStats;


    /**
     * Constructs a database connection pool wrapper.
     *
     * @param ds the pool to wrap.
     */
    public UniversalDataSource(DataSource ds) {

        this.ds = ds;

        boolean isApacheCommonsDBPoolAvail = false;
        try {
            Class.forName("org.apache.commons.dbcp.BasicDataSource");
            isApacheCommonsDBPoolAvail = true;
        } catch(ClassNotFoundException e) { /* ignore it */ }
        boolean isCauchoDBPoolAvail = false;
        try {
            Class.forName("com.caucho.sql.DBPool");
            isCauchoDBPoolAvail = true;
        } catch(ClassNotFoundException e) { /* ignore it */ }

        if(isApacheCommonsDBPoolAvail &&
           (ds instanceof org.apache.commons.dbcp.BasicDataSource))
            this.type = TYPE_DBCP;
        else if(isCauchoDBPoolAvail && (ds instanceof com.caucho.sql.DBPool))
            this.type = TYPE_RESIN;
        else
            this.type = TYPE_UNKNOWN;

        this.getConnectionStats = new EventStatistics(new EventHistory[] {
            new EventHistory(3456000L, 288),
            new EventHistory(24L * 3600000L, 288)
        });
    }


    //////////////////////////////////////////////////////////////////////////
    // Standard service
    //////////////////////////////////////////////////////////////////////////

    /**
     * Gets the undelying database connection pool type.
     *
     * @return one of <code>TYPE_*</code> constants.
     */
    public int getType() {

        return this.type;
    }

    /**
     * Gets the underlying data source.
     *
     * @return reference to the wrapped data source instance.
     */
    public DataSource getDataSource() {

        return this.ds;
    }

    /**
     * Closes connections in the undelying pool. If the pool has type
     * <code>TYPE_UNKNOWN</code> or does not support closing connections
     * this method does nothing.
     */
    public void close()
        throws SQLException {

        switch(this.type) {
        case TYPE_DBCP:
            ((org.apache.commons.dbcp.BasicDataSource)this.ds).close();
            break;
        case TYPE_RESIN:
            ((com.caucho.sql.DBPool)this.ds).close();
        }
    }


    //////////////////////////////////////////////////////////////////////////
    // DataSource interface
    //////////////////////////////////////////////////////////////////////////

    /**
     * Calls the underlying data source method.
     */
    public Connection getConnection()
        throws SQLException {

        this.getConnectionStats.eventStarted();

        return this.ds.getConnection();
    }

    /**
     * Calls the underlying data source method.
     */
    public Connection getConnection(String username, String password)
        throws SQLException {

        this.getConnectionStats.eventStarted();

        return this.ds.getConnection(username, password);
    }

    /**
     * Calls the underlying data source method.
     */
    public PrintWriter getLogWriter()
        throws SQLException {

        return this.ds.getLogWriter();
    }

    /**
     * Calls the underlying data source method.
     */
    public void setLogWriter(PrintWriter out)
        throws SQLException {

        this.ds.setLogWriter(out);
    }

    /**
     * Calls the underlying data source method.
     */
    public void setLoginTimeout(int seconds)
        throws SQLException {

        this.ds.setLoginTimeout(seconds);
    }

    /**
     * Calls the underlying data source method.
     */
    public int getLoginTimeout()
        throws SQLException {

        return this.ds.getLoginTimeout();
    }


    //////////////////////////////////////////////////////////////////////////
    // StatusProvider interface
    //////////////////////////////////////////////////////////////////////////

    /**
     * Gets pool info as an XML fragment.
     *
     * @return connection pool implementation specific XML, or an empty array
     * if the connection pool implementation is unknown.
     *
     * @throws UnsupportedEncodingException if UTF-8 is not supported on this
     * platform, which is practically impossible.
     */
    public byte [] getStats()
        throws UnsupportedEncodingException {

        ByteArrayOutputStream out = new ByteArrayOutputStream(4096);
        PrintStream res = new PrintStream(out, false, "UTF-8");

        switch(this.type) {
        case TYPE_DBCP:
            org.apache.commons.dbcp.BasicDataSource commonsDS =
                (org.apache.commons.dbcp.BasicDataSource)this.ds;
            res.println("<commonsDBCP>");
            res.print("  <autoCommit>"); res.print(commonsDS.getDefaultAutoCommit()); res.println("</autoCommit>");
            res.print("  <readOnly>"); res.print(commonsDS.getDefaultReadOnly()); res.println("</readOnly>");
            String txIsolationName = this.getTxIsolationLevelName(commonsDS.getDefaultTransactionIsolation());
            if(txIsolationName != null) {
                res.print("  <txIsolation>"); res.print(txIsolationName); res.println("</txIsolation>");
            }
            if(commonsDS.getDefaultCatalog() != null) {
                res.print("  <catalog><![CDATA["); res.print(commonsDS.getDefaultReadOnly()); res.println("]]></catalog>");
            }
            res.print("  <driverClassName>"); res.print(commonsDS.getDriverClassName()); res.println("</driverClassName>");
            res.print("  <maxActive>"); res.print(commonsDS.getMaxActive()); res.println("</maxActive>");
            res.print("  <maxIdle>"); res.print(commonsDS.getMaxIdle()); res.println("</maxIdle>");
            res.print("  <minIdle>"); res.print(commonsDS.getMinIdle()); res.println("</minIdle>");
            res.print("  <maxWait>"); res.print(commonsDS.getMaxWait()); res.println("</maxWait>");
            res.print("  <poolPS>"); res.print(commonsDS.isPoolPreparedStatements()); res.println("</poolPS>");
            res.print("  <maxOpenPSs>"); res.print(commonsDS.getMaxOpenPreparedStatements()); res.println("</maxOpenPSs>");
            res.print("  <testOnBorrow>"); res.print(commonsDS.getTestOnBorrow()); res.println("</testOnBorrow>");
            res.print("  <testOnReturn>"); res.print(commonsDS.getTestOnReturn()); res.println("</testOnReturn>");
            res.print("  <timeBetweenEvictionRuns>"); res.print(commonsDS.getTimeBetweenEvictionRunsMillis()); res.println("</timeBetweenEvictionRuns>");
            res.print("  <testsPerEvictionRun>"); res.print(commonsDS.getNumTestsPerEvictionRun()); res.println("</testsPerEvictionRun>");
            res.print("  <minEvictableIdleTime>"); res.print(commonsDS.getMinEvictableIdleTimeMillis()); res.println("</minEvictableIdleTime>");
            res.print("  <testWhileIdle>"); res.print(commonsDS.getTestWhileIdle()); res.println("</testWhileIdle>");
            res.print("  <url><![CDATA["); res.print(commonsDS.getUrl()); res.println("]]></url>");
            res.print("  <username><![CDATA["); res.print(commonsDS.getUsername()); res.println("]]></username>");
            if(commonsDS.getValidationQuery() != null) {
                res.print("  <valQuery><![CDATA["); res.print(commonsDS.getValidationQuery()); res.println("  ]]></valQuery>");
            }
            res.print("  <numActive>"); res.print(commonsDS.getNumActive()); res.println("</numActive>");
            res.print("  <numIdle>"); res.print(commonsDS.getNumIdle()); res.println("</numIdle>");
            res.println("</commonsDBCP>");
            break;

        case TYPE_RESIN:
            com.caucho.sql.DBPool resinDS =
                (com.caucho.sql.DBPool)this.ds;
            res.println("<resinDBPool>");
            res.print("  <name><![CDATA["); res.print(resinDS.getName()); res.println("]]></name>");
            res.print("  <driverName>"); res.print(resinDS.getDriverName()); res.println("</driverName>");
            res.print("  <url><![CDATA["); res.print(resinDS.getURL()); res.println("]]></url>");
            res.print("  <user><![CDATA["); res.print(resinDS.getUser()); res.println("]]></user>");
            res.print("  <connectionWaitTime>"); res.print(resinDS.getConnectionWaitTime()); res.println("</connectionWaitTime>");
            res.print("  <maxOverflowConnections>"); res.print(resinDS.getMaxOverflowConnections()); res.println("</maxOverflowConnections>");
            res.print("  <maxIdleTime>"); res.print(resinDS.getMaxIdleTime()); res.println("</maxIdleTime>");
            res.print("  <maxActiveTime>"); res.print(resinDS.getMaxActiveTime()); res.println("</maxActiveTime>");
            if(resinDS.getPingTable() != null) {
                res.print("  <pingTable><![CDATA["); res.print(resinDS.getPingTable()); res.println("]]></pingTable>");
            }
            res.print("  <pingOnReuse>"); res.print(resinDS.getPingOnReuse()); res.println("</pingOnReuse>");
            res.print("  <pingOnIdle>"); res.print(resinDS.getPingOnIdle()); res.println("</pingOnIdle>");
            res.print("  <pingInterval>"); res.print(resinDS.getPingInterval()); res.println("</pingInterval>");
            res.print("  <activeCons>"); res.print(resinDS.getActiveConnections()); res.println("</activeCons>");
            res.print("  <totalCons>"); res.print(resinDS.getTotalConnections()); res.println("</totalCons>");
            res.print("  <maxCons>"); res.print(resinDS.getMaxConnections()); res.println("</maxCons>");
            res.print("  <transactional>"); res.print(resinDS.isTransactional()); res.println("</transactional>");
            res.println("</resinDBPool>");
        }

        res.flush();
        return out.toByteArray();
    }

    /**
     * Gets name of a transaction isolation level.
     *
     * @param level one of <code>Connection.TRANSACTION_xxx</code> constants.
     *
     * @return the isolation level's name, or <code>null</code>.
     */
    private String getTxIsolationLevelName(int level) {

        switch(level) {
        case Connection.TRANSACTION_NONE:
            return "NONE";
        case Connection.TRANSACTION_READ_COMMITTED:
            return "READ COMMITTED";
        case Connection.TRANSACTION_READ_UNCOMMITTED:
            return "READ UNCOMMITTED";
        case Connection.TRANSACTION_REPEATABLE_READ:
            return "REPEATABLE READ";
        case Connection.TRANSACTION_SERIALIZABLE:
            return "SERIALIZABLE";
        default:
            return null;
        }
    }

    /**
     * Gets stats for an event. This method accepts the following events:
     *
     * <ul>
     *
     * <li><b>getConnection</b> - get connection from the data source
     * event. The returned stats object has also two history timelines:
     * last 24 hours and last hour.</li>
     *
     * </ul>
     */
    public EventStatistics getEventStats(String eventName) {

        if("getConnection".equals(eventName))
            return this.getConnectionStats;

        return null;
    }
}
