package com.boylesoftware.cb2.admin.servlet;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.nio.CharBuffer;
import java.io.ByteArrayInputStream;
import java.io.PrintWriter;
import java.io.InputStream;
import java.io.SequenceInputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.awt.image.BufferedImage;
import javax.servlet.UnavailableException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;
import javax.imageio.ImageIO;

import com.boylesoftware.cb2.StatusProvider;
import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.BLManager;
import com.boylesoftware.cb2.presentation.servlet.CB2Application;
import com.boylesoftware.cb2.presentation.servlet.CB2PlugIn;
import com.boylesoftware.cb2.util.HttpResponseUtils;
import com.boylesoftware.cb2.util.StatsGraph;
import com.boylesoftware.cb2.util.EventHistory;
import com.boylesoftware.cb2.util.TextUtils;


/**
 * Servlet, which displays the CB2 Application Status/Administration
 * page and processes adminitstrative actions submitted from it.
 *
 * <p>In order to access the servlet you must define it in the
 * <code>web.xml</code> file of your web-application and map it to a
 * URL ending with <code>/cb2-admin</code>. Below is an example:
 *
 * <pre>
 * ...
 * &lt;web-app&gt;
 *   ...
 *   &lt;servlet&gt;
 *     &lt;servlet-name&gt;admin&lt;/servlet-name&gt;
 *     &lt;display-name&gt;The CB2 Application Status/Admin Page&lt;/display-name&gt;
 *     &lt;servlet-class&gt;com.boylesoftware.cb2.admin.servlet.AdminServlet&lt;/servlet-class&gt;
 *     &lt;!-- load it AFTER the main controller servlet --&gt;
 *     &lt;load-on-startup&gt;15&lt;/load-on-startup&gt;
 *   &lt;/servlet&gt;
 *   ...
 *   &lt;servlet-mapping&gt;
 *     &lt;servlet-name&gt;admin&lt;/servlet-name&gt;
 *     &lt;url-pattern&gt;/cb2-admin&lt;/url-pattern&gt;
 *   &lt;/servlet-mapping&gt;
 *   ...
 * &lt;/web-app&gt;
 * </pre>
 *
 * <p>Note that it is very important that this servlet is loaded AFTER
 * the main action servlet, because in its initialization it counts on
 * the fact that the application context and all subsystems are already
 * initialized.
 *
 * <p>Note also, that it is usually a good idea to restrict access to this
 * servlet, especially in production environments.
 *
 * @author Lev Himmelfarb
 * @version $Id: AdminServlet.java,v 1.13 2004/10/11 18:17:36 levahim Exp $
 */
public final class AdminServlet
    extends HttpServlet {

    /**
     * Date format for timezone info.
     */
    private static final DateFormat TZ_DF =
        new SimpleDateFormat("z ('GMT'Z)");

    /**
     * Name of session attribute with the last data for a subsystem.
     */
    private static final String LAST_DATA_PREFIX =
        "com.boylesoftware.cb2.admin.servlet.LAST_DATA.";

    /**
     * Header for stats XMLs.
     */
    private static final byte [] XML_HEADER;
    static {
        try {
            XML_HEADER =
                "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n".getBytes("UTF-8");
        } catch(UnsupportedEncodingException e) {
            throw new Error("UTF-8 is not supported on this platform.");
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Internal Data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Describes a subsystem.
     */
    private class SubsystemDescriptor {

        /**
         * Reference to the status provider.
         */
        public StatusProvider provider;

        /**
         * Map with info pages descriptors. Keys are info names (objects of
         * <code>String</code>), values are info descriptors (objects of
         * <code>InfoDescriptor</code>).
         */
        public final Map infos;


        /**
         * Constructs a descriptor.
         *
         * @param provider reference to the status provider.
         * @param infos map with info pages descriptors.
         */
        public SubsystemDescriptor(StatusProvider provider,
                                   Map infos) {

            this.provider = provider;
            this.infos = infos;
        }
    }


    /**
     * Describes a info page within a subsystem.
     */
    private class InfoDescriptor {

        /**
         * XSL file name.
         */
        public final String xslFile;

        /**
         * Default <code>sortBy</code> paramater value or <code>null</code>.
         */
        public final String defaultSortBy;

        /**
         * Specifies supplementary parameters names (same
         * in the HTTP request and for the XSL).
         */
        public final String [] supplParamNames;

        /**
         * Tells if we can use last fetched data for this info.
         */
        public final boolean useLastData;


        /**
         * Constructs a descriptor.
         *
         * @param xslFile XSL file name.
         * @param defaultSortBy default <code>sortBy</code> parameter.
         * @param supplParamNames supplementary parameters names.
         * @param useLastData tells if we can use last fetched data.
         */
        public InfoDescriptor(String xslFile,
                              String defaultSortBy,
                              String [] supplParamNames,
                              boolean useLastData) {

            this.xslFile = xslFile;
            this.defaultSortBy = defaultSortBy;
            this.supplParamNames = supplParamNames;
            this.useLastData = useLastData;
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Servlet Data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Prefix for window title, which normally contains the host name
     * (or address). Can be an empty string.
     */
    private String windowTitlePrefix;

    /**
     * XSL transformer factory.
     */
    private TransformerFactory transformerFactory;

    /**
     * Contains compiled templates. Keys are XSL file names (objects
     * of <code>String</code>), values are compiled templates (objects
     * of <code>Templates</code>).
     */
    private Map templates;

    /**
     * Collection of subsystem descriptors. Keys are subsystem names
     * (objects of <code>String</code>), values are subsystem descriptors
     * (objects of <code>SubsystemDescriptor</code>).
     */
    private Map subsystems;

    /**
     * Cluster controller instance.
     */
    private ClusterController cluster;


    ///////////////////////////////////////////////////////////////////////////
    // Servlet Interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Initializes the servlet.
     */
    public void init()
        throws UnavailableException {

        // find CB2 application environment
        CB2Application app = CB2PlugIn.getApplication(this.getServletContext());
        if(app == null)
            throw new UnavailableException("Could not find CB2 application environment," +
                                           " probably the main action servlet is not loaded.");

        // install cluster controller
        this.cluster = new ClusterController(app);

        // form window title prefix
        this.windowTitlePrefix = "";
        try {
            InetAddress localHost = InetAddress.getLocalHost();
            this.windowTitlePrefix = localHost.getHostName() + " - ";
        } catch(UnknownHostException e) { /* ignore it */ }

        // setup basic admin pages

        this.transformerFactory = TransformerFactory.newInstance();
        this.templates = new HashMap();

        this.subsystems = new HashMap();

        Map infos = new HashMap(12);
        infos.put("jvmInfo",
            new InfoDescriptor("jvminfo.xsl",           null,        null,                         false));
        infos.put("sysProps",
            new InfoDescriptor("sysprops.xsl",          null,        null,                         false));
        infos.put("sysPropValue",
            new InfoDescriptor("syspropvalue.xsl",      null,        new String [] { "propName" }, true));
        infos.put("appProps",
            new InfoDescriptor("appprops.xsl",          null,        null,                         false));
        infos.put("appPropValue",
            new InfoDescriptor("apppropvalue.xsl",      null,        new String [] { "propName" }, true));
        infos.put("clChain",
            new InfoDescriptor("clchain.xsl",           null,        null,                         false));
        infos.put("jndi",
            new InfoDescriptor("jndi.xsl",              null,        null,                         false));
        infos.put("jndiValue",
            new InfoDescriptor("jndivalue.xsl",         null,        new String [] { "jndiName" }, true));
        infos.put("dbInfo",
            new InfoDescriptor("dbinfo.xsl",            null,        null,                         false));
        infos.put("dbConPools",
            new InfoDescriptor("dbconpools.xsl",        null,        null,                         false));
        infos.put("activeStmts",
            new InfoDescriptor("activestmts.xsl",       null,        null,                         false));
        infos.put("activeStmtDetails",
            new InfoDescriptor("activestmtdetails.xsl", null,        new String [] { "stmtId" },   true));
        this.subsystems.put("appCtx", new SubsystemDescriptor(null, infos));

        infos = new HashMap(5);
        infos.put("blos",
            new InfoDescriptor("blos.xsl",              "name",      null,                         false));
        infos.put("bloInitParams",
            new InfoDescriptor("bloinitparams.xsl",     null,        new String [] { "bloName" },  false));
        infos.put("blSessions",
            new InfoDescriptor("blsessions.xsl",        null,        null,                         false));
        infos.put("queries",
            new InfoDescriptor("queries.xsl",           "name",      null,                         false));
        infos.put("queryText",
            new InfoDescriptor("querytext.xsl",         null,        new String [] { "dsName", "n" }, true));
        this.subsystems.put("blMgr", new SubsystemDescriptor(null, infos));

        infos = new HashMap(6);
        infos.put("servletInfo",
            new InfoDescriptor("servletinfo.xsl",       null,        null,                         false));
        infos.put("servletAttrs",
            new InfoDescriptor("servletattrs.xsl",      null,        null,                         false));
        infos.put("servletAttrValue",
            new InfoDescriptor("servletattrvalue.xsl",  null,        new String [] { "attrName" }, true));
        infos.put("userSessions",
            new InfoDescriptor("usersessions.xsl",      "createdOn", null,                         false));
        infos.put("lastErrors",
            new InfoDescriptor("lasterrors.xsl",        null,        new String [] { "sessionId" }, true));
        infos.put("requests",
            new InfoDescriptor("requests.xsl",          null,        null,                         false));
        this.subsystems.put("ctrlr", new SubsystemDescriptor(null, infos));

        infos = new HashMap(1);
        infos.put("clusterMembers",
            new InfoDescriptor("clusterMembers.xsl",    null,        null,                         false));
        this.subsystems.put("cluster", new SubsystemDescriptor(null, infos));
    }

    /**
     * Called when the servlet goes out of service.
     */
    public void destroy() {

        this.templates.clear();
        this.templates = null;
        this.transformerFactory = null;
        this.subsystems.clear();
        this.cluster = null;
    }

    /**
     * Handles a <code>GET</code> request.
     *
     * @param request the HTTP request.
     * @param response the HTTP response.
     *
     * @throws ServletException if an error happened during the
     * request processing.
     * @throws IOException if an input/output error has happened
     * during the request processing.
     */
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException {

        // get basic subsystems
        CB2Application app = CB2PlugIn.getApplication(this.getServletContext());
        if(app == null)
            throw new ServletException("Could not find CB2 application environment," +
                                       " probably the main action servlet is not loaded.");
        BLManager blManager = app.getBLManager();
        ApplicationContext appCtx = app.getApplicationContext();

        // setup cluster controller
        this.cluster.setLocalAdminServletURL(request);

        // update basic subsys descriptors
        ((SubsystemDescriptor)this.subsystems.get("appCtx")).provider = appCtx;
        ((SubsystemDescriptor)this.subsystems.get("blMgr")).provider = blManager;
        ((SubsystemDescriptor)this.subsystems.get("ctrlr")).provider = app;
        ((SubsystemDescriptor)this.subsystems.get("cluster")).provider = this.cluster;

        // action or display?
        String type = request.getParameter("action");
        if(type != null) {
            this.processAction(request,
                               response,
                               appCtx,
                               blManager,
                               app);
            return;
        }
        type = request.getParameter("disp");
        String subsys = request.getParameter("ss");
        String compName = request.getParameter("name");
        String sortBy = request.getParameter("sortBy");

        try {

            // analyze type
            if("info".equals(type)) {

                // check basic parameters
                if((subsys == null) || (compName == null))
                    throw new ServletException("Invalid request parameters.");

                // find requested info descriptors
                SubsystemDescriptor subsysDesc = (SubsystemDescriptor)this.subsystems.get(subsys);
                if(subsysDesc == null)
                    throw new ServletException("Invalid ss request parameter.");
                InfoDescriptor  infoDesc = (InfoDescriptor)subsysDesc.infos.get(compName);
                if(infoDesc == null)
                    throw new ServletException("Invalid name request parameter.");

                // set additional parameters
                Map params = new HashMap();
                params.put("timezone", TZ_DF.format(new java.util.Date()));
                if(sortBy == null)
                    sortBy = infoDesc.defaultSortBy;
                if(sortBy != null)
                    params.put("sortBy", sortBy);
                if(infoDesc.supplParamNames != null) {
                    for(int i = 0; i < infoDesc.supplParamNames.length; i++) {
                        String paramValue = request.getParameter(infoDesc.supplParamNames[i]);
                        if(paramValue == null)
                            throw new ServletException("Invalid number of parameters.");
                        params.put(infoDesc.supplParamNames[i], paramValue);
                    }
                }

                // adjust response
                response.setContentType("text/html; charset=ISO-8859-1");
                HttpResponseUtils.setNoCache(response);

                // get data
                byte [] data = null;
                if(infoDesc.useLastData)
                    data = (byte [])request.getSession().
                        getAttribute(LAST_DATA_PREFIX + subsys);
                if(data == null) {
                    data = subsysDesc.provider.getStats();
                    request.getSession().setAttribute(LAST_DATA_PREFIX + subsys, data);
                }

                // send the response
                this.sendTransformed(response.getWriter(), infoDesc.xslFile, data, params);

            } else if("css".equals(type)) {

                response.setContentType("text/css; charset=ISO-8859-1");
                this.sendFile(response.getWriter(), "stylesheet.css", null);

            } else if("pic".equals(type)) {

                // get picture index
                String indS = request.getParameter("ind");
                if(indS == null)
                    throw new ServletException("Invalid request parameters.");
                int ind;
                try {
                    ind = Integer.parseInt(indS);
                } catch(NumberFormatException e) {
                    throw new ServletException("Invalid ind request parameter.");
                }

                // get common max sample
                String maxSampleS = request.getParameter("maxSample");
                long maxSample;
                if(maxSampleS == null)
                    maxSample = -1L;
                else {
                    try {
                        maxSample = Long.parseLong(maxSampleS);
                    } catch(NumberFormatException e) {
                        throw new ServletException("Invalid maxSample request parameter.");
                    }
                }

                // check basic parameters
                if((subsys == null) || (compName == null))
                    throw new ServletException("Invalid request parameters.");

                // find requested subsys descriptor
                SubsystemDescriptor subsysDesc = (SubsystemDescriptor)this.subsystems.get(subsys);
                if(subsysDesc == null)
                    throw new ServletException("Invalid ss request parameter.");

                // picture size and data
                EventHistory data = subsysDesc.provider.getEventStats(compName).getTimelines()[ind];
                BufferedImage image;
                String size = request.getParameter("size");
                if("S".equals(size)) {
                    image = StatsGraph.generateTinyImage(data, 60, 10, maxSample);
                } else if("N".equals(size)) {
                    image = StatsGraph.generateImage(data, 288, 80);
                } else
                    throw new ServletException("Invalid size request parameter.");

                // adjust the response
                response.setContentType("image/png");
                response.setHeader("Cache-Control", "no-cache");
                response.setDateHeader("Expires", System.currentTimeMillis());

                // write the image data
                ImageIO.write(image, "PNG", response.getOutputStream());

            } else if("xml".equals(type)) {

                // check basic parameters
                if(subsys == null)
                    throw new ServletException("Invalid request parameters.");

                // find requested subsys descriptor
                SubsystemDescriptor subsysDesc = (SubsystemDescriptor)this.subsystems.get(subsys);
                if(subsysDesc == null)
                    throw new ServletException("Invalid ss request parameter.");

                // adjust response
                response.setContentType("text/xml; charset=ISO-8859-1");
                HttpResponseUtils.setNoCache(response);

                // send the response
                response.getWriter().print(subsysDesc.provider.getStats());

            } else { // html

                // adjust response
                response.setContentType("text/html; charset=ISO-8859-1");

                // what component was requested?
                if((compName == null) || (compName.length() == 0) || compName.equals("main")) {
                    Map params = new HashMap(1);
                    params.put("windowTitlePrefix", this.windowTitlePrefix);
                    this.sendFile(response.getWriter(), "frameset.html", params);

                } else if(compName.equals("nav")) {
                    this.sendFile(response.getWriter(), "nav.html", null);

                } else if(compName.equals("header")) {
                    Map params = new HashMap(2);
                    params.put("currentTimeMillis", String.valueOf(System.currentTimeMillis()));
                    params.put("Lock/Unlock", (app.isLocked() ? "Unlock" : "Lock"));
                    this.sendFile(response.getWriter(), "header.html", params);
                } else
                    throw new ServletException("Invalid request parameters.");
            }

        } catch(Exception e) {
            if(e instanceof ServletException)
                throw (ServletException)e;
            else if(e instanceof IOException)
                throw (IOException)e;
            else
                throw new ServletException("Got Error.", e);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Actions
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Checks if there is an "action" parameter in the request and if
     * so performs the requested action. After that it redirects to
     * the same page from which the action was requested and returns
     * <code>true</code>. If no action was requested then the method
     * returns <code>false</code>.
     *
     * @param request the HTTP request.
     * @param response the HTTP response.
     * @param appCtx the application context.
     * @param blManager the BL manager.
     * @param app the <code>CB2Application</code> instance.
     *
     * @throws ServletException if an error happened during the
     * request processing.
     * @throws IOException if an input/output error has happened
     * during the request processing.
     */
    private void processAction(HttpServletRequest request,
                               HttpServletResponse response,
                               ApplicationContext appCtx,
                               BLManager blManager,
                               CB2Application app)
        throws ServletException, IOException {

        String action = request.getParameter("action");

        try {

            if(action.equals("ReconfDAL")) {
                blManager.reconfigureDAL(null);

            } else if(action.equals("InitApp")) {
                appCtx.reconfigure(appCtx.getConfigDirPath());

            } else if(action.equals("CancelStmt")) {
                String stmtIdS = request.getParameter("stmtId");
                if(stmtIdS == null)
                    throw new ServletException("Required parameter stmtId is missing.");
                appCtx.cancelStatement(new Long(stmtIdS));

            } else if(action.equals("CancelAllStmts")) {
                appCtx.cancelAllStatements();

            } else if(action.equals("LockApp")) {
                app.setLock(!app.isLocked());

            } else if(action.equals("ClusterCancelAllStmts")) {
                this.cluster.sendClusterCommand(ClusterController.CR_CANCEL_ALL_STMTS);

            } else if(action.equals("ClusterLockApp")) {
                this.cluster.sendClusterCommand(ClusterController.CR_LOCK);

            } else {
                throw new ServletException("Invalid action [" + action + "]");
            }
        } catch(Exception e) {
            if(e instanceof ServletException)
                throw (ServletException)e;

            throw new ServletException("Could not perform the requested action.", e);
        }

        response.sendRedirect(request.getRequestURI());
    }


    ///////////////////////////////////////////////////////////////////////////
    // Internal Service
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Sends contents of the specified resource file to the output.
     *
     * @param out the response output stream.
     * @param fileName name of the file to send without a path. The
     * file is looked up in the directory of the class.
     * @param params optional parameters for replacement.
     *
     * @throws IOException if an input/output error has happened
     * writing to the output.
     */
    private void sendFile(PrintWriter out, String fileName, Map params)
        throws IOException {

        char [] buffer = new char[4096];

        InputStreamReader in = null;
        try {
            in = new InputStreamReader(this.getClass().getResourceAsStream(fileName));

            int p = 0;
            int n;
            while((n = in.read(buffer, p, buffer.length - p)) >= 0) {
                if((p += n) >= buffer.length)
                    throw new IOException("The input file " + fileName + " is too big.");
            }

            if(params != null)
                out.write(TextUtils.replaceMacros(CharBuffer.wrap(buffer, 0, p),
                                                  new TextUtils.MapValueGetter(params)));
            else
                out.write(buffer, 0, p);

        } finally {
            if(in != null)
                in.close();
        }
    }

    /**
     * Transforms XML data through an XSL template and sends the
     * result to the output.
     *
     * @param out the response output stream.
     * @param templateFileName name of the XSL template file without a
     * path. The file is looked up in the directory of the class.
     * @param data the XML data to be transformed.
     * @param params additional template parameters.
     *
     * @throws ServletException if an error has happened.
     * @throws IOException if an input/output error has happened
     * writing to the output.
     */
    private void sendTransformed(PrintWriter out,
                                 String templateFileName,
                                 byte [] data,
                                 Map params)
        throws ServletException, IOException {

        try {
            Transformer transformer = this.getTransformer(templateFileName);
            if(params != null) {
                Iterator i = params.entrySet().iterator();
                while(i.hasNext()) {
                    Map.Entry entry = (Map.Entry)i.next();
                    transformer.setParameter((String)entry.getKey(), entry.getValue());
                }
            }
            transformer.
                transform(new StreamSource(new SequenceInputStream(new ByteArrayInputStream(XML_HEADER),
                                                                   new ByteArrayInputStream(data))),
                          new StreamResult(out));

        } catch(TransformerConfigurationException e) {
            throw new ServletException("Got XSL transformer configuration exception.", e);

        } catch(TransformerException e) {
            throw new ServletException("Got XSL transformer exception.", e);
        }
    }

    /**
     * Gets a transformer instance for the specified XSL file.
     *
     * @param templateFileName name of the XSL template file without a
     * path. The file is looked up in the directory of the class.
     *
     * @return transformer to be used for transformations by the
     * specified XSL template.
     *
     * @throws IOException if an input/output error has happened
     * reading the XSL template resource.
     * @throws TransformerConfigurationException if a transformer
     * cannot be created.
     */
    private Transformer getTransformer(String templateFileName)
        throws IOException, TransformerConfigurationException {

        Templates t;
        synchronized(this.templates) {
            t = (Templates)this.templates.get(templateFileName);
            if(t == null) {
                InputStream in = null;
                try {
                    in = this.getClass().getResourceAsStream(templateFileName);
                    t = this.transformerFactory.newTemplates(new StreamSource(in));
                    this.templates.put(templateFileName, t);
                } finally {
                    if(in != null)
                        in.close();
                }
            }
        }

        return t.newTransformer();
    }
}
