package com.boylesoftware.cb2.admin.servlet;

import java.io.Serializable;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Date;
import java.net.InetAddress;
import java.net.UnknownHostException;
import javax.servlet.http.HttpServletRequest;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.BroadcastMessagingTerminal;
import com.boylesoftware.cb2.StatusProvider;
import com.boylesoftware.cb2.presentation.servlet.CB2Application;
import com.boylesoftware.cb2.util.EventStatistics;
import com.boylesoftware.cb2.util.UniqueIdGenerator;


/**
 * This singleton provides services for getting information about
 * the cluster to the admin servet.
 *
 * @author Lev Himmelfarb
 * @version $Id: ClusterController.java,v 1.7 2004/10/11 18:17:36 levahim Exp $
 */
class ClusterController
    implements BroadcastMessagingTerminal, StatusProvider {

    /**
     * Name of the BCM channel used by the controller for sending requests
     * to cluster members.
     */
    private static final String BCM_CHANNEL_REQUEST = "com.boylesoftware.cb2.admin.servlet.ClusterController.REQUEST";

    /**
     * Name of the BCM channel used by the controller for sending responses
     * with information about local instance.
     */
    private static final String BCM_CHANNEL_RESPONSE = "com.boylesoftware.cb2.admin.servlet.ClusterController.RESPONSE";

    /**
     * Give this amount of milliseconds to cluster members to respond after
     * a request has been sent.
     */
    private static final long WAIT_FOR_RESPONSE_MILLIS = 5000L;

    // cluster commands

    /**
     * List cluster members.
     */
    private static final int CR_LIST     = 1;

    /**
     * Cancel all active database statements.
     */
    static final int CR_CANCEL_ALL_STMTS = 4;

    /**
     * Lock the application.
     */
    static final int CR_LOCK             = 5;


    /**
     * Reference to our CB2 application.
     */
    private final CB2Application app;

    /**
     * Reference to our application context.
     */
    private final ApplicationContext appCtx;

    /**
     * Generator of request keys.
     */
    private final UniqueIdGenerator requestKeyGen;

    /**
     * Indicates it we are collecting responses from cluster members at
     * this moment.
     */
    private boolean receiving;

    /**
     * Synchronization object for the <code>receiving</code> flag access.
     */
    private final Object receivingLock = new Object();

    /**
     * Request key for the request we just sent and are waiting for
     * responses.
     */
    private String currentRequestKey;

    /**
     * Collected messages (objects of <code>ClusterControllerMessage</code>)
     * from cluster members.
     */
    private final List membersInfo;

    /**
     * URL of admin servlet for the local instance.
     */
    private String localAdminServletURL;

    /**
     * Synchronization object for <code>localAdminServletURL</code> access.
     */
    private final Object localAdminServletURLLock = new Object();

    /**
     * Local host name.
     */
    private final String localHostName;

    /**
     * Local host IP address.
     */
    private final String localHostAddress;

    /**
     * Timestamp when this instance was started.
     */
    private final Date startedTs;


    /**
     * Creates and initializes an instance of cluster controller.
     *
     * @param app reference to our <code>CB2Application</code>.
     */
    public ClusterController(CB2Application app) {

        this.app = app;
        this.appCtx = app.getApplicationContext();
        this.requestKeyGen = new UniqueIdGenerator();
        this.receiving = false;
        this.membersInfo = new ArrayList();
        this.localAdminServletURL = null;
        this.startedTs = new Date();

        InetAddress localHost = null;
        try {
            localHost = InetAddress.getLocalHost();
        } catch(UnknownHostException e) { /* ignore */ }
        this.localHostName = (localHost != null ? localHost.getHostName() : "unknown");
        this.localHostAddress = (localHost != null ? localHost.getHostAddress() : "unknown");

        this.appCtx.registerBCMTerminal(BCM_CHANNEL_REQUEST, this);
        this.appCtx.registerBCMTerminal(BCM_CHANNEL_RESPONSE, this);
    }

    /**
     * Analizes a request to the admin servlet and if required sets
     * the <code>localAdminServletURL</code> feild.
     *
     * @param adminServletRequest request to the admin servlet.
     */
    void setLocalAdminServletURL(HttpServletRequest adminServletRequest) {

        synchronized(this.localAdminServletURLLock) {
            if(this.localAdminServletURL == null) {
                this.localAdminServletURL = adminServletRequest.getRequestURL().toString();
            }
        }
    }

    /**
     * Sends a command to all cluster members.
     *
     * @param command one of <code>CR_xxx</code> constants.
     */
    void sendClusterCommand(int command) {

        this.appCtx.broadcastBCMMessage(BCM_CHANNEL_REQUEST,
                                        new ClusterControllerMessage(null, command));
    }

    /**
     * Processes requests and responses. If it is a request the method
     * broadcasts information about the local instance. If it is a response
     * and a request was just sent then the received information is
     * added to the list of known cluster members.
     *
     * @param channelName BCM channel name, can be either the request channel
     * or response channel.
     * @param source information about the source of the message, ignored by
     * this method.
     * @param messageParam instance of <code>ClusterControllerMessage</code>.
     */
    public void processMessage(String channelName, String source, Serializable messageParam) {

        ClusterControllerMessage msg = (ClusterControllerMessage)messageParam;

        if(channelName.equals(BCM_CHANNEL_REQUEST)) {

            switch(msg.type) {
            case CR_CANCEL_ALL_STMTS:
                this.appCtx.cancelAllStatements();
                break;
            case CR_LOCK:
                this.app.setLock(!this.app.isLocked());
                break;
            default: // CR_LIST
                msg.hostName = this.localHostName;
                msg.hostAddress = this.localHostAddress;
                msg.adminServletURL = this.localAdminServletURL;
                msg.nSessions = this.app.getNumberOfSessions();
                msg.requestsPerMinute = this.app.getEventStats("requests").getEventsPerMinuteAvg();
                msg.startedTs = this.startedTs;
                msg.nErrors = this.app.getNumberOfErrors();
                msg.isLocked = this.app.isLocked();
                this.appCtx.broadcastBCMMessage(BCM_CHANNEL_RESPONSE, msg);
            }

        } else {

            synchronized(this.receivingLock) {
                if(this.receiving && msg.key.equals(this.currentRequestKey)) {
                    this.membersInfo.add(msg);
                }
            }
        }
    }

    /**
     * Forms XML with information about cluster and its members. This method
     * first broadcasts a request, then wait a certain amount of time giving
     * chance to all cluster members to respond and then builds XML basing
     * on the received information.
     *
     * @return cluster info XML.
     *
     * @throws UnsupportedEncodingException if UTF-8 is not supported on this
     * platform, which is practically impossible.
     */
    public synchronized byte [] getStats()
        throws UnsupportedEncodingException {

        // generate new request key
        this.currentRequestKey = this.requestKeyGen.next();

        // open for receiving
        synchronized(this.receivingLock) {
            this.receiving = true;
        }

        // send request
        this.appCtx.broadcastBCMMessage(BCM_CHANNEL_REQUEST,
                                        new ClusterControllerMessage(this.currentRequestKey,
                                                                     CR_LIST));

        // give cluster members time to respond
        try {
            Thread.sleep(WAIT_FOR_RESPONSE_MILLIS);
        } catch(InterruptedException e) { /* ignore it */ }

        // close receiving
        synchronized(this.receivingLock) {
            this.receiving = false;
        }

        // build XML from the collected data
        ByteArrayOutputStream out = new ByteArrayOutputStream(1024);
        PrintStream res = new PrintStream(out, false, "UTF-8");
        res.println("<stats>");
        res.println("  <members>");
        Iterator i = this.membersInfo.iterator();
        int n = 0;
        while(i.hasNext()) {
            ClusterControllerMessage msg = (ClusterControllerMessage)i.next();
            res.print("    <member n=\""); res.print(n++); res.println("\">");
            res.print("      <hostName>"); res.print(msg.hostName); res.println("</hostName>");
            res.print("      <hostAddress>"); res.print(msg.hostAddress); res.println("</hostAddress>");
            if(msg.adminServletURL != null) {
                res.print("      <adminServletURL>"); res.print(msg.adminServletURL); res.println("</adminServletURL>");
            }
            res.print("      <nSessions>"); res.print(msg.nSessions); res.println("</nSessions>");
            res.print("      <requestsPerMinute>"); res.print(msg.requestsPerMinute); res.println("</requestsPerMinute>");
            res.print("      <startedOn>"); res.print(StatusProvider.TIMESTAMP_FORMAT.format(msg.startedTs)); res.println("</startedOn>");
            res.print("      <nErrors>"); res.print(msg.nErrors); res.println("</nErrors>");
            res.print("      <isLocked>"); res.print(msg.isLocked ? "yes" : "no"); res.println("</isLocked>");
            res.println("    </member>");
        }
        res.println("  </members>");
        res.println("</stats>");

        // clear collected info
        this.membersInfo.clear();

        // done
        res.flush();
        return out.toByteArray();
    }

    /**
     * Does nothing in this implementation.
     *
     * @param eventName ignored.
     *
     * @return <code>null</code>.
     */
    public EventStatistics getEventStats(String eventName) {

        return null;
    }
}
