package com.boylesoftware.cb2.auth.servlet.resin;

import java.io.PrintStream;
import java.security.Principal;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Set;
import java.util.HashSet;
import javax.sql.DataSource;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.caucho.http.security.AbstractAuthenticator;

import com.boylesoftware.cb2.util.PrincipalWithRoles;


/**
 * Authenticator implementation for Caucho Resin servlet engine,
 * which uses database as the user accounts storage.
 *
 * <p>Resin already has an authenticator implementation called
 * <code>com.caucho.http.security.JdbcAuthenticator</code>. However,
 * that imlpementation is virtually unusable since it queries
 * the database for each <code>isUserInRole</code> call, which is
 * exteremely inefficient. Our implementation fetches the whole
 * list of user roles at the user login and caches it in the
 * session. It also means that any updates to a user roles will
 * be reflected only with the next login.
 *
 * <p>In order to use this authenticator for servlet container
 * managed authentication you should add it to the
 * <code>resin.conf</code> file. Here is an example:
 *
 * <pre>
 * &lt;caucho.com&gt;
 *
 *   ...
 *
 *   &lt;resource-ref&gt;
 *     &lt;res-ref-name&gt;jdbc/dataSource&lt;/res-ref-name&gt;
 *     &lt;res-type&gt;javax.sql.DataSource&lt;/res-type&gt;
 *     ...
 *   &lt;/resource-ref&gt;
 *
 *   ...
 *
 *   &lt;http-server&gt;
 *
 *     ...
 *
 *     &lt;host id=""&gt;
 *
 *       ...
 *
 *       &lt;authenticator id="com.boylesoftware.cb2.auth.servlet.resin.JdbcAuthenticator"&gt;
 *         &lt;pool-name&gt;jdbc/dataSource&lt;/pool-name&gt;
 *         &lt;password-query&gt;
 *           SELECT password FROM tbUser WHERE loginName = ?
 *         &lt;/password-query&gt;
 *           SELECT roleName FROM tbUserRole INNER JOIN tbUser ON tbUserRole.userId = tbUser.userId WHERE loginName = ?
 *         &lt;roles-query&gt;
 *         &lt;/roles-query&gt;
 *       &lt;/authenticator&gt;
 *
 *       ...
 *
 *     &lt;/host&gt;
 *
 *   &lt;/http-server&gt;
 *
 * &lt;/caucho.com&gt;
 * </pre>
 *
 * <p>Note that Resin requires authenticator implementation class to be
 * in the startup class path, so this class is included in the
 * <code>cb2-auth.jar</code>, which has to be added to the Resin's
 * startup class path.
 *
 * @author Lev Himmelfarb
 * @version $Id: JdbcAuthenticator.java,v 1.2 2004/04/04 20:13:45 levahim Exp $
 */
public class JdbcAuthenticator
    extends AbstractAuthenticator {

    /**
     * Output for error messages.
     */
    private static final PrintStream log = System.err;

    /**
     * The database connection pool's name in the JNDI.
     */
    private String poolName;

    /**
     * SQL query, which fetches a user's password by login name.
     */
    private String passwordQuery;

    /**
     * SQL query, which fetches the list of user roles by the user's
     * login name. This parameter is optional and may be
     * <code>null</code>.
     */
    private String rolesQuery;


    /**
     * The datasource for database connections. 
     */
    protected DataSource dataSource;


    /**
     * Gets the database connection pool name.
     *
     * @return the database connection pool name.
     */
    public String getPoolName() {

        return this.poolName;
    }

    /**
     * Sets the database connection pool name.
     *
     * @param poolName the pool name in the JNDI.
     */
    public void setPoolName(String poolName) {

        this.poolName = poolName;
    }

    /**
     * Gets the password query.
     *
     * @return the SQL query, which fetches the user's password
     * by login name. 
     */
    public String getPasswordQuery() {

        return this.passwordQuery;
    }

    /**
     * Sets the password query.
     *
     * @param passwordQuery the SQL query, which fetches user's
     * password by login name.
     */
    public void setPasswordQuery(String passwordQuery) {

        this.passwordQuery = passwordQuery;
    }

    /**
     * Gets the role query.
     *
     * @return the SQL query, which fetches the list of roles for
     * a user by the user's login name.
     */
    public String getRolesQuery() {

        return this.rolesQuery;
    }

    /**
     * Sets the role query.
     *
     * @param rolesQuery the SQL query, which fetches the list of
     * roles for a user by the user's login name. 
     */
    public void setRolesQuery(String rolesQuery) {

        this.rolesQuery = rolesQuery;
    }


    /**
     * Initialize the authenticator.
     *
     * @throws ServletException if an error happened.
     */
    public void init()
        throws ServletException {
        super.init();

        if(this.poolName == null) {
            log.println("JdbcAuthenticator error: pool-name property is not specified.");
            throw new ServletException("pool-name property is not specified.");
        }
        if(this.passwordQuery == null) {
            log.println("JdbcAuthenticator error: password-query property is not specified.");
            throw new ServletException("password-query property is not specified.");
        }

        try {
            InitialContext ic = null;
            try {
                ic = new InitialContext();
                this.dataSource = (DataSource)ic.lookup("java:comp/env/" + this.poolName);
            } finally {
                if(ic != null)
                    ic.close();
            }
        } catch(NamingException e) {
            log.println("JdbcAuthenticator error: got JNDI error: " + e);
            throw new ServletException("Got JNDI error.", e);
        }
    }

    /**
     * Authenticates the user given the request.
     *
     * @param request the HTTP request.
     * @param response the HTTP response.
     * @param application the Servlet context.
     * @param username the user name for the login.
     * @param password the password for the login.
     *
     * @return the authenticated user or <code>null</code> for a failure.
     *
     * @throws ServletException if an error happened.
     */
    public Principal loginImpl(HttpServletRequest request,
                               HttpServletResponse response,
                               ServletContext application,
                               String username,
                               String password)
        throws ServletException {

        String dbPassword = this.getPassword(username);

        if ((dbPassword == null) || (!dbPassword.equals(password)))
            return null;

        return new PrincipalWithRoles(username, this.getRoles(username));
    }

    /**
     * Returns the password from the database.
     *
     * @param request the HTTP request.
     * @param response the HTTP response.
     * @param application the Servlet context.
     * @param username the user's login name.
     * @param realm the realm.
     *
     * @return the password or <code>null</code> for a failure.
     *
     * @throws ServletException if an error happened.
     */
    protected String getDigestPassword(HttpServletRequest request,
                                       HttpServletResponse response,
                                       ServletContext application,
                                       String username,
                                       String realm)
        throws ServletException {

        return this.getPassword(username);
    }

    /**
     * Gets a user's password from the database,
     *
     * @param username the user's login name.
     *
     * @return the user's password as it is stored in the database
     * or <code>null</code> if no such user.
     *
     * @throws ServletException if an error happened.
     */
    protected String getPassword(String username)
        throws ServletException {

        Connection con = null;
        PreparedStatement pstmt = null;
        ResultSet rs = null;

        try {
            con = this.dataSource.getConnection();

            pstmt = con.prepareStatement(this.passwordQuery);
            pstmt.setString(1, username);

            rs = pstmt.executeQuery();
            if(!rs.next())
                return null;

            return rs.getString(1);

        } catch(SQLException e) {
            log.println("JdbcAuthenticator error: got JDBC error: " + e);
            throw new ServletException("Got JDBC error.", e);

        } finally {
            if(rs != null) {
                try {
                    rs.close();
                } catch(SQLException e) {
                    log.println("JdbcAuthenticator error: got JDBC error: " + e);
                    throw new ServletException("Got JDBC error.", e);
                }
            }
            if(pstmt != null) {
                try {
                    pstmt.close();
                } catch(SQLException e) {
                    log.println("JdbcAuthenticator error: got JDBC error: " + e);
                    throw new ServletException("Got JDBC error.", e);
                }
            }
            if(con != null) {
                try {
                    con.close();
                } catch(SQLException e) {
                    log.println("JdbcAuthenticator error: got JDBC error: " + e);
                    throw new ServletException("Got JDBC error.", e);
                }
            }
        }
    }

    /**
     * Gets a collection of user's security roles from the database.
     *
     * @param username the user's login name.
     *
     * @return set of user's security roles (objects of
     * <code>String</code>). Mat return an empty set, but never
     * a <code>null</code>.
     *
     * @throws ServletException if an error happened.
     */
    protected Set getRoles(String username)
        throws ServletException {

        if(this.rolesQuery == null)
            return new HashSet(0);

        Connection con = null;
        PreparedStatement pstmt = null;
        ResultSet rs = null;

        try {
            con = this.dataSource.getConnection();

            pstmt = con.prepareStatement(this.rolesQuery);
            pstmt.setString(1, username);

            Set res = new HashSet();
            rs = pstmt.executeQuery();
            while(rs.next())
                res.add(rs.getString(1));

            return res;

        } catch(SQLException e) {
            log.println("JdbcAuthenticator error: got JDBC error: " + e);
            throw new ServletException("Got JDBC error.", e);

        } finally {
            if(rs != null) {
                try {
                    rs.close();
                } catch(SQLException e) {
                    log.println("JdbcAuthenticator error: got JDBC error: " + e);
                    throw new ServletException("Got JDBC error.", e);
                }
            }
            if(pstmt != null) {
                try {
                    pstmt.close();
                } catch(SQLException e) {
                    log.println("JdbcAuthenticator error: got JDBC error: " + e);
                    throw new ServletException("Got JDBC error.", e);
                }
            }
            if(con != null) {
                try {
                    con.close();
                } catch(SQLException e) {
                    log.println("JdbcAuthenticator error: got JDBC error: " + e);
                    throw new ServletException("Got JDBC error.", e);
                }
            }
        }
    }

    /**
     * Checks if a user plays a security role.
     *
     * @param request the HTTP request.
     * @param response the HTTP response.
     * @param application the Servlet context.
     * @param principal the user principal, must be
     * <code>PrincipalWithRoles</code> instance.
     * @param role the role to check.
     *
     * @return <code>true</code> if the user plays the role.
     */
    public boolean isUserInRole(HttpServletRequest request,
                                HttpServletResponse response,
                                ServletContext application,
                                Principal principal,
                                String role) {

        if((principal == null) || (role == null) ||
           !(principal instanceof PrincipalWithRoles))
            return false;

        return ((PrincipalWithRoles)principal).isInRole(role);
    }
}
