package com.boylesoftware.cb2.bcmproviders.http;

import java.io.DataOutputStream;
import java.io.BufferedInputStream;
import java.io.CharArrayWriter;
import java.io.Serializable;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Map;
import java.util.HashMap;
import java.util.Properties;
import java.util.Enumeration;
import java.util.Iterator;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.net.MalformedURLException;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.BroadcastMessagingProvider;
import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.ApplicationException;
import com.boylesoftware.cb2.util.CodeUtils;
import com.boylesoftware.cb2.util.pool.AdaptiveObjectPool;
import com.boylesoftware.cb2.util.pool.PooledThreadFactory;
import com.boylesoftware.cb2.util.pool.PooledThread;


/**
 * Implementation of the <code>BroadcastMessagingProvider</code>
 * interface which uses HTTP to deliver messages to the members of the
 * cluster.
 *
 * <p>This provider implementation requires a special servlet deployed
 * on every member of the cluster. The servlet is implemented by
 * <code>ReceiverServlet</code> class and can be deployed in the
 * <code>web.xml</code> file as follows:
 *
 * <pre>
 * ...
 *
 * &lt;servlet&gt;
 *   &lt;servlet-name&gt;bcmreceiver&lt;/servlet-name&gt;
 *   &lt;display-name&gt;The CB2 BCM Receiver&lt;/display-name&gt;
 *   &lt;servlet-class&gt;com.boylesoftware.cb2.bcmproviders.http.ReceiverServlet&lt;/servlet-class&gt;
 *   &lt;load-on-startup&gt;12&lt;/load-on-startup&gt;
 * &lt;/servlet&gt;
 *
 * ...
 *
 * &lt;servlet-mapping&gt;
 *   &lt;servlet-name&gt;bcmreceiver&lt;/servlet-name&gt;
 *   &lt;url-pattern&gt;/cb2-bcm&lt;/url-pattern&gt;
 * &lt;/servlet-mapping&gt;
 *
 * ...</pre>
 *
 * <p>It is not necessary to protect the receiver servlet's URL,
 * because it always checks that the request comes with the same
 * password specified by
 * <code>com.boylesoftware.cb2.broadcastMessaging.http.password</code>
 * application property. This also means that this property has to have the
 * same value on all the cluster members. The password is passed as a
 * request parameter and is clear text though, so for better security
 * you may consider protecting the URL restricting access, for example,
 * only to the cluster members' IP addresses.
 *
 * <p>This implementation uses the following application parameters:
 *
 * <ul>
 *
 * <li><code><b>com.boylesoftware.cb2.broadcastMessaging.http.password</b></code> -
 * password, which is used to protect internal cluster communication. This
 * password must be the same on all cluster members.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.broadcastMessaging.http.<i>xxx</i>.url</b></code> -
 * URL of <code>bmcreceiver</code> servlet of a cluster member. The
 * <code>xxx</code> is the member's name, which can be any string. Specify
 * one such property for each member of the cluster including the local
 * member itself (so the set of properties is exactly the same on all
 * cluster members).</li>
 *
 * </ul>
 *
 * @author Lev Himmelfarb
 * @version $Id: HttpBCMProvider.java,v 1.6 2004/10/11 18:17:36 levahim Exp $
 */
public final class HttpBCMProvider
    implements BroadcastMessagingProvider {


    ///////////////////////////////////////////////////////////////////////////
    // Internal classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Internal factory for message deliverer threads.
     */
    static final class MessageDelivererFactory
        extends PooledThreadFactory {

        /**
         * The log.
         */
        private final Log log;


        /**
         * Constructs the factory.
         *
         * @param log the log to be used in the deliverers.
         */
        MessageDelivererFactory(Log log) {
            super("HTTP BCM Deliverers");

            this.log = log;
        }


        /**
         * Creates a new instance of message deliverer.
         *
         * @return new message deliverer instance.
         */
        public PooledThread createPooledThread() {

            this.log.debug("creating new deliverer thread");

            return new MessageDeliverer(this.log);
        }
    }


    /**
     * Internal message deliverer implementation.
     */
    static final class MessageDeliverer
        extends PooledThread {

        /**
         * The log.
         */
        private final Log log;

        /**
         * Fully formed target URL.
         */
        private URL url;

        /**
         * Message parameter, may be <code>null</code>.
         */
        private Serializable param;


        /**
         * Constructs a deliverer.
         *
         * @param log the log.
         */
        MessageDeliverer(Log log) {

            this.log = log;
        }


        /**
         * Setup the deliverer.
         *
         * @param baseUrl base target URL without channel.
         * @param channel channel name.
         * @param param optional message parameter, mey be <code>null</code>.
         */
        void setup(String baseUrl, String channel, Serializable param) {

            try {
                this.url = new URL(baseUrl + URLEncoder.encode(channel, "UTF-8"));
            } catch(UnsupportedEncodingException e) { /* never can happen */
            } catch(MalformedURLException e) { /* never can happen */ }
            this.param = param;
        }

        /**
         * Delivers a message.
         */
        protected void execute() {

            this.log.debug("message deliverer to [" + this.url.toString() + "]: started");

            // send the message
            BufferedInputStream in = null;
            DataOutputStream out = null;
            CharArrayWriter resp = null;
            try {

                // setup connection
                URLConnection urlConn = this.url.openConnection();
                urlConn.setDoInput(true);
                urlConn.setDoOutput(this.param != null);
                urlConn.setUseCaches(false);

                // send parameter
                if(this.param != null) {
                    urlConn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
                    out = new DataOutputStream(urlConn.getOutputStream());
                    out.writeBytes("param=" + CodeUtils.objectToString(this.param));
                    out.flush();
                    out.close();
                    out = null;
                }

                // read response
                in = new BufferedInputStream(urlConn.getInputStream(), 100);
                resp = new CharArrayWriter(100);
                int c;
                while((c = in.read()) >= 0)
                    out.write(c);

            } catch(IOException e) {
                this.log.error("Got error sending message to a cluster member.", e);
                resp = null;

            } finally {
                if(in != null) {
                    try {
                        in.close();
                    } catch(IOException e) {
                        this.log.error("Error closing the input stream, continuing operation.", e);
                    }
                }
                if(out != null) {
                    try {
                        out.close();
                    } catch(IOException e) {
                        this.log.error("Error closing the output stream, continuing operation.", e);
                    }
                }
            }

            // process the response
            if(resp != null) {
                this.log.debug("got response [" + resp.toString() + "]");
            }

            // all done
            this.log.debug("message deliverer [" + this.url.toString() + "]: finished");
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Provider data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * URLs of the cluster members excluding our own URL. The keys are
     * member names (objects of <code>String</code>), the values are
     * URLs without the channel name added (objects of
     * <code>String</code>).
     */
    private final Map memberUrls = new HashMap();

    /**
     * Our log.
     */
    private Log log;

    /**
     * Pool of deliverer threads.
     */
    private AdaptiveObjectPool delivererThreadsPool;

    /**
     * Id of this provider, unique in the cluster.
     */
    private String providerId;


    ///////////////////////////////////////////////////////////////////////////
    // Provider interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Initialize the provider instance.
     *
     * @param appCtx the application context.
     *
     * @throws ApplicationException if a serious error has happened
     * and the provider cannot be initialized.
     */
    public void init(ApplicationContext appCtx)
        throws ApplicationException {

        this.log = appCtx.getLog(LOG_SUBSYS);

        this.log.debug("initializing HTTP Broadcast Messaging provider");

        // form the provider id
        try {
            StringBuffer providerId = new StringBuffer();
            byte [] myAddress = java.net.InetAddress.getLocalHost().getAddress();
            for(int i = 0; i < myAddress.length; i++) {
                String octet = Integer.toHexString(myAddress[i] & 0xFF);
                if(octet.length() < 2)
                    providerId.append('0');
                providerId.append(octet);
            }
            providerId.append(Long.toHexString(System.currentTimeMillis()));

            this.providerId = providerId.toString();

            this.log.debug("this provider's uniqie id will be [" + this.providerId + "]");

        } catch(java.net.UnknownHostException e) {
            this.log.fatal("Could not determine local IP address.", e);
            throw new ApplicationException("Could not determine local IP address.", e);
        }

        // build cluster members' URLs
        Properties props = appCtx.getProperties();
        String password = props.getProperty("com.boylesoftware.cb2.broadcastMessaging.http.password");
        if(password == null) {
            this.log.fatal("The com.boylesoftware.cb2.broadcastMessaging.http.password property is not defined.");
            throw new ApplicationException("The com.boylesoftware.cb2.broadcastMessaging.http.password property is not defined.");
        }
        Enumeration en = props.propertyNames();
        while(en.hasMoreElements()) {
            String propName = (String)en.nextElement();
            if(propName.startsWith("com.boylesoftware.cb2.broadcastMessaging.http.") &&
               propName.endsWith(".url")) {

                String memberName = propName.substring("com.boylesoftware.cb2.broadcastMessaging.http.".length(),
                                                       propName.length() - ".url".length());

                StringBuffer memberUrlSB = new StringBuffer(props.getProperty(propName));
                try {
                    memberUrlSB.append("?fromId=");
                    memberUrlSB.append(URLEncoder.encode(this.providerId, "UTF-8"));
                    memberUrlSB.append("&toName=");
                    memberUrlSB.append(URLEncoder.encode(memberName, "UTF-8"));
                    memberUrlSB.append("&password=");
                    memberUrlSB.append(URLEncoder.encode(password, "UTF-8"));
                    memberUrlSB.append("&channel=");
                } catch(UnsupportedEncodingException e) { /* never can happen */ }

                // check the URL
                try {
                    new URL(memberUrlSB.toString());
                } catch(MalformedURLException e) {
                    this.log.fatal("Invalid cluster member URL [" + props.getProperty(propName) + "]", e);
                    throw new ApplicationException("Invalid cluster member URL.", e);
                }

                this.memberUrls.put(memberName, memberUrlSB.toString());

                this.log.debug("added cluster member name [" + memberName +
                               "], url [" + props.getProperty(propName) + "]");
            }
        }

        // create deliverers thread pool
        this.delivererThreadsPool =
            new AdaptiveObjectPool(new MessageDelivererFactory(this.log), 30, 0, 300);

        // all done
        this.log.debug("successfully initialized.");
    }

    /**
     * Destroy the provider.
     */
    public void destroy() {

        this.log.debug("destroying the HTTP Broadcast Messaging provider instance");

        this.delivererThreadsPool.discard();
        this.memberUrls.clear();
        this.log = null;
    }

    /**
     * Send a message to all cluster members.
     *
     * @param channelName logical name of the channel, to which the message is
     * targeted.
     * @param messageParam optional message parameter, may be
     * <code>null</code>.
     */
    public void broadcastMessage(String channelName,
                                 Serializable messageParam) {

        this.log.debug("sending broadcast message on channel " + channelName);

        Iterator i = this.memberUrls.values().iterator();
        while(i.hasNext()) {
            String baseUrl = (String)i.next();

            MessageDeliverer deliverer = (MessageDeliverer)this.delivererThreadsPool.get();
            deliverer.setup(baseUrl, channelName, messageParam);
            deliverer.doExecute();
        }
    }
}
