package com.boylesoftware.cb2.bcmproviders.sockethub;

import java.net.Socket;
import java.net.ServerSocket;
import java.net.InetAddress;
import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Date;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.ApplicationException;
import com.boylesoftware.cb2.util.EventStatistics;


/**
 * Standalone program, which plays the role of a message hub, or server.
 * In the BCM implementation represented by this package cluster nodes connect
 * to a centralized server. When a node needs to broadcast a message it sends
 * it to the server and then the server sends it to all connected nodes.
 * This approach allows keeping cluster nodes without knowledge of each other
 * - all a node needs to know to be able to send and receive messages is the
 * address of the hub.
 *
 * <p>This hub program is a CB2 based console application, which means it
 * needs <code>cb2app.properties</code> file. It takes one command line
 * option - directory, which contains the configuration file.
 * If there are no command line arguments, <code>user.dir</code> system
 * property is used, which points to the current working directory. The
 * option is "-d" followed by the directory path.
 *
 * <p>To shutdown a running hub instance, run this class again and pass
 * "shutdown" command line argument. To get information on currently connected
 * clients give it "status" argument. The complete command line is:
 *
 * <pre>java com.boylesoftware.cb2.bcmproviders.sockethub.Hub [-d &lt;config dir&gt;] [shutdown|status]</pre>
 *
 * <p>This program requires one application property -
 * <code>com.boylesoftware.cb2.broadcastMessaging.sockethub.port</code>, which
 * specifies what TCP port the hub uses to listen for cluster members'
 * connections.
 *
 * @author Lev Himmelfarb
 * @version $Id: Hub.java,v 1.10 2004/10/11 18:17:36 levahim Exp $
 */
public final class Hub {


    ///////////////////////////////////////////////////////////////////////////
    // program startup
    ///////////////////////////////////////////////////////////////////////////

    /**
     * The program entry point.
     *
     * @param args command line arguments.
     *
     * @throws ApplicationException if the program exits abnormally.
     */
    public static void main(String [] args)
        throws ApplicationException {

        String configDir = null;
        boolean doShutdown = false;
        boolean doStatus = false;

        for(int i = 0; i < args.length; i++) {
            if(args[i].startsWith("-d")) {
                if(args[i].length() > 2)
                    configDir = args[i].substring(2);
                else {
                    if(args.length <= (i + 1))
                        usage();
                    else
                        configDir = args[++i];
                }
            } else if(args[i].equalsIgnoreCase("shutdown")) {
                doShutdown = true;
            } else if(args[i].equalsIgnoreCase("status")) {
                doStatus = true;
            } else
                usage();
        }
        if(configDir == null)
            configDir = System.getProperty("user.dir");

        Hub app = new Hub(configDir);

        if(doShutdown)
            app.shutdown();
        else if(doStatus)
            app.status();
        else
            app.run();
    }

    /**
     * Prints the usage message and terminates the JVM. This method
     * never returns.
     */
    private static void usage() {

        System.out.println("usage:");
        System.out.println("");
        System.out.println("    java com.boylesoftware.cb2.bcmproviders.sockethub.Hub [-d <config dir>] [shutdown|status]");

        System.exit(1);
    }


    ///////////////////////////////////////////////////////////////////////////
    // internal classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Represents a client connection data.
     */
    private static final class ConnectionDescriptor {

        /**
         * Next socket descriptor id to assign.
         */
        private static long nextId = 1L;


        /**
         * Unique connection id.
         */
        public final Long id;

        /**
         * When the connection was established.
         */
        public final Date establishedOn;

        /**
         * The socket.
         */
        public final Socket socket;

        /**
         * The message input.
         */
        public final ObjectInputStream input;

        /**
         * The message output.
         */
        private final ObjectOutputStream output;

        /**
         * Pings statistics.
         */
        public final EventStatistics pingStats;

        /**
         * The connection's description for logging. 
         */
        public final String desc;


        /**
         * Creates a new instance of a descriptor for a socket.
         *
         * @param socket the socket to wrap.
         *
         * @throws IOException if could not establish object
         * input/output channel.
         */
        ConnectionDescriptor(Socket socket)
            throws IOException {

            synchronized(this.getClass()) {
                this.id = new Long(nextId++);
            }
            this.establishedOn = new Date();
            this.socket = socket;
            this.input = new ObjectInputStream(socket.getInputStream());
            this.output = new ObjectOutputStream(socket.getOutputStream());
            this.pingStats = new EventStatistics();
            this.desc = socket.getInetAddress().getHostAddress() + ":" + socket.getPort();
        }


        /**
         * Sends a message.
         *
         * @param msg the message to send.
         *
         * @throws IOException if there was an I/O error.
         */
        public synchronized void sendMessage(Message msg)
            throws IOException {

            this.output.writeObject(msg);
            this.output.flush();
        }
    }


    /**
     * Connection status descriptor in the "get status" message
     * response.
     */
    private static final class ConnectionStatus
        implements Serializable {

        /**
         * IP address of the client.
         */
        public String remoteAddress;

        /**
         * Remote port.
         */
        public int remotePort;

        /**
         * Local port.
         */
        public int localPort;

        /**
         * When the connection was established.
         */
        public Date establishedOn;

        /**
         * Ping statistics.
         */
        public EventStatistics pingStats;
    }


    /**
     * The client connection service thread.
     */
    private final class Client
        implements Runnable {

        /**
         * Our socket.
         */
        private final ConnectionDescriptor c;


        /**
         * Constructs a new client object.
         *
         * @param c the client connection descriptor.
         */
        Client(ConnectionDescriptor c) {

            this.c = c;
        }


        /**
         * The thread entry point.
         */
        public void run() {

            try {

                Message msg;
                do {
                    msg = (Message)this.c.input.readObject();
                    if(log.isDebugEnabled())
                        log.debug("got message from [" + this.c.desc +
                                  "]: type=[" + msg.type +
                                  "], channel=[" + msg.channelName +
                                  "], param=[" + (msg.messageParam != null ?
                                                  "byte[" + msg.messageParam.length + "]" :
                                                  "null") + "]");
                    switch(msg.type) {
                    case Message.TYPE_REGULAR:
                        msg.source = this.c.desc;
                        synchronized(messageQueue) {
                            messageQueue.addLast(msg);
                            messageQueue.notify();
                        }
                        break;
                    case Message.TYPE_PING:
                        this.c.pingStats.eventFinished(((Long)msg.getMessageParam()).longValue());
                        break;
                    case Message.TYPE_CLOSE:
                        log.info("got request to terminate this client connection from [" + this.c.desc + "]");
                        break;
                    default:
                        log.error("got invalid message type " + msg.type +
                                  " from [" + this.c.desc + "], continuing operation");
                    }
                } while(msg.type != Message.TYPE_CLOSE);

            } catch(Exception e) {
                log.error("Got error serving client connection [" + this.c.desc +
                          "], terminating it.", e);
            } finally {
                log.debug("closing connection");
                try {
                    this.c.socket.close();
                } catch(IOException e) {
                    log.error("Could not close the connection, finishing the thread anyway.", e);
                }
            }
        }
    }


    /**
     * Sends messages from the message queue.
     */
    private final class Sender
        implements Runnable {

        /**
         * The thread entry point.
         */
        public void run() {

            log.debug("started sender thread");

            Message msg;
            do {

                // get a message from the queue
                synchronized(messageQueue) {
                    while(messageQueue.size() == 0) {
                        log.debug("waiting for messages in the queue");
                        try {
                            messageQueue.wait();
                        } catch(InterruptedException e) { /* ignore */ }
                    }
                    msg = (Message)messageQueue.removeFirst();
                }

                // send the message to everybody
                if(msg.type != Message.TYPE_SHUTDOWN) {
                    ConnectionDescriptor [] c;
                    synchronized(openSockets) {
                        c = (ConnectionDescriptor [])openSockets.values().
                            toArray(new ConnectionDescriptor[openSockets.size()]);
                    }
                    log.debug("sending message to all cluster members");
                    for(int i = 0; i < c.length; i++) {
                        try {
                            if(c[i].socket.isClosed()) {
                                log.debug("removing closed connection [" + c[i].desc + "]");
                                synchronized(openSockets) {
                                    openSockets.remove(c[i].id);
                                }
                            } else {
                                log.debug("sending message to [" + c[i].desc + "]");
                                c[i].sendMessage(msg);
                            }
                        } catch(Exception e) {
                            log.error("Got error sending message to [" + c[i].desc +
                                      "], continuing operation.", e);
                        }
                    }
                    log.debug("the message has been sent to all cluster members");
                }

            } while(msg.type != Message.TYPE_SHUTDOWN);

            log.debug("shutting down the sender thread");
        }
    }


    /**
     * Daemon thread that peroidically pings all registered connections
     * gathering statistics and closing them in case unexcpected errors.
     */
    private final class Pinger
        implements Runnable {

        /**
         * Ping connections every this amount of milliseconds.
         */
        private static final long PING_PERIOD = 60000L;


        /**
         * The thread's entry point.
         */
        public void run() {

            log.debug("started pinger thread");

            try {

                Message msg;
                do {
                    try {
                        Thread.sleep(PING_PERIOD);
                    } catch(InterruptedException e) { /* ignore it */ }

                    ConnectionDescriptor [] c;
                    synchronized(openSockets) {
                        c = (ConnectionDescriptor [])openSockets.values().
                            toArray(new ConnectionDescriptor[openSockets.size()]);
                    }
                    for(int i = 0; i < c.length; i++) {
                        if(c[i].socket.isClosed()) {
                            log.debug("removing closed connection [" + c[i].desc + "]");
                            synchronized(openSockets) {
                                openSockets.remove(c[i].id);
                            }
                        } else {
                            msg = new Message();
                            msg.type = Message.TYPE_PING;
                            msg.setMessageParam(new Long(c[i].pingStats.eventStarted()));
                            try {
                                c[i].sendMessage(msg);
                            } catch(IOException e) {
                                log.error("Got I/O error sending a ping to [" + c[i].desc +
                                          "], closing the connection and removing it from the list of registered clients.", e);
                                synchronized(openSockets) {
                                    try {
                                        c[i].socket.close();
                                    } catch(IOException e1) {
                                        log.error("Got error closing connection after another error, ignoring it.", e1);
                                    }
                                    openSockets.remove(c[i].id);
                                }
                            }
                        }
                    }
                } while(true); // it's a daemon thread

            } catch(Throwable e) {
                log.error("A fatal error happened in the pinger," +
                          " the thread will be shutdown and no pings will be performed," +
                          " the rest of the application will continue functioning.", e);
            }
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // program data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Our application context.
     */
    private final ApplicationContext appCtx;

    /**
     * Our log.
     */
    final Log log;

    /**
     * The server TCP port number.
     */
    private final int port;

    /**
     * Queue of messages to send.
     */
    final LinkedList messageQueue = new LinkedList();

    /**
     * Set of all open (and closed) client sockets. It gets cleaned up
     * from the closed sockets automatically in the <code>Sender</code>
     * thread. The objects in the set are of type
     * <code>ConnectionDescriptor</code>.
     */
    final Map openSockets = new HashMap();


    ///////////////////////////////////////////////////////////////////////////
    // main body
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructs application instanse and intializes the application
     * context.
     *
     * @param configDir path to the configuration directory.
     *
     * @throws ApplicationException if application context could not
     * be initialized.
     */
    private Hub(String configDir)
        throws ApplicationException {

        this.appCtx = new ApplicationContext("SocketHub", configDir);
        this.log = this.appCtx.getLog("SocketHub");

        // get the port number
        String portS = this.appCtx.getProperty("com.boylesoftware.cb2.broadcastMessaging.sockethub.port");
        if(portS == null)
            throw new ApplicationException("Required application property com.boylesoftware.cb2.broadcastMessaging.sockethub.port is not defined.");
        try {
            this.port = Integer.parseInt(portS);
        } catch(NumberFormatException e) {
            throw new ApplicationException("Required application property com.boylesoftware.cb2.broadcastMessaging.sockethub.port must be a number.");
        }
    }


    /**
     * The server main body. Runs as a daemon thread.
     */
    private void run() {

        ServerSocket ss = null;
        Thread senderThread = null;
        try {
            // start sender thread
            senderThread = new Thread(new Sender(), "Sender");
            senderThread.start();

            // start pinger thread
            Thread pingerThread = new Thread(new Pinger(), "Pinger");
            pingerThread.setDaemon(true);
            pingerThread.start();

            // open server socket
            this.log.debug("opening server port for listening");
            ss = new ServerSocket(this.port);

            // main loop
            this.log.info("accepting connections");
            boolean shutdown = false;
            do {
                Socket s = ss.accept();
                this.log.debug("got new connection from " + s.getInetAddress().getHostAddress());
                try {
                    // setup the connection
                    s.setKeepAlive(true);
                    ConnectionDescriptor c = new ConnectionDescriptor(s);

                    // read initial command, which can be OPEN or SHUTDOWN
                    Message initCmd = (Message)c.input.readObject();
                    switch(initCmd.type) {
                    case Message.TYPE_OPEN:
                        this.log.info("opening new session for [" + c.desc + "]");
                        Thread clientThread = new Thread(new Client(c));
                        this.openSockets.put(c.id, c);
                        clientThread.setDaemon(true);
                        clientThread.start();
                        break;
                    case Message.TYPE_SHUTDOWN:
                        this.log.info("received shutdown command");
                        shutdown = true;
                        s.close();
                        break;
                    case Message.TYPE_STATUS:
                        this.log.debug("received get status command");
                        ConnectionDescriptor [] cd;
                        synchronized(this.openSockets) {
                            cd = (ConnectionDescriptor [])this.openSockets.values().
                                toArray(new ConnectionDescriptor[this.openSockets.size()]);
                        }
                        List csList = new ArrayList(cd.length);
                        for(int i = 0; i < cd.length; i++) {
                            if(cd[i].socket.isClosed()) {
                                this.log.debug("removing closed connection [" + cd[i].desc + "]");
                                synchronized(this.openSockets) {
                                    this.openSockets.remove(cd[i].id);
                                }
                            } else {
                                ConnectionStatus cs = new ConnectionStatus();
                                cs.establishedOn = cd[i].establishedOn;
                                cs.remoteAddress = cd[i].socket.getInetAddress().getHostAddress();
                                cs.remotePort = cd[i].socket.getPort();
                                cs.localPort = cd[i].socket.getLocalPort();
                                cs.pingStats = cd[i].pingStats;
                                csList.add(cs);
                            }
                        }
                        Message msg = new Message();
                        msg.setMessageParam(csList.toArray(new ConnectionStatus[csList.size()]));
                        c.sendMessage(msg);
                        s.close();
                        break;
                    default:
                        throw new IOException("Received inappropriate at the beginning of the session message of type " + initCmd.type);
                    }

                } catch(Exception e) {
                    this.log.error("Got error initializing client connection, closing it and continuing operation.", e);
                    try {
                        s.close();
                    } catch(IOException e1) {
                        this.log.error("Could not close connection, continuing operation.", e1);
                    }
                }
            } while(!shutdown);

        } catch(IOException e) {
            this.log.fatal("Got fatal I/O error, terminating the server.", e);

        } finally {

            // close the server socket
            if(ss != null) {
                this.log.debug("closing server socket");
                try {
                    ss.close();
                } catch(IOException e) {
                    this.log.error("Could not close the server socket.", e);
                }
            }

            // tell the sender thread to shutdown
            if(senderThread != null) {
                synchronized(this.messageQueue) {
                    Message shutdownMsg = new Message();
                    shutdownMsg.type = Message.TYPE_SHUTDOWN;
                    this.messageQueue.addFirst(shutdownMsg);
                    this.messageQueue.notify();
                }
                try {
                    senderThread.join();
                } catch(InterruptedException e) { /* ignore, exit */ }
            }

            // shutdown the app
            this.appCtx.destroy();
        }
    }

    /**
     * Shuts down a running server.
     */
    private void shutdown()
        throws ApplicationException {

        try {

            Socket s = null;
            try {
                s = new Socket("127.0.0.1", this.port);
                s.shutdownInput();

                ObjectOutputStream oos = new ObjectOutputStream(s.getOutputStream());
                Message msg = new Message();
                msg.type = Message.TYPE_SHUTDOWN;
                oos.writeObject(msg);
                oos.flush();

            } finally {
                if(s != null)
                    s.close();
            }

        } catch(IOException e) {
            throw new ApplicationException("Got I/O error shutting down the server.", e);

        } finally {
            this.appCtx.destroy();
        }
    }

    /**
     * Prints the hub's status information.
     */
    private void status()
        throws ApplicationException {

        try {

            Socket s = null;
            try {
                s = new Socket("127.0.0.1", this.port);

                ObjectOutputStream oos = new ObjectOutputStream(s.getOutputStream());
                ObjectInputStream ois = new ObjectInputStream(s.getInputStream());
                Message msg = new Message();
                msg.type = Message.TYPE_STATUS;
                oos.writeObject(msg);
                oos.flush();

                msg = (Message)ois.readObject();

                s.close();
                s = null;

                ConnectionStatus [] cs = (ConnectionStatus [])msg.getMessageParam();
                StringBuffer sb = new StringBuffer();
                for(int i = 0; i < cs.length; i++) {
                    InetAddress addr = InetAddress.getByName(cs[i].remoteAddress);
                    sb.append(i + 1);
                    sb.append(". remote: ");
                    sb.append(addr.getHostName());
                    sb.append(":");
                    sb.append(cs[i].remotePort);
                    sb.append(" (");
                    sb.append(addr.getHostAddress());
                    sb.append(":");
                    sb.append(cs[i].remotePort);
                    sb.append("), local: *:");
                    sb.append(cs[i].localPort);
                    sb.append(", stats: ");
                    sb.append(cs[i].pingStats.getNTimes());
                    sb.append(" ");
                    sb.append(cs[i].pingStats.getProcTimeMin());
                    sb.append("/");
                    sb.append(cs[i].pingStats.getProcTimeAvg());
                    sb.append("/");
                    sb.append(cs[i].pingStats.getProcTimeMax());
                    System.out.println(sb.toString());
                    sb.delete(0, sb.length());
                }

            } finally {
                if(s != null)
                    s.close();
            }

        } catch(ClassNotFoundException e) {
            throw new ApplicationException("Incompatible hub instance.", e);

        } catch(IOException e) {
            throw new ApplicationException("Got I/O error shutting down the server.", e);

        } finally {
            this.appCtx.destroy();
        }
    }
}
