package com.boylesoftware.cb2.bcmproviders.sockethub;

import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.SocketException;
import java.util.LinkedList;

import org.apache.commons.logging.Log;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.ApplicationException;
import com.boylesoftware.cb2.BroadcastMessagingProvider;


/**
 * Implementation of BCM provider, which uses a centralized hub to broadcast
 * messages to all registered cluster members over TCP.
 *
 * <p>This provider implementation requires two application properties
 * that point it to the location of the hub. The properties are:
 *
 * <ul>
 *
 * <li><code><b>com.boylesoftware.cb2.broadcastMessaging.sockethub.host</b></code> -
 * name or IP address of the host where the hub is running.</li>
 *
 * <li><code><b>com.boylesoftware.cb2.broadcastMessaging.sockethub.port</b></code> -
 * TCP port on which the hub is listening for connections.</li>
 *
 * </ul>
 *
 * @author Lev Himmelfarb
 * @version $Id: SocketHubBCMProvider.java,v 1.7 2004/10/11 18:17:36 levahim Exp $
 */
public class SocketHubBCMProvider
    implements BroadcastMessagingProvider {


    ///////////////////////////////////////////////////////////////////////////
    // Inner classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Represents the connection to the hub.
     */
    private final class Connection {

        /**
         * Retry reconnect to the hub every this amount of seconds
         * in case of problems.
         */
        private static final long RETRY_SEC = 30;


        /**
         * Host where the hub is running.
         */
        private final String host;

        /**
         * Port the hub is listening for incoming connections.
         */
        private final int port;

        /**
         * Connection to the hub. If this field is <code>null</code>
         * it means that the connection is not established.
         */
        private Socket socket;

        /**
         * Stream to send message to.
         */
        private ObjectOutputStream output;

        /**
         * Stream to read messages from.
         */
        private ObjectInputStream input;

        /**
         * Sycnhronization object for accessing the socket.
         */
        private final Object socketLock = new Object();

        /**
         * Indicates that we are shutting down.
         */
        private boolean shuttingDown = false;

        /**
         * Indicates that the connection should be close and reopened.
         */
        private boolean reconnect = false;


        /**
         * Constructs a connection object.
         *
         * @param host host where the hub is running.
         * @param port port the hub listens for incoming connections.
         */
        Connection(String host, int port) {

            this.host = host;
            this.port = port;
        }


        /**
         * Checks if we are connected and if not establishes a connection.
         * This method tries to connect indefinitely until success.
         */
        private boolean ensureConnection() {

            boolean success = false;
            do {
                synchronized(this.socketLock) {

                    if(this.reconnect && (this.socket != null)) {
                        log.debug("told to reconnect, closing the current connection");
                        try {
                            this.socket.close();
                        } catch(Exception e) {
                            log.error("got error closing the connection, ignoring it", e);
                        }
                        this.socket = null;
                        this.reconnect = false;
                    }

                    if(this.socket == null) {

                        if(this.shuttingDown) {
                            log.debug("told to shutdown, aborting reestablishing connection");
                            return false;
                        }

                        try {
                            log.debug("trying to open connection to the hub at " + this.host + ":" + this.port);
                            this.socket = new Socket(this.host, this.port);
                            this.socket.setKeepAlive(true);
                            this.output = new ObjectOutputStream(this.socket.getOutputStream());
                            this.input = new ObjectInputStream(this.socket.getInputStream());

                            log.debug("sending session open command to the hub");
                            Message msg = new Message();
                            msg.type = Message.TYPE_OPEN;
                            this.output.writeObject(msg);
                            this.output.flush();

                            this.socketLock.notifyAll();

                            success = true;

                        } catch(Exception e) {
                            log.error("got error opening connection to the hub, will try again after " + RETRY_SEC + " seconds", e);
                            if(this.socket != null) {
                                try {
                                    this.socket.close();
                                } catch(Exception e1) {
                                    log.error("could not close socket after another error, ignoring", e1);
                                }
                                this.socket = null;
                            }
                            try {
                                this.socketLock.wait(RETRY_SEC * 1000L);
                            } catch(InterruptedException e1) {
                                log.debug("got interrupted, will try to reconnect now");
                            }
                        }
                    } else
                        success = true;
                }
            } while(!success);

            return true;
        }

        /**
         * Flags that the connection should be closed and reopnened.
         */
        private void setReconnect() {

            synchronized(this.socketLock) {
                this.reconnect = true;
                this.socketLock.notifyAll();
            }
        }

        /**
         * Flags that we are shutting down and no attempts to establish
         * new connections should be made.
         */
        private void setShuttingDown() {

            synchronized(this.socketLock) {
                this.shuttingDown = true;
                this.socketLock.notifyAll();
            }
        }

        /**
         * Closes the connection if it is open.
         */
        private void closeConnection() {

            synchronized(this.socketLock) {
                if(this.socket != null) {
                    try {
                        this.socket.close();
                    } catch(Exception e) {
                        log.error("got error closing connection to the hub, ignoring it", e);
                    }
                    this.socket = null;
                }
                this.socketLock.notifyAll();
            }
        }

        /**
         * Delivers a message to the hub.
         *
         * @param msg the message to send.
         */
        void sendMessage(Message msg) {

            if(msg.type == Message.TYPE_CLOSE)
                this.setShuttingDown();

            boolean success = false;
            do {
                if(this.ensureConnection()) {
                    try {
                        this.output.writeObject(msg);
                        this.output.flush();
                        log.debug("sent message of type [" + msg.type + "] to the hub");
                        if(msg.type == Message.TYPE_CLOSE)
                            this.closeConnection();
                        success = true;
                    } catch(Exception e) {
                        log.error("got error sending message to the hub, will close the connection and try to reconnect", e);
                        this.setReconnect();
                    }
                } else {
                    Thread.yield();
                    success = true;
                }
            } while(!success);
        }

        /**
         * Reads a message from the hub. Blocks until a message is
         * available.
         *
         * @return the message received from the hub.
         */
        Message readMessage() {

            Message msg = null;

            boolean success = false;
            do {
                if(this.ensureConnection()) {
                    try {
                        msg = (Message)this.input.readObject();
                        success = true;
                    } catch(Exception e) {
                        if(this.shuttingDown && (e instanceof SocketException)) {
                            msg = new Message();
                            msg.type = Message.TYPE_NOP;
                            success = true;
                        } else {
                            log.error("got error reading message from the hub, will close the connection and try to reconnect", e);
                            this.setReconnect();
                        }
                    }

                } else {
                    msg = new Message();
                    msg.type = Message.TYPE_NOP;
                    Thread.yield(); // give other threads chance to disable the receiver
                    success = true;
                }
            } while(!success);

            return msg;
        }
    }


    /**
     * Thread that reads messages from the queue and sends them to the hub.
     */
    private final class Sender
        implements Runnable {

        /**
         * The message queue, contains objects of <code>Message</code>.
         */
        private final LinkedList queue = new LinkedList();

        /**
         * Tells if the sender is disabled.
         */
        private boolean disabled = false;


        /**
         * The thread entry point.
         */
        public void run() {

            log.debug("started sender thread");

            try {

                Message msg;
                do {
                    // wait for a message in the queue
                    synchronized(this.queue) {
                        while(this.queue.size() == 0) {
                            log.debug("waiting for messages in the queue");
                            try {
                                this.queue.wait();
                            } catch(InterruptedException e) { /* ignore it */ }
                        }
                        msg = (Message)this.queue.removeFirst();
                    }

                    // send the message to the hub
                    log.debug("sending a message to the hub");
                    connection.sendMessage(msg);

                } while(msg.type != Message.TYPE_CLOSE);

            } catch(RuntimeException e) {
                synchronized(this.queue) {
                    this.disabled = true;
                }
                log.fatal("got unexpected fatal exception in the sender thread," +
                          " terminating the thread so no messages will be sent" +
                          " but the application will continue working.", e);

            } finally {
                log.debug("sender thread is finished");
            }
        }

        /**
         * Puts a message to the send queue.
         *
         * @param msg the message to send.
         */
        void sendMessage(Message msg) {

            synchronized(this.queue) {
                if(!this.disabled) {
                    this.queue.addLast(msg);
                    log.debug("added a message to the send queue, the new queue size is " + this.queue.size());
                    this.queue.notify();
                }
            }
        }

        /**
         * Tells the sender thread to shutdown. It causes the sender to
         * terminate the session with the hub, close the connection and
         * exit the thread. However, this method returns immediately
         * without waiting until the thread is dead.
         */
        void shutdown() {

            Message msg = new Message();
            msg.type = Message.TYPE_CLOSE;
            this.sendMessage(msg);

            synchronized(this.queue) {
                this.disabled = true;
            }
        }
    }


    /**
     * Thread, that reads messages from the hub and delivers
     * them to the terminals. The thread is supposed to be a daemon
     * thread.
     */
    private final class Receiver
        implements Runnable {

        /**
         * Tells if the receiver is disabled.
         */
        private boolean disabled = false;


        /**
         * The thread entry point.
         */
        public void run() {

            try {
                log.debug("started the receiver thread");

                // it's a daemon thread, loop forever
                while(true) {

                    Message msg = connection.readMessage();

                    synchronized(this) {
                        if(!this.disabled) {
                            try {

                                if(msg.type == Message.TYPE_NOP)
                                    continue;

                                if(msg.type == Message.TYPE_PING) {
                                    sender.sendMessage(msg);
                                    continue;
                                }

                                if(log.isDebugEnabled())
                                    log.debug("got message: type=[" + msg.type +
                                              "], channel=[" + msg.channelName +
                                              "], param=[" + (msg.messageParam != null ?
                                                              "byte[" + msg.messageParam.length + "]" :
                                                              "null") +
                                              "], source=[" + msg.source + "]");

                                if(msg.type != Message.TYPE_REGULAR) {
                                    log.error("got message of invalid type " + msg.type +
                                              ", ignoring it and continuing operation");
                                } else {
                                    appCtx.processBCMMessage(msg.channelName, msg.source, msg.getMessageParam());
                                }

                            } catch(Exception e) {
                                log.error("got error processing a message, continuing operation", e);
                            }
                        } else
                            return;
                    }
                }

            } catch(RuntimeException e) {
                log.fatal("got unexpected fatal exception in the receiver thread," +
                          " terminating the thread so no messages will be received" +
                          " but the application will continue working.", e);
            }
        }

        /**
         * Disables the receiver. Any new messages received after the receiver
         * was disables are ignored. If a message is already being processed
         * this method will wait until the processing is finished.
         */
        void disable() {

            synchronized(this) {
                this.disabled = true;
                log.debug("the receiver is disabled, ready to be shut down");
            }
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // The provider's data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Our application context.
     */
    ApplicationContext appCtx;

    /**
     * Connection to the hub.
     */
    Connection connection;

    /**
     * Our log.
     */
    Log log;

    /**
     * The sender.
     */
    Sender sender;

    /**
     * The sender's thread.
     */
    private Thread senderThread;

    /**
     * The receiver.
     */
    private Receiver receiver;


    ///////////////////////////////////////////////////////////////////////////
    // Main interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Initialize the provider.
     *
     * @param appCtx the application context.
     *
     * @throws ApplicationException if the provider cannot be intialized.
     */
    public void init(ApplicationContext appCtx)
        throws ApplicationException {

        this.appCtx = appCtx;
        this.log = appCtx.getLog(LOG_SUBSYS);

        this.log.debug("initializing SocketHub BCM provider");

        // read application properties
        String host = appCtx.getProperty("com.boylesoftware.cb2.broadcastMessaging.sockethub.host");
        if(host == null)
            throw new ApplicationException("Required application property com.boylesoftware.cb2.broadcastMessaging.sockethub.host is not defined.");
        String portS = appCtx.getProperty("com.boylesoftware.cb2.broadcastMessaging.sockethub.port");
        if(portS == null)
            throw new ApplicationException("Required application property com.boylesoftware.cb2.broadcastMessaging.sockethub.port is not defined.");
        int port;
        try {
            port = Integer.parseInt(portS);
        } catch(NumberFormatException e) {
            throw new ApplicationException("Application property com.boylesoftware.cb2.broadcastMessaging.sockethub.port must be a number.");
        }

        // create connection object
        this.connection = new Connection(host, port);

        // create and start sender thread
        this.senderThread = new Thread((this.sender = new Sender()),
                                       "SocketHub BCM Sender");
        this.senderThread.start();

        // create and start receiver thread
        Thread receiverThread = new Thread((this.receiver = new Receiver()),
                                           "SocketHub BCM Receiver");
        receiverThread.setDaemon(true);
        receiverThread.start();

        // done
        this.log.debug("successfully initialized.");
    }

    /**
     * Destroy the provider instance and close the connection to the hub.
     */
    public void destroy() {

        this.log.debug("destroying the SocketHub BCM provider instance");

        // disable the receiver
        this.receiver.disable();

        // shutdown the sender
        this.sender.shutdown();
        this.log.debug("waiting the sender thread to shutdown");
        try {
            this.senderThread.join();
        } catch(InterruptedException e) { /* ignore it */ }

        this.log.debug("the SocketHub BCM provider instance has been destroyed");

        // clear the fields
        this.appCtx = null;
        this.log = null;
    }

    /**
     * Broadcast a message to all cluster members. This implementation
     * actually sends the message to the hub and then the hub sends it to all
     * connected clients including this one. The message is not sent
     * immediately by this method, but placed into a queue so the method
     * may return before the message is actually delivered (even to the local
     * instance itself).
     *
     * @param channelName target BCM channel name.
     * @param messageParam optional message parameter, may be
     * <code>null</code>.
     */
    public void broadcastMessage(String channelName,
                                 Serializable messageParam) {

        Message msg = new Message();
        msg.type = Message.TYPE_REGULAR;
        msg.channelName = channelName;
        try {
            msg.setMessageParam(messageParam);
        } catch(IOException e) {
            this.log.error("Got error serializing message parameter, will send null instead.", e);
            msg.messageParam = null;
        }

        this.sender.sendMessage(msg);
    }
}
