package com.boylesoftware.cb2.commons.property;

import java.util.Properties;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.BLObject;
import com.boylesoftware.cb2.BLException;
import com.boylesoftware.cb2.util.BTreeMap;


/**
 * Generic BLO for managing database-stored application properties, that are
 * used by the application context via
 * <code>com.boylesoftware.cb2.appProperties.query</code> property. This BLO
 * can be deployed as a shared BLO since does not contain any session specific
 * data. The BLO needs a number of initialization parameters that name the
 * database table, which contains additional application properties, the column
 * in the table, which contains property names, and the column, which contains
 * property values. Here is how the BLO can be deployed in the
 * <code>blo-config.xml</code> file:
 *
 * <pre>
 * &lt;blo name="property" shared="true"&gt;
 *   &lt;init-param&gt;
 *     &lt;param-name&gt;table&lt;/param-name&gt;
 *     &lt;param-value&gt;Property&lt;/param-value&gt;
 *   &lt;/init-param&gt;
 *   &lt;init-param&gt;
 *     &lt;param-name&gt;nameColumn&lt;/param-name&gt;
 *     &lt;param-value&gt;name&lt;/param-value&gt;
 *   &lt;/init-param&gt;
 *   &lt;init-param&gt;
 *     &lt;param-name&gt;valueColumn&lt;/param-name&gt;
 *     &lt;param-value&gt;value&lt;/param-value&gt;
 *   &lt;/init-param&gt;
 * &lt;/blo&gt;</pre>
 *
 * <p>Note, that if the application does not use database-stored application
 * properties, that is <code>com.boylesoftware.cb2.appProperties.query</code>
 * is not set, the <code>init</code> method of this BLO throws an exception.
 *
 * @author Lev Himmelfarb
 * @version $Id: PropertySharedBLO.java,v 1.3 2004/04/29 19:32:47 levahim Exp $
 */
public class PropertySharedBLO
    extends BLObject {

    /**
     * This value is returned by <code>getPropertyType</code> method when type
     * of a non existant property is requested.
     */
    public static final int PROPTYPE_UNKNOWN = -1;

    /**
     * Property type, returned by <code>getPropertyType</code> method, which
     * indicates that the property comes from the
     * <code>cb2app.properties</code> file and its value cannot be changed,
     * because it is used by the application context before it processes
     * additional properties from the database.
     */
    public static final int PROPTYPE_IMMUTABLE = 0;

    /**
     * Property type, returned by <code>getPropertyType</code> method, which
     * indicates that the property comes from the
     * <code>cb2app.properties</code> file.
     */
    public static final int PROPTYPE_BASIC = 1;

    /**
     * Property type, returned by <code>getPropertyType</code> method, which
     * indicates that the property comes from the
     * <code>cb2app.properties</code> file, but its value has been overridden
     * by a property with the same name stored in the database.
     */
    public static final int PROPTYPE_OVERRIDDEN = 2;

    /**
     * Property type, returned by <code>getPropertyType</code> method, which
     * indicates that the property comes from the database.
     */
    public static final int PROPTYPE_DB = 3;


    /**
     * Set of name prefixes for application properties that cannot be stored in
     * the database and can be specified only in the
     * <code>cb2app.properties</code> file.
     */
    protected static final BTreeMap NONCHANGABLE_PROPS;
    static {
        NONCHANGABLE_PROPS = new BTreeMap();
        Object o = new Object();
        NONCHANGABLE_PROPS.put("com.boylesoftware.cb2.closeJNDIContexts", o);
        NONCHANGABLE_PROPS.put("com.boylesoftware.cb2.dataSource.", o);
        NONCHANGABLE_PROPS.put("com.boylesoftware.cb2.appProperties.", o);
    }


    /**
     * Name of the data source corresponding to the database where the
     * properties are stored. Set in the <code>init</code> method from
     * <code>com.boylesoftware.cb2.appProperties.dataSource</code> application
     * property. If the property is not defined, set to "default".
     */
    protected String dataSourceName;

    /**
     * Name of the database table, which contains application properties. Set
     * in the <code>init</code> method from the "table" BLO initialization
     * parameter.
     */
    protected String table;

    /**
     * Name of the column, which contains application property names. Set
     * in the <code>init</code> method from the "nameColumn" BLO initialization
     * parameter.
     */
    protected String nameColumn;

    /**
     * Name of the column, which contains application property values. Set
     * in the <code>init</code> method from the "valueColumn" BLO
     * initialization parameter.
     */
    protected String valueColumn;

    /**
     * SQL query used to insert new properties;
     */
    private String insertPropertyQuery;

    /**
     * SQL query used to update existing properties;
     */
    private String updatePropertyQuery;

    /**
     * SQL query used to delete properties;
     */
    private String deletePropertyQuery;


    /**
     * Initializes the BLO.
     */
    protected void init()
        throws BLException {

        if(this.getApplicationContext().
            getProperty("com.boylesoftware.cb2.appProperties.query") == null)
            throw new BLException("No com.boylesoftware.cb2.appProperties.query " +
                                  "application property is defined, the BLO is unusable.");

        this.dataSourceName = this.getApplicationContext().
            getProperty("com.boylesoftware.cb2.appProperties.dataSource");
        if(this.dataSourceName == null)
            this.dataSourceName = "default";

        this.table = this.getInitParameter("table");
        if(this.table == null)
            throw new BLException("Required BLO init parameter \"table\" is not defined.");
        this.nameColumn = this.getInitParameter("nameColumn");
        if(this.nameColumn == null)
            throw new BLException("Required BLO init parameter \"nameColumn\" is not defined.");
        this.valueColumn = this.getInitParameter("valueColumn");
        if(this.valueColumn == null)
            throw new BLException("Required BLO init parameter \"valueColumn\" is not defined.");

        this.insertPropertyQuery =
            "INSERT INTO " + this.table + " (" + this.nameColumn + ", " +
            this.valueColumn + ") VALUES (?, ?)";
        this.updatePropertyQuery =
            "UPDATE " + this.table + " SET " + this.valueColumn + " = ? WHERE " +
            this.nameColumn + " = ?";
        this.deletePropertyQuery =
            "DELETE FROM " + this.table + " WHERE " + this.nameColumn + " = ?";
    }


    /**
     * Tells the specified application property's type, which is where its
     * value comes from.
     *
     * @param name the application property's name.
     *
     * @return one of <code>PROPTYPE_<i>xxx</i></code> values.
     */
    public int getPropertyType(String name) {

        ApplicationContext appCtx = this.getApplicationContext();

        Properties props = appCtx.getProperties();
        if(!props.containsKey(name))
            return PROPTYPE_UNKNOWN;

        if(NONCHANGABLE_PROPS.get(name) != null)
            return PROPTYPE_IMMUTABLE;

        String curVal = props.getProperty(name);
        String basVal = appCtx.getBasicProperty(name);

        if(curVal.equals(basVal))
            return PROPTYPE_BASIC;

        if(basVal != null)
            return PROPTYPE_OVERRIDDEN;

        return PROPTYPE_DB;
    }

    /**
     * Gets all application properties as a sorted map.
     *
     * @return map with all application properties sorted by their names.
     */
    public SortedMap getProperties() {

        return new TreeMap(this.getApplicationContext().getProperties());
    }

    /**
     * Saves properties from the specified set of properties that have
     * different from current application properties values to the database.
     * It also saves properties that are not currently defined at all. This
     * method can insert new values into the database, and it can update
     * existing values. When all database operations have been finished
     * successfully, the method calls <code>setProperty</code> on the
     * applicatrion context for all changed properties.
     *
     * @param newProps collection of application property values.
     *
     * @throws BLException if a database error happens, or in case of an
     * attempt to change an immutable property.
     */
    public void saveProperties(Properties newProps)
        throws BLException {

        ApplicationContext appCtx = this.getApplicationContext();
        Properties props = appCtx.getProperties();
        Map propsToSet = new HashMap(newProps.size());

        Iterator i = newProps.entrySet().iterator();
        while(i.hasNext()) {
            Map.Entry entry = (Map.Entry)i.next();
            String propName = (String)entry.getKey();
            String propValue = (String)entry.getValue();
            if(!(props.containsKey(propName) &&
                 propValue.equals(props.getProperty(propName)))) {

                if(NONCHANGABLE_PROPS.get(propName) != null)
                    throw new BLException("Application property \"" + propName +
                                          "\" is immutable, you cannot change its value.");

                this.log.debug("updating application property \"" + propName +
                               "\" in the database");
                int n = this.getDAO().
                    executeUpdate(this.updatePropertyQuery,
                                  new Object [] { propValue, propName });
                if(n == 0) {
                    this.log.debug("adding new application property \"" + propName +
                                   "\" to the database");
                    this.getDAO().
                        executeUpdate(this.insertPropertyQuery,
                                      new Object [] { propName, propValue });
                }

                propsToSet.put(propName, propValue);
            }
        }

        i = propsToSet.entrySet().iterator();
        while(i.hasNext()) {
            Map.Entry entry = (Map.Entry)i.next();
            appCtx.setProperty((String)entry.getKey(),
                               (String)entry.getValue());
        }
    }

    /**
     * Resets an application property's value. This method deletes the value
     * from the database, which in the case of a database property removes it
     * from the set of application properties, and in the case of an overridden
     * basic property resets its value to the original one as defined in the
     * <code>cb2app.properties</code> file. This method has no effect if the
     * property is not stored in the database and it calls
     * <code>setProperty</code> on the application context if any change has
     * been made.
     *
     * @param name the application property's name.
     *
     * @throws BLException if a database error happens.
     */
    public void resetProperty(String name)
        throws BLException {

        int n = this.getDAO().
            executeUpdate(this.deletePropertyQuery,
                          new Object [] { name });

        if(n > 0) {
            ApplicationContext appCtx = this.getApplicationContext();
            String origValue = appCtx.getBasicProperty(name);
            appCtx.setProperty(name, origValue);
        }
    }
}
