package com.boylesoftware.cb2.commons.user;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.security.Principal;

import com.boylesoftware.cb2.BLObject;
import com.boylesoftware.cb2.AuthenticationListener;
import com.boylesoftware.cb2.BLErrors;
import com.boylesoftware.cb2.BLException;


/**
 * Generic implementation of an application user account business
 * entity. This BLO allows to login/logout users and manage their
 * accounts.
 *
 * <p>This BLO requires the following initialization parameters:
 * <ul>
 * <li><i>fetchUserDetailsByLoginNameQN</i> -- names the query in the
 * configuration file used to fetch user details data model by a
 * user's login name. The data model returned by the query should be
 * <code>UserBasicDM</code> or derived from it.</li>
 * <li><i>fetchUserDescriptorsQN</i> -- names the query in the
 * configuration file used to fetch descriptors for all available in
 * the system user accounts. The data models returned by the query
 * should be <code>UserBasicDM</code> or derived from it.</li>
 * </ul>
 *
 * <p>This BLO implements <code>SecurityBLI</code> interface so it can
 * be used with CB2's support for Servlet Specification user
 * authentication/authorization when deployed under name "user".
 *
 * @author Lev Himmelfarb
 * @version $Id: UserBLO.java,v 1.2 2004/10/11 18:17:36 levahim Exp $
 *
 * @see UserBasicDM
 * @see AuthenticationListener
 */
public class UserBLO
    extends BLObject
    implements AuthenticationListener {

    /**
     * Base value for error codes.
     */
    protected final static int ERR_BASE = 100000;

    /**
     * Error code indicating that the login name was invalid.
     */
    public final static int ERR_INVALID_USERNAME = ERR_BASE + 0;

    /**
     * Error code indicating that the password was invalid.
     */
    public final static int ERR_INVALID_PASSWORD = ERR_BASE + 1;

    /**
     * Error code indicating that the user is already logged in and
     * you are trying to log in again.
     */
    public final static int ERR_ALREADY_LOGGEDIN = ERR_BASE + 2;


    /**
     * List of all registered UserSessionStateEvent listeners.
     */
    private List userSessionStateListeners;


    /**
     * Name of the select query in the configuration file which is
     * used for getting a user details data model by the user's login
     * name. This field is populated from
     * <i>fetchUserDetailsByLoginNameQN</i> BLO initialization
     * parameter.
     */
    protected String fetchUserDetailsByLoginNameQN;

    /**
     * DM of the logged in user or <code>null</code> if no user is
     * logged in.
     */
    protected UserBasicDM loggedInUserDM;


    ///////////////////////////////////////////////////////////////////////////
    // Initialization
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Initializes the BLO. Call this method if you override it in
     * your child class.
     *
     * @throws BLException indicates a serious error, the BLO can not
     * be initialized.
     */
    protected void init()
        throws BLException {

        this.fetchUserDetailsByLoginNameQN = this.getInitParameter("fetchUserDetailsByLoginNameQN");
        if(this.fetchUserDetailsByLoginNameQN == null)
            throw new BLException("BLO initialization parameter \"fetchUserDetailsByLoginNameQN\" is required.");

        this.userSessionStateListeners = new ArrayList();

        this.loggedInUserDM = null;
    }

    /**
     * Free all used by the BLO resources. Call this method if you
     * override it in your child class.
     */
    protected void destroy() {

        this.loggedInUserDM = null;

        this.userSessionStateListeners.clear();
        this.userSessionStateListeners = null;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Events
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Registers the specified listener for user session state events.
     *
     * @param listener instance of your listener which will be
     * receiving the notifications.
     */
    public void addUserSessionStateListener(UserSessionStateListener listener) {

        this.userSessionStateListeners.add(listener);
    }

    /**
     * Removes the specified listener for user session state events
     * from the internal list of listeners.
     *
     * @param listener instance of the listener to be removed from the
     * notification list.
     */
    public void removeUserSessionStateListener(UserSessionStateListener listener) {

        this.userSessionStateListeners.remove(listener);
    }

    /**
     * Call <code>userLoggedIn()</code> method on all registered user
     * session state event listeners.
     *
     * @param userId id of the user logged in.
     *
     * @throws BLException if any of the listeners threw it on
     * <code>userLoggedIn()</code> method call.
     */
    protected void fireUserLoggedInEvent(long userId)
        throws BLException {

        UserSessionStateEvent usse = new UserSessionStateEvent(this, userId);

        Iterator i = this.userSessionStateListeners.iterator();
        while(i.hasNext())
            ((UserSessionStateListener)i.next()).userLoggedIn(usse);
    }

    /**
     * Call <code>userLoggedOut()</code> method on all registered user
     * session state event listeners.
     *
     * @param userId id of the user logged out.
     *
     * @throws BLException if any of the listeners threw it on
     * <code>userLoggedOut()</code> method call.
     */
    protected void fireUserLoggedOutEvent(long userId)
        throws BLException {

        UserSessionStateEvent usse = new UserSessionStateEvent(this, userId);

        Iterator i = this.userSessionStateListeners.iterator();
        while(i.hasNext())
            ((UserSessionStateListener)i.next()).userLoggedOut(usse);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Main Service Interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Tries to log in the specified by login name user. If the method
     * returned <code>false</code> use <code>getLastErrors()</code>
     * method to get the reason why login was unsuccessful. Possible
     * error codes this method sets are
     * <code>ERR_INVALID_USERNAME</code>,
     * <code>ERR_INVALID_PASSWORD</code> and
     * <code>ERR_ALREADY_LOGGEDIN</code>.
     *
     * <p>Use this method only if you implement user
     * authentication/authorization yourself. If you use Servlet
     * Specification security, that is an
     * <code>SecurityAuthenticator</code> implementation - use the
     * <code>SecurityBLI</code> interface instead.
     *
     * @param username login name of the user to log in.
     * @param password password to be used for log in.
     *
     * @return <code>true</code> if the user was successfully
     * authenticated and logged in.
     *
     * @throws BLException indicates a serious problem in the system.
     */
    public boolean login(String username, String password)
        throws BLException {

        BLErrors errors = new BLErrors();

        // are we already logged in?
        if(this.isLoggedIn()) {
            errors.addError(ERR_ALREADY_LOGGEDIN);
            this.setErrors(errors);
            return false;
        }

        // fetch the user account data
        UserBasicDM [] dms = (UserBasicDM [])this.getDAO().fetch(this.fetchUserDetailsByLoginNameQN,
                                                                 new Object [] { username });

        // did we find the user?
        if(dms.length == 0) {
            errors.addError(ERR_INVALID_USERNAME);
            this.setErrors(errors);
            this.log.info("Unsuccessful user login attempt: " + username + "/" + password);
            return false;
        }

        // check the password
        if(!dms[0].password.equals(password)) {
            errors.addError(ERR_INVALID_PASSWORD);
            this.setErrors(errors);
            this.log.info("Unsuccessful user login attempt: " + username + "/" + password);
            return false;
        }

        // register login
        this.registerLogin(dms[0]);

        // successful login
        return true;
    }

    /**
     * Logs out the currently logged in user. The BLO changes to the
     * non-autheticated "guest" mode.
     */
    public void logout() {

        if(!this.isLoggedIn())
            return;

        try {
            this.fireUserLoggedOutEvent(this.loggedInUserDM.userId);
        } catch(BLException e) {
            this.log.error("Caught BLException sending logout notification event, ignoring it.", e);
        }

        this.log.debug("User " + this.loggedInUserDM.loginName +
                       " [" + this.loggedInUserDM.userId + "] has logged out");

        this.loggedInUserDM = null;
    }

    /**
     * Returns <code>true</code> if a user is authenticated and
     * successfully logged in meaning that the BLO is in authenticated
     * mode.
     *
     * @return <code>true</code> if the user is logged in.
     */
    public boolean isLoggedIn() {

        return (this.loggedInUserDM != null);
    }

    /**
     * Returns <code>true</code> if currently logged in user is in the
     * specified by name role. If no user is currently logged in this
     * method always returns <code>false</code>. Also, if a user is
     * logged in this method always returns <code>true</code> for role
     * "user".
     *
     * @param role name of the role to check. Note that the role name
     * is case-sensitive.
     *
     * @return <code>true</code> if the user is in the specified role.
     */
    public boolean isInRole(String role) {

        if(!this.isLoggedIn())
            return false;

        if(role.equals("user"))
            return true;

        for(int i = 0; i < this.loggedInUserDM.roles.length; i++)
            if(role.equals(this.loggedInUserDM.roles[i].name))
                return true;

        return false;
    }


    ///////////////////////////////////////////////////////////////////////////
    // The AuthenticationListener Interface Implementation
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Notify the BLO about a user login event. This implementation
     * fetches the user's DM from the database using the configured
     * DAO query (the <code>fetchUserDetailsByLoginNameQN</code>
     * init parameter) and registers the login by a
     * <code>registerLogin</code> method call.
     *
     * @param principal the logged in principal.
     *
     * @throws BLException if an error happened and the user cannot be
     * logged in.
     */
    public void processUserLogin(Principal principal)
        throws BLException {

        // fetch the user account data
        UserBasicDM [] dms = (UserBasicDM [])this.getDAO().fetch(this.fetchUserDetailsByLoginNameQN,
                                                                 new Object [] { principal.getName() });

        // did we find the user? should never happen
        if(dms.length == 0)
            throw new BLException("There is no user with login name \"" + principal.getName() + "\".");

        // register the login
        this.registerLogin(dms[0]);
    }

    /**
     * Notify the BLO about a user logout event. This implementation
     * just calls <code>logout</code>.
     */
    public void processUserLogout() {

        this.logout();
    }


    ///////////////////////////////////////////////////////////////////////////
    // Internal Service Methods
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Performes all internal BLO operations for a user login action.
     * This implementation sets the <code>loggedInUserDM</code>
     * internal field and fires "user logged in" event by calling
     * <code>fireUserLoggedInEvent</code> protected method.
     *
     * @param dm the logged in user's DM.
     *
     * @throws BLException if the login could not be processed.
     */
    protected void registerLogin(UserBasicDM dm)
        throws BLException {

        // cache the account info internally
        this.loggedInUserDM = dm;

        // notify the listeners
        try {
            this.fireUserLoggedInEvent(this.loggedInUserDM.userId);
        } catch(BLException e) {
            // we are not logged in unless everybody's happy about it
            this.loggedInUserDM = null;
            throw e;
        }

        // log the event
        this.log.debug("User " + this.loggedInUserDM.loginName +
                       " [" + this.loggedInUserDM.userId + "] has logged in");
    }
}
