package com.boylesoftware.cb2.db.tsql;

import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import com.boylesoftware.cb2.GenericFetchExecutor;


/**
 * Fetch executor implementation for Sybase and MS SQL Server that
 * does not use scrollable result sets for paginated fetches.
 *
 * <p>Some JDBC driver implementations have issues with using
 * scrollable result sets, for example Sybase's jConnect reads the
 * whole result set into the client's memory when the
 * <code>ResultSet.last()</code> method, used by the
 * <code>GenericFetchExecutor</code> to determine the total rows
 * count, is executed, which slows down the performance and may cause
 * out of memory errors. This fetch executor implementation avoids
 * using scrollable result sets by performing a <code>SELECT
 * INTO</code> based on the original request and copying data into a
 * temporary table first, and then selecting the requested page from
 * that temporary table.
 *
 * <p>Please note, that since this executor uses <code>SELECT
 * INTO</code> the DAO's fetch methods cannot be called from within a
 * transaction.
 *
 * <p>Also note, that the field name separator for the DAO, which uses
 * this fetch executor implementation, HAS TO BE a dollar sign ($).
 *
 * <p>This fetch executor allows paginated result sets containing no
 * more than a million rows.
 *
 * @author Lev Himmelfarb
 * @version $Id: NoCursorFetchExecutor.java,v 1.4 2004/04/20 16:00:58 levahim Exp $
 */
public class NoCursorFetchExecutor
    extends GenericFetchExecutor {

    /**
     * Name of the temporary table to use.
     */
    protected final static String TMP_TABLE_NAME = "#GFETmp";

    /**
     * Regex pattern used to parse and modify the query.
     */
    protected final static Pattern REP_PATTERN =
        Pattern.compile("^(.*?)(SELECT\\s)(.+?)(\\sFROM\\s)", Pattern.DOTALL | Pattern.CASE_INSENSITIVE);

    /**
     * Replacement string for the query modification.
     */
    protected final static String REP_STRING =
        "$1SELECT recn\\$ = IDENTITY(6), $3 INTO " + TMP_TABLE_NAME + " FROM ";

    /**
     * SQL query used to drop the temp table.
     */
    protected final static String DROP_TEMP_TABLE_QUERY =
        "DROP TABLE " + TMP_TABLE_NAME;

    /**
     * SQL query used to select the final result from the temporary
     * table. Part 1.
     */
    protected final static String SELECT_PAGE_QUERY_1 =
        "SELECT * FROM " + TMP_TABLE_NAME + " WHERE recn$ BETWEEN ";

    /**
     * SQL query used to select the final result from the temporary
     * table. Part 2.
     */
    protected final static String SELECT_PAGE_QUERY_2 =
        " AND ";


    /**
     * Prepares a statement object before the <code>execute</code>
     * method call. If a paginated result set is requested, this
     * implementation actually modifies the original query by adding a
     * <code>INTO</code> clause to it.
     *
     * @return the statement object, which may be an instance of
     * <code>PreparedStatement</code> if <code>queryParams</code> is
     * specified.
     *
     * @throws SQLException if a database error happened.
     */
    public Statement prepareStatement()
        throws SQLException {

        // do special preparations for paginated requests
        if(this.pageSize > 0) {

            // try to drop the temp table
            Statement tmpstmt = null;
            try {
                tmpstmt = this.con.createStatement();
                tmpstmt.execute(DROP_TEMP_TABLE_QUERY);
            } catch(SQLException e) {
                // ignore it
            } finally {
                if(tmpstmt != null)
                    tmpstmt.close();
            }

            // modify the original query
            Matcher m = REP_PATTERN.matcher(this.query);
            if(!m.find())
                throw new SQLException("Could not parse the input query.");
            StringBuffer res = new StringBuffer();
            m.appendReplacement(res, REP_STRING);
            m.appendTail(res);
            this.query = res.toString();
        }

        // prepare the statement
        this.stmt = (this.usePreparedStatement ?
                     this.con.prepareStatement(this.query) :
                     this.con.createStatement());

        // done
        return this.stmt;
    }

    /**
     * Executes the query and returns the statement actually used for
     * getting the result set. The <code>getResultSet</code> method
     * should be called to get the result set.
     *
     * @return the statement owning the result set.
     *
     * @throws SQLException if a database error happened.
     */
    public Statement execute()
        throws SQLException {

        // if paginated result set is requested
        if(this.pageSize > 0) {

            // do the select into
            int nRows = (this.usePreparedStatement ?
                         ((PreparedStatement)this.stmt).executeUpdate() :
                         this.stmt.executeUpdate(this.query));

            // close the select into statement
            this.stmt.close();
            this.stmt = null;

            // set the total rows
            if(this.fetchResultDescriptor != null)
                this.fetchResultDescriptor.setRowsTotal(nRows);

            // prepare the select from temp table statement
            this.stmt = this.con.createStatement();
            int ps = this.page * this.pageSize;
            StringBuffer finalQuerySB = new StringBuffer(SELECT_PAGE_QUERY_1);
            finalQuerySB.append(ps + 1);
            finalQuerySB.append(SELECT_PAGE_QUERY_2);
            finalQuerySB.append(ps + this.pageSize);
            String finalQuery = finalQuerySB.toString();

            // select the data, execute the fetch
            try {
                this.rs = this.stmt.executeQuery(finalQuery);
            } catch(SQLException e) {
                this.stmt.close();
                throw e;
            }

        } else {

            // execute the fetch
            this.rs = (this.usePreparedStatement ?
                       ((PreparedStatement)this.stmt).executeQuery() :
                       this.stmt.executeQuery(this.query));
        }

        // return the used statement
        return this.stmt;
    }

    /**
     * Positions the result set cursor to the first row of the
     * requested data.
     *
     * @return <code>true</code> if the cursor is set or
     * <code>false</code> if there is not data (for example the result
     * set is empty or the requested page does not exist).
     *
     * @throws SQLException if a database error happened.
     */
    public boolean setToTheFirstRow()
        throws SQLException {

        if(!this.rs.next()) {

            if((this.pageSize <= 0) && (this.fetchResultDescriptor != null))
                this.fetchResultDescriptor.setRowsTotal(0);

            return false;
        }

        return true;
    }
}
