package com.boylesoftware.cb2.presentation.servlet;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Locale;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.struts.Globals;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionMessages;
import org.apache.struts.config.ModuleConfig;
import org.apache.struts.config.ActionConfig;
import org.apache.struts.util.MessageResources;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.BLOContainer;
import com.boylesoftware.cb2.BLObject;


/**
 * Represents context of a CB2 action execution within a request
 * processing. A correct instance of this class is always provided to
 * actions and presentation elements.
 *
 * @author Lev Himmelfarb
 * @version $Id: ActionContext.java,v 1.13 2004/10/11 18:17:36 levahim Exp $
 */
public class ActionContext {

    /**
     * Indicates the web-application scope for <code>setAttribute</code>,
     * <code>getAttribute</code> and <code>removeAttribute</code> methods.
     */
    public static final int APPLICATION_SCOPE = 1;

    /**
     * Indicates the HTTP session scope for <code>setAttribute</code>,
     * <code>getAttribute</code> and <code>removeAttribute</code> methods.
     */
    public static final int SESSION_SCOPE = 2;

    /**
     * Indicates the HTTP request scope for <code>setAttribute</code>,
     * <code>getAttribute</code> and <code>removeAttribute</code> methods.
     */
    public static final int REQUEST_SCOPE = 3;


    /**
     * An empty hash map.
     */
    private static final Map EMPTY_MAP = new HashMap(0);


    /**
     * Used to call Struts action methods.
     */
    private static final class FakeAction
        extends org.apache.struts.action.Action {

        /**
         * Protected method wrapper.
         */
        public Locale getLocale(HttpServletRequest request) {

            return super.getLocale(request);
        }

        /**
         * Protected method wrapper.
         */
        public MessageResources getResources(HttpServletRequest request) {

            return super.getResources(request);
        }

        /**
         * Protected method wrapper.
         */
        public MessageResources getResources(HttpServletRequest request,
                                             String key) {

            return super.getResources(request, key);
        }

        /**
         * Protected method wrapper.
         */
        public void saveErrors(HttpServletRequest request,
                               ActionErrors errors) {

            super.saveErrors(request, errors);
        }

        /**
         * Protected method wrapper.
         */
        public void saveMessages(HttpServletRequest request,
                                 ActionMessages messages) {

            super.saveMessages(request, messages);
        }
    }

    /**
     * Used to call stateless Struts action methods.
     */
    private static final FakeAction FAKE_ACTION = new FakeAction();


    /**
     * Reference to the current request processing context.
     */
    private RequestProcessingContext requestCtx;

    /**
     * Action mapping used to select the action.
     */
    private ActionMapping actionMapping;

    /**
     * Form bean submitted to the action or <code>null</code> if no
     * form is required by the action.
     */
    private ActionForm form;

    /**
     * The HTTP request.
     */
    private HttpServletRequest request;

    /**
     * The HTTP response.
     */
    private HttpServletResponse response;

    /**
     * BLO container for the current session.
     */
    private BLOContainer bloContainer;

    /**
     * Log to be used in the action.
     */
    private Log log;

    /**
     * Reference to the Struts module configuration.
     */
    private ModuleConfig moduleConfig;


    /**
     * Resets all internal fields.
     */
    void reset() {

        this.requestCtx = null;
        this.actionMapping = null;
        this.form = null;
        this.request = null;
        this.response = null;
        this.bloContainer = null;
        this.log = null;
        this.moduleConfig = null;
    }


    /**
     * Sets internal fields for the context.
     *
     * @param moduleConfig reference to the module config.
     * @param request the HTTP request.
     * @param response the HTTP response.
     * @param actionMapping action mapping used to select the action.
     * @param form form bean submitted to the action or
     * <code>null</code> if no form is required by the action.
     * @param requestCtx reference to the request processing context
     * instance.
     * @param bloContainer BLO container for the current session.
     * @param log Log to be used in the action.
     */
    void setup(ModuleConfig moduleConfig,
               HttpServletRequest request,
               HttpServletResponse response,
               ActionMapping actionMapping,
               ActionForm form,
               RequestProcessingContext requestCtx,
               BLOContainer bloContainer,
               Log log) {

        this.requestCtx = requestCtx;
        this.actionMapping = actionMapping;
        this.form = form;
        this.request = request;
        this.response = response;
        this.bloContainer = bloContainer;
        this.log = log;
        this.moduleConfig = moduleConfig;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Request context access
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets the application context.
     *
     * @return reference to the application context.
     */
    public ApplicationContext getApplicationContext() {

        return this.requestCtx.getApplicationContext();
    }

    /**
     * Returns the specified by name BLO from the
     * <code>BLManager</code>'s shared BLO container or
     * <code>null</code>. The BLO names are case-insensitive.
     *
     * <p>In the most cases you should use the <code>getBLO</code>
     * method on the <code>BLOContainer</code> from the action
     * context, which can return both regular and shared BLOs. This
     * method is here for those rare cases, when you do not have
     * access to the <code>BLOContainer</code> but still need access
     * to some shared BLOs. Such a situation is possible, for example,
     * in a detached action.
     *
     * @param bloName the requested BLO's deployment name,
     * case-insensitive.
     *
     * @return reference to the shared BLO or <code>null</code>.
     *
     * @see BLOContainer#getBLO
     */
    public BLObject getSharedBLO(String bloName) {

        return this.requestCtx.getBLManager().getSharedBLO(bloName);
    }

    /**
     * Gets complete URL requested by the client browser. This
     * method returns the very first requested URL regardless of any
     * forwards or includes during the request processing.
     *
     * @return complete URL requested by the client browser.
     */
    public String getRequestURL() {

        return this.requestCtx.getRequestURL();
    }

    /**
     * Gets the UIState bean instance for the current session.
     *
     * @return reference to the UIState bean instance.
     */
    public UIState getUIState() {

        return this.requestCtx.getUIState();
    }

    /**
     * Finds corresponding BL session id for a request. The session
     * ids in the <code>BLManager</code> are not the same as the HTTP
     * session ids assigned by the Servlet Container. This method allows
     * to find the BL session id for the HTTP request being processed.
     *
     * @return the BL session id.
     */
    public String getBLSessionId() {

        return this.requestCtx.getBLSessionId();
    }


    ///////////////////////////////////////////////////////////////////////////
    // Access to action context specific data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets action mapping used to select the action.
     *
     * @return action mapping used to select the action.
     */
    public ActionMapping getActionMapping() {

        return this.actionMapping;
    }

    /**
     * Gets form bean submitted to the action or <code>null</code> if
     * no form is required by the action.
     *
     * @return form bean submitted to the action or <code>null</code>.
     */
    public ActionForm getForm() {

        return this.form;
    }

    /**
     * Gets the HTTP request.
     *
     * @return the HTTP request.
     */
    public HttpServletRequest getRequest() {

        return this.request;
    }

    /**
     * Gets the HTTP response.
     *
     * @return the HTTP response.
     */
    public HttpServletResponse getResponse() {

        return this.response;
    }

    /**
     * Gets BLO container for the current session.
     *
     * @return BLO container for the current session.
     */
    public BLOContainer getBLOContainer() {

        return this.bloContainer;
    }

    /**
     * Gets log to be used in the action.
     *
     * @return log to be used in the action.
     */
    public Log getLog() {

        return this.log;
    }

    /**
     * Gets the Struts module config.
     *
     * @return reference to the module config.
     */
    ModuleConfig getModuleConfig() {

        return this.moduleConfig;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Access to action service interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets the user's currently selected locale. The same as Struts'
     * <code>Action.getLocale</code> method.
     *
     * @return the user's locale.
     */
    public Locale getLocale() {

        return FAKE_ACTION.getLocale(this.request);
    }

    /**
     * Gets the default message resources for the current module. The same
     * as Struts' <code>Action.getResources</code> method.
     *
     * @return the message resources.
     */
    public MessageResources getResources() {

        return FAKE_ACTION.getResources(this.request);
    }

    /**
     * Gets the specified message resources for the current module. The
     * same as Struts' <code>Action.getResources</code> method.
     *
     * @param key the key specified in the
     * <code>&lt;message-resources&gt;</code> element for the requested bundle.
     *
     * @return the message resources.
     */
    public MessageResources getResources(String key) {

        return FAKE_ACTION.getResources(this.request, key);
    }

    /**
     * Save the specified error messages keys into the appropriate request
     * attribute for use by the <code>&lt;html:errors&gt;</code> tag, if any
     * messages are required. Otherwise, ensure that the request attribute
     * is not created. The same as Struts' <code>Action.saveErrors</code>
     * method.
     *
     * @param errors errors to set, <code>null</code> or empty structure
     * to clear the errors.
     */
    public void saveErrors(ActionErrors errors) {

        FAKE_ACTION.saveErrors(this.request, errors);
    }

    /**
     * Save the specified messages keys into the appropriate request attribute
     * for use by the <code>&lt;html:messages&gt;</code> tag
     * (if messages="true" is set), if any messages are required. Otherwise,
     * ensure that the request attribute is not created. The same as Struts'
     * <code>Action.saveMessages</code> method.
     *
     * @param messages messages to set, <code>null</code> or empty structure
     * to clear the messages.
     */
    public void saveMessages(ActionMessages messages) {

        FAKE_ACTION.saveMessages(this.request, messages);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Additional service to actions
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets the named BLO. This is just a convenience method and its effect
     * is exactly the same as calling
     * <code>getBLOContainer().getBLO(bloName)</code>.
     *
     * @param bloName the BLO's deployment name.
     *
     * @return the BLO instance.
     *
     * @see BLOContainer#getBLO
     */
    public BLObject getBLO(String bloName) {

        return this.bloContainer.getBLO(bloName);
    }

    /**
     * Same as the <code>saveErrors</code> method, but does not overwrite
     * any existing errors in the request. If there are already some errors
     * in the request this method will add the specified bundle to them.
     *
     * @param errors the errors to set/add.
     */
    public void addErrors(ActionErrors errors) {

        ActionErrors origErrors = (ActionErrors)this.request.
            getAttribute(Globals.ERROR_KEY);

        if(origErrors == null) {
            FAKE_ACTION.saveErrors(this.request, errors);
        } else {
            origErrors.add(errors);
        }
    }

    /**
     * Same as the <code>saveMessages</code> method, but does not overwrite
     * any existing messages in the request. If there are already some messages
     * in the request this method will add the specified bundle to them.
     *
     * @param messages the messages to set/add.
     */
    public void addMessages(ActionMessages messages) {

        ActionMessages origMessages = (ActionMessages)this.request.
            getAttribute(Globals.MESSAGE_KEY);

        if(origMessages == null) {
            FAKE_ACTION.saveMessages(this.request, messages);
        } else {
            origMessages.add(messages);
        }
    }

    /**
     * Convenience method, which calls <code>findForward</code> on the
     * action mapping in this context.
     *
     * @param name name of the action forward to find.
     *
     * @return named action forward from the context's action mapping.
     */
    public ActionForward findForward(String name) {

        return this.actionMapping.findForward(name);
    }

    /**
     * Finds action mapping for the named page from pages configuration file.
     * Note that this method looks only among pages defined for this
     * Struts module.
     *
     * @param pageName name of the page.
     *
     * @return action mapping that points to the page or <code>null</code>
     * if there is no page with the specified name.
     */
    public ActionConfig findPageMapping(String pageName) {

        PageDesc page = CB2PlugIn.getModule(this.request.getSession().getServletContext(),
                                        this.moduleConfig).
            getPLConfig().getPage(pageName);

        return (page != null ?
                this.moduleConfig.findActionConfig(page.path) :
                null);
    }

    /**
     * A shortcut method returning a named HTTP request parameter. Calling
     * this method is the same as calling
     * <code>actionCtx.getRequest().getParameter(paramName)</code>.
     *
     * @param paramName the HTTP request parameter name.
     *
     * @return the named parameter value or <code>null</code>.
     */
    public String getRequestParameter(String paramName) {

        return this.request.getParameter(paramName);
    }

    /**
     * Gets an attribute value from the specified scope.
     *
     * @param attrName name of the attribute to retrieve.
     * @param scope the scope, one of <code>xxx_SCOPE</code> constants.
     *
     * @return the attribtue value or <code>null</code>.
     */
    public Object getAttribute(String attrName, int scope) {

        switch(scope) {
        case APPLICATION_SCOPE:
            return this.request.getSession().getServletContext().getAttribute(attrName);
        case SESSION_SCOPE:
            return this.request.getSession().getAttribute(attrName);
        case REQUEST_SCOPE:
            return this.request.getAttribute(attrName);
        }

        throw new IllegalArgumentException("The argument scope should be one of ActionContext.xxx_SCOPE constants.");
    }

    /**
     * Sets an attribute in the specified scope.
     *
     * @param attrName name of the attribute to set.
     * @param value the value to set in the attribute.
     * @param scope the scope, one of <code>xxx_SCOPE</code> constants.
     */
    public void setAttribute(String attrName, Object value, int scope) {

        switch(scope) {
        case APPLICATION_SCOPE:
            this.request.getSession().getServletContext().setAttribute(attrName, value);
            break;
        case SESSION_SCOPE:
            this.request.getSession().setAttribute(attrName, value);
            break;
        case REQUEST_SCOPE:
            this.request.setAttribute(attrName, value);
            break;
        default:
            throw new IllegalArgumentException("The argument scope should be one of ActionContext.xxx_SCOPE constants.");
        }
    }

    /**
     * Removes an attribute from the specified scope.
     *
     * @param attrName name of the attribute to remove.
     * @param scope the scope, one of <code>xxx_SCOPE</code> constants.
     */
    public void removeAttribute(String attrName, int scope) {

        switch(scope) {
        case APPLICATION_SCOPE:
            this.request.getSession().getServletContext().removeAttribute(attrName);
            break;
        case SESSION_SCOPE:
            this.request.getSession().removeAttribute(attrName);
            break;
        case REQUEST_SCOPE:
            this.request.removeAttribute(attrName);
            break;
        default:
            throw new IllegalArgumentException("The argument scope should be one of ActionContext.xxx_SCOPE constants.");
        }
    }

    /**
     * Calls <code>addComponentRef(refName, componentName, null)</code>.
     *
     * @param refName reference name to be used with the
     * <code>&lt;cb2:insert&gt;</code> tag.
     * @param componentName name of the corresponding component as defined
     * in the <code>pages-config.xml</code>.
     *
     * @throws ServletException if no component is defined under the
     * specified name.
     *
     * @see #addComponentRef(String, String, Map)
     */
    public void addComponentRef(String refName,
                                String componentName)
        throws ServletException {

        this.addComponentRef(refName, componentName, null);
    }

    /**
     * Adds a component reference to the list of component references
     * associated with the page, which will be displayed as the result of
     * this request processing. The static list of component references
     * is associated with a page in its descriptors using
     * <code>&lt;componentref&gt;</code> elements. This method allows to add
     * additional component references to the list dynamically from actions,
     * page pre-actions and presentation elements. Note, that this method
     * adds a component reference to the end of the list, meaning that all
     * the corresponding PEs, including any global PEs on which the
     * component depends, will be initialized after all global and local
     * PEs required by any components refered by the page statically
     * via its descriptor, as well as after all PEs required by any
     * components already added to the list using this method.
     *
     * @param refName reference name to be used with the
     * <code>&lt;cb2:insert&gt;</code> tag.
     * @param componentName name of the corresponding component as defined
     * in the <code>pages-config.xml</code>.
     * @param refAttributes component reference attributes, which can be used
     * with component's presentation elements as input parameters with "from"
     * attribute set to "crefattribute" or "any". This argument can be
     * <code>null</code>.
     *
     * @throws ServletException if no component is defined under the
     * specified name.
     */
    public void addComponentRef(String refName,
                                String componentName,
                                Map refAttributes)
        throws ServletException {

        ComponentDesc component = CB2PlugIn.getModule(this.request.getSession().getServletContext(),
                                                      this.moduleConfig).
            getPLConfig().getComponent(componentName);
        if(component == null)
            throw new ServletException("No component is defined under name \"" +
                                       componentName + "\".");

        List addlCompRefs = (List)this.request.
            getAttribute(ShowPageAction.RQKEY_ADDL_COMPREFS);
        if(addlCompRefs == null) {
            addlCompRefs = new ArrayList();
            this.request.setAttribute(ShowPageAction.RQKEY_ADDL_COMPREFS,
                                      addlCompRefs);
        }

        addlCompRefs.add(new PageDesc.ComponentRefDesc(refName,
                                                       component,
                                                       (refAttributes != null ?
                                                        refAttributes :
                                                        EMPTY_MAP)));
    }
}
