package com.boylesoftware.cb2.presentation.servlet;

import java.util.regex.Pattern;
import java.io.IOException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.struts.action.Action;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.config.ModuleConfig;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.BLManager;
import com.boylesoftware.cb2.BLOContainer;
import com.boylesoftware.cb2.ApplicationException;
import com.boylesoftware.cb2.BLException;
import com.boylesoftware.cb2.NoBLOContainerException;
import com.boylesoftware.cb2.util.ExceptionUtils;


/**
 * Abstract parent class for CB2/Struts action handlers. This class
 * extends Struts' action class functionality and adds such features
 * as action context and database transaction management. Extending
 * this class makes easy to communicate with CB2 business level. In
 * the most cases you will need only to redefine the abstract
 * <code>execute</code> method in your concrete action handler
 * implementation.
 *
 * <p>Normally, execution of actions is serialized within a session,
 * meaning that if there is an action being executed in its
 * user-defined <code>execute</code> method and a new request comes
 * within the same session, then this new request will wait until the
 * first action is finished. This is implemented through locking the
 * BLO container right before calling the <code>execute</code> method
 * and releasing it right after. This behaviour can be overriden with
 * <code>[detached]</code> action flag (see below).
 *
 * <p>Unless <code>[noTX]</code> (see below) is specified, on
 * redefined <code>doAutoTransaction</code> method returns
 * <code>false</code>, a database transaction will be started before
 * calling <code>execute</code> method. If the <code>execute</code>
 * returns successfully and does not throw any exception, the
 * transaction will be committed right after its return. If it throws
 * an exception the transaction will be automatically rolled back.
 *
 * <p>You can specify special flags and pass them to the action with
 * <code>parameter</code> attribute of the corresponding
 * <code>action</code> element in the
 * <code>struts-config.xml</code>. Here is an example:
 *
 * <pre>
 * &lt;action
 *   path="/app/addRecord"
 *   type="my.application.actions.GotoAction"
 *   parameter="[noTX][detached]"&gt;
 *   &lt;forward name="success" path="addRecordForm"/&gt;
 * &lt;/action&gt;</pre>
 *
 * <p>This implementation recognizes the following flags:
 *
 * <ul>
 *
 * <li><code>[noTX]</code> - do not automatically manage database
 * transaction for the action. Usually you specify this flag for
 * actions that do not do anything with the database (through
 * BLOs). Note, that there is an alternative way to enable/disable
 * automatic transaction handling for an action: you can override the
 * <code>doAutoTransaction</code> method in your action
 * implementation. The <code>[noTX]</code> then will be ignored.</li>
 *
 * <li><code>[detached]</code> - do not aquire and lock the BLO
 * container for this action. The <code>bloContainer</code> property
 * (<code>getBLOContainer()</code> method on the
 * <code>ActionContext</code>) in the action context passed to to the
 * <code>execute</code> method will be <code>null</code>. You still
 * can get access to shared BLOs through the <code>getSharedBLO</code>
 * protected method. This flag is useful in situation when you need a
 * long running action. Since normally actions are executed
 * sequencially with the BLO container locking, it is highly
 * recommended to keep your actions as "short" as possible, so they
 * release the BLO container as soon as possible after they lock
 * it. Now, sometimes, you need an action that performes a long and
 * complex operation. Then you declare this action "detached" and it
 * won't have access to the session's BLO container and thus won't
 * lock it. If you need any data from business level, which is not
 * provided by your shared BLOs, you should fetch all that data in
 * advance in a regular action and then pass it to the "detached"
 * action in an application dependent manner. Note also, that you can
 * achieve the effect of this flag by redefining the
 * <code>isDetached</code> method in your action implementation.
 *
 * </ul>
 *
 * <p>Sometimes when multiple requests are being processed for the
 * same session, one of the actions may destroy the session (for
 * example user requested logout from a browser window while
 * request is still being processed in another window). If this
 * happens and the action cannot get BLO container, because it was
 * just destroyed, a special forward named "nosession" is found
 * and returned by the action without calling user defined
 * <code>execute</code> method. If no "nosession" forward is defined
 * an exception is thrown by the action. Note also, that the action
 * enforces "redirect" flag on the "nosession" forward.
 *
 * @author Lev Himmelfarb
 * @version $Id: CB2Action.java,v 1.15 2004/10/11 18:17:36 levahim Exp $
 */
public abstract class CB2Action
    extends Action {

    /**
     * Subsystem name prefix for the logger to be used in action
     * handlers. This implementation adds <code>getActionName</code>
     * method's result to this prefix to request a logger for the
     * action handler.
     */
    private final static String LOG_SUBSYS_PREFIX = "com.boylesoftware.cb2.ACTION";


    /**
     * CB2 implementation of the Struts' action <code>execute</code>
     * method. It prepares an action context and calls user-defined
     * <code>execute(ActionContext actionCtx)</code> method
     * internally. It also manages a database transaction if required.
     *
     * @param mapping the action mapping used to select this instance.
     * @param form form bean instance for this request. May be
     * <code>null</code> if no form is associated with this action in
     * the mapping.
     * @param request the HTTP request.
     * @param response the HTTP response.
     *
     * @throws ServletException if an error happened in the action.
     * @throws IOException if an input/output error happened in the
     * action.
     */
    public final ActionForward execute(ActionMapping mapping,
                                       ActionForm form,
                                       HttpServletRequest request,
                                       HttpServletResponse response)
        throws ServletException, IOException {

        // get the basic subsystems
        CB2Application app = CB2PlugIn.getApplication(request.getSession().getServletContext());
        RequestProcessingContext requestCtx = RequestProcessingContext.get(request);
        ApplicationContext appCtx = requestCtx.getApplicationContext();
        BLManager blManager = requestCtx.getBLManager();
        UserSession userSession = requestCtx.getUserSession();

        // create the log
        String actionName = this.getActionName();
        Log log = appCtx.getLog(actionName != null ?
                                LOG_SUBSYS_PREFIX + "." + actionName :
                                LOG_SUBSYS_PREFIX);

        // log the action call
        log.debug("called action " + this.getClass().getName());

        // get action mapping parameter with flags
        String parameter = mapping.getParameter();

        // check if we want transactions for this action
        boolean doTX = (this.doAutoTransaction(request, parameter) && appCtx.isDatabaseAvailable());
        boolean inTX = false;

        // register action call event
        long startedTs = userSession.getRequestStats().eventStarted();

        // do the action
        ActionForward res = null;
        try {

            // begin tx if needed
            if(doTX) {
                appCtx.beginTransaction();
                inTX = true;
            }

            // create action context
            BLOContainer bloContainer = null;
            ActionContext actionCtx = null;
            try {

                // get and lock the BLO container instance
                if(!this.isDetached(request, parameter))
                    bloContainer = blManager.getBLOContainer(requestCtx.getBLSessionId());

                // prepare action context
                actionCtx = app.getActionContext();
                actionCtx.setup((ModuleConfig)request.getAttribute(CB2RequestProcessor.MODULE_CONFIG),
                                request,
                                response,
                                mapping,
                                form,
                                requestCtx,
                                bloContainer,
                                log);

                // do the action
                res = this.execute(actionCtx);

            } catch(NoBLOContainerException e) {

                log.info("could not get BLO container for the session, " +
                         "trying to return \"nosession\" forward");
                res = mapping.findForward("nosession");
                if(res == null) {
                    log.warn("no \"nosession\" forward is defined, throwing exception");
                    throw e;
                }

            } finally {

                // release the BLO container
                if(bloContainer != null)
                    bloContainer.release();

                // return the action context back to the pool
                if(actionCtx != null) {
                    actionCtx.reset();
                    app.returnActionContext(actionCtx);
                }
            }

            // commit tx if needed
            if(inTX) {
                ApplicationContext.commitTransaction();
                inTX = false;
            }

        } catch(Throwable e) {

            // log the error
            log.error("Got error performing action.", e);

            // stats
            userSession.incrementNErrors(e);

            // try to rollback the tx if needed
            if(inTX) {
                try {
                    ApplicationContext.rollbackTransaction();
                } catch(ApplicationException e2) {
                    log.error("Got an error rolling back a transaction due to another error. The original cause was: " + e, e2);
                    throw new ServletException("Got an error rolling back a transaction due to another error.", e2);
                }
            }

            // convert the error and throw it up
            ExceptionUtils.
                rethrowServletException(e, "Got an error performing an action.");

        } finally {

            // register action processing time
            userSession.getRequestStats().eventFinished(startedTs);
        }

        // log successful action return
        log.debug("action is returning " + res);

        // return the result ActionForward
        return res;
    }

    /**
     * Redefine this method in your child action handler. It will be
     * called by the framework for each request mapped to this action
     * in the <code>struts-config.xml</code> configuration file.
     *
     * @param actionCtx action context.
     *
     * @throws BLException if an error happened in the business
     * level. Usually it is an exception thrown by a BLO called from
     * within the action.
     * @throws ServletException if an error happened in the action.
     * @throws IOException if an input/output error happened in the
     * action.
     */
    protected abstract ActionForward execute(ActionContext actionCtx)
        throws BLException, ServletException, IOException;


    ///////////////////////////////////////////////////////////////////////////
    // Overridable action properties
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Returns this action implementation's logical name. The default
     * implementation returns <code>null</code>.
     *
     * <p>By default the log passed to the user-defined
     * <code>execute</code> method with the action context is set to
     * "com.boylesoftware.cb2.ACTION" subsystem name. You can redefine
     * this method to return name which will be added to this prefix
     * after an additional dot to form subsystem name for the
     * logger. This allows you to have a different log configuration,
     * such as verbosity, on per action level.
     *
     * @return action name. This implementation returns
     * <code>null</code>.
     */
    protected String getActionName() {

        return null;
    }

    /**
     * Checks if <code>[noTX]</code> flag is specified in the action
     * mapping parameter and tells whether an automatic transaction
     * should be handled for the action.
     *
     * <p>You can redefine this method in your action implementation
     * to involve some other mechanism of telling whether the
     * trasaction should be handled automatically or not.
     *
     * @param request current request, the default implementation
     * ignores this parameter.
     * @param mappingParameter value of the Struts action mapping's
     * <code>parameter</code> attribute from the
     * <code>struts-config.xml</code> file. May be <code>null</code>.
     *
     * @return <code>true</code> if the action shuold be executed in a
     * transaction. This implementation returns <code>true</code> by
     * default when the <code>mappingParameter</code> argument is
     * <code>null</code> or does not contain <code>[noTX]</code>
     * substring.
     */
    protected boolean doAutoTransaction(HttpServletRequest request, String mappingParameter) {

        return ((mappingParameter == null) || (mappingParameter.indexOf("[noTX]") < 0));
    }

    /**
     * Checks if <code>[detached]</code> flag is specified in the
     * action mapping parameter and tells whether the session's BLO
     * container should be aquired and locked for the action
     * execution.
     *
     * <p>You can redefine this method in your action implementation
     * to involve some other mechanism of telling whether the action
     * should be "detached" or not.
     *
     * @param request current request, the default implementation
     * ignores this parameter.
     * @param mappingParameter value of the Struts action mapping's
     * <code>parameter</code> attribute from the
     * <code>struts-config.xml</code> file. May be <code>null</code>.
     *
     * @return <code>true</code> if the action should be
     * "detached". This implementation returns <code>false</code> by
     * default when the <code>mappingParameter</code> argument is
     * <code>null</code> and it returns <code>true</code> if the
     * <code>mappingParameter</code> is not <code>null</code> and
     * contains <code>[detached]</code> substring.
     */
    protected boolean isDetached(HttpServletRequest request, String mappingParameter) {

        return ((mappingParameter != null) && (mappingParameter.indexOf("[detached]") >= 0));
    }


    ///////////////////////////////////////////////////////////////////////////
    // Service to action implementations
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Logs out the currently authenticated in user. Calling this
     * method also results in the HTTP session invalidation at the end
     * of the request processing. If there is no authenticated user
     * this method still invalidates the session.
     *
     * @param actionCtx the action context.
     */
    protected final void logout(ActionContext actionCtx) {

        // put the logout token to the request
        actionCtx.getRequest().setAttribute(CB2RequestProcessor.LOGOUT_TOKEN, new Object());
    }


    ///////////////////////////////////////////////////////////////////////////
    // Validators
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Precompiled regular expression pattern for valid email
     * addresses.
     */
    private final static Pattern EMAIL_ADDRESS_PATTERN =
        Pattern.compile("^[a-zA-Z0-9][a-zA-Z0-9_.]*@[a-zA-Z0-9][a-zA-Z0-9_.]+[a-zA-Z0-9]$");


    /**
     * Checks if the specified value is a valid email address.
     *
     * @param value form field value to check.
     *
     * @return <code>true</code> if the specified string is not
     * <code>null</code> and is a valid email address (synthax-wise).
     *
     * @deprecated use
     * {@link com.boylesoftware.cb2.util.TextUtils#isValidEmailAddress} static
     * method instead.
     */
    protected boolean isValidEmailAddress(String value) {

        return (value != null) && EMAIL_ADDRESS_PATTERN.matcher(value).matches();
    }

    /**
     * Checks if the specified value is a valid US phone number.
     *
     * @param value form field to check.
     *
     * @return <code>true</code> if the specified string is not
     * <code>null</code> and is a valid US phone number.
     *
     * @deprecated no direct replacement for this method, should be
     * implemented in the application specific way.
     */
    protected boolean isValidUSPhoneNumber(String value) {

        if(value == null)
            return false;

        int nDigits = 0;
        char [] phoneChars = value.toCharArray();
        for(int i = 0; i < phoneChars.length; i++) {
            if(Character.isDigit(phoneChars[i]))
                nDigits++;
        }

        return nDigits == 10;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Request utils
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Checks if a parameter is in a request and returns its value if
     * so. Otherwise, throws an exception.
     *
     * @param request HTTP request to get the parameter from.
     * @param paramName name of the request parameter.
     *
     * @return the specified parameter's value.
     *
     * @throws ServletException if the specified parameter is not
     * present in the specified request.
     */
    protected String getParameterNoNull(HttpServletRequest request, String paramName)
        throws ServletException {

        String paramValue = request.getParameter(paramName);
        if(paramValue == null)
            throw new ServletException("Required request parameter \"" +
                                       paramName + "\" is not specified.");

        return paramValue;
    }

    /**
     * Retreives a request parameter as a long number.
     *
     * @param request HTTP request to get the parameter from.
     * @param paramName name of the request parameter.
     *
     * @return the specified parameter's value as a long number or
     * <code>null</code> if the parameter is not present in the
     * request or its value is empty.
     *
     * @throws ServletException if the specified parameter is present
     * in the request but it is not a valid long number.
     */
    protected Long getParameterAsLong(HttpServletRequest request, String paramName)
        throws ServletException {

        String paramValueS = request.getParameter(paramName);
        if((paramValueS == null) || (paramValueS.length() == 0))
            return null;

        Long paramValue;
        try {
            paramValue = new Long(paramValueS);
        } catch(NumberFormatException e) {
            throw new ServletException("Request parameter \"" +
                                       paramName + "\" is not a number.");
        }

        return paramValue;
    }

    /**
     * Retreives a request parameter as a long number.
     *
     * @param request HTTP request to get the parameter from.
     * @param paramName name of the request parameter.
     *
     * @return the specified parameter's value as a long number.
     *
     * @throws ServletException if the specified parameter is not
     * present in the request or it is not a valid long number.
     */
    protected long getParameterAsLongNoNull(HttpServletRequest request, String paramName)
        throws ServletException {

        String paramValueS = this.getParameterNoNull(request, paramName);

        long paramValue;
        try {
            paramValue = Long.parseLong(paramValueS);
        } catch(NumberFormatException e) {
            throw new ServletException("Request parameter \"" +
                                       paramName + "\" is not a number.");
        }

        return paramValue;
    }

    /**
     * Retreives a request parameter as an int number.
     *
     * @param request HTTP request to get the parameter from.
     * @param paramName name of the request parameter.
     *
     * @return the specified parameter's value as an int number or
     * <code>null</code> if the parameter is not present in the
     * request or its value is empty.
     *
     * @throws ServletException if the specified parameter is present
     * in the request but it is not a valid int number.
     */
    protected Integer getParameterAsInt(HttpServletRequest request, String paramName)
        throws ServletException {

        String paramValueS = request.getParameter(paramName);
        if((paramValueS == null) || (paramValueS.length() == 0))
            return null;

        Integer paramValue;
        try {
            paramValue = new Integer(paramValueS);
        } catch(NumberFormatException e) {
            throw new ServletException("Request parameter \"" +
                                       paramName + "\" is not a number.");
        }

        return paramValue;
    }

    /**
     * Retreives a request parameter as an int number.
     *
     * @param request HTTP request to get the parameter from.
     * @param paramName name of the request parameter.
     *
     * @return the specified parameter's value as an int number.
     *
     * @throws ServletException if the specified parameter is not
     * present in the request or it is not a valid int number.
     */
    protected int getParameterAsIntNoNull(HttpServletRequest request, String paramName)
        throws ServletException {

        String paramValueS = this.getParameterNoNull(request, paramName);

        int paramValue;
        try {
            paramValue = Integer.parseInt(paramValueS);
        } catch(NumberFormatException e) {
            throw new ServletException("Request parameter \"" +
                                       paramName + "\" is not a number.");
        }

        return paramValue;
    }
}
