package com.boylesoftware.cb2.presentation.servlet;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import javax.servlet.ServletContext;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.BLManager;
import com.boylesoftware.cb2.StatusProvider;
import com.boylesoftware.cb2.ApplicationException;
import com.boylesoftware.cb2.util.EventStatistics;
import com.boylesoftware.cb2.util.EventHistory;
import com.boylesoftware.cb2.util.pool.AdaptiveObjectPool;


/**
 * This web-application wide singleton represents this CB2 based
 * application. When you use Struts modules you may end up with
 * setting up CB2 plugin multiple times and then the plugin will
 * install it's request processor (<code>ControllerServlet</code>)
 * multiple times for different modules. Regardless of how many
 * modules and CB2 request processors you have they all will share
 * one <code>CB2Application</code> instance, which will provide
 * them with access to the application context and business logic
 * level.
 *
 * <p>This class is not of any interest to the user application code.
 *
 * @author Lev Himmelfarb
 * @version $Id: CB2Application.java,v 1.6 2004/04/30 18:54:08 levahim Exp $
 */
public final class CB2Application
    implements StatusProvider {


    ///////////////////////////////////////////////////////////////////////////
    // Application data
    ///////////////////////////////////////////////////////////////////////////

	/**
	 * Our application context.
	 */
	private final ApplicationContext appCtx;

	/**
	 * Business logic level.
	 */
	private final BLManager blManager;

    /**
     * Reference to the web-application's servlet context.
     */
    private final ServletContext servletCtx;

    /**
     * All registered HTTP sessions. The keys are BL session ids (objects
     * of <code>String</code>) and the values are session descriptors
     * (objects of <code>UserSession</code>).
     */
    private final Map userSessions;

    /**
     * Number of requests stats and history.
     */
    private final EventStatistics requestStats;

    /**
     * Indicates if the web-application is temporarily locked.
     */
    private boolean isLocked;

    /**
     * Tells if the application is initialized.
     */
    private boolean isInitialized;

    /**
     * Synchronization object for the <code>isInitialized</code> flag.
     */
    private final Object isInitializedLock = new Object();

    /**
     * Pool of action context instances used by the
     * <code>CB2Action</code>.
     */
    private final AdaptiveObjectPool actionContextPool;


    ///////////////////////////////////////////////////////////////////////////
    // Application lifecycle
    ///////////////////////////////////////////////////////////////////////////

	/**
	 * Constructs an instance and initializes the CB2 application
	 * environment.
	 *
	 * @param servletCtx servlet context.
	 *
	 * @throws ApplicationException if the application cannot be
	 * initialized.
	 */
	CB2Application(ServletContext servletCtx)
		throws ApplicationException {

        this.servletCtx = servletCtx;

		this.appCtx = new ApplicationContext(servletCtx.getRealPath("/"),
											 servletCtx.getRealPath("/WEB-INF"));

		try {
			this.blManager = new BLManager(this.appCtx);
        } catch(Throwable e) {

            this.appCtx.getLog(this.getClass().getName()).
                fatal("Could not initialize business level, throwing exception up.", e);

            this.appCtx.destroy();

            if(e instanceof ApplicationException)
                throw (ApplicationException)e;
            else if(e instanceof RuntimeException)
                throw (RuntimeException)e;
            else if(e instanceof Error)
                throw (Error)e;
            else
                throw new ApplicationException("Error.", e); // should never happen
		}

        this.userSessions = new HashMap();
        this.requestStats = new EventStatistics(new EventHistory[] {
            new EventHistory(3456000L, 288),
            new EventHistory(24L * 3600000L, 288),
            new EventHistory(144L * 3600000L, 288)
        });
        this.isLocked = false;
        this.isInitialized = true;
        this.actionContextPool = new AdaptiveObjectPool(ActionContext.class);
	}

	/**
	 * Destroys the application environment in order to prepare
	 * for the application's shutdown.
	 */
	void destroy() {

        synchronized(this.isInitializedLock) {
            this.isInitialized = false;
        }

		this.blManager.destroy();
		this.appCtx.destroy();
	}


    ///////////////////////////////////////////////////////////////////////////
    // Service interface
    ///////////////////////////////////////////////////////////////////////////

	/**
	 * Gets our application context.
	 *
	 * @return reference to our application context.
	 */
	public ApplicationContext getApplicationContext() {

		return this.appCtx;
	}

	/**
	 * Gets our business logic level manager.
	 *
	 * @return reference to out BL manager.
	 */
	public BLManager getBLManager() {

		return this.blManager;
	}

    /**
     * Register a fact of a single request arrived to the application. For
     * statistics.
     */
    void registerRequest() {

        this.requestStats.eventStarted();
    }

    /**
     * Adds a session to the collection of registered sessions.
     *
     * @param userSession the user session object.
     */
    void registerSession(UserSession userSession) {

        synchronized(this.userSessions) {
            this.userSessions.put(userSession.getBLSessionId(), userSession);
        }
    }

    /**
     * Checks if the session is not registered in our
     * session list (this may happen when a session was restored
     * from serialized form) and if so, adds it to the list and returns
     * <code>true</code>.
     *
     * @param userSession the user session object.
     *
     * @return <code>true</code> if the session was restored in the collection
     * of registered sessions, <code>false</code> if the session was already
     * in the collection.
     */
    boolean restoreSession(UserSession userSession) {

        synchronized(this.userSessions) {
            if(!this.userSessions.containsKey(userSession.getBLSessionId())) {
                this.userSessions.put(userSession.getBLSessionId(), userSession);
                return true;
            }
        }

        return false;
    }

    /**
     * Unregisters a session. This method is called by a
     * <code>UserSession</code> instance when it gets unbound from a
     * session (meaning that the session is being destroyed).
     *
     * @param blSessionId the corresponding BL session id.
     */
    void unregisterSession(String blSessionId) {

        // the counterpart of this synchronization is in the destroy()
        // method
        synchronized(this.isInitializedLock) {

            // check if the application is initialized. (Some servlet
            // containers, when going down, first call the destroy()
            // method on the servlet and only then unbind session
            // binding listeners from the existing sessions. We don't
            // want to do anything if the application is shutting down
            // and the controller has already been destroyed.)
            if(this.isInitialized) {

                if(blSessionId != null) {

                    // remove the session from the list of registered
                    synchronized(this.userSessions) {
                        this.userSessions.remove(blSessionId);
                    }

                    // destroy the corresponding business level session
                    this.blManager.destroyUserSession(blSessionId);
                }
            }
        }
    }

    /**
     * Gets the action context instance for the <code>CB2Action</code>.
     * Instances retrived through this method must be returned by calling
     * <code>returnActionContext</code>.
     *
     * @return an action context instance.
     */
    ActionContext getActionContext() {

        return (ActionContext)this.actionContextPool.get();
    }

    /**
     * Returns an action context instance previously got through
     * the <code>getActionContext</code> method back to the pool.
     *
     * @param actionCtx the instance to return.
     */
    void returnActionContext(ActionContext actionCtx) {

        this.actionContextPool.returnObject(actionCtx);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Status provider interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Retrieves information about the servlet container and user
     * sessions and returns it as an XML text.
     *
     * @return XML for the servlet container and user sessions info.
     *
     * @throws UnsupportedEncodingException if UTF-8 is not supported on this
     * platform, which is practically impossible.
     */
    public byte [] getStats()
        throws UnsupportedEncodingException {

        ByteArrayOutputStream out = new ByteArrayOutputStream(1024);
        PrintStream res = new PrintStream(out, false, "UTF-8");

        res.println("<stats>");
        res.println("  <servletInfo>");
        res.print("    <container>"); res.print(this.servletCtx.getServerInfo()); res.println("</container>");
        res.print("    <apiVersion>");
        res.print(this.servletCtx.getMajorVersion());
        res.print('.');
        res.print(this.servletCtx.getMinorVersion());
        res.println("</apiVersion>");
        res.print("    <contextClass>"); res.print(this.servletCtx.getClass().getName()); res.println("</contextClass>");
        res.print("    <webapp>"); res.print(this.servletCtx.getServletContextName()); res.println("</webapp>");
        javax.servlet.jsp.JspFactory jspFactory = javax.servlet.jsp.JspFactory.getDefaultFactory();
        if(jspFactory != null) {
            String jspVersion = jspFactory.getEngineInfo().getSpecificationVersion();
            if(jspVersion != null) {
                res.print("    <jspVersion>"); res.print(jspVersion); res.println("</jspVersion>");
            }
        }
        res.println("  </servletInfo>");
        res.println("  <servletAttrs>");
        java.util.Enumeration en = this.servletCtx.getAttributeNames();
        while(en.hasMoreElements()) {
            String name = (String)en.nextElement();
            Object value = this.servletCtx.getAttribute(name);
            res.print("    <attr name=\""); res.print(name); res.println("\">");
            if(value != null) {
                res.print("      <stringValue>"); res.print(value.toString()); res.println("</stringValue>");
                res.print("      <class>"); res.print(value.getClass().getName()); res.println("</class>");
            }
            res.println("    </attr>");
        }
        res.println("  </servletAttrs>");
        res.println("  <sessions>");
        long maxSamples[] = null;
        UserSession [] userSessions;
        synchronized(this.userSessions) {
            userSessions = (UserSession [])this.userSessions.values().
                toArray(new UserSession[this.userSessions.size()]);
        }
        for(int i = 0; i < userSessions.length; i++) {
            UserSession us = userSessions[i];
            synchronized(us) {
                if(!us.isNew()) {
                    res.print("    <session id=\""); res.print(us.getSessionId()); res.println("\">");
                    res.print("      <blSessionId>"); res.print(us.getBLSessionId()); res.println("</blSessionId>");
                    res.print("      <createdOn>"); res.print(StatusProvider.TIMESTAMP_FORMAT.format(us.getCreatedOn())); res.println("</createdOn>");
                    if(us.getUser() != null) {
                        res.print("      <user><![CDATA["); res.print(us.getUser().getName()); res.println("]]></user>");
                        res.print("      <authedOn>"); res.print(StatusProvider.TIMESTAMP_FORMAT.format(us.getAuthenticatedOn())); res.println("</authedOn>");
                    }
                    res.print("      <nActs>"); res.print(us.getRequestStats().getNTimes()); res.println("</nActs>");
                    res.print("      <nErrors>"); res.print(us.getNErrors()); res.println("</nErrors>");
                    res.println("      <lastErrors>");
                    UserSession.ErrorDescriptor [] errors = us.getLastErrors();
                    for(int j = 0; j < errors.length; j++) {
                        res.println("        <error>");
                        res.print("          <ts>"); res.print(StatusProvider.TIMESTAMP_FORMAT.format(errors[j].ts)); res.println("</ts>");
                        res.print("          <type>"); res.print(errors[j].throwable.getClass().getName()); res.println("</type>");
                        res.print("          <message><![CDATA["); res.print(errors[j].throwable.getMessage()); res.println("]]></message>");
                        res.println("        </error>");
                    }
                    res.println("      </lastErrors>");
                    res.println("    </session>");

                    EventHistory [] eh = us.getRequestStats().getTimelines();
                    if(maxSamples == null)
                        maxSamples = new long[eh.length];
                    for(int j = 0; j < eh.length; j++) {
                        synchronized(eh[j]) {
                            if(eh[j].getMaxSample() > maxSamples[j])
                                maxSamples[j] = eh[j].getMaxSample();
                        }
                    }
                }
            }
        }
        res.println("  </sessions>");
        res.println("  <maxSamples>");
        if(maxSamples != null) {
            for(int j = 0; j < maxSamples.length; j++) {
                res.print("    <actMaxSample ind=\""); res.print(j); res.print("\">");
                res.print(maxSamples[j]);
                res.println("</actMaxSample>");
            }
        }
        res.println("  </maxSamples>");
        res.println("</stats>");

        res.flush();
        return out.toByteArray();
    }

    /**
     * Interface method implementation. Accepts the following event names:
     *
     * <ul>
     *
     * <li><code>requests</code> - number of requests.</li>
     *
     * <li><code>sessionActivity_xxx</code> - number of requests for the
     * session with BL session id xxx.</li>
     *
     * </ul>
     */
    public EventStatistics getEventStats(String eventName) {

        if(eventName.equals("requests"))
            return this.requestStats;
        else if(eventName.startsWith("sessionActivity_")) {
            synchronized(this.userSessions) {
                UserSession us = (UserSession)this.userSessions.get(eventName.substring(16));
                return (us != null ? us.getRequestStats() : null);
            }
        } else
            return null;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Interface for admin servlet
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Tells if the application is locked by a <code>setLock</code> call.
     *
     * @return <code>true</code> if the application is locked.
     */
    public synchronized boolean isLocked() {

        return this.isLocked;
    }

    /**
     * Sets the application "locked" status. When the application is locked
     * the request processor does not let any requests into actions, instead
     * it immediately sends a Service Unavailable error code back to the
     * browser (or alternatively forwards to the configured by
     * <code>lockedUrl</code> CB2 plugin property page).
     *
     * @param lock the new lock state -- <code>true</code> locks the
     * application, <code>false</code> unlocks it.
     */
    public synchronized void setLock(boolean lock) {

        this.isLocked = lock;
    }

    /**
     * Gets number of currently registered user sessions. This method can
     * be used for statistics reports.
     *
     * @return number of sessions.
     */
    public int getNumberOfSessions() {

        synchronized(this.userSessions) {
            return this.userSessions.size();
        }
    }

    /**
     * Gets the number of registered errors in all user sessions. This
     * method can be used for statistics reports.
     *
     * @return total number of errors in all sessions.
     */
    public int getNumberOfErrors() {

        int res = 0;
        synchronized(this.userSessions) {
            Iterator i = this.userSessions.values().iterator();
            while(i.hasNext()) {
                res += ((UserSession)i.next()).getNErrors();
            }
        }

        return res;
    }
}
