package com.boylesoftware.cb2.presentation.servlet;

import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;

import org.apache.struts.action.PlugIn;
import org.apache.struts.action.ActionServlet;
import org.apache.struts.action.RequestProcessor;
import org.apache.struts.config.ModuleConfig;

import com.boylesoftware.cb2.ApplicationException;


/**
 * Struts plugin to enable CB2 presentation level for a Struts module.
 * This plugin should be set up for each module that uses CB2. Regardless
 * of how any instances of the plugin you install for different modules
 * they all will share the same application context, business logic level
 * and other basic application-wide subsystems, however each module will
 * have its own presentation level subsystems such as CB2 pages.
 *
 * <p>The plugin expects the following init parameters:
 *
 * <ul>
 *
 * <li><i>pages-config</i> - web-application context relative URL of the
 * pages definition file. If not specified defaults to
 * <code>/WEB-INF/pages-config.xml</code>.</li>
 *
 * <li><i>controller</i> - name of the controller class. The controller
 * class should extend
 * <code>com.boylesoftware.cb2.presentation.servlet.CB2RequestProcessor</code>.
 * If not specified defaults to <code>CB2RequestProcessor</code>.</li>
 *
 * <li><i>uistate</i> - name of the UI state bean class. The class must
 * extend <code>com.boylesoftware.cb2.presentation.servlet.UIState</code>.
 * If not specified the module will be provided with the default
 * <code>UIState</code> implementation.</li>
 *
 * <li><i>lockedUrl</i> - URL of a page that should be displayed when
 * a request comes to the module while the application is locked via the
 * admin interface. If it is not specified the system sends a Service
 * Unavailable error code in response to the request.</li>
 *
 * <li><i>requestCharEncoding</i></li> - character encoding for requests.
 * If specified, CB2 will explicitely set it on every request using
 * <code>ServletRequest.setCharacterEncoding</code> method before passing
 * control to any action or page.</li>
 *
 * <li><i>responseContentType</i></li> - response content type string
 * (optionally with encoding). If specified, CB2 will explicitely set it
 * on every response using <code>ServletResponse.setContentType</code>, as
 * well as set HTTP header "Content-Type" using
 * <code>HttpServletResponse.setHeader</code>. The difference of this
 * mechanism from Struts' controller "contentType" attribute is that Struts
 * only calls <code>setContentType</code> and does not explicitely sets
 * the "Content-Type" HTTP response header, which may be required in some
 * Servlet Container implementations for proper results. As in Struts,
 * CB2 sets the content type before passing control to any action or page,
 * which, if desired, may override it. Note, that if this parameter is
 * specified, the framework sets content type on <strong>every</strong>
 * response. More often it is better to control this on per page basis via
 * "contentType" attribute of <code>&lt;page&gt;</code> elements in the PL
 * configuration file, and since "contentType" attribute is inheritable, you
 * can achieve the same effect by specifying it for an abstract page, which is
 * extended by all other pages.</li>
 *
 * </ul>
 *
 * <p>Below is a sample fragment of a struts config file:
 *
 * <pre>
 * ...
 * &lt;plug-in className="com.boylesoftware.cb2.presentation.servlet.CB2PlugIn"&gt;
 *   &lt;set-property
 *     property="pages-config"
 *     value="/WEB-INF/pages-config-mymodule.xml"/&gt;
 *   &lt;set-property
 *     property="uistate"
 *     value="my.package.MyUIState"/&gt;
 * &lt;/plug-in&gt;
 * ...
 * </pre>
 *
 * <p>Note that this plugin installs its own implementation of the
 * <code>RequestProcessor</code> for the module and therefore it is
 * not compatible with other plugins that do the same, for example with
 * <code>TilesPlugin</code>.
 *
 * <p>This implementation supports persistent sessions provided that you
 * set <code>com.boylesoftware.cb2.keepBLSessions</code> application
 * property to <code>true</code> so the <code>BLManager</code> does not
 * destroy BLO containers when the application goes down. Note that if you
 * turn on persistent sessions in your servlet container implementation
 * and forget to set the <code>com.boylesoftware.cb2.keepBLSessions</code>
 * property you will get errors when using restored sessions.
 *
 * @author Lev Himmelfarb
 * @version $Id: CB2PlugIn.java,v 1.8 2004/04/28 15:12:08 levahim Exp $
 */
public final class CB2PlugIn
    implements PlugIn {


    ///////////////////////////////////////////////////////////////////////////
    // Constants
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Name of servlet context attribute holding the
     * <code>CB2Application</code> instance.
     */
    private static final String APPLICATION = "com.boylesoftware.cb2.presentation.servlet.APPLICATION";

    /**
     * Name of servlet context attribute holding a map with plugin instances
     * for modules. The keys are module prefixes (objects of
     * <code>String</code>), the values are plugin instances (objects of
     * <code>CB2PlugIn</code>). The map is synchronized.
     */
    private static final String MODULES = "com.boylesoftware.cb2.presentation.servlet.MODULES";


    /**
     * Stored in the servlet context and is used to count number of installed
     * and initialized plugins.
     */
    private static final class SetupCounter
        implements java.io.Serializable {

        /**
         * Number of currently set up plugins.
         */
        private int count = 0;


        /**
         * Increments the count.
         */
        public synchronized void increment() {

            this.count++;
        }

        /**
         * Decrements the count.
         *
         * @return <code>true</code> if there are no more plugins installed.
         */
        public synchronized boolean decrement() {

            return (--this.count <= 0);
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Init parameters
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Init parameter -- pages config file URL.
     */
    private String pagesConfig = "/WEB-INF/pages-config.xml";

    /**
     * Init parameter -- controller class name.
     */
    private String controllerClassName = (CB2RequestProcessor.class).getName();

    /**
     * Init parameter -- UIState bean class name.
     */
    private String uiStateClassName = (UIState.class).getName();

    /**
     * Init parameter -- URL of the "application is locked" page.
     */
    private String lockedUrl = null;

    /**
     * Optional request character encoding.
     */
    private String requestCharEncoding = null;

    /**
     * Optional response content type.
     */
    private String responseContentType = null;


    /**
     * Sets the pages config file URL init parameter.
     *
     * @param pagesConfig pages config file servlet context relative URL.
     */
    public void setPagesConfig(String pagesConfig) {

        this.pagesConfig = pagesConfig;
    }

    /**
     * Sets the controller class name init parameter.
     *
     * @param controllerClassName the controller class name.
     */
    public void setController(String controllerClassName) {

        this.controllerClassName = controllerClassName;
    }

    /**
     * Sets the UI state bean class name init parameter.
     *
     * @param uiStateClassName the UI state bean class name.
     */
    public void setUistate(String uiStateClassName) {

        this.uiStateClassName = uiStateClassName;
    }

    /**
     * Sets the URL of the "application is locked" page.
     *
     * @param lockedUrl the page's URL.
     */
    public void setLockedUrl(String lockedUrl) {

        this.lockedUrl = lockedUrl;
    }

    /**
     * Sets optional request character encoding.
     *
     * @param requestCharEncoding the request character encoding.
     */
    public void setRequestCharEncoding(String requestCharEncoding) {

        this.requestCharEncoding = requestCharEncoding;
    }

    /**
     * Sets optional response content type.
     *
     * @param responseContentType the reponse content type string.
     */
    public void setResponseContentType(String responseContentType) {

        this.responseContentType = responseContentType;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Plugin data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Tells if the plugin is initialized and <code>destroy</code> method
     * should destroy it when called.
     */
    private boolean isInitialized = false;

    /**
     * Reference to the Struts action servlet.
     */
    private ActionServlet actionServlet;

    /**
     * This module's config.
     */
    private ModuleConfig moduleConfig;

    /**
     * UI state bean implementation class.
     */
    private Class uiStateClass;

    /**
     * The PL config for this module.
     */
    private PLConfig plConfig = null;


    ///////////////////////////////////////////////////////////////////////////
    // Plugin lifecycle
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Initialize the plugin for a module. If this is the first, or
     * the only module, then this method also creates and initializes
     * such shared subsystems as the application context and business
     * logic level manager.
     *
     * @param servlet reference to the Struts action servlet.
     * @param config the module configuration.
     *
     * @throws ServletException if the plugin cannot be initialized.
     */
	public void init(ActionServlet servlet, ModuleConfig config)
        throws ServletException {

        servlet.log("CB2 plugin is initializing for module with prefix [" + config.getPrefix() + "]");

        this.actionServlet = servlet;
        this.moduleConfig = config;

        ServletContext servletCtx = servlet.getServletContext();

        // setup request processor
        Class currentProcessorClass;
        try {
            currentProcessorClass = Class.forName(config.getControllerConfig().getProcessorClass());
        } catch(ClassNotFoundException e) {
            throw new ServletException("Could not load current request processor class.", e);
        }
        if(currentProcessorClass.getName().equals((RequestProcessor.class).getName())) {
            Class processorClass;
            try {
                processorClass = Class.forName(this.controllerClassName);
            } catch(ClassNotFoundException e) {
                throw new ServletException("Could not load controller class.", e);
            }
            if(!(CB2RequestProcessor.class).isAssignableFrom(processorClass))
                throw new ServletException("Specified controller class does not extend CB2RequestProcessor.");
            config.getControllerConfig().setProcessorClass(processorClass.getName());
        } else if(!(CB2RequestProcessor.class).isAssignableFrom(currentProcessorClass))
            throw new ServletException("Current request processor is not compatible with CB2," +
                                       " it should be CB2RequestProcessor or a subclass.");

        // load UI state bean class and put it to the servlet context
        try {
            this.uiStateClass = Class.forName(this.uiStateClassName);
        } catch(ClassNotFoundException e) {
            throw new ServletException("Could not load UI state bean class.", e);
        }
        if(!(UIState.class).isAssignableFrom(this.uiStateClass))
            throw new ServletException("Specified UI state class does extend UIState.");

        // initialize PL configuration
        this.plConfig = new PLConfig(servletCtx, this.pagesConfig, config);

        // setup the application
        CB2Application app;
        synchronized(this.getClass()) {
            app = (CB2Application)servletCtx.getAttribute(APPLICATION);
            if(app == null) {
                try {
                    app = new CB2Application(servletCtx);
                } catch(ApplicationException e) {
                    throw new ServletException("Could not initialize CB2 application environment.", e);
                }
                servletCtx.setAttribute(APPLICATION, app);
            }
        }

        // register the plugin
        SetupCounter sc;
        synchronized(this.getClass()) {
            sc = (SetupCounter)servletCtx.getAttribute((SetupCounter.class).getName());
            if(sc == null) {
                sc = new SetupCounter();
                servletCtx.setAttribute((SetupCounter.class).getName(), sc);
            }
        }
        sc.increment();

        // turn on the destroy method
        this.isInitialized = true;

        // if any of the following parts are not successfull the destroy
        // will be called
        boolean success = false;
        try {

            // add me to the module->plugin map
            Map modules;
            synchronized(this.getClass()) {
                modules = (Map)servletCtx.getAttribute(MODULES);
                if(modules == null) {
                    modules = Collections.synchronizedMap(new HashMap());
                    servletCtx.setAttribute(MODULES, modules);
                }
            }
            if(modules.put(config.getPrefix(), this) != null)
                throw new ServletException("CB2 plugin is set more than once for module with prefix [" +
                                           config.getPrefix() + "].");

            // all done
            success = true;

        } finally {
            if(!success)
                this.destroy();
        }
	}

    /**
     * Destroys the plugin when the application goes down. If this is the
     * last, or the only module, then this method also destroys application
     * context and business logic level manager.
     */
    public void destroy() {

        if(this.isInitialized) {
            this.isInitialized = false;

            ServletContext servletCtx = this.actionServlet.getServletContext();

            // forget PL config
            this.plConfig = null;

            // remove me from the module->plugin map
            Map modules = (Map)servletCtx.getAttribute(MODULES);
            if(modules != null)
                modules.remove(this.moduleConfig.getPrefix());

            // unregister the plugin
            SetupCounter sc = (SetupCounter)servletCtx.getAttribute((SetupCounter.class).getName());
            if(sc != null) {
                if(sc.decrement()) {
                    // destroy application
                    CB2Application app = (CB2Application)servletCtx.getAttribute(APPLICATION);
                    if(app != null) {
                        servletCtx.removeAttribute(APPLICATION);
                        app.destroy();
                    }
                }
            }
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Interface to internal subsystems
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets the UI state bean implementation class for this module.
     *
     * @return the UI state bean class.
     */
    Class getUIStateClass() {

        return this.uiStateClass;
    }

    /**
     * Gets PL configuration for this module.
     *
     * @return the PL configuration.
     */
    PLConfig getPLConfig() {

        return this.plConfig;
    }

    /**
     * Gets the "application is locked" page URL.
     *
     * @return the page's URL or <code>null</code>.
     */
    String getLockedUrl() {

        return this.lockedUrl;
    }

    /**
     * Gets optional request character encoding.
     *
     * @return the request character encoding or <code>null</code>.
     */
    String getRequestCharEncoding() {

        return this.requestCharEncoding;
    }

    /**
     * Gets optional response content type.
     *
     * @return the reponse content type string or <code>null</code>.
     */
    String getResponseContentType() {

        return this.responseContentType;
    }

    /**
     * Finds the <code>CB2Plugin</code> instance for a Struts module.
     *
     * @param servletCtx the web-application's servlet context.
     * @param moduleConfig the module's config.
     *
     * @return reference to the appropriate instance of <code>CB2PlugIn</code>.
     */
    static CB2PlugIn getModule(ServletContext servletCtx,
                               ModuleConfig moduleConfig) {

        return (CB2PlugIn)((Map)servletCtx.getAttribute(MODULES)).
            get(moduleConfig.getPrefix());
    }


    ///////////////////////////////////////////////////////////////////////////
    // Public interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Finds the <code>CB2Application</code> instance for the web-application.
     *
     * @param servletCtx the web-application's servlet context.
     *
     * @return reference to the <code>CB2Application<code> instance.
     */
    public static CB2Application getApplication(ServletContext servletCtx) {

        return (CB2Application)servletCtx.getAttribute(APPLICATION);
    }
}
