package com.boylesoftware.cb2.presentation.servlet;

import java.util.Map;
import java.util.HashMap;
import java.util.Collections;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.net.URLEncoder;
import java.io.IOException;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.struts.action.RequestProcessor;
import org.apache.struts.action.Action;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionServlet;
import org.apache.struts.config.ModuleConfig;
import org.apache.struts.config.ForwardConfig;

import com.boylesoftware.cb2.BLManager;
import com.boylesoftware.cb2.BLOContainer;
import com.boylesoftware.cb2.BLException;
import com.boylesoftware.cb2.NoBLOContainerException;
import com.boylesoftware.cb2.util.ExceptionUtils;


/**
 * CB2 Struts request processor. This processor is automatically set up
 * for a module by the <code>CB2PlugIn</code>. You application can extend
 * this class to provide additional request processing logic. If you
 * subclass this class you will have to specify the your class name to
 * the <code>CB2PlugIn</code> by setting its <code>controller</code>
 * property (or, as an alternative, by providing <code>processorClass</code>
 * attribute to the <code>controller</code> element in the Struts
 * configuration file).
 *
 * <p>This class logs messages under "com.boylesoftware.cb2.CTRLER"
 * log channel.
 *
 * @author Lev Himmelfarb
 * @version $Id: CB2RequestProcessor.java,v 1.6 2004/10/11 18:17:36 levahim Exp $
 */
public class CB2RequestProcessor
    extends RequestProcessor {


    ///////////////////////////////////////////////////////////////////////////
    // Constants
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Subsystem name for the logger.
     */
    private static final String LOG_SUBSYS = "com.boylesoftware.cb2.CTRLER";

    /**
     * Regex pattern used for finding <code>[request:xxx]</code>
     * tokens in the forward path.
     */
    private static final Pattern FORWARD_PARAM_REQUEST = Pattern.compile("\\[(request|rqparam|rqattribute|session|any):([\\w.]+)\\]");

    /**
     * Name of session attribute where we put BLO container instance. We do it
     * so we have a serialized BLO container when the session gets "passivated"
     * for this or that reason.
     */
    private static final String BLOCONTAINER = "com.boylesoftware.cb2.presentation.servlet.BLOCONTAINER";

    /**
     * Name of request attribute where the controller puts reference to
     * the module config so it can be picked up later by <code>CB2Action</code>
     * and set in the action context.
     */
    static final String MODULE_CONFIG = "com.boylesoftware.cb2.presentation.servlet.ModuleConfig";

    /**
     * Name of session attribute where controller servlet stores instances
     * of UIState bean for sessions. The UIState bean instance is also
     * available for actions and presentation elements through the action
     * context, but direct access to the attribute may be needed in custom
     * JSP tag implementations.
     */
    private static final String UISTATE = "com.boylesoftware.cb2.presentation.servlet.UIState";

    /**
     * Name of request attributes where actions put a token (just any object)
     * if they want the controller to perform logout in the
     * <Code>processPostprocess</code>.
     */
    static final String LOGOUT_TOKEN = "com.boylesoftware.cb2.presentation.servlet.LogoutToken";

    /**
     * Name of request attribute where <code>processActionPerform</code> puts
     * page instances so <code>ShowPageAction</code> can pick it up later.
     */
    static final String PAGE = "com.boylesoftware.cb2.presentation.servlet.PAGE";


    ///////////////////////////////////////////////////////////////////////////
    // CB2RequestProcessor data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * The Struts module configuration.
     */
    private ModuleConfig moduleConfig;

    /**
     * The log.
     */
    private Log log = null;

    /**
     * Map between pages names and corresponding action forwards.
     * The keys are page names (objects of <code>String</code>), the
     * values are action forwards pointed to the pages (objects of
     * <code>ForwardConfig</code>).
     */
    private Map pagesToForwards;


    ///////////////////////////////////////////////////////////////////////////
    // CB2RequestProcessor lifecycle
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Initializes the request processor. Called by Struts framework.
     *
     * @throws ServletException if an error occurs during the initialization.
     */
    public void init(ActionServlet actionServlet,
                     ModuleConfig moduleConfig)
        throws ServletException {

        // initialize Struts' part
        super.init(actionServlet, moduleConfig);

        this.moduleConfig = moduleConfig;
        this.log = CB2PlugIn.getApplication(this.getServletContext()).
            getApplicationContext().getLog(LOG_SUBSYS);
        this.pagesToForwards = Collections.synchronizedMap(new HashMap());
    }

    /**
     * Gracefully shutdown the application. This method is called when
     * the container unloads the <code>ControllerServlet</code>, which
     * usually happens when the web-application goes down.
     */
    public void destroy() {

        super.destroy();

        this.pagesToForwards.clear();
        this.pagesToForwards = null;
        this.log = null;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Request processing
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Process a request.
     *
     * @param request the HTTP request to be processed.
     * @param response the HTTP response to be sent back to the
     * client.
     *
     * @throws ServletException indicates a serious error in the
     * application.
     * @throws IOException if an I/O error occurs.
     */
    public void process(HttpServletRequest request, HttpServletResponse response)
        throws IOException, ServletException {

        try {

            // find basic subsystems
            CB2Application app = CB2PlugIn.getApplication(this.getServletContext());
            CB2PlugIn module = CB2PlugIn.getModule(this.getServletContext(), this.moduleConfig);
            BLManager blManager = app.getBLManager();

            // request stats
            app.registerRequest();

            // check if the application is locked
            if(app.isLocked()) {
                String lockedUrl = module.getLockedUrl();
                if(lockedUrl != null) {
                    RequestDispatcher rd = this.getServletContext().
                        getRequestDispatcher(lockedUrl);
                    if(rd == null)
                        throw new ServletException("The lockedUrl CB2 plugin parameter is misconfigured, " +
                                                   "could not get a request dispatcher for the specified URL.");
                    rd.forward(request, response);

                } else {
                    response.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE,
                                       "The web-application is currently locked.");
                }
            }

            // basic session info
            HttpSession session = request.getSession();
            String sessionId = session.getId();

            // log the request
            if(this.log.isDebugEnabled()) {
                String hdr = null;
                String src = "";
                if((hdr = request.getHeader("X-Forwarded-For")) != null)
                    src = " (X-Forwarded-For: " + hdr + ")";
                else if((hdr = request.getHeader("Client-ip")) != null)
                    src = " (Client-ip: " + hdr + ")";
                else if((hdr = request.getHeader("Forwarded")) != null)
                    src = " (Forwarded: " + hdr + ")";
                java.security.Principal p = request.getUserPrincipal();
                this.log.debug("processing request [" + request.getRequestURL() +
                               "] from " + request.getRemoteAddr() +
                               src +
                               ", session " + sessionId +
                               (p != null ? ", user " + p.getName() : ", not authenticated"));
            }

            // set encodings on request and response
            if(module.getRequestCharEncoding() != null)
                request.setCharacterEncoding(module.getRequestCharEncoding());
            if(module.getResponseContentType() != null) {
                response.setContentType(module.getResponseContentType());
                response.setHeader("Content-Type", module.getResponseContentType());
            }

            // get the user session object
            UserSession userSession;
            synchronized(app) {
                userSession = (UserSession)session.getAttribute((UserSession.class).getName());
                if(userSession == null) {
                    userSession = new UserSession(session);
                    session.setAttribute((UserSession.class).getName(), userSession);
                }
            }

            // do the session adjustments
            synchronized(userSession) {

                // check if this session is new and do initialization if so
                if(userSession.isNew()) {

                    // log the new session creation
                    this.log.debug("registering new session " + sessionId);

                    // create a corresponding session in the business level
                    String blSessionId;
                    try {
                        blSessionId = blManager.createUserSession();
                    } catch(BLException e) {
                        throw new ServletException("Could not create session in the business level.", e);
                    }
                    userSession.setBLSessionId(blSessionId);

                    // put BLO container instance to the session, so it gets
                    // serialized with the session
                    BLOContainer bloc = null;
                    try {
                        bloc = blManager.getBLOContainer(blSessionId);
                        session.setAttribute(BLOCONTAINER, bloc);
                    } finally {
                        if(bloc != null)
                            bloc.release();
                    }

                    // prepare a UI state bean for the session
                    try {
                        session.setAttribute(UISTATE, module.getUIStateClass().newInstance());
                    } catch(InstantiationException e) {
                        throw new ServletException("Could not instantiate UIState bean.", e);
                    } catch(IllegalAccessException e) {
                        throw new ServletException("Could not instantiate UIState bean.", e);
                    }

                    // put the session to the list of the registered
                    app.registerSession(userSession);

                    // reset the isNew flag
                    userSession.resetIsNew();
                }

                // check if the session is not new, but not registered in our
                // session list (this may happen when a session was restored
                // from serialized form)
                if(app.restoreSession(userSession)) {
                    this.log.info("registered session " + sessionId + " as restored from serialized form");

                    // resotre BL session if needed
                    BLOContainer bloc = (BLOContainer)session.getAttribute(BLOCONTAINER);
                    if(bloc == null) // should never happen
                        throw new ServletException("There is no BLO container instance in the restored from serialized form session.");
                    blManager.restoreUserSession(userSession.getBLSessionId(), bloc);
                }

                // adjust session's "authenticated" status and sync with the BL
                BLOContainer bloc = null;
                try {

                    switch(userSession.checkAuthenticationStatus(request)) {

                    case UserSession.AUTHACTION_LOGIN:
                        bloc = blManager.getBLOContainer(userSession.getBLSessionId());
                        bloc.processUserLogin(userSession.getUser());
                        break;

                    case UserSession.AUTHACTION_LOGOUT:
                        bloc = blManager.getBLOContainer(userSession.getBLSessionId());
                        bloc.processUserLogout();
                    }

                } catch(NoBLOContainerException e) {
                    // this may happen if another request invalidated
                    // the session and removed the BLO container from BL
                    // manager, then we just discard this request
                    this.log.info("could not get BLO container for the session," +
                                   " aborting request processing");
                    return;

                } catch(BLException e) {
                    throw new ServletException("Could not process user login/logout in the business level.", e);

                } finally {
                    if(bloc != null)
                        bloc.release();
                }
            }

            // create request processing context if needed
            if(request.getAttribute(RequestProcessingContext.RPCTX_KEY) == null) {
                request.setAttribute(RequestProcessingContext.RPCTX_KEY,
                                     new RequestProcessingContext(app.getApplicationContext(),
                                                                  blManager,
                                                                  this.getMyURL(request),
                                                                  userSession,
                                                                  (UIState)session.getAttribute(UISTATE)));
            }

        } catch(Throwable e) {

            this.log.error("Got an error preparing for request processing, invalidating the whole session.", e);

            // invalidate the session
            request.getSession().invalidate();

            // rethrow the error
            ExceptionUtils.
                rethrowServletException(e, "Got an unexpected error preparing for request processing.");
        }

        // all done, let Struts do its magic
        super.process(request, response);
    }

    /**
     * Overrides the basic Struts method that calls actions. This method
     * checks if the action is actually a page defined in
     * <code>pages-config.xml</code> and if so calls the appropriate preactions
     * first.
     *
     * @param request request we are processing.
     * @param response response we are creating.
     * @param action action to be called.
     * @param form form to pass to the action.
     * @param mapping action mapping to pass to the action.
     *
     * @return action forward returned by the action.
     *
     * @throws IOException if I/O error occurs.
     * @throws ServletException if other error occurs.
     */
    protected ActionForward processActionPerform(HttpServletRequest request,
                                                 HttpServletResponse response,
                                                 Action action,
                                                 ActionForm form,
                                                 ActionMapping mapping)
        throws IOException, ServletException {

        // put module config to the request so CB2Action can pick it up
        // later and put to the action context
        request.setAttribute(MODULE_CONFIG, this.moduleConfig);

        // call preactions for pages
        if(action instanceof ShowPageAction) {

            // what is the page name
            String pageName = mapping.getParameter();
            int nameEndInd = pageName.indexOf('|');
            if(nameEndInd >= 0)
                pageName = pageName.substring(0, nameEndInd);

            // find the page descriptor
            PageDesc page = CB2PlugIn.getModule(this.getServletContext(), this.moduleConfig).
                getPLConfig().getPage(pageName);

            // call preactions
            for(int i = 0; i < page.preactions.length; i++) {
                ActionForward res = page.preactions[i].action.execute(page.preactions[i].actionMapping,
                                                                      null,
                                                                      request,
                                                                      response);
                if(res != null)
                    return res;
            }

            // pass the page descriptor to the ShowPageAction via the request
            request.setAttribute(PAGE, page);
        }

        // call the action
        return super.processActionPerform(request,
                                          response,
                                          action,
                                          form,
                                          mapping);
    }

    /**
     * Performs request post-processing after the action has been
     * executed and returned a forward. This method is called from
     * the overriden <code>processForwardConfig</code> and adds
     * additional functionality, such as special tokens in the
     * forward's <code>path</code> attribute.
     *
     * <p>In the <code>path</code> attribute of a <code>forward</code>
     * element of your action mapping you can specify special lexical
     * tokens, which are recognized by this method and replaced by
     * some values, depending on the token, before the path is passed
     * further to Struts. Currently CB2 recognizes the following
     * tokens:
     *
     * <ul>
     *
     * <li><code>[request:xxx]</code> - will be substituted by the
     * value of request parameter named <code>xxx</code>. If there is
     * no such parameter in the request, then the method will try to
     * get it from the request's attributes. If it is still not found
     * an empty string will be put in place of the token.
     *
     * <li><code>[rqparam:xxx]</code> - will be substituted by the
     * value of request parameter named <code>xxx</code>. If there is
     * no such parameter in the request an empty string will be put in
     * place of the token.
     *
     * <li><code>[rqattribute:xxx]</code> - will be substituted by the
     * string value of request attribute named <code>xxx</code>. If
     * there is no such attribute in the request an empty string will
     * be put in place of the token.
     *
     * <li><code>[session:xxx]</code> - will be substituted by the
     * string value of session attribute named <code>xxx</code>. If
     * there is no such attribute in the session an empty string will
     * be put in place of the token.
     *
     * <li><code>[any:xxx]</code> - will be substituted by the value
     * looked up by the name <code>xxx</code> first among the request
     * parameters, if not found, then in the request attribtues, and
     * if still not found, in the session. If the value cannot be
     * found in any of these sources an empty string will be put in
     * place of the token.
     *
     * </ul>
     *
     * <p>Also, this method actually performs the logout operations if
     * the action asked for it.
     *
     * @param request request we are processing.
     * @param response response we are creating.
     * @param forward forward controlling where we go next.
     *
     * @return possibly corrected forward to be passed to Struts.
     *
     * @throws IOException if I/O error occurs.
     * @throws ServletException if other error occurs.
     */
    protected ForwardConfig processPostprocess(HttpServletRequest request,
                                               HttpServletResponse response,
                                               ForwardConfig forward)
        throws ServletException, IOException {

        try {

            // fix forward's path if it is actually a page name
            if(forward != null) {
                String path = forward.getPath();
                int qInd = path.indexOf('?');
                if(qInd >= 0)
                    path = path.substring(0, qInd);
                ForwardConfig pageForward = (ForwardConfig)this.pagesToForwards.
                    get(path);
                if(pageForward != null)
                    forward = pageForward;
                else {
                    PageDesc page = CB2PlugIn.getModule(this.servlet.getServletContext(),
                                                    this.moduleConfig).
                        getPLConfig().getPage(path);
                    if(page != null) { // it's a page!
                        // find extension
                        String pagePath = page.path;
                        String servletPath = request.getServletPath();
                        int indPoint = servletPath.lastIndexOf('.');
                        if((indPoint >= 0) && (indPoint > servletPath.lastIndexOf('/'))) {
                            String ext = servletPath.substring(indPoint);
                            // check if the page path doesn't have extention
                            indPoint = pagePath.lastIndexOf('.');
                            if((indPoint < 0) || (indPoint < pagePath.lastIndexOf('/')))
                                pagePath += ext;
                        }
                        pageForward = new ActionForward(forward.getName(),
                                                        (qInd >= 0 ?
                                                         pagePath + path.substring(qInd) :
                                                         pagePath),
                                                        forward.getRedirect(),
                                                        forward.getContextRelative());
                        this.pagesToForwards.put(path, pageForward);
                        forward = pageForward;
                    }
                }
            }

            // replace macroses in the forward
            if(forward != null) {
                StringBuffer resPath = new StringBuffer();

                Matcher matcher = FORWARD_PARAM_REQUEST.matcher(forward.getPath());
                while(matcher.find()) {
                    String tokenType = matcher.group(1);
                    String tokenParam = matcher.group(2);
                    Object val;
                    if(tokenType.equals("rqparam")) {
                        val = request.getParameter(tokenParam);
                    } else if(tokenType.equals("rqattribute")) {
                        val = request.getAttribute(tokenParam);
                    } else if(tokenType.equals("session")) {
                        val = request.getSession().getAttribute(tokenParam);
                    } else if(tokenType.equals("request")) {
                        if((val = request.getParameter(tokenParam)) == null)
                            val = request.getAttribute(tokenParam);
                    } else if(tokenType.equals("any")) {
                        if((val = request.getParameter(tokenParam)) == null) {
                            if((val = request.getAttribute(tokenParam)) == null)
                                val = request.getSession().getAttribute(tokenParam);
                        }
                    } else {
                        val = matcher.group();
                    }
                    String valS = (val != null ? val.toString() : "");
                    matcher.appendReplacement(resPath, URLEncoder.encode(valS, "UTF-8"));
                }
                matcher.appendTail(resPath);

                String resPathS = resPath.toString();
                if(!resPathS.equals(forward.getPath())) {
                    ForwardConfig newForward = new ActionForward(forward.getName(),
                                                                 resPathS,
                                                                 forward.getRedirect(),
                                                                 forward.getContextRelative());
                    forward = newForward;
                }
            }

            // do logout if it was called in the action
            if(request.getAttribute(LOGOUT_TOKEN) != null) {

                // do servlet container specific logout (the nasty part)
                if(this.getServletContext().getAttribute("caucho.authenticator") != null) { // Caucho Resin?
                    ((com.caucho.http.security.ServletAuthenticator)this.getServletContext().getAttribute("caucho.authenticator")).
                        logout(request, response, this.getServletContext(), request.getUserPrincipal());
                }

                // invalidate the session
                request.getSession().invalidate();
            }

        } catch(Throwable e) {

            this.log.error("Got en error finalizing request processing.", e);

            ExceptionUtils.
                rethrowServletException(e, "Got an unexpected error finalizing request processing.");
        }

        // done
        return forward;
    }

    /**
     * Forward to the specified destination by the specified mechanism
     * if an <code>ActionForward</code> instance was returned by the
     * <code>Action</code>. This method overrides the original one from
     * Struts and adds <code>processPostprocess</code> method call.
     *
     * @param request request we are processing.
     * @param response response we are creating.
     * @param forward forward controlling where we go next.
     *
     * @throws IOException if I/O error occurs.
     * @throws ServletException if other error occurs.
     *
     * @see #processPostprocess
     */
    protected void processForwardConfig(HttpServletRequest request,
                                        HttpServletResponse response,
                                        ForwardConfig forward)
        throws ServletException, IOException {

        // let Struts to do the forward magic
        super.processForwardConfig(request,
                                   response,
                                   this.processPostprocess(request,
                                                           response,
                                                           forward));
    }


    ///////////////////////////////////////////////////////////////////////////
    // Internal service
    ///////////////////////////////////////////////////////////////////////////

    /**
     * This utility method returns a context-relative URL for the
     * specified request. It returns <code>null</code> if it is unable
     * to parse the request URL string.
     *
     * <p>Note, that this method does not take POST request parameters
     * in account, but only those, that are in the request's query
     * string (that is only GET parameters).
     *
     * @param request the request to get URL for.
     *
     * @return the URL or <code>null</code> if this methos did not
     * manage to parse the request (which should never happen
     * actually).
     */
    private String getMyURL(HttpServletRequest request) {

        StringBuffer requestURLSB = request.getRequestURL();
        if(requestURLSB == null) {
            if(request instanceof org.apache.struts.upload.MultipartRequestWrapper) {
                requestURLSB = ((org.apache.struts.upload.MultipartRequestWrapper)request).
                    getRequest().getRequestURL();
                if(requestURLSB == null)
                    return null;
            } else
                return null;
        }

        String requestURL = requestURLSB.toString();
        int i = requestURL.indexOf("//");
        if(i < 0)
            return null;
        if((i = requestURL.indexOf('/', i + 2)) < 0)
            return null;
        String contextPath = request.getContextPath();
        if(contextPath != null)
            i += contextPath.length();
        if((i > requestURL.length() - 1) || (requestURL.charAt(i) != '/'))
            return null;

        StringBuffer res = new StringBuffer(requestURL.substring(i));

        String queryString = request.getQueryString();
        if((queryString != null) && (queryString.length() > 0)) {
            res.append('?');
            res.append(queryString);
        }

        return res.toString();
    }
}
