package com.boylesoftware.cb2.presentation.servlet;

import java.util.Map;
import java.util.Set;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.beans.IntrospectionException;
import javax.servlet.ServletException;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.tagext.TagSupport;

import org.apache.commons.logging.Log;
import org.apache.struts.Globals;


/**
 * Implementation of the <code>&lt;cb2:insert&gt;</code> JSP tag. This tag
 * inserts a component or a page attribtue value, which may be configured
 * for the page in the <code>pages-config.xml</code> file. The input for
 * this tag is a name of a component reference or a page attribute. The
 * name can be specified statically right in the tag instance using
 * <code>name</code> attribute, or it can be provided as a bean property
 * named by <code>nameProperty</code> and <code>nameBean</code> attributes.
 * The tag looks up the bean using <code>PageContext.findAttribute</code>
 * method. If only <code>nameProperty</code> is specified and no
 * <code>nameBean</code>, then an attribute with the specified by the
 * attribute name is assumed to be a <code>String</code> and is used as
 * the component reference or page attribute name.
 * 
 * <p>This tag works in tight conjunction with the
 * <code>ShowPageAction</code> action handler.
 *
 * @author Lev Himmelfarb
 * @version $Id: InsertTag.java,v 1.7 2004/10/11 18:17:36 levahim Exp $
 */
public final class InsertTag
    extends TagSupport {

    /**
     * Name of the component reference or the attribute as it is
     * configured by <code>componentref</code> or
     * <code>attribute</code> element in the configuration file.
     */
    private String name;

    /**
     * Name of the bean containing component reference or page attribute
     * name.
     */
    private String nameBean;

    /**
     * Property in the <code>nameBean</code> bean containing component
     * reference or page attribute name.
     */
    private String nameProperty;

    /**
     * Ignore the tag if no component nor attribute with the specified
     * name is defined for the page. If set to <code>false</code>, which
     * is the default, an exception will be thrown.
     */
    private boolean ignoreIfAbsent;


    /**
     * Gets the name of the component reference or the attribute as it
     * is configured by <code>componentref</code> or
     * <code>attribute</code> element in the configuration file.
     *
     * @return the componentref or attribute name.
     */
    public String getName() {

        return this.name;
    }

    /**
     * Sets the name of the component reference or the attribute as it
     * is configured by <code>componentref</code> or
     * <code>attribute</code> element in the configuration file.
     *
     * @param name the componentref or attribute name.
     */
    public void setName(String name) {

        this.name = name;
    }

    /**
     * Gets the name of the bean containing component reference or page
     * attribute name.
     *
     * @return the bean name.
     */
    public String getNameBean() {

        return this.nameBean;
    }

    /**
     * Sets the name of the bean containing component reference or page
     * attribute name.
     *
     * @param nameBean the bean name.
     */
    public void setNameBean(String nameBean) {

        this.nameBean = nameBean;
    }

    /**
     * Gets name of the property in the <code>nameBean</code> bean
     * containing component reference or page attribute name.
     *
     * @return the property name.
     */
    public String getNameProperty() {

        return this.nameProperty;
    }

    /**
     * Sets name of the property in the <code>nameBean</code> bean
     * containing component reference or page attribute name.
     *
     * @param nameProperty the property name.
     */
    public void setNameProperty(String nameProperty) {

        this.nameProperty = nameProperty;
    }

    /**
     * Gets the <code>ignoreIfAbsent</code> attribute for the tag.
     *
     * @return <code>true</code> if <code>ignoreIfAbsent</code> flag
     * is set.
     */
    public boolean getIgnoreIfAbsent() {

        return this.ignoreIfAbsent;
    }

    /**
     * Sets the <code>ignoreIfAbsent</code> attribute for the tag.
     * If set to <code>true</code> the tag will be ignored if no component
     * nor attribute with the specified name is defined for the page.
     * If set to <code>false</code>, which is the default, an exception
     * will be thrown in such a case.
     *
     * @param ignoreIfAbsent value for the <code>ignoreIfAbsent</code>
     * attribute.
     */
    public void setIgnoreIfAbsent(boolean ignoreIfAbsent) {

        this.ignoreIfAbsent = ignoreIfAbsent;
    }


    /**
     * This implementation just skips the tag's whole body doing
     * nothing. The functionality of the tag is implemented in the
     * <code>doEndTag</code> method.
     *
     * @return <code>Tag.SKIP_BODY</code>.
     */
    public int doStartTag() {

        return SKIP_BODY;
    }

    /**
     * Does the actual job at the tag's end preparing required beans
     * in the page context and inserting the content.
     *
     * @return <code>Tag.EVAL_PAGE</code>.
     *
     * @throws JspException if an error occured.
     */
    public int doEndTag()
        throws JspException {

        try {

            // find the ref or attr name
            String refName = null;
            if(this.name != null) {
                refName = this.name;
            } else if(this.nameBean == null) {
                Object refNameO = this.pageContext.findAttribute(this.nameProperty);
                if(refNameO == null)
                    throw new JspException("There is no componentref/attribute name under \"" +
                                           this.nameProperty + "\" in any scope.");
                refName = refNameO.toString();
            } else {
                Object refNameBean = this.pageContext.findAttribute(this.nameBean);
                if(refNameBean == null)
                    throw new JspException("There is no bean under \"" +
                                           this.nameBean + "\" in any scope.");
                try {
                    PropertyDescriptor [] pds = Introspector.getBeanInfo(refNameBean.getClass()).
                        getPropertyDescriptors();
                    for(int i = 0; i < pds.length; i++) {
                        if(pds[i].getName().equals(this.nameProperty)) {
                            Object refNameO = pds[i].getReadMethod().
                                invoke(refNameBean, null);
                            if(refNameO == null)
                                throw new JspException("There is no componentref/attribute name in bean \"" +
                                                       this.nameBean + "\" under \"" + this.nameProperty + "\" property.");
                            refName = refNameO.toString();
                            break;
                        }
                    }
                    if(refName == null)
                        throw new JspException("There is no \"" +
                                               this.nameProperty + "\" property in bean \"" +
                                               this.nameBean + "\".");

                } catch(InvocationTargetException e) {
                    throw new JspException("Cannot read the bean property.", e);

                } catch(IllegalAccessException e) {
                    throw new JspException("Cannot read the bean property.", e);

                } catch(IntrospectionException e) {
                    throw new JspException("Cannot introspect the bean.", e);
                }
            }

            // attribute
            if(((Set)this.pageContext.getAttribute(ShowPageAction.RQKEY_ATTRIBUTES,
                                                   PageContext.REQUEST_SCOPE)).
                   contains(refName)) {

                // find the attribute value
                Object attrValue = this.pageContext.
                    getAttribute(this.name, PageContext.REQUEST_SCOPE);

                // did we find the value?
                if(attrValue == null) {
                    if(this.ignoreIfAbsent)
                        return EVAL_PAGE;

                    throw new JspException("There is no attribute named \"" + refName +
                                           "\" defined for this page.");
                }

                // render the attribute's value
                try {
                    this.pageContext.getOut().print(attrValue);
                } catch(IOException e) {
                    throw new JspException("Could not output an attribute's content.", e);
                }

            // component
            } else {

                // find the component runtime wrapper
                ComponentRuntimeWrapper cw =
                    (ComponentRuntimeWrapper)((Map)this.pageContext.
                                                  getAttribute(ShowPageAction.RQKEY_COMPONENTS,
                                                               PageContext.REQUEST_SCOPE)).
                        get(refName);

                // did we find the component?
                if(cw == null) {
                    if(this.ignoreIfAbsent)
                        return EVAL_PAGE;

                    throw new JspException("There is no component named \"" + refName +
                                           "\" defined for this page.");
                }

                // move pelement beans to the request attribtues and store the current ones
                PresentationElement [] oldPes = new PresentationElement[cw.pelements.length];
                for(int i = 0; i < cw.pelements.length; i++) {
                    oldPes[i] = (PresentationElement)this.pageContext.
                        getAttribute(cw.pelementNames[i], PageContext.REQUEST_SCOPE);
                    this.pageContext.setAttribute(cw.pelementNames[i], cw.pelements[i], PageContext.REQUEST_SCOPE);
                }

                // include the component
                try {
                    this.pageContext.include(cw.jsp);
                } catch(ServletException e) {
                    throw new JspException("Could not include a component's content.", e);
                } catch(IOException e) {
                    throw new JspException("Could not include a component's content.", e);
                }

                // restore pelement beans in the request attribtues
                for(int i = 0; i < cw.pelements.length; i++) {
                    this.pageContext.setAttribute(cw.pelementNames[i], oldPes[i], PageContext.REQUEST_SCOPE);
                }
            }

            // done, continue page evaluation
            return EVAL_PAGE;

        } catch(JspException e) {

            // get the root cause
            Throwable cause = e.getRootCause();

            // set it for Struts
            this.pageContext.setAttribute(Globals.EXCEPTION_KEY,
                                          (cause != null ? cause : e),
                                          PageContext.REQUEST_SCOPE);

            // log the exception
            this.getLog().
                error("Got error in <cb2:insert> tag, aborting the tag execution and throwing a JspException.",
                      (cause != null ? cause : e));

            // throw it up
            throw e;
        }
    }


    /**
     * Release the tag.
     */
    public void release() {

        this.name = null;
        this.nameBean = null;
        this.nameProperty = null;
        this.ignoreIfAbsent = false;
    }


    /**
     * Gets log for logging errors.
     *
     * @return the log.
     */
    private Log getLog() {

        return CB2PlugIn.getApplication(this.pageContext.getServletContext()).
            getApplicationContext().getLog(this.getClass().getName());
    }
}
