package com.boylesoftware.cb2.presentation.servlet;

import java.io.InputStream;
import java.io.IOException;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.beans.IntrospectionException;
import java.lang.reflect.Method;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.config.ModuleConfig;
import org.apache.struts.config.FormBeanConfig;

import com.boylesoftware.cb2.util.CruelErrorHandler;
import com.boylesoftware.cb2.util.LocalResourceEntityResolver;


/**
 * Presentation Level configuration built from the
 * <code>pages-config.xml</code>.
 *
 * @author Lev Himmelfarb
 * @version $Id: PLConfig.java,v 1.5 2004/10/11 18:17:36 levahim Exp $
 */
final class PLConfig {

    /**
     * List of public ids and local DTDs for the configuration XML
     * files.
     */
    private final static String [] LOCAL_DTDS = new String[] {
        "-//Boyle Software, Inc.//DTD CB2 Common Entities 1.0//EN",
        "/com/boylesoftware/cb2/common-entities_1_0.dtd",
        "-//Boyle Software, Inc.//DTD CB2 Pages Configuration 1.0//EN",
        "/com/boylesoftware/cb2/presentation/servlet/pages-config_1_0.dtd" };


    ///////////////////////////////////////////////////////////////////////////
    // Config data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Components. Keys are component names (objects of <code>String</code>),
     * values are component descriptors (objects of
     * <code>ComponentDesc</code>).
     */
    private final Map components;

    /**
     * Pages. Keys are page names (objects of <code>String</code>), values
     * are page descriptors (objects of <code>PageDesc</code>).
     */
    private final Map pages;


    ///////////////////////////////////////////////////////////////////////////
    // Initialization
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructs the PL configuration singleton parsing the configuration
     * files. It also builds and adds appropriate action mappings to Struts
     * module configuration.
     *
     * @param servletCtx the action servlet context.
     * @param pagesConfigFileURL context relative URL of the main pages
     * configuration file.
     * @param moduleConfig the Struts module configuration.
     *
     * @throws ServletException if the object could not be initialized.
     */
    PLConfig(ServletContext servletCtx,
             String pagesConfigFileURL,
             ModuleConfig moduleConfig)
        throws ServletException {

        // check if we have the pages-config file
        try {
            if(servletCtx.getResource(pagesConfigFileURL) == null) {
                this.components = new HashMap(0);
                this.pages = new HashMap(0);
                return; 
            }
        } catch(java.net.MalformedURLException e) {
            throw new ServletException("The pages-config.xml file URL is invalid.", e);
        }

        // prepare collectors
        Map components = new HashMap();
        Map allPages = new HashMap();
        Map globalPElements = new HashMap();
        Set sessionScopePElementNames = new HashSet();
        Map formBeanRefs = new HashMap();
        Map inputForms = new HashMap();
        Map preactionInstances = new HashMap();

        // the stack of queues of pages-config elements to process
        LinkedList pcListStack = new LinkedList();

        // put the top-level pages-config.xml to the stack
        List pcList = new ArrayList();
        pcList.add(pagesConfigFileURL);
        pcListStack.add(pcList);

        // parse the XMLs
        int nConcretePages = 0;
        try {

            // get the parser
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setIgnoringComments(true);
            dbf.setValidating(true);
            DocumentBuilder db = dbf.newDocumentBuilder();
            db.setEntityResolver(new LocalResourceEntityResolver(this.getClass(), LOCAL_DTDS));
            db.setErrorHandler(new CruelErrorHandler());

            // loop through the pages-config file queues
            while(pcListStack.size() > 0) {

                Iterator i = ((List)pcListStack.removeLast()).iterator();
                while(i.hasNext()) {
                    String pcFileURL = (String)i.next();

                    try {
                        InputStream pagesConfigIS = servletCtx.getResourceAsStream(pcFileURL);
                        if(pagesConfigIS == null)
                            throw new ServletException("Cannot find file at " + pcFileURL);

                        Document pagesConfigDoc = db.parse(pagesConfigIS);

                        nConcretePages +=
                            this.processPagesConfigElement(pagesConfigDoc.getDocumentElement(),
                                                           moduleConfig,
                                                           globalPElements,
                                                           sessionScopePElementNames,
                                                           formBeanRefs,
                                                           inputForms,
                                                           components,
                                                           allPages,
                                                           preactionInstances);

                        pcList = new ArrayList();
                        NodeList nl = pagesConfigDoc.getDocumentElement().
                            getElementsByTagName("external-pages-config");
                        for(int j = 0; j < nl.getLength(); j++)
                            pcList.add(((Element)nl.item(j)).getAttribute("file"));
                        if(pcList.size() > 0)
                            pcListStack.add(pcList);

                    } catch(SAXException e) {
                        throw new ServletException("Could not parse pages configuration file at " +
                                                   pcFileURL + ".", e);
                    } catch(IOException e) {
                        throw new ServletException("Could not read pages configuration file at " +
                                                   pcFileURL + ".", e);
                    }
                }
            }

        } catch(ParserConfigurationException e) {
            throw new ServletException("Could not create XML parser.", e);
        }

        // free up some memory
        globalPElements.clear();
        sessionScopePElementNames.clear();
        formBeanRefs.clear();
        inputForms.clear();
        preactionInstances.clear();

        // create final collection of components
        this.components = new HashMap(components);
        components.clear();

        // create final collection of all concrete pages and add mappings
        this.pages = new HashMap(nConcretePages);
        Iterator i = allPages.values().iterator();
        while(i.hasNext()) {
            PageDesc page = (PageDesc)i.next();
            if(page.path != null) { // is concrete?
                this.pages.put(page.name, page);
                ActionMapping am = new ActionMapping();
                am.setPath(page.path);
                am.setType((ShowPageAction.class).getName());
                if(page.flags != null)
                    am.setParameter(page.name + "|" + page.flags);
                else
                    am.setParameter(page.name);
                if(page.inputForm != null) {
                    am.setName(page.inputForm.formBeanName);
                    am.setScope(page.inputForm.scope == FormBeanRefDesc.SCOPE_SESSION ?
                                "session" :
                                "request");
                    am.setValidate(false);
                }
                moduleConfig.addActionConfig(am);
            }
        }
        allPages.clear();
    }

    /**
     * Processes an XML <code>pages-config</code> element.
     *
     * @param pcElement the <code>pages-config</code> element.
     * @param moduleConfig Struts module config object, used for reference.
     * @param globalPElements in/out, collection of global pelements.
     * Keys are PE names (objects of <code>String</code>),
     * values are PE descriptors (objects of
     * <code>PresentationElementDesc</code>).
     * @param sessionScopePElementNames in/out, names
     * (objects of <code>String</code>) of all registered pelements of session
     * scope.
     * @param formBeanRefs out, all form bean references. The keys are
     * reference names (objects of <code>String</code>), the values are
     * reference descriptors (objects of <code>FormBeanRefDesc</code>).
     * @param inputForms out, all pelements with input forms. The keys
     * are descriptors of pelements with inputForms (objects of
     * <code>PresentationElementDesc</code>, the values are corresponding form
     * bean reference descriptors (objects of <code>FormBeanRefDesc</code>).
     * @param components in/out, collection of component descriptors. Keys
     * are component names (objects of <code>String</code>), values are
     * component descriptors (objects of <code>ComponentDesc</code>).
     * @param allPages in/out, collection of page descriptors. Keys are
     * page names (objects of <code>String</code>), values are page
     * descriptors (objects of <code>PageDesc</code>).
     * @param preactionInstances in/out, collection of pre-actions.
     *
     * @return number of concrete pages added to <code>allPages</code>
     * collection by this call.
     *
     * @throws ServletException if an error is detected.
     */
    private int processPagesConfigElement(Element pcElement,
                                          ModuleConfig moduleConfig,
                                          Map globalPElements,
                                          Set sessionScopePElementNames,
                                          Map formBeanRefs,
                                          Map inputForms,
                                          Map components,
                                          Map allPages,
                                          Map preactionInstances)
        throws ServletException {
        NodeList nl;

        // process form bean references
        nl = pcElement.getElementsByTagName("formbeanref");
        for(int i = 0; i < nl.getLength(); i++) {
            FormBeanRefDesc fbref =
                this.makeFormBeanRefDesc((Element)nl.item(i), moduleConfig);
            if(formBeanRefs.put(fbref.name, fbref) != null)
                throw new ServletException("Two or more form bean references are defined with name \"" +
                                           fbref.name + "\".");
        }

        // process global pelements
        nl = pcElement.getElementsByTagName("global-pelement");
        for(int i = 0; i < nl.getLength(); i++) {
            PresentationElementDesc gpe =
                this.makePresentationElementDesc((Element)nl.item(i),
                                                 globalPElements,
                                                 sessionScopePElementNames,
                                                 formBeanRefs,
                                                 inputForms,
                                                 moduleConfig,
                                                 true);
            if(globalPElements.put(gpe.name, gpe) != null)
                throw new ServletException("Two or more global presentation elements are defined with name \"" +
                                           gpe.name + "\".");
        }

        // process components
        nl = pcElement.getElementsByTagName("component");
        for(int i = 0; i < nl.getLength(); i++) {
            ComponentDesc component = makeComponentDesc((Element)nl.item(i),
                                                        globalPElements,
                                                        sessionScopePElementNames,
                                                        formBeanRefs,
                                                        inputForms,
                                                        moduleConfig);
            if(components.put(component.name, component) != null)
                throw new ServletException("Two or more components are defined with name \"" +
                                           component.name + "\".");
        }

        // process pages
        int nConcretePages = 0;
        nl = pcElement.getElementsByTagName("page");
        for(int i = 0; i < nl.getLength(); i++) {
            PageDesc page = this.makePageDesc((Element)nl.item(i),
                                              components,
                                              allPages,
                                              preactionInstances,
                                              inputForms,
                                              moduleConfig);
            if(allPages.put(page.name, page) != null)
                throw new ServletException("Two or more pages are defined with name \"" +
                                           page.name + "\".");
            if(page.path != null)
                nConcretePages++;
        }

        // done
        return nConcretePages;
    }

    /**
     * Creates a page descriptor from a <code>page</code> XML element.
     *
     * @param element the <code>page</code> XML element.
     * @param components collection of components. Keys are component names
     * (objects of <code>String</code>), values are component descriptors
     * (objects of <code>ComponentDesc</code>).
     * @param pages collection of pages. Keys are page names (objects of
     * <code>String</code>), values are page descriptors (objects of
     * <code>PageDesc</code>).
     * @param preactionInstances in/out, action instances for all preactions.
     * Keys are action class names (objects of <code>String</code>),
     * values are action instances (objects of <code>CB2Action</code>).
     * @param inputForms all pelements with input forms. The keys are
     * descriptors of pelements with inputForms (objects of
     * <code>PresentationElementDesc</code>, the values are corresponding form
     * bean reference descriptors (objects of <code>FormBeanRefDesc</code>).
     * @param moduleConfig the Struts module config.
     *
     * @return the page descriptor.
     *
     * @throws ServletException if the element is invalid.
     */
    private PageDesc makePageDesc(Element element,
                                  Map components,
                                  Map pages,
                                  Map preactionInstances,
                                  Map inputForms,
                                  ModuleConfig moduleConfig)
        throws ServletException {
        NodeList nl;

        // get the name
        String name = element.getAttribute("name");
        if(name.indexOf('|') >= 0)
            throw new ServletException("Page name \"" + name +
                                       "\" contains a pipe symbol in it, which is not allowed.");

        // get the path
        String path = (element.hasAttribute("path") ? element.getAttribute("path") : null);

        // get the content type
        String contentType = (element.hasAttribute("contentType") ? element.getAttribute("contentType") : null);

        // get the flags
        String flags = (element.hasAttribute("flags") ? element.getAttribute("flags") : null);

        // check attribute consistency
        if(element.hasAttribute("template") && element.hasAttribute("extends"))
            throw new ServletException("Page \"" + name + "\" has both \"template\" and \"extends\" attributes.");

        // find template component and create root action forward
        ActionForward root;
        ComponentDesc templateComponent = null;
        PageDesc parentPage = null;
        if(element.hasAttribute("template")) {
            templateComponent = (ComponentDesc)components.
                get(element.getAttribute("template"));
            if(templateComponent == null)
                throw new ServletException("Page \"" + name + "\" referes to a nonexistant or not yet defined component \"" +
                                           element.getAttribute("template") + "\" as its template component.");
            root = new ActionForward("success", templateComponent.src, false);

        } else if(element.hasAttribute("extends")) {
            parentPage = (PageDesc)pages.
                get(element.getAttribute("extends"));
            if(parentPage == null)
                throw new ServletException("Page \"" + name + "\" extends a nonexistant or not yet defined page \"" +
                                           element.getAttribute("extends") + "\".");
            templateComponent = parentPage.templateComponent;
            root = parentPage.root;

        } else
            throw new ServletException("Page \"" + name + "\" has neither template nor extends attribute.");

        // get component references and build final list of refered global pelements
        ArrayList globalPElementsList = new ArrayList();
        Set globalPElementNames = new HashSet();
        Set localPElementNames = new HashSet();
        Map compRefsMap = new HashMap();
        if(parentPage != null) {
            for(int i = 0; i < parentPage.componentRefs.length; i++) {
                compRefsMap.put(parentPage.componentRefs[i].name,
                                parentPage.componentRefs[i]);
                for(int j = 0; j < parentPage.componentRefs[i].component.pelements.length; j++)
                    localPElementNames.add(parentPage.componentRefs[i].component.pelements[j].name);
            }
            for(int i = 0; i < parentPage.globalPElements.length; i++) {
                globalPElementsList.add(parentPage.globalPElements[i]);
                globalPElementNames.add(parentPage.globalPElements[i].name);
            }
        } else {
            for(int i = 0; i < templateComponent.dependsOn.length; i++)
                if(!globalPElementNames.contains(templateComponent.dependsOn[i].name)) {
                    globalPElementsList.add(templateComponent.dependsOn[i]);
                    globalPElementNames.add(templateComponent.dependsOn[i].name);
                }
        }
        nl = element.getElementsByTagName("componentref");
        Map newCompRefsMap = new HashMap(nl.getLength() + compRefsMap.size());
        for(int i = 0; i < nl.getLength(); i++) {
            Element compRefElem = (Element)nl.item(i);
            ComponentDesc component = (ComponentDesc)components.
                get(compRefElem.getAttribute("component"));
            if(component == null)
                throw new ServletException("Page \"" + name + "\" referes to a nonexistant component \"" +
                                           compRefElem.getAttribute("component") + "\".");
            NodeList nl1 = compRefElem.getElementsByTagName("attribute");
            Map compRefAttrs = new HashMap(nl1.getLength());
            for(int j = 0; j < nl1.getLength(); j++) {
                Element compRefAttrElem = (Element)nl.item(j);
                if(compRefAttrs.put(compRefAttrElem.getAttribute("name"),
                                    compRefAttrElem.getAttribute("value")) != null)
                    throw new ServletException("Two or more attributes with name \"" +
                                               compRefAttrElem.getAttribute("name") +
                                               "\" are defined for component reference \"" +
                                               compRefElem.getAttribute("name") +
                                               "\" in page \"" + name + "\".");
            }
            PageDesc.ComponentRefDesc compRef =
                new PageDesc.ComponentRefDesc(compRefElem.getAttribute("name"),
                                              component,
                                              compRefAttrs);
            for(int j = 0; j < component.pelements.length; j++)
                localPElementNames.add(component.pelements[j].name);
            if(newCompRefsMap.put(compRef.name, compRef) != null)
                throw new ServletException("Two or more component references with name \"" + compRef.name +
                                           "\" are defined for page \"" + name + "\".");

            for(int j = 0; j < component.dependsOn.length; j++) {
                if(!globalPElementNames.contains(component.dependsOn[j].name)) {
                    globalPElementsList.add(component.dependsOn[j]);
                    globalPElementNames.add(component.dependsOn[j].name);
                }
            }
        }
        compRefsMap.putAll(newCompRefsMap);

        // get attributes
        Map attrsMap = new HashMap();
        if(parentPage != null) {
            for(int i = 0; i < parentPage.attributes.length; i++) {
                if(globalPElementNames.contains(parentPage.attributes[i].name))
                    throw new ServletException("One of parent pages of page \"" + name +
                                               "\" defines attribute \"" +
                                               parentPage.attributes[i].name +
                                               "\", which has the same name as one of global pelements" +
                                               " on which components on this page depend.");
                if(localPElementNames.contains(parentPage.attributes[i].name))
                    throw new ServletException("One of parent pages of page \"" + name +
                                               "\" defines attribute \"" +
                                               parentPage.attributes[i].name +
                                               "\", which has the same name as one of pelements" +
                                               " defined in components used on this page.");
                attrsMap.put(parentPage.attributes[i].name,
                             parentPage.attributes[i]);
            }
        }
        nl = element.getElementsByTagName("attribute");
        Map newAttrsMap = new HashMap(nl.getLength() + attrsMap.size());
        for(int i = 0; i < nl.getLength(); i++) {
            Element attrElem = (Element)nl.item(i);
            PageDesc.AttributeDesc attr =
                new PageDesc.AttributeDesc(attrElem.getAttribute("name"),
                                           attrElem.getAttribute("value"));
            if(globalPElementNames.contains(attr.name))
                throw new ServletException("Page \"" + name + "\" defines attribute \"" +
                                           attr.name + "\", which has the same name as one of global pelements" +
                                           " on which components on this page depend.");
            if(localPElementNames.contains(attr.name))
                throw new ServletException("Page \"" + name + "\" defines attribute \"" +
                                           attr.name + "\", which has the same name as one of pelements" +
                                           " defined in components used on this page.");
            if(newAttrsMap.put(attr.name, attr) != null)
                throw new ServletException("Two or more attributes with name \"" + attr.name +
                                           "\" are defined for page \"" + name + "\".");
        }
        attrsMap.putAll(newAttrsMap);
        Map attributesMap = new HashMap(attrsMap.size());
        Iterator n = attrsMap.values().iterator();
        while(n.hasNext()) {
            PageDesc.AttributeDesc attr = (PageDesc.AttributeDesc)n.next();
            attributesMap.put(attr.name, attr.value);
        }

        // find the input form
        FormBeanRefDesc inputForm = null;
        for(int i = 0; i < templateComponent.pelements.length; i++) { // search template components
            FormBeanRefDesc fbref = (FormBeanRefDesc)inputForms.
                get(templateComponent.pelements[i]);
            if(fbref != null) {
                if(inputForm != null)
                    throw new ServletException("Page \"" + name +
                                               "\" referes pelements expecting different input forms.");

                inputForm = fbref;
            }
        }
        for(Iterator i = globalPElementsList.iterator(); i.hasNext();) { // search global PEs
            FormBeanRefDesc fbref = (FormBeanRefDesc)inputForms.
                get(i.next());
            if(fbref != null) {
                if(inputForm != null)
                    throw new ServletException("Page \"" + name +
                                               "\" referes pelements expecting different input forms.");

                inputForm = fbref;
            }
        }
        for(Iterator i = compRefsMap.values().iterator(); i.hasNext();) { // search refered components
            PageDesc.ComponentRefDesc crd = (PageDesc.ComponentRefDesc)i.next();
            for(int j = 0; j < crd.component.pelements.length; j++) {
                FormBeanRefDesc fbref = (FormBeanRefDesc)inputForms.
                    get(crd.component.pelements[j]);
                if(fbref != null) {
                    if(inputForm != null)
                        throw new ServletException("Page \"" + name +
                                                   "\" referes pelements expecting different input forms.");

                    inputForm = fbref;
                }
            }
        }

        // get preactions
        List preactions = new ArrayList();
        if(parentPage != null) {
            for(int i = 0; i < parentPage.preactions.length; i++)
                preactions.add(parentPage.preactions[i]);
        }
        nl = element.getElementsByTagName("preaction");
        for(int i = 0; i < nl.getLength(); i++) {
            Element preactElem = (Element)nl.item(i);
            String className = preactElem.getAttribute("class");
            CB2Action action = (CB2Action)preactionInstances.get(className);
            if(action == null) {
                try {
                    action = (CB2Action)Class.forName(className).newInstance();
                } catch(ClassNotFoundException e) {
                    throw new ServletException("Could not load preaction class " +
                                               className +
                                               " for page \"" + name + "\".", e);
                } catch(InstantiationException e) {
                    throw new ServletException("Could not instnatiate preaction class " +
                                               className +
                                               " for page \"" + name + "\".", e);
                } catch(IllegalAccessException e) {
                    throw new ServletException("Could not instnatiate preaction class " +
                                               className +
                                               " for page \"" + name + "\".", e);
                }
                preactionInstances.put(className, action);
            }
            ActionMapping actionMapping = new ActionMapping();
            actionMapping.setModuleConfig(moduleConfig);
            actionMapping.setParameter(preactElem.getAttribute("flags"));
            actionMapping.setType(className);
            preactions.add(new PageDesc.PreactionDesc(action,
                                                      actionMapping));
        }

        // get onerror handlers
        List onerrorActions = new ArrayList();
        Set addedNames = new HashSet();
        try {
            nl = element.getElementsByTagName("onerror");
            for(int i = 0; i < nl.getLength(); i++) {
                Element onerrorElem = (Element)nl.item(i);
                Class exceptionClass = Class.forName(onerrorElem.getAttribute("class"));
                ActionForward forward =
                    new ActionForward("onerror:" + exceptionClass.getName(),
                                      onerrorElem.getAttribute("path"),
                                      Boolean.valueOf(onerrorElem.getAttribute("redirect")).booleanValue());
                onerrorActions.add(new PageDesc.OnerrorActionDesc(exceptionClass,
                                                                  forward));
                if(!addedNames.add(exceptionClass.getName()))
                    throw new ServletException("Two or more error handlers defined for exception class " +
                                               exceptionClass.getName() + " in page \"" + name + "\".");
            }
        } catch(ClassNotFoundException e) {
            throw new ServletException("Could not load exception class in an onerror element of page \"" +
                                       name + "\"", e);
        }
        if(parentPage != null) {
            for(int i = 0; i < parentPage.onerrorActions.length; i++) {
                if(!addedNames.contains(parentPage.onerrorActions[i].exceptionClass.getName()))
                    onerrorActions.add(parentPage.onerrorActions[i]);
            }
        }

        // create and return page descriptor
        return new PageDesc(name,
                            path,
                            contentType,
                            flags,
                            templateComponent,
                            (PresentationElementDesc [])globalPElementsList.
                                toArray(new PresentationElementDesc[globalPElementsList.size()]),
                            (PageDesc.ComponentRefDesc [])compRefsMap.values().
                                toArray(new PageDesc.ComponentRefDesc[compRefsMap.size()]),
                            (PageDesc.AttributeDesc [])attrsMap.values().
                                toArray(new PageDesc.AttributeDesc[attrsMap.size()]),
                            attributesMap,
                            inputForm,
                            root,
                            (PageDesc.PreactionDesc [])preactions.
                                toArray(new PageDesc.PreactionDesc[preactions.size()]),
                            (PageDesc.OnerrorActionDesc [])onerrorActions.
                                toArray(new PageDesc.OnerrorActionDesc[onerrorActions.size()]));
    }

    /**
     * Constructs a component descriptor from a <code>component</code>
     * XML element.
     *
     * @param element a <code>component</code> XML element.
     * @param globalPElements collection of global pelements.
     * Keys are PE names (objects of <code>String</code>),
     * values are PE descriptors (objects of
     * <code>PresentationElementDesc</code>).
     * @param sessionScopePElementNames names (objects of
     * <code>String</code>) of all registered pelements of session scope.
     * @param formBeanRefs all form bean references. The keys are reference
     * names (objects of <code>String</code>), the values are reference
     * descriptors (objects of <code>FormBeanRefDesc</code>).
     * @param inputForms all pelements with input forms. The keys are
     * descriptors of pelements with inputForms (objects of
     * <code>PresentationElementDesc</code>, the values are corresponding form
     * bean reference descriptors (objects of <code>FormBeanRefDesc</code>).
     * @param moduleConfig config for the Struts module.
     *
     * @throws ServletException if the element is invalid.
     */
    private ComponentDesc makeComponentDesc(Element element,
                                            Map globalPElements,
                                            Set sessionScopePElementNames,
                                            Map formBeanRefs,
                                            Map inputForms,
                                            ModuleConfig moduleConfig)
        throws ServletException {

        // get the name
        String name = element.getAttribute("name");

        // get the src
        String src = element.getAttribute("src");

        // get global pelement dependencies
        PresentationElementDesc [] dependsOn = this.
            makeDependsOnList(element, globalPElements);

        // get local pelements
        NodeList nl = element.getElementsByTagName("pelement");
        PresentationElementDesc [] pelements =
            new PresentationElementDesc[nl.getLength()];
        for(int i = 0; i < nl.getLength(); i++) {
            pelements[i] =
                this.makePresentationElementDesc((Element)nl.item(i),
                                                 globalPElements,
                                                 sessionScopePElementNames,
                                                 formBeanRefs,
                                                 inputForms,
                                                 moduleConfig,
                                                 false);
        }

        // create the component descriptor and return it
        return new ComponentDesc(name, src, dependsOn, pelements);
    }

    /**
     * Constructs a new pelement descriptor from an XML element.
     *
     * @param element the <code>global-pelement</code> or
     * <code>pelement</code> XML element.
     * @param globalPElements collection of already processed
     * global pelements. Keys are PE names (objects of <code>String</code>),
     * values are PE descriptors (objects of
     * <code>PresentationElementDesc</code>).
     * @param sessionScopePElementNames in/out, names (objects of
     * <code>String</code>) of all registered so far pelements
     * of session scope.
     * @param formBeanRefs all form bean references. The keys are reference
     * names (objects of <code>String</code>), the values are reference
     * descriptors (objects of <code>FormBeanRefDesc</code>).
     * @param inputForms out, all pelements with input forms. The keys are
     * descriptors of pelements with inputForms (objects of
     * <code>PresentationElementDesc</code>, the values are corresponding form
     * bean reference descriptors (objects of <code>FormBeanRefDesc</code>).
     * @param moduleConfig config for the Struts module.
     * @param isGlobal tells if this is a global pelement.
     *
     * @throws ServletException if the element is invalid.
     */
    private PresentationElementDesc makePresentationElementDesc(Element element,
                                                                Map globalPElements,
                                                                Set sessionScopePElementNames,
                                                                Map formBeanRefs,
                                                                Map inputForms,
                                                                ModuleConfig moduleConfig,
                                                                boolean isGlobal)
        throws ServletException {

        // get pelement's name
        String name = element.getAttribute("name");

        // get pelement's scope
        int scope;
        if("session".equalsIgnoreCase(element.getAttribute("scope"))) {
            scope = PresentationElementDesc.SCOPE_SESSION;
            if(!sessionScopePElementNames.add(name))
                throw new ServletException("There are two or more session scope presentation elements with name \"" +
                                           name + "\".");
        } else
            scope = PresentationElementDesc.SCOPE_REQUEST;

        // get pelement's implementation class
        Class pelementClass;
        try {
            pelementClass = Class.forName(element.getAttribute("class"));
        } catch(ClassNotFoundException e) {
            throw new ServletException("Could not load implementation class " +
                                       element.getAttribute("class") + " for pelement \"" +
                                       name + "\".", e);
        }
        if(!(PresentationElement.class).isAssignableFrom(pelementClass))
            throw new ServletException("Class " + pelementClass.getName() +
                                       ", used as implementation class for pelement \"" +
                                       name + "\", does not implement PresentationElement interface.");

        // get input form
        FormBeanRefDesc inputForm = null;
        if(element.hasAttribute("inputForm")) {
            inputForm = (FormBeanRefDesc)formBeanRefs.get(element.getAttribute("inputForm"));
            if(inputForm == null)
                throw new ServletException("Reference to nonexistant form bean reference \"" +
                                           element.getAttribute("inputForm") +
                                           "\" as an input form in pelement \"" +
                                           name + "\" definition.");
        }

        // get dependencies list
        PresentationElementDesc [] dependsOn = this.
            makeDependsOnList(element, globalPElements);

        // create pelement parameter descriptors
        PropertyDescriptor [] pelementBeanProps;
        try {
            pelementBeanProps = Introspector.getBeanInfo(pelementClass).getPropertyDescriptors();
        } catch(IntrospectionException e) {
            throw new ServletException("The pelement \"" + name +
                                       "\" class " + pelementClass.getName() +
                                       " could not be introspected.", e);
        }
        NodeList nl = element.getElementsByTagName("param");
        Set addedNames = new HashSet(nl.getLength());
        PresentationElementDesc.ParameterDesc [] parameters =
            new PresentationElementDesc.ParameterDesc[nl.getLength()];
        for(int i = 0; i < nl.getLength(); i++) {
            Element paramElement = (Element)nl.item(i);

            String paramName = paramElement.getAttribute("name");
            if(addedNames.contains(paramName))
                throw new ServletException("The pelement \"" + name +
                                           "\" has two or more parameters with name \"" +
                                           paramName + "\".");
            addedNames.add(paramName);

            boolean isRequired = "true".equals(paramElement.getAttribute("required"));

            String emptyStringAttr = paramElement.getAttribute("emptystring");
            int emptyString;
            if("disallow".equals(emptyStringAttr))
                emptyString = PresentationElementDesc.ParameterDesc.EMPTYSTRING_DISALLOW;
            else if("asis".equals(emptyStringAttr))
                emptyString = PresentationElementDesc.ParameterDesc.EMPTYSTRING_ASIS;
            else
                emptyString = PresentationElementDesc.ParameterDesc.EMPTYSTRING_NULL;

            String fromAttr = paramElement.getAttribute("from");
            int from;
            if("rqparam".equals(fromAttr))
                from = PresentationElementDesc.ParameterDesc.FROM_RQPARAM;
            else if("rqattribute".equals(fromAttr))
                from = PresentationElementDesc.ParameterDesc.FROM_RQATTRIBUTE;
            else if("session".equals(fromAttr))
                from = PresentationElementDesc.ParameterDesc.FROM_SESSION;
            else if("any".equals(fromAttr))
                from = PresentationElementDesc.ParameterDesc.FROM_ANY;
            else if("crefattribute".equals(fromAttr)) {
                if(isGlobal)
                    throw new ServletException("Global presentation element \"" +
                                               name + "\" has input parameter with \" +" +
                                               "from\" attribute set to \"" +
                                               "crefattribute\", which is invalid.");
                from = PresentationElementDesc.ParameterDesc.FROM_CREFATTRIBUTE;
            } else
                from = PresentationElementDesc.ParameterDesc.FROM_REQUEST;

            String fieldName = paramElement.getAttribute("fieldName");
            if((fieldName == null) || (fieldName.length() == 0))
                fieldName = paramName;
            Method setMethod = null;
            for(int j = 0; j < pelementBeanProps.length; j++) {
                if(pelementBeanProps[j].getName().equals(fieldName)) {
                    if((setMethod = pelementBeanProps[j].getWriteMethod()) == null)
                        throw new ServletException("No set method defined for param \"" + paramName +
                                                   "\" in pelement class " + pelementClass.getName() + ".");
                    break;
                }
            }
            if(setMethod == null)
                throw new ServletException("There is no bean property for param \"" + paramName +
                                           "\" in pelement class " + pelementClass.getName() + ".");

            parameters[i] = new PresentationElementDesc.ParameterDesc(paramName,
                                                                      setMethod,
                                                                      isRequired,
                                                                      emptyString,
                                                                      from);
        }
        addedNames.clear();

        // create pelement descriptor
        PresentationElementDesc res = new PresentationElementDesc(name,
                                                                  pelementClass,
                                                                  scope,
                                                                  inputForm,
                                                                  dependsOn,
                                                                  parameters);

        // add it to inputForms collection
        if(inputForm != null)
            inputForms.put(res, res.inputForm);

        // done, return the descriptor
        return res;
    }

    /**
     * Builds a filtered (containing no duplicates) list of global PE
     * dependencies including all nested dependencies basing on a set
     * of <code>depends</code> XML elements.
     *
     * @param element an XML element containing <code>depends</code>
     * subelements.
     * @param globalPElements collection of global pelements.
     * Keys are PE names (objects of <code>String</code>), values are
     * PE descriptors (objects of <code>PresentationElementDesc</code>).
     *
     * @return the dependency list.
     *
     * @throws ServletException if elements are invalid.
     */
    private PresentationElementDesc [] makeDependsOnList(Element element,
                                                         Map globalPElements)
        throws ServletException {

        List dependsOnList = new ArrayList();
        Set addedNames = new HashSet();
        NodeList nl = element.getElementsByTagName("depends");
        for(int i = 0; i < nl.getLength(); i++) {
            String on = ((Element)nl.item(i)).getAttribute("on");
            PresentationElementDesc pelement =
                (PresentationElementDesc)globalPElements.get(on);
            if(pelement == null)
                throw new ServletException("Found dependency on a nonexistant or not yet defined global pelement \"" +
                                           on + "\".");
            for(int j = 0; j < pelement.dependsOn.length; j++) {
                if(addedNames.add(pelement.dependsOn[j].name))
                    dependsOnList.add(pelement.dependsOn[j]);
            }
            if(addedNames.add(pelement.name))
                dependsOnList.add(pelement);
        }
        addedNames.clear();

        return (PresentationElementDesc [])dependsOnList.
            toArray(new PresentationElementDesc[dependsOnList.size()]);
    }

    /**
     * Creates a descriptor of form bean reference from a
     * <code>formbeanref</code> XML element. It also creates Struts form bean
     * definition if necessary.
     *
     * @param element the <code>formbeanref</code> XML element.
     * @param moduleConfig the Struts module config.
     *
     * @return the form bean reference descriptor.
     *
     * @throws ServletException if the element is invalid.
     */
    private FormBeanRefDesc makeFormBeanRefDesc(Element element,
                                                ModuleConfig moduleConfig)
        throws ServletException {

        // get the reference name
        String name = element.getAttribute("name");

        // get the form bean name
        String formBeanName = (element.hasAttribute("formBeanName") ?
                               element.getAttribute("formBeanName") :
                               name);

        // get the implementation class if specified
        Class formBeanClass = null;
        if(element.hasAttribute("class")) {
            try {
                formBeanClass = Class.forName(element.getAttribute("class"));
            } catch(ClassNotFoundException e) {
                throw new ServletException("Could not load implementation class " +
                                           element.getAttribute("class") +
                                           " for form bean \"" + formBeanName +
                                           "\".", e);
            }
        }

        // check/create Struts form bean definition
        FormBeanConfig fbc = moduleConfig.findFormBeanConfig(formBeanName);
        if(fbc != null) {
            if(formBeanClass != null)
                throw new ServletException("Form bean reference \"" + name +
                                           "\" has \"class\" attribute, while " +
                                           "there is a definition for the corresponding " +
                                           "form bean in Struts module configuration.");
        } else {
            if(formBeanClass == null)
                throw new ServletException("Cannot find corresponding form bean " +
                                           "definition in Struts module configuration " +
                                           "for form bean reference \"" + name +
                                           "\", nor \"class\" attribute is specified.");
            if(!(ActionForm.class).isAssignableFrom(formBeanClass))
                throw new ServletException("Class " + formBeanClass.getName() +
                                           ", specified in form bean reference \"" +
                                           name + "\", does not extend ActionForm.");
            fbc = new FormBeanConfig();
            fbc.setModuleConfig(moduleConfig);
            fbc.setName(formBeanName);
            fbc.setType(formBeanClass.getName());
            moduleConfig.addFormBeanConfig(fbc);
        }

        // get the scope
        int scope;
        if("session".equalsIgnoreCase(element.getAttribute("scope"))) {
            scope = FormBeanRefDesc.SCOPE_SESSION;
        } else
            scope = PresentationElementDesc.SCOPE_REQUEST;

        // create and return reference descriptor
        return new FormBeanRefDesc(name, formBeanName, formBeanClass, scope);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Service interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Gets the named component.
     *
     * <p>The method is thread-safe.
     *
     * @param name the component name.
     *
     * @return the requested component descriptor or <code>null</code> if
     * there is no component defined under this name.
     */
    ComponentDesc getComponent(String name) {

        synchronized(this.components) {

            return (ComponentDesc)this.components.get(name);
        }
    }

    /**
     * Gets the named page. This method is called from the
     * <code>ShowPageAction</code>, which gets the page name from its
     * action mapping parameter.
     *
     * <p>The method is thread-safe.
     *
     * @param name the page name.
     *
     * @return the requested page descriptor or <code>null</code> if
     * there is no such page defined.
     */
    PageDesc getPage(String name) {

        synchronized(this.pages) {

            return (PageDesc)this.pages.get(name);
        }
    }
}
