package com.boylesoftware.cb2.presentation.servlet;

import java.util.Map;

import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;


/**
 * Descriptor of a component-based page.
 *
 * @author Lev Himmelfarb
 * @version $Id: PageDesc.java,v 1.3 2004/05/20 19:19:52 levahim Exp $
 */
final class PageDesc {


    ///////////////////////////////////////////////////////////////////////////
    // Descriptor classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Descriptor of a component reference.
     */
    static final class ComponentRefDesc {

        /**
         * The reference name. The component will be available to the
         * page under this name.
         */
        final String name;

        /**
         * The referred component's descriptor.
         */
        final ComponentDesc component;

        /**
         * Component reference attributes. Keys are attribute names (objects
         * of <code>String</code>), values are attribute values (objects of
         * <code>Object</code>). Cannot be <code>null</code>.
         */
        final Map attributes;


        /**
         * Constructs a reference descriptor.
         *
         * @param name the reference name.
         * @param component the referred component's descriptor.
         * @param attributes component reference attributes.
         */
        ComponentRefDesc(String name,
                         ComponentDesc component,
                         Map attributes) {

            this.name = name;
            this.component = component;
            this.attributes = attributes;
        }
    }


    /**
     * Page attribute descriptor.
     */
    static final class AttributeDesc {

        /**
         * The attribute's name.
         */
        final String name;

        /**
         * The attribute's value.
         */
        final String value;


        /**
         * Constructs an attribute descriptor.
         *
         * @param name the attribute's name.
         * @param value the attribute's value.
         */
        AttributeDesc(String name,
                      String value) {

            this.name = name;
            this.value = value;
        }
    }


    /**
     * Descriptor of a page preaction.
     */
    static final class PreactionDesc {

        /**
         * The action object.
         */
        final CB2Action action;

        /**
         * "Fake" action mapping.
         */
        final ActionMapping actionMapping;


        /**
         * Constructs a descriptor.
         *
         * @param action the action.
         * @param actionMapping "fake" action mapping.
         */
        PreactionDesc(CB2Action action,
                      ActionMapping actionMapping) {

            this.action = action;
            this.actionMapping = actionMapping;
        }
    }


    /**
     * Descriptor of a page "onerror" action.
     */
    static final class OnerrorActionDesc {

        /**
         * The exception class to catch.
         */
        final Class exceptionClass;

        /**
         * The forward to return.
         */
        final ActionForward forward;


        /**
         * Constructs a descriptor.
         *
         * @param exceptionClass class of the exception to catch.
         * @param forward forward to return.
         */
        OnerrorActionDesc(Class exceptionClass,
                          ActionForward forward) {

            this.exceptionClass = exceptionClass;
            this.forward = forward;
        }
    }


    ///////////////////////////////////////////////////////////////////////////
    // Page data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * The page name.
     */
    final String name;

    /**
     * Path where this page is mapped. For abstract pages it is
     * <code>null</code>.
     */
    final String path;

    /**
     * The page content type, or <code>null</code> if should be left untouched.
     */
    final String contentType;

    /**
     * Page action flags. This string is also added to the page's
     * action mapping parameter after the page name and a pipe (|) symbol.
     * May be <code>null</code>.
     */
    final String flags;

    /**
     * Descriptor of the page's template component.
     */
    final ComponentDesc templateComponent;

    /**
     * Preprocessed list of descriptors of global pelements on which components
     * used with this page depend. This list contains no duplicates.
     */
    final PresentationElementDesc [] globalPElements;

    /**
     * List of descriptors of component references included in this page.
     */
    final ComponentRefDesc [] componentRefs;

    /**
     * List of page attributes.
     */
    final AttributeDesc [] attributes;

    /**
     * Page attributes organised into a map. The keys are attribute names
     * (objects of <code>String</code>), the values are attribute values
     * (objects of <code>String</code>).
     */
    final Map attributesMap;

    /**
     * The input form refefence or <code>null</code>.
     */
    final FormBeanRefDesc inputForm;

    /**
     * Struts action forward pointed to the page's top-level JSP.
     */
    final ActionForward root;

    /**
     * Ordered list of preactions for the page.
     */
    final PreactionDesc [] preactions;

    /**
     * Ordered list of onerror handlers.
     */
    final OnerrorActionDesc [] onerrorActions;


    /**
     * Constructs a page descriptor.
     *
     * @param name the page name.
     * @param path path where this page is mapped, <code>null</code> for
     * abstract pages.
     * @param contentType the page content type, <code>null</code> if should be
     * left untouched.
     * @param flags page action flags, may be <code>null</code>.
     * @param templateComponent descriptor of the page's template component.
     * @param globalPElements preprocessed list of descriptors of global
     * pelements on which components used with this page depend. This list
     * should contain no duplicates.
     * @param componentRefs list of descriptors of component references
     * included in this page.
     * @param attributes list of page attributes.
     * @param attributesMap page attributes as a map.
     * @param inputForm the input form reference or <code>null</code>.
     * @param root Struts action forward pointed to the page's top-level JSP.
     * @param preactions ordered list of preactions for the page.
     * @param onerrorActions ordered list of onerror handlers.
     */
    PageDesc(String name,
             String path,
             String contentType,
             String flags,
             ComponentDesc templateComponent,
             PresentationElementDesc [] globalPElements,
             ComponentRefDesc [] componentRefs,
             AttributeDesc [] attributes,
             Map attributesMap,
             FormBeanRefDesc inputForm,
             ActionForward root,
             PreactionDesc [] preactions,
             OnerrorActionDesc [] onerrorActions) {

        this.name = name;
        this.path = path;
        this.contentType = contentType;
        this.flags = flags;
        this.templateComponent = templateComponent;
        this.globalPElements = globalPElements;
        this.componentRefs = componentRefs;
        this.attributes = attributes;
        this.attributesMap = attributesMap;
        this.inputForm = inputForm;
        this.root = root;
        this.preactions = preactions;
        this.onerrorActions = onerrorActions;
    }
}
