package com.boylesoftware.cb2.presentation.servlet;

import java.lang.reflect.Method;


/**
 * Descriptor of a presentation element. This class is used for both
 * global and component local PEs.
 *
 * @author Lev Himmelfarb
 * @version $Id: PresentationElementDesc.java,v 1.3 2004/10/11 18:17:36 levahim Exp $
 */
final class PresentationElementDesc {

    // PE scopes
    static final int SCOPE_REQUEST = 1;
    static final int SCOPE_SESSION = 2;


    /**
     * Descriptor of a presentation element input parameter.
     */
    static final class ParameterDesc {

        // values for the "from" field
        static final int FROM_REQUEST       = 1;
        static final int FROM_RQPARAM       = 2;
        static final int FROM_RQATTRIBUTE   = 3;
        static final int FROM_SESSION       = 4;
        static final int FROM_CREFATTRIBUTE = 5;
        static final int FROM_ANY           = 6;

        // values for the "emptyString" field
        static final int EMPTYSTRING_NULL     = 1;
        static final int EMPTYSTRING_DISALLOW = 2;
        static final int EMPTYSTRING_ASIS     = 3;


        /**
         * The parameter name.
         */
        final String name;

        /**
         * Corresponding property set method on the presentation
         * element bean.
         */
        final Method setMethod;

        /**
         * Tells if the parameter is required.
         */
        final boolean isRequired;

        /**
         * Tells what to do with an empty string input.
         * One of <code>EMPTYSTRING_xxx</code> constants.
         */
        final int emptyString;

        /**
         * The parameter "from source". One of <code>FROM_xxx</code>
         * constants.
         */
        final int from;


        /**
         * Constructs a parameter descriptor.
         *
         * @param name input parameter name.
         * @param setMethod corresponding "set" method on the PE bean.
         * @param isRequired whether the parameter is required.
         * @param emptyString what to do with if the value is an empty
         * string (one of <code>EMPTYSTRING_xxx</code> constants).
         * @param from where to look for the parameter's value (one of
         * <code>FROM_xxx</code> constants).
         */
        ParameterDesc(String name,
                      Method setMethod,
                      boolean isRequired,
                      int emptyString,
                      int from) {

            this.name = name;
            this.setMethod = setMethod;
            this.isRequired = isRequired;
            this.emptyString = emptyString;
            this.from = from;
        }
    }


    /**
     * The presentation element's name.
     */
    final String name;

    /**
     * The presentation element implementation class. Must implement
     * <code>PresentationElement</code> interface.
     */
    final Class pelementClass;

    /**
     * The PE's scope, one of <code>SCOPE_xxx</code> constants.
     */
    final int scope;

    /**
     * Input form bean reference or <code>null</code>.
     */
    final FormBeanRefDesc inputForm;

    /**
     * List of descriptors of global PEs on which this one depends. This
     * ordered list also contains descriptors for all nested dependencies.
     * It is filtered and contains no duplicates. It is never
     * <code>null</code>, but may be empty.
     */
    final PresentationElementDesc [] dependsOn;

    /**
     * List of parameter descriptors.
     */
    final PresentationElementDesc.ParameterDesc [] parameters;


    /**
     * Constructs a new presentation element descriptor.
     *
     * @param name the PE's name.
     * @param pelementClass the PE implementation class. Must implement
     * <code>PresentationElement</code> interface.
     * @param scope the PE's scope (one of <code>SCOPE_xxx</code> constants).
     * @param inputForm input form bean reference or <code>null</code>.
     * @param dependsOn list of descriptors of global PEs on which this
     * one depends. This list should be ordered and also contain descriptors
     * for all nested dependencies. It should be filtered and contain
     * no duplicates. It is never <code>null</code>, but may be empty.
     * @param parameters descriptors of the PE's input parameters.
     */
    PresentationElementDesc(String name,
                            Class pelementClass,
                            int scope,
                            FormBeanRefDesc inputForm,
                            PresentationElementDesc [] dependsOn,
                            PresentationElementDesc.ParameterDesc [] parameters) {

        this.name = name;
        this.pelementClass = pelementClass;
        this.scope = scope;
        this.inputForm = inputForm;
        this.dependsOn = dependsOn;
        this.parameters = parameters;
    }
}
