package com.boylesoftware.cb2.presentation.servlet;

import javax.servlet.http.HttpServletRequest;

import com.boylesoftware.cb2.ApplicationContext;
import com.boylesoftware.cb2.BLManager;


/**
 * Request processor creates an instance of this class and keeps it as
 * a request attribute during the whole request processing procedure.
 * This class represents a collection of data and references that do not
 * change during single request processing.
 *
 * <p>This class is not used directly in actions and presentation elements,
 * instead <code>ActionContext</code> is used, which provides access to
 * everything an instance of this class can give you access to. However,
 * in certain situations when you don't have an action context (for example
 * in custom JSP tags implementations), you can use services of this class
 * directly given that its instance can be always easily found by calling
 * <code>get(HttpServletRequest request)</code> static method.
 *
 * @author Lev Himmelfarb
 * @version $Id: RequestProcessingContext.java,v 1.2 2003/10/02 21:51:08 levahim Exp $
 */
public class RequestProcessingContext {

    /**
     * Name of request attribute the request processor keeps an instance
     * of request processing context under.
     */
    static final String RPCTX_KEY = "com.boylesoftware.cb2.presentation.servlet.RequestProcessingContext";


    /**
     * Reference to our application context.
     */
    private final ApplicationContext appCtx;

    /**
     * Reference to our BL manager.
     */
    private final BLManager blManager;

    /**
     * The original request received from the client.
     */
    private final String requestURL;

    /**
     * Reference to the corresponding user session object.
     */
    private final UserSession userSession;

    /**
     * Reference to the UI state bean.
     */
    private final UIState uiState;


    /**
     * Creates an instance of the context.
     *
     * @param appCtx reference to our application context.
     * @param blManager reference to our BL manager.
     * @param requestURL the original request received from the client.
     * @param userSession reference to the corresponding user session
     * object.
     * @param uiState reference to the UI state bean.
     */
    RequestProcessingContext(ApplicationContext appCtx,
                             BLManager blManager,
                             String requestURL,
                             UserSession userSession,
                             UIState uiState) {

        this.appCtx = appCtx;
        this.blManager = blManager;
        this.requestURL = requestURL;
        this.userSession = userSession;
        this.uiState = uiState;
    }


    /**
     * Gets the instance of request processing context for the specified
     * request.
     *
     * @param request HTTP request to lookup context for.
     *
     * @return reference to the context instance or <code>null</code>, which
     * should never happen if the request went through the CB2 request
     * processor.
     */
    public static RequestProcessingContext get(HttpServletRequest request) {

        return (RequestProcessingContext)request.getAttribute(RPCTX_KEY);
    }


    /**
     * Gets out application context.
     *
     * @return reference to the application context.
     */
    public ApplicationContext getApplicationContext() { 

        return this.appCtx;
    }

    /**
     * Gets out business level manager.
     *
     * @return reference to the BL manager.
     */
    public BLManager getBLManager() {

        return this.blManager;
    }

    /**
     * Gets the requested by the client URL. This method returns the URL
     * originally requested by the client regardless any forwards and
     * includes happened during the request processing. Note that the
     * returned URL does not include an request parameters submitted using
     * HTTP POST method, only those using the GET method and thus included
     * in the request URL itself.
     *
     * @return the originally requested by the client URL.
     */
    public String getRequestURL() {

        return this.requestURL;
    }

    /**
     * Gets id of the corresponding session in the business level. BL
     * manager maintains its own list of sessions that corresponds to
     * the servlet container sessions and the business level sessions
     * have different ids.
     *
     * @return the BL session id.
     */
    public String getBLSessionId() {

        return this.userSession.getBLSessionId();
    }

    /**
     * Gets the corresponding user session object.
     *
     * @return reference to the user session.
     */
    UserSession getUserSession() {

        return this.userSession;
    }

    /**
     * Gets the UI state bean instance from the session.
     *
     * @return reference to the UI state instance.
     */
    public UIState getUIState() {

        return this.uiState;
    }
}
