package com.boylesoftware.cb2.presentation.servlet;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.math.BigDecimal;
import java.lang.reflect.Array;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;

import org.apache.commons.logging.Log;
import org.apache.struts.action.ActionForward;

import com.boylesoftware.cb2.BLException;


/**
 * This special action processes requests to pages mapped in the
 * <code>pages-config.xml</code> configuration file. Name of the
 * requested page is passed to it in the action mapping parameter. It
 * is not used by the user code.
 *
 * @author Lev Himmelfarb
 * @version $Id: ShowPageAction.java,v 1.17 2004/10/11 18:17:36 levahim Exp $
 */
public final class ShowPageAction
    extends CB2Action {

    /**
     * Empty hash map.
     */
    private static final Map EMPTY_MAP = new HashMap(0);


    /**
     * Name of the request attribute, which stores the component
     * runtime wrappers map. In that map the keys are component
     * reference names (objects of <code>String</code>) and the values
     * are component wrappers (objects of
     * <code>ComponentRuntimeWrapper</code>).
     */
    static final String RQKEY_COMPONENTS =
        "com.boylesoftware.cb2.presentation.servlet.COMPONENTS";

    /**
     * Name of the request attribute, which stores a set of attribute
     * names (objects of <code>String</code>).
     */
    static final String RQKEY_ATTRIBUTES =
        "com.boylesoftware.cb2.presentation.servlet.ATTRIBUTES";

    /**
     * Name of the request attribute, which stores a list of additional
     * component references added to the page via the
     * <code>ActionContext.addComponentRef</code> method. Elements
     * in the list are <code>Page.ComponentRefDescriptor</code> instances.
     */
    static final String RQKEY_ADDL_COMPREFS =
        "com.boylesoftware.cb2.presentation.servlet.ADDL_COMPREFS";


    /**
     * Initializes the presentation elements and puts them to the
     * request.
     *
     * @param actionCtx action context.
     *
     * @throws BLException if an error happened in the business level.
     * @throws ServletException if an error happened in the action.
     */
    protected ActionForward execute(ActionContext actionCtx)
        throws BLException, ServletException {

        // get the log
        Log log = actionCtx.getLog();

        // find the page descriptor
        PageDesc page = (PageDesc)actionCtx.getRequest().getAttribute(CB2RequestProcessor.PAGE);

        // log the page call
        log.debug("called page " + page.name);

        // set the content type if specified
        if(page.contentType != null) {
            actionCtx.getResponse().setContentType(page.contentType);
            actionCtx.getResponse().setHeader("Content-type", page.contentType);
        }

        // get the request object
        HttpServletRequest request = actionCtx.getRequest();

        // setup components and attributes
        try {

            // setup attributes
            Set attrNames = new HashSet(page.attributes.length);
            for(int i = 0; i < page.attributes.length; i++) {
                attrNames.add(page.attributes[i].name);
                request.setAttribute(page.attributes[i].name, page.attributes[i].value);
            }

            // setup global pelements
            for(int i = 0; i < page.globalPElements.length; i++) {
                if(log.isDebugEnabled())
                    log.debug("initializing global PE " + page.globalPElements[i].name +
                              " [" + page.globalPElements[i].pelementClass.getName() + "]");
                PresentationElement pe = this.setupPresentationElement(page.globalPElements[i],
                                                                       request,
                                                                       EMPTY_MAP);
                pe.init(actionCtx);
                request.setAttribute(page.globalPElements[i].name, pe);
            }

            // setup template component pelements and put them into the request
            for(int i = 0; i < page.templateComponent.pelements.length; i++) {
                if(log.isDebugEnabled())
                    log.debug("initializing template component's PE " +
                              page.templateComponent.pelements[i].name + " [" +
                              page.templateComponent.pelements[i].pelementClass.getName() + "]");
                PresentationElement pe = this.setupPresentationElement(page.templateComponent.pelements[i],
                                                                       request,
                                                                       page.attributesMap);
                pe.init(actionCtx);
                request.setAttribute(page.templateComponent.pelements[i].name, pe);
            }

            // setup components
            Map comps = new HashMap(page.componentRefs.length + 10);
            for(int i = 0; i < page.componentRefs.length; i++) {
                PresentationElementDesc [] pes = page.componentRefs[i].component.pelements;
                PresentationElement [] pelements = new PresentationElement[pes.length];
                String [] pelementNames = new String[pes.length];
                for(int j = 0; j < pes.length; j++) {
                    log.debug("initializing PE " + pes[j].name + " [" +
                              pes[j].pelementClass.getName() + "]");
                    pelements[j] = this.setupPresentationElement(pes[j],
                                                                 request,
                                                                 page.componentRefs[i].attributes);
                    pelements[j].init(actionCtx);
                    pelementNames[j] = pes[j].name;
                }
                comps.put(page.componentRefs[i].name,
                          new ComponentRuntimeWrapper(pelements,
                                                      pelementNames,
                                                      page.componentRefs[i].component.src));
            }

            // setup additional components added dynamically
            List addlCompRefs = (List)request.getAttribute(RQKEY_ADDL_COMPREFS);
            if(addlCompRefs != null) {
                for(int i = 0; i < addlCompRefs.size(); i++) {
                    PageDesc.ComponentRefDesc compRefDesc =
                        (PageDesc.ComponentRefDesc)addlCompRefs.get(i);

                    // global PE dependencies
                    for(int j = 0; j < compRefDesc.component.dependsOn.length; j++) {
                        Object o = request.getAttribute(compRefDesc.component.dependsOn[j].name);
                        if(o == null) {
                            if(log.isDebugEnabled())
                                log.debug("initializing dynamically added global PE " + compRefDesc.component.dependsOn[j].name +
                                          " [" + compRefDesc.component.dependsOn[j].pelementClass.getName() + "]");
                            PresentationElement pe = this.setupPresentationElement(compRefDesc.component.dependsOn[j],
                                                                                   request,
                                                                                   EMPTY_MAP);
                            pe.init(actionCtx);
                            request.setAttribute(compRefDesc.component.dependsOn[j].name, pe);
                        } else {
                            if(!(o instanceof PresentationElement))
                                throw new ServletException("Cannot setup global PE \"" + compRefDesc.component.dependsOn[j].name +
                                                           "\" because there is already an attribute in the request under that name.");
                        }
                    }

                    // local PEs
                    PresentationElement [] pelements = new PresentationElement[compRefDesc.component.pelements.length];
                    String [] pelementNames = new String[compRefDesc.component.pelements.length];
                    for(int j = 0; j < compRefDesc.component.pelements.length; j++) {
                        log.debug("initializing dynamically added PE " + compRefDesc.component.pelements[j].name + " [" +
                                  compRefDesc.component.pelements[j].pelementClass.getName() + "]");
                        pelements[j] = this.setupPresentationElement(compRefDesc.component.pelements[j],
                                                                     request,
                                                                     compRefDesc.attributes);
                        pelements[j].init(actionCtx);
                        pelementNames[j] = compRefDesc.component.pelements[j].name;
                    }
                    comps.put(compRefDesc.name,
                              new ComponentRuntimeWrapper(pelements,
                                                          pelementNames,
                                                          compRefDesc.component.src));
                }
            }

            // put the components map and attribute names set to the request
            request.setAttribute(RQKEY_COMPONENTS, comps);
            request.setAttribute(RQKEY_ATTRIBUTES, attrNames);

            // done, forward to the page's JSP
            return page.root;

        } catch(Throwable e) {

            // check onerror handlers
            for(int i = 0; i < page.onerrorActions.length; i++) {
                if(page.onerrorActions[i].exceptionClass.isInstance(e))
                    return page.onerrorActions[i].forward;
            }

            // rethrow the exception
            if(e instanceof BLException)
                throw (BLException)e;
            else if(e instanceof ServletException)
                throw (ServletException)e;
            else if(e instanceof RuntimeException)
                throw (RuntimeException)e;
            else if(e instanceof Error)
                throw (Error)e;
            else // never can happen
                throw new ServletException(e);
        }
    }


    /**
     * Gets an instance of a presentation element bean and sets
     * parameters on it.
     *
     * @param desc the presentation element's descriptor.
     * @param request the HTTP request.
     * @param refAttributes component reference attributes.
     *
     * @return instance of the presentation element with parameters
     * set.
     *
     * @throws ServletException if an error happened setting up the
     * presentation element, for example parameters are invalid.
     */
    private PresentationElement setupPresentationElement(PresentationElementDesc desc,
                                                         HttpServletRequest request,
                                                         Map refAttributes)
        throws ServletException {

        try {

            // get the bean's instance
            PresentationElement pe;
            if(desc.scope == PresentationElementDesc.SCOPE_SESSION) {
                pe = (PresentationElement)request.getSession().getAttribute(desc.name);
                if(pe == null) {
                    pe = (PresentationElement)desc.pelementClass.newInstance();
                    request.getSession().setAttribute(desc.name, pe);
                }
            } else {
                pe = (PresentationElement)request.getAttribute(desc.name);
                if(pe == null)
                    pe = (PresentationElement)desc.pelementClass.newInstance();
            }

            // reset parameters to their default values
            pe.reset();

            // set parameters
            Object [] setParams = new Object[1];
            for(int i = 0; i < desc.parameters.length; i++) {

                // find the parameter's value
                Object value;
                switch(desc.parameters[i].from) {

                case PresentationElementDesc.ParameterDesc.FROM_RQPARAM:
                    value = this.getRequestParameterValue(request, desc.parameters[i].name);
                    break;

                case PresentationElementDesc.ParameterDesc.FROM_RQATTRIBUTE:
                    value = request.getAttribute(desc.parameters[i].name);
                    break;

                case PresentationElementDesc.ParameterDesc.FROM_SESSION:
                    value = request.getSession().getAttribute(desc.parameters[i].name);
                    break;

                case PresentationElementDesc.ParameterDesc.FROM_CREFATTRIBUTE:
                    value = refAttributes.get(desc.parameters[i].name);
                    break;

                case PresentationElementDesc.ParameterDesc.FROM_ANY:
                    if((value = this.getRequestParameterValue(request, desc.parameters[i].name)) != null)
                        break;
                    if((value = request.getAttribute(desc.parameters[i].name)) != null)
                        break;
                    if((value = request.getSession().getAttribute(desc.parameters[i].name)) != null)
                        break;
                    value = refAttributes.get(desc.parameters[i].name);
                    break;

                default: // FROM_REQUEST
                    if((value = this.getRequestParameterValue(request, desc.parameters[i].name)) != null)
                        break;
                    value = request.getAttribute(desc.parameters[i].name);
                    break;
                }

                // check if it is required
                if(value == null) {
                    if(desc.parameters[i].isRequired)
                        throw new NoRequiredParamException("Required parameter " + desc.parameters[i].name +
                                                           " is not specified.");

                    continue;
                }

                // find out the target class
                Class targetClass = desc.parameters[i].setMethod.getParameterTypes()[0];

                // check if the input is an empty string
                if(value instanceof String)
                    value = this.processEmptyString((String)value,
                                                    desc.parameters[i],
                                                    targetClass);
                else if(value instanceof String []) {
                    String [] values = (String [])value;
                    Class myTargetClass = (targetClass.isArray() ?
                                           targetClass.getComponentType() :
                                           targetClass);
                    for(int j = 0; j < values.length; j++)
                        values[j] = this.processEmptyString(values[j],
                                                            desc.parameters[i],
                                                            myTargetClass);
                }

                // convert string value to target class
                if(value != null) {
                    if(value instanceof String) {
                        value = this.stringToObject((String)value, targetClass);
                    } else if((value instanceof String []) &&
                              !targetClass.isAssignableFrom(value.getClass())) {
                        if(!targetClass.isArray())
                            throw new InvalidParamValueException("Parameter \"" +
                                                                 desc.parameters[i].name +
                                                                 "\" value is an array, while the target field is not.");
                        String [] inValues = (String [])value;
                        Class componentType = targetClass.getComponentType();
                        value = Array.newInstance(componentType, inValues.length);
                        for(int j = 0; j < inValues.length; j++)
                            Array.set(value, j, this.stringToObject(inValues[j], componentType));
                    }
                }

                // set the parameter on the presentation element bean
                setParams[0] = value;
                try {
                    desc.parameters[i].setMethod.invoke(pe, setParams);
                } catch(Exception e) {
                    throw new ServletException("Could not set \"" + desc.parameters[i].name +
                                               "\" input parameter on \"" + desc.name +
                                               "\" presentation element bean.", e);
                }
            }

            // done
            return pe;

        } catch(InstantiationException e) {
            throw new ServletException("Could not create an instance of presentation element \"" +
                                       desc.name + "\".", e);
        } catch(IllegalAccessException e) {
            throw new ServletException("Could not access an instance of presentation element \"" +
                                       desc.name + "\".", e);
        }
    }

    /**
     * Gets request parameter value or values if it has multiple values.
     *
     * @param request the HTTP request.
     * @param name the parameter's name.
     *
     * @return the parameter's value, which can be a <code>String</code>
     * object, an array of <code>String</code> objects, or <code>null</code>.
     */
    private Object getRequestParameterValue(HttpServletRequest request,
                                            String name) {

        String [] values = request.getParameterValues(name);
        return (values != null ?
                (values.length == 1 ?
                 (Object)values[0] :
                 (Object)values) :
                null);
    }

    /**
     * Processes a string parameter value according to the
     * <code>emptyString</code> setting.
     *
     * @param value the input value.
     * @param paramDesc the parameter descriptor.
     * @param targetClass the target class.
     *
     * @return processed value, which is the original value or
     * <code>null</code>.
     *
     * @throws InvalidParamValueException the input value cannot be used with
     * the parameter.
     */
    private String processEmptyString(String value,
                                      PresentationElementDesc.ParameterDesc paramDesc,
                                      Class targetClass)
        throws InvalidParamValueException {

        if(value.trim().length() == 0) {
            switch(paramDesc.emptyString) {
                case PresentationElementDesc.ParameterDesc.EMPTYSTRING_DISALLOW:
                    throw new InvalidParamValueException("Value for parameter \"" + paramDesc.name +
                                                         "\" cannot be empty.");
                case PresentationElementDesc.ParameterDesc.EMPTYSTRING_NULL:
                    if(targetClass.isPrimitive())
                        throw new InvalidParamValueException("Cannot assign empty value to parameter \"" +
                                                             paramDesc.name + "\", because it has primitive type.");
                    return null;
            }
        }

        return value;
    }

    /**
     * Converts a string value to an object of the specified class. If the
     * specifed target class is an array, a new one-element arrays is created
     * to hold the converted value.
     *
     * @param strValue string representation of the value to convert.
     * @param targetClass class of the result object. It can be one of
     * Java basic classes, except <code>Void</code>, or
     * <code>String</code>.
     *
     * @return value represented by <code>strValue</code> converted to
     * class <code>targetClass</code>.
     *
     * @throws IllegalArgumentException if the specified target class
     * is not a Java basic class nor a <code>String</code>.
     * @throws InvalidParamValueException if <code>strValue</code> cannot be
     * converted to the specified class.
     */
    private Object stringToObject(String strValue, Class targetClass)
        throws InvalidParamValueException {

        boolean isTargetArray = targetClass.isArray();
        if(isTargetArray)
            targetClass = targetClass.getComponentType();

        Object value;
        try {

            if(targetClass.isAssignableFrom(String.class)) {
                value = strValue;

            } else if(targetClass.isAssignableFrom(Long.class) || targetClass.isAssignableFrom(Long.TYPE)) {
                value = new Long(strValue);

            } else if(targetClass.isAssignableFrom(Integer.class) || targetClass.isAssignableFrom(Integer.TYPE)) {
                value = new Integer(strValue);

            } else if(targetClass.isAssignableFrom(Boolean.class) || targetClass.isAssignableFrom(Boolean.TYPE)) {
                value = new Boolean(strValue);

            } else if(targetClass.isAssignableFrom(Character.class) || targetClass.isAssignableFrom(Character.TYPE)) {
                value = new Character(strValue.charAt(0));

            } else if(targetClass.isAssignableFrom(Double.class) || targetClass.isAssignableFrom(Double.TYPE)) {
                value = new Double(strValue);

            } else if(targetClass.isAssignableFrom(Float.class) || targetClass.isAssignableFrom(Float.TYPE)) {
                value = new Float(strValue);

            } else if(targetClass.isAssignableFrom(Byte.class) || targetClass.isAssignableFrom(Byte.TYPE)) {
                value = new Byte(strValue);

            } else if(targetClass.isAssignableFrom(Short.class) || targetClass.isAssignableFrom(Short.TYPE)) {
                value = new Short(strValue);

            } else if(targetClass.isAssignableFrom(BigDecimal.class)) {
                value = new BigDecimal(strValue);

            } else
                throw new IllegalArgumentException("Cannot convert string to " + targetClass.getName());

        } catch(NumberFormatException e) {
            throw new InvalidParamValueException("Could not convert \"" + strValue +
                                                 "\" to " + targetClass.getName(), e);
        }

        if(isTargetArray) {
            Object a = Array.newInstance(targetClass, 1);
            Array.set(a, 0, value);
            value = a;
        }

        return value;
    }
}
