package com.boylesoftware.cb2.presentation.servlet;

import java.util.Date;
import java.util.LinkedList;
import java.security.Principal;
import javax.servlet.http.HttpSessionBindingListener;
import javax.servlet.http.HttpSessionBindingEvent;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import com.boylesoftware.cb2.util.EventStatistics;
import com.boylesoftware.cb2.util.EventHistory;


/**
 * Represents a session in the <code>userSessions</code> map. The
 * <code>CB2Application</code> class maintains its onw list of all
 * registered sessions and objects of this class represent them.
 *
 * <p>This class also implements the
 * <code>HttpSessionBindingListener</code> interface and the
 * controller binds an instance of <code>UserSession</code> to every
 * new session. Implementing this interface is the only reliable way
 * to determine when the session is about to be destroyed for this or
 * that reason and to perform special operations, like destroying
 * the corresponding business level session, on that event.
 *
 * <p>Note, that this class requires external synchronization when
 * accessing. None of the class' methods is synchronized.
 */
class UserSession
    implements HttpSessionBindingListener, java.io.Serializable {

    /**
     * Descriptor of an error in the <code>lastErrors</code> list.
     */
    static final class ErrorDescriptor
        implements java.io.Serializable {

        /**
         * The error's cause.
         */
        public final Throwable throwable;

        /**
         * The timestamp.
         */
        public final Date ts;


        /**
         * Constructs a new descriptor.
         *
         * @param throwable the error's cause.
         */
        protected ErrorDescriptor(Throwable throwable) {

            this.throwable = throwable;
            this.ts = new Date();
        }
    }


    /**
     * No BL action is required.
     */
    public static final int AUTHACTION_NOP    = 0;

    /**
     * BL login must be performed.
     */
    public static final int AUTHACTION_LOGIN  = 1;

    /**
     * BL logout must be performed.
     */
    public static final int AUTHACTION_LOGOUT = 2;

    /**
     * Number of errors to keep in the <code>lastErrors</code> list.
     */
    private static final int KEEP_ERRORS = 10;


    /**
     * The session id.
     */
    private final String sessionId;

    /**
     * Timestamp when this session was registered.
     */
    private final Date createdOn;

    /**
     * Set to <code>true</code> by the constructor and reset by
     * the <code>valueBound</code> method.
     */
    private boolean isNew;

    /**
     * The corresponding BL session id.
     */
    private String blSessionId;

    /**
     * The authenticated user's principal object or
     * <code>null</code>.
     */
    private Principal user;

    /**
     * Timestamp when this session became authenticated.
     */
    private Date authenticatedOn;

    /**
     * Request statistics for the session.
     */
    private final EventStatistics reqStats;

    /**
     * Number of exceptions happened in actions for this session.
     */
    private int nErrors;

    /**
     * List of last <code>KEEP_ERRORS</code> errors (objects of
     * <code>ErrorDescriptor</code>). The first element in the list
     * is the last error.
     */
    private final LinkedList lastErrors;


    /**
     * Constructs a new descriptor.
     *
     * @param session the corresponding HTTP session.
     */
    public UserSession(HttpSession session) {

        this.sessionId = session.getId();
        this.createdOn = new Date(session.getCreationTime());
        this.reqStats = new EventStatistics(new EventHistory[] {
            new EventHistory(6L * 3600000L, 60)
        });
        this.nErrors = 0;
        this.lastErrors = new LinkedList();

        // it is new until resetIsNew is called
        this.isNew = true;

        // no BL session has been created yet
        this.blSessionId = null;

        // it is not authenticated at the beginning
        this.user = null;
        this.authenticatedOn = null;
    }


    /**
     * Notifies the object that it is being bound to a session and
     * identifies the session. This implementation does nothing.
     *
     * @param event the event object, is ignored.
     */
    public void valueBound(HttpSessionBindingEvent event) {

        // do nothing
    }

    /**
     * Notifies the object that it is being unbound from a session
     * and identifies the session. This implementation calls
     * <code>unregisterSession</code> on the parent controller
     * servlet.
     *
     * @param event the event object.
     */
    public void valueUnbound(HttpSessionBindingEvent event) {

        CB2PlugIn.getApplication(event.getSession().getServletContext()).
            unregisterSession(this.blSessionId);
    }


    /**
     * Resets the <code>isNew</code> flag to <code>false</code>.
     *
     */
    public void resetIsNew() {

        this.isNew = false;
    }

    /**
     * Tells if this session was not bound yet.
     *
     * @return <code>true</code> before the
     * <code>valueBound</code> method is called.
     */
    public boolean isNew() {

        return this.isNew;
    }

    /**
     * Sets the corresponding BL session id, which is initially
     * <code>null</code>.
     *
     * @param blSessionId the BL session id.
     */
    public void setBLSessionId(String blSessionId) {

        this.blSessionId = blSessionId;
    }

    /**
     * Gets the corresponding BL session id.
     *
     * @return the BL session id.
     */
    public String getBLSessionId() {

        return this.blSessionId;
    }

    /**
     * Gets the session id.
     *
     * @return the session id.
     */
    public String getSessionId() {

        return this.sessionId;
    }

    /**
     * Gets the timestamp when this session was registered.
     *
     * @return the timestamp when this session was registered.
     */
    public Date getCreatedOn() {

        return this.createdOn;
    }

    /**
     * Gets the authenticated user's principal object or
     * <code>null</code> if the session is not in the
     * "authenticated" state.
     *
     * @return the authenticated user's principal object or
     * <code>null</code>.
     */
    public Principal getUser() {

        return this.user;
    }

    /**
     * Gets timestamp when this session became authenticated, or
     * <code>null</code> if the session is not in the
     * "authenticated" state.
     *
     * @return timestamp when this session became authenticated or
     * <code>null</code>.
     */
    public Date getAuthenticatedOn() {

        return this.authenticatedOn;
    }


    /**
     * Checks if the "authenticated" status of the session has
     * changed and adjusts internal fields.
     *
     * @param request the current HTTP request.
     *
     * @return one of <code>AUTHACTION_XXX</code> constants
     * telling if a login, logout or nothing has happened.
     */
    public int checkAuthenticationStatus(HttpServletRequest request) {

        Principal newUser = request.getUserPrincipal();
        if(newUser != null) {
            if(!newUser.equals(this.user)) {
                this.user = newUser;
                this.authenticatedOn = new Date();
                return AUTHACTION_LOGIN;
            }
        } else {
            if(this.user != null) {
                this.user = null;
                this.authenticatedOn = null;
                return AUTHACTION_LOGOUT;
            }
        }

        return AUTHACTION_NOP;
    }

    /**
     * Gets request statistics for the session.
     *
     * @return request statistics.
     */
    public EventStatistics getRequestStats() {

        return this.reqStats;
    }

    /**
     * Gets number of exceptions happened in actions for the session.
     *
     * @return number of errors.
     */
    public int getNErrors() {

        return this.nErrors;
    }

    /**
     * Gets the list of last errors' descriptors.
     *
     * @return the list of last errors. The first element in the array
     * is the last happened error. May return an empty array, but never
     * a <code>null</code>.
     */
    public ErrorDescriptor [] getLastErrors() {

        return (ErrorDescriptor [])this.lastErrors.
            toArray(new ErrorDescriptor[this.lastErrors.size()]);
    }

    /**
     * Increments number of errors.
     *
     * @param t the cause of the error.
     */
    public synchronized void incrementNErrors(Throwable t) {

        this.nErrors++;

        if(this.lastErrors.size() >= KEEP_ERRORS)
            this.lastErrors.removeLast();
        this.lastErrors.addFirst(new ErrorDescriptor(t.getCause() == null ? t : t.getCause()));
    }
}
