package com.boylesoftware.cb2.util;


/**
 * This class provides service for synchronizing access to a guarded
 * object from multiple concurrent threads. Each thread may be either
 * a reader or a writer. When a reader has started working with the
 * target object by calling <code>enterShared</code> method any
 * writer trying to call <code>enterExclusive</code> will wait until
 * all readers free the object by calling <code>leaveShared</code>,
 * while readers continue to be able to register with
 * <code>enterShared</code> calls. When a writer gets access to the
 * object by calling <code>enterExclusive</code> any other reader or
 * writer will wait until the writer frees the object by calling
 * <code>leaveExclusive</code>.
 *
 * <p>If your code can throw an exception it is recommended to have
 * "leave" methods calls in <code>finally</code> blocks.
 *
 * @author Lev Himmelfarb
 * @version $Id: AccessGuard.java,v 1.2 2004/10/11 18:17:36 levahim Exp $
 */
public final class AccessGuard {

    /**
     * Number of currently registered readers.
     */
    private int nReaders;

    /**
     * True if the target object is locked by a writer.
     */
    private boolean isWriting;

    /**
     * the synchronization object.
     */
    private final Object sync;


    /**
     * Constructs a guard. You need one instance of the guard per each
     * target object instance.
     */
    public AccessGuard() {

        this.nReaders = 0;
        this.isWriting = false;
        this.sync = new Object();
    }


    /**
     * Enters shared access section.
     */
    public void enterShared() {

        synchronized(this.sync) {

            while(this.isWriting)
                try {
                    this.sync.wait();
                } catch(InterruptedException e) {
                    // do nothing
                }

            this.nReaders++;
        }
    }

    /**
     * Same as <code>enterShared</code>.
     *
     * @deprecated use <code>enterShared</code> instead.
     */
    public void startReading() {

        this.enterShared();
    }

    /**
     * Leaves shared access section.
     */
    public void leaveShared() {

        synchronized(this.sync) {

            if(--this.nReaders < 0)
                this.nReaders = 0;

            this.sync.notifyAll();
        }
    }

    /**
     * Same as <code>leaveShared</code>.
     *
     * @deprecated use <code>leaveShared</code> instead.
     */
    public void endReading() {

        this.leaveShared();
    }

    /**
     * Enters exclusive access section.
     */
    public void enterExclusive() {

        synchronized(this.sync) {

            while((this.nReaders > 0) || this.isWriting)
                try {
                    this.sync.wait();
                } catch(InterruptedException e) {
                    // do nothing
                }

            this.isWriting = true;
        }
    }

    /**
     * Same as <code>enterExclusive</code>.
     *
     * @deprecated use <code>enterExclusive</code> instead.
     */
    public void startWriting() {

        this.enterExclusive();
    }

    /**
     * Leaves exclusive access section.
     */
    public void leaveExclusive() {

        synchronized(this.sync) {

            this.isWriting = false;

            this.sync.notifyAll();
        }
    }

    /**
     * Same as <code>leaveExclusive</code>.
     *
     * @deprecated use <code>leaveExclusive</code> instead.
     */
    public void endWriting() {

        this.leaveExclusive();
    }
}
