package com.boylesoftware.cb2.util;

import java.util.HashMap;
import java.util.Iterator;
import java.io.Serializable;


/**
 * This class allows to associate values with string prefixes. You add elements
 * as to a map, where keys are strings and values are some objects, later
 * you can do lookups by the longest prefix. Here is an example:
 *
 * <p>You have the folowing elements in your map:
 *
 * <pre>
 * ab     => 1
 * abcd   => 2
 * zyx    => 3
 * zyxabc => 4
 * </pre>
 *
 * <p>Below are results of lookups by various input strings:
 *
 * <pre>
 * abcdefgh => 2
 * abc      => 1
 * a        => null
 * orange   => null
 * zyx      => 3
 * zyx_111  => 3
 * </pre>
 *
 * <p>This implementation is not thread-safe. In multithread environments
 * access an object of this class has to be synchronized externally.
 *
 * @author Lev Himmelfarb
 * @version $Id: BTreeMap.java,v 1.2 2003/09/17 19:16:05 levahim Exp $
 */
public class BTreeMap
    implements Serializable {

    /**
     * An entry.
     */
    protected class Entry
        implements Serializable {

        /**
         * The entrie's value. May be <code>null</code>.
         */
        public Object value;

        /**
         * Collection of entries for the next level. The keys are letters
         * (objects of <code>Character</code>), values are entries (objects
         * of <code>Entry</code>).
         */
        public final HashMap nextLevel;


        /**
         * Constructs an entry.
         */
        public Entry() {

            this.value = null;
            this.nextLevel = new HashMap();
        }
    }


    /**
     * The top level. Keys are objects of <code>Character</code>,
     * values are objects of <code>BTreeMap.Entry</code>.
     */
    protected final HashMap top;


    /**
     * Constructs a new empty map.
     */
    public BTreeMap() {

        this.top = new HashMap();
    }


    /**
     * Adds a new value to the map.
     *
     * @param prefix the key.
     * @param value the value.
     */
    public void put(String prefix, Object value) {

        char [] prefixChars = prefix.toCharArray();
        if(prefixChars.length == 0)
            throw new IllegalArgumentException("The prefix string cannot be empty.");
        HashMap currentLevel = this.top;
        Entry entry = null;
        for(int i = 0; i < prefixChars.length; i++) {
            Character c = new Character(prefixChars[i]);
            entry = (Entry)currentLevel.get(c);
            if(entry == null) {
                entry = new Entry();
                currentLevel.put(c, entry);
            }
            currentLevel = entry.nextLevel;
        }
        entry.value = value;
    }

    /**
     * Looks up the best value for the specified key.
     *
     * @param key the input key to look up a value for.
     *
     * @return value associated with the longest prefix in the map matching
     * the beginning of the specified key. May return <code>null</code>.
     */
    public Object get(String key) {

        Object res = null;

        char [] keyChars = key.toCharArray();
        HashMap currentLevel = this.top;
        for(int i = 0; i < keyChars.length; i++) {
            Character c = new Character(keyChars[i]);
            Entry entry = (Entry)currentLevel.get(c);
            if(entry == null)
                return res;
            if(entry.value != null)
                res = entry.value;
            currentLevel = entry.nextLevel;
        }

        return res;
    }

    /**
     * Clears the map.
     */
    public void clear() {

        this.clear(this.top);
    }

    /**
     * Clears one level.
     *
     * @param level the level.
     */
    protected void clear(HashMap level) {

        Iterator i = level.values().iterator();
        while(i.hasNext()) {
            Entry entry = (Entry)i.next();
            this.clear(entry.nextLevel);
        }
    }
}
