package com.boylesoftware.cb2.util;

import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectStreamException;
import java.io.IOException;


/**
 * Collection of data encoding/decoding utility functions.
 *
 * @author Lev Himmelfarb
 * @version $Id: CodeUtils.java,v 1.2 2004/10/11 18:17:36 levahim Exp $
 */
public class CodeUtils {

    /**
     * Conversion table for <code>bytesToString</code> method.
     */
    private static final char [] VALUE_TO_CHAR = new char[64];

    /**
     * Conversion table for <code>stringToBytes</code> method.
     */
    private static final int [] CHAR_TO_VALUE = new int[256];

    static {
        for(int i = 0; i <= 25; i++) {
            VALUE_TO_CHAR[i] = (char)('A' + i);
            VALUE_TO_CHAR[i + 26] = (char)('a' + i);
        }
        for(int i = 0; i <= 9; i++)
            VALUE_TO_CHAR[i + 52] = (char)('0' + i);
        VALUE_TO_CHAR[62] = '+';
        VALUE_TO_CHAR[63] = '*';
        for(int i = 0; i < 64; i++)
            CHAR_TO_VALUE[VALUE_TO_CHAR[i]] = i;
    }


    /**
     * Converts a byte array to a string of printable characters consisting
     * of symbols A-Z, a-z, 0-9, + and *. Later the bytes can be restored
     * from the string by <code>stringToBytes</code> method.
     *
     * @param input input bytes.
     *
     * @return string representation of the bytes.
     *
     * @see #stringToBytes
     */
    public static String bytesToString(byte [] input) {

        StringBuffer res = new StringBuffer((input.length/3 + 1) * 4);

        int len = (input.length / 3) * 3;
        int combined;
        for(int i = 0; i < len; i += 3) {

            combined = (input[i] & 0xFF) << 16 |
            (input[i + 1] & 0xFF) << 8 |
            (input[i + 2] & 0xFF);

            res.append(VALUE_TO_CHAR[(combined >>> 18) & 0x3F]);
            res.append(VALUE_TO_CHAR[(combined >>> 12) & 0x3F]);
            res.append(VALUE_TO_CHAR[(combined >>> 6) & 0x3F]);
            res.append(VALUE_TO_CHAR[combined & 0x3F]);
        }

        switch(input.length - len) {
        case 1:
            combined = (input[input.length - 1] & 0xFF) << 16;
            res.append(VALUE_TO_CHAR[(combined >>> 18) & 0x3F]);
            res.append(VALUE_TO_CHAR[(combined >>> 12) & 0x3F]);
            break;
        case 2:
            combined = (input[input.length - 2] & 0xFF) << 16 |
                       (input[input.length - 1] & 0xFF) << 8;
            res.append(VALUE_TO_CHAR[(combined >>> 18) & 0x3F]);
            res.append(VALUE_TO_CHAR[(combined >>> 12) & 0x3F]);
            res.append(VALUE_TO_CHAR[(combined >>> 6) & 0x3F]);
        }

        return res.toString();
    }

    /**
     * Restores a byte array from a string created by
     * <code>bytesToString</code> method.
     *
     * @param input the encoded string.
     *
     * @return restored binary data.
     *
     * @see #bytesToString
     */
    public static byte [] stringToBytes(String input) {

        int len = input.length();

        int resLen = (len / 4) * 3;
        switch(len % 4) {
        case 1:
            throw new IllegalArgumentException("The input string was not encoded by the bytesToString method.");
        case 2:
            resLen++;
            break;
        case 3:
            resLen += 2;
        }
        byte [] res = new byte[resLen];

        int curpos = 0;
        int combined = 0;
        int resI = 0;
        for(int i = 0; i < len; i++) {
            combined = (combined << 6) | CHAR_TO_VALUE[input.charAt(i)];
            if(++curpos >= 4) {
                res[resI++] = (byte)((combined >>> 16) & 0xFF);
                res[resI++] = (byte)((combined >>> 8) & 0xFF);
                res[resI++] = (byte)(combined & 0xFF);
                curpos = 0;
            }
        }

        switch(curpos) {
        case 2:
            combined <<= 12;
            res[resI] = (byte)((combined >>> 16) & 0xFF);
            break;
        case 3:
            combined <<= 6;
            res[resI++] = (byte)((combined >>> 16) & 0xFF);
            res[resI] = (byte)((combined >>> 8) & 0xFF);
        }

        return res;
    }

    /**
     * Serializes an object and returns string representation of its data.
     * The result string contains only symbols that may appear in the
     * output of <code>bytesToString</code> method. The object instance
     * can be restored with the <code>stringToObject</code> method.
     *
     * @param o the object to serialize.
     *
     * @return string representation of the serialized object.
     *
     * @throws ObjectStreamException if an error happened serializing the
     * object.
     *
     * @see #stringToObject
     */
    public static String objectToString(Object o)
        throws ObjectStreamException {

        try {
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream(bos);
            oos.writeObject(o);
            oos.flush();

            return bytesToString(bos.toByteArray());

        } catch(ObjectStreamException e) {
            throw e;

        } catch(IOException e) { // should never happen
            throw new RuntimeException("Got error.", e);
        }
    }

    /**
     * Restores an object instance from the string generated by the
     * <code>objectToString</code> method.
     *
     * @param input the encoded string.
     *
     * @return a new instance of the object.
     *
     * @throws ObjectStreamException if an error happened deserializing the
     * object.
     * @throws ClassNotFoundException the object's class not found.
     */
    public static Object stringToObject(String input)
        throws ObjectStreamException, ClassNotFoundException {

        try {

            return new ObjectInputStream(new ByteArrayInputStream(stringToBytes(input))).
                readObject();

        } catch(ObjectStreamException e) {
            throw e;

        } catch(IOException e) { // should never happen
            throw new RuntimeException("Got error.", e);
        }
    }
}
