package com.boylesoftware.cb2.util;


/**
 * Counts events and maintains a statistics for a certain period of time.
 * Canbe used together with <code>StatsGraph</code> for creating nice
 * event statistics images.
 *
 * <p>Note, that all get methods are not synchronized. Usually you call
 * a few get methods from a single <code>synchronized</code> on the
 * <code>EventHistory</code> instance block. The first call should be
 * <code>getSamples</code>, because it updates timestamps:
 *
 * <pre>
 * long totalSpan = eventHistory.getTotalSpan(); // don't need to be synchronized, since it never changes
 * long [] samples;
 * int lastSampleIndex;
 * int nSamples;
 * long lastSampleTs;
 * long maxSample;
 * synchronized(eventHistory) {
 *     samples = eventHistory.getSamples();
 *     lastSampleIndex = eventHistory.getLastSampleIndex();
 *     nSamples = eventHistory.getNSamples();
 *     lastSampleTs = eventHistory.getLastSampleTs();
 *     maxSample = eventHistory.getMaxSample();
 * }
 * </pre>
 *
 * @author Lev Himmelfarb
 * @version $Id: EventHistory.java,v 1.3 2004/10/11 18:17:36 levahim Exp $
 *
 * @see StatsGraph
 */
public class EventHistory
    implements java.io.Serializable {

    /**
     * Containes samples.
     */
    protected final long [] samples;

    /**
     * Number of milliseconds representing the whole timeline.
     */
    protected final long totalSpan;

    /**
     * Number of milliseconds for a single sample data.
     */
    protected final long sampleSpan;

    /**
     * Index of the currently being built sample.
     */
    protected int curSampleInd;

    /**
     * Number of samples in the <code>timeline</code> including the
     * current one (so it never can be zero).
     */
    protected int nSamples;

    /**
     * Timestamp when collecting data for the current sample was started.
     */
    protected long tsLastSampleStart;

    /**
     * Maximum sample value.
     */
    protected long maxSample;


    /**
     * Constructs and initializes a new event history object. The instance
     * starts collecting statistics for the event as soon as it is created.
     *
     * @param totalSpan number of milliseconds in the timeline, meaing that
     * the event history object will keep statistics for events for the
     * last specified number of milliseconds.
     * @param timelineSamples number of samples to keep, so this parameters
     * defines the resolution of the timeline. For example if the timeline
     * is 1 hour (<code>totalSpan = 3600000</code>) and the
     * <code>timelineSamples</code> is set to <code>60</code> then we will
     * have a sample for every minute. When you use <code>EventHistory</code>
     * with <code>StatsGraph</code> class for generating event timeline
     * images, for optimal mode it's better to set <code>timelineSamples</code>
     * to be the same as the width of the picture. This way you will have
     * a sample per each pixel on the horizontal axis of the generated image. 
     */
    public EventHistory(long totalSpan, int timelineSamples) {

        this.samples = new long[timelineSamples];
        this.totalSpan = totalSpan;
        this.sampleSpan = totalSpan/timelineSamples;
        this.curSampleInd = 0;
        this.nSamples = 1;
        this.tsLastSampleStart = System.currentTimeMillis();
        this.maxSample = 0;
    }


    /**
     * Registers an event happened now.
     *
     * @return timestamp of the event.
     */
    public long eventStarted() {

        long ts = System.currentTimeMillis();

        this.eventStarted(ts);

        return ts;
    }

    /**
     * Registers an event happened at the specified time.
     *
     * @param ts timestamp when the event happened.
     */
    public synchronized void eventStarted(long ts) {

        this.shiftPointer(ts);

        if(++this.samples[this.curSampleInd] > this.maxSample)
            this.maxSample = this.samples[this.curSampleInd];
    }

    /**
     * Updates current index and timestamps without registerin the event.
     *
     * @param ts timestamp.
     */
    private void shiftPointer(long ts) {

        boolean reevalMax = false;
        for(int i = (int)((ts - this.tsLastSampleStart)/this.sampleSpan); i > 0; i--) {
            this.curSampleInd = (this.curSampleInd + 1) % this.samples.length;
            if(this.nSamples < this.samples.length)
                this.nSamples++;
            else
                if(this.samples[this.curSampleInd] == this.maxSample)
                    reevalMax = true;
            this.samples[this.curSampleInd] = 0;
            this.tsLastSampleStart += this.sampleSpan;
        }

        if(reevalMax) {
            this.maxSample = this.samples[0];
            for(int i = 1; i < this.samples.length; i++)
                if(this.samples[i] > this.maxSample)
                    this.maxSample = this.samples[i];
        }
    }


    /**
     * Gets the samples.
     *
     * @return array of samples.
     */
    public long [] getSamples() {

        this.shiftPointer(System.currentTimeMillis());

        return this.samples;
    }

    /**
     * Gets the number of samples in the array of samples.
     *
     * @return the number of samples.
     */
    public int getNSamples() {

        return this.nSamples;
    }

    /**
     * Gets index of the last sample in the array of samples.
     *
     * @return the index of the last sample.
     */
    public int getLastSampleIndex() {

        return this.curSampleInd;
    }

    /**
     * Gets number of milliseconds in the whole timeline. This method does
     * not have to be in the <code>synchronized</code> block.
     *
     * @return number of milliseconds in the timeline.
     */
    public long getTotalSpan() {

        return this.totalSpan;
    }

    /**
     * Gets number of milliseconds of one sample. Thie method does not have
     * to be in the <code>synchronized</code> block.
     *
     * @return number of milliseconds per sample.
     */
    public long getSampleSpan() {

        return this.sampleSpan;
    }

    /**
     * Gets timestamp for the last sample.
     *
     * @return timestamp for the last sample.
     */
    public long getLastSampleTs() {

        return this.tsLastSampleStart;
    }

    /**
     * Gets the maximum sample value.
     *
     * @return maximum sample.
     */
    public long getMaxSample() {

        return this.maxSample;
    }
}
