package com.boylesoftware.cb2.util;


/**
 * This utility class can be used for collecting statistics on some
 * reqularly taking place event. Each event, when it happens, is
 * registered with an instance of this class and then various
 * parameters about this event regularity can be obtained.
 *
 * <p>There are two ways of using this class. First is for counting
 * "point" events, and the second is for gathering statistics on
 * events that take some time for processing and you want to have
 * statistics on that time. In the first case you just call
 * <code>eventStarted</code> method for each event. The
 * <code>getProcTimeXxx</code> methods then will always be returning
 * -1. In the second case you call <code>eventStarted</code> method
 * when the event processing is just started and then you call
 * <code>eventFinished</code> when the processing has just
 * finished. In this case the <code>getProcTimeXxx</code> method will
 * be returning valid computed values.
 *
 * <p>This class is thread safe. Basically all methods of this class
 * are instance-synchronized.
 *
 * @author Lev Himmelfarb
 * @version $Id: EventStatistics.java,v 1.5 2004/10/11 18:17:36 levahim Exp $
 */
public class EventStatistics
    implements java.io.Serializable {

    /**
     * Number of times the event has happened.
     */
    private long nTimes = 0;

    /**
     * Average time in milliseconds between events.
     */
    private long distAvg = -1;

    /**
     * Timestamp when the event happened last time.
     */
    private long tsLast = -1;

    /**
     * Minimum amount of time in milliseconds processing of the event
     * took.
     */
    private long procTimeMin = -1;

    /**
     * Average amount of time in milliseconds processing of the event
     * took.
     */
    private long procTimeAvg = -1;

    /**
     * Maximum amount of time in milliseconds processing of the event
     * took.
     */
    private long procTimeMax = -1;

    /**
     * Array of event history timlines.
     */
    private final EventHistory [] timelines;


    /**
     * Constructs an event statistics object.
     */
    public EventStatistics() {

        this(null);
    }

    /**
     * Constructs an event statistics object with event history timelines.
     *
     * @param timelines array of timelines to maintain.
     */
    public EventStatistics(EventHistory [] timelines) {

        this.timelines = timelines;
    }


    /**
     * Registers the fact that the event has happened.
     *
     * @return timestamp when the event happened.
     */
    public synchronized long eventStarted() {

        long ts = System.currentTimeMillis();

        this.nTimes++;

        if(this.tsLast >= 0) {
            long newDist = ts - this.tsLast;
            if(this.distAvg > 0)
                this.distAvg -= (this.distAvg - newDist)/this.nTimes;
            else
                this.distAvg = newDist;
        }

        if(this.timelines != null) {
            for(int i = this.timelines.length - 1; i >= 0; i--) {
                this.timelines[i].eventStarted(ts);
            }
        }

        return (this.tsLast = ts);
    }

    /**
     * Registers the fact that processing of the event has
     * finished. Calling this method is optional and is required only
     * if you need the <code>procTimeXxx</code> fields.
     *
     * @param startedTs value returned by <code>eventStarted</code>
     * method when registering the event's start.
     *
     * @return number of milliseconds the event processing took.
     */
    public synchronized long eventFinished(long startedTs) {

        long procTime = System.currentTimeMillis() - startedTs;
        if(this.nTimes <= 1) {
            this.procTimeMin = this.procTimeAvg = this.procTimeMax = procTime;
        } else {
            if(this.procTimeMin > procTime)
                this.procTimeMin = procTime;
            if(this.procTimeMax < procTime)
                this.procTimeMax = procTime;
            this.procTimeAvg -= (this.procTimeAvg - procTime)/this.nTimes;
        }

        return procTime;
    }


    /**
     * Gets the number of times the event has happened.
     *
     * @return the number of times the event has happened.
     */
    public synchronized long getNTimes() {

        return this.nTimes;
    }

    /**
     * Gets the average time in milliseconds between events.
     *
     * @return the average time in milliseconds between events.
     */
    public synchronized long getDistAvg() {

        return this.getDistAvgUnsync();
    }

    /**
     * Internal non synchronized version of <code>getDistAvg</code>.
     *
     * @return the average event distance.
     */
    private long getDistAvgUnsync() {

        return this.distAvg - (this.distAvg - (System.currentTimeMillis() - this.tsLast))/(this.nTimes + 1);
    }

    /**
     * Gets the timestamp when the event happened last time.
     *
     * @return the timestamp when the event happened last time.
     */
    public synchronized long getTsLast() {

        return this.tsLast;
    }

    /**
     * Gets the minimum amount of time in milliseconds processing of
     * the event took.
     *
     * @return the minimum amount of time in milliseconds processing
     * of the event took.
     */
    public synchronized long getProcTimeMin() {

        return this.procTimeMin;
    }

    /**
     * Gets the average amount of time in milliseconds processing of
     * the event took.
     *
     * @return the average amount of time in milliseconds processing
     * of the event took.
     */
    public synchronized long getProcTimeAvg() {

        return this.procTimeAvg;
    }

    /**
     * Gets the maximum amount of time in milliseconds processing of
     * the event took.
     *
     * @return the maximum amount of time in milliseconds processing
     * of the event took.
     */
    public synchronized long getProcTimeMax() {

        return this.procTimeMax;
    }

    /**
     * Gets the average number of events per minute.
     *
     * @return the average number of events per minute.
     */
    public synchronized long getEventsPerMinuteAvg() {

        long distAvg = this.getDistAvgUnsync();

        return (distAvg > 0 ? 60000 / distAvg : 0);
    }

    /**
     * Gets the event history timelines.
     *
     * @return array of event history timelines or <code>null</code>.
     */
    public synchronized EventHistory [] getTimelines() {

        return this.timelines;
    }
}
