package com.boylesoftware.cb2.util;

import java.security.Principal;
import java.util.Set;


/**
 * Simple implementation of Java's <code>Principal</code> interface,
 * which adds a list of security roles attached to the principal.
 *
 * @author Lev Himmelfarb
 * @version $Id: PrincipalWithRoles.java,v 1.1 2004/04/04 20:13:45 levahim Exp $
 */
public class PrincipalWithRoles
    implements Principal, java.io.Serializable {

    /**
     * The principal's name.
     */
    private final String name;

    /**
     * Colelction of user roles.
     */
    private final Set roles;


    /**
     * Constructs a principal.
     *
     * @param name the principal's unique name.
     * @param roles collection of principal's roles, contains
     * objects of <code>String</code>. Cannot be <code>null</code>.
     */
    public PrincipalWithRoles(String name, Set roles) {

        this.name = name;
        this.roles = roles;
    }


    /**
     * Gets the principal's unique name.
     *
     * @return the principal's name.
     */
    public String getName() {

        return this.name;
    }

    /**
     * Checks if this principal is the same as another.
     * This implementation assumes that principal's are
     * the same if the specified one is of class
     * <code>PrincipalWithRoles</code> or derived and has
     * equal name.
     *
     * @param obj another principal.
     *
     * @return <code>true</code> if principals are the same.
     */
    public boolean equals(Object obj) {

        return ((obj instanceof PrincipalWithRoles) &&
                ((PrincipalWithRoles)obj).getName().equals(this.name));
    }

    /**
     * Returns the object's hash code. This implementation
     * returns the principal name's hash code.
     *
     * @return the object's hash code.
     */
    public int hashCode() {

        return this.name.hashCode();
    }

    /**
     * Gets string representation of this principal. This
     * implementation returns the principal's name.
     *
     * @return string representation of this principal.
     */
    public String toString() {

        return this.name;
    }

    /**
     * Checks if this principal plays the specified role.
     * This implementation always returns <code>true</code> for
     * role "user" and always <code>false</code> for role "nobody".
     * 
     * @param role the role name.
     *
     * @return <code>true</code> if this principal plays the role.
     */
    public boolean isInRole(String role) {

        return "user".equals(role) ||
               (!"nobody".equals(role) && this.roles.contains(role));
    }
}
