package com.boylesoftware.cb2.util;

import java.awt.Color;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.font.TextLayout;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;
import java.util.Map;
import java.util.HashMap;


/**
 * Works with <code>EventHistory</code> objects and generates nice
 * images with event statistics.
 *
 * @author Lev Himmelfarb
 * @version $Id: StatsGraph.java,v 1.3 2004/10/11 18:17:36 levahim Exp $
 */
public class StatsGraph {

    /**
     * Width of the field to the left of the "data" area.
     */
    protected static final int MAIN_MARGIN_LEFT = 40;

    /**
     * Width of the field to the right of the "data" area.
     */
    protected static final int MAIN_MARGIN_RIGHT = 20;

    /**
     * Width of the field at the top of the "data" area.
     */
    protected static final int MAIN_MARGIN_TOP = 20;

    /**
     * Width of the field at the bottom of the "data" area.
     */
    protected static final int MAIN_MARGIN_BOTTOM = 20;

    /**
     * Minimum allowed distance in pixels between two vertical net-lines.
     */
    protected static final int MIN_STEPX = 15;


    // names of decor parameters

    public static final String MAIN_BG_COLOR     = "MAIN_BG_COLOR";
    public static final String DATA_BG_COLOR     = "DATA_BG_COLOR";
    public static final String DATA_BORDER_COLOR = "DATA_BORDER_COLOR";
    public static final String DATA_STROKE       = "DATA_STROKE";
    public static final String DATA_COLOR        = "DATA_COLOR";
    public static final String GRID_STROKE       = "GRID_STROKE";
    public static final String GRID_COLOR        = "GRID_COLOR";
    public static final String SCALE_LABEL_FONT  = "SCALE_LABEL_FONT";
    public static final String SCALE_LABEL_COLOR = "SCALE_LABEL_COLOR";
    public static final String LEVEL_MAX_COLOR   = "LEVEL_MAX_COLOR";
    public static final String LEVEL_MAX_STROKE  = "LEVEL_MAX_STROKE";
    public static final String LEVEL_MIN_COLOR   = "LEVEL_MIN_COLOR";
    public static final String LEVEL_MIN_STROKE  = "LEVEL_MIN_STROKE";
    public static final String LEVEL_AVG_COLOR   = "LEVEL_AVG_COLOR";
    public static final String LEVEL_AVG_STROKE  = "LEVEL_AVG_STROKE";

    /**
     * Default decoration.
     */
    private static final Map DEFAULT_DECOR;
    static {
        DEFAULT_DECOR = new HashMap(15);
        DEFAULT_DECOR.put(MAIN_BG_COLOR,     new Color(200, 200, 200));
        DEFAULT_DECOR.put(DATA_BG_COLOR,     Color.WHITE);
        DEFAULT_DECOR.put(DATA_BORDER_COLOR, Color.BLACK);
        DEFAULT_DECOR.put(DATA_STROKE,       new BasicStroke(1.0f));
        DEFAULT_DECOR.put(DATA_COLOR,        new Color(20, 170, 20));
        DEFAULT_DECOR.put(GRID_STROKE,       new BasicStroke(1.0f,
                                                             BasicStroke.CAP_BUTT,
                                                             BasicStroke.JOIN_BEVEL,
                                                             1.0f,
                                                             new float [] { 2.0f, 2.0f },
                                                             0.0f));
        DEFAULT_DECOR.put(GRID_COLOR,        new Color(120, 120, 120));
        DEFAULT_DECOR.put(SCALE_LABEL_FONT,  new Font("sansserif", Font.PLAIN, 10));
        DEFAULT_DECOR.put(SCALE_LABEL_COLOR, new Color(10, 10, 10));
        DEFAULT_DECOR.put(LEVEL_MAX_COLOR,   new Color(220, 0, 0));
        DEFAULT_DECOR.put(LEVEL_MAX_STROKE,  new BasicStroke(1.0f));
        DEFAULT_DECOR.put(LEVEL_MIN_COLOR,   new Color(0, 0, 220));
        DEFAULT_DECOR.put(LEVEL_MIN_STROKE,  new BasicStroke(1.0f));
        DEFAULT_DECOR.put(LEVEL_AVG_COLOR,   new Color(220, 0, 0));
        DEFAULT_DECOR.put(LEVEL_AVG_STROKE,  new BasicStroke(1.0f,
                                                             BasicStroke.CAP_BUTT,
                                                             BasicStroke.JOIN_BEVEL,
                                                             1.0f,
                                                             new float [] { 2.0f, 2.0f },
                                                             0.0f));
    }


    /**
     * Descriptor of a horizontal scale.
     */
    protected static final class ScaleDescriptor {

        /**
         * Scale step in milliseconds.
         */
        public final long stepMs;

        /**
         * Format for scale labels.
         */
        public final DateFormat dateFormat;

        /**
         * Maximum possible width in pixels of one label.
         */
        public final int labelWidth;


        /**
         * Constructs a scale descriptor.
         *
         * @param stepMs scale step in milliseconds.
         * @param dateFormat format for scale labels.
         * @param labelWidth maximum possible width in pixels of one label.
         */
        public ScaleDescriptor(long stepMs, DateFormat dateFormat, int labelWidth) {

            this.stepMs = stepMs;
            this.dateFormat = dateFormat;
            this.labelWidth = labelWidth;
        }
    }

    /**
     * Scales.
     */
    protected static final ScaleDescriptor [] SCALES = new ScaleDescriptor [] {
        new ScaleDescriptor(10000L,    new SimpleDateFormat("HH:mm:ss"),  50), // 10s
        new ScaleDescriptor(30000L,    new SimpleDateFormat("HH:mm:ss"),  50), // 30s
        new ScaleDescriptor(60000L,    new SimpleDateFormat("HH:mm"),     30), // 1min
        new ScaleDescriptor(300000L,   new SimpleDateFormat("HH:mm"),     30), // 5min
        new ScaleDescriptor(600000L,   new SimpleDateFormat("HH:mm"),     30), // 10min
        new ScaleDescriptor(1800000L,  new SimpleDateFormat("HH:mm"),     30), // 30min
        new ScaleDescriptor(3600000L,  new SimpleDateFormat("HH:mm"),     30), // 1h
        new ScaleDescriptor(10800000L, new SimpleDateFormat("HH:mm"),     30), // 3h
        new ScaleDescriptor(21600000L, new SimpleDateFormat("HH:mm"),     30), // 6h
        new ScaleDescriptor(86400000L, new SimpleDateFormat("M/d"),       30)  // 1d
    };


    /**
     * Calls <code>generateImage</code> with default decoration.
     */
    public static BufferedImage generateImage(EventHistory eventHistory,
                                              int width,
                                              int height) {

        return generateImage(eventHistory, width, height, null);
    }

    /**
     * Generates a graph image from the samples in the specified
     * <code>EventHistory</code> object.
     *
     * @param eventHistory the samples data.
     * @param width width in pixels of the data field of the generated image.
     * The overall width of the generated image will be greater, because
     * we add special fields around the graph itself with various information.
     * @param height height in pixels of the data field of the generated image.
     * The overall height of the generated image will be greater, because
     * we add special fields around the graph itself with various information.
     * @param decor collection of decoration elements that override the
     * default ones. Keys in the map are names of the elements and values
     * are decoration elements.
     *
     * @return the generated image.
     */
    public static BufferedImage generateImage(EventHistory eventHistory,
                                              int width,
                                              int height,
                                              Map decor) {

        // the result image
        BufferedImage res = new BufferedImage(width + MAIN_MARGIN_LEFT + MAIN_MARGIN_RIGHT,
                                              height + MAIN_MARGIN_TOP + MAIN_MARGIN_BOTTOM,
                                              BufferedImage.TYPE_BYTE_INDEXED);

        // get GCs
        Graphics2D gcMain = res.createGraphics();
        Graphics2D gc = (Graphics2D)gcMain.create(MAIN_MARGIN_LEFT,
                                                  MAIN_MARGIN_TOP,
                                                  width,
                                                  height);

        // fill the main area
        gcMain.setColor((Color)getDecorElement(MAIN_BG_COLOR, decor));
        gcMain.fill3DRect(0,
                          0,
                          width + MAIN_MARGIN_LEFT + MAIN_MARGIN_RIGHT,
                          height + MAIN_MARGIN_TOP + MAIN_MARGIN_BOTTOM,
                          true);

        // fill the data area
        gc.setColor((Color)getDecorElement(DATA_BG_COLOR, decor));
        gc.fillRect(0, 0, width, height);

        // get the events data
        long totalSpan = eventHistory.getTotalSpan();
        long [] samples;
        int lastSampleIndex;
        int nSamples;
        long lastSampleTs;
        long maxSample;
        synchronized(eventHistory) {
            samples = eventHistory.getSamples();
            lastSampleIndex = eventHistory.getLastSampleIndex();
            nSamples = eventHistory.getNSamples();
            lastSampleTs = eventHistory.getLastSampleTs();
            maxSample = eventHistory.getMaxSample();
        }

        // calculate statistics
        long minSample;
        long avgSample; 
        minSample = avgSample = samples[lastSampleIndex];
        for(int n = 1; n < nSamples; n++) {
            long sample = samples[(samples.length + (lastSampleIndex - n))%samples.length];
            if(sample < minSample)
                minSample = sample;
            avgSample -= (avgSample - sample)/(n + 1);
        }

        // draw the data
        gc.setColor((Color)getDecorElement(DATA_COLOR, decor));
        gc.setStroke((Stroke)getDecorElement(DATA_STROKE, decor));
        for(int x = width - 1; x >= 0; x--) {
            int i = (samples.length * x)/width;
            if((samples.length - i) > nSamples)
                break;
            i = (i + lastSampleIndex + 1) % samples.length;
            int y = (int)(((height - height/10) * samples[i])/(maxSample != 0 ? maxSample : 1));
            gc.drawLine(x, height - y, x, height);
        }

        // find appropriate scale
        ScaleDescriptor scale = null;
        long minStepMs = (totalSpan * MIN_STEPX)/width;
        for(int i = 0; i < SCALES.length; i++) {
            if(SCALES[i].stepMs >= minStepMs) {
                scale = SCALES[i];
                break;
            }
        }
        if(scale == null)
            throw new RuntimeException("Too big totalSpan.");

        // draw the horizontal scale
        gc.setColor((Color)getDecorElement(GRID_COLOR, decor));
        gc.setStroke((Stroke)getDecorElement(GRID_STROKE, decor));
        double a = (double)width/(double)totalSpan;
        double b = a * (lastSampleTs - totalSpan);
        long tzOffset = TimeZone.getDefault().getOffset(lastSampleTs + scale.stepMs);
        long t = ((lastSampleTs + tzOffset)/scale.stepMs + 1) * scale.stepMs - tzOffset;
        int stepx = (int)(a * scale.stepMs);
        int labelPeriod = 1;
        while(scale.labelWidth > (stepx * labelPeriod))
            labelPeriod *= 2;
        int x;
        int n = 0;
        do {
            x = (int)(a * t - b);
            if((x >= 0) && (x < width)) {
                gc.drawLine(x, height, x, 0);
                gcMain.setColor((Color)getDecorElement(DATA_BORDER_COLOR, decor));
                gcMain.drawLine(MAIN_MARGIN_LEFT + x,
                                MAIN_MARGIN_TOP + height,
                                MAIN_MARGIN_LEFT + x,
                                MAIN_MARGIN_TOP + height + 2);
                if(n++ % labelPeriod == 0) {
                    if(width - x + 1 + MAIN_MARGIN_RIGHT < scale.labelWidth/2) {
                        n--;
                    } else {
                        gcMain.setColor((Color)getDecorElement(SCALE_LABEL_COLOR, decor));
                        TextLayout text = new TextLayout(scale.dateFormat.format(new Date(t)),
                                                         (Font)getDecorElement(SCALE_LABEL_FONT, decor),
                                                         gcMain.getFontRenderContext());
                        text.draw(gcMain,
                                  (float)(MAIN_MARGIN_LEFT + x - text.getBounds().getWidth()/2),
                                  (float)(MAIN_MARGIN_TOP + height + text.getBounds().getHeight() + 4));
                    }
                }
            }
            t -= scale.stepMs;
        } while(x >= 0);

        // draw horizontal marks
        int y;
        TextLayout text;

        // draw the maximum value mark
        if(maxSample != 0) {
            y = height - (int)(((height - height/10) * maxSample)/maxSample);
            gc.setColor((Color)getDecorElement(LEVEL_MAX_COLOR, decor));
            gc.setStroke((Stroke)getDecorElement(LEVEL_MAX_STROKE, decor));
            gc.drawLine(0, y, width, y);
            gcMain.setColor((Color)getDecorElement(DATA_BORDER_COLOR, decor));
            gcMain.drawLine(MAIN_MARGIN_LEFT - 3, MAIN_MARGIN_TOP + y, MAIN_MARGIN_LEFT, MAIN_MARGIN_TOP + y);
            gcMain.setColor((Color)getDecorElement(SCALE_LABEL_COLOR, decor));
            text = new TextLayout(String.valueOf(maxSample),
                                  (Font)getDecorElement(SCALE_LABEL_FONT, decor),
                                  gcMain.getFontRenderContext());
            text.draw(gcMain,
                      (float)(MAIN_MARGIN_LEFT - text.getBounds().getWidth() - 7),
                      (float)(MAIN_MARGIN_TOP + y + text.getBounds().getHeight()/2));
        }

        // draw the minimum value mark
        y = height - (int)(((height - height/10) * minSample)/(maxSample != 0 ? maxSample : 1));
        if(minSample > 0) {
            gc.setColor((Color)getDecorElement(LEVEL_MIN_COLOR, decor));
            gc.setStroke((Stroke)getDecorElement(LEVEL_MIN_STROKE, decor));
            gc.drawLine(0, y, width, y);
        }
        gcMain.setColor((Color)getDecorElement(DATA_BORDER_COLOR, decor));
        gcMain.drawLine(MAIN_MARGIN_LEFT - 3, MAIN_MARGIN_TOP + y, MAIN_MARGIN_LEFT, MAIN_MARGIN_TOP + y);
        gcMain.setColor((Color)getDecorElement(SCALE_LABEL_COLOR, decor));
        text = new TextLayout(String.valueOf(minSample),
                              (Font)getDecorElement(SCALE_LABEL_FONT, decor),
                              gcMain.getFontRenderContext());
        text.draw(gcMain,
                  (float)(MAIN_MARGIN_LEFT - text.getBounds().getWidth() - 7),
                  (float)(MAIN_MARGIN_TOP + y + text.getBounds().getHeight()/2));

        // draw the average value mark
        y = height - (int)(((height - height/10) * avgSample)/(maxSample != 0 ? maxSample : 1));
        gc.setColor((Color)getDecorElement(LEVEL_AVG_COLOR, decor));
        gc.setStroke((Stroke)getDecorElement(LEVEL_AVG_STROKE, decor));
        gc.drawLine(0, y, width, y);
        gcMain.setColor((Color)getDecorElement(DATA_BORDER_COLOR, decor));
        gcMain.drawLine(MAIN_MARGIN_LEFT - 3, MAIN_MARGIN_TOP + y, MAIN_MARGIN_LEFT, MAIN_MARGIN_TOP + y);
        gcMain.setColor((Color)getDecorElement(SCALE_LABEL_COLOR, decor));
        text = new TextLayout(String.valueOf(avgSample),
                              (Font)getDecorElement(SCALE_LABEL_FONT, decor),
                              gcMain.getFontRenderContext());
        text.draw(gcMain,
                  (float)(MAIN_MARGIN_LEFT - text.getBounds().getWidth() - 7),
                  (float)(MAIN_MARGIN_TOP + y + text.getBounds().getHeight()/2));

        // draw a border around the data area
        gcMain.setColor((Color)getDecorElement(DATA_BORDER_COLOR, decor));
        gcMain.drawRect(MAIN_MARGIN_LEFT - 1, MAIN_MARGIN_TOP - 1, width + 1, height + 1);

        // draw the mark of the last sample timestamp
        gcMain.setColor((Color)getDecorElement(DATA_BORDER_COLOR, decor));
        gcMain.drawLine(MAIN_MARGIN_LEFT + width,
                        MAIN_MARGIN_TOP,
                        MAIN_MARGIN_LEFT + width,
                        MAIN_MARGIN_TOP - 10);
        gcMain.drawLine(MAIN_MARGIN_LEFT + width,
                        MAIN_MARGIN_TOP - 10,
                        MAIN_MARGIN_LEFT + width - 8,
                        MAIN_MARGIN_TOP - 10);
        text = new TextLayout((new Date(lastSampleTs)).toString(),
                              (Font)getDecorElement(SCALE_LABEL_FONT, decor),
                              gcMain.getFontRenderContext());
        text.draw(gcMain,
                  (float)(MAIN_MARGIN_LEFT + width - text.getBounds().getWidth() - 12),
                  MAIN_MARGIN_TOP - 5);

        // all done, dispose the GCs
        gc.dispose();
        gcMain.dispose();

        // return the result
        return res;
    }

    /**
     * Calls <code>generateTinyImage</code> with default decoration.
     */
    public static BufferedImage generateTinyImage(EventHistory eventHistory,
                                                  int width,
                                                  int height,
                                                  long maxSample) {

        return generateTinyImage(eventHistory, width, height, maxSample, null);
    }

    /**
     * Analogue of the <code>generateImage</code> method, but the output
     * image contains only the samples data and no scales, fields and
     * other decorations.
     *
     * @param eventHistory the samples data.
     * @param width the output image width.
     * @param height the output image height.
     * @param maxSample assumed value of maximum sample. If this parameter
     * is less than zero the "real" maximum is taken from the event history
     * data.
     * @param decor collection of decoration elements that override the
     * default ones. Keys in the map are names of the elements and values
     * are decoration elements.
     *
     * @return the resulting image.
     */
    public static BufferedImage generateTinyImage(EventHistory eventHistory,
                                                  int width,
                                                  int height,
                                                  long maxSample,
                                                  Map decor) {

        // the result image
        BufferedImage res = new BufferedImage(width + 2, height + 2,
                                              BufferedImage.TYPE_BYTE_INDEXED);

        // get GC
        Graphics2D gcMain = res.createGraphics();
        Graphics2D gc = (Graphics2D)gcMain.create(1, 1, width, height);

        // fill the data area
        gc.setColor((Color)getDecorElement(DATA_BG_COLOR, decor));
        gc.fillRect(0, 0, width, height);

        // draw the border
        gcMain.setColor((Color)getDecorElement(DATA_BORDER_COLOR, decor));
        gcMain.drawRect(0, 0, width + 1, height + 1);

        // get the events data
        long [] samples;
        int lastSampleIndex;
        int nSamples;
        synchronized(eventHistory) {
            samples = eventHistory.getSamples();
            lastSampleIndex = eventHistory.getLastSampleIndex();
            nSamples = eventHistory.getNSamples();
            if(maxSample < 0)
                maxSample = eventHistory.getMaxSample();
        }

        // draw the data
        gc.setColor((Color)getDecorElement(DATA_COLOR, decor));
        gc.setStroke((Stroke)getDecorElement(DATA_STROKE, decor));
        for(int x = width - 1; x >= 0; x--) {
            int i = (samples.length * x)/width;
            if((samples.length - i) > nSamples)
                break;
            i = (i + lastSampleIndex + 1) % samples.length;
            int y = (int)(((height - height/10) * samples[i])/(maxSample != 0 ? maxSample : 1));
            gc.drawLine(x, height - y, x, height);
        }

        // all done, dispose the GC
        gc.dispose();

        // return the result
        return res;
    }

    /**
     * Returns the named decoration element.
     *
     * @param name name of the element requested.
     * @param elements collection of elements, may be <code>null</code>.
     *
     * @return the requested element.
     */
    protected static final Object getDecorElement(String name, Map elements) {

        if(elements != null) {
            Object res = elements.get(name);
            if(res != null)
                return res;
        }

        return DEFAULT_DECOR.get(name);
    }
}
