package com.boylesoftware.cb2.util;

import java.util.Map;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.text.MessageFormat;


/**
 * Collection of useful text and string processing static functions.
 *
 * @author Lev Himmelfarb
 * @version $Id: TextUtils.java,v 1.6 2004/04/04 20:13:45 levahim Exp $
 */
public class TextUtils {

    /**
     * Pattern for macros used in <code>replaceMacros</code> function.
     */
    private static final Pattern PARAM_PATTERN =
        Pattern.compile("\\$\\{([^{}]+)\\}");

    // formats for time intervals
    private static final MessageFormat TIME_INTERVAL_FORMAT_1 =
        new MessageFormat("{0}.{1,number,000}s");
    private static final MessageFormat TIME_INTERVAL_FORMAT_2 =
        new MessageFormat("{0,number,00}:{1,number,00}:{2,number,00}");
    private static final MessageFormat TIME_INTERVAL_FORMAT_3 =
        new MessageFormat("{0}d {1,number,00}:{2,number,00}:{3,number,00}");

    /**
     * Pattern for email addresses.
     */
    private static final Pattern EMAIL_ADDRESS_PATTERN =
        Pattern.compile("^[\\w!#$%&'*+-/=?^`{|}~]+(?:\\.[\\w!#$%&'*+-/=?^`{|}~]+)*@[A-Za-z](?:[A-Za-z\\d-]{0,61}[A-Za-z\\d])?(?:\\.[A-Za-z](?:[A-Za-z\\d-]{0,61}[A-Za-z\\d])?)*$");


    /**
     * Interface used by the <code>replaceMacros</code> methods to get values
     * for macros.
     */
    public static interface ValueGetter {

        /**
         * Gets the value.
         *
         * @param name the macro name.
         *
         * @return value for the specified macro or <code>null</code>.
         */
        public String get(String name);
    }


    /**
     * Standard implementation of <code>ValueGetter</code> interface, which
     * takes values for macros from a map object. The values in the map
     * are automatically converted to strings using <code>toString</code>
     * method.
     */
    public static class MapValueGetter
        implements ValueGetter {

        /**
         * The map with macro values.
         */
        private final Map values;


        /**
         * Constructs a getter instance.
         *
         * @param values the map with macro values by macro names.
         */
        public MapValueGetter(Map values) {

            this.values = values;
        }


        /**
         * Gets the specified macro's value from the map.
         *
         * @param name the macro name.
         *
         * @return the macro value or <code>null</code> if there is no
         * value associated with this name in the map.
         */
        public String get(String name) {

            Object value = this.values.get(name);

            return (value != null ? value.toString() : null);
        }
    }


    /**
     * This method replaces all occurencies of <code>${name}</code> in the
     * input string with respective values provided by the specified
     * value getter. If no value is returned by the getter for the name
     * <code>name</code> then the <code>${name}</code> is replaced with
     * an empty string.
     *
     * @param input the input string containing macros in the form
     * <code>${name}</code>.
     * @param valueGetter implementation of <code>TextUtils.ValueGetter</code>
     * interface, which provides values for macros.
     *
     * @return the input string with macros replaces with the corresponding
     * values.
     */
    public static String replaceMacros(CharSequence input,
                                       ValueGetter valueGetter) {

        return replaceMacros(input, valueGetter, false);
    }

    /**
     * This method replaces all occurencies of <code>${name}</code> in the
     * input string with respective values provided by the specified
     * value getter.
     *
     * @param input the input string containing macros in the form
     * <code>${name}</code>.
     * @param valueGetter implementation of <code>TextUtils.ValueGetter</code>
     * interface, which provides values for macros.
     * @param leaveNotFoundMacros if <code>true</code> and no value was
     * found for the macro, that is the value getter returned
     * <code>null</code>, the macro tag is copied to the output
     * intact, otherwise the tag is replaced with an empty string.
     *
     * @return the input string with macros replaces with the corresponding
     * values.
     */
    public static String replaceMacros(CharSequence input,
                                       ValueGetter valueGetter,
                                       boolean leaveNotFoundMacros) {

        StringBuffer res = new StringBuffer();

        Matcher m = PARAM_PATTERN.matcher(input);
        while(m.find()) {
            String rep = valueGetter.get(m.group(1));
            if(rep == null) {
                if(leaveNotFoundMacros)
                    m.appendReplacement(res, m.group());
            } else
                m.appendReplacement(res, rep);
        }
        m.appendTail(res);

        return res.toString();
    }

    /**
     * Calls <code>arrayToString(src, 0, src.length)</code>.
     *
     * @param src the input array.
     *
     * @return readable representation of the data in the array.
     */
    public static String arrayToString(Object [] src) {

        return (src != null ? arrayToString(src, 0, src.length) : "null");
    }

    /**
     * Returns string representation of data in the specified
     * array of objects. The elements of the specified array
     * may be arrays too, but not of any primitive type. This method
     * is convenient for logging when you want to see contents
     * of an array in a readable format in a log or debug printout. 
     *
     * @param src the input array.
     * @param startIndex index of the first element in the array.
     * @param length number of array elements to process.
     *
     * @return readable representation of the data in the
     * array.
     */
    public static String arrayToString(Object [] src, int startIndex, int length) {

        if(src == null)
            return "null";

        StringBuffer res = new StringBuffer();

        int lastIndex = startIndex + length;
        if(lastIndex > src.length)
            lastIndex = src.length;
        for(int i = startIndex; i < lastIndex; i++) {
            res.append('{');
            if(src[i] == null)
                res.append("null");
            else {
                if(src[i].getClass().isArray()) {
                    res.append(src[i].getClass().getComponentType().getName());
                    res.append("[]/");
                    res.append(arrayToString((Object [])src[i]));
                } else {
                    res.append(src[i].getClass().getName());
                    res.append('/');
                    res.append(src[i].toString());
                }
            }
            res.append('}');
        }

        return res.toString();
    }

    /**
     * Render a readable string representation of a time interval. This
     * method can be used when you want to display a time interval value
     * in the user interface, log, or any other place where a human
     * would see it.
     *
     * <p>Examples of the method's output:
     *
     * <pre>
     * 0.028s
     * 50.315s
     * 00:40:53
     * 15:20:45
     * 21d 15:20:45
     * </pre>
     *
     * @param interval the time interval in milliseconds.
     *
     * @return human-readable representation of the specified time interval.
     */
    public static String timeIntervalToString(long interval) {

        if(interval < 60000L) {
            long sec = interval/1000L;
            long millis = interval%1000L;
            return TIME_INTERVAL_FORMAT_1.format(new Long [] { new Long(sec), new Long(millis) });
        } else if(interval < 24L * 3600000L) {
            long hrs = interval/3600000L;
            interval %= 3600000L;
            long min = interval/60000L;
            interval %= 60000L;
            long sec = interval/1000L;
            return TIME_INTERVAL_FORMAT_2.format(new Long [] { new Long(hrs), new Long(min), new Long(sec) });
        } else {
            long days = interval/(24L * 3600000L);
            interval %= 24L * 3600000L;
            long hrs = interval/3600000L;
            interval %= 3600000L;
            long min = interval/60000L;
            interval %= 60000L;
            long sec = interval/1000L;
            return TIME_INTERVAL_FORMAT_3.format(new Long [] { new Long(days), new Long(hrs), new Long(min), new Long(sec) });
        }
    }

    /**
     * Checks if the specified string is "empty". A string is considered
     * "empty" if it is <code>null</code>, zero length or consists of
     * only whitespace characters.
     *
     * @param s string to check, can be <code>null</code>.
     *
     * @return <code>true</code> if the specified string is empty.
     */
    public static boolean isEmpty(String s) {

        return ((s == null) || (s.trim().length() == 0));
    }

    /**
     * If the input string is not <code>null</code>, not zero length and
     * does not consist of only whitespace characters, this method trims it
     * and returns the result. Otherwise, it returns <code>null</code>.
     *
     * @param s the input string, can be <code>null</code>.
     *
     * @return trimmed input string or <code>null</code> if the input string
     * is "empty" in terms of <code>TextUtils.isEmpty</code> method.
     */
    public static String getNullIfEmpty(String s) {

        if(s == null)
            return null;

        s = s.trim();

        return (s.length() > 0 ? s : null);
    }

    /**
     * Tells if the specified string is valid email address.
     *
     * @param s the string to test, should be trimmed, allowed
     * to be <code>null</code>.
     *
     * @return <code>true</code> if the specified string is not
     * <code>null</code> and denotes a syntactically valid email
     * address.
     */
    public static boolean isValidEmailAddress(String s) {

        return (s != null) && EMAIL_ADDRESS_PATTERN.matcher(s).matches();
    }
}
