package com.boylesoftware.cb2.util;

import java.util.Random;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;


/**
 * Generates sequence of unique ids, which can be used, for example,
 * as session ids.
 *
 * <p>This implementation is thread-safe.
 *
 * @author Lev Himmelfarb
 * @version $Id: UniqueIdGenerator.java,v 1.2 2004/10/11 18:17:36 levahim Exp $
 */
public class UniqueIdGenerator {

    /**
     * Default number of bytes in the generated uds.
     */
    private static final int DEFAULT_NBYTES = 16;


    /**
     * Number of bytes in the generated ids.
     */
    protected final int nBytes;

    /**
     * Random number generator.
     */
    protected final Random random;

    /**
     * Digest used to postprocess new ids.
     */
    protected final MessageDigest digest;


    /**
     * Constructs and initializes a new generator.
     *
     * @param nBytes number of bytes in the ids.
     */
    public UniqueIdGenerator(int nBytes) {

        this.nBytes = nBytes;

        long seed = System.currentTimeMillis();
        char [] entropy = this.toString().toCharArray();
        for(int i = 0; i < entropy.length; i++)
            seed ^= ((byte)entropy[i]) << ((i % 8) * 8);
        this.random = new Random(seed);

        MessageDigest digest = null;
        try {
            digest = MessageDigest.getInstance("MD5");
        } catch(NoSuchAlgorithmException e) { /* ignore */ }
        this.digest = digest;
    }

    /**
     * Constructs and initializes a new generator for 16 bytes
     * in the generated ids.
     */
    public UniqueIdGenerator() {

        this(DEFAULT_NBYTES);
    }


    /**
     * Generates next id.
     *
     * @return the id.
     */
    public String next() {

        byte [] bytes = new byte[this.nBytes];
        synchronized(this) {
            this.random.nextBytes(bytes);
        }
        bytes = this.digest.digest(bytes);

        StringBuffer res = new StringBuffer(bytes.length * 2);
        for(int i = 0; i < bytes.length; i++) {
            byte b1 = (byte)((bytes[i] & 0xF0) >> 4);
            byte b2 = (byte)(bytes[i] & 0x0F);
            if(b1 < 10)
                res.append((char)('0' + b1));
            else
                res.append((char)('A' + b1));
            if(b2 < 10)
                res.append((char)('0' + b2));
            else
                res.append((char)('A' + b2));
        }

        return res.toString();
    }
}
