package com.boylesoftware.cb2.util.pool;


/**
 * A general purpose object pool. This implementation allows dynamic
 * growing of the internal storage of idle objects, but it never
 * shrinks.
 *
 * <p>This implementation is thread-safe and can be used by a number
 * of concurrent threads.
 *
 * @author Lev Himmelfarb
 * @version $Id: AdaptiveObjectPool.java,v 1.6 2004/10/11 18:17:36 levahim Exp $
 */
public class AdaptiveObjectPool {

    ///////////////////////////////////////////////////////////////////////////
    // Internal classes
    ///////////////////////////////////////////////////////////////////////////

    /**
     * The pool cleaning thread.
     */
    static final class PoolCleaner
        implements Runnable {

        /**
         * The pool we are taking care of.
         */
        private final AdaptiveObjectPool pool;

        /**
         * How often in seconds the threads awakes and tries to shrink
         * the pool.
         */
        private final long period;


        /**
         * Constructs the cleaner.
         *
         * @param pool the pool we're taking care of.
         * @param maxIdleTime pool's <code>maxIdleTime</code> parameter
         * in seconds, cannot be zero.
         */
        PoolCleaner(AdaptiveObjectPool pool,
                    long maxIdleTime) {

            this.pool = pool;
            this.period = (maxIdleTime < 1000 ?
                          maxIdleTime :
                          (long)(5.0d * Math.sqrt(maxIdleTime) - 4.0d) * 1000l);
        }


        /**
         * The thread's entry point.
         */
        public void run() {

            while(true) {

                try {
                    Thread.sleep(this.period);
                } catch(InterruptedException e) { /* ignore */ }

                if(this.pool.isDiscarded())
                    break;

                this.pool.shrink();
            }
        }
    }


    /**
     * Objects are stored in the pool wrapped into this class.
     */
    private class PooledObjectWrapper {

        /**
         * The pooled object.
         */
        public Object object;

        /**
         * Timestamp since when the object is in the pool idling.
         */
        public long idlingSince;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Internal pool data
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Factory to be used for new instances creation.
     */
    private final PooledObjectFactory objectFactory;

    /**
     * Minimum number of objects in the pool.
     */
    private final int minObjects;

    /**
     * Maximum time in millis allowed for instances to idle.
     * 0 for no limit.
     */
    private final long maxIdleTime;

    /**
     * The instances storage.
     */
    private PooledObjectWrapper [] pool;

    /**
     * Index of the first idle instance.
     */
    private int head;

    /**
     * Index just before the last idling instance.
     */
    private int tail;

    /**
     * Tells if the pool is discarded at should be destroyed.
     */
    private boolean isDiscarded;


    // pool statistics

    /**
     * Number of the pool storage array increments.
     */
    private int nIncrements;

    /**
     * Number of new object allocations.
     */
    private long nAllocs;

    /**
     * Number of get object requests.
     */
    private long nGets;

    /**
     * Number of objects being used it the moment (retrieved from the
     * pool through the <code>get</code> method).
     */
    private int nUsed;


    ///////////////////////////////////////////////////////////////////////////
    // Constructors
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructs a pool with the specified initial capacity and all other
     * passible parameters. This is the main contructor internally called
     * from all others.
     *
     * <p>Note, that if you specify <code>maxIdleTime</code> parameter, you
     * should call <code>discrad</code> method if you want to destroy
     * the pool, abandoning the reference to the pool used in you application
     * is not sufficient to free the instance's memory and kill the cleaner
     * thread.
     *
     * @param objectFactory the factory to be used for new instances
     * creation.
     * @param poolSize initial capacity of the pool's internal storage.
     * The pool can hold this many idle objects without the internal
     * storage being reallocated, which is a relatively expensive
     * operation.
     * @param minObjects minimum number of instances the pool should
     * maintain. This includes both idling and allocated instances.
     * When the pool initializes it creates this number of instances
     * and prepopulates itself with them.
     * @param maxIdleTime maximum allowed time in seconds for
     * instances to be idle. If an instance was not allocated for this
     * amount of time it is removed from the pool, thus the pool
     * shrinks. If this parameters is <code>0</code> there is no
     * limit and the pool never shrinks.
     */
    public AdaptiveObjectPool(PooledObjectFactory objectFactory,
                              int poolSize,
                              int minObjects,
                              long maxIdleTime) {

        // check parameters
        if(poolSize < 2)
            throw new IllegalArgumentException("Pool's initial capacity must be at least 2.");

        // intialize internal data
        this.objectFactory = objectFactory;
        this.minObjects = minObjects;
        this.maxIdleTime = maxIdleTime * 1000;
        this.pool = new PooledObjectWrapper[poolSize + 1];
        this.head = this.tail = 0;
        this.isDiscarded = false;

        // intialize statistics
        this.nIncrements = 0;
        this.nAllocs = 0;
        this.nGets = 0;
        this.nUsed = 0;

        // create the pool cleaner thread
        if(maxIdleTime > 0) {
            PoolCleaner cleaner = new PoolCleaner(this, maxIdleTime);
            Thread cleanerThread = new Thread(cleaner,
                                              "AdaptiveObjectPoolCleaner-" + cleaner.hashCode());
            cleanerThread.setDaemon(true);
            cleanerThread.start();
        }

        // intialize factory
        this.objectFactory.setPool(this);
    }

    /**
     * Constructs a pool with initial capacity of 10 instances, 0 minimum
     * objects limit and no idle timeout.
     *
     * @param objectFactory the factory to be used for new instances
     * creation.
     */
    public AdaptiveObjectPool(PooledObjectFactory objectFactory) {

        this(objectFactory, 10, 0, 0);
    }

    /**
     * Constructs a pool with the specified intial capacity.
     *
     * @param objectClass class of pooled objects.
     * @param poolSize the maximum number of pooled, idle objects.
     * @param minObjects minimum number of objects in the pool.
     * @param maxIdleTime idle timeout in seconds.
     */
    public AdaptiveObjectPool(Class objectClass,
                              int poolSize,
                              int minObjects,
                              long maxIdleTime) {

        this(new SimplePooledObjectFactory(objectClass),
             poolSize,
             minObjects,
             maxIdleTime);
    }

    /**
     * Constructs a pool with initial capacity of 10 instances, 0 minimum
     * objects limit and no idle timeout.
     *
     * @param objectClass class of pooled objects.
     */
    public AdaptiveObjectPool(Class objectClass) {

        this(objectClass, 10, 0, 0);
    }


    ///////////////////////////////////////////////////////////////////////////
    // Pool interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Returns an object from the pool. The object must be returned
     * back to the pool by calling the <code>returnObject</code>
     * method.
     *
     * @return reference to the object instance.
     */
    public synchronized Object get() {

        Object res;

        if(this.head == this.tail) { // the pool is empty
            res = this.objectFactory.createObject();
            this.nAllocs++;
        } else {
            PooledObjectWrapper w = this.pool[this.head];
            res = w.object;
            w.object = null;
            if(--this.head < 0)
                this.head = this.pool.length - 1;
        }

        this.nGets++;
        this.nUsed++;

        return res;
    }

    /**
     * Returns a borrowed object back to the pool.
     *
     * @param obj reference to the object to return.
     *
     * @throws IllegalArgumentException if <code>obj</code> parameter is
     * <code>null</code>.
     */
    public synchronized void returnObject(Object obj) {

        if(obj == null)
            throw new IllegalArgumentException("The object being returned cannot be null.");

        this.objectFactory.resetObject(obj);

        if(++this.head >= this.pool.length)
            this.head = 0;

        if(this.head == this.tail) { // the pool is full
            double l = Math.log(this.pool.length);
            PooledObjectWrapper [] newPool =
                new PooledObjectWrapper[this.pool.length + (int)(l * l) * 2 + 1];
            int midIndex = 0;
            if(++this.tail < this.pool.length) {
                midIndex = this.pool.length - this.tail;
                System.arraycopy(this.pool, this.tail, newPool, 1, midIndex);
            }
            this.tail = 0;
            if(this.head > 0) {
                System.arraycopy(this.pool, 0, newPool, midIndex + 1, this.head);
            }
            this.head = this.pool.length;
            this.pool = newPool;
            this.nIncrements++;
        }

        PooledObjectWrapper w = this.pool[this.head];
        if(w == null)
            this.pool[this.head] = w = new PooledObjectWrapper();

        w.object = obj;
        w.idlingSince = System.currentTimeMillis();

        this.nUsed--;
    }

    /**
     * Destroys the pool. If you use <code>maxIdleTime</code> pool parameter
     * the system creates a special daemon thread, which periodically checks
     * the pool and shrinks it if needed. The thread is not terminated
     * automatically after you abandon the pool instance in your application,
     * because it has an internal reference to the pool instance. Therefore,
     * if you want to destroy a pool in your application you should
     * explitiately call this <code>discard</code> method and then the
     * instance will be freed and the cleaner thread killed.
     *
     * <p>There is no need to call this method if you don't use the
     * <code>maxIdleTime</code> parameter (that is it is zero).
     */
    public synchronized void discard() {

        this.isDiscarded = true;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Internal pool interface
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Check for instances that's been idling for too long and remove
     * them.
     */
    synchronized void shrink() {

        long currentTime = System.currentTimeMillis();

        while(this.tail != this.head) {
            int i = this.tail + 1;
            if(i >= this.pool.length)
                i = 0;
            PooledObjectWrapper w = this.pool[i];
            if(currentTime - w.idlingSince >= this.maxIdleTime) {
                this.objectFactory.destroyObject(w.object);
                w.object = null;
                this.tail = i;
            } else
                break;
        }
    }

    /**
     * Tells if the pool was discarded.
     *
     * @return <code>true</code> if the pool was discarded.
     */
    synchronized boolean isDiscarded() {

        return this.isDiscarded;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Pool statistics
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Returns the number of idle objects in the pool.
     *
     * @return the number of idle objects.
     */
    public synchronized int getNIdles() {

        int res = this.head - this.tail;
        if(res < 0)
            res = this.pool.length - res;

        return res;
    }

    /**
     * Returns the number of objects currently being assigned through
     * the <code>get</code> method.
     *
     * @return the number of used objects.
     */
    public synchronized int getNUsed() {

        return this.nUsed;
    }

    /**
     * Returns number of idle objects this pool can hold without
     * reallocating the pool's internal storage of idle objects.
     *
     * @return the pool's internal storage size.
     */
    public synchronized int getPoolSize() {

        return this.pool.length;
    }

    /**
     * The hit rate (the relation of the number of new object
     * allocations to the total number of pooled object requests) in
     * 0.01%.
     *
     * @return the pool's hit rate.
     */
    public synchronized int getHitRate() {

        return 10000 - (int)((this.nAllocs * 10000L)/this.nGets);
    }

    /**
     * Returns the number of times the pool's internal storage of idle
     * objects had to be increased because it could not hold the
     * required number of idle objects.
     *
     * @return the number of pool internal storage increments.
     */
    public synchronized int getNIncrements() {

        return this.nIncrements;
    }
}
