package com.boylesoftware.cb2.util.pool;


/**
 * Factories implementing this interface are used by object pool
 * implementations for instantiating the objects.
 *
 * <p>Rarely you need to implement this interface directly, usually
 * you extend <code>PooledObjectFactoryAdapter</code> class.
 *
 * @author Lev Himmelfarb
 * @version $Id: PooledObjectFactory.java,v 1.2 2003/07/23 22:18:08 levahim Exp $
 *
 * @see AdaptiveObjectPool
 * @see PooledObjectFactoryAdapter
 */
public interface PooledObjectFactory {

    /**
     * Instantiates an object for a pool.
     *
     * @return the new object instance.
     */
    public Object createObject();

    /**
     * Resets an object after it had been used. The pool calls this
     * method on the factory and passes an object to it each time the
     * object is returned to the pool.
     *
     * <p>Note, this method is called from a synchronized block, so
     * it should be short.
     *
     * @param o the object being returned to the pool after use.
     */
    public void resetObject(Object o);

    /**
     * Destroyes the object before it is removed from the pool. The
     * pool calls this method when it removes an object, usually because
     * of idle timeout.
     *
     * <p>Note, this method is called from a synchronized block, so
     * it should be short.
     *
     * @param o the object being removed from the pool.
     */
    public void destroyObject(Object o);

    /**
     * This method is called when the factory instance is used when
     * creating a new pool. The <code>AdaptiveObjectPool</code> contructor
     * calls the method.
     *
     * @param pool reference to the pool instance, which is going to
     * use this factory.
     */
    public void setPool(AdaptiveObjectPool pool);
}
