package com.boylesoftware.cb2.util.pool;


/**
 * Abstract parent class for threads to be used with
 * <code>PooledThreadFactory</code> class. The thread sleeps most
 * of the time and calling <code>doExecute</code> method from another
 * thread triggers the user-implemented <code>execute</code> method
 * call from within the pooled thread. After the <code>execute</code>
 * method returns the thread returns itself to the pool and goes into
 * the sleep mode again.
 *
 * <p>Note, you should not return the thread back to the pool, it
 * does it automatically after the user-defined <code>execute</code>
 * method returns.
 *
 * <p>The two thread control methods, <code>doExecute</code> and
 * <code>kill</code> should only be called on instances just taken
 * from the pool. That is you call <code>get</code> method on the pool
 * to get an instance and then call one of the control methods once.
 *
 * @author Lev Himmelfarb
 * @version $Id: PooledThread.java,v 1.5 2004/10/11 18:17:36 levahim Exp $
 */
public abstract class PooledThread
    implements Runnable {

    /**
     * Reference to the pool this thread belongs to.
     */
    private AdaptiveObjectPool pool;

    /**
     * The trigger object.
     */
    private final Object trigger = new Object();

    /**
     * The kill flag.
     */
    private boolean killMe = false;

    /**
     * Used to avoid "thread race condition".
     */
    private boolean readyToExecute = false;


    /**
     * This method is automatically called from the factory, it sets the
     * internal reference to the pool so this implementation can return
     * objects after execution.
     *
     * @param pool reference to the pool this thread belongs to.
     */
    public final void setPool(AdaptiveObjectPool pool) {

        this.pool = pool;
    }

    /**
     * The thread enrty point. This method contains the main
     * sleep/execute loop.
     */
    public final void run() {

        while(!this.killMe) {

            // sleep and wait for the trigger
            synchronized(this.trigger) {
                this.readyToExecute = true;

                try {
                    this.trigger.wait();
                } catch(InterruptedException e) { /* ignore it */ }

                // don't do anything if the thread is requested to die
                if(this.killMe)
                    break;
            }

            // call the user-defined execute
            this.execute();

            // return the thread back to the pool
            this.pool.returnObject(this);
        }
    }

    /**
     * Triggers the user-defined <code>execute</code> method call. Note,
     * that this method does nothing if the thread is already in the
     * user-defined <code>execute</code>.
     */
    public final void doExecute() {

        boolean cont;
        do {
            synchronized(this.trigger) {
                cont = this.readyToExecute; 
            }
            if(!cont) {
                try {
                    Thread.sleep(100);
                } catch(InterruptedException e) { /* ignore */ }
            }
        } while(!cont);

        synchronized(this.trigger) {
            this.trigger.notify();
        }
    }

    /**
     * Kills the thread. If the thread is executing the user-defined
     * <code>execute</code> method it will not be destroyed until the
     * <code>execute</code> method finishes and returns. Note, that
     * you should call this method only on instances just returned from
     * the pool (the <code>get</code> method on the pool) and never
     * immediately after <code>doExecute</code> call.
     */
    public final void kill() {

        synchronized(this.trigger) {
            this.killMe = true;
            this.trigger.notify();
        }
    }

    /**
     * Define this method and implement the thread's logic in it.
     * This method is called each time the <code>doExecute</code>
     * triggering method is called from another thread.
     */
    protected abstract void execute();
}
