package com.boylesoftware.cb2.util.pool;


/**
 * Parent class for thread pool factories. When you are implementing
 * a thread pool you implement a factory object extending this class
 * and the pooled thread object extending the <code>PooledThread</code>
 * class.
 *
 * @author Lev Himmelfarb
 * @version $Id: PooledThreadFactory.java,v 1.5 2003/08/29 13:08:53 levahim Exp $
 */
public abstract class PooledThreadFactory
    extends PooledObjectFactoryAdapter {

    /**
     * Our thread group;
     */
    private final ThreadGroup threadGroup;


    /**
     * Constructs the pool. It must be called from your pool
     * implementation's constructor.
     *
     * @param threadGroupName name of the thread group all threads in
     * this pool will belong.
     */
    public PooledThreadFactory(String threadGroupName) {

        // create the thread group
        this.threadGroup = new ThreadGroup(threadGroupName);
    }


    /**
     * Creates a new instance of the pooled thread class and
     * starts a thread on it.
     *
     * @return the new thread object instance.
     */
    public final Object createObject() {

        PooledThread pooledThread = this.createPooledThread();
        pooledThread.setPool(this.pool);

        Thread thread = new Thread(this.threadGroup,
                                   pooledThread,
                                   this.threadGroup.getName() + "-Thread-" + pooledThread.hashCode());
        thread.setDaemon(true);
        thread.start();

        return pooledThread;
    }

    /**
     * User-defined method that creates a new instance of the pooled thread
     * object. The imlpementation should not start the thread, just create
     * an instance of the class.
     *
     * @return the new pooled thread instance.
     */
    protected abstract PooledThread createPooledThread();


    /**
     * Kills a thread.
     *
     * @param o the thread to kill, should be instance of
     * <code>PooledThread</code> or derived.
     */
    public void destroyObject(Object o) {

        ((PooledThread)o).kill();
    }
}
