package com.boylesoftware.cb2.util.sql;

import java.sql.*;
import java.util.Calendar;
import java.math.BigDecimal;
import java.io.InputStream;
import java.io.Reader;
import java.net.URL;


/**
 * Wrapper around a JDBC's <code>CallableStatement</code> interface
 * implementation. You can derive your statement wrapper classes from
 * this one and override some of its methods, while all the others
 * will be calling appropriate methods on the wrapped, "real" statement.
 *
 * <p>This class also provides some "hook" methods called in certain
 * situations. Note that if your redefine interface methods in your subclass
 * and you use hooks you will have to call the base method from your
 * redefined one in order to make the hooks invoked. Alternatively you
 * call them yourself directly from the redefined interface method.
 *
 * @author Lev Himmelfarb
 * @version $Id: CallableStatementWrapper.java,v 1.5 2004/10/11 18:17:36 levahim Exp $
 */
public class CallableStatementWrapper
    extends PreparedStatementWrapper
    implements CallableStatement {

    /**
     * Reference to the wrapped statement as a callable statement
     * so we don't have to do class casting all the time.
     */
    protected CallableStatement cstmt;


    /**
     * Creates a new wrapper. The wrapped statement should be set
     * by a <code>setWrappedStatement</code> call before using the
     * instance.
     */
    public CallableStatementWrapper() { /* do nothing */ }

    /**
     * Creates a new wrapper and sets the wrapped statement to the
     * specified one.
     *
     * @param cstmt the callable statement to be wrapped.
     */
    public CallableStatementWrapper(CallableStatement cstmt) {

        this.stmt = this.pstmt = this.cstmt = cstmt;
    }

    /**
     * Sets the wrapped statement.
     *
     * @param stmt statement to wrap, must be a callable statement.
     *
     * @throws IllegalArgumentException if the specified argument is not
     * a prepared statement.
     */
    public void setWrappedStatement(Statement stmt) {

        if(!(stmt instanceof CallableStatement))
            throw new IllegalArgumentException("The argument is not a callable statement.");

        this.stmt = this.pstmt = this.cstmt = (CallableStatement)stmt;
    }

    /**
     * Returns the wrapped callable statement.
     *
     * @return the wrapped callable statement.
     */
    public CallableStatement getWrappedCallableStatement() {

        return this.cstmt;
    }


    ///////////////////////////////////////////////////////////////////////////
    // The interface implementation
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Interface method wrapper.
     */
    public void registerOutParameter(int parameterIndex, int sqlType)
        throws SQLException {

        this.cstmt.registerOutParameter(parameterIndex, sqlType);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void registerOutParameter(int parameterIndex, int sqlType, int scale)
        throws SQLException {

        this.cstmt.registerOutParameter(parameterIndex, sqlType, scale);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public boolean wasNull()
        throws SQLException {

        boolean res = this.cstmt.wasNull();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public String getString(int parameterIndex)
        throws SQLException {

        String res = this.cstmt.getString(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public boolean getBoolean(int parameterIndex)
        throws SQLException {

        boolean res = this.cstmt.getBoolean(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public byte getByte(int parameterIndex)
        throws SQLException {

        byte res = this.cstmt.getByte(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public short getShort(int parameterIndex)
        throws SQLException {

        short res = this.cstmt.getShort(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public int getInt(int parameterIndex)
        throws SQLException {

        int res = this.cstmt.getInt(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public long getLong(int parameterIndex)
        throws SQLException {

        long res = this.cstmt.getLong(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public float getFloat(int parameterIndex)
        throws SQLException {

        float res = this.cstmt.getFloat(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public double getDouble(int parameterIndex)
        throws SQLException {

        double res = this.cstmt.getDouble(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * @deprecated
     */
    public BigDecimal getBigDecimal(int parameterIndex, int scale)
        throws SQLException {

        BigDecimal res = this.cstmt.getBigDecimal(parameterIndex, scale);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public byte [] getBytes(int parameterIndex)
        throws SQLException {

        byte [] res = this.cstmt.getBytes(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Date getDate(int parameterIndex)
        throws SQLException {

        java.sql.Date res = this.cstmt.getDate(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Time getTime(int parameterIndex)
        throws SQLException {

        java.sql.Time res = this.cstmt.getTime(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Timestamp getTimestamp(int parameterIndex) 
        throws SQLException {

        java.sql.Timestamp res = this.cstmt.getTimestamp(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Object getObject(int parameterIndex)
        throws SQLException {

        Object res = this.cstmt.getObject(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public BigDecimal getBigDecimal(int parameterIndex)
        throws SQLException {

        BigDecimal res = this.cstmt.getBigDecimal(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Object getObject(int i, java.util.Map map)
        throws SQLException {

        Object res = this.cstmt.getObject(i, map);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Ref getRef(int i)
        throws SQLException {

        Ref res = this.cstmt.getRef(i);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Blob getBlob(int i)
        throws SQLException {

        Blob res = this.cstmt.getBlob(i);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Clob getClob(int i)
        throws SQLException {

        Clob res = this.cstmt.getClob(i);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Array getArray(int i)
        throws SQLException {

        Array res = this.cstmt.getArray(i);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Date getDate(int parameterIndex, Calendar cal) 
        throws SQLException {

        java.sql.Date res = this.cstmt.getDate(parameterIndex, cal);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Time getTime(int parameterIndex, Calendar cal) 
        throws SQLException {

        java.sql.Time res = this.cstmt.getTime(parameterIndex, cal);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Timestamp getTimestamp(int parameterIndex, Calendar cal) 
        throws SQLException {

        java.sql.Timestamp res = this.cstmt.getTimestamp(parameterIndex, cal);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void registerOutParameter(int paramIndex, int sqlType, String typeName)
        throws SQLException {

        this.cstmt.registerOutParameter(paramIndex, sqlType, typeName);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void registerOutParameter(String parameterName, int sqlType)
        throws SQLException {

        this.cstmt.registerOutParameter(parameterName, sqlType);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void registerOutParameter(String parameterName, int sqlType, int scale)
        throws SQLException {

        this.cstmt.registerOutParameter(parameterName, sqlType, scale);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void registerOutParameter(String parameterName, int sqlType, String typeName)
        throws SQLException {

        this.cstmt.registerOutParameter(parameterName, sqlType, typeName);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public URL getURL(int parameterIndex)
        throws SQLException {

        URL res = this.cstmt.getURL(parameterIndex);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setURL(String parameterName, URL val)
        throws SQLException {

        this.cstmt.setURL(parameterName, val);
        this.postCallHook();
    }
    
    /**
     * Interface method wrapper.
     */
    public void setNull(String parameterName, int sqlType)
        throws SQLException {

        this.setParameterHook(parameterName, null);
        this.cstmt.setNull(parameterName, sqlType);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBoolean(String parameterName, boolean x)
        throws SQLException {

        this.cstmt.setBoolean(parameterName,
                              ((Boolean)this.setParameterHook(parameterName, new Boolean(x))).
                                  booleanValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setByte(String parameterName, byte x)
        throws SQLException {

        this.cstmt.setByte(parameterName,
                           ((Byte)this.setParameterHook(parameterName, new Byte(x))).
                               byteValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setShort(String parameterName, short x)
        throws SQLException {

        this.cstmt.setShort(parameterName,
                            ((Short)this.setParameterHook(parameterName, new Short(x))).
                                shortValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setInt(String parameterName, int x)
        throws SQLException {

        this.cstmt.setInt(parameterName,
                          ((Integer)this.setParameterHook(parameterName, new Integer(x))).
                              intValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setLong(String parameterName, long x)
        throws SQLException {

        this.cstmt.setLong(parameterName,
                           ((Long)this.setParameterHook(parameterName, new Long(x))).
                               longValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setFloat(String parameterName, float x)
        throws SQLException {

        this.cstmt.setFloat(parameterName,
                            ((Float)this.setParameterHook(parameterName, new Float(x))).
                                floatValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setDouble(String parameterName, double x)
        throws SQLException {

        this.cstmt.setDouble(parameterName,
                             ((Double)this.setParameterHook(parameterName, new Double(x))).
                                 doubleValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBigDecimal(String parameterName, BigDecimal x)
        throws SQLException {

        this.cstmt.setBigDecimal(parameterName,
                                 (BigDecimal)this.setParameterHook(parameterName, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setString(String parameterName, String x)
        throws SQLException {

        this.cstmt.setString(parameterName,
                             (String)this.setParameterHook(parameterName, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBytes(String parameterName, byte [] x)
        throws SQLException {

        this.cstmt.setBytes(parameterName,
                            (byte [])this.setParameterHook(parameterName, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setDate(String parameterName, java.sql.Date x)
        throws SQLException {

        this.cstmt.setDate(parameterName,
                           (java.sql.Date)this.setParameterHook(parameterName, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTime(String parameterName, java.sql.Time x) 
        throws SQLException {

        this.cstmt.setTime(parameterName,
                           (java.sql.Time)this.setParameterHook(parameterName, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTimestamp(String parameterName, java.sql.Timestamp x)
        throws SQLException {

        this.cstmt.setTimestamp(parameterName,
                                (java.sql.Timestamp)this.setParameterHook(parameterName, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setAsciiStream(String parameterName, InputStream x, int length)
        throws SQLException {

        this.cstmt.setAsciiStream(parameterName,
                                  (InputStream)this.setParameterHook(parameterName, x),
                                  length);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBinaryStream(String parameterName, InputStream x, int length)
        throws SQLException {

        this.cstmt.setBinaryStream(parameterName,
                                   (InputStream)this.setParameterHook(parameterName, x),
                                   length);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setObject(String parameterName, Object x, int targetSqlType, int scale)
        throws SQLException {

        this.cstmt.setObject(parameterName,
                             this.setParameterHook(parameterName, x),
                             targetSqlType,
                             scale);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setObject(String parameterName, Object x, int targetSqlType) 
        throws SQLException {

        this.cstmt.setObject(parameterName,
                             this.setParameterHook(parameterName, x),
                             targetSqlType);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setObject(String parameterName, Object x)
        throws SQLException {

        this.cstmt.setObject(parameterName,
                             this.setParameterHook(parameterName, x));
        this.postCallHook();
    }
   
    /**
     * Interface method wrapper.
     */
    public void setCharacterStream(String parameterName, Reader reader, int length)
        throws SQLException {

        this.cstmt.setCharacterStream(parameterName,
                                      (Reader)this.setParameterHook(parameterName, reader),
                                      length);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setDate(String parameterName, java.sql.Date x, Calendar cal)
        throws SQLException {

        this.cstmt.setDate(parameterName,
                           (java.sql.Date)this.setParameterHook(parameterName, x),
                           cal);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTime(String parameterName, java.sql.Time x, Calendar cal) 
        throws SQLException {

        this.cstmt.setTime(parameterName,
                           (java.sql.Time)this.setParameterHook(parameterName, x),
                           cal);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTimestamp(String parameterName, java.sql.Timestamp x, Calendar cal)
        throws SQLException {

        this.cstmt.setTimestamp(parameterName,
                                (java.sql.Timestamp)this.setParameterHook(parameterName, x),
                                cal);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setNull(String parameterName, int sqlType, String typeName) 
        throws SQLException {

        this.setParameterHook(parameterName, null);
        this.cstmt.setNull(parameterName, sqlType, typeName);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public String getString(String parameterName)
        throws SQLException {

        String res = this.cstmt.getString(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public boolean getBoolean(String parameterName)
        throws SQLException {

        boolean res = this.cstmt.getBoolean(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public byte getByte(String parameterName)
        throws SQLException {

        byte res = this.cstmt.getByte(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public short getShort(String parameterName)
        throws SQLException {

        short res = this.cstmt.getShort(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public int getInt(String parameterName)
        throws SQLException {

        int res = this.cstmt.getInt(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public long getLong(String parameterName)
        throws SQLException {

        long res = this.cstmt.getLong(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public float getFloat(String parameterName)
        throws SQLException {

        float res = this.cstmt.getFloat(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public double getDouble(String parameterName)
        throws SQLException {

        double res = this.cstmt.getDouble(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public byte [] getBytes(String parameterName)
        throws SQLException {

        byte [] res = this.cstmt.getBytes(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Date getDate(String parameterName)
        throws SQLException {

        java.sql.Date res = this.cstmt.getDate(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Time getTime(String parameterName)
        throws SQLException {

        java.sql.Time res = this.cstmt.getTime(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Timestamp getTimestamp(String parameterName)
        throws SQLException {

        java.sql.Timestamp res = this.cstmt.getTimestamp(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Object getObject(String parameterName)
        throws SQLException {

        Object res = this.cstmt.getObject(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public BigDecimal getBigDecimal(String parameterName)
        throws SQLException {

        BigDecimal res = this.cstmt.getBigDecimal(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Object getObject(String parameterName, java.util.Map map)
        throws SQLException {

        Object res = this.cstmt.getObject(parameterName, map);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Ref getRef(String parameterName)
        throws SQLException {

        Ref res = this.cstmt.getRef(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Blob getBlob(String parameterName)
        throws SQLException {

        Blob res = this.cstmt.getBlob(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Clob getClob(String parameterName)
        throws SQLException {

        Clob res = this.cstmt.getClob(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Array getArray(String parameterName)
        throws SQLException {

        Array res = this.cstmt.getArray(parameterName);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Date getDate(String parameterName, Calendar cal) 
        throws SQLException {

        java.sql.Date res = this.cstmt.getDate(parameterName, cal);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Time getTime(String parameterName, Calendar cal) 
        throws SQLException {

        java.sql.Time res = this.cstmt.getTime(parameterName, cal);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public java.sql.Timestamp getTimestamp(String parameterName, Calendar cal) 
        throws SQLException {

        java.sql.Timestamp res = this.cstmt.getTimestamp(parameterName, cal);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public URL getURL(String parameterName)
        throws SQLException {

        URL res = this.cstmt.getURL(parameterName);
        this.postCallHook();
        return res;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Hooks
    ///////////////////////////////////////////////////////////////////////////

    /**
     * This hook is called by all methods that set named parameters. The
     * default implementation just returns the <code>param</code>.
     *
     * @param name the parameter's name.
     * @param param the param's value, which is properly wrapped if it is
     * a primitive type, <code>null</code> if <code>setNull</code> is called. 
     *
     * @return value, which will be passed down to the underlying statement
     * implementation. Has to be of the proper runtime type.
     *
     * @throws SQLException if an error happens.
     */
    protected Object setParameterHook(String name, Object param)
        throws SQLException {

        return param;
    }
}
