package com.boylesoftware.cb2.util.sql;

import java.sql.*;
import java.util.Map;


/**
 * Wrapper around a JDBC's <code>Connection</code> interface
 * implementation. You can derive your connection wrapper classes from
 * this one and override some of its methods, while all the others
 * will be calling appropriate methods on the wrapped, "real" connection.
 *
 * <p>This class also provides some "hook" methods called in certain
 * situations. Note that if your redefine interface methods in your subclass
 * and you use hooks you will have to call the base method from your
 * redefined one in order to make the hooks invoked. Alternatively you
 * call them yourself directly from the redefined interface method.
 *
 * @author Lev Himmelfarb
 * @version $Id: ConnectionWrapper.java,v 1.4 2004/10/11 18:17:36 levahim Exp $
 */
public class ConnectionWrapper
    implements Connection {

    /**
     * The wrapped connection.
     */
    protected Connection con;


    /**
     * Creates a new wrapper. The wrapped connection should be set
     * by a <code>setWrappedConnection</code> call before using the
     * instance.
     */
    public ConnectionWrapper() { /* do nothing */ }

    /**
     * Creates a new wrapper and sets the wrapped connection to the
     * specified one.
     *
     * @param con the connection to be wrapped.
     */
    public ConnectionWrapper(Connection con) {

        this.con = con;
    }

    /**
     * Sets the wrapped connection.
     *
     * @param con connection to wrap.
     */
    public void setWrappedConnection(Connection con) {

        this.con = con;
    }

    /**
     * Returns the wrapped connection.
     *
     * @return the wrapped connection.
     */
    public Connection getWrappedConnection() {

        return this.con;
    }

    /**
     * Called by the garbage collector. Tries to close the connection.
     *
     * @throws SQLException if the connection could not be closed.
     */
    protected void finalize()
        throws SQLException {

        if((this.con != null) && !this.isClosed())
            this.close();
    }


    ///////////////////////////////////////////////////////////////////////////
    // The interface implementation
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public Statement createStatement()
        throws SQLException {

        Statement res = this.createStatementHook(this.con.createStatement(), null);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public PreparedStatement prepareStatement(String sql)
        throws SQLException {

        PreparedStatement res = (PreparedStatement)this.createStatementHook(this.con.prepareStatement(sql), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public CallableStatement prepareCall(String sql)
        throws SQLException {

        CallableStatement res = (CallableStatement)this.createStatementHook(this.con.prepareCall(sql), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public String nativeSQL(String sql)
        throws SQLException {

        String res = this.con.nativeSQL(sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setAutoCommit(boolean autoCommit)
        throws SQLException {

        this.con.setAutoCommit(autoCommit);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public boolean getAutoCommit()
        throws SQLException {

        boolean res = this.con.getAutoCommit();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void commit()
        throws SQLException {

        this.con.commit();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void rollback()
        throws SQLException {

        this.con.rollback();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void close()
        throws SQLException {

        this.con.close();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public boolean isClosed()
        throws SQLException {

        boolean res = this.con.isClosed();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public DatabaseMetaData getMetaData()
        throws SQLException {

        DatabaseMetaData res = this.con.getMetaData();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setReadOnly(boolean readOnly)
        throws SQLException {

        this.con.setReadOnly(readOnly);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public boolean isReadOnly()
        throws SQLException {

        boolean res = this.con.isReadOnly();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setCatalog(String catalog)
        throws SQLException {

        this.con.setCatalog(catalog);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public String getCatalog()
        throws SQLException {

        String res = this.con.getCatalog();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setTransactionIsolation(int level)
        throws SQLException {

        this.con.setTransactionIsolation(level);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public int getTransactionIsolation()
        throws SQLException {

        int res = this.con.getTransactionIsolation();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public SQLWarning getWarnings()
        throws SQLException {

        SQLWarning res = this.con.getWarnings();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void clearWarnings()
        throws SQLException {

        this.con.clearWarnings();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency)
        throws SQLException {

        Statement res = this.createStatementHook(this.con.createStatement(resultSetType, resultSetConcurrency), null);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency)
        throws SQLException {

        PreparedStatement res = (PreparedStatement)this.createStatementHook(this.con.prepareStatement(sql, resultSetType, resultSetConcurrency), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency)
        throws SQLException {

        CallableStatement res = (CallableStatement)this.createStatementHook(this.con.prepareCall(sql, resultSetType, resultSetConcurrency), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Map getTypeMap()
        throws SQLException {

        Map res = this.con.getTypeMap();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setTypeMap(Map map)
        throws SQLException {

        this.con.setTypeMap(map);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setHoldability(int holdability)
        throws SQLException {

        this.con.setHoldability(holdability);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public int getHoldability()
        throws SQLException {

        int res = this.con.getHoldability();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Savepoint setSavepoint()
        throws SQLException {

        Savepoint res = this.con.setSavepoint();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Savepoint setSavepoint(String name)
        throws SQLException {

        Savepoint res = this.con.setSavepoint(name);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void rollback(Savepoint savepoint)
        throws SQLException {

        this.con.rollback(savepoint);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void releaseSavepoint(Savepoint savepoint)
        throws SQLException {

        this.con.releaseSavepoint(savepoint);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public Statement createStatement(int resultSetType, int resultSetConcurrency, int resultSetHoldability)
        throws SQLException {

        Statement res = this.createStatementHook(this.con.createStatement(resultSetType, resultSetConcurrency, resultSetHoldability), null);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency, int resultSetHoldability)
        throws SQLException {

        PreparedStatement res = (PreparedStatement)this.createStatementHook(this.con.prepareStatement(sql, resultSetType, resultSetConcurrency, resultSetHoldability), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency, int resultSetHoldability)
        throws SQLException {

        CallableStatement res = (CallableStatement)this.createStatementHook(this.con.prepareCall(sql, resultSetType, resultSetConcurrency, resultSetHoldability), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys)
        throws SQLException {

        PreparedStatement res = (PreparedStatement)this.createStatementHook(this.con.prepareStatement(sql, autoGeneratedKeys), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public PreparedStatement prepareStatement(String sql, int [] columnIndexes)
        throws SQLException {

        PreparedStatement res = (PreparedStatement)this.createStatementHook(this.con.prepareStatement(sql, columnIndexes), sql);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>createStatementHook</code> hook.
     */
    public PreparedStatement prepareStatement(String sql, String [] columnNames)
        throws SQLException {

        PreparedStatement res = (PreparedStatement)this.createStatementHook(this.con.prepareStatement(sql, columnNames), sql);
        this.postCallHook();
        return res;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Hooks
    ///////////////////////////////////////////////////////////////////////////

    /**
     * This hook is invoked after every interface method call. It is not
     * called if the underlying implementation method has thrown an exception.
     * The default implementation does nothing.
     *
     * @throws SQLException to indicate an error in the method.
     */
    protected void postCallHook()
        throws SQLException {

        // do nothing
    }

    /**
     * This hook method is called from all the methods that create all kinds
     * of database statements, including prepared and callable statements.
     * The default implementation just returns the passed <code>stmt</code>
     * parameter back, but you can override this method in a subclass.
     *
     * @param stmt statement returned by the underlying wrapped connection.
     * @param sql SQL text passed to a <code>prepareStatement<code> or
     * <code>prepareCall</code> method, <code>null</code> if it was
     * <code>createStatement</code> call.
     *
     * @return object that should be returned to the caller.
     *
     * @throws SQLException if an error happened.
     */
    protected Statement createStatementHook(Statement stmt, String sql)
        throws SQLException {

        return stmt;
    }
}
