package com.boylesoftware.cb2.util.sql;

import java.sql.*;
import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.util.Calendar;
import java.net.URL;


/**
 * Wrapper around a JDBC's <code>PreparedStatement</code> interface
 * implementation. You can derive your statement wrapper classes from
 * this one and override some of its methods, while all the others
 * will be calling appropriate methods on the wrapped, "real" statement.
 *
 * <p>This class also provides some "hook" methods called in certain
 * situations. Note that if your redefine interface methods in your subclass
 * and you use hooks you will have to call the base method from your
 * redefined one in order to make the hooks invoked. Alternatively you
 * call them yourself directly from the redefined interface method.
 *
 * @author Lev Himmelfarb
 * @version $Id: PreparedStatementWrapper.java,v 1.7 2004/10/11 18:17:36 levahim Exp $
 */
public class PreparedStatementWrapper
    extends StatementWrapper
    implements PreparedStatement {

    /**
     * Reference to the wrapped statement as a prepared statement
     * so we don't have to do class casting all the time.
     */
    protected PreparedStatement pstmt;


    /**
     * Creates a new wrapper. The wrapped statement should be set
     * by a <code>setWrappedStatement</code> call before using the
     * instance.
     */
    public PreparedStatementWrapper() { /* do nothing */ }

    /**
     * Creates a new wrapper and sets the wrapped statement to the
     * specified one.
     *
     * @param pstmt the prepared statement to be wrapped.
     */
    public PreparedStatementWrapper(PreparedStatement pstmt) {

        this.stmt = this.pstmt = pstmt;
    }

    /**
     * Sets the wrapped statement.
     *
     * @param stmt statement to wrap, must be a prepared statement.
     *
     * @throws IllegalArgumentException if the specified argument is not
     * a prepared statement.
     */
    public void setWrappedStatement(Statement stmt) {

        if(!(stmt instanceof PreparedStatement))
            throw new IllegalArgumentException("The argument is not a prepared statement.");

        this.stmt = this.pstmt = (PreparedStatement)stmt;
    }

    /**
     * Returns the wrapped prepared statement.
     *
     * @return the wrapped prepared statement.
     */
    public PreparedStatement getWrappedPreparedStatement() {

        return this.pstmt;
    }


    ///////////////////////////////////////////////////////////////////////////
    // The interface implementation
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public ResultSet executeQuery()
        throws SQLException {

        this.preexecuteHook(EXECUTE_TYPE_QUERY, null);
        ResultSet rs = this.pstmt.executeQuery();
        this.postexecuteHook(rs, -1, true);
        this.postCallHook();
        return rs;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public int executeUpdate()
        throws SQLException {

        this.preexecuteHook(EXECUTE_TYPE_UPDATE, null);
        int cnt = this.pstmt.executeUpdate();
        this.postexecuteHook(null, cnt, false);
        this.postCallHook();
        return cnt;
    }

    /**
     * Interface method wrapper.
     */
    public void setNull(int parameterIndex, int sqlType)
        throws SQLException {

        this.setParameterHook(parameterIndex, null);
        this.pstmt.setNull(parameterIndex, sqlType);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBoolean(int parameterIndex, boolean x)
        throws SQLException {

        this.pstmt.setBoolean(parameterIndex,
                              ((Boolean)this.setParameterHook(parameterIndex,
                                                              new Boolean(x))).
                                  booleanValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setByte(int parameterIndex, byte x)
        throws SQLException {

        this.pstmt.setByte(parameterIndex,
                           ((Byte)this.setParameterHook(parameterIndex,
                                                        new Byte(x))).
                               byteValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setShort(int parameterIndex, short x)
        throws SQLException {

        this.pstmt.setShort(parameterIndex,
                            ((Short)this.setParameterHook(parameterIndex,
                                                          new Short(x))).
                                shortValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setInt(int parameterIndex, int x)
        throws SQLException {

        this.pstmt.setInt(parameterIndex,
                          ((Integer)this.setParameterHook(parameterIndex,
                                                          new Integer(x))).
                              intValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setLong(int parameterIndex, long x)
        throws SQLException {

        this.pstmt.setLong(parameterIndex,
                           ((Long)this.setParameterHook(parameterIndex,
                                                        new Long(x))).
                               longValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setFloat(int parameterIndex, float x)
        throws SQLException {

        this.pstmt.setFloat(parameterIndex,
                            ((Float)this.setParameterHook(parameterIndex,
                                                          new Float(x))).
                                floatValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setDouble(int parameterIndex, double x)
        throws SQLException {

        this.pstmt.setDouble(parameterIndex,
                             ((Double)this.setParameterHook(parameterIndex,
                                                            new Double(x))).
                                 doubleValue());
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBigDecimal(int parameterIndex, BigDecimal x)
        throws SQLException {

        this.pstmt.setBigDecimal(parameterIndex,
                                 (BigDecimal)this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setString(int parameterIndex, String x)
        throws SQLException {

        this.pstmt.setString(parameterIndex,
                             (String)this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBytes(int parameterIndex, byte [] x)
        throws SQLException {

        this.pstmt.setBytes(parameterIndex,
                            (byte [])this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setDate(int parameterIndex, java.sql.Date x)
        throws SQLException {

        this.pstmt.setDate(parameterIndex,
                           (java.sql.Date)this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTime(int parameterIndex, java.sql.Time x) 
        throws SQLException {

        this.pstmt.setTime(parameterIndex,
                           (java.sql.Time)this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTimestamp(int parameterIndex, java.sql.Timestamp x)
        throws SQLException {

        this.pstmt.setTimestamp(parameterIndex,
                                (java.sql.Timestamp)this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setAsciiStream(int parameterIndex, InputStream x, int length)
        throws SQLException {

        this.pstmt.setAsciiStream(parameterIndex,
                                  (InputStream)this.setParameterHook(parameterIndex, x),
                                  length);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     *
     * @deprecated
     */
    public void setUnicodeStream(int parameterIndex, InputStream x, int length)
        throws SQLException {

        this.pstmt.setUnicodeStream(parameterIndex,
                                    (InputStream)this.setParameterHook(parameterIndex, x),
                                    length);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBinaryStream(int parameterIndex, InputStream x, int length)
        throws SQLException {

        this.pstmt.setBinaryStream(parameterIndex,
                                   (InputStream)this.setParameterHook(parameterIndex, x),
                                   length);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void clearParameters()
        throws SQLException {

        this.pstmt.clearParameters();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setObject(int parameterIndex, Object x, int targetSqlType, int scale)
        throws SQLException {

        this.pstmt.setObject(parameterIndex,
                             this.setParameterHook(parameterIndex, x),
                             targetSqlType,
                             scale);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setObject(int parameterIndex, Object x, int targetSqlType) 
        throws SQLException {

        this.pstmt.setObject(parameterIndex,
                             this.setParameterHook(parameterIndex, x),
                             targetSqlType);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setObject(int parameterIndex, Object x)
        throws SQLException {

        this.pstmt.setObject(parameterIndex,
                             this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public boolean execute()
        throws SQLException {

        this.preexecuteHook(EXECUTE_TYPE_GENERIC, null);
        boolean isResultSet = this.pstmt.execute();
        this.postexecuteHook(null, -1, isResultSet);
        this.postCallHook();
        return isResultSet;
    }

    /**
     * Interface method wrapper.
     */
    public void addBatch()
        throws SQLException {

        this.pstmt.addBatch();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setCharacterStream(int parameterIndex, Reader reader, int length)
        throws SQLException {

        this.pstmt.setCharacterStream(parameterIndex,
                                      (Reader)this.setParameterHook(parameterIndex, reader),
                                      length);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setRef(int i, Ref x)
        throws SQLException {

        this.pstmt.setRef(i, (Ref)this.setParameterHook(i, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setBlob(int i, Blob x)
        throws SQLException {

        this.pstmt.setBlob(i, (Blob)this.setParameterHook(i, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setClob(int i, Clob x)
        throws SQLException {

        this.pstmt.setClob(i, (Clob)this.setParameterHook(i, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setArray(int i, Array x)
        throws SQLException {

        this.pstmt.setArray(i, (Array)this.setParameterHook(i, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public ResultSetMetaData getMetaData()
        throws SQLException {

        ResultSetMetaData res = this.pstmt.getMetaData();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setDate(int parameterIndex, java.sql.Date x, Calendar cal)
        throws SQLException {

        this.pstmt.setDate(parameterIndex,
                           (java.sql.Date)this.setParameterHook(parameterIndex, x),
                           cal);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTime(int parameterIndex, java.sql.Time x, Calendar cal) 
        throws SQLException {

        this.pstmt.setTime(parameterIndex,
                           (java.sql.Time)this.setParameterHook(parameterIndex, x),
                           cal);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setTimestamp(int parameterIndex, java.sql.Timestamp x, Calendar cal)
        throws SQLException {

        this.pstmt.setTimestamp(parameterIndex,
                                (java.sql.Timestamp)this.setParameterHook(parameterIndex, x),
                                cal);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setNull(int paramIndex, int sqlType, String typeName) 
        throws SQLException {

        this.setParameterHook(paramIndex, null);
        this.pstmt.setNull(paramIndex, sqlType, typeName);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setURL(int parameterIndex, URL x)
        throws SQLException {

        this.pstmt.setURL(parameterIndex,
                          (URL)this.setParameterHook(parameterIndex, x));
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public ParameterMetaData getParameterMetaData()
        throws SQLException {

        ParameterMetaData res = this.pstmt.getParameterMetaData();
        this.postCallHook();
        return res;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Hooks
    ///////////////////////////////////////////////////////////////////////////

    /**
     * This hook is called by all methods that set query parameters. The
     * default implementation just returns the <code>param</code>.
     *
     * @param index index of the parameter.
     * @param param the param's value, which is properly wrapped if it is
     * a primitive type, <code>null</code> if <code>setNull</code> is called. 
     *
     * @return value, which will be passed down to the underlying statement
     * implementation. Has to be of the proper runtime type.
     *
     * @throws SQLException if an error happens.
     */
    protected Object setParameterHook(int index, Object param)
        throws SQLException {

        return param;
    }
}
