package com.boylesoftware.cb2.util.sql;

import java.sql.*;


/**
 * Wrapper around a JDBC's <code>Statement</code> interface
 * implementation. You can derive your statement wrapper classes from
 * this one and override some of its methods, while all the others
 * will be calling appropriate methods on the wrapped, "real" statement.
 *
 * <p>This class also provides some "hook" methods called in certain
 * situations. Note that if your redefine interface methods in your subclass
 * and you use hooks you will have to call the base method from your
 * redefined one in order to make the hooks invoked. Alternatively you
 * call them yourself directly from the redefined interface method.
 *
 * @author Lev Himmelfarb
 * @version $Id: StatementWrapper.java,v 1.4 2003/10/01 22:02:36 levahim Exp $
 */
public class StatementWrapper
    implements Statement {

    /**
     * Indicates <code>executeQuery</code> call.
     */
    public static final int EXECUTE_TYPE_QUERY   = 1;

    /**
     * Indicates <code>executeUpdate</code> call.
     */
    public static final int EXECUTE_TYPE_UPDATE  = 2;

    /**
     * Indicates generic <code>execute</code> call.
     */
    public static final int EXECUTE_TYPE_GENERIC = 3;

    /**
     * Indicates <code>executeBatch</code> call.
     */
    public static final int EXECUTE_TYPE_BATCH   = 4;


    /**
     * Connection, which will be returned by the <code>getConnection</code>
     * method. If not set, then <code>getConnection</code> on the
     * wrapped statement will be used.
     */
    protected Connection con;

    /**
     * The wrapped statement.
     */
    protected Statement stmt;


    /**
     * Creates a new wrapper. The wrapped statement should be set
     * by a <code>setWrappedStatement</code> call before using the
     * instance.
     */
    public StatementWrapper() {

        this.con = null;
        this.stmt = null;
    }

    /**
     * Creates a new wrapper and sets the wrapped statement to the
     * specified one.
     *
     * @param con connection, which will be returned by the
     * <code>getConnection</code> method. If <code>null</code>,
     * then <code>getConnection</code> on the wrapped statement
     * will be used.
     * @param stmt the statement to be wrapped.
     */
    public StatementWrapper(Connection con, Statement stmt) {

        this.con = con;
        this.stmt = stmt;
    }

    /**
     * Sets the connection, which will be returned by the
     * <code>getConnection</code> method. If <code>null</code>,
     * then <code>getConnection</code> on the wrapped statement
     * will be used.
     *
     * @param con the connection to set.
     */
    public void setConnection(Connection con) {

        this.con = con;
    }

    /**
     * Sets the wrapped statement.
     *
     * @param stmt statement to wrap.
     */
    public void setWrappedStatement(Statement stmt) {

        this.stmt = stmt;
    }

    /**
     * Returns the wrapped statement.
     *
     * @return the wrapped statement.
     */
    public Statement getWrappedStatement() {

        return this.stmt;
    }


    ///////////////////////////////////////////////////////////////////////////
    // The interface implementation
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public ResultSet executeQuery(String sql)
        throws SQLException {

        ResultSet rs = this.stmt.executeQuery(this.preexecuteHook(EXECUTE_TYPE_QUERY, sql));
        this.postexecuteHook(rs, -1, true);
        this.postCallHook();
        return rs;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public int executeUpdate(String sql)
        throws SQLException {

        int cnt = this.stmt.executeUpdate(this.preexecuteHook(EXECUTE_TYPE_UPDATE, sql));
        this.postexecuteHook(null, cnt, false);
        this.postCallHook();
        return cnt;
    }

    /**
     * Interface method wrapper.
     */
    public void close()
        throws SQLException {

        this.stmt.close();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public int getMaxFieldSize()
        throws SQLException {

        int res = this.stmt.getMaxFieldSize();
        this.postCallHook();
        return res;
    }
    
    /**
     * Interface method wrapper.
     */
    public void setMaxFieldSize(int max)
        throws SQLException {

        this.stmt.setMaxFieldSize(max);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public int getMaxRows()
        throws SQLException {

        int res = this.stmt.getMaxRows();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setMaxRows(int max)
        throws SQLException {

        this.stmt.setMaxRows(max);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setEscapeProcessing(boolean enable)
        throws SQLException {

        this.stmt.setEscapeProcessing(enable);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public int getQueryTimeout()
        throws SQLException {

        int res = this.stmt.getQueryTimeout();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setQueryTimeout(int seconds)
        throws SQLException {

        this.stmt.setQueryTimeout(seconds);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void cancel()
        throws SQLException {

        this.stmt.cancel();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public SQLWarning getWarnings()
        throws SQLException {

        SQLWarning res = this.stmt.getWarnings();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void clearWarnings()
        throws SQLException {

        this.stmt.clearWarnings();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void setCursorName(String name)
        throws SQLException {

        this.stmt.setCursorName(name);
        this.postCallHook();
    }
    
    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public boolean execute(String sql)
        throws SQLException {

        boolean isResultSet = this.stmt.execute(this.preexecuteHook(EXECUTE_TYPE_GENERIC, sql));
        this.postexecuteHook(null, -1, isResultSet);
        this.postCallHook();
        return isResultSet;
    }
    
    /**
     * Interface method wrapper.
     */
    public ResultSet getResultSet()
        throws SQLException { 

        ResultSet res = this.stmt.getResultSet();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public int getUpdateCount()
        throws SQLException { 

        int res = this.stmt.getUpdateCount();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public boolean getMoreResults()
        throws SQLException { 

        boolean res = this.stmt.getMoreResults();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setFetchDirection(int direction)
        throws SQLException {

        this.stmt.setFetchDirection(direction);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public int getFetchDirection()
        throws SQLException {

        int res = this.stmt.getFetchDirection();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void setFetchSize(int rows)
        throws SQLException {

        this.stmt.setFetchSize(rows);
        this.postCallHook();
    }
  
    /**
     * Interface method wrapper.
     */
    public int getFetchSize()
        throws SQLException {

        int res = this.stmt.getFetchSize();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public int getResultSetConcurrency()
        throws SQLException {

        int res = this.stmt.getResultSetConcurrency();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public int getResultSetType()
        throws SQLException {

        int res = this.stmt.getResultSetType();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public void addBatch(String sql)
        throws SQLException {

        this.stmt.addBatch(sql);
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     */
    public void clearBatch()
        throws SQLException {

        this.stmt.clearBatch();
        this.postCallHook();
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public int [] executeBatch()
        throws SQLException {

        this.preexecuteHook(EXECUTE_TYPE_BATCH, null);
        int [] res = this.stmt.executeBatch();
        this.postexecuteHook(null, -1, false);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public Connection getConnection()
        throws SQLException {

        Connection res = (this.con != null ? this.con : this.stmt.getConnection());
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public boolean getMoreResults(int current)
        throws SQLException {

        boolean res = this.stmt.getMoreResults(current);
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     */
    public ResultSet getGeneratedKeys()
        throws SQLException {

        ResultSet res = this.stmt.getGeneratedKeys();
        this.postCallHook();
        return res;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public int executeUpdate(String sql, int autoGeneratedKeys)
        throws SQLException {

        int cnt = this.stmt.executeUpdate(this.preexecuteHook(EXECUTE_TYPE_UPDATE, sql),
                                          autoGeneratedKeys);
        this.postexecuteHook(null, cnt, false);
        this.postCallHook();
        return cnt;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public int executeUpdate(String sql, int [] columnIndexes)
        throws SQLException {

        int cnt = this.stmt.executeUpdate(this.preexecuteHook(EXECUTE_TYPE_UPDATE, sql),
                                          columnIndexes);
        this.postexecuteHook(null, cnt, false);
        this.postCallHook();
        return cnt;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public int executeUpdate(String sql, String [] columnNames)
        throws SQLException {

        int cnt = this.stmt.executeUpdate(this.preexecuteHook(EXECUTE_TYPE_UPDATE, sql),
                                          columnNames);
        this.postexecuteHook(null, cnt, false);
        this.postCallHook();
        return cnt;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public boolean execute(String sql, int autoGeneratedKeys)
        throws SQLException {

        boolean isResultSet = this.stmt.execute(this.preexecuteHook(EXECUTE_TYPE_GENERIC, sql),
                                                autoGeneratedKeys);
        this.postexecuteHook(null, -1, isResultSet);
        this.postCallHook();
        return isResultSet;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public boolean execute(String sql, int [] columnIndexes)
        throws SQLException {

        boolean isResultSet = this.stmt.execute(this.preexecuteHook(EXECUTE_TYPE_GENERIC, sql),
                                                columnIndexes);
        this.postexecuteHook(null, -1, isResultSet);
        this.postCallHook();
        return isResultSet;
    }

    /**
     * Interface method wrapper.
     *
     * <p>Also calles <code>preexecuteHook</code> and
     * <code>postexecuteHook</code> hooks.
     */
    public boolean execute(String sql, String [] columnNames)
        throws SQLException {

        boolean isResultSet = this.stmt.execute(this.preexecuteHook(EXECUTE_TYPE_GENERIC, sql),
                                                columnNames);
        this.postexecuteHook(null, -1, isResultSet);
        this.postCallHook();
        return isResultSet;
    }

    /**
     * Interface method wrapper.
     */
    public int getResultSetHoldability()
        throws SQLException {

        int res = this.stmt.getResultSetHoldability();
        this.postCallHook();
        return res;
    }


    ///////////////////////////////////////////////////////////////////////////
    // Hooks
    ///////////////////////////////////////////////////////////////////////////

    /**
     * This hook is invoked after every interface method call. It is not
     * called if the underlying implementation method has thrown an exception.
     * The default implementation does nothing.
     *
     * @throws SQLException to indicate an error in the method.
     */
    protected void postCallHook()
        throws SQLException {

        // do nothing
    }

    /**
     * This hook method is called from all the methods that cause the statement
     * execution before the control is actually passed down to the underlying
     * statement. The default implementation just returns the passed
     * <code>sql</code> parameter back, but you can override this method
     * in a subclass.
     *
     * @param executeType one of <code>EXECUTE_TYPE_xxx</code> constants
     * indicating the type of the call.
     * @param sql SQL text passed to the execute method by the caller, or
     * <code>null</code> if not applicable (for example it is a prepared
     * statement)
     *
     * @return SQL text which will be passed to the underlying execute
     * method.
     *
     * @throws SQLException if an error happened.
     */
    protected String preexecuteHook(int executeType, String sql)
        throws SQLException {

        return sql;
    }

    /**
     * This hook method is called from all the methods that cause the statement
     * execution after the actuall method on the underlying statement
     * returns. The default implementation does nothing.
     *
     * <p>Note that this hook is not called if there was an exception in
     * the underlying statement.
     *
     * @param rs result set returned by the underlying statement or
     * <code>null</code> if not applicable.
     * @param cnt update count returned by the underlying statement or
     * <code>-1</code> if not applicable.
     * @param isResultSet <code>true</code> if the first result is a
     * <code>ResultSet</code> object, <code>false</code> if it is an update
     * count or there are no results.
     *
     * @throws SQLException if an error happened.
     */
    protected void postexecuteHook(ResultSet rs, int cnt, boolean isResultSet)
        throws SQLException {

        // do nothing
    }
}
